//==============================================================================
//
//  This file is part of GNSSTk, the ARL:UT GNSS Toolkit.
//
//  The GNSSTk is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 3.0 of the License, or
//  any later version.
//
//  The GNSSTk is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with GNSSTk; if not, write to the Free Software Foundation,
//  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
//
//  This software was developed by Applied Research Laboratories at the
//  University of Texas at Austin.
//  Copyright 2004-2022, The Board of Regents of The University of Texas System
//
//==============================================================================


//==============================================================================
//
//  This software was developed by Applied Research Laboratories at the
//  University of Texas at Austin, under contract to an agency or agencies
//  within the U.S. Department of Defense. The U.S. Government retains all
//  rights to use, duplicate, distribute, disclose, or release this software.
//
//  Pursuant to DoD Directive 523024
//
//  DISTRIBUTION STATEMENT A: This software has been approved for public
//                            release, distribution is unlimited.
//
//==============================================================================

#include <cmath>
#include "CCIR.hpp"
#include "Exception.hpp"
#include "CivilTime.hpp"
#include "GNSSconstants.hpp"
#include "BasicTimeSystemConverter.hpp"
#include "DebugTrace.hpp"
#include "TimeString.hpp"

/// Epsilon for UT hour comparison
constexpr double UTHourEpsilon = 1.0e-20;
/// Epsilon for effective ionisation level comparison
constexpr double AZEpsilon = 1.0e-10;

namespace gnsstk
{
   CCIR ::
   CCIR()
         : cacheMonth(-1),
           cacheHour(-1),
           cacheSunSpots(std::numeric_limits<double>::quiet_NaN()),
           cacheGood(false),
           cacheFourierGood(false)
   {
   }


   void CCIR ::
   interpolate(const CommonTime& when, double effSunSpots)
   {
      validateCache(when, effSunSpots);
      if (!cacheGood)
      {
         interpolateF2(when, effSunSpots);
         interpolateFm3(when, effSunSpots);
         cacheGood = true;
      }
   }


   void CCIR ::
   fourier(const CommonTime& when, double effSunSpots)
   {
      DEBUGTRACE_FUNCTION();
      DEBUGTRACE("when=" << gnsstk::printTime(when, "%4Y/%02m/%02d %02H:%02M:%02S"));
      DEBUGTRACE("effSunSpots=" << effSunSpots);
      gnsstk::BasicTimeSystemConverter btsc;
      CivilTime whenUTC(when);
      if (whenUTC.getTimeSystem() != gnsstk::TimeSystem::UTC)
      {
         GNSSTK_ASSERT(whenUTC.changeTimeSystem(gnsstk::TimeSystem::UTC,&btsc));
      }
         // Compute the interpolated ITU-R coefficients AF2, Am3
         // (cacheF2 and cacheFM3)
      interpolate(whenUTC, effSunSpots);
         // interpolate should have called validateCache so we have a
         // valid cacheFourierGood setting.  validateCache will also
         // have set cacheHour to the value corresponding to when, so
         // we don't need to recompute it.
      if (!cacheFourierGood)
      {
         unsigned idx = 0; // computed cache 1D array index
         double t = (15.0 * cacheHour - 180.0) * PI / 180.0;            //eq.49
            // compute Fourier time series for foF2
         for (unsigned degree = 0; degree < F2MaxDegree; degree++)
         {
            cacheCF2[degree] = cacheF2[degree*F2MaxOrder];              //eq.50
            DEBUGTRACE("i=" << degree);
            DEBUGTRACE("  CF2[i] (init) = " << cacheCF2[degree]);
            for (unsigned order = 0; order < 6; order++)
            {
                  // idx = i,2k = degree,2*order, add 1 because
                  // k=[1,6] while order/index=[0,5]
               idx = degree*F2MaxOrder+(order*2+1);
               DEBUGTRACE("  j=" << order);
               DEBUGTRACE("    term 1=" << (cacheF2[idx] * sin((order+1)*t)));
               DEBUGTRACE("    term 2=" << (cacheF2[idx+1] * cos((order+1)*t)));
               cacheCF2[degree] += cacheF2[idx] * sin((order+1)*t) +
                  cacheF2[idx+1] * cos((order+1)*t);
               DEBUGTRACE("    CF2[i] (final) = " << cacheCF2[degree]);
            }
         }
            // compute Fourier time series for M(3000)F2
         for (unsigned degree = 0; degree < FM3MaxDegree; degree++)
         {
            cacheCM3[degree] = cacheFM3[degree*FM3MaxOrder];            //eq.51
            for (unsigned order = 0; order < 4; order++)
            {
               idx = degree*FM3MaxOrder+(order*2+1);
               cacheCM3[degree] += cacheFM3[idx] * sin((order+1)*t) +
                  cacheFM3[idx+1] * cos((order+1)*t);
            }
         }
         cacheFourierGood = true;
      }
   }


   void CCIR ::
   interpolateF2(const CommonTime& when, double effSunSpots)
   {
                                                                        //eq.44
      double effSunSpotCount = effSunSpots / 100.0;
      for (unsigned degree = 0; degree < F2MaxDegree; degree++)
      {
         for (unsigned order = 0; order < F2MaxOrder; order++)
         {
               // cacheF2 is a vector so we translate 2D addressing to 1D.
               // cacheHour/cacheMonth/cacheSunSpots are assumed to
               // ahve been updated by validateCache prior to calling.
            cacheF2[degree*F2MaxOrder+order] =
               ccirF2(cacheMonth, LowSolarActIdx, degree, order) *
               (1.0 - effSunSpotCount) +
               ccirF2(cacheMonth, HighSolarActIdx, degree, order) *
               effSunSpotCount;
         }
      }
   }


   void CCIR ::
   interpolateFm3(const CommonTime& when, double effSunSpots)
   {
      DEBUGTRACE_FUNCTION();
      DEBUGTRACE("effSunSpots=" << effSunSpots);
                                                                        //eq.46
      double effSunSpotCount = effSunSpots / 100.0;
      for (unsigned degree = 0; degree < FM3MaxDegree; degree++)
      {
         for (unsigned order = 0; order < FM3MaxOrder; order++)
         {
               // cacheFm3 is a vector so we translate 2D addressing to 1D.
               // cacheHour/cacheMonth/cacheSunSpots are assumed to
               // have been updated by validateCache prior to calling.
            cacheFM3[degree*FM3MaxOrder+order] =
               ccirFm3(cacheMonth, LowSolarActIdx, degree, order) *
               (1.0 - effSunSpotCount) +
               ccirFm3(cacheMonth, HighSolarActIdx, degree, order) *
               effSunSpotCount;
            DEBUGTRACE("month=" << cacheMonth << " degree=" << degree
                       << " order=" << order << " idx="
                       << (degree*FM3MaxOrder+order) << " low="
                       << ccirFm3(cacheMonth, LowSolarActIdx, degree, order)
                       << " high="
                       << ccirFm3(cacheMonth, HighSolarActIdx, degree, order)
                       << " cacheFM3=" << cacheFM3[degree*FM3MaxOrder+order]);
         }
      }
   }


   void CCIR ::
   validateCache(const CommonTime& when, double effSunSpots)
   {
      DEBUGTRACE_FUNCTION();
      DEBUGTRACE("when=" << gnsstk::printTime(when, "%4Y/%02m/%02d %02H:%02M:%02S"));
      DEBUGTRACE("effSunSpots=" << effSunSpots);
      GNSSTK_ASSERT(when.getTimeSystem() == TimeSystem::UTC);
      if (cacheMonth == -1)
      {
            // Set the cache vector sizes the first time through.
         cacheF2.resize(F2MaxDegree*F2MaxOrder,0.0);
         cacheFM3.resize(FM3MaxDegree*FM3MaxOrder,0.0);
         cacheCF2.resize(F2MaxDegree,0.0);
         cacheCM3.resize(FM3MaxDegree,0.0);
            // dirty cache I want you
         cacheGood = cacheFourierGood = false;
      }
         // Convert to civil time to get hour and month, which is how
         // the model's cycle works (daily cycles + annual cycles).
      CivilTime civ(when);
         // Compute fractional hour of day.
      double utHour = civ.getUTHour();
      DEBUGTRACE("cacheMonth=" << cacheMonth << "  civ.month=" << civ.month);
      DEBUGTRACE("cacheHour=" << cacheHour << "  utHour=" << utHour);
      DEBUGTRACE("cacheSunSpots=" << cacheSunSpots << "  effSunSpots=" << effSunSpots);
         // check to see if our cached interpolated coefficients are still valid
      if ((cacheMonth != civ.month) ||
          (fabs(cacheHour - utHour) > UTHourEpsilon) ||
          (fabs(cacheSunSpots - effSunSpots) > AZEpsilon))
      {
            // dirty cache I need you, ohh
         cacheGood = cacheFourierGood = false;
      }
         // I'm not keen on updating the cache ID before the data, but
         // you need to pick an order and this helps the down-stream
         // interpolation.
      cacheMonth = civ.month;
      cacheHour = utHour;
      cacheSunSpots = effSunSpots;
      DEBUGTRACE("cacheGood=" << cacheGood);
      DEBUGTRACE("cacheFourierGood=" << cacheFourierGood);
   }


   double CCIR ::
   ccirF2(unsigned month, int cond, int deg, int ord)
   {
      GNSSTK_ASSERT(month >= 1 && month <= 12);
      switch (month)
      {
         case  1: return ccir11F2(cond, deg, ord);
         case  2: return ccir12F2(cond, deg, ord);
         case  3: return ccir13F2(cond, deg, ord);
         case  4: return ccir14F2(cond, deg, ord);
         case  5: return ccir15F2(cond, deg, ord);
         case  6: return ccir16F2(cond, deg, ord);
         case  7: return ccir17F2(cond, deg, ord);
         case  8: return ccir18F2(cond, deg, ord);
         case  9: return ccir19F2(cond, deg, ord);
         case 10: return ccir20F2(cond, deg, ord);
         case 11: return ccir21F2(cond, deg, ord);
         case 12: return ccir22F2(cond, deg, ord);
      }
         // shouldn't be reachable
      return std::numeric_limits<double>::quiet_NaN();
   }


   double CCIR ::
   ccirFm3(unsigned month, int cond, int deg, int ord)
   {
      GNSSTK_ASSERT(month >= 1 && month <= 12);
      switch (month)
      {
         case  1: return ccir11Fm3(cond, deg, ord);
         case  2: return ccir12Fm3(cond, deg, ord);
         case  3: return ccir13Fm3(cond, deg, ord);
         case  4: return ccir14Fm3(cond, deg, ord);
         case  5: return ccir15Fm3(cond, deg, ord);
         case  6: return ccir16Fm3(cond, deg, ord);
         case  7: return ccir17Fm3(cond, deg, ord);
         case  8: return ccir18Fm3(cond, deg, ord);
         case  9: return ccir19Fm3(cond, deg, ord);
         case 10: return ccir20Fm3(cond, deg, ord);
         case 11: return ccir21Fm3(cond, deg, ord);
         case 12: return ccir22Fm3(cond, deg, ord);
      }
         // shouldn't be reachable
      return std::numeric_limits<double>::quiet_NaN();
   }


   double CCIR ::
   ccir11F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.52396593E+01,-0.56523629E-01,-0.18704617E-01, 0.12128915E-01,
              0.79412190E-02,-0.10031431E-01, 0.21567261E-01,-0.68602660E-02,
              0.37022342E-02, 0.78359339E-02, 0.63161603E-02,-0.10695397E-01,
              0.29390149E-01, },
            { 0.93325400E+00,-0.28997503E-01, 0.10946778E+00,-0.30769527E+00,
             -0.37993383E+00,-0.23273268E+00, 0.89480691E-01, 0.33896305E-01,
              0.32839003E+00,-0.81993349E-01,-0.14348941E+00,-0.27823284E-01,
              0.11266428E-01, },
            { 0.80531130E+01, 0.13981724E+01, 0.47361958E+00,-0.11388183E+00,
              0.77816737E+00,-0.17388150E+00, 0.29099104E+00, 0.29059123E-01,
             -0.37210885E+00,-0.11191850E+00,-0.43733008E-01, 0.12193082E+00,
             -0.32639468E+00, },
            {-0.13390853E+02, 0.62356526E+00, 0.24597554E+01, 0.44970918E+01,
              0.85659552E+01, 0.40355296E+01,-0.79231381E+00,-0.67203265E+00,
             -0.47442737E+01, 0.21074810E+01, 0.33913586E+01,-0.10161761E+00,
             -0.69574153E+00, },
            {-0.32273560E+02,-0.56248417E+01,-0.58702965E+01,-0.43174982E+01,
             -0.32914643E+01, 0.14731911E+01,-0.84157687E+00,-0.18629679E+00,
              0.36630037E+01,-0.49589828E+00, 0.42669845E+00,-0.39217409E+00,
              0.15235479E+01, },
            {-0.10364820E+00,-0.17449905E+02,-0.18375044E+02,-0.16429411E+02,
             -0.42843178E+02,-0.18512333E+02, 0.29954524E+01, 0.44427662E+01,
              0.22331482E+02,-0.11312363E+02,-0.17385393E+02, 0.13743354E+01,
              0.49287653E+01, },
            { 0.35317551E+02, 0.78040113E+01, 0.22017223E+02, 0.19007372E+02,
              0.51486945E+01,-0.32355106E+01,-0.70289177E+00,-0.18949838E-01,
             -0.10833933E+02, 0.30533838E+01,-0.15614654E+01, 0.34174812E+00,
             -0.33425496E+01, },
            { 0.10955015E+03, 0.64495300E+02, 0.39307384E+02, 0.21447481E+02,
              0.88321732E+02, 0.36022030E+02,-0.49279366E+01,-0.11266347E+02,
             -0.45660416E+02, 0.24368761E+02, 0.36402855E+02,-0.37914002E+01,
             -0.12354781E+02, },
            {-0.15069658E+02,-0.54928751E+01,-0.30891773E+02,-0.26297506E+02,
             -0.28576374E+01, 0.23907247E+01, 0.27161193E+01, 0.10669554E+01,
              0.12239296E+02,-0.43662004E+01, 0.20293267E+01, 0.21851628E+00,
              0.34469433E+01, },
            {-0.18371085E+03,-0.78854736E+02,-0.32724571E+02,-0.86944246E+01,
             -0.80539780E+02,-0.31638607E+02, 0.32864110E+01, 0.12187758E+02,
              0.42042595E+02,-0.23398046E+02,-0.33907742E+02, 0.42148323E+01,
              0.12834593E+02, },
            { 0.22537875E+01, 0.19722799E+01, 0.14214872E+02, 0.11752549E+02,
              0.97573824E-01,-0.41853258E+00,-0.14487737E+01,-0.91950291E+00,
             -0.47050514E+01, 0.19138944E+01,-0.85748482E+00,-0.30080923E+00,
             -0.13293968E+01, },
            { 0.86102806E+02, 0.31215933E+02, 0.92085867E+01,-0.52750909E+00,
              0.26795364E+02, 0.10308076E+02,-0.60259044E+00,-0.47401175E+01,
             -0.14323977E+02, 0.83509979E+01, 0.11643826E+02,-0.16808408E+01,
             -0.47212696E+01, },
            { 0.53214836E-02, 0.17079000E+01, 0.16235816E+01, 0.33274181E-01,
             -0.12954264E-01,-0.36316011E-02, 0.45543164E-01,-0.43191575E-02,
             -0.25532667E-01, 0.89464840E-02, 0.23123844E-01, 0.16921325E-01,
              0.33485752E-01, },
            {-0.22577250E+00,-0.16961778E+01, 0.17585011E+01,-0.36464985E-01,
             -0.46494432E-01,-0.19080851E-02,-0.31143922E-01, 0.10176010E-01,
             -0.69593927E-02,-0.66427351E-02, 0.23588389E-02,-0.55617536E-02,
             -0.19860143E-01, },
            {-0.17363087E+01, 0.18277783E+01, 0.16650461E+01,-0.48928317E-01,
             -0.43136498E+00,-0.47681876E-01, 0.25482827E+00,-0.18712924E+00,
              0.47013280E+00,-0.17417172E+00, 0.78237861E-01, 0.10381135E+00,
              0.17819388E-01, },
            {-0.14371649E+01,-0.80177718E+00, 0.61360812E+00,-0.59573984E+00,
              0.52904451E+00,-0.10058827E+00, 0.21083134E+00, 0.78204811E-01,
              0.61967522E-02, 0.11077180E+00,-0.40769756E-01, 0.21807816E-01,
              0.20648648E-02, },
            { 0.46911736E+01, 0.78173614E+01, 0.12199426E+02,-0.16422656E+01,
              0.37755153E+01,-0.40263680E+00,-0.52920246E+00,-0.21391793E+00,
              0.57929236E+00, 0.80246784E-01,-0.18910319E+00,-0.50804520E+00,
             -0.88265932E+00, },
            { 0.69275224E+00,-0.12002475E+02, 0.57374411E+01, 0.13528378E+01,
              0.44981871E+01,-0.47366244E+00, 0.36056972E+00, 0.79039741E+00,
             -0.22344507E-01,-0.64753883E-01, 0.46970192E+00,-0.40660375E+00,
              0.44191428E-01, },
            { 0.39014656E+02,-0.28619871E+02,-0.15200562E+01,-0.19682378E+01,
              0.13416008E+02, 0.27089481E+01,-0.59305787E+01, 0.31192074E+01,
             -0.91367588E+01, 0.26589971E+01,-0.11606407E+00,-0.35579140E+01,
             -0.89003706E+00, },
            { 0.34678040E+02,-0.52889085E+00,-0.57936335E+01, 0.87519836E+01,
             -0.10267422E+02, 0.12299103E+01,-0.30112040E+01, 0.11044223E+01,
             -0.21466107E+01, 0.45218792E-01, 0.30407925E+01,-0.71307260E+00,
             -0.15156338E+01, },
            {-0.52359825E+02,-0.65902946E+02,-0.30542086E+02, 0.45587659E+01,
             -0.33256973E+02, 0.32021556E+01, 0.60264249E+01, 0.39931140E+01,
             -0.21957397E+01,-0.14460717E+01,-0.62892991E+00, 0.19364247E+01,
              0.71591692E+01, },
            { 0.99375763E+01, 0.36078590E+02,-0.38465443E+02,-0.34042759E+01,
             -0.46145630E+02, 0.20836172E+01,-0.37764747E+01,-0.62155070E+01,
             -0.36285675E+00, 0.83748114E+00,-0.11298743E+01, 0.38863559E+01,
              0.14642847E+01, },
            {-0.22182343E+03, 0.11115142E+03,-0.16951048E+02, 0.17158335E+02,
             -0.61797066E+02,-0.21725040E+02, 0.33751064E+02,-0.20662024E+02,
              0.52077259E+02,-0.10576082E+02, 0.19523579E+00, 0.24283333E+02,
              0.39174995E+01, },
            {-0.22131866E+03, 0.17304182E+02,-0.28861012E+02,-0.44697052E+02,
              0.53147095E+02,-0.61317348E+01, 0.16753126E+02,-0.89905128E+01,
              0.20872980E+02,-0.63277907E+01,-0.26276550E+02, 0.64804440E+01,
              0.11094950E+02, },
            { 0.17550497E+03, 0.16372993E+03,-0.21463629E+02, 0.27199810E+01,
              0.99251183E+02,-0.79424796E+01,-0.21751915E+02,-0.18336351E+02,
              0.69923621E+00, 0.50953045E+01, 0.72605686E+01,-0.15412294E+01,
             -0.22803534E+02, },
            {-0.56229095E+02, 0.14818197E+01, 0.69807755E+02, 0.33564434E+01,
              0.14655196E+03,-0.27966247E+01, 0.12579897E+02, 0.17929468E+02,
              0.43725858E+01,-0.22176220E+01,-0.37626834E+01,-0.10943099E+02,
             -0.52088380E+01, },
            { 0.53665625E+03,-0.20060770E+03, 0.17310247E+02,-0.54237488E+02,
              0.99595345E+02, 0.66468887E+02,-0.80023064E+02, 0.58134552E+02,
             -0.12709147E+03, 0.15493525E+02,-0.31282263E+01,-0.65706467E+02,
             -0.49680519E+01, },
            { 0.55392444E+03,-0.46329002E+02, 0.14449484E+03, 0.10233988E+03,
             -0.11556944E+03, 0.13084748E+02,-0.44251144E+02, 0.22732943E+02,
             -0.62439163E+02, 0.21050632E+02, 0.77459229E+02,-0.19942715E+02,
             -0.28916513E+02, },
            {-0.23239845E+03,-0.18225694E+03, 0.10545853E+03,-0.16537470E+02,
             -0.12225861E+03, 0.72478418E+01, 0.29107874E+02, 0.29596212E+02,
              0.43757663E+01,-0.56305561E+01,-0.14893911E+02,-0.16775293E+01,
              0.30470707E+02, },
            { 0.95573692E+02,-0.82114143E+02,-0.54173027E+02,-0.28702703E+01,
             -0.18174019E+03,-0.33480763E+00,-0.15407213E+02,-0.20705093E+02,
             -0.86905804E+01, 0.13393050E+01, 0.11856750E+02, 0.11862597E+02,
              0.56412406E+01, },
            {-0.58746796E+03, 0.17872774E+03, 0.44170815E+02, 0.71773781E+02,
             -0.55947525E+02,-0.85275742E+02, 0.84508293E+02,-0.71050819E+02,
              0.13868575E+03,-0.73421726E+01, 0.69255672E+01, 0.76826965E+02,
              0.51061606E+00, },
            {-0.58909216E+03, 0.41609570E+02,-0.19081412E+03,-0.10350248E+03,
              0.11202155E+03,-0.11497140E+02, 0.53446163E+02,-0.24178514E+02,
              0.73052414E+02,-0.24337910E+02,-0.93695763E+02, 0.24911678E+02,
              0.31728121E+02, },
            { 0.10805177E+03, 0.77080933E+02,-0.67569344E+02, 0.11093394E+02,
              0.53251934E+02,-0.18096049E+01,-0.12874349E+02,-0.15589551E+02,
             -0.36182413E+01, 0.18153871E+01, 0.87414274E+01, 0.19512452E+01,
             -0.14214747E+02, },
            {-0.53066933E+02, 0.58636524E+02, 0.16770851E+02, 0.15051444E+01,
              0.77045029E+02, 0.18778749E+01, 0.61496325E+01, 0.81532841E+01,
              0.46847639E+01, 0.32412454E+00,-0.78524041E+01,-0.42911925E+01,
             -0.19247583E+01, },
            { 0.23769147E+03,-0.61533985E+02,-0.47820061E+02,-0.34214542E+02,
              0.40193143E+01, 0.38487625E+02,-0.32399452E+02, 0.30905638E+02,
             -0.55419487E+02,-0.24066789E+00,-0.41797304E+01,-0.32275726E+02,
              0.15222611E+01, },
            { 0.22423343E+03,-0.12581701E+02, 0.80397057E+02, 0.37280334E+02,
             -0.41082157E+02, 0.31536264E+01,-0.23678221E+02, 0.94953613E+01,
             -0.29353230E+02, 0.94907122E+01, 0.40155750E+02,-0.10989323E+02,
             -0.12388154E+02, },
            { 0.20844085E-01,-0.80697089E-01, 0.30533599E-01,-0.64598656E+00,
             -0.16657443E+00,-0.21123195E-01, 0.92248200E-02,-0.25574602E-02,
             -0.73007592E-02, 0.12166507E-01,-0.12164736E-01, 0.68045142E-02,
             -0.77586402E-02, },
            {-0.69821566E-01, 0.10285664E-01, 0.30625435E-01, 0.12556811E+00,
             -0.77144945E+00, 0.14403922E-01, 0.32232669E-02,-0.18554740E-01,
              0.41907090E-02,-0.10089985E-02, 0.43220972E-02, 0.18566685E-02,
             -0.25719402E-01, },
            {-0.36431205E+00,-0.18156898E+00,-0.11474409E+00, 0.64775425E+00,
              0.20361809E+00, 0.58565461E-02,-0.85165583E-01,-0.11972090E+00,
             -0.25655270E+00,-0.12931146E+00,-0.37106670E-01,-0.15892604E+00,
             -0.45120206E-01, },
            { 0.10515987E+01, 0.61001372E+00, 0.23907986E+00,-0.96539223E+00,
              0.35442737E+00,-0.22381260E+00, 0.19057482E+00,-0.75943330E-02,
              0.47220100E-01,-0.86490205E-02, 0.44078577E-01,-0.37896223E-01,
             -0.78807129E-02, },
            { 0.27450144E+00,-0.69624442E+00,-0.14316448E+01, 0.18575617E+01,
              0.37408116E+01,-0.57409221E+00,-0.48675659E+00, 0.22489709E+00,
             -0.55995405E-01, 0.24701399E+00, 0.39949957E-01, 0.95097050E-02,
              0.15562671E+00, },
            { 0.29598875E+01,-0.22428038E+01, 0.16291965E+01,-0.38134632E+01,
              0.55432544E+01,-0.76043957E+00,-0.81076163E+00, 0.97178990E+00,
              0.18208525E+00,-0.16950764E-01, 0.58102113E+00,-0.36129999E+00,
              0.14220279E+00, },
            { 0.34898536E+01, 0.21215761E+01, 0.22773728E+01,-0.46729574E+01,
              0.32191403E+01, 0.10207694E+01, 0.40245906E+00, 0.71286887E+00,
              0.25590413E+01, 0.10251747E+01,-0.47245717E+00, 0.82578117E+00,
             -0.63747823E-01, },
            {-0.69263606E+01,-0.71795802E+01, 0.26847789E+01, 0.42841086E+01,
             -0.33002298E+01, 0.25364313E+01,-0.16150088E+01, 0.56352544E+00,
              0.55770433E+00, 0.44689545E+00,-0.62278980E+00,-0.38403094E+00,
             -0.24888094E+00, },
            {-0.28185658E+01, 0.10438052E+02, 0.53413100E+01, 0.12666808E+02,
              0.10997744E+02, 0.27631536E+01, 0.28622413E+01,-0.12079479E+01,
              0.16985104E+01,-0.21927736E+01, 0.75003189E+00, 0.98793454E-01,
             -0.10628952E+01, },
            {-0.15150089E+02, 0.23410828E+02,-0.12058938E+02,-0.97661982E+01,
             -0.59994440E+01, 0.12919568E+01, 0.10240656E+01,-0.46654758E+01,
             -0.17236120E+01, 0.58620757E+00,-0.33529608E+01, 0.26049089E+01,
             -0.11627923E+00, },
            {-0.76791863E+01,-0.53989253E+01,-0.93963947E+01, 0.89348888E+01,
             -0.13714699E+02,-0.40793180E+01,-0.13703756E+01,-0.11329166E+01,
             -0.66214280E+01,-0.19741457E+01, 0.19676812E+01,-0.15802444E+01,
              0.74486417E+00, },
            { 0.10005783E+02, 0.22066399E+02,-0.93126631E+01,-0.34220302E+01,
              0.73809199E+01,-0.56643796E+01, 0.28039861E+01,-0.20841560E+01,
             -0.25536554E+01,-0.20768545E+01, 0.14609894E+01, 0.22773330E+01,
              0.10445656E+01, },
            { 0.88902111E+01,-0.21831713E+02,-0.75237002E+01,-0.45947899E+02,
             -0.52254841E+02,-0.43777122E+01,-0.65832415E+01, 0.27934115E+01,
             -0.51957211E+01, 0.45358448E+01,-0.29035003E+01,-0.80293524E+00,
              0.31588891E+01, },
            { 0.30450523E+02,-0.57946457E+02, 0.24090494E+02, 0.50759991E+02,
             -0.16110195E+02, 0.35295739E+01, 0.28397505E+01, 0.93683434E+01,
              0.46508346E+01,-0.98472559E+00, 0.71794248E+01,-0.56259923E+01,
              0.42344132E+00, },
            { 0.37496376E+01, 0.41398544E+01, 0.89929867E+01,-0.21592083E+01,
              0.15710482E+02, 0.37366185E+01, 0.12688895E+01, 0.56009912E+00,
              0.45878363E+01, 0.10043666E+01,-0.15857160E+01, 0.11541928E+01,
             -0.63341975E+00, },
            {-0.16841691E+01,-0.19766069E+02, 0.61021037E+01,-0.47736259E+01,
             -0.30741682E+01, 0.37985227E+01,-0.14054892E+01, 0.19982285E+01,
              0.24144678E+01, 0.21827767E+01,-0.78220928E+00,-0.24638696E+01,
             -0.81352621E+00, },
            {-0.86120291E+01, 0.12768568E+02, 0.40449328E+01, 0.35735741E+02,
              0.45281525E+02, 0.14536483E+01, 0.48580317E+01,-0.21178229E+01,
              0.40526733E+01,-0.26759613E+01, 0.23512807E+01, 0.86008638E+00,
             -0.27273736E+01, },
            {-0.23491453E+02, 0.42392094E+02,-0.13589549E+02,-0.45290329E+02,
              0.22402483E+02,-0.52239609E+01,-0.39634597E+01,-0.66355700E+01,
             -0.30386250E+01, 0.24522121E+00,-0.52094841E+01, 0.38612597E+01,
             -0.76240212E+00, },
            {-0.14890511E+00,-0.24325432E-01,-0.38893390E-01, 0.66723570E-01,
              0.59333660E-01,-0.36929664E+00,-0.60848296E+00, 0.94836196E-02,
              0.54205763E-02,-0.15333358E-01,-0.11106117E-02, 0.22575622E-02,
              0.23593083E-02, },
            {-0.22660287E-01, 0.45373976E-01, 0.50671652E-01,-0.36677148E-01,
              0.63135192E-01, 0.61916250E+00,-0.39381266E+00,-0.22737507E-01,
              0.37895903E-01,-0.11128758E-01,-0.91112573E-02,-0.10985604E-01,
              0.15489354E-01, },
            {-0.22802021E+00, 0.26512158E+00,-0.22216140E+00, 0.10585243E+00,
             -0.16293724E+00,-0.69953322E+00,-0.20355964E+00, 0.23573788E-01,
              0.29221939E-01,-0.58141425E-01,-0.10479219E+00, 0.38031027E-01,
              0.34238581E-01, },
            {-0.32278037E+00, 0.10433105E+00,-0.70636101E-01,-0.45763806E-01,
              0.49903285E-01, 0.49318537E+00,-0.80187595E+00,-0.83227634E-01,
              0.23710161E-01,-0.30420637E-01,-0.10782068E-01, 0.52331045E-01,
              0.40990692E-01, },
            { 0.75721157E+00,-0.10054682E+00,-0.43790352E+00, 0.35622928E-01,
             -0.68008196E+00, 0.13344156E+01, 0.10325201E+01, 0.21211481E+00,
             -0.52811496E-01, 0.58822650E-01,-0.29041562E-01, 0.13278130E+00,
              0.60832474E-01, },
            {-0.51921064E+00, 0.73404215E-01,-0.14045806E+00, 0.56511676E+00,
             -0.10780083E+01,-0.90128416E+00, 0.18347827E+01, 0.23589993E-01,
              0.19664790E+00, 0.10371149E+00, 0.16768134E+00, 0.25422749E+00,
             -0.27516479E-01, },
            { 0.47642750E+00,-0.12712888E+01, 0.50919574E+00,-0.37558299E-01,
              0.41861004E+00, 0.72344553E+00, 0.58304822E+00,-0.37630741E-01,
             -0.21993767E-01, 0.15301681E+00, 0.14597137E+00,-0.88676058E-01,
             -0.34135389E-02, },
            { 0.10227042E+01,-0.10414064E+00, 0.21676075E+00,-0.18772216E+00,
              0.27981552E+00,-0.11495651E+01, 0.94243652E+00, 0.13660996E+00,
             -0.18090218E+00, 0.32121949E-01, 0.75706489E-01,-0.24605010E+00,
             -0.13862634E+00, },
            {-0.76268691E+00, 0.10962813E+01, 0.10399640E+01, 0.96028954E-01,
              0.77267742E+00,-0.18069416E+01,-0.60100126E+00,-0.48312178E+00,
             -0.72206438E-01, 0.13322057E-02,-0.17916759E-01,-0.28112444E+00,
             -0.19206632E-02, },
            { 0.68990123E+00,-0.17288908E+00, 0.42733526E+00,-0.42356339E+00,
              0.20443857E+01, 0.25324249E+00,-0.22968805E+01, 0.21025427E+00,
             -0.64606416E+00,-0.20289107E+00,-0.37085348E+00,-0.52354544E+00,
             -0.47585368E-01, },
            { 0.59121337E-01, 0.75014145E-02, 0.32973826E-01, 0.16474618E-01,
              0.32099034E-02, 0.20523887E-01, 0.14850210E-01, 0.15989561E+00,
             -0.22012538E+00, 0.11682349E-01, 0.37532360E-02, 0.28839896E-02,
              0.90501476E-02, },
            { 0.25123442E-01,-0.54974113E-01,-0.40254004E-01,-0.64294338E-01,
              0.18476812E-01, 0.54143360E-02, 0.26773563E-01, 0.22185430E+00,
              0.18453135E+00, 0.40824950E-03,-0.20066464E-01, 0.10584317E-01,
             -0.42071915E-02, },
            {-0.11119860E-01,-0.84521756E-01,-0.72065502E-01,-0.10357253E+00,
              0.25481060E-01,-0.31697039E-01,-0.34495186E-01,-0.12500069E+00,
             -0.29316071E+00,-0.50189231E-01, 0.20690957E-01, 0.20984372E-01,
              0.46142563E-01, },
            { 0.31204326E-02,-0.15957630E+00,-0.45523688E-01, 0.11955297E-02,
              0.21015655E-01, 0.35082828E-01,-0.32445975E-01, 0.22741143E+00,
             -0.11440502E+00, 0.16439863E-01,-0.42733331E-02, 0.36579225E-01,
             -0.35476536E-01, },
            { 0.20448800E-01, 0.88741951E-01,-0.45464959E-01,-0.15399729E-01,
              0.49765386E-01,-0.19545292E-01, 0.20340797E-01,-0.11448290E-01,
              0.21605456E-01, 0.11625878E+00, 0.15443808E+00,-0.68518179E-02,
              0.68731960E-02, },
            { 0.44177953E-01, 0.17190436E-01,-0.34843344E-01,-0.55796955E-01,
             -0.18618047E-01,-0.13446327E-02,-0.13807907E-01, 0.16442720E-01,
             -0.63370585E-01,-0.13994679E+00, 0.10775572E+00, 0.18059421E-01,
             -0.11631262E-02, },
            { 0.56581877E-01,-0.43296669E-01, 0.87594809E-02,-0.19183895E-01,
              0.22148756E-01, 0.16138326E-01, 0.35961125E-01, 0.62827743E-02,
             -0.29091935E-02, 0.17914118E-01, 0.23472050E-01,-0.40293451E-01,
              0.52220732E-01, },
            {-0.28325748E-01, 0.18121796E-01,-0.55669233E-01,-0.43550450E-01,
             -0.41511986E-01, 0.28219909E-03,-0.45169748E-01, 0.30541521E-01,
             -0.75302455E-02, 0.11561214E-01,-0.71855863E-02,-0.42805452E-01,
             -0.74895360E-01, },
            {-0.40362079E-01, 0.39089788E-02,-0.43586571E-01,-0.13430195E-01,
              0.19074345E-01,-0.14927471E-02,-0.83385846E-02, 0.12412880E-01,
             -0.39789891E-02,-0.93350494E-02, 0.31809614E-03, 0.69165085E-02,
              0.16948413E-01, },
            { 0.40239409E-01, 0.21818378E-02,-0.23886045E-01,-0.19706257E-02,
              0.17834106E-01, 0.75256284E-02,-0.88512655E-02, 0.25555318E-01,
              0.24699425E-01,-0.68906411E-02,-0.15585792E-02, 0.12531394E-01,
             -0.15920257E-01, },
            { 0.71578033E-01, 0.36469545E-01,-0.18202318E-01, 0.23860386E-01,
             -0.15428271E-02,-0.31324875E-01, 0.12997811E-01, 0.49806456E-02,
             -0.13169400E-01, 0.15163812E-01, 0.45060008E-02, 0.96705575E-02,
              0.32339197E-01, },
            {-0.71236528E-01, 0.24793777E-03, 0.38303085E-01, 0.23001375E-01,
             -0.34666590E-02, 0.19711124E-01, 0.14156577E-01,-0.15163544E-02,
              0.14914329E-01,-0.16906057E-01, 0.68336162E-02, 0.72287614E-02,
             -0.10352528E-01, },
         },
         {
            { 0.84432182E+01,-0.17606992E-01,-0.12995030E-01, 0.30051462E-01,
              0.45477942E-01, 0.12136347E-01,-0.20886991E-01,-0.25218520E-01,
              0.38802756E-02, 0.50634779E-02, 0.17426111E-01,-0.19343946E-01,
              0.21120729E-01, },
            { 0.13067631E+01,-0.19031613E+00,-0.47110280E+00, 0.15309569E+00,
             -0.25000983E+00,-0.21639533E+00, 0.13572131E+00, 0.16442597E-01,
              0.29920632E+00,-0.35695668E-01, 0.72213798E-02, 0.36948250E-03,
             -0.19350252E+00, },
            { 0.20192820E+02, 0.10371314E+00,-0.12936909E+01,-0.47384906E+00,
              0.43617851E+00,-0.88357627E+00, 0.84384918E+00, 0.40130630E+00,
             -0.24254835E+00, 0.19504283E+00,-0.32403544E+00, 0.68768036E+00,
             -0.70447671E+00, },
            { 0.10385976E+02, 0.26844606E+01, 0.91865940E+01,-0.28847570E+01,
              0.69040632E+01, 0.30880930E+01,-0.16791331E+01,-0.77926904E+00,
             -0.40792027E+01, 0.16587726E+01, 0.11048369E+01, 0.56877963E-01,
              0.26553495E+01, },
            {-0.94894234E+02,-0.74020910E+00, 0.97572689E+01,-0.13988616E+01,
             -0.48308020E+01, 0.61252832E+01,-0.29142206E+01,-0.31117852E+01,
              0.10199926E+01,-0.25860064E+01, 0.15488667E+01,-0.52199402E+01,
              0.62431121E+01, },
            {-0.15947099E+03,-0.33153683E+02,-0.38277382E+02, 0.16296080E+02,
             -0.37176064E+02,-0.16504719E+02, 0.97844954E+01, 0.60243282E+01,
              0.18602617E+02,-0.88784866E+01,-0.70023069E+01,-0.85667688E+00,
             -0.10270858E+02, },
            { 0.14434691E+03, 0.40043654E+01,-0.22543146E+02, 0.10819620E+02,
              0.13452056E+02,-0.14324649E+02, 0.14989966E+01, 0.83210297E+01,
             -0.11521393E+01, 0.86834040E+01,-0.32280769E+01, 0.13568054E+02,
             -0.17570894E+02, },
            { 0.46371674E+03, 0.11013431E+03, 0.54929173E+02,-0.38762981E+02,
              0.82882599E+02, 0.38311665E+02,-0.25024160E+02,-0.15210931E+02,
             -0.36635254E+02, 0.18874268E+02, 0.14921486E+02, 0.31062043E+01,
              0.17111540E+02, },
            {-0.10361338E+03,-0.10559738E+02, 0.20353815E+02,-0.17083015E+02,
             -0.12879775E+02, 0.13625260E+02, 0.28179276E+01,-0.85639162E+01,
             -0.39836466E+00,-0.10996532E+02, 0.32205973E+01,-0.14092367E+02,
              0.19663298E+02, },
            {-0.51334308E+03,-0.13410316E+03,-0.27003073E+02, 0.40749172E+02,
             -0.81037025E+02,-0.38719093E+02, 0.27226254E+02, 0.15593900E+02,
              0.32171307E+02,-0.18132610E+02,-0.12962888E+02,-0.39625747E+01,
             -0.13030040E+02, },
            { 0.30432690E+02, 0.74267569E+01,-0.65251384E+01, 0.81919823E+01,
              0.35932457E+01,-0.45463281E+01,-0.21767290E+01, 0.29457211E+01,
              0.77555543E+00, 0.47232051E+01,-0.12480311E+01, 0.50437918E+01,
             -0.76494961E+01, },
            { 0.19720856E+03, 0.54842201E+02, 0.14811361E+01,-0.15508772E+02,
              0.28584126E+02, 0.13968620E+02,-0.10371297E+02,-0.56411085E+01,
             -0.10385420E+02, 0.65630751E+01, 0.39054112E+01, 0.16419491E+01,
              0.37288151E+01, },
            { 0.73617727E-01, 0.15868278E+01, 0.15948020E+01,-0.51909208E-01,
              0.10965753E+00, 0.42061005E-01, 0.86369058E-02,-0.35956562E-01,
             -0.24439633E-01, 0.13813865E-01, 0.33473477E-01,-0.12541716E-01,
              0.13831583E-01, },
            {-0.22122271E+00,-0.16716869E+01, 0.17568765E+01,-0.10806262E+00,
             -0.60551725E-01, 0.32627724E-01,-0.40097758E-01, 0.32800335E-01,
             -0.22351671E-01,-0.30436333E-01, 0.85034929E-02, 0.44959588E-02,
             -0.57923377E-01, },
            {-0.13319713E+01, 0.20320992E+00, 0.28417442E+01,-0.26053941E+00,
              0.41736737E-01, 0.48073849E+00, 0.30752096E+00, 0.27773425E+00,
              0.58028370E+00,-0.27988768E+00, 0.20987786E+00, 0.21454689E+00,
             -0.21824501E+00, },
            {-0.32366142E+00,-0.18102030E+01, 0.73649567E+00,-0.12784365E+01,
              0.13987561E+00, 0.29401943E+00,-0.11734486E+00, 0.30285731E+00,
              0.47187995E-01,-0.40725458E+00, 0.69726735E-01,-0.14392924E-01,
             -0.16440730E+00, },
            { 0.43222728E+01, 0.25277067E+02,-0.94375896E+01, 0.12424446E+01,
              0.19729581E-01,-0.89677191E+00,-0.36436412E+00, 0.90389627E+00,
             -0.13076818E+00, 0.15358768E+00,-0.36496457E+00, 0.83983088E+00,
             -0.97137824E-01, },
            {-0.39563198E+01, 0.94551926E+01, 0.17336029E+02, 0.19375322E+01,
              0.29811153E+01, 0.36434084E+00, 0.96826065E+00, 0.76319680E-01,
              0.63908470E+00, 0.66578782E+00,-0.31862479E+00,-0.63156575E+00,
              0.88741764E-01, },
            { 0.31899084E+02, 0.26338453E+02,-0.26924992E+02,-0.28954151E+01,
              0.11245874E+01,-0.62383070E+01,-0.71671276E+01,-0.44009361E+01,
             -0.11893276E+02, 0.59641714E+01,-0.12007704E+01,-0.48183508E+01,
              0.41470280E+01, },
            { 0.29249310E+01, 0.10189669E+02, 0.64222164E+01, 0.16846102E+02,
             -0.56195569E+01,-0.38048911E+01, 0.29025004E+01,-0.38915317E+01,
             -0.16548055E+01, 0.71972709E+01, 0.18794578E+01, 0.20752627E+00,
              0.15136871E+01, },
            {-0.47960709E+02,-0.14527844E+03, 0.13558260E+03,-0.15012978E+02,
             -0.89980879E+01, 0.33358767E+01, 0.65263834E+01,-0.57562585E+01,
              0.23737571E+01,-0.32089403E+01, 0.66337514E+00,-0.98810377E+01,
             -0.12185401E+01, },
            { 0.54375412E+02,-0.13382069E+03,-0.69432632E+02,-0.15594945E+02,
             -0.28076584E+02,-0.55643253E+01,-0.81322355E+01,-0.26028937E+00,
             -0.52387090E+01,-0.39641640E+01, 0.58208838E+01, 0.58059587E+01,
              0.43214092E+01, },
            {-0.18676460E+03,-0.16449841E+03, 0.18679526E+03, 0.44028065E+02,
              0.67239275E+01, 0.20764881E+02, 0.43572437E+02, 0.15105406E+02,
              0.72933273E+02,-0.35417358E+02,-0.33102422E+01, 0.33491127E+02,
             -0.28131252E+02, },
            { 0.13132272E+02,-0.71697983E+02,-0.14838391E+02,-0.88297775E+02,
              0.35739063E+02, 0.24429436E+02,-0.19637230E+02, 0.15767020E+02,
              0.12106133E+02,-0.38460899E+02,-0.17444717E+02,-0.57627004E+00,
             -0.71105013E+01, },
            { 0.16358971E+03, 0.30532544E+03,-0.45578516E+03, 0.51725403E+02,
              0.42052006E+02,-0.12835170E+01,-0.28778358E+02, 0.11572843E+02,
             -0.11792854E+02, 0.97184315E+01, 0.28023026E+01, 0.34010338E+02,
              0.76306810E+01, },
            {-0.19060997E+03, 0.45207547E+03, 0.60605202E+02, 0.49964787E+02,
              0.83181343E+02, 0.17889486E+02, 0.23796255E+02,-0.10524486E+01,
              0.17759972E+02, 0.98242826E+01,-0.24883154E+02,-0.17036594E+02,
             -0.18363985E+02, },
            { 0.47393274E+03, 0.36210413E+03,-0.54417682E+03,-0.15719064E+03,
             -0.48869652E+02,-0.19227535E+02,-0.11137081E+03,-0.14016641E+02,
             -0.18484474E+03, 0.86145981E+02, 0.26243605E+02,-0.93299721E+02,
              0.76153831E+02, },
            {-0.10607853E+03, 0.18373109E+03,-0.71254395E+02, 0.21552979E+03,
             -0.88725677E+02,-0.66744774E+02, 0.48730526E+02,-0.28787111E+02,
             -0.34096436E+02, 0.84370399E+02, 0.52670143E+02,-0.46487325E+00,
              0.18449718E+02, },
            {-0.22080930E+03,-0.29196872E+03, 0.61042780E+03,-0.71494499E+02,
             -0.68313850E+02,-0.85744381E+01, 0.44273830E+02,-0.82676754E+01,
              0.20448513E+02,-0.80523005E+01,-0.84449396E+01,-0.45622372E+02,
             -0.12797120E+02, },
            { 0.25156476E+03,-0.59986975E+03, 0.34923374E+02,-0.63288609E+02,
             -0.98415474E+02,-0.20724583E+02,-0.28114664E+02, 0.44651847E+01,
             -0.23789305E+02,-0.11882715E+02, 0.37959538E+02, 0.19314947E+02,
              0.25713858E+02, },
            {-0.54166937E+03,-0.33657809E+03, 0.70614557E+03, 0.21372903E+03,
              0.80236282E+02,-0.66625953E+01, 0.12710117E+03,-0.65174923E+01,
              0.20572482E+03,-0.91904800E+02,-0.43451694E+02, 0.11079334E+03,
             -0.87262184E+02, },
            { 0.19350169E+03,-0.21782204E+03, 0.19717972E+03,-0.23897552E+03,
              0.95249496E+02, 0.76332985E+02,-0.49345135E+02, 0.26176369E+02,
              0.39268757E+02,-0.81754929E+02,-0.66510406E+02, 0.24783335E+01,
             -0.23144762E+02, },
            { 0.10398347E+03, 0.10713031E+03,-0.28724725E+03, 0.34307171E+02,
              0.36974197E+02, 0.81658230E+01,-0.22095419E+02, 0.13036780E+01,
             -0.11200826E+02, 0.97571468E+00, 0.54882064E+01, 0.21151997E+02,
              0.67120991E+01, },
            {-0.11383261E+03, 0.27709833E+03,-0.45287556E+02, 0.26823547E+02,
              0.40176460E+02, 0.79526610E+01, 0.11477461E+02,-0.35184112E+01,
              0.10715305E+02, 0.56945720E+01,-0.19063519E+02,-0.72694077E+01,
             -0.11852451E+02, },
            { 0.22522438E+03, 0.11369879E+03,-0.33281451E+03,-0.10020547E+03,
             -0.39448936E+02, 0.11468914E+02,-0.53004147E+02, 0.10001610E+02,
             -0.83203133E+02, 0.35805698E+02, 0.21695692E+02,-0.46803165E+02,
              0.35515347E+02, },
            {-0.10584347E+03, 0.98553574E+02,-0.12213636E+03, 0.96482056E+02,
             -0.38166542E+02,-0.30556103E+02, 0.17174086E+02,-0.97377214E+01,
             -0.15434036E+02, 0.29186680E+02, 0.30085251E+02,-0.17299670E+01,
              0.10782919E+02, },
            {-0.27865281E-01,-0.14160953E+00,-0.46167344E-01,-0.83359647E+00,
              0.54028988E+00, 0.17667174E-01,-0.23999680E-02,-0.50582860E-01,
             -0.34205623E-01, 0.24929952E-01,-0.20637264E-01,-0.54828986E-02,
             -0.16638445E-02, },
            {-0.99061303E-01,-0.61199144E-02, 0.16169669E-01,-0.50423771E+00,
             -0.87746030E+00,-0.14626432E-02, 0.91258921E-02, 0.41810900E-01,
              0.27339887E-02,-0.15622666E-02, 0.32905261E-02, 0.18039158E-01,
             -0.38937837E-01, },
            {-0.11977397E+00,-0.37973851E-01, 0.51543677E+00,-0.18759683E+01,
              0.63566899E+00, 0.28123093E+00,-0.29566622E+00,-0.18840617E-01,
             -0.36266062E-01,-0.26874974E+00,-0.17266946E+00, 0.10599524E-01,
             -0.83736926E-01, },
            { 0.49956435E+00, 0.68339938E+00, 0.18975423E+00,-0.11983259E+01,
             -0.20601561E+01, 0.58113441E-01, 0.15165167E+00, 0.12512311E+00,
              0.19267312E+00,-0.55146988E-01, 0.15980954E+00, 0.82305297E-01,
             -0.42632781E-01, },
            { 0.23057373E+01, 0.76537257E+00, 0.43733449E+01,-0.50523310E+01,
              0.28448110E+01, 0.51284897E+00,-0.16459599E+01, 0.74293798E+00,
              0.66224128E-01,-0.41510153E+00, 0.26375905E+00, 0.37355468E+00,
             -0.17590469E+00, },
            { 0.11372099E+01,-0.12647940E+00, 0.25441618E+01,-0.39932528E+01,
             -0.41477337E+01, 0.13616855E+01,-0.13583241E+00,-0.12699013E+01,
             -0.66975029E-02,-0.43774159E-02, 0.72404754E+00,-0.79435611E+00,
              0.15491714E+00, },
            { 0.19683877E+01, 0.29083934E+01,-0.72367439E+01, 0.59465566E+01,
             -0.19175700E+01,-0.14585569E+01, 0.16693540E+01,-0.56365860E+00,
              0.17722517E+01, 0.32326121E+01, 0.53729033E+00,-0.88220710E+00,
             -0.56293678E+00, },
            {-0.41296825E+01,-0.31065300E+01, 0.19166176E+01, 0.91364346E+01,
              0.66980739E+01, 0.77398407E+00,-0.30196124E+00,-0.13194689E+01,
             -0.16660863E+01, 0.88819361E+00,-0.11270262E+01,-0.17106761E+01,
              0.95284730E-01, },
            {-0.11591570E+02, 0.10206953E+02,-0.44386238E+02, 0.47346340E+02,
             -0.79038370E+00,-0.90883474E+01, 0.11115670E+02,-0.19898731E+01,
              0.37733054E+00, 0.13326268E+01,-0.29739472E+00,-0.21107407E+01,
              0.19746448E+01, },
            {-0.77565346E+01, 0.19142870E+02,-0.19318724E+02, 0.79751496E+01,
              0.45732677E+02,-0.11700178E+02,-0.21409187E+01, 0.95194101E+01,
             -0.12891829E+01,-0.70309061E+00,-0.49068875E+01, 0.54949450E+01,
              0.72287112E+00, },
            {-0.51350603E+01,-0.88256512E+01, 0.16213120E+02,-0.95362787E+01,
             -0.49770875E+01, 0.10020863E+01,-0.38989060E+01, 0.27658885E+01,
             -0.65731020E+01,-0.83604755E+01,-0.15559013E+00, 0.25949471E+01,
              0.29289787E+01, },
            { 0.96731539E+01, 0.52264996E+01,-0.65303130E+01,-0.14551664E+02,
             -0.86602306E+01,-0.16371939E+01,-0.16511111E+01, 0.34544625E+01,
              0.35504460E+01,-0.31492507E+01, 0.25533750E+01, 0.52310605E+01,
             -0.48787493E-01, },
            { 0.20374853E+02,-0.30132755E+02, 0.10810119E+03,-0.10528988E+03,
             -0.18006445E+02, 0.24039795E+02,-0.24310305E+02, 0.27824404E+01,
             -0.91876078E+00,-0.13190975E+01,-0.37454176E+00, 0.39671352E+01,
             -0.47484903E+01, },
            { 0.22527531E+02,-0.60154240E+02, 0.38476719E+02, 0.45293006E+00,
             -0.10749376E+03, 0.29633495E+02, 0.56231313E+01,-0.19200817E+02,
              0.49150085E+01, 0.26779370E+01, 0.99957170E+01,-0.12485248E+02,
             -0.21405015E+01, },
            { 0.15377275E+01, 0.74568505E+01,-0.10354218E+02, 0.10922906E+02,
              0.13785338E+02, 0.83337110E+00, 0.27636354E+01,-0.27910285E+01,
              0.50508113E+01, 0.60138631E+01,-0.15283442E+00,-0.17226232E+01,
             -0.27341883E+01, },
            {-0.54970169E+01,-0.42347341E+01, 0.50668473E+01, 0.30705661E+00,
              0.80789337E+01, 0.52706480E+00, 0.30988753E+01,-0.23904357E+01,
             -0.23165927E+01, 0.27464480E+01,-0.15124305E+01,-0.44464526E+01,
              0.11784693E+00, },
            {-0.12952716E+02, 0.21251139E+02,-0.73948364E+02, 0.71904915E+02,
              0.22982941E+02,-0.17152084E+02, 0.16589590E+02,-0.19711390E+01,
              0.14029673E+00, 0.53307498E+00, 0.45210415E+00,-0.22865257E+01,
              0.31447535E+01, },
            {-0.21541494E+02, 0.48531136E+02,-0.20743677E+02,-0.97520990E+01,
              0.76132256E+02,-0.22496307E+02,-0.30357060E+01, 0.11248891E+02,
             -0.39207585E+01,-0.22813702E+01,-0.63165536E+01, 0.86511211E+01,
              0.12434313E+01, },
            {-0.15622102E+00,-0.59632380E-01,-0.90208650E-01, 0.19414164E-01,
              0.73269002E-01,-0.56129473E+00,-0.63541585E+00,-0.16588993E-01,
             -0.34606468E-01,-0.21938419E-02,-0.21088144E-01,-0.10219838E-01,
              0.62936130E-02, },
            { 0.10229758E+00, 0.31424812E-02, 0.12619580E+00,-0.10478984E-01,
              0.10643311E+00, 0.63880980E+00,-0.58100104E+00,-0.80984796E-03,
             -0.32982638E-02,-0.10336742E-01,-0.12900264E-01,-0.13739295E-01,
              0.22639144E-01, },
            { 0.11596611E+00, 0.24706095E+00,-0.15466779E+00,-0.18783578E+00,
             -0.65919489E-01,-0.62681031E+00,-0.66545194E+00, 0.32444045E-01,
              0.13160099E+00, 0.13321731E-01,-0.10927579E+00, 0.71161732E-01,
              0.11962270E-01, },
            {-0.13758811E+00,-0.70412166E-01, 0.23057573E+00,-0.63733466E-01,
             -0.23197371E-02, 0.77353519E+00,-0.66607153E+00,-0.11856486E+00,
              0.19452100E+00, 0.44829205E-01, 0.53664832E-02, 0.81992827E-01,
             -0.22440353E-01, },
            { 0.17074040E+01, 0.56844890E+00,-0.19279140E+00,-0.48811182E+00,
             -0.82032895E+00, 0.17906553E+01, 0.46447378E+00, 0.62934273E+00,
              0.11496823E+00, 0.72347268E-01, 0.16785286E+00, 0.25202709E+00,
              0.64511701E-01, },
            {-0.21264596E+01, 0.78551662E+00,-0.27990738E+00, 0.20393796E+00,
             -0.17369808E+01,-0.49138665E+00, 0.25081100E+01,-0.13376953E+00,
              0.84815377E+00, 0.67785099E-01,-0.21081494E-01, 0.27407336E+00,
             -0.19572496E+00, },
            { 0.22916576E+00,-0.89228165E+00, 0.14403734E+00, 0.62578565E+00,
              0.42053275E-01,-0.14299740E+00, 0.10011269E+01,-0.10695215E+00,
             -0.26902208E+00, 0.31509314E-01, 0.18015152E+00,-0.22677259E+00,
              0.45944661E-01, },
            { 0.78097403E+00, 0.17563751E+00,-0.79983306E+00,-0.23202594E-01,
              0.51307887E+00,-0.12497675E+01, 0.63095026E-01, 0.27101776E+00,
             -0.64966005E+00,-0.12470666E+00, 0.57328966E-01,-0.25540158E+00,
              0.30125923E-01, },
            {-0.36262932E+01, 0.51153991E-04, 0.65369076E+00, 0.12630241E+01,
              0.59161860E+00,-0.33515153E+01, 0.12459384E+00,-0.92280024E+00,
              0.16300046E+00,-0.98683946E-02,-0.39698192E+00,-0.33663470E+00,
             -0.30247193E-01, },
            { 0.35409751E+01,-0.94228369E+00, 0.59330601E+00,-0.47572304E-01,
              0.26288931E+01,-0.11341671E+00,-0.40985465E+01, 0.55726737E+00,
             -0.15782918E+01,-0.92821464E-01,-0.86877599E-01,-0.41220182E+00,
              0.23814109E+00, },
            { 0.40369980E-01, 0.69174074E-01, 0.90490915E-02, 0.33020664E-01,
              0.11527482E-01, 0.12345885E-01, 0.26705259E-01, 0.19554184E+00,
             -0.36173043E+00, 0.22097254E-01,-0.12866687E-01, 0.62658121E-02,
              0.24714425E-01, },
            { 0.16548369E-01,-0.36080949E-01,-0.25610084E-01,-0.21156788E-01,
              0.43353230E-01,-0.31337328E-01, 0.13042549E-01, 0.34459081E+00,
              0.21506374E+00, 0.16185859E-01,-0.13301040E-01, 0.14696219E-01,
              0.17140709E-01, },
            { 0.11640607E+00,-0.35488162E-01,-0.61050300E-01,-0.10260254E+00,
              0.45410369E-01,-0.78796092E-02,-0.90719406E-02,-0.14850970E+00,
             -0.58185756E+00,-0.61733756E-01,-0.32026162E-02, 0.21957425E-01,
              0.52263454E-01, },
            { 0.55307180E-01,-0.17401587E-01,-0.41585356E-01, 0.76646626E-01,
             -0.57702390E-02,-0.12208957E+00, 0.50012428E-01, 0.47269493E+00,
              0.20059482E-02, 0.50178025E-01,-0.18123774E-02, 0.58514785E-01,
             -0.23311080E-01, },
            { 0.39524496E-01, 0.80552071E-01,-0.42998645E-01, 0.16749263E-01,
              0.31579159E-01,-0.44717076E-02,-0.84887655E-03,-0.49433589E-01,
              0.10765205E-01, 0.21353492E+00, 0.13728055E+00,-0.10284929E-01,
              0.20548493E-01, },
            { 0.99338487E-01, 0.17302408E-02, 0.41451098E-04,-0.84679388E-01,
             -0.11557483E-01,-0.22574198E-01,-0.23921095E-01,-0.82226349E-02,
             -0.69644853E-01,-0.13849451E+00, 0.20419706E+00, 0.16953168E-01,
              0.82117142E-02, },
            { 0.72829023E-01,-0.41823316E-01, 0.10375716E-01,-0.56614611E-01,
              0.41825403E-01, 0.20270178E-01, 0.12032609E-01,-0.24085421E-01,
             -0.12075845E-01, 0.61627743E-02, 0.59546165E-01,-0.69686007E-02,
              0.10623483E+00, },
            { 0.54462738E-02,-0.64947754E-02,-0.79987003E-02,-0.12688862E-01,
             -0.63994959E-01,-0.21709125E-01,-0.43794498E-01, 0.31835452E-01,
             -0.27470827E-01,-0.70910002E-02, 0.91979457E-02,-0.11307489E+00,
             -0.22885030E-01, },
            {-0.13536142E-01, 0.18244112E-01,-0.90564989E-01, 0.45806570E-02,
              0.17037233E-01, 0.58723753E-02,-0.15382156E-01,-0.81376935E-03,
             -0.10111791E-01,-0.13766921E-01, 0.88332891E-02,-0.10245698E-01,
              0.28617415E-02, },
            { 0.21085773E-01, 0.25080677E-01,-0.33543896E-01, 0.24984222E-01,
             -0.45952465E-01,-0.19729603E-01, 0.25285017E-01, 0.50980866E-01,
             -0.13921154E-01,-0.91980677E-02, 0.12159725E-01,-0.17162612E-03,
             -0.10890553E-01, },
            { 0.40087324E-01, 0.82861669E-01,-0.52096929E-01,-0.26000416E-01,
             -0.59353504E-01,-0.32921359E-01, 0.14816161E-01,-0.14302545E-01,
             -0.12839498E-01, 0.24855364E-01,-0.25451362E-01, 0.88601554E-03,
              0.27688909E-01, },
            {-0.54229580E-01, 0.40404364E-01, 0.11072406E+00, 0.53175781E-01,
             -0.27590901E-01, 0.21070372E-01, 0.30208087E-01, 0.12800893E-01,
              0.44277799E-02,-0.28233783E-01, 0.11693147E-02, 0.20908635E-01,
             -0.11176485E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir11Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30831590E+01, 0.36423469E-02, 0.95374268E-02,-0.43797647E-02,
             -0.18526374E-02, 0.12529454E-02,-0.21983865E-02, 0.29568156E-03,
              0.14598125E-02, },
            { 0.46192214E+00, 0.57096808E-03, 0.68593971E-01, 0.65610953E-01,
             -0.10962675E-01,-0.29499030E-01,-0.40161178E-01, 0.90372534E-02,
              0.10792332E-01, },
            { 0.59456217E+00,-0.58928527E-01,-0.10835457E+00, 0.15614104E-01,
              0.38582690E-01, 0.23725567E-01, 0.37873339E-01,-0.45640215E-01,
             -0.25683858E-01, },
            {-0.58345467E+00,-0.27269027E+00,-0.13519897E+00,-0.14928319E+00,
              0.25888463E-01, 0.10878225E+00, 0.17128439E+00,-0.17555457E-01,
             -0.41636076E-01, },
            {-0.47605807E+00, 0.18886384E+00, 0.18850458E+00, 0.15938819E-02,
             -0.82812428E-01,-0.16693853E-01,-0.74811786E-01, 0.13277733E+00,
              0.68222940E-01, },
            { 0.97611375E-01, 0.27176991E+00, 0.73444307E-01, 0.79687804E-01,
             -0.56013255E-02,-0.85544743E-01,-0.14200655E+00, 0.61242115E-02,
              0.28192669E-01, },
            {-0.89809746E-01,-0.11081778E+00,-0.60269386E-01,-0.10322292E-01,
              0.58627542E-01,-0.12565220E-01, 0.40889379E-01,-0.88073589E-01,
             -0.47065701E-01, },
            { 0.50374349E-02,-0.12933061E+00,-0.36687320E+00, 0.13463667E-02,
             -0.39664879E-02, 0.91926754E-02,-0.25788704E-02, 0.42144703E-02,
              0.51479922E-02, },
            { 0.33288829E-01, 0.38233611E+00,-0.14407398E+00, 0.14271198E-01,
              0.11049573E-01,-0.93471622E-02,-0.82618268E-02,-0.22260593E-02,
             -0.10530115E-02, },
            {-0.19392334E-02, 0.48438212E+00, 0.80777749E-01, 0.84854387E-01,
              0.67027077E-01, 0.96795214E-02,-0.77769174E-02,-0.18757645E-01,
             -0.34126833E-02, },
            { 0.10197649E-01,-0.10353541E-01, 0.50584143E+00,-0.11593369E+00,
             -0.44522554E-01, 0.55989962E-01,-0.29334029E-01, 0.31971790E-01,
             -0.19206982E-01, },
            {-0.19130749E+00, 0.67898899E+00, 0.13500415E+01,-0.41417088E-01,
             -0.41389436E-01,-0.11354158E+00, 0.12479115E+00,-0.15671545E+00,
             -0.16941839E+00, },
            {-0.69109064E+00,-0.15827085E+01, 0.77499503E+00,-0.54360222E-01,
             -0.76154426E-01, 0.21804528E-01, 0.12542558E+00, 0.97710431E-01,
              0.15336756E-01, },
            { 0.46886271E+00,-0.25143082E+01, 0.51227117E+00, 0.20739187E+00,
             -0.24356622E+00,-0.41955873E+00,-0.88255458E-01,-0.11599702E+00,
             -0.16374972E+00, },
            { 0.60307807E+00,-0.83648968E+00,-0.31556292E+01, 0.47265437E+00,
              0.62394643E+00,-0.54705179E+00, 0.20412104E+00,-0.19938691E+00,
              0.61526425E-01, },
            { 0.66733706E+00,-0.84022284E+00,-0.36363804E+00,-0.47948975E-01,
              0.91792926E-01, 0.36664313E+00,-0.50559223E+00, 0.43860993E+00,
              0.55686808E+00, },
            { 0.20355382E+01, 0.71406555E+00,-0.12435471E+01, 0.21483092E+00,
             -0.53310115E-01,-0.59347369E-01,-0.33315611E+00,-0.25482255E+00,
             -0.24464151E-02, },
            {-0.78910613E+00, 0.37705996E+01,-0.86769462E+00,-0.12256935E+01,
              0.33561519E+00, 0.13111099E+01, 0.17731486E+00, 0.52530521E+00,
              0.42007926E+00, },
            {-0.23053181E+01, 0.16463206E+01, 0.60083599E+01,-0.67770940E+00,
             -0.16302882E+01, 0.13140205E+01,-0.39495325E+00, 0.35773265E+00,
             -0.20983079E-01, },
            {-0.48089162E+00, 0.39525765E+00,-0.48140696E+00, 0.16104813E+00,
             -0.11935355E-01,-0.33358160E+00, 0.43848920E+00,-0.33588856E+00,
             -0.45440468E+00, },
            {-0.16640769E+01, 0.70364147E+00, 0.72135723E+00,-0.26408896E+00,
              0.13320316E+00, 0.60090333E-01, 0.20128188E+00, 0.18900408E+00,
             -0.34622148E-01, },
            { 0.22574455E+00,-0.15068560E+01, 0.90492386E+00, 0.10402929E+01,
             -0.16921978E+00,-0.10404446E+01,-0.66429354E-01,-0.44397834E+00,
             -0.24560237E+00, },
            { 0.19569390E+01,-0.13762980E+01,-0.33979950E+01, 0.37711099E+00,
              0.11398213E+01,-0.88892829E+00, 0.24327640E+00,-0.19420274E+00,
             -0.70902784E-02, },
            {-0.20724727E-01,-0.35504639E-01,-0.82354777E-03, 0.36352221E-01,
             -0.11773663E+00,-0.11033534E-01, 0.95688127E-03,-0.97791515E-02,
             -0.58923671E-02, },
            { 0.11510791E-01,-0.53077191E-02,-0.21679617E-01, 0.12743253E+00,
              0.43944508E-01,-0.99055991E-02,-0.21604956E-02,-0.19020389E-02,
              0.15028869E-02, },
            { 0.34215365E-01, 0.19715223E+00, 0.12483910E+00,-0.11022672E+00,
             -0.31943321E-01, 0.11501916E-02,-0.29224209E-01,-0.21727003E-01,
             -0.48656184E-01, },
            {-0.12077173E+00, 0.44407237E-01,-0.50276604E-01, 0.74237496E-01,
             -0.13876393E+00, 0.50061855E-01,-0.89441317E-04,-0.19618968E-01,
              0.77808690E-02, },
            { 0.11498174E+00, 0.30138892E+00, 0.84145188E-01, 0.52387480E-01,
             -0.16962323E+00, 0.13028961E-01, 0.83306350E-01, 0.12248267E+00,
             -0.22928387E-01, },
            {-0.18141045E+00,-0.73630512E-01, 0.12350223E-01, 0.22824010E+00,
              0.88408612E-01,-0.56969866E-01, 0.44197991E-01, 0.42088978E-01,
             -0.97412907E-01, },
            {-0.15992832E+00,-0.10725375E+01,-0.37769392E+00, 0.20365877E+00,
             -0.10712838E+00,-0.52147180E-01, 0.11348375E+00, 0.12272292E+00,
              0.21829335E+00, },
            { 0.90741438E+00, 0.27918214E+00, 0.21284491E+00,-0.16413610E+00,
              0.20872620E+00,-0.26607731E-01,-0.13638993E+00, 0.10097218E+00,
             -0.90743758E-01, },
            {-0.16678724E+00,-0.49267471E+00,-0.25228688E+00,-0.20995039E+00,
              0.35344949E+00,-0.42392898E-01,-0.14291544E+00,-0.18131788E+00,
              0.37399542E-01, },
            { 0.22921523E+00,-0.11525336E-01, 0.10362252E+00,-0.50349134E+00,
             -0.24221870E+00, 0.16809084E+00,-0.94889142E-01,-0.70770048E-01,
              0.13145503E+00, },
            { 0.21850903E+00, 0.14192160E+01, 0.37012824E+00,-0.17118199E-01,
              0.49541610E+00, 0.17686056E+00,-0.86773776E-01,-0.13545221E+00,
             -0.20977929E+00, },
            {-0.10618936E+01,-0.52702361E+00,-0.14100182E+00,-0.16452017E+00,
              0.87792031E-01,-0.11034585E+00, 0.22220004E+00,-0.97200051E-01,
              0.10797976E+00, },
            {-0.20363385E-01,-0.53697578E-02, 0.68575093E-02, 0.68112370E-02,
              0.10468053E-01, 0.19557819E-01,-0.29966215E-01, 0.26615837E-02,
             -0.39665838E-03, },
            {-0.27748954E-02,-0.12629431E-01,-0.34522379E-02,-0.19603441E-01,
             -0.74741081E-02, 0.36278360E-01, 0.38917966E-01, 0.62133339E-02,
              0.52181524E-02, },
            { 0.47905695E-01, 0.42130247E-01, 0.10669809E-01,-0.42659171E-01,
             -0.48365989E-02,-0.12414151E+00, 0.83759082E-02, 0.88774720E-02,
             -0.24281682E-02, },
            { 0.65088286E-02,-0.18100817E-01,-0.99380631E-02,-0.49212319E-02,
              0.54735452E-03,-0.14428278E-01,-0.12441483E+00,-0.92067998E-02,
             -0.31998189E-03, },
            { 0.24574004E-01,-0.10163294E+00, 0.31056121E-01, 0.11012295E-01,
             -0.45570403E-01,-0.14161536E+00,-0.41898366E-01, 0.22085367E-02,
             -0.13771434E-01, },
            {-0.37635207E-01,-0.79316162E-02,-0.12239707E-01, 0.54416325E-01,
             -0.23432048E-01, 0.64405173E-01,-0.15231898E+00,-0.39993403E-02,
              0.10142059E-01, },
            {-0.18101418E-01,-0.84799835E-02, 0.34984993E-03, 0.89346385E-02,
              0.47468557E-03, 0.28572041E-02,-0.35473125E-02, 0.14134602E-01,
             -0.84402598E-02, },
            {-0.24490006E-01, 0.59575271E-02,-0.83975289E-02,-0.10783280E-01,
              0.10519189E-01, 0.37295434E-02, 0.78448253E-02, 0.11155130E-01,
              0.20005925E-01, },
            { 0.62797785E-01, 0.11900892E-01,-0.27039561E-01,-0.19209160E-01,
              0.50548892E-02,-0.49631188E-04,-0.94857474E-03,-0.61566597E-02,
             -0.72145365E-01, },
            { 0.32358173E-01,-0.47734085E-01, 0.26209857E-01, 0.28988877E-01,
              0.74443221E-02,-0.39548054E-02, 0.37050038E-02, 0.76541923E-01,
             -0.17525801E-02, },
            { 0.52079731E-02,-0.92345122E-02, 0.14466225E-02,-0.19333605E-02,
             -0.76571708E-02, 0.36316388E-02,-0.43371287E-02,-0.36561657E-02,
              0.67056152E-02, },
            { 0.16525778E-02, 0.18420495E-01,-0.17318184E-01,-0.72389068E-02,
             -0.36425991E-02, 0.67126690E-02, 0.62284856E-02,-0.25143016E-02,
             -0.54201265E-02, },
            {-0.49585272E-02, 0.72677066E-02, 0.95763793E-02, 0.33443144E-02,
              0.68702078E-02, 0.38347980E-02,-0.47801477E-02,-0.56901621E-02,
              0.15504239E-02, },
            {-0.65200361E-02,-0.13484796E-01,-0.16534761E-01,-0.93058543E-02,
             -0.70031788E-02, 0.61732728E-03, 0.68263346E-02, 0.85278694E-02,
              0.13020536E-03, },
         },
         {
            { 0.26661365E+01,-0.11117877E-02, 0.34222321E-02,-0.20794710E-02,
              0.24010155E-02,-0.40831408E-02,-0.47719344E-02,-0.29675064E-02,
             -0.20296483E-02, },
            { 0.41117510E+00, 0.10291318E-01, 0.47319636E-01, 0.31215122E-01,
              0.78935334E-02,-0.25743436E-01,-0.38405076E-01, 0.98825051E-02,
              0.67149191E-02, },
            { 0.74506956E+00, 0.48187740E-01,-0.71195945E-01, 0.25079396E-01,
             -0.21259580E-01, 0.67766912E-01, 0.48439432E-01,-0.66096769E-02,
              0.20697504E-01, },
            {-0.39045444E+00,-0.26972118E+00,-0.70841588E-01,-0.73414013E-01,
             -0.12673208E-01, 0.81827506E-01, 0.14244869E+00,-0.26457297E-01,
             -0.33021890E-01, },
            {-0.66143775E+00,-0.15479179E+00, 0.11289432E+00,-0.54646458E-01,
              0.28970562E-01,-0.14369094E+00,-0.10134431E+00, 0.25818255E-01,
             -0.48019908E-01, },
            { 0.53726800E-01, 0.25038397E+00, 0.37336465E-01, 0.40437397E-01,
              0.14270316E-01,-0.59865799E-01,-0.11061210E+00, 0.18749857E-01,
              0.24053220E-01, },
            {-0.40976707E-01, 0.14224784E+00,-0.30323656E-01, 0.30423727E-01,
             -0.44295667E-02, 0.87912515E-01, 0.58721401E-01,-0.13916852E-01,
              0.28919090E-01, },
            { 0.74064271E-02,-0.27877745E+00,-0.28779912E+00, 0.39281864E-02,
             -0.14419556E-02,-0.11547093E-02, 0.31535963E-02,-0.72778366E-03,
              0.34911863E-02, },
            { 0.19075764E-01, 0.28820461E+00,-0.28949183E+00, 0.22228628E-01,
              0.63713072E-02,-0.21684170E-02,-0.55248677E-02,-0.13032343E-02,
              0.91081448E-02, },
            {-0.81327841E-01, 0.27268976E+00, 0.13500665E+00, 0.54546636E-01,
              0.37922066E-01,-0.32600060E-01,-0.10949604E-01, 0.19963372E-01,
              0.26835289E-01, },
            { 0.87298211E-02,-0.22350943E-01, 0.29820484E+00, 0.20726329E-01,
             -0.52162141E-01, 0.49364645E-01,-0.39931941E-02, 0.40589698E-01,
             -0.22952456E-02, },
            {-0.10790928E+00, 0.83676583E+00, 0.63060778E+00,-0.68363361E-01,
             -0.99837855E-02,-0.66430098E-02,-0.12207449E-01,-0.76347262E-01,
             -0.15965886E+00, },
            {-0.35241818E+00,-0.65478712E+00, 0.11052341E+01,-0.12858538E+00,
              0.48155166E-01, 0.30722397E-02, 0.79846002E-01, 0.70853211E-01,
             -0.98708250E-01, },
            { 0.10073023E+01,-0.17845337E+01,-0.80092168E+00,-0.46188459E-01,
             -0.14296983E+00, 0.64327158E-01,-0.35780571E-01,-0.21253628E+00,
             -0.38209790E+00, },
            { 0.28986445E+00, 0.20012251E+00,-0.21629679E+01,-0.28332931E+00,
              0.53812021E+00,-0.36693877E+00, 0.39688271E-01,-0.34019363E+00,
             -0.44371117E-01, },
            { 0.21445227E+00,-0.62239212E+00, 0.50726289E+00, 0.12845744E+00,
             -0.50081648E-01, 0.10264898E-01,-0.10842126E-01, 0.25324762E+00,
              0.47339380E+00, },
            { 0.98013788E+00,-0.77179390E+00,-0.15949878E+01, 0.26540795E+00,
             -0.23207283E+00,-0.71822710E-01,-0.27238446E+00,-0.25130799E+00,
              0.26421374E+00, },
            {-0.21762612E+01, 0.31028821E+01, 0.20128744E+01,-0.15856740E+00,
              0.15671612E+00,-0.83239703E-02, 0.12464356E+00, 0.52468282E+00,
              0.88304269E+00, },
            {-0.13928696E+01,-0.54484439E+00, 0.43236532E+01, 0.72512794E+00,
             -0.13013941E+01, 0.80244911E+00,-0.15787378E+00, 0.71791559E+00,
              0.87640740E-01, },
            {-0.90506822E-01, 0.29343450E+00,-0.64716011E+00,-0.85126469E-02,
              0.86700074E-01,-0.19545069E-01, 0.22539426E-01,-0.21825901E+00,
             -0.35119224E+00, },
            {-0.77049851E+00, 0.13025295E+01, 0.10195137E+01,-0.17070553E+00,
              0.19100779E+00, 0.10148965E+00, 0.19473387E+00, 0.22605897E+00,
             -0.21595261E+00, },
            { 0.12541903E+01,-0.14151653E+01,-0.76735270E+00, 0.20479736E+00,
             -0.57303514E-01,-0.61315097E-01,-0.67472473E-01,-0.35682014E+00,
             -0.55211598E+00, },
            { 0.12503939E+01,-0.19497246E+00,-0.25232329E+01,-0.48834419E+00,
              0.95700103E+00,-0.50639701E+00, 0.15538096E+00,-0.43736929E+00,
             -0.24403557E-02, },
            { 0.58859168E-02,-0.32317564E-01, 0.18789638E-02, 0.81562921E-01,
             -0.10188310E+00,-0.16970166E-02, 0.10929941E-01,-0.80213174E-02,
             -0.64339088E-02, },
            { 0.12977694E-01,-0.20233879E-01,-0.87966016E-02, 0.10940550E+00,
              0.84022991E-01,-0.43360288E-02, 0.49938275E-02,-0.24055217E-02,
              0.34415221E-02, },
            { 0.13273510E-01, 0.11690062E+00, 0.11301346E+00,-0.53855572E-01,
              0.71545341E-02, 0.26712853E-02,-0.65927980E-02,-0.43166103E-02,
             -0.16314678E-01, },
            {-0.76610744E-01, 0.31374011E-01,-0.51425263E-01, 0.29726738E-01,
             -0.10468841E+00, 0.35839197E-02, 0.31228568E-01,-0.15449940E-02,
              0.10545071E-01, },
            { 0.31877279E-01, 0.21496819E+00, 0.69084771E-01,-0.39078769E+00,
              0.10721206E+00,-0.50794119E-02,-0.28492521E-01, 0.88592894E-01,
              0.17780269E-01, },
            {-0.14039850E+00,-0.13015021E-01,-0.97526431E-01,-0.13987683E+00,
             -0.35487783E+00,-0.44129573E-01, 0.44898996E-02, 0.47013644E-01,
             -0.73569492E-01, },
            {-0.59302974E-01,-0.65177011E+00,-0.36309096E+00,-0.23424511E+00,
             -0.13430981E+00,-0.32746039E-01, 0.60699228E-01, 0.33282310E-01,
              0.85893124E-01, },
            { 0.49057353E+00, 0.21044838E+00, 0.23838104E+00, 0.10888345E-01,
             -0.39147530E-01, 0.10908667E+00,-0.18488690E+00, 0.53481314E-01,
             -0.13602662E+00, },
            {-0.10266943E+00,-0.34125081E+00,-0.22209170E+00, 0.21323264E+00,
             -0.73764898E-01,-0.42999983E-01, 0.28910641E-01,-0.12368568E+00,
             -0.24522578E-01, },
            { 0.17222378E+00,-0.45565147E-01, 0.26528457E+00, 0.10084933E+00,
              0.18232718E+00, 0.10636824E+00,-0.40960647E-01,-0.71175881E-01,
              0.91745026E-01, },
            { 0.48435759E-01, 0.88375968E+00, 0.36494932E+00, 0.37522808E+00,
              0.38870537E+00, 0.11076623E+00,-0.87760679E-01,-0.44668164E-01,
             -0.12374668E+00, },
            {-0.56823593E+00,-0.31550634E+00,-0.23784252E+00,-0.28017625E+00,
              0.20619076E+00,-0.18672663E+00, 0.23123896E+00,-0.66964269E-01,
              0.16415413E+00, },
            {-0.73595000E-02,-0.61682197E-02, 0.89407256E-02, 0.50872318E-02,
              0.11877637E-01, 0.35297502E-01,-0.39307773E-01, 0.14423406E-02,
              0.44433944E-03, },
            {-0.74655698E-04,-0.16662820E-02,-0.57203509E-02,-0.14137458E-01,
             -0.46855919E-02, 0.45804936E-01, 0.35633929E-01, 0.30760856E-02,
             -0.23319600E-02, },
            { 0.12941354E-01, 0.44963390E-01, 0.54211006E-02,-0.12041884E-01,
             -0.95001562E-02,-0.11241351E+00,-0.19248098E-01, 0.27269039E-02,
             -0.30035619E-02, },
            { 0.59964410E-02,-0.13195523E-02,-0.46101669E-02, 0.39474596E-02,
             -0.64694295E-02, 0.15680529E-01,-0.11271558E+00,-0.32720950E-02,
             -0.51417295E-02, },
            { 0.48678972E-01,-0.73544301E-01, 0.30441817E-01, 0.96263876E-02,
             -0.39520711E-01,-0.17735021E+00,-0.45016337E-01, 0.51114690E-02,
             -0.11839547E-01, },
            {-0.81728892E-02,-0.23926420E-01, 0.60986411E-02, 0.40335935E-01,
             -0.35920464E-02, 0.41258320E-01,-0.14260805E+00,-0.34774819E-02,
              0.18476740E-01, },
            {-0.10441234E-01,-0.13517550E-01, 0.34676618E-02,-0.14249106E-02,
              0.14872112E-02,-0.32633506E-02,-0.21993518E-02, 0.17342664E-01,
             -0.17490787E-02, },
            {-0.15205597E-01, 0.49438239E-02,-0.87333824E-02,-0.87372363E-02,
              0.14296352E-02, 0.37051606E-03, 0.55464786E-02, 0.15304491E-03,
              0.21174494E-01, },
            { 0.39107330E-01, 0.80265282E-02,-0.22919752E-01,-0.76330574E-02,
             -0.77880705E-02,-0.70988761E-04, 0.10868077E-01, 0.58998396E-02,
             -0.64436145E-01, },
            { 0.41344304E-01,-0.33050463E-01, 0.35609432E-01, 0.15621077E-01,
              0.52525150E-02,-0.76148482E-02, 0.10637491E-01, 0.63588083E-01,
              0.85605606E-02, },
            { 0.37736224E-02,-0.12569182E-02,-0.73067157E-03,-0.55869166E-02,
             -0.57651363E-02,-0.35552061E-03,-0.52995873E-02,-0.63682459E-02,
              0.34755855E-02, },
            { 0.33890572E-02, 0.20352295E-01,-0.10258360E-01,-0.16343703E-02,
             -0.76323333E-02, 0.61682411E-02, 0.74977572E-02,-0.10669790E-02,
             -0.31994097E-02, },
            {-0.56459666E-02, 0.10195384E-01, 0.95578721E-02, 0.39921715E-02,
              0.18846111E-02,-0.28111471E-03,-0.14161777E-03,-0.63958345E-02,
             -0.16010746E-03, },
            { 0.83076116E-02,-0.32676375E-02,-0.81046103E-02,-0.20230603E-02,
             -0.86054020E-02, 0.55735982E-02, 0.49315910E-02, 0.46230308E-02,
              0.80582756E-03, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir12F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.58852773E+01,-0.96291333E-01,-0.24346260E-02, 0.59140641E-01,
             -0.63641076E-02,-0.26449412E-02,-0.31942986E-01,-0.10096008E-02,
              0.14919642E-01,-0.31421110E-02,-0.46499600E-02,-0.67822053E-02,
              0.15032781E-01, },
            { 0.16458440E+01,-0.18067592E+00, 0.30474961E+00,-0.21605857E+00,
             -0.49374822E+00,-0.62353790E-01,-0.34158051E-01,-0.12912387E+00,
              0.10613752E+00, 0.29181531E-01,-0.23050727E-01, 0.40541522E-01,
             -0.11224715E+00, },
            { 0.10501716E+02, 0.15741066E+01,-0.14308903E+01,-0.48761311E+00,
              0.94652450E+00,-0.30434468E+00, 0.10839933E+01,-0.10534965E+00,
             -0.51447499E+00,-0.91966093E-01, 0.29239580E+00, 0.20149432E-01,
             -0.39589244E+00, },
            {-0.23581297E+02, 0.36677198E+01,-0.80632642E-01, 0.43846774E+01,
              0.81138687E+01, 0.20209019E+01, 0.32430317E+01, 0.27812088E+01,
             -0.52667046E+00, 0.26251921E+00, 0.75711900E+00,-0.62242740E+00,
              0.95958692E+00, },
            {-0.55303654E+02,-0.63127465E+01, 0.83682976E+01,-0.49186325E+01,
             -0.69345760E+01, 0.62089677E+01,-0.60723081E+01, 0.13838018E+01,
              0.38213999E+01,-0.14319529E+00,-0.18019359E+01, 0.50669366E+00,
              0.23809679E+01, },
            { 0.59686283E+02,-0.30755751E+02,-0.12747594E+02,-0.16911802E+02,
             -0.32807728E+02,-0.12712322E+02,-0.21192408E+02,-0.12855251E+02,
             -0.71364388E-01,-0.14239579E+01,-0.39535253E+01, 0.33811421E+01,
             -0.32126002E+01, },
            { 0.87687027E+02, 0.13251464E+02,-0.15531278E+02, 0.25012060E+02,
              0.18020908E+02,-0.21276220E+02, 0.13908660E+02,-0.47290397E+01,
             -0.11022418E+02, 0.16285949E+01, 0.42395010E+01,-0.24118855E+01,
             -0.54046264E+01, },
            {-0.23993408E+02, 0.89517502E+02, 0.42279510E+02, 0.22479494E+02,
              0.54392288E+02, 0.30301125E+02, 0.50255486E+02, 0.23656767E+02,
              0.32564282E+01, 0.28625791E+01, 0.82307177E+01,-0.78404694E+01,
              0.53785987E+01, },
            {-0.66045982E+02,-0.16009594E+02, 0.10073885E+02,-0.35849022E+02,
             -0.18812138E+02, 0.25561331E+02,-0.14358003E+02, 0.62677326E+01,
              0.13022478E+02,-0.24407010E+01,-0.45216794E+01, 0.35019362E+01,
              0.52684317E+01, },
            {-0.57186363E+02,-0.10306866E+03,-0.47426033E+02,-0.90657454E+01,
             -0.39237549E+02,-0.30380714E+02,-0.50804012E+02,-0.19044092E+02,
             -0.49842196E+01,-0.29108648E+01,-0.74697671E+01, 0.79306436E+01,
             -0.42882524E+01, },
            { 0.20778374E+02, 0.76332736E+01,-0.15185139E+01, 0.16232574E+02,
              0.66888428E+01,-0.10187649E+02, 0.54846730E+01,-0.28345735E+01,
             -0.53408065E+01, 0.10432419E+01, 0.18286176E+01,-0.16316253E+01,
             -0.18638300E+01, },
            { 0.42999359E+02, 0.40907448E+02, 0.17649250E+02,-0.67038941E+00,
              0.99691086E+01, 0.10768867E+02, 0.18582966E+02, 0.55952592E+01,
              0.21914599E+01, 0.12064843E+01, 0.24707305E+01,-0.28870997E+01,
              0.12680758E+01, },
            { 0.10030658E+00, 0.17061027E+01, 0.14686522E+01, 0.67665540E-01,
              0.21641403E-01, 0.18892784E-01, 0.10023636E-01,-0.12924850E-01,
             -0.28599406E-01,-0.64662793E-02, 0.11706679E-01,-0.93984343E-02,
              0.17487410E-01, },
            {-0.23059970E+00,-0.14013827E+01, 0.18092790E+01,-0.91769502E-01,
              0.27110616E-01, 0.17508439E-02,-0.26233229E-02, 0.22305546E-01,
              0.89679928E-02, 0.32947335E-01, 0.14318242E-01, 0.24811620E-01,
             -0.15941351E-02, },
            {-0.90277523E+00, 0.23125539E+01, 0.16092964E+01,-0.14453238E+00,
             -0.20665842E+00, 0.51915783E+00, 0.49450490E+00,-0.19916792E+00,
              0.18426664E+00, 0.25234304E-01, 0.38722817E-01,-0.42649661E-02,
             -0.26182031E-01, },
            {-0.56744313E+00, 0.44611600E+00, 0.11730177E+01,-0.47925121E+00,
              0.65121371E+00,-0.45957804E-01, 0.45501205E+00, 0.25943026E+00,
              0.15283296E+00, 0.99556014E-01,-0.21310527E-01, 0.37541655E+00,
             -0.42286009E-01, },
            { 0.20739954E+01, 0.21575235E+02, 0.34665103E+01,-0.20396402E+01,
              0.26834292E+01, 0.28675493E-01, 0.42604324E+00, 0.28829846E+00,
              0.78976136E+00, 0.17124640E-01,-0.31698342E-01, 0.52003258E+00,
             -0.44703317E+00, },
            {-0.10670326E+00,-0.74782805E+01, 0.15177947E+02, 0.26179512E+01,
              0.20770681E+01,-0.10762349E+01,-0.44632065E+00, 0.60240370E+00,
             -0.58553094E+00,-0.10602016E+01,-0.88225491E-01,-0.11664101E+01,
             -0.34999716E+00, },
            { 0.25958813E+02,-0.27454485E+02,-0.54103351E+01, 0.15304747E+01,
              0.93126850E+01,-0.87061596E+01,-0.62785430E+01, 0.45617943E+01,
             -0.37337399E+01,-0.49913475E+00, 0.41627291E+00,-0.20554547E+00,
              0.59320398E-01, },
            { 0.14931066E+02,-0.24322105E+02,-0.18978451E+02, 0.66896195E+01,
             -0.11191239E+02, 0.25101948E+01,-0.91349773E+01,-0.48819108E+01,
             -0.36994448E+01, 0.46167007E+00, 0.15925080E+01,-0.77618051E+01,
             -0.19510709E+01, },
            {-0.18649813E+02,-0.15214691E+03, 0.53326965E+02, 0.57327814E+01,
             -0.26576279E+02,-0.19274548E+00, 0.15754475E+00, 0.28831229E+01,
             -0.84430466E+01,-0.36592329E+00, 0.27528572E+01,-0.42047176E+01,
              0.30515225E+01, },
            { 0.14867091E+02,-0.28944750E+02,-0.87118622E+02,-0.89018812E+01,
             -0.27807983E+02, 0.10147781E+02, 0.44683981E+01,-0.29643438E+01,
              0.63139424E+01, 0.88454924E+01,-0.19191800E+01, 0.97981710E+01,
              0.39162433E+01, },
            {-0.15608711E+03, 0.10050726E+03, 0.31733864E+02, 0.67168102E+01,
             -0.41432091E+02, 0.47361160E+02, 0.28008722E+02,-0.28612593E+02,
              0.24360193E+02, 0.73379183E+01,-0.59432845E+01, 0.31006491E+01,
              0.61421555E+00, },
            {-0.10763945E+03, 0.12835529E+03, 0.69382744E+02,-0.43361359E+02,
              0.56389217E+02,-0.13272401E+02, 0.51118816E+02, 0.27908688E+02,
              0.29593452E+02,-0.90813322E+01,-0.13741107E+02, 0.45609165E+02,
              0.17076014E+02, },
            { 0.49405956E+02, 0.36311176E+03,-0.26420923E+03, 0.41176744E+01,
              0.81804565E+02,-0.21720428E+01,-0.76053777E+01,-0.19046164E+02,
              0.28341925E+02, 0.31033659E+01,-0.15050348E+02, 0.12041607E+02,
             -0.86199579E+01, },
            {-0.58249134E+02, 0.21876076E+03, 0.15515247E+03, 0.21008591E+02,
              0.99964432E+02,-0.32595772E+02,-0.15503604E+02, 0.48927865E+01,
             -0.20159599E+02,-0.26689791E+02, 0.95943127E+01,-0.29238544E+02,
             -0.13632691E+02, },
            { 0.39192972E+03,-0.18186565E+03,-0.12129234E+03,-0.57435074E+02,
              0.56430367E+02,-0.11213171E+03,-0.54315083E+02, 0.70347794E+02,
             -0.67493736E+02,-0.30095579E+02, 0.18334648E+02,-0.94074984E+01,
             -0.31521268E+01, },
            { 0.27239188E+03,-0.28597064E+03,-0.12101062E+03, 0.12336749E+03,
             -0.12470569E+03, 0.26024103E+02,-0.12104427E+03,-0.59888294E+02,
             -0.85995743E+02, 0.29719172E+02, 0.37946323E+02,-0.11199417E+03,
             -0.46891853E+02, },
            {-0.50657452E+02,-0.37947943E+03, 0.38515732E+03,-0.23199799E+02,
             -0.10095910E+03, 0.50711770E+01, 0.14311047E+02, 0.33573044E+02,
             -0.37079788E+02,-0.60635815E+01, 0.25830387E+02,-0.14170277E+02,
              0.10315926E+02, },
            { 0.79237267E+02,-0.35955161E+03,-0.10834153E+03,-0.33607841E+02,
             -0.13744238E+03, 0.42688866E+02, 0.22823519E+02,-0.24746652E+01,
              0.25041647E+02, 0.33539783E+02,-0.14638429E+02, 0.35408268E+02,
              0.18950537E+02, },
            {-0.44368756E+03, 0.16978250E+03, 0.19025841E+03, 0.10587995E+03,
             -0.16499306E+02, 0.12066158E+03, 0.45466454E+02,-0.74402390E+02,
              0.81906021E+02, 0.44979858E+02,-0.21132385E+02, 0.10045572E+02,
              0.49591422E+01, },
            {-0.27212534E+03, 0.29296640E+03, 0.10778039E+03,-0.14709779E+03,
              0.12872285E+03,-0.21798769E+02, 0.12947227E+03, 0.51874657E+02,
              0.10150548E+03,-0.36791374E+02,-0.41881748E+02, 0.12185918E+03,
              0.52135880E+02, },
            { 0.17340685E+02, 0.14838107E+03,-0.17882977E+03, 0.15429540E+02,
              0.43413673E+02,-0.26469874E+01,-0.74090319E+01,-0.18376600E+02,
              0.16659889E+02, 0.33483205E+01,-0.13998299E+02, 0.58607697E+01,
             -0.43770027E+01, },
            {-0.36554958E+02, 0.17909183E+03, 0.23982536E+02, 0.19982359E+02,
              0.63960342E+02,-0.19503317E+02,-0.11998042E+02,-0.21119624E+00,
             -0.10841630E+02,-0.14879585E+02, 0.71827703E+01,-0.14873817E+02,
             -0.91093102E+01, },
            { 0.18463353E+03,-0.63689983E+02,-0.10000366E+03,-0.59487541E+02,
             -0.83337955E+01,-0.48281857E+02,-0.12824591E+02, 0.28396828E+02,
             -0.35691708E+02,-0.22169371E+02, 0.81618071E+01,-0.34571950E+01,
             -0.25228517E+01, },
            { 0.91842926E+02,-0.11355025E+03,-0.39262459E+02, 0.60758072E+02,
             -0.51565228E+02, 0.64481277E+01,-0.51903679E+02,-0.14617392E+02,
             -0.41567421E+02, 0.15849867E+02, 0.16073263E+02,-0.48541443E+02,
             -0.20367502E+02, },
            { 0.33230793E-01,-0.53760577E-02,-0.43271579E-01,-0.82253754E+00,
              0.10927148E+00,-0.62830341E-02, 0.32827888E-01, 0.20675356E-02,
              0.11074953E-01, 0.14030251E-01,-0.93817338E-02, 0.58153598E-02,
             -0.90415105E-02, },
            {-0.72160540E-02, 0.44827171E-01, 0.94338991E-01,-0.15674749E+00,
             -0.88599294E+00, 0.21040715E-01, 0.11656060E-01,-0.43813023E-02,
              0.28248457E-02, 0.18651808E-01, 0.15717305E-01, 0.10290854E-01,
             -0.74585471E-02, },
            { 0.32696807E+00, 0.55384904E+00,-0.20601006E+00, 0.22753012E+00,
             -0.38133556E+00, 0.28293085E+00,-0.22165217E+00,-0.19600345E+00,
             -0.30326167E+00,-0.53036101E-01,-0.64323887E-01,-0.25258502E-01,
              0.27758265E-01, },
            { 0.84868371E+00, 0.71870285E+00, 0.10409747E+00,-0.34669647E+00,
              0.68305522E-01, 0.17950530E+00, 0.17103158E+00,-0.16034061E+00,
              0.11909673E+00,-0.90939038E-01, 0.77516790E-02, 0.50704526E-02,
             -0.44594165E-01, },
            { 0.19773264E+01,-0.63938504E+00, 0.54355240E+00, 0.34600697E+01,
              0.14479409E+01,-0.65015870E+00,-0.12161551E+01, 0.79876208E+00,
             -0.60558927E+00, 0.30918708E+00, 0.64186800E+00, 0.23445684E+00,
              0.25567999E+00, },
            {-0.23858164E+00,-0.21537189E+01,-0.21538348E+00, 0.19827148E+00,
              0.62378974E+01,-0.14610817E+01,-0.79967242E+00, 0.97321606E+00,
              0.28241432E+00,-0.71557999E+00,-0.40822551E-01,-0.35631898E+00,
             -0.30096257E+00, },
            { 0.27642247E+00,-0.35841665E+01, 0.20953624E+01, 0.21262348E+01,
              0.82936811E+01,-0.25825257E+01, 0.12673424E+01, 0.22965710E+01,
              0.32046340E+01, 0.59451181E+00, 0.29327148E+00, 0.60926776E-01,
             -0.18911134E+00, },
            {-0.77316437E+01,-0.86913109E+01, 0.70329523E+00, 0.29154107E+00,
              0.37658503E+01,-0.15267447E+01,-0.19037495E+01, 0.18705777E+01,
             -0.38128939E+00, 0.16374694E+01, 0.94264515E-01,-0.12111475E+01,
              0.54479063E+00, },
            {-0.11841908E+02, 0.12982108E+02,-0.46878176E+01, 0.11409554E+02,
              0.11456189E+02, 0.28962121E+01, 0.70374961E+01,-0.53894963E+01,
              0.32251689E+01,-0.26807201E+01,-0.33563082E+01,-0.14099928E+01,
             -0.16444161E+01, },
            { 0.45781121E+01, 0.22667099E+02, 0.25675006E+01,-0.19983486E+02,
             -0.88170242E+00, 0.60731277E+01, 0.46235948E+01,-0.60956397E+01,
             -0.36376829E+01, 0.50197077E+01, 0.35221526E+00, 0.22886684E+01,
              0.17762291E+01, },
            {-0.40170121E+01, 0.62177038E+01,-0.81950235E+01,-0.85565033E+01,
             -0.26114573E+02, 0.52662058E+01,-0.33671968E+01,-0.57750220E+01,
             -0.89394646E+01,-0.96039099E+00,-0.44176674E+00,-0.38639593E+00,
              0.38240880E+00, },
            { 0.16040434E+02, 0.22648796E+02,-0.38160076E+01, 0.32987599E+01,
             -0.11050160E+02, 0.41704860E+01, 0.42172189E+01,-0.41252751E+01,
              0.30131927E+00,-0.52030268E+01,-0.83111125E+00, 0.49567075E+01,
             -0.11959333E+01, },
            { 0.23355219E+02,-0.34282352E+02, 0.11538242E+02,-0.49848263E+02,
             -0.40598743E+02,-0.39050004E+01,-0.14526564E+02, 0.10634094E+02,
             -0.49407711E+01, 0.58253703E+01, 0.57160439E+01, 0.24022803E+01,
              0.39250429E+01, },
            {-0.89171867E+01,-0.57658352E+02,-0.79275866E+01, 0.56511944E+02,
             -0.33630772E+02,-0.73403568E+01,-0.97860651E+01, 0.12670938E+02,
              0.95948782E+01,-0.10938619E+02,-0.13207883E+01,-0.39759080E+01,
             -0.31204889E+01, },
            { 0.31317453E+01,-0.26248660E+01, 0.75624285E+01, 0.10054381E+02,
              0.23560154E+02,-0.29456441E+01, 0.29910414E+01, 0.40049529E+01,
              0.69776058E+01, 0.23051453E+00, 0.35383257E+00, 0.45363480E+00,
             -0.17931676E+00, },
            {-0.85573292E+01,-0.16594076E+02, 0.34918075E+01,-0.66410427E+01,
              0.96621037E+01,-0.31863921E+01,-0.24729841E+01, 0.26573007E+01,
              0.11516500E+00, 0.40185380E+01, 0.10184885E+01,-0.46337085E+01,
              0.70501119E+00, },
            {-0.15520618E+02, 0.24251535E+02,-0.78417487E+01, 0.39622379E+02,
              0.31915640E+02, 0.84966403E+00, 0.10047779E+02,-0.63451653E+01,
              0.19262159E+01,-0.35821593E+01,-0.31948893E+01,-0.13733149E+01,
             -0.27612700E+01, },
            { 0.26133041E+01, 0.42539505E+02, 0.70027027E+01,-0.41647896E+02,
              0.34347599E+02, 0.24487431E+01, 0.63871117E+01,-0.86457167E+01,
             -0.66765871E+01, 0.73050599E+01, 0.12362036E+01, 0.21852674E+01,
              0.16885310E+01, },
            {-0.16071810E+00, 0.14123527E-01,-0.13230798E+00, 0.34138896E-01,
              0.59729513E-01,-0.50168329E+00,-0.57280529E+00, 0.43802693E-01,
             -0.13694671E-01, 0.17453475E-01, 0.33055807E-02, 0.23122183E-02,
              0.62880949E-02, },
            { 0.53357907E-01,-0.19458123E-01, 0.69014654E-01, 0.63686967E-02,
              0.54668400E-01, 0.58147484E+00,-0.50669259E+00,-0.12543346E-01,
              0.45777824E-01,-0.29289942E-01,-0.17757392E-01,-0.45786155E-03,
             -0.10528302E-01, },
            {-0.54030935E-02, 0.18703187E+00,-0.11604700E+00,-0.90121031E-01,
             -0.24077909E+00,-0.58423525E+00,-0.78927018E-01,-0.25518278E-01,
              0.17570584E+00, 0.64967521E-01,-0.28771313E-01, 0.21496497E-01,
              0.58259081E-01, },
            {-0.21440378E+00, 0.46478152E-01, 0.59997041E-01,-0.10545159E+00,
              0.87879352E-01, 0.31408742E+00,-0.76883548E+00,-0.22707506E+00,
              0.97065166E-01,-0.84792078E-01,-0.42242599E-02,-0.36972161E-01,
              0.13577213E-01, },
            { 0.13984452E+01, 0.21562593E+00, 0.81785202E+00, 0.42129427E-01,
             -0.45740336E+00, 0.19257108E+01, 0.10962818E+01,-0.13253964E+00,
              0.64441629E-01,-0.76614603E-01, 0.80381967E-01, 0.60075335E-01,
             -0.42204820E-02, },
            {-0.88858914E+00, 0.68705350E+00,-0.11165400E+00,-0.39081043E-02,
             -0.11788681E+01,-0.96058542E+00, 0.22603960E+01, 0.79304010E-01,
             -0.20290291E+00, 0.10524015E+00, 0.22646762E+00, 0.12000297E-01,
              0.20639112E-02, },
            {-0.38604550E-01,-0.64558333E+00, 0.15047131E+00, 0.23604788E+00,
              0.51776189E+00, 0.65631807E+00,-0.64743772E-01, 0.23293781E+00,
             -0.31635416E+00,-0.24677651E+00, 0.73522218E-01,-0.12383265E+00,
             -0.23125414E-01, },
            { 0.63801259E+00,-0.17993659E+00, 0.41443504E-01,-0.30626392E+00,
              0.13222785E+00,-0.38505167E+00, 0.11026433E+01, 0.38552052E+00,
             -0.14338492E+00, 0.16477841E+00, 0.44638623E-01,-0.27766574E-01,
             -0.14409494E-01, },
            {-0.16779745E+01, 0.35500047E+00,-0.75000972E+00, 0.44501577E-01,
              0.16619511E+00,-0.29565687E+01,-0.11524156E+01, 0.10906345E+00,
              0.63353598E-01, 0.74749373E-01,-0.98273449E-01,-0.23684590E+00,
              0.99928290E-01, },
            { 0.12862395E+01,-0.10268888E+01,-0.34922286E-03, 0.42514104E+00,
              0.18869687E+01, 0.92553735E+00,-0.33864348E+01,-0.96434183E-01,
              0.88395439E-01,-0.27669014E-01,-0.47842038E+00,-0.87917507E-01,
             -0.50919611E-01, },
            { 0.67794308E-01, 0.43764468E-01,-0.85234791E-02, 0.55448759E-01,
             -0.75007221E-02, 0.26949584E-02, 0.40005311E-01, 0.17787793E+00,
             -0.21966216E+00, 0.13119747E-01, 0.17033216E-01, 0.48258617E-02,
              0.93731331E-02, },
            { 0.41226272E-01,-0.45745876E-01,-0.44874894E-02,-0.56989055E-01,
              0.66878618E-03, 0.49436111E-02,-0.93553104E-02, 0.20852254E+00,
              0.17780304E+00,-0.10008897E-01, 0.40826789E-03,-0.44579688E-02,
             -0.96555213E-02, },
            { 0.32994185E-01,-0.50840318E-01,-0.82209967E-01,-0.16789146E+00,
             -0.35853390E-01,-0.74349888E-01, 0.36870703E-01,-0.67959286E-01,
             -0.23842365E+00,-0.75435199E-01, 0.58410540E-01, 0.19143853E-01,
              0.72812706E-01, },
            {-0.35581663E-01,-0.75313292E-01,-0.37497939E-02,-0.53745732E-02,
              0.86594537E-01,-0.85272796E-01,-0.75303704E-01, 0.15578288E+00,
             -0.57227295E-01,-0.19145612E-01, 0.38526919E-01,-0.33226525E-02,
             -0.25027826E-01, },
            {-0.24258215E-01, 0.56879047E-01,-0.87987788E-01,-0.18437518E-01,
             -0.26629930E-02, 0.94258245E-02, 0.11299050E-01,-0.10933794E-01,
              0.11623712E-02, 0.17373119E+00, 0.16964652E+00,-0.10166550E-01,
             -0.62725325E-02, },
            { 0.30380953E-01,-0.23155006E-01,-0.33405375E-01,-0.44372205E-01,
              0.21909758E-01,-0.50504357E-02,-0.34750249E-01, 0.32919927E-02,
             -0.52613996E-01,-0.17605481E+00, 0.17251751E+00, 0.14515134E-01,
              0.11617797E-01, },
            { 0.42745028E-01,-0.67311842E-02,-0.12199854E-01,-0.23588119E-01,
              0.42619001E-01, 0.23207851E-01, 0.16206969E-01,-0.71131140E-02,
             -0.15745321E-01, 0.93277032E-02, 0.54416276E-03,-0.92576854E-01,
              0.75096644E-01, },
            {-0.57170991E-01,-0.11325012E-01,-0.45626007E-01,-0.51696181E-01,
             -0.54596667E-02, 0.18047408E-02,-0.57304051E-01, 0.28688202E-01,
             -0.29154113E-02,-0.48920629E-02,-0.11745230E-02,-0.72187759E-01,
             -0.10533948E+00, },
            {-0.46112318E-01,-0.93352431E-02,-0.77548169E-01, 0.10806181E-01,
              0.16295983E-01,-0.91340821E-02,-0.37854966E-01, 0.59323609E-02,
             -0.70786551E-02, 0.43499945E-02,-0.12444421E-01,-0.23848300E-02,
              0.11797853E-01, },
            {-0.69076233E-02,-0.14922633E-01,-0.16607357E-01,-0.28959988E-01,
              0.37629224E-01, 0.27335903E-02,-0.21362375E-01, 0.14246407E-01,
              0.54619005E-02, 0.60001425E-02, 0.29482830E-01,-0.15042797E-01,
             -0.14008236E-01, },
            {-0.39528869E-02, 0.18516526E-02,-0.39649025E-01, 0.28208220E-01,
             -0.11501377E-01,-0.40830545E-01, 0.14463947E-01, 0.18475296E-02,
             -0.75901337E-02, 0.18725945E-01, 0.10570663E-01,-0.12890499E-04,
              0.23008062E-01, },
            {-0.51339738E-01, 0.28110512E-01,-0.61789113E-02, 0.81891287E-02,
             -0.29036777E-01, 0.28971057E-01, 0.64823423E-02, 0.76583945E-02,
              0.27483683E-02,-0.17355928E-01, 0.15277426E-01, 0.51793233E-02,
             -0.15141426E-01, },
         },
         {
            { 0.88938065E+01,-0.15078264E-01,-0.25140066E-01, 0.22020804E-01,
              0.25030833E-02,-0.20642305E-01,-0.30851010E-01, 0.35875994E-02,
             -0.87216607E-03, 0.24480885E-01, 0.15102470E-01,-0.19631794E-01,
              0.20391542E-01, },
            { 0.61357880E+00, 0.13776004E+00,-0.57032752E+00,-0.28906608E-01,
             -0.35331145E+00,-0.21978439E+00,-0.14012583E+00, 0.54715615E-01,
              0.14554127E+00,-0.24320902E-02,-0.52757952E-01,-0.18868657E-01,
             -0.78913517E-01, },
            { 0.26584980E+02,-0.91476983E+00,-0.10767018E+01, 0.83116639E+00,
              0.18041835E+01,-0.63989371E+00, 0.51932621E+00, 0.50343698E+00,
             -0.63642442E+00,-0.97818619E+00,-0.29772168E+00, 0.35066122E+00,
             -0.68079138E+00, },
            { 0.31529946E+01,-0.22581005E+01, 0.95738211E+01, 0.81222594E+00,
              0.97320185E+01, 0.23350554E+01, 0.16431698E+01, 0.61718774E+00,
             -0.20582492E+01,-0.12018207E-01, 0.19452527E+01,-0.24020819E+00,
              0.93675071E+00, },
            {-0.12845282E+03, 0.76379828E+01, 0.81046038E+01,-0.12054643E+02,
             -0.15387237E+02, 0.68140392E+01,-0.15058470E+01,-0.48598957E+01,
              0.49816790E+01, 0.52011762E+01, 0.22433043E+01,-0.60794902E+00,
              0.40673318E+01, },
            {-0.78451271E+02,-0.14396160E+02,-0.34657215E+02,-0.25212047E+01,
             -0.49333851E+02,-0.11871571E+02,-0.53334084E+01,-0.59590955E+01,
              0.88634005E+01, 0.10821542E+01,-0.11399439E+02, 0.47652416E+01,
             -0.32855816E+01, },
            { 0.21243150E+03,-0.14004525E+02,-0.17843674E+02, 0.43910168E+02,
              0.43445202E+02,-0.19488865E+02, 0.13466698E+01, 0.12091710E+02,
             -0.13215184E+02,-0.10340988E+02,-0.57268853E+01,-0.15506169E+01,
             -0.88107119E+01, },
            { 0.24853136E+03, 0.78768333E+02, 0.34186371E+02, 0.19544324E+01,
              0.10425354E+03, 0.27955261E+02, 0.85331831E+01, 0.15262192E+02,
             -0.16240059E+02,-0.35532913E+01, 0.25211710E+02,-0.16891586E+02,
              0.51664090E+01, },
            {-0.16578394E+03, 0.30796990E+01, 0.14645020E+02,-0.58538593E+02,
             -0.48412685E+02, 0.21371475E+02,-0.61169732E+00,-0.10880286E+02,
              0.14118615E+02, 0.89713211E+01, 0.56740236E+01, 0.40566039E+01,
              0.80941858E+01, },
            {-0.28710046E+03,-0.11082728E+03, 0.47830529E+01, 0.12881908E+01,
             -0.99203979E+02,-0.28727617E+02,-0.78436465E+01,-0.15181051E+02,
              0.13331136E+02, 0.39052055E+01,-0.23992605E+02, 0.21218477E+02,
             -0.36269221E+01, },
            { 0.51397850E+02, 0.44691315E+01,-0.40350652E+01, 0.25961407E+02,
              0.18408005E+02,-0.80434484E+01, 0.32910940E+00, 0.31243167E+01,
             -0.52550664E+01,-0.28733051E+01,-0.18864105E+01,-0.22640367E+01,
             -0.26912494E+01, },
            { 0.11311129E+03, 0.48871571E+02,-0.13456262E+02,-0.14472980E+01,
              0.34871849E+02, 0.10456634E+02, 0.32485316E+01, 0.52217917E+01,
             -0.40892677E+01,-0.13860580E+01, 0.83094473E+01,-0.88487806E+01,
              0.87602508E+00, },
            { 0.69597103E-01, 0.16769730E+01, 0.15545819E+01,-0.55449349E-02,
              0.69705747E-01, 0.23653561E-01, 0.12845129E-01, 0.86718919E-02,
             -0.45932859E-01, 0.31570107E-01, 0.45554500E-01,-0.21642283E-01,
              0.12889816E-01, },
            {-0.21921287E+00,-0.16087893E+01, 0.17624192E+01,-0.13460632E+00,
             -0.59290775E-02,-0.83502904E-02,-0.34621302E-01, 0.11610980E-01,
             -0.14968859E-01,-0.25111865E-01, 0.17095666E-01,-0.55068266E-02,
             -0.23103973E-01, },
            {-0.48869601E+00, 0.17512943E+01, 0.14671189E+01,-0.29812467E+00,
              0.43485570E+00, 0.74070942E+00, 0.47886825E+00,-0.15530139E-01,
              0.28587043E+00,-0.90405986E-01, 0.12089233E+00,-0.10466526E+00,
             -0.22864467E+00, },
            { 0.26475123E+00,-0.10255442E+01, 0.12023131E+01,-0.79612106E+00,
             -0.14503480E+00, 0.48347488E-01, 0.24352114E+00, 0.33930928E+00,
              0.17734116E+00,-0.29564553E+00,-0.23605779E-01, 0.26259270E+00,
             -0.14518072E+00, },
            { 0.32442267E+01, 0.19304132E+02,-0.17855993E+02, 0.15392791E+01,
              0.12891808E+01,-0.13034651E+01, 0.78133240E-01,-0.28128734E+00,
              0.13791049E+01,-0.29651865E+00,-0.13774163E+01, 0.90447134E+00,
             -0.48520055E+00, },
            {-0.38545475E+01, 0.17391617E+02, 0.15856245E+02, 0.32506826E+01,
              0.17471519E+01, 0.54408711E+00,-0.87338269E-01, 0.11546892E+01,
             -0.39241776E+00, 0.82400136E-01,-0.41304421E+00,-0.31734011E+00,
             -0.47280574E+00, },
            { 0.15891956E+02,-0.71328111E+01,-0.82937202E+01,-0.10363106E+01,
             -0.37894361E+01,-0.13196981E+02,-0.97192078E+01, 0.57880574E+00,
             -0.50687685E+01, 0.15093237E+00,-0.15454117E+01, 0.22428346E+01,
              0.37367301E+01, },
            {-0.37251842E+01, 0.30780756E+00,-0.38300519E+01, 0.13541659E+02,
             -0.41292334E+01, 0.24990664E+01,-0.36101906E+01,-0.68872380E+01,
             -0.33844929E+01, 0.48587584E+01, 0.16653656E+01,-0.54408031E+01,
             -0.14610332E+01, },
            {-0.35525444E+02,-0.55185818E+02, 0.18234506E+03,-0.17191862E+02,
             -0.18231380E+02, 0.10715060E+02, 0.41820860E+00, 0.49597664E+01,
             -0.10105695E+02, 0.44890794E+00, 0.11448022E+02,-0.59232845E+01,
              0.21196377E+01, },
            { 0.48971519E+02,-0.17984668E+03,-0.10086093E+02,-0.28173100E+02,
             -0.18571411E+02,-0.34781046E+01, 0.46446908E+00,-0.91365309E+01,
              0.32334213E+01, 0.29283717E+00, 0.53791575E+01, 0.33645318E+01,
              0.52979574E+01, },
            {-0.98831841E+02, 0.35010017E+02, 0.81809647E+02, 0.42378185E+02,
              0.32127880E+02, 0.71276741E+02, 0.47292530E+02,-0.79754658E+01,
              0.26064089E+02, 0.57404060E+01, 0.62453079E+01,-0.13935455E+02,
             -0.23157148E+02, },
            { 0.37681507E+02,-0.22015532E+02, 0.41568684E+02,-0.89377480E+02,
              0.34201309E+02,-0.14441486E+02, 0.12126778E+02, 0.37069691E+02,
              0.21612261E+02,-0.29737093E+02,-0.11194304E+02, 0.31738426E+02,
              0.18420141E+02, },
            { 0.11772128E+03,-0.16025244E+02,-0.54052008E+03, 0.52717884E+02,
              0.66749588E+02,-0.31852327E+02,-0.27905669E+01,-0.18864178E+02,
              0.26960806E+02, 0.23407796E+00,-0.36213844E+02, 0.12973186E+02,
             -0.41116295E+01, },
            {-0.15707156E+03, 0.53367804E+03,-0.18824304E+03, 0.96771606E+02,
              0.59688560E+02, 0.96216154E+01, 0.33605173E+01, 0.27191839E+02,
             -0.63597393E+01,-0.57947147E+00,-0.20041174E+02,-0.10695272E+02,
             -0.15405689E+02, },
            { 0.25402480E+03,-0.12169586E+03,-0.29696808E+03,-0.17765755E+03,
             -0.10246051E+03,-0.16805841E+03,-0.91728348E+02, 0.29707182E+02,
             -0.58331951E+02,-0.25739067E+02,-0.14582021E+02, 0.34748569E+02,
              0.59372269E+02, },
            {-0.14180334E+03, 0.90961342E+02,-0.17497099E+03, 0.24299307E+03,
             -0.93683586E+02, 0.25937101E+02,-0.10940236E+02,-0.83610329E+02,
             -0.59765331E+02, 0.78051521E+02, 0.30359779E+02,-0.77565620E+02,
             -0.57330196E+02, },
            {-0.15175063E+03, 0.14136122E+03, 0.67243622E+03,-0.63942410E+02,
             -0.95142883E+02, 0.37566513E+02, 0.25746937E+01, 0.25226887E+02,
             -0.30201876E+02, 0.11459297E+01, 0.47971264E+02,-0.10312352E+02,
              0.43723435E+01, },
            { 0.18913745E+03,-0.64964203E+03, 0.40084787E+03,-0.13516272E+03,
             -0.75677673E+02,-0.12140692E+02,-0.98627472E+01,-0.34648926E+02,
              0.30085969E+01,-0.86283243E+00, 0.28143951E+02, 0.13283780E+02,
              0.17091984E+02, },
            {-0.29154059E+03, 0.17865944E+03, 0.44016376E+03, 0.25832797E+03,
              0.12581499E+03, 0.18186659E+03, 0.75668427E+02,-0.42528137E+02,
              0.58712421E+02, 0.38270672E+02, 0.19351198E+02,-0.37923859E+02,
             -0.65868088E+02, },
            { 0.21195496E+03,-0.14264771E+03, 0.25605206E+03,-0.27863412E+03,
              0.10939057E+03,-0.14928993E+02,-0.43717604E+01, 0.85104988E+02,
              0.72570885E+02,-0.90291161E+02,-0.36729885E+02, 0.84085503E+02,
              0.70076569E+02, },
            { 0.67721512E+02,-0.90769173E+02,-0.30035886E+03, 0.26721811E+02,
              0.47102436E+02,-0.15028220E+02,-0.20477100E+00,-0.11208223E+02,
              0.12107578E+02,-0.20266440E+01,-0.22364059E+02, 0.21444893E+01,
             -0.19921613E+01, },
            {-0.77146408E+02, 0.27997412E+03,-0.22317946E+03, 0.64471100E+02,
              0.32618526E+02, 0.58598595E+01, 0.59874973E+01, 0.15801620E+02,
              0.81914449E+00, 0.12750365E+01,-0.13325417E+02,-0.56135464E+01,
             -0.63814468E+01, },
            { 0.12182439E+03,-0.87194756E+02,-0.22319070E+03,-0.12528762E+03,
             -0.51943058E+02,-0.73651070E+02,-0.21513721E+02, 0.20706060E+02,
             -0.21602533E+02,-0.18770321E+02,-0.10114022E+02, 0.15116210E+02,
              0.26225403E+02, },
            {-0.10590909E+03, 0.75341286E+02,-0.12101205E+03, 0.11209223E+03,
             -0.47404507E+02, 0.42671257E+00, 0.63914065E+01,-0.32075302E+02,
             -0.31445272E+02, 0.37990154E+02, 0.16272755E+02,-0.33418064E+02,
             -0.29758162E+02, },
            { 0.24743646E-01,-0.24780034E+00, 0.32477383E-01,-0.70758212E+00,
              0.77269542E+00, 0.21704473E-01, 0.73262020E-02,-0.54433793E-02,
             -0.36484734E-02, 0.15552454E-01, 0.11511402E-01, 0.95175020E-02,
             -0.77193896E-02, },
            {-0.48369352E-01,-0.53220801E-01, 0.27144585E-01,-0.81413311E+00,
             -0.71973467E+00,-0.70226772E-04,-0.60768458E-02, 0.16783381E-01,
             -0.50986288E-02,-0.15607438E-01, 0.57154228E-02, 0.16545281E-01,
             -0.27624261E-01, },
            { 0.42223513E+00,-0.98611221E-01, 0.50277317E+00,-0.16155951E+01,
              0.44995168E+00, 0.60947478E+00,-0.26007742E+00,-0.34951422E+00,
             -0.61507735E-01,-0.29398164E+00,-0.57447203E-01,-0.17995073E-01,
             -0.72372705E-01, },
            { 0.12762964E+01, 0.26164582E+00, 0.47269124E+00,-0.10257167E+01,
             -0.19976238E+01, 0.18526942E+00, 0.29332149E+00,-0.20163956E+00,
              0.15762205E+00,-0.21016265E+00,-0.20759076E-01, 0.75109035E-01,
             -0.30883980E+00, },
            { 0.23981123E+01, 0.29899118E+01, 0.38449600E+01,-0.60386233E+01,
              0.53788686E+01,-0.93953538E+00,-0.15277848E+01, 0.59264237E+00,
              0.45906556E+00, 0.24649295E+00,-0.10305211E+00, 0.32214963E+00,
              0.46169207E+00, },
            { 0.16696631E+01, 0.25706765E+00, 0.28278365E+01,-0.46109791E+01,
             -0.72205920E+01, 0.34772471E+00,-0.62179738E+00,-0.14041898E-01,
              0.11932110E+01, 0.11926308E+00, 0.61168361E-01,-0.12217858E+00,
              0.74488103E-01, },
            {-0.42425346E+01, 0.35861960E+00,-0.69054904E+01, 0.77307906E+01,
              0.14447305E+01,-0.51289444E+01, 0.42940632E+00, 0.37216494E+01,
              0.11370006E+01, 0.24206262E+01,-0.55311292E-01, 0.61113574E-01,
              0.43589985E+00, },
            {-0.11186955E+02, 0.14551842E+01,-0.46096087E+01, 0.70924859E+01,
              0.82799072E+01, 0.54562348E+00,-0.19091228E+01, 0.63213462E+00,
             -0.97554809E+00, 0.27594109E+01, 0.64862490E-01,-0.13592871E+01,
              0.23405161E+01, },
            {-0.17754059E+02,-0.32034547E+01,-0.41930241E+02, 0.41079243E+02,
             -0.18051516E+02, 0.25384116E+01, 0.90485382E+01,-0.25573390E+01,
             -0.52479219E+01,-0.22733593E+01, 0.16716585E+01,-0.19437449E+01,
             -0.34734039E+01, },
            {-0.11303488E+02, 0.17556179E+02,-0.22370165E+02, 0.15778043E+02,
              0.52471008E+02,-0.31812954E+01, 0.27775536E+01,-0.23356988E+00,
             -0.10081247E+02, 0.12472678E+01,-0.56565022E+00, 0.53981122E-01,
              0.12582926E+00, },
            { 0.12362515E+02,-0.29632866E+01, 0.15124092E+02,-0.18801899E+02,
             -0.13831272E+02, 0.12536344E+02, 0.10004860E+01,-0.93945045E+01,
             -0.36175582E+01,-0.51194916E+01, 0.14203264E+01,-0.24957810E-01,
             -0.76139688E+00, },
            { 0.25606970E+02,-0.82127504E+01, 0.11993701E+02,-0.10493736E+02,
             -0.15479230E+02,-0.24436879E+01, 0.37523029E+01, 0.70967054E+00,
              0.28187840E+01,-0.75910454E+01, 0.22493189E-01, 0.46240821E+01,
             -0.53346510E+01, },
            { 0.38124302E+02,-0.81857109E+01, 0.10164820E+03,-0.78550056E+02,
              0.12163009E+02, 0.46451297E-01,-0.17560549E+02, 0.37782583E+01,
              0.13861211E+02, 0.52900162E+01,-0.42033973E+01, 0.37402480E+01,
              0.76729035E+01, },
            { 0.27311661E+02,-0.54518929E+02, 0.45628468E+02,-0.14268176E+02,
             -0.10502533E+03, 0.78711667E+01,-0.59193988E+01, 0.21227322E+01,
              0.23136654E+02,-0.59244928E+01, 0.94676512E+00, 0.14095348E+01,
              0.77875800E-01, },
            {-0.12606331E+02, 0.52987642E+01,-0.10864910E+02, 0.19476118E+02,
              0.18500385E+02,-0.88927727E+01,-0.17200902E+01, 0.67320638E+01,
              0.26062746E+01, 0.29826615E+01,-0.15832294E+01,-0.24634519E+00,
              0.30302539E+00, },
            {-0.16824041E+02, 0.83268948E+01,-0.88709040E+01, 0.17254645E+00,
              0.14521973E+02, 0.19288079E+01,-0.16344380E+01,-0.16025696E+01,
             -0.28889387E+01, 0.53860931E+01, 0.74160628E-01,-0.41329098E+01,
              0.36643453E+01, },
            {-0.25644232E+02, 0.11694123E+02,-0.68230270E+02, 0.47731461E+02,
              0.49299583E+01,-0.33478410E+01, 0.10701688E+02,-0.15946394E+01,
             -0.10438581E+02,-0.35907745E+01, 0.26091225E+01,-0.24007337E+01,
             -0.50249095E+01, },
            {-0.22422714E+02, 0.41399738E+02,-0.25003633E+02, 0.22523800E+00,
              0.66093285E+02,-0.54813409E+01, 0.45410838E+01,-0.25613079E+01,
             -0.15686406E+02, 0.53152466E+01,-0.38827559E+00,-0.17855603E+01,
             -0.56493247E+00, },
            {-0.25739521E+00,-0.51444806E-01,-0.39007924E-01, 0.43471336E-01,
              0.74577101E-01,-0.69018120E+00,-0.77091724E+00, 0.18789891E-01,
             -0.30679217E-01, 0.34278046E-01, 0.70777461E-02, 0.17007646E-02,
              0.12116459E-01, },
            { 0.30931152E-01, 0.82304358E-01, 0.23780592E-01,-0.61587142E-02,
              0.14854147E+00, 0.73189169E+00,-0.68215245E+00, 0.67674126E-02,
              0.98737925E-02,-0.13763633E-01,-0.28919926E-01,-0.11603950E-02,
              0.28128898E-01, },
            { 0.15583675E+00,-0.97631486E-02,-0.42719718E-01, 0.55609424E-01,
             -0.76401152E-01,-0.72838420E+00,-0.35610983E+00, 0.81875414E-01,
              0.17080715E+00,-0.30333633E-01,-0.31908482E-01, 0.57391785E-01,
              0.27769849E-01, },
            {-0.52173845E-01, 0.24948262E+00, 0.12350982E+00,-0.15202740E+00,
              0.21870527E+00, 0.53911513E+00,-0.62828916E+00,-0.21726982E+00,
             -0.81490837E-02,-0.29228795E-01,-0.10110049E-01,-0.15235767E-01,
              0.31288322E-01, },
            { 0.17942052E+01,-0.47697356E+00,-0.24683045E-01,-0.73759890E+00,
             -0.31411227E-01, 0.23973184E+01, 0.15254406E+01, 0.46166003E+00,
             -0.19089140E+00,-0.40453413E+00, 0.24325491E+00, 0.47338758E-01,
             -0.12414197E+00, },
            {-0.10261521E+01, 0.34687656E+00,-0.13719416E+00,-0.56109297E+00,
             -0.19106919E+01,-0.12889602E+01, 0.28899510E+01, 0.31973645E+00,
              0.39896643E+00, 0.91862530E-01, 0.74149244E-01, 0.17311741E+00,
             -0.22030464E+00, },
            { 0.19139686E+00,-0.29637602E+00,-0.89744121E-01, 0.34440780E-03,
              0.12633979E+00, 0.17837116E+00, 0.34260508E+00,-0.52784592E-01,
             -0.35431859E+00, 0.11950550E+00, 0.18184212E+00,-0.15287527E+00,
             -0.12142629E-02, },
            { 0.38059655E+00,-0.70801187E+00,-0.47004503E+00, 0.36706582E-01,
             -0.10262799E+00,-0.90509468E+00,-0.71891233E-01, 0.41418588E+00,
              0.51909392E-02, 0.73546261E-01, 0.61649170E-01,-0.10198504E+00,
             -0.11583487E+00, },
            {-0.30275664E+01, 0.20220103E+01, 0.34253958E+00, 0.11947123E+01,
             -0.63510936E+00,-0.44640880E+01,-0.18234404E+01,-0.84736711E+00,
              0.57081997E+00, 0.75149286E+00,-0.48013139E+00,-0.24022443E-01,
              0.20686179E+00, },
            { 0.15877696E+01,-0.71108854E+00, 0.99898326E+00, 0.13696940E+01,
              0.26519346E+01, 0.18384987E+01,-0.53081517E+01,-0.50006735E+00,
             -0.82502490E+00,-0.16206498E+00,-0.16357306E+00,-0.40980041E+00,
              0.16522449E+00, },
            {-0.80586374E-02, 0.50992239E-01,-0.46632238E-01, 0.57858139E-01,
              0.29589590E-01,-0.29866926E-01, 0.44593327E-01, 0.20724247E+00,
             -0.33956417E+00, 0.48757136E-01, 0.48252712E-02, 0.75157746E-02,
              0.82745338E-02, },
            { 0.56242988E-01,-0.98413797E-02,-0.85247047E-02,-0.81652971E-02,
              0.10179361E-01,-0.32878406E-01, 0.63941120E-02, 0.30926964E+00,
              0.19301084E+00, 0.51544094E-02,-0.95572891E-02, 0.14922715E-01,
             -0.64172191E-02, },
            { 0.19160556E-02,-0.24454545E-02,-0.16382514E+00,-0.21728696E-01,
              0.84944502E-01,-0.70919216E-01, 0.61331507E-01,-0.12808955E+00,
             -0.38097063E+00,-0.58370437E-01, 0.67949370E-01, 0.16139146E-01,
              0.56471489E-01, },
            {-0.82536153E-02,-0.59724025E-01, 0.79807997E-01, 0.33527713E-01,
              0.72642304E-01,-0.10801914E+00,-0.34897956E-02, 0.36352578E+00,
             -0.57958063E-01,-0.35505172E-01, 0.12366363E-02, 0.40800598E-01,
             -0.18490655E-01, },
            {-0.46349317E-02, 0.14545528E-01,-0.12120489E+00, 0.26173390E-01,
              0.17550857E-02,-0.28536871E-01, 0.22547282E-01,-0.36351152E-01,
             -0.60174363E-02, 0.28420082E+00, 0.16398448E+00,-0.12288245E-01,
              0.14442492E-01, },
            { 0.60371339E-01, 0.38244709E-01, 0.32651130E-01,-0.77243634E-01,
             -0.35003498E-01, 0.14984454E-01, 0.15117691E-01,-0.19099513E-01,
             -0.49697436E-01,-0.15599817E+00, 0.25729859E+00, 0.33024959E-02,
             -0.99182769E-03, },
            { 0.71319506E-01,-0.75824670E-01, 0.36760561E-01,-0.49347986E-01,
              0.15868425E-01, 0.38285628E-01, 0.23324162E-01,-0.39113779E-01,
             -0.30983375E-01, 0.13260117E-01, 0.37964746E-01,-0.74268147E-01,
              0.11632289E+00, },
            {-0.17346490E-01, 0.15348023E-01,-0.14875906E-01,-0.52558880E-01,
             -0.71550235E-01,-0.61885449E-02,-0.24182137E-01, 0.37889171E-02,
             -0.25238233E-01,-0.21548860E-02, 0.12035027E-01,-0.14105137E+00,
             -0.92943028E-01, },
            {-0.53878661E-01,-0.32849424E-01,-0.84194556E-01, 0.31174347E-01,
             -0.46395538E-02, 0.12672110E-01,-0.28512165E-01,-0.13265373E-01,
             -0.14615447E-01, 0.36233847E-02,-0.12471826E-01, 0.34783301E-02,
              0.54068253E-02, },
            {-0.35035559E-02, 0.55042438E-01, 0.23584509E-01,-0.25574235E-01,
              0.69150315E-02,-0.15520843E-01, 0.11701881E-01, 0.12512361E-01,
             -0.23976658E-01, 0.11502410E-02, 0.91396850E-02, 0.85549625E-02,
             -0.30781742E-01, },
            { 0.11815184E-01, 0.77101663E-01, 0.19891476E-01,-0.14418868E-01,
             -0.48151329E-01,-0.46777282E-01, 0.32677677E-01, 0.10109333E-01,
              0.22690778E-03, 0.29887985E-01,-0.23127506E-01, 0.21396790E-01,
              0.23422489E-01, },
            {-0.10521832E+00, 0.46248104E-01,-0.85954962E-04, 0.24128256E-01,
              0.10098126E-01, 0.15944390E-01, 0.24146425E-01, 0.21453386E-01,
              0.39750654E-02,-0.21658875E-01, 0.59070806E-02, 0.15631018E-01,
             -0.13315894E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir12Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30724745E+01, 0.76932684E-02, 0.10582104E-01,-0.58678435E-02,
              0.86863544E-02, 0.23237041E-02,-0.16314250E-02, 0.22063474E-02,
              0.30517471E-02, },
            { 0.31540647E+00,-0.10881314E-01, 0.51532704E-01, 0.60170814E-01,
              0.84916512E-02,-0.18651348E-01,-0.24970861E-01, 0.24080003E-01,
             -0.41201661E-03, },
            { 0.70605505E+00,-0.14244080E+00,-0.29646529E-01,-0.12706631E-01,
             -0.95824571E-02,-0.33969921E-02, 0.17386436E-01,-0.78530975E-01,
             -0.11982952E-01, },
            {-0.42120489E+00,-0.20498924E+00,-0.70140377E-01,-0.13760924E+00,
             -0.42224452E-01, 0.85511714E-01, 0.73424190E-01,-0.67239940E-01,
              0.15587630E-01, },
            {-0.89627123E+00, 0.41788438E+00,-0.50209500E-01, 0.68723954E-01,
              0.28690726E-01, 0.34792382E-01,-0.43728635E-01, 0.22312871E+00,
              0.29185694E-01, },
            { 0.65906473E-01, 0.19188379E+00, 0.25963048E-01, 0.69747426E-01,
              0.35327666E-01,-0.73112071E-01,-0.53085875E-01, 0.43332018E-01,
             -0.21014430E-01, },
            { 0.26427588E+00,-0.27628177E+00, 0.10059510E+00,-0.50584864E-01,
             -0.29864950E-01,-0.44461355E-01, 0.33647098E-01,-0.14697210E+00,
             -0.26645195E-01, },
            { 0.16860802E-01,-0.18605264E+00,-0.38714597E+00,-0.19328046E-03,
              0.40253001E-03, 0.75271179E-03, 0.54476932E-02, 0.79560261E-02,
              0.55059040E-03, },
            { 0.20514991E-01, 0.39720568E+00,-0.19083656E+00,-0.12558009E-01,
              0.56398655E-02,-0.12297944E-01,-0.14556028E-01,-0.61569130E-02,
              0.10029857E-01, },
            { 0.10009073E-01, 0.30681634E+00, 0.34232047E+00, 0.98892421E-01,
              0.69215246E-01, 0.82950536E-02,-0.18969623E-01, 0.55852693E-01,
             -0.31100798E-01, },
            { 0.77327024E-02,-0.19496195E+00, 0.34152469E+00,-0.10556676E+00,
             -0.18567836E-01, 0.55548723E-03,-0.31484790E-01, 0.53078610E-01,
              0.60641043E-01, },
            {-0.30647790E+00, 0.81640363E+00, 0.18649312E+01,-0.98093107E-01,
              0.15951095E-01, 0.17376818E-01,-0.39560460E-01,-0.20320363E+00,
             -0.74637689E-01, },
            {-0.73014015E+00,-0.19144686E+01, 0.81253171E+00, 0.19620964E+00,
             -0.52492674E-01, 0.88517673E-01, 0.17049298E+00, 0.12281351E+00,
             -0.20673332E+00, },
            { 0.48533568E+00,-0.15024738E+01,-0.17705139E+01,-0.81606116E-03,
             -0.53666764E+00,-0.37227347E+00, 0.75344280E-02,-0.40956607E+00,
              0.40251580E+00, },
            { 0.89858896E+00, 0.10686440E+01,-0.16356423E+01, 0.29005781E+00,
              0.37916356E+00,-0.13871599E-01, 0.33012232E+00,-0.34580472E+00,
             -0.33433393E+00, },
            { 0.10481230E+01,-0.77934951E+00,-0.19115534E+01, 0.12393999E+00,
              0.55828355E-01,-0.70680223E-01, 0.11313992E+00, 0.58049357E+00,
              0.31924832E+00, },
            { 0.22701538E+01, 0.16197052E+01,-0.96915621E+00,-0.33665845E+00,
             -0.14552668E+00,-0.21495460E+00,-0.52067465E+00,-0.32353064E+00,
              0.57284433E+00, },
            {-0.12134343E+01, 0.19404646E+01, 0.37635965E+01,-0.62870938E+00,
              0.10511738E+01, 0.11378073E+01,-0.29707570E-01, 0.86152846E+00,
             -0.11126765E+01, },
            {-0.29988582E+01,-0.20589967E+01, 0.24830685E+01,-0.22141758E-01,
             -0.89734024E+00, 0.17300197E+00,-0.71507871E+00, 0.67783284E+00,
              0.56816947E+00, },
            {-0.79848272E+00, 0.23965521E+00, 0.80937093E+00, 0.18386710E-01,
             -0.11181234E+00, 0.65899365E-01,-0.10855011E+00,-0.42357123E+00,
             -0.30929345E+00, },
            {-0.19041708E+01,-0.49399413E-01, 0.43354791E+00, 0.93478829E-01,
              0.24635497E+00, 0.13785353E+00, 0.41312331E+00, 0.21510728E+00,
             -0.42922243E+00, },
            { 0.70117193E+00,-0.47726721E+00,-0.19355594E+01, 0.54693979E+00,
             -0.56250012E+00,-0.84684205E+00, 0.39356496E-01,-0.54770625E+00,
              0.81452745E+00, },
            { 0.24183495E+01, 0.72689992E+00,-0.10589318E+01,-0.23238608E+00,
              0.54398221E+00,-0.20291325E+00, 0.43034214E+00,-0.41715705E+00,
             -0.29115310E+00, },
            {-0.90023009E-02,-0.35391398E-01,-0.47154608E-02, 0.19855529E-01,
             -0.11723321E+00,-0.10261102E-01, 0.60877344E-02,-0.11590884E-01,
             -0.67245788E-02, },
            { 0.96493689E-02,-0.22004317E-01,-0.56211897E-02, 0.11911504E+00,
              0.25769388E-01,-0.13206676E-01,-0.57847810E-03,-0.63171289E-02,
             -0.14825935E-02, },
            { 0.97160034E-01, 0.14064242E+00, 0.12654921E+00, 0.90184147E-02,
              0.10824892E+00, 0.98466850E-03,-0.39058451E-01, 0.21360939E-01,
             -0.68189781E-02, },
            {-0.69581389E-01, 0.23204980E-01,-0.33772092E-01,-0.23201568E-01,
             -0.16431106E-01, 0.35661325E-01,-0.24787607E-01,-0.15851652E-01,
              0.41659139E-01, },
            {-0.95431058E-03, 0.32939354E+00, 0.11949618E+00,-0.34582306E-01,
             -0.11245041E+00, 0.58247324E-01, 0.44278201E-01, 0.32697480E-01,
              0.79446398E-02, },
            {-0.27802193E+00, 0.99982269E-01,-0.18884204E+00, 0.14989616E+00,
             -0.30045878E-01,-0.12822229E-01, 0.70558451E-01, 0.67012236E-02,
             -0.54927107E-01, },
            {-0.52163047E+00,-0.76617777E+00,-0.38275307E+00, 0.93988568E-01,
             -0.80029559E+00,-0.46852093E-01, 0.19828326E+00,-0.19820786E+00,
             -0.20384854E-01, },
            { 0.73622519E+00, 0.23355688E+00, 0.24254851E+00, 0.40519950E+00,
              0.11749786E-01,-0.76143019E-01, 0.65943278E-01, 0.78227133E-01,
             -0.24041747E+00, },
            {-0.48859108E-01,-0.53506434E+00,-0.33032450E+00,-0.10122342E+00,
              0.79143211E-01,-0.14064236E+00,-0.84364615E-01, 0.29860295E-02,
             -0.22253854E-01, },
            { 0.39246804E+00,-0.24812710E+00, 0.40944198E+00,-0.17673942E+00,
             -0.78235663E-01, 0.72178833E-01,-0.16227914E+00, 0.10302575E-01,
              0.85755751E-01, },
            { 0.64045298E+00, 0.11233286E+01, 0.36713567E+00,-0.16331942E+00,
              0.12432953E+01, 0.20128106E+00,-0.23927681E+00, 0.27385575E+00,
              0.72312951E-01, },
            {-0.94199336E+00,-0.33690757E+00,-0.17330156E+00,-0.81343490E+00,
              0.69725722E-01, 0.41456182E-01,-0.43618813E-03,-0.84605128E-01,
              0.24982797E+00, },
            {-0.22805035E-01,-0.59726271E-02, 0.38705333E-02, 0.30002026E-02,
              0.19159815E-02, 0.18420041E-01,-0.37724916E-01, 0.35234531E-02,
             -0.43825819E-02, },
            {-0.67331134E-02,-0.10876593E-01,-0.79003610E-02,-0.13564546E-01,
             -0.40359148E-02, 0.42511482E-01, 0.42690061E-01, 0.78176362E-02,
             -0.44241268E-02, },
            { 0.61757568E-01, 0.38225956E-01, 0.16911132E-01,-0.45007326E-01,
              0.18286875E-01,-0.83427288E-01, 0.43182421E-01, 0.80693150E-02,
             -0.15373636E-02, },
            { 0.29610468E-01,-0.18918579E-01,-0.18943191E-01,-0.15242503E-01,
             -0.12218372E-01,-0.21460386E-01,-0.98122187E-01,-0.44038394E-03,
              0.17151907E-01, },
            { 0.91914488E-02,-0.88108525E-01, 0.29710799E-01, 0.37761092E-01,
             -0.42912245E-01,-0.14925539E+00,-0.86556077E-01, 0.32728366E-02,
             -0.97887143E-02, },
            {-0.35554547E-01,-0.30815808E-01,-0.29502632E-01, 0.11506282E-01,
             -0.96649574E-02, 0.11226377E+00,-0.22778022E+00, 0.95754518E-03,
              0.19735217E-01, },
            {-0.19949598E-01,-0.14575391E-01, 0.80723669E-02, 0.10257792E-01,
              0.98831486E-03, 0.21253310E-02, 0.44467166E-03, 0.30331459E-01,
             -0.68215737E-02, },
            {-0.16475892E-01, 0.66600149E-02,-0.96909413E-02,-0.23368848E-02,
              0.10803202E-01, 0.18352592E-03, 0.94600813E-02, 0.34825657E-02,
              0.29143477E-01, },
            { 0.68724401E-01, 0.15268687E-01,-0.34583665E-02,-0.26248895E-01,
             -0.41180071E-02,-0.69923680E-02, 0.78354776E-02,-0.15067245E-01,
             -0.50502356E-01, },
            { 0.26632205E-01,-0.38117900E-01, 0.12489504E-01, 0.38269710E-01,
             -0.23778917E-01,-0.45978669E-02,-0.10741331E-01, 0.55188518E-01,
             -0.19089222E-01, },
            { 0.50233081E-02,-0.68038455E-02, 0.13171528E-01,-0.79407496E-02,
             -0.15828494E-02, 0.69405884E-02,-0.26731647E-02,-0.18833794E-02,
              0.10679700E-01, },
            { 0.77929986E-02, 0.25170868E-01,-0.12785459E-01,-0.85407840E-02,
             -0.88489456E-02, 0.11496038E-01,-0.64758729E-03,-0.79743098E-02,
             -0.92306349E-03, },
            {-0.26192807E-03, 0.58631459E-02, 0.11213572E-01, 0.88310177E-02,
              0.19518093E-02, 0.98038837E-03,-0.22055174E-02, 0.15052836E-02,
              0.49419207E-02, },
            {-0.49593612E-02,-0.28198517E-02,-0.11953374E-01,-0.12248964E-01,
             -0.72131357E-02, 0.10401811E-01,-0.17990725E-02,-0.28054372E-02,
              0.18662043E-02, },
         },
         {
            { 0.27312081E+01,-0.36635617E-03, 0.16809828E-02, 0.35598597E-02,
              0.75279046E-02, 0.53007563E-03, 0.95740217E-03,-0.22847112E-02,
             -0.28182086E-03, },
            { 0.21439911E+00, 0.15668221E-01, 0.39947953E-01, 0.18864455E-01,
              0.18349148E-01,-0.73838946E-02,-0.29964812E-01, 0.16750785E-01,
              0.58596502E-02, },
            { 0.98324704E+00,-0.38888179E-01,-0.89483615E-02,-0.30634863E-01,
             -0.57423547E-01, 0.13029807E-01, 0.31853255E-01,-0.39872225E-01,
              0.66098981E-02, },
            {-0.71482211E-01,-0.26525632E+00,-0.19658718E-01,-0.33415090E-01,
             -0.51787488E-01, 0.35274658E-01, 0.85669100E-01,-0.48909988E-01,
             -0.21501379E-01, },
            {-0.13290060E+01, 0.17191923E+00,-0.47023818E-01, 0.52566610E-01,
              0.13115127E+00,-0.12865191E-01,-0.10077955E+00, 0.11846802E+00,
              0.88568972E-02, },
            {-0.13872875E+00, 0.21173365E+00,-0.49534766E-02, 0.75430055E-02,
              0.31798299E-01,-0.33703383E-01,-0.59091616E-01, 0.35570819E-01,
              0.11184796E-01, },
            { 0.40687090E+00,-0.11250211E+00, 0.64376906E-01,-0.24939286E-01,
             -0.85930578E-01,-0.26775303E-02, 0.72211958E-01,-0.74090555E-01,
             -0.24171596E-01, },
            { 0.18235693E-01,-0.32615599E+00,-0.29267040E+00, 0.70642778E-02,
              0.11300360E-02,-0.19541867E-02, 0.65887659E-02, 0.25641054E-03,
              0.29901171E-03, },
            { 0.10992618E-01, 0.29850316E+00,-0.32224572E+00,-0.13626590E-01,
             -0.15000732E-02,-0.56463708E-02,-0.13415891E-01, 0.31011614E-02,
              0.13834988E-01, },
            {-0.92273327E-02, 0.24425700E+00, 0.56881353E-01, 0.36258149E-03,
              0.63643664E-01,-0.37808134E-03,-0.32947324E-01, 0.38247403E-01,
              0.19941090E-01, },
            {-0.57335582E-01, 0.87612174E-01, 0.25150666E+00,-0.19707844E-01,
             -0.44863459E-01, 0.16167358E-01,-0.10398516E-01, 0.50344802E-01,
              0.68640865E-01, },
            {-0.23492485E+00, 0.11852225E+01, 0.87067664E+00,-0.62521182E-01,
              0.42712395E-02, 0.31809572E-01, 0.21999139E-01,-0.92716359E-01,
             -0.10461421E+00, },
            {-0.57572049E+00,-0.91379106E+00, 0.12106148E+01, 0.14713533E+00,
              0.49721982E-01, 0.31173525E-01, 0.14988387E+00, 0.74236974E-01,
             -0.24920762E+00, },
            { 0.37146536E+00,-0.15066245E+01,-0.33582768E+00, 0.13240382E+00,
             -0.49595794E+00,-0.17519762E+00, 0.14915365E-01,-0.31701419E+00,
             -0.11277883E+00, },
            { 0.11722958E+01,-0.33704272E+00,-0.16070234E+01,-0.73385984E-01,
              0.44775617E+00,-0.89496598E-01, 0.11555021E+00,-0.42885178E+00,
             -0.43839565E+00, },
            { 0.61520100E+00,-0.11530799E+01, 0.13902967E+00, 0.11537025E+00,
             -0.13625573E-01,-0.11211418E+00,-0.90618208E-01, 0.34417906E+00,
              0.43181029E+00, },
            { 0.19090471E+01,-0.30285779E+00,-0.14655380E+01,-0.29066685E+00,
             -0.22451898E+00,-0.12008682E+00,-0.50865197E+00,-0.31707212E+00,
              0.65367371E+00, },
            {-0.86626554E+00, 0.25727086E+01, 0.11919091E+01,-0.25942805E+00,
              0.98266923E+00, 0.50214672E+00, 0.11235259E+00, 0.75492084E+00,
              0.13290592E+00, },
            {-0.35606391E+01, 0.50929976E+00, 0.30332544E+01, 0.46690202E+00,
             -0.93843836E+00, 0.33491963E+00,-0.23293011E+00, 0.92977613E+00,
              0.81740057E+00, },
            {-0.38443810E+00, 0.52484918E+00,-0.42313319E+00,-0.31547219E-01,
             -0.23082055E-01, 0.98553799E-01, 0.48045248E-01,-0.29371861E+00,
             -0.38383645E+00, },
            {-0.16131206E+01, 0.93166280E+00, 0.74328411E+00, 0.15688561E+00,
              0.21910542E+00, 0.11658286E+00, 0.44385305E+00, 0.28195140E+00,
             -0.46593127E+00, },
            { 0.43430272E+00,-0.11887656E+01,-0.43771577E+00, 0.14430624E+00,
             -0.56061119E+00,-0.36049953E+00,-0.12475537E+00,-0.52396154E+00,
             -0.43358773E-01, },
            { 0.28472085E+01,-0.78438592E+00,-0.16446294E+01,-0.40643516E+00,
              0.59259111E+00,-0.32095200E+00, 0.95902123E-01,-0.59256905E+00,
             -0.44462144E+00, },
            { 0.10362735E-02,-0.26521128E-01, 0.73475315E-03, 0.78305684E-01,
             -0.87907061E-01,-0.87850457E-02, 0.68914322E-02,-0.58439686E-02,
             -0.10781294E-01, },
            { 0.83343238E-02,-0.22290355E-01, 0.50302162E-02, 0.79541668E-01,
              0.76331988E-01,-0.69553126E-02,-0.10722160E-02,-0.23844005E-02,
             -0.24587914E-03, },
            { 0.73374845E-01, 0.11703365E+00, 0.75943522E-01,-0.60151584E-01,
              0.15490431E+00, 0.13585863E-01,-0.18192250E-01, 0.11195567E-01,
              0.70080832E-02, },
            {-0.29040046E-01, 0.32617461E-01,-0.60124036E-01,-0.11059102E+00,
             -0.43481726E-01, 0.27728144E-01, 0.21420596E-01,-0.26929742E-01,
              0.35609290E-01, },
            {-0.37986711E-01, 0.21261556E+00, 0.65299794E-01,-0.47157770E+00,
              0.18923242E+00, 0.89936638E-02, 0.30793028E-02, 0.21662014E-01,
              0.45056507E-01, },
            {-0.18105258E+00, 0.52625749E-01,-0.21388224E+00,-0.12751260E+00,
             -0.44269633E+00,-0.68748944E-01, 0.48228782E-01, 0.46695493E-01,
             -0.61157331E-01, },
            {-0.31314614E+00,-0.65667820E+00,-0.16174650E+00, 0.18607917E+00,
             -0.82700682E+00,-0.49523894E-01, 0.11539982E+00,-0.11283565E+00,
             -0.42533660E-02, },
            { 0.31864041E+00, 0.13107453E+00, 0.33589387E+00, 0.72368956E+00,
              0.48834573E-01,-0.47290842E-02,-0.90480998E-01, 0.11117186E+00,
             -0.19858018E+00, },
            { 0.16404297E-01,-0.35947222E+00,-0.23393697E+00, 0.24125105E+00,
             -0.40167990E+00,-0.59507195E-01,-0.12976588E-01, 0.16296888E-02,
             -0.64027645E-01, },
            { 0.21480629E+00,-0.16956283E+00, 0.42124182E+00, 0.31029803E+00,
              0.24321973E+00, 0.14924949E+00,-0.11708120E+00,-0.62503330E-01,
              0.91057926E-01, },
            { 0.29138744E+00, 0.99653769E+00, 0.11597233E+00,-0.23700126E+00,
              0.11012855E+01, 0.13280974E+00,-0.17497236E+00, 0.16628051E+00,
             -0.11821241E-01, },
            {-0.40243325E+00,-0.14012007E+00,-0.34927446E+00,-0.10288903E+01,
             -0.43111134E-01,-0.44211321E-01, 0.12559946E+00,-0.10422827E+00,
              0.19184583E+00, },
            {-0.12884370E-01,-0.28929866E-02, 0.62589170E-02,-0.43417537E-02,
              0.16956747E-02, 0.30290781E-01,-0.54206152E-01, 0.47972486E-02,
             -0.54383771E-02, },
            {-0.18483959E-02,-0.49267081E-02,-0.35054756E-02,-0.10494787E-02,
             -0.31706650E-03, 0.53946707E-01, 0.46015918E-01, 0.77006100E-02,
             -0.74489904E-02, },
            { 0.28184274E-01, 0.48737317E-01, 0.56083137E-02,-0.11658832E-01,
              0.10535412E-01,-0.77447735E-01, 0.14809875E-01,-0.54364088E-02,
              0.26957258E-02, },
            { 0.12003129E-01,-0.48028270E-03,-0.13672524E-01,-0.10574406E-01,
             -0.94266171E-02,-0.67775045E-02,-0.85156538E-01, 0.66752793E-04,
              0.83940942E-02, },
            { 0.25840148E-01,-0.82805589E-01, 0.31506196E-01, 0.41513219E-01,
             -0.28803995E-01,-0.18279599E+00,-0.70030414E-01, 0.43120938E-02,
             -0.44541989E-03, },
            {-0.10199835E-01,-0.34188926E-01,-0.19336939E-01, 0.11237562E-01,
             -0.19958831E-01, 0.87408088E-01,-0.20651507E+00,-0.62467060E-02,
              0.28983794E-01, },
            {-0.82333675E-02,-0.15517812E-01, 0.84783202E-02, 0.52443543E-03,
              0.13247513E-02, 0.27244100E-02, 0.20155314E-03, 0.27045680E-01,
              0.32722121E-02, },
            {-0.98290918E-02, 0.24848564E-02,-0.12258626E-01, 0.43838951E-02,
              0.21730985E-02, 0.14691027E-03, 0.10895627E-01,-0.70574167E-02,
              0.27331531E-01, },
            { 0.34686182E-01, 0.11546563E-01,-0.10812423E-01,-0.11996835E-01,
             -0.61492608E-02,-0.16286846E-01, 0.90343263E-02,-0.16508182E-02,
             -0.44392467E-01, },
            { 0.24883408E-01,-0.41023172E-01, 0.30125706E-01, 0.19361736E-01,
             -0.84361620E-02,-0.86162388E-02,-0.12216836E-01, 0.51103357E-01,
             -0.75027507E-04, },
            { 0.11774480E-01,-0.37965027E-02, 0.73318430E-02,-0.37572165E-02,
             -0.29526060E-02, 0.51077693E-02,-0.20714712E-02,-0.67285765E-02,
              0.69300095E-02, },
            { 0.76303934E-02, 0.18072840E-01,-0.92502227E-02,-0.23002878E-02,
             -0.76491982E-02, 0.66094738E-02, 0.57324427E-02,-0.36398084E-02,
              0.18937787E-02, },
            { 0.12494961E-03, 0.95274886E-02, 0.79445858E-02, 0.87751141E-02,
             -0.22462981E-02, 0.10836435E-02, 0.13684512E-02,-0.88894583E-03,
              0.21372179E-02, },
            {-0.11187471E-02,-0.17962216E-02,-0.84776217E-02,-0.32081937E-02,
             -0.57096705E-02, 0.51954137E-02, 0.20163737E-02,-0.19013528E-02,
              0.13429531E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir13F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.65998969E+01,-0.96180730E-01,-0.18148595E-01, 0.64034291E-01,
             -0.14069435E-01,-0.11540678E-01, 0.66070836E-02, 0.26406836E-01,
             -0.39686030E-03, 0.23210552E-01,-0.22450823E-02,-0.10104239E-01,
             -0.12631579E-01, },
            {-0.42413343E-01, 0.80642931E-01, 0.27393836E+00,-0.16345827E+00,
             -0.16899700E+00,-0.32240987E+00, 0.70448697E-01,-0.77763997E-01,
             -0.13244358E-01, 0.30214982E-01,-0.13424671E+00,-0.17871724E-01,
              0.15353709E-01, },
            { 0.17452030E+02, 0.14823139E+01,-0.10489835E+01,-0.31613788E+00,
              0.14407470E+01,-0.17907827E+00, 0.29510629E+00,-0.22444358E-01,
              0.21991573E+00,-0.10238969E+01, 0.21552382E+00, 0.43258831E+00,
              0.30413870E-01, },
            {-0.42824140E+01,-0.43370719E+01, 0.27843063E+01, 0.32597795E+01,
              0.52011843E+01, 0.38541880E+01,-0.19715509E+01, 0.21992905E+01,
              0.23029692E+01,-0.60228556E+00, 0.22066264E+01, 0.23972336E-01,
             -0.80908841E+00, },
            {-0.11069540E+03,-0.47234511E+01, 0.81807337E+01,-0.57918835E+01,
             -0.15985530E+02, 0.41825180E+01, 0.31950969E+00,-0.15979006E+01,
             -0.29805335E+00, 0.65984240E+01,-0.15701975E+01,-0.28786166E+01,
             -0.55098748E+00, },
            { 0.53160348E+01, 0.72372947E+01,-0.31718040E+02,-0.84363041E+01,
             -0.20782209E+02,-0.15876794E+02, 0.10275279E+02,-0.12953021E+02,
             -0.18293257E+02, 0.30118990E+01,-0.10225961E+02, 0.10002422E+01,
              0.57159657E+01, },
            { 0.22316840E+03, 0.88046808E+01,-0.19268261E+02, 0.27535299E+02,
              0.49745659E+02,-0.14210365E+02,-0.63509369E+01, 0.59100389E+01,
             -0.75693429E+00,-0.15410734E+02, 0.43323641E+01, 0.69434185E+01,
              0.26285546E+01, },
            { 0.29579544E+02, 0.18237551E+02, 0.82377556E+02, 0.84427804E+00,
              0.30343164E+02, 0.29563707E+02,-0.19898548E+02, 0.30717596E+02,
              0.48212425E+02,-0.59438219E+01, 0.19914419E+02,-0.42029824E+01,
             -0.14977191E+02, },
            {-0.20669727E+03,-0.11006776E+02, 0.17986933E+02,-0.39454205E+02,
             -0.59185928E+02, 0.16635517E+02, 0.11510051E+02,-0.69961228E+01,
              0.14310032E+01, 0.15335248E+02,-0.50885744E+01,-0.70506644E+01,
             -0.39190328E+01, },
            {-0.65092957E+02,-0.45056271E+02,-0.80958855E+02, 0.13474419E+02,
             -0.17513617E+02,-0.25489252E+02, 0.16221642E+02,-0.31921728E+02,
             -0.51784859E+02, 0.50485601E+01,-0.17261196E+02, 0.55830841E+01,
              0.16545364E+02, },
            { 0.74110527E+02, 0.55856385E+01,-0.58897896E+01, 0.18034472E+02,
              0.23959206E+02,-0.63647890E+01,-0.57852674E+01, 0.26369393E+01,
             -0.59193814E+00,-0.55258875E+01, 0.21282353E+01, 0.25619016E+01,
              0.18311784E+01, },
            { 0.34186604E+02, 0.23974506E+02, 0.27236195E+02,-0.89847775E+01,
              0.29195013E+01, 0.82554817E+01,-0.46716986E+01, 0.12058103E+02,
              0.19580334E+02,-0.15368044E+01, 0.54791775E+01,-0.23879936E+01,
             -0.65005140E+01, },
            {-0.24007400E-01, 0.21675975E+01, 0.11725757E+01, 0.50758090E-01,
              0.13093476E-01,-0.23201449E-01, 0.38342886E-01, 0.48389513E-01,
              0.63410657E-02, 0.44442564E-01, 0.19522274E-01,-0.27899284E-01,
             -0.14491891E-02, },
            {-0.34519996E-01,-0.11496038E+01, 0.21471732E+01,-0.14192484E+00,
              0.68746604E-01,-0.33114925E-01,-0.61117645E-01, 0.42638432E-01,
              0.90309959E-02,-0.28206121E-01,-0.93916431E-02,-0.10890185E-01,
              0.86654341E-02, },
            {-0.62337194E-01, 0.79887128E+00, 0.91313583E+00,-0.18493378E+00,
              0.26627880E+00, 0.28355405E+00, 0.51949298E+00, 0.11120747E+00,
              0.27354652E+00,-0.12354252E+00,-0.27485764E-01,-0.33132059E-02,
              0.10892819E+00, },
            {-0.26281741E+00, 0.44478518E+00, 0.60648137E+00,-0.41248816E+00,
              0.25587833E+00,-0.95881902E-01,-0.33539433E-01, 0.57145733E+00,
              0.13682042E+00,-0.25388572E+00,-0.10539193E+00, 0.25895587E+00,
             -0.85921496E-01, },
            { 0.27832451E+01, 0.16864960E+02, 0.40619507E+01,-0.20387976E+01,
              0.93537444E+00, 0.16030459E+01, 0.35865560E+00,-0.70939946E+00,
              0.63179672E+00,-0.16262046E+01, 0.22060844E+00, 0.10916179E+01,
             -0.26890588E+00, },
            {-0.38157611E+01,-0.31960356E+01, 0.17024626E+02, 0.35198452E+01,
              0.14320093E+01,-0.42652923E+00, 0.44432792E+00,-0.12067032E+00,
             -0.51051909E+00, 0.10688665E+01, 0.34004807E+00, 0.25413656E+00,
             -0.20112856E+00, },
            { 0.69541912E+01,-0.18067736E+02, 0.91149321E+01, 0.38701576E+00,
              0.36137350E+01,-0.53189068E+01,-0.11495462E+02,-0.10826284E+01,
             -0.27792187E+01, 0.27802820E+01, 0.12037225E+01,-0.14036686E+01,
             -0.12014115E+01, },
            { 0.12424923E+02,-0.22949505E+02,-0.12072407E+02, 0.37552617E+01,
             -0.35413680E+01, 0.57876983E+01, 0.79935116E+00,-0.93507385E+01,
             -0.10703611E+01, 0.37773585E+01, 0.11066062E+01,-0.40062089E+01,
             -0.51845562E+00, },
            {-0.17640097E+02,-0.10663563E+03, 0.76676788E+02, 0.15154503E+02,
             -0.11779245E+02,-0.14249015E+02,-0.23536947E+01, 0.50884314E+01,
             -0.77559652E+01, 0.10476252E+02,-0.28326011E+01,-0.80197392E+01,
              0.17742889E+01, },
            { 0.41513287E+02,-0.97553841E+02,-0.95567078E+02,-0.19530249E+02,
             -0.16382975E+02, 0.50004249E+01, 0.46859521E+00, 0.59307051E+00,
              0.33864222E+01,-0.85503206E+01,-0.49351854E+01,-0.36469856E+00,
              0.17972651E+01, },
            {-0.50485680E+02, 0.11387515E+03,-0.35196548E+02, 0.13647220E+02,
             -0.18866581E+02, 0.35348217E+02, 0.61966629E+02, 0.28466079E+01,
              0.12943146E+02,-0.17173031E+02,-0.69480538E+01, 0.11160711E+02,
              0.54356551E+01, },
            {-0.10127907E+03, 0.11782328E+03, 0.68567123E+02,-0.27491049E+02,
              0.10630718E+02,-0.37230160E+02,-0.21067333E+01, 0.51232288E+02,
              0.83989620E+01,-0.19422342E+02,-0.21974294E+01, 0.19062983E+02,
              0.78485546E+01, },
            { 0.37536377E+02, 0.21429555E+03,-0.35160019E+03,-0.42891769E+02,
              0.40961739E+02, 0.44047215E+02, 0.81962329E+00,-0.13827250E+02,
              0.28208015E+02,-0.26970650E+02, 0.91622019E+01, 0.21888941E+02,
             -0.39068761E+01, },
            {-0.13467381E+03, 0.43239706E+03, 0.16293761E+03, 0.55668144E+02,
              0.53183456E+02,-0.12969423E+02,-0.54627337E+01,-0.22841866E+01,
             -0.78471823E+01, 0.25674978E+02, 0.17792437E+02,-0.14446113E+01,
             -0.57079659E+01, },
            { 0.15792365E+03,-0.30741367E+03, 0.12066244E+02,-0.73300522E+02,
              0.21158731E+02,-0.10123546E+03,-0.13733231E+03, 0.33357602E-01,
             -0.29374790E+02, 0.42309250E+02, 0.14585931E+02,-0.30827669E+02,
             -0.13427900E+02, },
            { 0.27333496E+03,-0.24579082E+03,-0.17028867E+03, 0.87392265E+02,
             -0.19682516E+02, 0.87100410E+02, 0.28480635E+01,-0.12049124E+03,
             -0.31395855E+02, 0.46579369E+02,-0.16486955E+01,-0.39088516E+02,
             -0.23570745E+02, },
            {-0.31926086E+02,-0.18912199E+03, 0.49377734E+03, 0.51254738E+02,
             -0.55447197E+02,-0.56824291E+02, 0.64984670E+01, 0.14788384E+02,
             -0.39874298E+02, 0.31675289E+02,-0.11539883E+02,-0.25417494E+02,
              0.36541965E+01, },
            { 0.17872554E+03,-0.60567108E+03,-0.10871112E+03,-0.77059822E+02,
             -0.71071373E+02, 0.11870968E+02, 0.10181161E+02, 0.38203001E+01,
              0.61157165E+01,-0.31524561E+02,-0.23383059E+02, 0.37743144E+01,
              0.61833124E+01, },
            {-0.22133595E+03, 0.37228479E+03, 0.54766731E+02, 0.12162727E+03,
              0.21483109E+01, 0.12726328E+03, 0.13672382E+03,-0.70452433E+01,
              0.30163021E+02,-0.44665485E+02,-0.12745265E+02, 0.35419662E+02,
              0.16225985E+02, },
            {-0.29494229E+03, 0.23506215E+03, 0.19200447E+03,-0.10832544E+03,
              0.29829510E+02,-0.84662743E+02,-0.59570804E+01, 0.12735707E+03,
              0.48301052E+02,-0.52132950E+02, 0.77952442E+01, 0.36343361E+02,
              0.26745314E+02, },
            { 0.91377707E+01, 0.66141739E+02,-0.22206757E+03,-0.21895306E+02,
              0.25961342E+02, 0.26107534E+02,-0.57225599E+01,-0.50952439E+01,
              0.19250315E+02,-0.14170377E+02, 0.50944929E+01, 0.10638086E+02,
             -0.13205591E+01, },
            {-0.84391235E+02, 0.27314240E+03, 0.24791868E+02, 0.38730942E+02,
              0.33692513E+02,-0.33924561E+01,-0.61104679E+01,-0.23151672E+01,
             -0.92081583E+00, 0.13415736E+02, 0.10256060E+02,-0.23337302E+01,
             -0.19138920E+01, },
            { 0.11088846E+03,-0.16574039E+03,-0.44082218E+02,-0.64954468E+02,
             -0.86842918E+01,-0.57481571E+02,-0.51243740E+02, 0.55801930E+01,
             -0.11060497E+02, 0.16775259E+02, 0.39023740E+01,-0.14520377E+02,
             -0.73669624E+01, },
            { 0.11177200E+03,-0.84463303E+02,-0.81101242E+02, 0.44699486E+02,
             -0.19768431E+02, 0.28911020E+02, 0.49595222E+01,-0.49829720E+02,
             -0.25354982E+02, 0.21757786E+02,-0.52241926E+01,-0.12592972E+02,
             -0.10411530E+02, },
            { 0.30574529E-01,-0.31148717E-01,-0.33901516E-01,-0.87049460E+00,
              0.20368204E+00,-0.32607842E-01,-0.16634515E-01, 0.27669171E-01,
              0.36947880E-01, 0.47631604E-02, 0.34091142E-02,-0.23116291E-01,
              0.21287642E-01, },
            { 0.58895037E-01, 0.15586518E-01, 0.69088846E-01,-0.21127127E+00,
             -0.80712038E+00,-0.14075822E-01,-0.17555034E-01,-0.16835840E-01,
              0.10679643E-01,-0.12162154E-01,-0.73469672E-02,-0.37808849E-02,
              0.10991140E-02, },
            { 0.62006795E+00, 0.13560387E+00, 0.16468419E+00, 0.99273920E-01,
             -0.26552621E+00, 0.37287197E+00,-0.35812023E+00,-0.19624999E-01,
              0.44019237E-01,-0.16461620E+00, 0.20510979E-01,-0.11776922E+00,
              0.38004402E-01, },
            { 0.10694618E+01, 0.31023821E+00, 0.14203931E+00,-0.43358117E-01,
              0.10166671E+00, 0.41650143E+00, 0.48760930E+00,-0.12005379E+00,
              0.97967356E-01,-0.15563516E+00,-0.22586857E+00, 0.49013641E-01,
             -0.20813970E-01, },
            { 0.18206578E+01,-0.41084620E+00, 0.27921236E+01, 0.24209125E+01,
              0.16685303E+01,-0.28808552E+00,-0.90014291E+00,-0.50827670E+00,
             -0.45300135E+00,-0.35001542E-01, 0.36338080E-01, 0.24220033E+00,
             -0.31733364E+00, },
            { 0.43253836E+00,-0.63174385E+00, 0.50231868E+00,-0.17871255E+01,
              0.42565956E+01,-0.41577759E+00,-0.12443371E-01, 0.90635931E+00,
             -0.81871720E-02, 0.12240939E+00,-0.58664852E+00, 0.16192600E-01,
             -0.34534615E-01, },
            {-0.39647148E+01, 0.33445108E+00,-0.17875113E+01, 0.35496018E+01,
              0.51491537E+01,-0.33950496E+01, 0.13043875E+01, 0.25871399E+00,
              0.68523347E-01, 0.13834658E+01, 0.79948354E+00, 0.91803938E+00,
             -0.45803875E-01, },
            {-0.64467540E+01,-0.20690763E+01, 0.49838591E+00,-0.82725680E+00,
              0.50504017E+01,-0.24308875E+01,-0.33799708E+01, 0.58349323E+00,
             -0.12550083E+01, 0.10462335E+01, 0.23345358E+01,-0.70762014E+00,
              0.51471233E+00, },
            {-0.15008719E+02, 0.14641236E+02,-0.20451660E+02, 0.19731981E+02,
              0.55634995E+01, 0.25407569E+01, 0.62443280E+01, 0.28349652E+01,
              0.24219623E+01,-0.96738261E+00,-0.41830257E+00,-0.83132789E-01,
              0.83228946E+00, },
            {-0.45921922E+01, 0.10040653E+02,-0.35055692E+01,-0.40767598E+01,
              0.60929508E+01, 0.75083405E+00, 0.31848945E-01,-0.46174974E+01,
             -0.34868580E+00, 0.49051598E+00, 0.39066932E+01, 0.55248684E+00,
             -0.44293070E+00, },
            { 0.63590775E+01,-0.54019079E+01, 0.35216527E+01,-0.11525576E+02,
             -0.16511587E+02, 0.75384369E+01,-0.19127115E+01,-0.46861181E+00,
             -0.16359047E+01,-0.28776963E+01,-0.27371666E+01,-0.23561432E+01,
             -0.15143438E+00, },
            { 0.10229012E+02, 0.11192629E+01,-0.24896045E+01, 0.21574695E+01,
             -0.16082794E+02, 0.50335708E+01, 0.60995483E+01, 0.46528451E-01,
              0.44228539E+01,-0.24760296E+01,-0.58701057E+01, 0.17978436E+01,
             -0.14681013E+01, },
            { 0.34721951E+02,-0.42912174E+02, 0.43504799E+02,-0.63420731E+02,
             -0.24599474E+02,-0.63787656E+01,-0.11826031E+02,-0.60871482E+01,
             -0.45813766E+01, 0.32331924E+01, 0.11356505E+01,-0.13086146E+01,
             -0.20494525E+00, },
            { 0.13705240E+02,-0.26223623E+02, 0.63174248E+01, 0.21032364E+02,
             -0.33421906E+02, 0.69371116E+00, 0.33373719E+00, 0.82973967E+01,
              0.57793902E-02,-0.35463052E+01,-0.80484915E+01,-0.15746634E+01,
              0.11740108E+01, },
            {-0.34624617E+01, 0.80138206E+01,-0.30727046E+01, 0.99697714E+01,
              0.14296803E+02,-0.47378635E+01, 0.91392756E+00, 0.25849405E+00,
              0.23256736E+01, 0.17453274E+01, 0.22083228E+01, 0.17524531E+01,
              0.13290420E+00, },
            {-0.44886589E+01, 0.27600312E+01, 0.26766870E+01,-0.16193961E+01,
              0.12873305E+02,-0.35492828E+01,-0.32382324E+01,-0.82338858E+00,
             -0.40521255E+01, 0.16536398E+01, 0.42015963E+01,-0.11780934E+01,
              0.11460971E+01, },
            {-0.24425175E+02, 0.32594193E+02,-0.26974190E+02, 0.44860214E+02,
              0.18794016E+02, 0.41103082E+01, 0.71137915E+01, 0.43121295E+01,
              0.26971650E+01,-0.24290116E+01,-0.88320661E+00, 0.13616098E+01,
             -0.47190940E+00, },
            {-0.12868726E+02, 0.17487787E+02,-0.35085900E+01,-0.16604246E+02,
              0.24305555E+02,-0.93955803E+00,-0.42676285E+00,-0.49368086E+01,
              0.68438685E+00, 0.36197155E+01, 0.52519341E+01, 0.11007043E+01,
             -0.74931699E+00, },
            {-0.15135266E+00,-0.28601021E-02,-0.89310557E-01,-0.33103667E-01,
              0.79293363E-01,-0.50901473E+00,-0.82532144E+00, 0.27812915E-01,
              0.20369733E-01, 0.24551898E-01, 0.18150743E-01,-0.15981859E-01,
              0.17871849E-01, },
            { 0.47308975E-02,-0.10820430E+00, 0.30828558E-01, 0.20033805E-01,
              0.29000906E-01, 0.75493073E+00,-0.47701424E+00, 0.32587885E-02,
              0.46372235E-01,-0.10304347E-01,-0.20734092E-01, 0.25965128E-01,
             -0.58355234E-02, },
            {-0.30175501E-01, 0.29149768E+00,-0.15922238E+00,-0.23880073E+00,
             -0.31934550E+00,-0.21410786E+00,-0.14080049E+00,-0.13714211E+00,
              0.11907785E+00, 0.39285958E-01, 0.14231239E+00, 0.47506593E-01,
             -0.21241790E-01, },
            {-0.11470716E+00, 0.12420133E+00,-0.40786356E-01,-0.17982394E+00,
              0.13026784E+00, 0.35654244E+00,-0.31161553E+00,-0.22107504E+00,
              0.22054937E-01,-0.22563413E-01,-0.27322270E-01, 0.78594657E-02,
              0.29850721E-01, },
            { 0.91586936E+00, 0.45026922E+00, 0.85786003E+00, 0.57185078E+00,
             -0.34676486E+00, 0.13745323E+01, 0.23583150E+01,-0.16617003E+00,
             -0.55329833E-01,-0.98012596E-01,-0.80390774E-01, 0.74126594E-01,
             -0.24377277E-01, },
            {-0.41545358E+00, 0.12106431E+01,-0.93434161E+00, 0.17056143E+00,
             -0.64522797E+00,-0.23836486E+01, 0.16357204E+01,-0.10281745E+00,
             -0.50418901E+00, 0.29982610E-01,-0.91422498E-01,-0.29078582E+00,
              0.26920128E-02, },
            {-0.32012030E+00,-0.33131790E+00, 0.44400707E+00, 0.35786450E+00,
              0.46332926E+00, 0.39637893E+00, 0.22233669E+00, 0.42943186E+00,
             -0.11153209E+00,-0.72281212E-01,-0.15658084E+00,-0.14401633E+00,
              0.95212579E-01, },
            { 0.34326747E+00,-0.40543489E-01, 0.20303135E+00,-0.67969397E-01,
             -0.69856882E-01,-0.80425566E+00, 0.52323300E+00, 0.22219919E+00,
              0.59583042E-01, 0.29945867E-01, 0.43922316E-01,-0.15566495E-01,
             -0.91002490E-02, },
            {-0.44131538E+00,-0.30327237E+00,-0.92330801E+00,-0.77402663E+00,
              0.13792758E+00,-0.20475721E+01,-0.27376485E+01, 0.10802826E+00,
             -0.65257788E-01, 0.78867815E-01, 0.10399983E+00, 0.14911513E-01,
             -0.53068724E-01, },
            { 0.66848314E+00,-0.16395499E+01, 0.17826374E+01,-0.61707690E-01,
              0.10942951E+01, 0.34093125E+01,-0.27390239E+01, 0.11946615E+00,
              0.77287668E+00, 0.62772550E-01, 0.15083092E+00, 0.45553604E+00,
             -0.37019581E-01, },
            { 0.21583674E-01, 0.64543181E-03,-0.51576743E-03, 0.58179975E-01,
              0.27645398E-01, 0.17042680E-01, 0.23695769E-01, 0.30250359E+00,
             -0.15979145E+00, 0.11132521E-01, 0.16834151E-01,-0.14365312E-01,
             -0.20031256E-02, },
            { 0.37123345E-01, 0.31039350E-01,-0.32179166E-01, 0.90723867E-02,
             -0.57725228E-01, 0.15170093E-01, 0.13725638E-01, 0.11982990E+00,
              0.30412433E+00, 0.16133197E-01, 0.21785465E-02, 0.67538638E-02,
             -0.11140665E-01, },
            {-0.14433794E-02, 0.11381988E+00,-0.10765873E+00,-0.10239576E+00,
             -0.81107587E-01,-0.61356392E-01, 0.38089089E-01, 0.56286909E-01,
              0.27529791E-01,-0.90354122E-01, 0.81266269E-01, 0.29821135E-03,
              0.13847735E-01, },
            { 0.18724556E+00, 0.95570683E-02,-0.21760223E-01, 0.13237505E+00,
              0.12431903E+00,-0.66988245E-01,-0.10429621E+00,-0.10871855E-01,
             -0.82563132E-03,-0.46499770E-01,-0.35336360E-01, 0.74498006E-02,
             -0.50428421E-02, },
            {-0.87156013E-01, 0.10371949E-01,-0.10520714E+00, 0.13535339E-01,
             -0.12469280E-01, 0.89040743E-02, 0.31624246E-01, 0.37560670E-02,
             -0.11854895E-01, 0.15747434E+00, 0.22601800E+00,-0.16246822E-01,
             -0.11745606E-01, },
            { 0.61789237E-01, 0.55235770E-01,-0.75306934E-02,-0.16015360E-01,
             -0.22144977E-01, 0.42483758E-01,-0.94584087E-02,-0.29955334E-04,
             -0.20362385E-01,-0.22954023E+00, 0.14123079E+00, 0.11846971E-01,
             -0.48866938E-02, },
            { 0.41830935E-01,-0.25836842E-01, 0.33995997E-01,-0.30328436E-02,
             -0.43513537E-02, 0.32394242E-01, 0.24555236E-01, 0.15828688E-01,
             -0.14602708E-01, 0.79171928E-02,-0.24579484E-02,-0.12720145E+00,
              0.61430719E-01, },
            { 0.11831303E-01, 0.48629563E-01,-0.36074687E-01, 0.10915471E-01,
              0.65989299E-02, 0.81955465E-02,-0.26336169E-01,-0.81899911E-02,
             -0.11951329E-01,-0.18219600E-01,-0.14673570E-01,-0.55252794E-01,
             -0.12856482E+00, },
            {-0.18628046E-01,-0.58445823E-02,-0.82438067E-01, 0.24269113E-01,
              0.24777478E-01,-0.10445634E-01,-0.56395708E-02, 0.88527258E-02,
             -0.72764154E-02, 0.19107535E-03,-0.24804156E-01, 0.88188238E-02,
              0.33777929E-02, },
            { 0.22041786E-01, 0.75239100E-01, 0.32701142E-01,-0.36235764E-02,
              0.28052095E-01, 0.17236874E-01, 0.44026822E-03,-0.37383433E-01,
              0.11369155E-02, 0.21209640E-02, 0.54265251E-02, 0.13008385E-02,
              0.85698189E-02, },
            { 0.10458428E-01,-0.10118044E-02,-0.77243894E-01, 0.30200144E-04,
              0.16539751E-01,-0.37585138E-02, 0.45206282E-01,-0.19975781E-01,
              0.64657149E-02, 0.45526298E-02,-0.13365079E-01, 0.17035095E-01,
              0.60681752E-02, },
            {-0.31197891E-01, 0.27110072E-01,-0.81724217E-02, 0.53124398E-01,
             -0.37347758E-03,-0.60616857E-02, 0.15380551E-01, 0.24982348E-01,
              0.43096016E-02,-0.35201445E-01,-0.23587819E-01,-0.47319382E-02,
              0.29201475E-02, },
         },
         {
            { 0.93594637E+01, 0.11758081E-01, 0.84277675E-01, 0.73569529E-01,
              0.11903162E-01,-0.44155386E-02,-0.15138906E-01, 0.74017155E-02,
             -0.15588634E-01, 0.36506962E-01,-0.14251899E-02,-0.65558045E-02,
             -0.18775851E-02, },
            { 0.88674414E+00, 0.51994503E-01, 0.62100869E-02,-0.30662936E+00,
             -0.13489079E+00,-0.16742069E+00,-0.27921739E+00,-0.73416762E-01,
             -0.51260434E-01, 0.34743927E-01,-0.64074233E-01,-0.14990614E-01,
             -0.33514559E-01, },
            { 0.29466911E+02,-0.10430388E+01,-0.17538368E+01, 0.64629501E+00,
              0.15527004E+01,-0.81919432E+00, 0.24933697E+00, 0.21972425E+00,
              0.14215918E+00,-0.14595740E+01, 0.24961294E+00, 0.30417073E+00,
             -0.55456981E-01, },
            {-0.67446027E+01, 0.40613857E+00, 0.46058164E+01, 0.29142275E+01,
              0.43435960E+01, 0.19936315E+01, 0.27405980E+01, 0.19893142E+01,
              0.26587334E+01,-0.36289024E+00, 0.87958097E+00, 0.43432501E+00,
              0.45760855E+00, },
            {-0.14369508E+03, 0.82375011E+01, 0.86440678E+01,-0.11147148E+02,
             -0.15416179E+02, 0.78732028E+01, 0.98172742E+00,-0.10455229E+01,
              0.79332030E+00, 0.88577414E+01,-0.20928278E+01,-0.17485448E+01,
             -0.37013805E+00, },
            {-0.38804274E+01,-0.24045439E+02,-0.22236847E+02,-0.39119647E+01,
             -0.14422157E+02,-0.11314327E+02,-0.11605090E+02,-0.11023640E+02,
             -0.19990587E+02, 0.18404472E+01,-0.36433988E+01,-0.19584160E+01,
             -0.16701306E+01, },
            { 0.24726463E+03,-0.13912273E+02,-0.16638878E+02, 0.39193115E+02,
              0.45437634E+02,-0.20975456E+02,-0.71870809E+01, 0.10358952E+01,
             -0.41288266E+01,-0.20344755E+02, 0.61743193E+01, 0.37249963E+01,
              0.23654854E+01, },
            { 0.73438400E+02, 0.85878952E+02, 0.28821287E+02,-0.12087472E+02,
              0.15061758E+02, 0.27009087E+02, 0.25284414E+02, 0.25017273E+02,
              0.52237434E+02,-0.39412918E+01, 0.60525670E+01, 0.32629383E+01,
              0.20258255E+01, },
            {-0.20673763E+03, 0.37566130E+01, 0.14237963E+02,-0.50470264E+02,
             -0.52279778E+02, 0.21282154E+02, 0.11163842E+02, 0.52002543E+00,
              0.55029087E+01, 0.20240622E+02,-0.74278593E+01,-0.33429351E+01,
             -0.35941296E+01, },
            {-0.11962262E+03,-0.10679832E+03,-0.54719863E+01, 0.29441711E+02,
             -0.20143695E+01,-0.27501472E+02,-0.26564098E+02,-0.25437061E+02,
             -0.56287434E+02, 0.34835873E+01,-0.41210556E+01,-0.22583303E+01,
             -0.41311565E+00, },
            { 0.69616798E+02, 0.29591665E+01,-0.46051245E+01, 0.21710459E+02,
              0.20603981E+02,-0.72812696E+01,-0.51570253E+01,-0.75040758E+00,
             -0.23034899E+01,-0.73399339E+01, 0.31145618E+01, 0.10590287E+01,
              0.16675494E+01, },
            { 0.55920959E+02, 0.44685562E+02,-0.57169757E+01,-0.16177469E+02,
             -0.28279176E+01, 0.99622288E+01, 0.10504270E+02, 0.95728416E+01,
              0.21457167E+02,-0.10309284E+01, 0.88172013E+00, 0.53643727E+00,
             -0.38652459E+00, },
            {-0.91051519E-01, 0.16604546E+01, 0.15699625E+01, 0.76715745E-01,
              0.43537375E-01,-0.13598571E-01, 0.49196340E-01, 0.54983906E-02,
             -0.43234859E-01, 0.35610504E-01, 0.33476464E-01,-0.23877524E-01,
             -0.35067947E-03, },
            { 0.12912868E+00,-0.15785012E+01, 0.15362040E+01,-0.14700162E+00,
              0.59193391E-01,-0.35832047E-01,-0.21853298E-01, 0.66302014E-02,
              0.16508466E-01, 0.10124203E-01, 0.11756688E-01,-0.16161023E-01,
             -0.59979502E-02, },
            {-0.84079015E+00, 0.12545848E+01,-0.53319615E+00,-0.38466421E+00,
              0.71295905E+00, 0.44927454E+00, 0.46575686E+00, 0.96385442E-01,
              0.24827370E+00,-0.49310956E-01,-0.60446393E-01,-0.82794242E-01,
              0.53030346E-01, },
            { 0.93256634E+00,-0.20340858E+00, 0.24310616E+00,-0.10172791E+00,
              0.75127995E+00, 0.97861588E-01, 0.30854341E+00, 0.39208809E+00,
              0.25966071E-01,-0.73293775E-01,-0.10588093E+00, 0.91211334E-01,
             -0.16803257E+00, },
            { 0.31397121E+01, 0.17180098E+02,-0.13800441E+02, 0.70740975E-01,
              0.46341306E+00, 0.14315274E+00,-0.44296429E-01, 0.59252042E+00,
              0.15348610E+01,-0.11630143E+01,-0.22830707E+00, 0.94050133E+00,
             -0.27207699E+00, },
            {-0.83226357E+01, 0.16027706E+02, 0.18227182E+02, 0.24127638E+01,
              0.16124903E+01, 0.64143932E+00,-0.48962837E+00, 0.14949485E+01,
             -0.68322599E+00,-0.83393383E+00,-0.21246544E+00, 0.36369675E+00,
              0.15471225E-01, },
            { 0.25553751E+02,-0.20131190E+01, 0.26365784E+02,-0.37747600E+01,
             -0.64699745E+01,-0.78558393E+01,-0.10889877E+02,-0.23848505E+00,
             -0.30194924E+01, 0.15077212E+01, 0.17867323E+01, 0.61363947E+00,
             -0.83897740E+00, },
            {-0.97228460E+01,-0.83611345E+01, 0.63290434E+01, 0.16351886E+01,
             -0.20032827E+02, 0.44843369E+01,-0.30204687E+01,-0.67688684E+01,
              0.79144502E+00, 0.25467318E+00, 0.99470073E+00,-0.17565970E+01,
              0.14626296E+01, },
            {-0.19355894E+02,-0.47454216E+02, 0.12098093E+03,-0.89022665E+01,
             -0.42745748E+01,-0.45515871E+01, 0.30617993E-01,-0.31281152E+01,
             -0.12787308E+02, 0.67629743E+01, 0.81136113E+00,-0.72239995E+01,
              0.21611481E+01, },
            { 0.77719887E+02,-0.15284087E+03,-0.28094072E+02,-0.12935537E+02,
             -0.18593321E+02,-0.32531598E+01, 0.55940328E+01,-0.11318358E+02,
              0.32409825E+01, 0.75526614E+01, 0.48537755E+00,-0.15223314E+01,
              0.88688987E+00, },
            {-0.16509648E+03,-0.13600343E+02,-0.14347475E+03, 0.67285439E+02,
              0.29329416E+02, 0.45942951E+02, 0.55822559E+02,-0.64459381E+01,
              0.13185073E+02,-0.11480048E+02,-0.14453205E+02,-0.30418870E+00,
              0.47256918E+01, },
            { 0.55850414E+02, 0.47293289E+02,-0.38787140E+02,-0.14584154E+02,
              0.12430083E+03,-0.31622999E+02, 0.13420869E+02, 0.35762138E+02,
             -0.36125880E+00,-0.15479681E+01,-0.12744308E+01, 0.86030378E+01,
             -0.44953423E+01, },
            { 0.38700024E+02,-0.28222193E+02,-0.30419247E+03, 0.39386944E+02,
              0.15715430E+02, 0.20906570E+02,-0.27541671E+01, 0.65622759E+01,
              0.40180477E+02,-0.16013824E+02,-0.15492724E+01, 0.20418951E+02,
             -0.60262113E+01, },
            {-0.24083032E+03, 0.41435196E+03,-0.13463223E+03, 0.38008816E+02,
              0.62823856E+02, 0.78923759E+01,-0.17364424E+02, 0.29752794E+02,
             -0.36583562E+01,-0.22033194E+02,-0.86091971E+00, 0.23167665E+01,
             -0.40712471E+01, },
            { 0.43789868E+03, 0.37758240E+02, 0.30222437E+03,-0.25211705E+03,
             -0.68283501E+02,-0.12247941E+03,-0.11544855E+03, 0.29691120E+02,
             -0.26862541E+02, 0.32532574E+02, 0.40526028E+02,-0.55368447E+01,
             -0.12097772E+02, },
            {-0.16987953E+03,-0.98175888E+02, 0.93504860E+02, 0.44424995E+02,
             -0.31534625E+03, 0.75678284E+02,-0.27929533E+02,-0.75439156E+02,
             -0.14984094E+02, 0.86548948E+01,-0.38562276E+01,-0.19092201E+02,
              0.82323360E+01, },
            {-0.28264292E+02, 0.16086139E+03, 0.33062650E+03,-0.60156818E+02,
             -0.25561420E+02,-0.33909615E+02, 0.69856234E+01,-0.67728324E+01,
             -0.52560555E+02, 0.17889961E+02, 0.16138614E+01,-0.24223919E+02,
              0.70134716E+01, },
            { 0.30368719E+03,-0.46959631E+03, 0.33666513E+03,-0.55255219E+02,
             -0.86959816E+02,-0.90647640E+01, 0.22722628E+02,-0.31628498E+02,
             -0.20313892E+01, 0.26141891E+02, 0.15387295E+01,-0.10764118E+01,
              0.55649915E+01, },
            {-0.51269446E+03,-0.24774500E+02,-0.26924051E+03, 0.35456693E+03,
              0.72959488E+02, 0.14845412E+03, 0.10644101E+03,-0.44661758E+02,
              0.25646076E+02,-0.38323627E+02,-0.45829033E+02, 0.11802900E+02,
              0.13949016E+02, },
            { 0.23760629E+03, 0.85215302E+02,-0.10850819E+03,-0.47027061E+02,
              0.35732697E+03,-0.74357651E+02, 0.24879040E+02, 0.67221268E+02,
              0.34626274E+02,-0.15760853E+02, 0.95299788E+01, 0.20570358E+02,
             -0.99656239E+01, },
            { 0.57932053E+01,-0.10340132E+03,-0.13253651E+03, 0.30239813E+02,
              0.14212955E+02, 0.18223082E+02,-0.44180136E+01, 0.29476216E+01,
              0.24200890E+02,-0.79207344E+01,-0.70972711E+00, 0.10236231E+02,
             -0.29160755E+01, },
            {-0.13413786E+03, 0.19037309E+03,-0.19506718E+03, 0.28727461E+02,
              0.42836147E+02, 0.36783912E+01,-0.11200485E+02, 0.11366664E+02,
              0.33454440E+01,-0.10977008E+02,-0.10146281E+01,-0.15341072E+00,
             -0.23435018E+01, },
            { 0.21834229E+03,-0.54074174E+00, 0.81919060E+02,-0.17072711E+03,
             -0.27865332E+02,-0.65729042E+02,-0.36894901E+02, 0.22338470E+02,
             -0.91411514E+01, 0.15863236E+02, 0.18038870E+02,-0.67405629E+01,
             -0.59032989E+01, },
            {-0.11627517E+03,-0.24264631E+02, 0.48176723E+02, 0.13980582E+02,
             -0.15089851E+03, 0.25507414E+02,-0.73551226E+01,-0.20702442E+02,
             -0.20904865E+02, 0.87487831E+01,-0.55933886E+01,-0.86189528E+01,
              0.53419061E+01, },
            { 0.78886345E-01,-0.29637566E+00,-0.60464900E-01,-0.71206862E+00,
              0.59372735E+00,-0.20558244E-01, 0.48958015E-01, 0.17571801E-01,
              0.42676944E-02,-0.61578038E-02, 0.21608399E-01, 0.51746960E-03,
             -0.57962183E-02, },
            {-0.11099774E-01,-0.77226818E-01,-0.35840608E-01,-0.68324232E+00,
             -0.71531457E+00,-0.41512378E-01,-0.81707286E-02,-0.76341666E-02,
              0.20536058E-01, 0.75761302E-04, 0.10275762E-01,-0.93738467E-03,
              0.10511260E-02, },
            { 0.13842398E+01,-0.49234234E-01, 0.12677437E-01,-0.10571032E+01,
              0.10398120E+00, 0.59301776E+00,-0.14417675E+00,-0.12642118E+00,
             -0.50918173E-01,-0.13518077E+00, 0.30521570E-01,-0.12413521E+00,
             -0.55086080E-01, },
            { 0.13468280E+01, 0.45321479E+00, 0.33050817E-01,-0.19486007E+00,
             -0.84212863E+00, 0.33315787E+00, 0.56156081E+00,-0.21251965E+00,
              0.18369167E+00,-0.45716692E-01,-0.15525161E+00,-0.21887375E-02,
             -0.13814416E-01, },
            { 0.13547544E+01, 0.35052302E+01, 0.39791517E+01,-0.39452157E+01,
              0.91488066E+01,-0.59917301E+00,-0.23404021E+01, 0.19902436E+00,
              0.28914595E+00, 0.32449687E+00,-0.30973130E+00,-0.13093148E+00,
              0.24543998E+00, },
            { 0.93270177E+00, 0.10100644E+01, 0.21535051E+01,-0.79689913E+01,
             -0.27236195E+01, 0.71202469E+00,-0.46545181E+00, 0.11434031E+01,
             -0.18245190E+00,-0.40822005E+00,-0.56782430E+00,-0.14069527E-01,
             -0.91296121E-01, },
            {-0.10542483E+02, 0.11633001E+00,-0.27338638E+01, 0.60308566E+01,
              0.38688369E+01,-0.47813067E+01,-0.80322623E-01, 0.11000689E+01,
              0.13760308E+01, 0.10471877E+01,-0.10195704E+00, 0.84878850E+00,
              0.67848611E+00, },
            {-0.11052456E+02,-0.35547700E+01,-0.17437516E+01,-0.14164073E+01,
              0.29000556E+01,-0.17477272E+00,-0.31872108E+01, 0.15479761E+01,
             -0.11199074E+01, 0.35770193E-01, 0.15186522E+01,-0.51870143E+00,
              0.42309412E+00, },
            {-0.12578057E+02,-0.68359613E+01,-0.30894140E+02, 0.29231524E+02,
             -0.37497784E+02, 0.42103472E+01, 0.13634277E+02,-0.23105447E+01,
             -0.25411386E+01,-0.20200241E+01, 0.17991817E+01, 0.12031139E+01,
             -0.22861779E+01, },
            {-0.45937929E+01, 0.20954571E+01,-0.12322830E+02, 0.30053967E+02,
              0.19175726E+02,-0.63447361E+01, 0.20542054E+01,-0.71390734E+01,
             -0.82132196E+00, 0.34279492E+01, 0.34167354E+01, 0.95250791E+00,
             -0.24310125E+00, },
            { 0.24285011E+02,-0.33831913E+01, 0.87184391E+01,-0.14099850E+02,
             -0.16883636E+02, 0.11646293E+02, 0.59411901E+00,-0.24929807E+01,
             -0.56281371E+01,-0.20639105E+01, 0.36040890E+00,-0.18116419E+01,
             -0.17643514E+01, },
            { 0.24702211E+02, 0.38919735E+01, 0.61029816E+01, 0.65146122E+01,
             -0.22756846E+01,-0.12218447E+01, 0.49939814E+01,-0.24445646E+01,
              0.28680899E+01,-0.10368446E+00,-0.39225392E+01, 0.16193440E+01,
             -0.12233027E+01, },
            { 0.28191441E+02, 0.78694773E+00, 0.64988586E+02,-0.59622383E+02,
              0.50546680E+02,-0.93377047E+01,-0.25691008E+02, 0.49692044E+01,
              0.66733646E+01, 0.38342435E+01,-0.33193257E+01,-0.28077452E+01,
              0.59086132E+01, },
            { 0.68308530E+01,-0.11145398E+02, 0.20142715E+02,-0.35770203E+02,
             -0.34637836E+02, 0.14650714E+02,-0.28380070E+01, 0.14030188E+02,
              0.30224795E+01,-0.93840456E+01,-0.67304063E+01,-0.30187075E+01,
              0.13366692E+01, },
            {-0.19719049E+02, 0.68867254E+01,-0.97942524E+01, 0.12571602E+02,
              0.16866741E+02,-0.80991716E+01,-0.60612911E+00, 0.14781255E+01,
              0.55502706E+01, 0.12907434E+01,-0.25038749E+00, 0.10915368E+01,
              0.11269934E+01, },
            {-0.16954172E+02, 0.31211653E+01,-0.45404301E+01,-0.60521469E+01,
              0.81650001E+00, 0.11353194E+01,-0.19928211E+01, 0.10430180E+01,
             -0.25809250E+01, 0.46962984E-01, 0.30409884E+01,-0.11915781E+01,
              0.96175605E+00, },
            {-0.18254309E+02, 0.40652580E+01,-0.38976635E+02, 0.37998058E+02,
             -0.22253872E+02, 0.54544687E+01, 0.15347375E+02,-0.28659797E+01,
             -0.52428341E+01,-0.22536197E+01, 0.18733271E+01, 0.18808317E+01,
             -0.43471789E+01, },
            {-0.42598801E+01, 0.68740029E+01,-0.86680889E+01, 0.13091699E+02,
              0.19776112E+02,-0.94230890E+01, 0.93041652E+00,-0.87512522E+01,
             -0.19371971E+01, 0.73051834E+01, 0.41424298E+01, 0.23743393E+01,
             -0.12347583E+01, },
            {-0.10308004E+00,-0.11345814E+00,-0.42388607E-01,-0.54478396E-01,
              0.37894849E-01,-0.71249014E+00,-0.99690706E+00, 0.42135198E-01,
             -0.17244507E-01, 0.18175337E-01, 0.22615835E-01, 0.40037422E-02,
             -0.35507255E-02, },
            {-0.27820380E-01,-0.77944390E-01, 0.14688867E+00, 0.16636178E-01,
              0.53211000E-01, 0.95628875E+00,-0.70051461E+00, 0.17104762E-01,
              0.31993400E-01,-0.93262866E-02,-0.18731829E-01, 0.20413624E-01,
             -0.50872733E-03, },
            { 0.29756284E+00, 0.12934245E+00,-0.18983012E+00, 0.71824878E-01,
             -0.41808215E+00,-0.11720640E+00,-0.41143602E+00,-0.28631784E-01,
              0.21437567E+00, 0.25426921E-01, 0.33951670E-01, 0.75097620E-01,
             -0.22934020E-01, },
            { 0.60831457E-01, 0.26703396E+00, 0.11492737E+00,-0.14872836E+00,
              0.15820681E+00, 0.73452604E+00,-0.15326773E+00,-0.20186380E+00,
              0.79028957E-01,-0.39597573E-02,-0.69052177E-02,-0.35170142E-02,
              0.39248485E-01, },
            { 0.77993286E+00, 0.11083744E+01, 0.43744072E+00, 0.59175074E+00,
              0.27468514E+00, 0.27732401E+01, 0.21687400E+01,-0.25144777E+00,
              0.20495271E-01,-0.25761271E-01, 0.13951994E-02,-0.70093565E-01,
              0.10886874E+00, },
            {-0.10175124E+01, 0.14403582E+01,-0.19194679E+01, 0.23482410E-01,
             -0.27122623E+00,-0.25169380E+01, 0.30849991E+01,-0.93018368E-01,
             -0.32557723E+00,-0.32845687E-01, 0.71069151E-01,-0.18071032E+00,
             -0.11195786E+00, },
            {-0.68460888E+00, 0.29701370E+00, 0.60023886E+00, 0.74086197E-01,
              0.98181754E+00, 0.17232137E-01, 0.71498251E+00, 0.15950713E+00,
             -0.32126063E+00,-0.25459521E-01, 0.10034926E+00,-0.21438281E+00,
              0.80607370E-01, },
            { 0.29350964E-01,-0.52844387E+00,-0.36183974E+00,-0.28149527E+00,
              0.92464551E-01,-0.15573705E+01,-0.51209643E-01, 0.32866797E+00,
             -0.38855325E-01, 0.12696608E-01, 0.45670737E-01,-0.69922097E-02,
             -0.22250839E-01, },
            {-0.49448678E+00,-0.11309274E+01,-0.43056571E+00,-0.97913963E+00,
             -0.75081772E+00,-0.50914221E+01,-0.24884198E+01, 0.29986811E+00,
              0.74715838E-02,-0.13069284E-01,-0.12120257E+00, 0.20640135E+00,
             -0.19321387E+00, },
            { 0.16908376E+01,-0.17448000E+01, 0.30932491E+01,-0.34096658E+00,
              0.17136328E-01, 0.35928824E+01,-0.54552412E+01, 0.32137137E-01,
              0.51082975E+00, 0.19546562E+00,-0.10672491E+00, 0.28139511E+00,
              0.41073278E-01, },
            {-0.29096520E-01, 0.21864381E-01, 0.12243150E-01, 0.32921236E-01,
              0.26921557E-01,-0.18893564E-01, 0.16802772E-02, 0.28694057E+00,
             -0.29211941E+00, 0.16393879E-01, 0.94393184E-02,-0.60703461E-02,
             -0.10121063E-02, },
            { 0.22870019E-01, 0.40416889E-01, 0.20207793E-02, 0.28009053E-01,
             -0.47511235E-01, 0.16151302E-01, 0.41487203E-02, 0.24980807E+00,
              0.29251510E+00,-0.41429587E-02, 0.31435888E-02, 0.86209821E-02,
             -0.10191371E-01, },
            {-0.31774394E-01, 0.34211457E-01,-0.82030557E-01, 0.80120198E-01,
              0.52378416E-01,-0.12363158E+00, 0.10513846E+00, 0.79557329E-01,
             -0.65318130E-01,-0.10646767E+00, 0.76470129E-01, 0.25804231E-02,
              0.20964760E-01, },
            { 0.12791419E+00, 0.45520272E-02, 0.28757790E-01, 0.62985376E-01,
              0.84955081E-01,-0.12073985E+00,-0.98188937E-01, 0.52158490E-01,
              0.69911316E-01,-0.42228647E-01,-0.41942727E-01,-0.53663673E-04,
              0.20629911E-01, },
            {-0.84972277E-01, 0.74791508E-02,-0.72876997E-01, 0.54825362E-01,
              0.78998553E-02, 0.21651424E-02, 0.11295615E-01,-0.22708066E-01,
             -0.24967110E-01, 0.23709922E+00, 0.20721011E+00,-0.75820675E-02,
             -0.66711716E-02, },
            { 0.91338418E-01, 0.75325817E-01,-0.11707541E-01,-0.18225781E-02,
             -0.68179360E-02, 0.26803995E-01, 0.29882748E-01,-0.50773583E-02,
             -0.38290662E-02,-0.20912644E+00, 0.21416292E+00,-0.53014228E-03,
              0.60032657E-02, },
            { 0.42857230E-01,-0.81398398E-01,-0.70250421E-02, 0.28814852E-01,
              0.10183385E-01, 0.45627076E-01, 0.34210838E-01,-0.13469204E-01,
             -0.16962208E-01, 0.95584551E-02, 0.76674414E-02,-0.11811175E+00,
              0.10944934E+00, },
            { 0.37916854E-01, 0.34320731E-01,-0.38995236E-01,-0.11582067E-01,
              0.56862938E-02, 0.19178797E-01,-0.19646440E-01,-0.19834880E-02,
             -0.70472504E-02,-0.87856855E-02,-0.38532328E-02,-0.12558445E+00,
             -0.99743478E-01, },
            {-0.44582158E-01,-0.84324472E-01,-0.50034769E-01, 0.46942960E-01,
              0.46193022E-02, 0.71702013E-02,-0.12611548E-01, 0.95665868E-03,
             -0.12036865E-01,-0.67090063E-04,-0.20105047E-01, 0.25720501E-02,
              0.19281610E-02, },
            {-0.26651995E-01, 0.87436706E-01, 0.20321304E-01,-0.69048993E-01,
              0.10091052E-02, 0.10883666E-01,-0.17108636E-01,-0.87443851E-02,
              0.57800952E-02, 0.46012336E-02, 0.20599568E-01,-0.53997170E-02,
              0.58125644E-02, },
            {-0.57532750E-02, 0.34642294E-01, 0.31471007E-01,-0.11714598E-01,
             -0.16319931E-01,-0.50641306E-01, 0.25461499E-01, 0.27247524E-01,
              0.11968782E-01, 0.18001130E-01,-0.22403058E-01, 0.21385524E-01,
              0.13279283E-01, },
            {-0.25084522E-01, 0.21860242E-01, 0.28222946E-02, 0.35862558E-01,
              0.60812447E-01, 0.43446873E-02, 0.19660845E-01, 0.30596415E-01,
             -0.88513829E-02,-0.24848016E-01,-0.96459324E-02,-0.90757394E-02,
             -0.20335526E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir13Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.29977396E+01,-0.81961853E-02, 0.18906312E-01, 0.31932625E-02,
             -0.40838039E-02,-0.23357221E-03, 0.76571512E-02,-0.79752160E-02,
              0.34400288E-03, },
            { 0.20014945E+00,-0.29716199E-01, 0.52193105E-01, 0.37738375E-01,
             -0.50039222E-02,-0.29981710E-01,-0.13986629E-01, 0.22926540E-02,
             -0.21308563E-02, },
            { 0.13428990E+01, 0.44498194E-01,-0.18523586E+00,-0.17925259E+00,
              0.45948785E-01, 0.47860365E-01,-0.33176940E-01, 0.42967819E-01,
              0.43051202E-01, },
            {-0.47216845E+00,-0.16368178E+00,-0.10566967E+00,-0.59037436E-01,
             -0.33112377E-01, 0.11583864E+00, 0.14226819E-01, 0.78643629E-04,
              0.69223372E-02, },
            {-0.23492818E+01, 0.67165658E-01, 0.47879127E+00, 0.52358502E+00,
             -0.15943331E+00,-0.14764257E+00, 0.48829701E-01,-0.51714528E-01,
             -0.10612427E+00, },
            { 0.24342524E+00, 0.16540094E+00, 0.64394549E-01, 0.85904840E-02,
              0.40568676E-01,-0.88596366E-01,-0.20141602E-02,-0.10498879E-02,
             -0.26079479E-02, },
            { 0.11590286E+01,-0.87880895E-01,-0.30248937E+00,-0.36426729E+00,
              0.13489676E+00, 0.10174665E+00,-0.20617876E-01, 0.17740101E-01,
              0.64807892E-01, },
            { 0.14391306E-01,-0.21309042E+00,-0.33884123E+00, 0.10892765E-01,
             -0.42811530E-02,-0.25248926E-02, 0.12565411E-01,-0.44771405E-02,
              0.69199535E-02, },
            { 0.26753715E-02, 0.35680091E+00,-0.20014954E+00,-0.98086214E-02,
              0.87794177E-02,-0.34690208E-02,-0.91818944E-02, 0.56939302E-02,
              0.10811757E-01, },
            {-0.10302394E+00, 0.19964647E+00, 0.12284161E+00, 0.62227905E-01,
              0.39830696E-01,-0.31292230E-01,-0.18277895E-01, 0.34071945E-01,
             -0.24559934E-01, },
            { 0.73183656E-01,-0.31019021E-01, 0.22225291E+00,-0.79470731E-01,
             -0.39866902E-01, 0.20936769E-01,-0.13295945E-01, 0.57061128E-01,
              0.55259675E-01, },
            {-0.22905074E+00, 0.90782636E+00, 0.16330606E+01, 0.19163447E-02,
              0.35776563E-01, 0.49570456E-01,-0.11902388E+00, 0.45402091E-01,
             -0.11314365E+00, },
            {-0.39456546E+00,-0.17365061E+01, 0.86957514E+00, 0.16775037E+00,
             -0.85004143E-01, 0.42578947E-01, 0.72186656E-01,-0.78504875E-01,
             -0.12058398E+00, },
            { 0.13912698E+01,-0.96152496E+00,-0.60672927E+00,-0.13106681E-01,
             -0.41902941E+00, 0.11963305E+00, 0.62974300E-02,-0.18483321E+00,
              0.25544432E+00, },
            { 0.54170632E+00, 0.34240028E+00,-0.99347818E+00, 0.24775450E+00,
              0.47574925E+00,-0.18045484E+00, 0.13595828E+00,-0.61139083E+00,
             -0.26010543E+00, },
            { 0.74438137E+00,-0.80979353E+00,-0.17163999E+01,-0.35758594E+00,
             -0.14151997E-01,-0.13134675E+00, 0.27793384E+00,-0.11998469E+00,
              0.27071473E+00, },
            { 0.13506950E+01, 0.16267354E+01,-0.12598553E+01,-0.31339702E+00,
              0.68517044E-01,-0.18701576E+00,-0.86923413E-01, 0.14469229E+00,
              0.22702868E+00, },
            {-0.34414785E+01, 0.10698559E+01, 0.14466070E+01,-0.78836642E-01,
              0.10688086E+01,-0.78987226E-01, 0.20188455E+00, 0.36264941E+00,
             -0.72390538E+00, },
            {-0.21533570E+01,-0.86708015E+00, 0.16319656E+01,-0.28797507E+00,
             -0.10868007E+01, 0.50646031E+00,-0.39900690E+00, 0.15120949E+01,
              0.29860815E+00, },
            {-0.48577079E+00, 0.13015094E+00, 0.95722884E+00, 0.44391346E+00,
             -0.32366667E-01, 0.12413590E+00,-0.17610566E+00, 0.96929386E-01,
             -0.16703936E+00, },
            {-0.11894507E+01,-0.38839000E+00, 0.71710134E+00, 0.12543808E+00,
              0.23892194E-01, 0.17996967E+00,-0.33655792E-01,-0.64367644E-01,
             -0.11553254E+00, },
            { 0.22641020E+01,-0.19882113E-01,-0.99476331E+00,-0.88516802E-01,
             -0.73695379E+00, 0.47051879E-02,-0.22439910E+00,-0.23656236E+00,
              0.55085856E+00, },
            { 0.17834074E+01, 0.62063682E+00,-0.85358530E+00, 0.12144197E+00,
              0.67288089E+00,-0.39924723E+00, 0.35274625E+00,-0.10284065E+01,
             -0.55081531E-01, },
            {-0.42461534E-02,-0.19090576E-01,-0.21750687E-02, 0.77972598E-01,
             -0.10230899E+00,-0.61414940E-02,-0.60802395E-02,-0.60765678E-02,
             -0.31842440E-02, },
            { 0.36602344E-02,-0.20476623E-01, 0.19559988E-02, 0.12126583E+00,
              0.80235869E-01,-0.91823516E-02,-0.76314104E-02, 0.35730617E-02,
             -0.18788265E-02, },
            { 0.73799282E-01, 0.18658288E+00, 0.25844129E-01, 0.38106717E-01,
              0.13187513E-01, 0.33492066E-01,-0.36260985E-01,-0.13196414E-01,
             -0.99324873E-02, },
            { 0.14192873E-01,-0.25389520E-01,-0.18034412E-01, 0.66986732E-01,
              0.37657667E-01, 0.83836555E-01,-0.19943716E-01,-0.24610309E-01,
              0.22301378E-01, },
            {-0.17031059E+00, 0.24293567E+00,-0.47562119E-01,-0.18710293E+00,
             -0.21173187E+00, 0.45424961E-01, 0.10562468E+00, 0.11877582E-01,
              0.27717501E-01, },
            {-0.23334758E+00, 0.18515380E-01,-0.18031678E+00, 0.16339774E+00,
             -0.16562817E+00, 0.22401256E-01, 0.12215625E+00,-0.20599376E-01,
             -0.22505935E-01, },
            {-0.23864314E+00,-0.10410759E+01, 0.20875806E+00, 0.41495808E-02,
             -0.14400753E+00,-0.13934614E-01, 0.98016091E-01, 0.64151272E-01,
             -0.90540049E-03, },
            { 0.43185464E+00, 0.47100690E+00, 0.23188816E+00,-0.12820059E+00,
              0.30625871E+00,-0.37186635E+00, 0.27746350E-01, 0.18685450E+00,
             -0.14788827E+00, },
            { 0.31099787E+00,-0.51439357E+00, 0.33028484E-02,-0.17644372E-01,
              0.89867815E-01,-0.63249975E-01,-0.14556959E+00,-0.11969471E-01,
             -0.39798412E-01, },
            { 0.37246844E+00,-0.99704146E-01, 0.28348404E+00,-0.27310552E-01,
             -0.17952356E-02, 0.30666633E-01,-0.20847893E+00,-0.28230352E-02,
              0.19477863E-01, },
            { 0.20202787E+00, 0.14707644E+01,-0.31790829E+00,-0.87093808E-01,
              0.18147080E+00,-0.21470476E-01,-0.88236138E-01,-0.84560640E-01,
              0.35590544E-01, },
            {-0.68945140E+00,-0.61575735E+00,-0.83507515E-01, 0.61113432E-01,
             -0.57205093E+00, 0.39900196E+00, 0.70120350E-01,-0.22128372E+00,
              0.17749649E+00, },
            {-0.23316426E-01, 0.12048603E-01, 0.23783310E-02, 0.92227728E-03,
              0.47883224E-02, 0.22830907E-01,-0.44083364E-01,-0.45512458E-02,
              0.17387259E-02, },
            {-0.70657874E-02,-0.15885700E-01,-0.40105851E-02,-0.69217128E-02,
             -0.58886250E-02, 0.35231218E-01, 0.42952802E-01, 0.33667521E-02,
             -0.20392444E-02, },
            { 0.59208658E-01, 0.63783467E-01, 0.12437732E-01,-0.64228050E-01,
              0.13783633E-01,-0.51416833E-01, 0.18383345E-01, 0.69502792E-02,
              0.81083290E-02, },
            { 0.40749948E-01,-0.75332168E-02,-0.27416045E-01,-0.33444658E-01,
             -0.19954119E-01, 0.60441419E-02,-0.56821901E-01, 0.11292337E-02,
              0.17509496E-01, },
            {-0.67723908E-01,-0.14539140E+00,-0.30557400E-01, 0.45087757E-02,
             -0.46182219E-01,-0.15280160E+00,-0.10240623E+00, 0.26795991E-01,
             -0.21112165E-02, },
            {-0.47966205E-01, 0.10730190E-01,-0.31358361E-01, 0.15586140E-01,
             -0.61677694E-02, 0.13321459E+00,-0.20194066E+00, 0.10347781E-01,
             -0.12390453E-01, },
            {-0.18478956E-01,-0.15848843E-01, 0.85952645E-02, 0.16202679E-01,
              0.12485572E-02,-0.19915239E-02, 0.31446822E-02, 0.44241477E-01,
             -0.30390822E-03, },
            {-0.94869835E-02, 0.91496035E-02,-0.60107494E-02,-0.70505976E-02,
             -0.37957488E-02, 0.63174270E-03, 0.55268044E-02,-0.33853750E-03,
              0.48207454E-01, },
            { 0.65101489E-01, 0.29953860E-01,-0.47032386E-02,-0.26011158E-01,
             -0.31840345E-02, 0.30755354E-02, 0.13277406E-02, 0.22132441E-01,
              0.82672704E-02, },
            {-0.78570237E-02,-0.45788117E-01, 0.51256320E-02, 0.33361401E-01,
             -0.11823519E-01,-0.16863131E-02,-0.15676618E-01, 0.17375020E-01,
              0.15721165E-01, },
            { 0.64684800E-02,-0.96570859E-02, 0.89065656E-02,-0.50537391E-02,
             -0.49813055E-02, 0.20753632E-02, 0.76234308E-02,-0.90143485E-02,
              0.38806573E-02, },
            { 0.70466385E-02, 0.23270652E-01,-0.98308958E-02,-0.23814661E-02,
             -0.12868521E-02, 0.14961665E-01, 0.29478413E-02,-0.25250819E-02,
              0.35902048E-02, },
            { 0.68825735E-02, 0.69126990E-02, 0.12519725E-01, 0.12947415E-01,
             -0.24092742E-02, 0.11634666E-02,-0.38903584E-02,-0.92853449E-03,
              0.12939579E-02, },
            {-0.27606012E-02,-0.23899183E-02,-0.13654518E-01,-0.14725209E-02,
              0.18952837E-02, 0.95515465E-02, 0.27220524E-02,-0.41225250E-02,
              0.64941798E-02, },
         },
         {
            { 0.26891699E+01,-0.77640088E-02, 0.11127225E-01, 0.56581018E-02,
              0.55255590E-03, 0.19188958E-02, 0.38992222E-02,-0.63712653E-02,
             -0.10736038E-02, },
            { 0.89743264E-01, 0.20814057E-01, 0.49664252E-01, 0.12002765E-01,
              0.16872061E-01,-0.14828340E-01,-0.20098081E-01,-0.71201292E-02,
             -0.10063166E-01, },
            { 0.14451987E+01, 0.18207384E-01,-0.12469007E+00,-0.77557176E-01,
              0.96017895E-02, 0.50489246E-02, 0.77703637E-02, 0.43605078E-01,
              0.20670578E-01, },
            {-0.10420049E+00,-0.30086073E+00,-0.98388202E-01,-0.17734094E-01,
             -0.10223059E+00, 0.44458374E-01, 0.44485763E-01, 0.28927306E-01,
              0.21576228E-01, },
            {-0.26058183E+01, 0.11795595E+00, 0.28667596E+00, 0.16859519E+00,
             -0.69919094E-01,-0.51369867E-02,-0.43885093E-01,-0.75206392E-01,
             -0.15287774E-01, },
            { 0.39084069E-02, 0.23944090E+00, 0.64139619E-01, 0.46739038E-02,
              0.91723315E-01,-0.28238665E-01,-0.27856352E-01,-0.21242142E-01,
             -0.13344122E-01, },
            { 0.13095468E+01,-0.10214921E+00,-0.16770977E+00,-0.98850533E-01,
              0.73869824E-01,-0.47578220E-02, 0.37430078E-01, 0.40278375E-01,
             -0.92505598E-02, },
            { 0.23524527E-01,-0.34971893E+00,-0.28617576E+00, 0.24448626E-02,
              0.10531447E-02,-0.24429685E-02, 0.79786703E-02,-0.32044223E-02,
              0.65662437E-02, },
            {-0.49985661E-02, 0.30120927E+00,-0.33876929E+00,-0.98431129E-02,
             -0.94157760E-03,-0.78012803E-04,-0.10437771E-01, 0.98567735E-02,
              0.11818309E-01, },
            {-0.57822518E-01, 0.21672082E+00, 0.88802382E-01,-0.38946371E-02,
              0.58250647E-01,-0.24885843E-01,-0.17417107E-01, 0.15775116E-01,
             -0.64615905E-02, },
            { 0.32139458E-01, 0.32517146E-01, 0.21158411E+00,-0.15073596E-01,
             -0.36884345E-01, 0.46076268E-01, 0.17812065E-02, 0.35551559E-01,
              0.46800584E-01, },
            {-0.24651740E+00, 0.11098006E+01, 0.93898606E+00, 0.13167660E+00,
              0.78329649E-02, 0.21393843E-01, 0.26152806E-01, 0.99524148E-02,
             -0.16420493E+00, },
            {-0.35180086E+00,-0.94679403E+00, 0.12385637E+01, 0.81863344E-01,
              0.52147958E-01,-0.94125001E-02, 0.86879626E-01,-0.13224545E+00,
             -0.17574912E+00, },
            { 0.77992594E+00,-0.13512278E+01,-0.66809833E+00, 0.17822575E-01,
             -0.50182998E+00, 0.64230546E-01,-0.65235958E-01,-0.17756595E+00,
              0.55901699E-01, },
            { 0.60895658E+00,-0.39232608E-01,-0.12162886E+01,-0.19391639E-01,
              0.24424776E+00,-0.32840198E+00, 0.16300706E-01,-0.27074844E+00,
             -0.22464836E+00, },
            { 0.55456519E+00,-0.69234842E+00,-0.40266684E+00,-0.63226157E+00,
              0.10810311E-01,-0.15206470E-01,-0.15813012E+00, 0.76063015E-01,
              0.54846871E+00, },
            { 0.12965285E+01,-0.14871565E+00,-0.16574056E+01,-0.77499501E-01,
             -0.24706507E+00,-0.49161386E-01,-0.18116425E+00, 0.29508817E+00,
              0.42510369E+00, },
            {-0.19151396E+01, 0.23017201E+01, 0.19343640E+01, 0.18132132E+00,
              0.97290933E+00,-0.32005332E-01, 0.30257878E+00, 0.47717166E+00,
             -0.22451325E+00, },
            {-0.23164320E+01,-0.23464796E-02, 0.23696494E+01, 0.11396227E+00,
             -0.30829358E+00, 0.81678474E+00,-0.14368920E+00, 0.57714278E+00,
              0.30708337E+00, },
            {-0.28491089E+00, 0.36383953E-01, 0.31162423E+00, 0.57425320E+00,
             -0.42527661E-01, 0.73678866E-02, 0.16507234E+00,-0.12021207E+00,
             -0.42351913E+00, },
            {-0.11174608E+01, 0.66397589E+00, 0.92702854E+00,-0.17574133E-01,
              0.21021613E+00, 0.75819165E-01, 0.74323684E-01,-0.16377261E+00,
             -0.30484450E+00, },
            { 0.11824274E+01,-0.11248287E+01,-0.14524229E+01,-0.21500514E+00,
             -0.54644227E+00,-0.13700272E-01,-0.25452733E+00,-0.35112619E+00,
              0.19990946E+00, },
            { 0.20050185E+01, 0.13328118E-01,-0.14572992E+01,-0.51557049E-01,
              0.12367892E+00,-0.57814145E+00, 0.17719761E+00,-0.37120983E+00,
             -0.99483021E-01, },
            {-0.84307930E-03,-0.17258719E-01,-0.53407173E-02, 0.81290640E-01,
             -0.76485693E-01,-0.47752857E-02, 0.17251115E-02,-0.37566551E-02,
             -0.43733157E-02, },
            { 0.26733198E-02,-0.18344466E-01, 0.50127055E-02, 0.71141437E-01,
              0.87314069E-01,-0.61744666E-02,-0.70299716E-02, 0.26243473E-02,
             -0.23496000E-02, },
            { 0.28444359E-01, 0.11505325E+00,-0.33933113E-02, 0.28973157E-02,
              0.92887551E-01, 0.31396382E-01,-0.23507848E-01,-0.87130796E-02,
             -0.41673426E-02, },
            { 0.46873268E-01, 0.13442473E-01,-0.25793346E-01,-0.76118015E-01,
              0.18148199E-01, 0.74972913E-01, 0.14298553E-01,-0.31826302E-01,
              0.11931223E-01, },
            {-0.10709218E+00, 0.17803708E+00, 0.89736991E-02,-0.42456055E+00,
              0.21876013E+00,-0.10819510E-01, 0.28348537E-01,-0.17255733E-01,
              0.18319869E-01, },
            {-0.19474298E+00, 0.41860364E-01,-0.13339257E+00,-0.18719760E+00,
             -0.42420575E+00,-0.77197952E-02, 0.99332146E-01,-0.31550217E-01,
             -0.35813104E-01, },
            { 0.40004592E-01,-0.68339717E+00, 0.27681801E+00,-0.17132305E-01,
             -0.32585517E+00, 0.31075010E-02, 0.79180390E-01, 0.84248841E-01,
             -0.97170295E-02, },
            { 0.93430899E-01, 0.14520720E+00, 0.17805339E+00, 0.36324471E+00,
              0.13691723E+00,-0.32484740E+00,-0.81565991E-01, 0.24830067E+00,
             -0.83268017E-01, },
            { 0.17182921E+00,-0.39017051E+00,-0.75088419E-01, 0.10350134E+00,
             -0.59314120E+00, 0.26216136E-01,-0.36645230E-01, 0.51128760E-01,
             -0.23116680E-01, },
            { 0.29115209E+00,-0.16628721E+00, 0.17287152E+00, 0.57455140E+00,
              0.10316521E+00, 0.24566704E-01,-0.18611246E+00, 0.36977723E-01,
              0.51306799E-01, },
            {-0.20561792E+00, 0.10611655E+01,-0.37965357E+00, 0.59568016E-02,
              0.27356151E+00,-0.27140494E-01,-0.93819186E-01,-0.12059696E+00,
              0.53143971E-01, },
            {-0.24612656E+00,-0.89114860E-01,-0.68045080E-01,-0.36101612E+00,
             -0.25655359E+00, 0.38696358E+00, 0.15423262E+00,-0.32276914E+00,
              0.91858663E-01, },
            {-0.19476928E-01, 0.73561035E-02, 0.17873573E-02,-0.35678830E-02,
             -0.33237985E-02, 0.33766452E-01,-0.63296571E-01,-0.19452671E-02,
             -0.13239883E-02, },
            {-0.42364779E-02,-0.10886639E-01, 0.13287691E-02, 0.24953796E-03,
             -0.21379865E-02, 0.55756997E-01, 0.50980330E-01, 0.43777414E-02,
             -0.78862291E-02, },
            { 0.35293180E-01, 0.57927500E-01, 0.91379685E-02,-0.24990179E-01,
              0.15874520E-01,-0.32756235E-01,-0.91940789E-02, 0.63448343E-02,
              0.59108045E-02, },
            { 0.20577457E-01, 0.78460500E-02,-0.80252262E-02,-0.28437277E-01,
             -0.12757894E-01, 0.18312816E-01,-0.39697696E-01,-0.13474633E-02,
              0.13109337E-01, },
            {-0.83176987E-02,-0.11117948E+00,-0.57078397E-03, 0.10578160E-01,
             -0.20841746E-01,-0.17377385E+00,-0.81262551E-01, 0.16026536E-01,
              0.58632647E-02, },
            {-0.14602411E-01,-0.52710978E-03,-0.37079420E-01, 0.15373140E-01,
              0.46844836E-03, 0.11484052E+00,-0.20478855E+00,-0.60416027E-02,
              0.15950160E-01, },
            {-0.92366859E-02,-0.13327929E-01, 0.76846951E-02, 0.44647725E-02,
              0.99554425E-03,-0.25372533E-02, 0.11280873E-02, 0.34911953E-01,
              0.61604744E-02, },
            {-0.65696407E-02, 0.27554489E-02,-0.89025469E-02,-0.58459694E-03,
             -0.36023045E-02, 0.17695033E-02, 0.71463874E-02,-0.67619323E-02,
              0.35612658E-01, },
            { 0.37786972E-01, 0.27549380E-01,-0.10036921E-01,-0.17617412E-01,
             -0.51135998E-02,-0.85136052E-02, 0.82170218E-02, 0.34347039E-01,
              0.35131623E-02, },
            { 0.34137515E-02,-0.39485246E-01, 0.19134531E-01, 0.13404503E-01,
             -0.16808830E-01,-0.10705445E-01,-0.19780356E-01, 0.15510093E-01,
              0.21839621E-01, },
            { 0.12911428E-01,-0.73543200E-02, 0.50285622E-02, 0.59204112E-03,
             -0.34981014E-02, 0.37478306E-02, 0.20511295E-02,-0.60793972E-02,
              0.34043847E-02, },
            { 0.10783734E-01, 0.17769093E-01,-0.77495449E-02, 0.85491629E-03,
             -0.31315386E-02, 0.84124962E-02, 0.57312618E-02, 0.12188161E-02,
              0.74360068E-02, },
            { 0.63715721E-02, 0.54682349E-02, 0.96778516E-02, 0.52643046E-02,
             -0.51632943E-02, 0.32400473E-02,-0.58748294E-03,-0.52953459E-03,
              0.25913932E-02, },
            { 0.41622827E-02,-0.12601028E-02,-0.82769860E-02, 0.12908098E-02,
             -0.11074101E-02, 0.28636472E-02, 0.26873229E-02,-0.14917087E-02,
              0.50641391E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir14F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.64934897E+01,-0.60000785E-01, 0.15284570E+00, 0.42398740E-01,
              0.80449143E-02,-0.12943848E-02,-0.18471053E-01, 0.14364937E-01,
             -0.27232725E-01, 0.34205071E-02, 0.44790650E-03, 0.49032934E-03,
             -0.11499410E-01, },
            {-0.88309318E+00,-0.13880616E+00, 0.72883040E+00,-0.23786409E+00,
              0.56492403E-01,-0.44968966E+00, 0.27552197E-01,-0.10558825E+00,
             -0.26403388E+00,-0.71968939E-02,-0.76252706E-01,-0.75245716E-01,
              0.36521424E-01, },
            { 0.10725157E+02, 0.11520151E+01,-0.29717491E+01, 0.77741110E+00,
              0.10456994E+01,-0.56724101E+00, 0.10612051E+01,-0.19255696E-01,
              0.20865384E+00, 0.83688861E-02, 0.17841756E+00, 0.38568929E-01,
             -0.13791542E+00, },
            { 0.65042410E+01,-0.35215244E+01,-0.41706271E+01, 0.52065024E+01,
              0.25852561E+00, 0.73438549E+01,-0.40593761E+00, 0.99215275E+00,
              0.34198158E+01, 0.59768510E+00, 0.13837192E+01, 0.21450295E+01,
             -0.30246952E+00, },
            {-0.73952766E+02,-0.10371047E+02, 0.20006662E+02,-0.58430343E+01,
             -0.17244541E+02, 0.25785179E+01,-0.66231856E+01,-0.10432464E+01,
              0.73055929E+00, 0.10648139E+01,-0.16124163E+01, 0.77939086E-01,
              0.78449029E+00, },
            { 0.16648845E+00, 0.12172329E+02,-0.27850802E+01,-0.25546795E+02,
              0.91742992E+01,-0.32795715E+02, 0.28933215E+00,-0.56761956E+01,
             -0.18448353E+02,-0.33526461E+01,-0.67829127E+01,-0.12907632E+02,
             -0.90398967E+00, },
            { 0.15081433E+03, 0.33412430E+02,-0.50513580E+02, 0.15113485E+02,
              0.58774567E+02,-0.36906857E+01, 0.14214766E+02, 0.30996857E+01,
             -0.45489955E+01,-0.52512527E+01, 0.42320633E+01,-0.65626699E+00,
             -0.12561433E+01, },
            {-0.37230701E+02,-0.76325502E+01, 0.37164608E+02, 0.49207062E+02,
             -0.43934433E+02, 0.60087318E+02, 0.41347008E+01, 0.15909638E+02,
              0.43547508E+02, 0.70065861E+01, 0.12691342E+02, 0.29472052E+02,
              0.48083038E+01, },
            {-0.14633238E+03,-0.42288403E+02, 0.54256516E+02,-0.17434313E+02,
             -0.73183670E+02, 0.16437659E+01,-0.12896775E+02,-0.25577700E+01,
              0.64732404E+01, 0.78860893E+01,-0.44402709E+01, 0.10190881E+01,
              0.83452731E+00, },
            { 0.52198563E+02,-0.78648896E+01,-0.54521660E+02,-0.41549355E+02,
              0.62281200E+02,-0.48637409E+02,-0.83191090E+01,-0.19453861E+02,
             -0.45155918E+02,-0.66038699E+01,-0.10019678E+02,-0.28811026E+02,
             -0.57783585E+01, },
            { 0.55789318E+02, 0.18161959E+02,-0.20868525E+02, 0.74471416E+01,
              0.30560638E+02, 0.63638061E-01, 0.42821302E+01, 0.46981829E+00,
             -0.28367572E+01,-0.37410161E+01, 0.16539934E+01,-0.48298737E+00,
             -0.21111564E+00, },
            {-0.20312288E+02, 0.70529175E+01, 0.23697180E+02, 0.12980741E+02,
             -0.27796104E+02, 0.14427422E+02, 0.43019180E+01, 0.83882170E+01,
              0.16908133E+02, 0.23868093E+01, 0.27971561E+01, 0.10196642E+02,
              0.21167967E+01, },
            {-0.23634213E+00, 0.20315306E+01, 0.14751929E+01, 0.51506594E-01,
             -0.41991584E-01,-0.66530528E-02, 0.60239173E-01, 0.42047124E-01,
              0.14694092E-01, 0.14501678E-01, 0.22909958E-01, 0.50705005E-02,
             -0.13813758E-01, },
            { 0.24778880E-01,-0.13336085E+01, 0.20401320E+01,-0.29250428E-01,
              0.11255962E+00,-0.27715957E-01,-0.61806727E-01, 0.33551306E-02,
              0.49672600E-01,-0.88648414E-02, 0.15672091E-01,-0.17772952E-01,
              0.25840361E-01, },
            {-0.94002926E+00, 0.96187347E+00,-0.48695213E+00,-0.14855824E-01,
              0.85204057E-01, 0.68687528E-01, 0.59923762E+00,-0.17397812E+00,
              0.52780789E+00,-0.97602285E-01,-0.50955068E-01, 0.97110130E-01,
              0.78837104E-01, },
            {-0.15056743E+00, 0.25372839E+00, 0.77975589E+00, 0.44844508E-01,
              0.66648591E+00, 0.29808590E-01,-0.28804496E-01, 0.27313724E+00,
              0.57353723E+00,-0.14783476E+00,-0.54335695E-01, 0.18213356E+00,
             -0.17622781E+00, },
            { 0.44821749E+01, 0.12306819E+02, 0.71697435E+01,-0.92180455E+00,
              0.17149884E+01, 0.12936370E+01,-0.14673869E+00,-0.16734644E+01,
             -0.90978986E+00,-0.64830935E+00,-0.10648321E+01, 0.25684693E+00,
             -0.10137751E+00, },
            { 0.85512914E-01,-0.10835258E+02, 0.85481062E+01, 0.11747242E+01,
              0.18459188E+01, 0.27393374E+00, 0.21028144E+01,-0.48073640E+00,
             -0.68387592E+00, 0.73008960E+00,-0.79940535E-01, 0.11825157E+01,
             -0.94839269E+00, },
            { 0.13821822E+02,-0.18135992E+02, 0.25114826E+02, 0.74971118E+01,
              0.18632157E+01, 0.13507416E+01,-0.10556159E+02, 0.19317433E+01,
             -0.81330175E+01, 0.36078362E+01, 0.21540523E+00,-0.27909291E+01,
             -0.10886441E+01, },
            { 0.62252507E+01,-0.13018351E+02,-0.10101440E+02,-0.10007334E+01,
             -0.14628852E+02, 0.11357673E+01, 0.36046584E+01,-0.42910991E+01,
             -0.89776678E+01,-0.89598792E-02,-0.42933542E+00,-0.15849814E+01,
              0.34161615E+01, },
            {-0.17371538E+02,-0.66748375E+02, 0.62607899E+02, 0.20953897E+02,
             -0.16884054E+02,-0.15206277E+02,-0.11621350E+01, 0.13002543E+02,
              0.13039217E+02, 0.52742462E+01, 0.93021603E+01,-0.34476848E+01,
              0.56655067E+00, },
            {-0.56244588E+00,-0.33539303E+02,-0.27122942E+02,-0.17939253E+02,
             -0.18481981E+02,-0.38500461E+00,-0.14842595E+02,-0.31616006E+01,
              0.47638435E+01,-0.45677834E+01,-0.55476618E+00,-0.94443140E+01,
              0.57917881E+01, },
            {-0.74924995E+02, 0.14071364E+03,-0.14038925E+03,-0.58120224E+02,
              0.71764197E+01,-0.49531412E+01, 0.62206001E+02,-0.40747013E+01,
              0.32049103E+02,-0.27928366E+02, 0.17490292E+01, 0.17929644E+02,
              0.32690539E+01, },
            {-0.65806305E+02, 0.54215317E+02, 0.98710693E+02,-0.88018885E+01,
              0.75170929E+02,-0.11637873E+02,-0.30119808E+02, 0.31769405E+02,
              0.47102097E+02, 0.81819410E+01, 0.76862321E+01, 0.71875310E+01,
             -0.20791466E+02, },
            { 0.89678268E+01, 0.10135834E+03,-0.35098581E+03,-0.86679405E+02,
              0.51673248E+02, 0.53351051E+02, 0.76851803E+00,-0.36545685E+02,
             -0.47825413E+02,-0.15288162E+02,-0.24478294E+02, 0.11542589E+02,
              0.20585668E+01, },
            { 0.16705963E+02, 0.25283752E+03,-0.40345543E+02, 0.73295845E+02,
              0.65839951E+02, 0.55188236E+01, 0.37025734E+02, 0.24288336E+02,
             -0.10951288E+02, 0.91241064E+01, 0.23609199E+00, 0.27590469E+02,
             -0.15265544E+02, },
            { 0.19687543E+03,-0.38724133E+03, 0.30177777E+03, 0.15378911E+03,
             -0.65229530E+02,-0.34766839E+01,-0.15835892E+03, 0.39712989E+01,
             -0.47045303E+02, 0.82022125E+02,-0.10522800E+02,-0.45835922E+02,
             -0.15153832E+01, },
            { 0.19607632E+03,-0.66794991E+02,-0.28728564E+03, 0.45004837E+02,
             -0.16881575E+03, 0.29734268E+02, 0.87140694E+02,-0.93818230E+02,
             -0.11390152E+03,-0.33501694E+02,-0.25811827E+02,-0.17500467E+02,
              0.51393471E+02, },
            { 0.36721821E+02,-0.58378582E+02, 0.52954004E+03, 0.12650403E+03,
             -0.61309460E+02,-0.72215317E+02, 0.55067840E+01, 0.43052776E+02,
              0.65146362E+02, 0.18467991E+02, 0.23828384E+02,-0.14804021E+02,
             -0.66496506E+01, },
            {-0.41053894E+02,-0.39513184E+03, 0.14958513E+03,-0.11126383E+03,
             -0.10337821E+03,-0.17559303E+02,-0.34940395E+02,-0.43032936E+02,
              0.85262613E+01,-0.69466352E+01, 0.36357474E+01,-0.34809723E+02,
              0.18168346E+02, },
            {-0.24683015E+03, 0.44674811E+03,-0.29277634E+03,-0.16985976E+03,
              0.11071173E+03, 0.20931087E+02, 0.18121008E+03,-0.38683255E+01,
              0.23711794E+02,-0.10216876E+03, 0.16843342E+02, 0.51287098E+02,
             -0.42596765E+01, },
            {-0.21369885E+03, 0.15854466E+02, 0.32373230E+03,-0.56573444E+02,
              0.18293462E+03,-0.24894985E+02,-0.10696294E+03, 0.11558176E+03,
              0.12890714E+03, 0.48120480E+02, 0.32083454E+02, 0.20970270E+02,
             -0.55741127E+02, },
            {-0.36825500E+02, 0.12671242E+02,-0.24902853E+03,-0.60534359E+02,
              0.24415829E+02, 0.33357067E+02,-0.56761322E+01,-0.18138479E+02,
             -0.29999857E+02,-0.79702997E+01,-0.73311758E+01, 0.64502964E+01,
              0.43373952E+01, },
            { 0.26744591E+02, 0.18625591E+03,-0.93930122E+02, 0.56447613E+02,
              0.58127296E+02, 0.13249874E+02, 0.98932362E+01, 0.23001041E+02,
             -0.14396309E+01, 0.16029669E+01,-0.36048818E+01, 0.15927850E+02,
             -0.79116158E+01, },
            { 0.11669273E+03,-0.18738850E+03, 0.10586761E+03, 0.65915176E+02,
             -0.55447582E+02,-0.14503896E+02,-0.76371826E+02, 0.23939581E+01,
             -0.49679360E+00, 0.45396572E+02,-0.83436613E+01,-0.20962414E+02,
              0.35527298E+01, },
            { 0.76042992E+02, 0.11706135E+02,-0.12662367E+03, 0.20020100E+02,
             -0.78973969E+02, 0.45572462E+01, 0.47576389E+02,-0.49966724E+02,
             -0.54524891E+02,-0.23131273E+02,-0.13527062E+02,-0.95180111E+01,
              0.22111988E+02, },
            { 0.19034376E-01, 0.98584309E-01, 0.10540415E-01,-0.10316912E+01,
             -0.50497755E-01,-0.28466880E-02, 0.14425301E-01, 0.89508243E-03,
              0.40291805E-01,-0.58668968E-02,-0.37617797E-02, 0.81910536E-03,
              0.14230346E-01, },
            { 0.55068702E-01,-0.61659035E-02, 0.25711088E-01, 0.34259062E-01,
             -0.75383478E+00, 0.40433593E-02,-0.13901708E-01,-0.40712301E-01,
             -0.29276965E-01,-0.10735627E-01,-0.72969968E-03,-0.98825945E-02,
             -0.17654425E-02, },
            { 0.12820829E+00, 0.71660620E+00, 0.71653306E+00, 0.37673324E+00,
             -0.52130508E+00, 0.26109612E+00,-0.33626252E+00, 0.12658983E+00,
              0.23496895E+00, 0.45112655E-01, 0.10876163E-01,-0.79171509E-01,
             -0.49059778E-01, },
            { 0.84820724E+00, 0.11983800E+00, 0.93958420E+00, 0.63321990E+00,
              0.95848083E+00, 0.16888373E+00, 0.38933194E+00,-0.13175000E+00,
             -0.13978599E+00,-0.28500167E+00,-0.16895266E+00, 0.28273294E-01,
             -0.13272938E+00, },
            { 0.10719464E+01,-0.10473355E+01, 0.16151112E+01, 0.88533993E+01,
              0.29226627E+01,-0.44834813E+00,-0.18926809E+01,-0.10330509E+01,
              0.70661664E-01, 0.34604049E+00, 0.24440205E+00,-0.55603737E+00,
             -0.15387839E+00, },
            { 0.15618073E+01, 0.14701888E+00,-0.74968077E-01,-0.39555376E+01,
              0.69049392E+01,-0.19420260E+00, 0.57892245E+00,-0.79809892E+00,
              0.60062045E+00, 0.22479720E-01,-0.32704258E+00, 0.23038152E+00,
              0.36810508E+00, },
            {-0.29060516E+01,-0.35181143E+01,-0.57364321E+01, 0.55235324E+01,
              0.45822948E+00,-0.15032139E+01, 0.21021144E+01,-0.25328857E+00,
             -0.18473371E+01,-0.84338069E-01, 0.39706534E+00, 0.18311095E+00,
              0.12498163E+01, },
            {-0.99295187E+01,-0.10716367E+01,-0.49983807E+01,-0.20832248E+01,
              0.18830384E+01, 0.86717820E+00,-0.15240573E+01, 0.21804466E+01,
              0.58699912E+00, 0.14913265E+01, 0.91169506E+00, 0.18899290E+00,
              0.18160110E+01, },
            {-0.94457245E+01, 0.17658319E+02,-0.11660739E+02,-0.68379431E+01,
             -0.55005379E+01, 0.43817210E+01, 0.11818675E+02, 0.47689600E+01,
              0.28519669E+00,-0.19000353E+01,-0.25939715E+01, 0.37662926E+01,
              0.75847077E+00, },
            {-0.56515350E+01, 0.91205730E+01,-0.19686940E+01, 0.13128128E+02,
             -0.68311524E+00, 0.99185145E+00,-0.16249559E+01, 0.51543722E+01,
             -0.11013597E+01,-0.84779017E-01, 0.18501135E+01,-0.38142797E+00,
             -0.30122881E+01, },
            { 0.75019464E+01, 0.35646596E+01, 0.11501075E+02,-0.18637688E+02,
             -0.69856608E+00, 0.35613034E+01,-0.47264366E+01,-0.46824551E+00,
              0.42374134E+01,-0.10472504E+00,-0.11728551E+01,-0.23341645E+00,
             -0.39178739E+01, },
            { 0.26873817E+02,-0.89032382E+00, 0.74988832E+01, 0.33407247E+01,
             -0.96325502E+01,-0.34091041E+01, 0.21045973E+01,-0.56645589E+01,
              0.19026136E+00,-0.22414279E+01,-0.13311152E+01,-0.59583026E+00,
             -0.46499009E+01, },
            { 0.21163473E+02,-0.48009987E+02, 0.25364653E+02,-0.28718987E+02,
              0.28193629E+00,-0.14275020E+02,-0.20585054E+02,-0.64663568E+01,
             -0.17794694E+01, 0.29850135E+01, 0.60325065E+01,-0.71860552E+01,
             -0.16418031E+01, },
            { 0.39766910E+01,-0.27315857E+02, 0.82747126E+01,-0.15480450E+02,
             -0.37333546E+02,-0.14049913E+01,-0.78134075E-01,-0.88510962E+01,
             -0.22243371E+01, 0.97602409E+00,-0.25714512E+01,-0.92084908E+00,
              0.73255792E+01, },
            {-0.59779377E+01, 0.12392722E+01,-0.78840604E+01, 0.12213486E+02,
             -0.45400086E+00,-0.31542869E+01, 0.34382269E+01, 0.87975073E+00,
             -0.27262006E+01, 0.43419167E+00, 0.68464506E+00, 0.22095339E+00,
              0.30523138E+01, },
            {-0.20467285E+02, 0.39570870E+01,-0.23940721E+01, 0.22183473E+00,
              0.45692406E+01, 0.23968148E+01,-0.16324166E+01, 0.37603843E+01,
             -0.13567244E+01, 0.12711176E+01, 0.64020485E+00, 0.41453242E+00,
              0.32386906E+01, },
            {-0.14195784E+02, 0.34288692E+02,-0.15830330E+02, 0.32217907E+02,
              0.34044676E+01, 0.12452764E+02, 0.10785024E+02, 0.26360891E+01,
              0.13603797E+01,-0.15146811E+01,-0.39355423E+01, 0.41184163E+01,
              0.11697747E+01, },
            { 0.83907932E+00, 0.17864233E+02,-0.90199413E+01, 0.53105712E+01,
              0.36594128E+02, 0.70343792E+00, 0.17940173E+01, 0.47861300E+01,
              0.37236953E+01,-0.12472179E+01, 0.73165965E+00, 0.12939287E+01,
             -0.52513423E+01, },
            {-0.64029008E-01, 0.89368589E-01,-0.10483639E+00,-0.41617226E-01,
              0.42851824E-01,-0.37522101E+00,-0.74830681E+00,-0.35888155E-02,
              0.50473481E-01, 0.39554734E-01, 0.17244495E-01,-0.62780860E-02,
              0.40925271E-02, },
            { 0.47223818E-01,-0.15007500E+00,-0.77078842E-01, 0.23586081E-01,
             -0.40981129E-01, 0.73113239E+00,-0.37994549E+00,-0.29909762E-01,
              0.14559225E-01, 0.11881361E-01,-0.21748168E-01, 0.17178882E-01,
              0.25832307E-01, },
            {-0.12422315E+00, 0.75573784E+00, 0.11975990E+00,-0.26279071E+00,
             -0.63855439E+00,-0.10346416E+00,-0.24757487E+00, 0.83832882E-01,
             -0.37987664E-01, 0.47420461E-01, 0.51827621E-01,-0.82543463E-01,
             -0.48918199E-01, },
            {-0.25143120E-01,-0.32514688E+00,-0.23532422E+00,-0.13309413E-01,
             -0.85015304E-01, 0.38035685E+00, 0.60264971E-01,-0.13409159E-01,
             -0.31570796E-01,-0.10386517E+00,-0.10889217E-01, 0.21442087E-01,
             -0.58925855E-02, },
            { 0.51993853E+00,-0.57693455E-01, 0.28187773E+00,-0.73443003E-01,
             -0.19520199E+00,-0.11351212E-01, 0.58080941E+00,-0.54905391E+00,
              0.22085562E+00,-0.66554368E-01,-0.17385939E+00, 0.18870950E-03,
             -0.11992918E+00, },
            {-0.21958467E+00, 0.21047795E+01,-0.20394135E+00,-0.57054359E+00,
              0.54538924E-01,-0.12259457E+01, 0.53822305E-01,-0.47958690E+00,
             -0.65027052E+00,-0.24385016E-01, 0.14002994E+00,-0.22192319E+00,
             -0.13525110E+00, },
            {-0.54393631E+00,-0.13514569E+01, 0.26100159E+00, 0.43233761E+00,
              0.97404122E+00, 0.40088597E+00, 0.80763835E+00,-0.98028913E-01,
              0.11027520E+00,-0.10748786E+00,-0.36647193E-01, 0.19190548E+00,
              0.88344067E-01, },
            {-0.25658110E+00, 0.72416735E+00, 0.57697695E+00,-0.34691748E+00,
              0.32392946E+00,-0.13627889E+01, 0.11442543E+00,-0.10774225E+00,
              0.10521625E+00, 0.98740987E-01, 0.29976496E-01, 0.26290998E-01,
             -0.56078196E-01, },
            { 0.27019140E+00, 0.45024633E+00,-0.17958543E+00, 0.12233192E+00,
              0.51881069E+00, 0.42524201E+00, 0.30084953E+00, 0.94178420E+00,
             -0.65111023E+00,-0.27122760E-01, 0.76508380E-01,-0.85360545E-04,
              0.12641452E+00, },
            { 0.79744786E-01,-0.28821568E+01, 0.73260629E+00, 0.53177392E+00,
              0.33241987E+00, 0.14034445E+01, 0.37998933E+00, 0.94033825E+00,
              0.12633942E+01,-0.60186643E-01,-0.11780193E+00, 0.26741916E+00,
              0.93822643E-01, },
            {-0.40619418E-01, 0.80723641E-02,-0.33849481E-01, 0.31438507E-01,
              0.26720112E-01, 0.34434196E-01, 0.15387811E-01, 0.30153814E+00,
             -0.68199985E-01, 0.15819095E-01, 0.84541179E-02,-0.12130376E-01,
             -0.27178498E-02, },
            { 0.10475045E+00, 0.49542687E-02, 0.35022680E-01, 0.87533630E-01,
             -0.61479896E-01,-0.14135311E-01,-0.12770660E-01, 0.36308039E-01,
              0.26710710E+00, 0.10513934E-01,-0.55483915E-02, 0.76266043E-02,
             -0.12753191E-01, },
            {-0.21390147E-02, 0.58438715E-01,-0.11032433E+00, 0.12841828E+00,
             -0.22302932E+00,-0.96188486E-01, 0.17974999E-01, 0.29908022E+00,
              0.44135422E+00,-0.46303540E-01, 0.41154433E-01,-0.14057259E-01,
              0.87432843E-02, },
            { 0.30396742E+00, 0.75595453E-01,-0.15522315E+00,-0.17690331E-02,
              0.20859620E+00,-0.87131839E-02,-0.10404948E+00,-0.36777064E+00,
              0.11611282E+00,-0.29208448E-02,-0.80868147E-01, 0.16911151E-01,
              0.13700172E-01, },
            {-0.13175465E+00,-0.61796591E-01,-0.48755106E-01,-0.12370612E-01,
             -0.54931860E-01,-0.18966548E-01, 0.25214590E-01, 0.38970198E-01,
             -0.42744722E-01, 0.83304532E-01, 0.22285062E+00,-0.54338336E-03,
              0.11550711E-02, },
            { 0.12497649E-01, 0.33643261E-01,-0.49855210E-01, 0.77534869E-01,
             -0.20823942E-01, 0.56232158E-02,-0.27867474E-01, 0.31055154E-01,
              0.45609649E-01,-0.22880867E+00, 0.88289447E-01, 0.14045586E-01,
             -0.77454462E-02, },
            {-0.15834920E-01,-0.47239352E-01, 0.55206638E-01,-0.21747222E-01,
              0.19386897E-01, 0.35340972E-01, 0.19234702E-01, 0.55831406E-01,
              0.12493882E-01,-0.18281758E-01,-0.14992007E-01,-0.12327413E+00,
              0.10836018E-01, },
            { 0.29655593E-01, 0.94482973E-02,-0.85593509E-02, 0.62130149E-02,
              0.62354233E-01, 0.73285494E-02,-0.53194746E-01,-0.14881391E-01,
              0.54990896E-02,-0.13634078E-01,-0.12771355E-01, 0.14701579E-02,
             -0.12577879E+00, },
            { 0.20759542E-01, 0.37858854E-02,-0.38624063E-01, 0.25079571E-01,
              0.28830525E-01, 0.14798064E-01, 0.11077589E-01, 0.17425489E-01,
              0.35621811E-01,-0.25656393E-01,-0.24061346E-01, 0.10520253E-01,
              0.45533017E-02, },
            { 0.66021487E-01,-0.21064222E-01,-0.72724107E-02, 0.14031813E-01,
              0.59114888E-01, 0.15700987E-01, 0.68930597E-02,-0.33762053E-01,
             -0.16519623E-01,-0.90381289E-02,-0.59326985E-02, 0.11942729E-01,
              0.13869079E-01, },
            { 0.63189775E-01, 0.68125673E-01,-0.60883962E-01,-0.21250362E-01,
             -0.66251666E-02,-0.73823389E-02, 0.39793655E-01,-0.44493157E-01,
              0.10462924E-01,-0.12947634E-01,-0.25500301E-02,-0.15238015E-01,
              0.15929751E-01, },
            {-0.17856486E-01,-0.52744098E-01,-0.53626947E-01, 0.39534133E-01,
              0.30350417E-01, 0.46890290E-02,-0.20292008E-02, 0.65144026E-02,
              0.41108787E-01,-0.98453555E-02,-0.10881119E-01, 0.51510613E-03,
              0.14489324E-01, },
         },
         {
            { 0.97736454E+01, 0.14032146E+00, 0.20471638E+00, 0.45970183E-01,
              0.19347202E-01, 0.34263041E-02,-0.56153454E-01, 0.16434982E-01,
             -0.13611992E-01, 0.26763946E-01, 0.57403501E-02, 0.63922638E-02,
             -0.11557203E-01, },
            { 0.21873847E+00,-0.40463670E-02, 0.25696555E+00,-0.28938836E+00,
              0.35062745E+00, 0.15157066E+00,-0.29073560E+00, 0.10218962E+00,
             -0.17300349E+00,-0.40608421E-01,-0.18231006E-01,-0.10038075E-01,
              0.41188154E-01, },
            { 0.30748093E+02,-0.22748404E+01,-0.31234460E+01, 0.27615957E+01,
              0.15846493E+01,-0.34613076E+00, 0.56108177E+00, 0.11689792E+00,
              0.14760590E+00,-0.10534291E+01,-0.31982246E+00, 0.22981022E+00,
             -0.44722270E-01, },
            {-0.14903673E+02, 0.10495951E+00, 0.34792299E+01, 0.47500277E+01,
             -0.45592580E+01,-0.62566572E+00, 0.40229535E+01,-0.11005296E+01,
              0.26008980E+01,-0.20069683E+00, 0.20519117E+00, 0.83855027E+00,
             -0.12517659E+00, },
            {-0.15924529E+03, 0.13519261E+02, 0.10932475E+02,-0.29060259E+02,
             -0.17306383E+02, 0.23026779E+01,-0.11505018E+01, 0.89456850E+00,
              0.19126704E+00, 0.56772127E+01, 0.24429238E+01,-0.23409443E+01,
              0.10148532E+01, },
            { 0.11584834E+03,-0.17900440E+02,-0.13041323E+02,-0.16672869E+02,
              0.35105484E+02,-0.30701976E+01,-0.21194830E+02, 0.64325070E+01,
             -0.15774626E+02, 0.32018068E+01, 0.94925404E+00,-0.46957250E+01,
             -0.13125659E+01, },
            { 0.29437540E+03,-0.25372292E+02,-0.15352101E+02, 0.91963020E+02,
              0.53816139E+02,-0.39363284E+01,-0.11707987E+01,-0.62234478E+01,
             -0.21492360E+01,-0.11163957E+02,-0.70858817E+01, 0.70973535E+01,
             -0.29095371E+01, },
            {-0.29224625E+03, 0.61603146E+02,-0.42039781E+01, 0.19493481E+02,
             -0.98674675E+02, 0.13333192E+02, 0.47858620E+02,-0.16327671E+02,
              0.39196560E+02,-0.93861637E+01,-0.60532708E+01, 0.96561794E+01,
              0.45459685E+01, },
            {-0.26182547E+03, 0.17990252E+02, 0.11552253E+02,-0.11288570E+03,
             -0.65355690E+02, 0.19973891E+01, 0.40330243E+01, 0.10058838E+02,
              0.31824422E+01, 0.93387594E+01, 0.84235001E+01,-0.84775333E+01,
              0.29146054E+01, },
            { 0.29994781E+03,-0.74359123E+02, 0.36368931E+02,-0.57280774E+01,
              0.11249139E+03,-0.15975358E+02,-0.48075100E+02, 0.17203218E+02,
             -0.41371578E+02, 0.10335050E+02, 0.90007267E+01,-0.86931496E+01,
             -0.47164741E+01, },
            { 0.91385628E+02,-0.41139150E+01,-0.41653366E+01, 0.47360004E+02,
              0.27213379E+02,-0.10529798E-02,-0.22038872E+01,-0.48922997E+01,
             -0.13381232E+01,-0.28368783E+01,-0.34559801E+01, 0.34892297E+01,
             -0.93806028E+00, },
            {-0.10869084E+03, 0.30786814E+02,-0.22838774E+02,-0.15038877E+01,
             -0.44714828E+02, 0.61636329E+01, 0.17758163E+02,-0.62771416E+01,
              0.15508571E+02,-0.38858967E+01,-0.40972447E+01, 0.29156959E+01,
              0.15558701E+01, },
            {-0.13369410E+00, 0.17268354E+01, 0.14351931E+01, 0.76106519E-01,
             -0.28437419E-01,-0.11006788E-01, 0.26756642E-01, 0.27960541E-01,
              0.89221104E-03, 0.14117885E-01, 0.26185049E-01,-0.14624753E-02,
             -0.14191492E-01, },
            { 0.24574746E+00,-0.14273102E+01, 0.15711827E+01,-0.83912954E-01,
              0.46123277E-01,-0.25364889E-01,-0.29690681E-01,-0.29524004E-01,
              0.19983063E-01,-0.17720954E-01, 0.16691638E-01,-0.70638601E-02,
             -0.72944113E-02, },
            { 0.39862120E+00,-0.13170791E+00,-0.64909148E+00,-0.32600096E+00,
              0.21109910E+00,-0.12451132E+00, 0.43011883E+00,-0.18416311E+00,
              0.24612106E+00,-0.10645313E+00, 0.16143291E+00,-0.21634487E-01,
              0.63935556E-01, },
            { 0.92620957E+00,-0.49406093E+00, 0.64722395E+00,-0.72104082E-01,
              0.28068078E+00, 0.60120350E+00, 0.27893120E+00, 0.30756903E+00,
              0.17852144E-01,-0.17372723E+00,-0.12843073E-01, 0.60005605E-01,
             -0.44218376E+00, },
            { 0.10201948E+01, 0.13495907E+02,-0.10630937E+02,-0.10683585E+01,
              0.58663207E+00,-0.10076985E+01,-0.52236602E-01,-0.47220096E+00,
              0.46965632E+00,-0.10257162E+01,-0.77649891E+00, 0.21886133E+00,
             -0.13192793E-01, },
            {-0.63123832E+01, 0.98021412E+01, 0.17635956E+02, 0.41804461E+01,
              0.17669466E+01, 0.96404308E+00,-0.45744419E-01, 0.97980118E+00,
             -0.73717684E+00,-0.27884778E+00,-0.14686050E+00, 0.75753582E+00,
              0.43112877E+00, },
            {-0.29365113E+01,-0.14322780E+01, 0.14004352E+02, 0.79823327E+00,
             -0.58754277E+01, 0.25628228E+01,-0.86770887E+01, 0.29869020E+01,
             -0.79320574E+00, 0.71983641E+00,-0.20778570E+01, 0.96054983E+00,
             -0.10125142E+01, },
            {-0.14787666E+02, 0.89715195E+01,-0.16094648E+02, 0.62427344E+01,
             -0.78737440E+01,-0.20880353E+01,-0.53182220E+00,-0.72572389E+01,
              0.20186639E+01,-0.28493693E+00,-0.16519644E+01,-0.56926078E+00,
              0.75286026E+01, },
            { 0.38437169E+01, 0.23742902E+01, 0.10843651E+03, 0.32352233E+01,
              0.29753842E+01, 0.82216110E+01,-0.24196384E+01, 0.54554963E+01,
             -0.16184169E+00, 0.67759509E+01, 0.53559599E+01,-0.44717285E+00,
             -0.89659041E+00, },
            { 0.51552006E+02,-0.11602977E+03,-0.19592438E+02,-0.36331181E+02,
             -0.12795036E+02,-0.63453674E+01, 0.43711376E+01,-0.10106875E+02,
              0.51247444E+01, 0.35840931E+01, 0.18366957E+01,-0.68241401E+01,
             -0.46224647E+01, },
            {-0.85184917E+01, 0.16360107E+02,-0.82426064E+02, 0.12709790E+02,
              0.42946472E+02,-0.15440465E+02, 0.50130878E+02,-0.12952686E+02,
             -0.69280114E+01,-0.39795120E+01, 0.73542218E+01,-0.55133834E+01,
              0.81558447E+01, },
            { 0.96721992E+02,-0.39015026E+02, 0.11213722E+03,-0.45417725E+02,
              0.37435158E+02,-0.13082533E+01,-0.10213682E+02, 0.49293102E+02,
             -0.15153333E+02, 0.73512311E+01, 0.14430303E+02, 0.19037811E+01,
             -0.38897942E+02, },
            {-0.43353905E+02,-0.20246556E+03,-0.28853781E+03, 0.21355968E+01,
             -0.19108803E+02,-0.21631128E+02, 0.12639037E+02,-0.18742435E+02,
             -0.96571703E+01,-0.16555702E+02,-0.14417870E+02, 0.12324684E+00,
              0.51283345E+01, },
            {-0.14933170E+03, 0.32959485E+03,-0.18182323E+03, 0.11664076E+03,
              0.33371815E+02, 0.17004457E+02,-0.19989794E+02, 0.32714710E+02,
             -0.11863819E+02,-0.12437937E+02,-0.82264671E+01, 0.19701031E+02,
              0.14783766E+02, },
            { 0.80268501E+02,-0.52966335E+02, 0.19053438E+03,-0.56769230E+02,
             -0.11784628E+03, 0.38850357E+02,-0.12247976E+03, 0.22450397E+02,
              0.35795444E+02, 0.12220045E+02,-0.89107647E+01, 0.82319517E+01,
             -0.27355555E+02, },
            {-0.27648264E+03, 0.89760971E+02,-0.27464084E+03, 0.11196627E+03,
             -0.69628716E+02, 0.97918758E+01, 0.44225208E+02,-0.12664152E+03,
              0.35483727E+02,-0.18650520E+02,-0.41325413E+02,-0.23082571E+01,
              0.85997475E+02, },
            { 0.88594856E+02, 0.39292142E+03, 0.33457990E+03,-0.12988249E+02,
              0.27334524E+02, 0.24206512E+02,-0.20228155E+02, 0.24981485E+02,
              0.22630320E+02, 0.17233301E+02, 0.16416178E+02,-0.32048091E+00,
             -0.83845139E+01, },
            { 0.18617633E+03,-0.39346274E+03, 0.42378693E+03,-0.15928719E+03,
             -0.40951443E+02,-0.20998573E+02, 0.32289707E+02,-0.42694340E+02,
              0.11828459E+02, 0.17220362E+02, 0.13173159E+02,-0.22972109E+02,
             -0.18645224E+02, },
            {-0.14528056E+03, 0.72072128E+02,-0.21711536E+03, 0.77514603E+02,
              0.13667310E+03,-0.45385811E+02, 0.13474231E+03,-0.16270960E+02,
             -0.53262211E+02,-0.17100195E+02, 0.23942339E+01,-0.10488991E+01,
              0.37738712E+02, },
            { 0.35145352E+03,-0.85622932E+02, 0.26572525E+03,-0.10509781E+03,
              0.58970482E+02,-0.46502337E+01,-0.66478355E+02, 0.13732690E+03,
             -0.33438587E+02, 0.15821194E+02, 0.48546875E+02, 0.14652005E+01,
             -0.85408585E+02, },
            {-0.52937946E+02,-0.20882060E+03,-0.14351868E+03, 0.98669796E+01,
             -0.11305204E+02,-0.10061109E+02, 0.10184311E+02,-0.11492324E+02,
             -0.13785835E+02,-0.64393106E+01,-0.66763282E+01, 0.56769723E+00,
              0.43437328E+01, },
            {-0.82705460E+02, 0.16970370E+03,-0.24499571E+03, 0.77111313E+02,
              0.21147839E+02, 0.97495165E+01,-0.17787193E+02, 0.19586727E+02,
             -0.44074783E+01,-0.82658730E+01,-0.68889670E+01, 0.94197063E+01,
              0.82319374E+01, },
            { 0.79956360E+02,-0.38184700E+02, 0.95916229E+02,-0.35911259E+02,
             -0.56887783E+02, 0.20183622E+02,-0.54853699E+02, 0.38453090E+01,
              0.25204611E+02, 0.86927290E+01, 0.11558259E+01,-0.29337678E+01,
             -0.18029606E+02, },
            {-0.15973770E+03, 0.28304308E+02,-0.87137711E+02, 0.30577978E+02,
             -0.21899446E+02,-0.35690374E+01, 0.34283478E+02,-0.53517769E+02,
              0.10929289E+02,-0.38510764E+01,-0.20168242E+02,-0.64146167E+00,
              0.31352381E+02, },
            { 0.29265393E-01,-0.24591343E+00, 0.14553222E+00,-0.84675032E+00,
              0.61966658E+00,-0.31919140E-01, 0.56226421E-01, 0.23503466E-01,
              0.53592073E-02,-0.20415217E-01, 0.15285027E-01,-0.63779615E-02,
              0.55775273E-03, },
            { 0.90108313E-01,-0.20454387E+00,-0.73520780E-01,-0.64477378E+00,
             -0.82056785E+00,-0.32559566E-01,-0.23924757E-01,-0.15155605E-01,
             -0.23472775E-01,-0.29544417E-01, 0.71342289E-02,-0.47043166E-02,
             -0.47027934E-02, },
            { 0.52882087E+00, 0.75878024E+00, 0.37392050E+00, 0.52214783E+00,
             -0.73013431E+00, 0.44710663E+00, 0.54939032E-01, 0.22959816E+00,
              0.20311107E+00,-0.81182420E-01,-0.88519581E-01,-0.12205946E+00,
             -0.90271063E-01, },
            { 0.76053393E+00, 0.47488001E+00, 0.52843851E+00, 0.69257706E+00,
              0.27872598E+00, 0.68038344E-01, 0.28850755E+00,-0.78320973E-01,
              0.16159263E+00,-0.23097005E+00,-0.32177996E-01, 0.43377087E-01,
             -0.46754386E-01, },
            { 0.45139912E+00, 0.33630769E+01,-0.12169751E+01,-0.37143755E+01,
              0.72003312E+01,-0.30781597E+00,-0.21212363E+01,-0.96283758E+00,
              0.84034169E+00, 0.23972341E+00,-0.28099918E+00, 0.56343917E-01,
             -0.53734177E+00, },
            {-0.14882063E+01, 0.13920959E+01, 0.12130704E+01,-0.79513636E+01,
             -0.25826595E+01,-0.42678672E+00, 0.51285851E+00, 0.19756992E+00,
              0.12423240E+01, 0.54061139E+00,-0.37421107E-01, 0.29954505E+00,
              0.11390962E+00, },
            {-0.56358533E+01,-0.57385144E+01,-0.52712636E+01,-0.10486714E+01,
              0.53347268E+01,-0.35143905E+01,-0.12000906E+01,-0.28657036E+01,
             -0.12631081E+01, 0.94290608E+00, 0.92692262E+00, 0.13676528E+01,
              0.12747043E+01, },
            {-0.10084604E+02,-0.83411729E+00,-0.68800192E+01,-0.56028543E+01,
             -0.31238866E+00, 0.13777543E+01,-0.14698855E+01, 0.52827829E+00,
             -0.40625021E+00, 0.18055102E+01,-0.78214049E+00,-0.69355123E-01,
              0.65176880E+00, },
            { 0.11241800E+01,-0.65113230E+01,-0.21158099E+01, 0.29776365E+02,
             -0.24963110E+02, 0.54256268E+01, 0.11766754E+02, 0.49691448E+01,
             -0.47331429E+01, 0.25715870E+00, 0.14093016E+01,-0.11038080E+01,
              0.35657845E+01, },
            { 0.14730919E+02,-0.30367286E+01,-0.11354827E+02, 0.30109867E+02,
              0.22487963E+02, 0.39788237E+01,-0.24708533E+01,-0.19585212E+01,
             -0.73399401E+01,-0.29647706E+01, 0.20387471E+00,-0.30212843E+01,
             -0.14654021E+01, },
            { 0.14754129E+02, 0.11049473E+02, 0.13400968E+02, 0.58017559E+01,
             -0.10485391E+02, 0.91833000E+01, 0.26157830E+01, 0.80078135E+01,
              0.25257504E+01,-0.23380883E+01,-0.18219419E+01,-0.34472325E+01,
             -0.37856765E+01, },
            { 0.27359425E+02,-0.55443020E+01, 0.20240791E+02, 0.79822865E+01,
              0.63297043E+01,-0.35024116E+01, 0.38299043E+01, 0.60368818E+00,
              0.41653635E-02,-0.47033868E+01, 0.24803977E+01,-0.54888380E+00,
             -0.18400835E+01, },
            {-0.71220417E+01,-0.11084239E+01, 0.93044796E+01,-0.59619061E+02,
              0.21860548E+02,-0.16687052E+02,-0.19368769E+02,-0.89723320E+01,
              0.81729593E+01,-0.26830549E+01,-0.22497013E+01, 0.32522850E+01,
             -0.77899055E+01, },
            {-0.42522598E+02, 0.84326611E+01, 0.25657207E+02,-0.35328533E+02,
             -0.45498726E+02,-0.12340075E+02, 0.45494366E+01, 0.38336654E+01,
              0.14014010E+02, 0.50293469E+01,-0.80758733E+00, 0.73919759E+01,
              0.33322282E+01, },
            {-0.14168424E+02,-0.38411441E+01,-0.11940656E+02,-0.86237640E+01,
              0.33731959E+01,-0.65763731E+01,-0.14734651E+01,-0.60773010E+01,
             -0.16627012E+01, 0.18981390E+01, 0.91158038E+00, 0.23963671E+01,
              0.29741066E+01, },
            {-0.22353624E+02, 0.11713838E+02,-0.14435802E+02, 0.11054903E+01,
             -0.10635337E+02, 0.18735406E+01,-0.32202849E+01,-0.17535979E+01,
              0.22505578E+00, 0.38518550E+01,-0.17070565E+01, 0.92721128E+00,
              0.15198689E+01, },
            { 0.53186111E+01, 0.61862016E+01,-0.31186655E+01, 0.37600212E+02,
             -0.35645136E+00, 0.12633383E+02, 0.96007862E+01, 0.52255836E+01,
             -0.41970530E+01, 0.22828405E+01, 0.11919301E+01,-0.24510043E+01,
              0.52333202E+01, },
            { 0.34840076E+02,-0.12539686E+02,-0.16330975E+02, 0.10885726E+02,
              0.29065947E+02, 0.10912358E+02,-0.25189373E+01,-0.15795352E+01,
             -0.82213926E+01,-0.28210602E+01, 0.71790761E+00,-0.55111480E+01,
             -0.20255141E+01, },
            {-0.19228468E-01,-0.13114972E+00, 0.15714884E-01,-0.59424613E-01,
              0.61318103E-01,-0.58697629E+00,-0.10644103E+01, 0.20145899E-01,
             -0.13622616E-01,-0.18642439E-01, 0.22151224E-01, 0.47150520E-02,
              0.11869197E-02, },
            { 0.58520351E-01,-0.16852538E+00,-0.50097305E-01, 0.11990506E-01,
             -0.65762773E-02, 0.10272388E+01,-0.58508122E+00, 0.10597577E-01,
              0.35774235E-01,-0.15361957E-02,-0.45238879E-01,-0.18923517E-01,
             -0.59281811E-02, },
            { 0.25976369E+00, 0.61606842E+00, 0.13208783E-01, 0.22894323E+00,
             -0.39829186E+00, 0.10561230E+00,-0.10849887E+00, 0.61109383E-01,
              0.13916729E+00, 0.60511831E-01, 0.12184703E-01, 0.34330254E-02,
              0.49652576E-01, },
            { 0.33294815E+00, 0.93975160E-02,-0.15443997E-01, 0.12950849E-01,
             -0.53923000E-01, 0.16462359E+00, 0.16396619E+00,-0.73470592E-01,
              0.80365367E-01,-0.35155836E-01,-0.88913888E-01,-0.47349852E-01,
              0.39145373E-01, },
            { 0.11885962E+01, 0.70961654E+00, 0.16401079E+00, 0.73770237E+00,
             -0.50291091E+00, 0.21559596E+01, 0.16197531E+01,-0.89583582E+00,
              0.20669129E+00, 0.29641485E+00, 0.17593935E+00, 0.21165982E-01,
             -0.99231154E-02, },
            {-0.12865351E+01, 0.23329020E+01,-0.46353462E+00,-0.75524963E-01,
              0.23183054E+00,-0.19695925E+01, 0.19060562E+01,-0.13436994E+00,
             -0.10136840E+01, 0.37376542E-01, 0.39470100E+00, 0.21998765E+00,
             -0.11666382E-01, },
            {-0.71975428E+00,-0.56198847E+00, 0.62459812E-01,-0.95555380E-01,
              0.97378629E+00, 0.58153582E+00, 0.55926204E+00,-0.48008882E-01,
             -0.39891291E+00,-0.46037629E-01, 0.33642620E-01, 0.59914336E-01,
             -0.43716524E-01, },
            {-0.10643944E+01,-0.18275885E+00,-0.33081704E+00,-0.55505341E+00,
              0.37782079E+00,-0.80053759E+00, 0.45424259E+00,-0.85584493E-02,
             -0.85956886E-01, 0.12678237E+00, 0.19393581E+00, 0.18442687E+00,
             -0.24153914E-01, },
            {-0.15753589E+01,-0.46785453E+00,-0.12589587E+01,-0.16451536E+01,
              0.12443190E+01,-0.42283177E+01,-0.14672083E+01, 0.14503810E+01,
             -0.24458455E+00,-0.60518199E+00,-0.42549494E+00,-0.76019049E-01,
             -0.11163369E+00, },
            { 0.21944077E+01,-0.25643041E+01, 0.99713415E+00,-0.11333123E+00,
             -0.71992117E+00, 0.24609313E+01,-0.36890426E+01, 0.18535003E+00,
              0.17068619E+01,-0.43102510E-01,-0.49058500E+00,-0.35204008E+00,
             -0.33120699E-01, },
            {-0.58546931E-01,-0.61199530E-02, 0.13920870E-01,-0.22593593E-01,
              0.60600404E-01,-0.26639467E-01,-0.16022459E-01, 0.32998070E+00,
             -0.22570471E+00, 0.51776585E-02, 0.17488200E-01,-0.34345095E-02,
             -0.48399698E-02, },
            { 0.23667336E-01, 0.10178508E+00,-0.29486222E-01, 0.57037488E-01,
             -0.34963217E-01, 0.11159874E-01,-0.30660912E-01, 0.19712472E+00,
              0.30310646E+00,-0.16301766E-01, 0.55935443E-03,-0.90922392E-03,
              0.42374507E-02, },
            {-0.16880974E+00, 0.95758550E-01,-0.14629462E+00, 0.12069291E+00,
              0.56210205E-01,-0.12945679E+00, 0.17435406E+00, 0.96609093E-01,
              0.45201021E+00,-0.36387533E-01,-0.25372004E-01, 0.18637785E-02,
             -0.35717620E-02, },
            { 0.18735176E+00,-0.15473643E-01, 0.31935621E-01, 0.47426168E-02,
              0.11311485E+00,-0.13744952E+00,-0.12981303E+00,-0.40770856E+00,
              0.16419126E-01, 0.27462881E-03,-0.82576886E-01,-0.18355066E-01,
             -0.51394594E-02, },
            {-0.22159746E-01,-0.74835721E-03,-0.36677279E-01, 0.25976226E-01,
              0.33930016E-02, 0.16898612E-01, 0.23251474E-02, 0.38304754E-01,
             -0.42265587E-01, 0.20900027E+00, 0.22713178E+00,-0.21249074E-02,
             -0.10710889E-01, },
            { 0.22197489E-01, 0.83272249E-01,-0.76565236E-01, 0.21317841E-02,
             -0.13129898E-01, 0.16477922E-01, 0.48060704E-01, 0.35126519E-02,
              0.52965067E-01,-0.20686191E+00, 0.17197247E+00,-0.42362534E-02,
             -0.16804585E-01, },
            { 0.21828212E-01,-0.31586766E-01, 0.81997104E-01, 0.30568482E-01,
              0.17527269E-01, 0.51112954E-01,-0.15168884E-02, 0.15925393E-02,
             -0.10897530E-01,-0.70789969E-02, 0.73832730E-02,-0.12407729E+00,
              0.85962117E-01, },
            { 0.66300765E-01, 0.51962450E-01,-0.13351660E-01,-0.23629975E-01,
              0.20142781E-01, 0.38463738E-01,-0.13412732E-01,-0.93066981E-02,
             -0.41893595E-02, 0.41461214E-02,-0.17357355E-01,-0.11255018E+00,
             -0.12668885E+00, },
            {-0.59766680E-01,-0.32285359E-01,-0.46988070E-01, 0.39100412E-01,
              0.98173115E-02, 0.30527687E-01,-0.36675990E-01,-0.82789976E-02,
              0.30238917E-01,-0.19059830E-02,-0.11280150E-01,-0.14733355E-01,
             -0.83649457E-02, },
            { 0.45781240E-01, 0.46038061E-01, 0.41310016E-01,-0.11225826E+00,
             -0.28089361E-01, 0.13309368E-01,-0.41037355E-01,-0.30710904E-01,
             -0.95726689E-02,-0.23963351E-01, 0.16758924E-01, 0.11331435E-01,
              0.23001438E-01, },
            { 0.68212301E-02, 0.87390676E-01, 0.28464913E-01,-0.32604266E-01,
             -0.19913841E-01,-0.69480464E-01,-0.14800048E-01, 0.22902381E-01,
              0.37979659E-01, 0.20491343E-01,-0.21194531E-01,-0.35256979E-02,
             -0.19661651E-02, },
            {-0.39640926E-01, 0.19599617E-01,-0.60924832E-01,-0.32056216E-01,
              0.47802251E-01, 0.17946452E-01,-0.80087418E-02, 0.32511208E-01,
             -0.46744170E-02, 0.21583992E-02,-0.19618925E-01,-0.14446047E-01,
             -0.65722289E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir14Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30586901E+01,-0.16637227E-01, 0.11471306E-01, 0.27490901E-02,
              0.25678012E-02, 0.37677912E-02, 0.98502170E-02,-0.31296548E-02,
             -0.10140381E-02, },
            {-0.13978875E+00, 0.40312845E-01, 0.71543925E-01, 0.19734301E-01,
             -0.78674164E-02,-0.28102145E-01,-0.14239720E-01,-0.12600211E-01,
              0.16756015E-01, },
            { 0.90984887E+00, 0.30035329E+00,-0.44702735E-01,-0.11969292E+00,
             -0.10557233E+00,-0.41341111E-01,-0.80306910E-01, 0.44076685E-01,
              0.34757949E-01, },
            { 0.86042881E-01,-0.43596470E+00,-0.28588781E+00,-0.25438232E-01,
              0.25069173E-01, 0.82588397E-01, 0.67566939E-01, 0.33206843E-01,
             -0.41824643E-01, },
            {-0.12932888E+01,-0.72340810E+00, 0.77358051E-02, 0.28751051E+00,
              0.31436363E+00, 0.10259442E+00, 0.24149394E+00,-0.12717050E+00,
             -0.11702042E+00, },
            {-0.22331176E-01, 0.38658845E+00, 0.28259498E+00, 0.19711521E-02,
             -0.24714001E-01,-0.56775019E-01,-0.61311480E-01,-0.21844221E-01,
              0.23831621E-01, },
            { 0.39463902E+00, 0.47750130E+00, 0.76429829E-01,-0.16718166E+00,
             -0.21960321E+00,-0.68737417E-01,-0.17916369E+00, 0.90928286E-01,
              0.88356271E-01, },
            { 0.15348783E-01,-0.19566035E+00,-0.33543330E+00, 0.15999714E-01,
             -0.25417623E-02, 0.83187250E-02, 0.10059298E-01,-0.18240046E-02,
             -0.25734284E-02, },
            { 0.16147807E-01, 0.36533716E+00,-0.16898610E+00,-0.20461163E-01,
              0.10633610E-01,-0.42438932E-03,-0.20938257E-01,-0.21641953E-02,
              0.55525000E-02, },
            {-0.45814320E-01, 0.26786938E-01, 0.10879262E+00, 0.84931575E-01,
              0.15261321E-01,-0.21805953E-01,-0.14455589E-01, 0.77350605E-02,
              0.63701227E-01, },
            { 0.53684928E-01, 0.14995110E-01,-0.27922710E-01,-0.33047084E-01,
              0.29755291E-01,-0.18964428E-01,-0.24144307E-01, 0.18880604E-01,
              0.58857322E-01, },
            {-0.22821718E+00, 0.75741506E+00, 0.14800967E+01, 0.46163443E-01,
             -0.14732170E+00,-0.15568149E+00,-0.56558043E-01, 0.70298553E-01,
              0.11026065E+00, },
            {-0.79187071E+00,-0.15295109E+01, 0.65210593E+00, 0.17441598E+00,
              0.41551366E-01,-0.66639587E-01, 0.12556066E+00,-0.61385304E-01,
              0.27926465E-01, },
            { 0.87673557E+00,-0.39955699E+00,-0.11206909E+01,-0.28584492E+00,
             -0.59748320E-02, 0.23729971E+00,-0.19509710E+00, 0.81449501E-01,
             -0.73962659E+00, },
            { 0.28284615E+00, 0.89057451E+00, 0.18807463E+00,-0.24254571E+00,
             -0.17398730E-01, 0.79301950E-02, 0.20122550E+00, 0.40711228E-01,
             -0.34922609E+00, },
            { 0.67434180E+00,-0.67298329E+00,-0.74801010E+00,-0.38288945E+00,
              0.36076793E+00, 0.47676265E+00, 0.13727280E+00,-0.16624431E+00,
             -0.43988281E+00, },
            { 0.26705718E+01, 0.34237185E+00,-0.97559363E+00,-0.27165601E+00,
             -0.16340867E+00, 0.21915103E+00,-0.12436141E+00, 0.95994301E-01,
             -0.12902556E+00, },
            {-0.24572020E+01, 0.30924758E+00, 0.28079758E+01, 0.50721401E+00,
             -0.17149813E+00,-0.52098680E+00, 0.85138261E+00,-0.28973010E+00,
              0.17749029E+01, },
            {-0.11304549E+01,-0.23089054E+01,-0.42302907E+00, 0.84942168E+00,
              0.32151359E+00, 0.73436782E-01,-0.39552429E+00,-0.25851825E+00,
              0.62756401E+00, },
            {-0.39601427E+00, 0.25266504E+00,-0.31376165E+00, 0.29245359E+00,
             -0.21280552E+00,-0.37739208E+00,-0.73750824E-01, 0.96640058E-01,
              0.40330350E+00, },
            {-0.21711268E+01, 0.10492510E+01, 0.67013448E+00, 0.88845745E-01,
              0.71819268E-01,-0.15133914E+00,-0.11708713E-01,-0.57785516E-02,
              0.95103070E-01, },
            { 0.18162587E+01, 0.38207385E+00,-0.23117478E+01,-0.38997951E+00,
              0.19229445E+00, 0.28452805E+00,-0.74004084E+00, 0.22057386E+00,
             -0.11552705E+01, },
            { 0.10284771E+01, 0.18343008E+01, 0.33841428E+00,-0.62726456E+00,
             -0.48801002E+00,-0.87205552E-01, 0.21157789E+00, 0.21693024E+00,
             -0.32993799E+00, },
            {-0.18698256E-01,-0.71124723E-02,-0.16465351E-02, 0.46282008E-01,
             -0.15294395E+00,-0.13111427E-03,-0.99337585E-02,-0.43608667E-02,
             -0.90814838E-02, },
            { 0.45116893E-02,-0.97028855E-02, 0.12505715E-01, 0.16136165E+00,
              0.81213906E-01,-0.58312347E-03,-0.18872097E-01,-0.72849258E-02,
              0.14784365E-02, },
            { 0.11196061E+00, 0.20533665E+00, 0.13476495E+00, 0.33945632E+00,
             -0.11701981E+00, 0.27578942E-01,-0.47568362E-01, 0.25997533E-01,
              0.16754409E-01, },
            {-0.93068995E-01, 0.18283080E-01, 0.31589814E-01, 0.22644915E+00,
              0.42842603E+00, 0.72851159E-01, 0.11030859E-01, 0.23866531E-02,
              0.14397724E-01, },
            { 0.61859138E-01, 0.55517521E-01,-0.38489584E-01,-0.42788506E-01,
             -0.22592755E+00,-0.99609442E-01, 0.99531204E-01, 0.29087320E-01,
              0.16361941E+00, },
            {-0.23381135E+00, 0.14897207E-01,-0.29369393E+00, 0.21109748E+00,
             -0.18685365E+00,-0.38392194E-01, 0.10634159E+00,-0.28300032E-02,
              0.66435995E-03, },
            {-0.47440812E+00,-0.11250492E+01,-0.27856839E+00,-0.13569555E+01,
              0.21064770E+00,-0.44342969E-01, 0.76229990E-01,-0.24290409E-01,
             -0.11261434E+00, },
            { 0.10018148E+01, 0.22767542E+00,-0.75611047E-01,-0.59486103E+00,
             -0.14215121E+01,-0.24381903E+00,-0.25383967E+00, 0.49857941E-01,
             -0.13947746E+00, },
            {-0.18549798E+00,-0.21002911E+00,-0.15347001E-01,-0.16834821E+00,
              0.21803890E+00, 0.20929639E+00,-0.10061080E+00,-0.31904507E-01,
             -0.28207496E+00, },
            { 0.25608513E+00,-0.18180054E+00, 0.42851421E+00,-0.17312264E+00,
             -0.17903680E-01, 0.86439624E-01,-0.80150083E-01, 0.17098183E-01,
             -0.30419683E-01, },
            { 0.51279235E+00, 0.15700254E+01, 0.14896685E+00, 0.12490706E+01,
             -0.17830937E+00, 0.75656466E-01,-0.44224098E-01,-0.61300103E-01,
              0.12280394E+00, },
            {-0.14596163E+01,-0.35159817E+00, 0.27934030E+00, 0.59198815E+00,
              0.11541491E+01, 0.17868274E+00, 0.44150430E+00,-0.87764300E-01,
              0.17973503E+00, },
            {-0.24559209E-01, 0.12929835E-01, 0.68207067E-02,-0.12768010E-01,
             -0.38065603E-02, 0.26851548E-01,-0.10830905E-01, 0.77535969E-03,
             -0.17365512E-02, },
            {-0.15774000E-01,-0.33989731E-01,-0.11249687E-01, 0.58662505E-02,
             -0.10439691E-01, 0.19332505E-02, 0.34207720E-01,-0.10329138E-01,
             -0.22380955E-02, },
            { 0.55513047E-01, 0.53344555E-01, 0.27011273E-01,-0.84630392E-01,
              0.11442303E-01, 0.11869611E-01, 0.18336920E-01, 0.21410547E-02,
              0.52863415E-02, },
            { 0.21707227E-01,-0.38661321E-02,-0.25270691E-01,-0.34020606E-01,
             -0.14334359E-02,-0.38540121E-02, 0.28737679E-01, 0.37115016E-02,
             -0.18286736E-02, },
            {-0.17681256E-02,-0.14221156E+00,-0.59075002E-01, 0.19836439E-01,
             -0.76047629E-02,-0.12995811E+00,-0.11560405E+00,-0.16809883E-01,
              0.35188191E-01, },
            { 0.23344200E-01, 0.10418752E+00,-0.43988552E-01,-0.21166109E-01,
             -0.33218442E-02, 0.14282972E+00,-0.15798105E+00, 0.98103182E-02,
              0.10149175E-01, },
            {-0.20782379E-01,-0.12256938E-01, 0.90859635E-02, 0.14533856E-01,
             -0.24522969E-02, 0.11254709E-01, 0.13583394E-02, 0.43356489E-01,
              0.26643327E-02, },
            {-0.14128673E-01, 0.47900616E-02,-0.10882850E-01,-0.47548087E-02,
             -0.53821979E-02, 0.31365620E-02, 0.12613773E-01,-0.60953344E-02,
              0.45044251E-01, },
            { 0.63900851E-01,-0.59823156E-02, 0.71507203E-02,-0.16160239E-01,
             -0.33681545E-01,-0.40392941E-02, 0.60944878E-02, 0.60271841E-01,
              0.34214541E-01, },
            { 0.23164382E-01,-0.68351336E-01, 0.12046548E-01, 0.31958293E-01,
              0.36123714E-02,-0.11482013E-01,-0.12627610E-01,-0.46289042E-01,
              0.64354017E-01, },
            { 0.10392379E-02,-0.16863159E-02, 0.73957373E-02,-0.10698938E-02,
             -0.64807963E-02, 0.99908616E-02, 0.22006985E-02,-0.50359694E-02,
             -0.14203113E-02, },
            { 0.12199912E-01, 0.20179294E-01,-0.62138624E-02, 0.78815650E-02,
             -0.74640587E-02, 0.73885913E-02, 0.22004370E-02, 0.85453289E-02,
              0.76826192E-02, },
            { 0.31520717E-02,-0.12767011E-02, 0.30879308E-02, 0.64275470E-02,
             -0.43601147E-02, 0.98584695E-02, 0.17043862E-02, 0.84848665E-02,
              0.36111218E-02, },
            {-0.43704277E-02,-0.78293349E-03,-0.17360970E-01,-0.11529316E-02,
              0.11252892E-01,-0.21595629E-02,-0.14455176E-02, 0.98986849E-04,
              0.87585188E-02, },
         },
         {
            { 0.26772809E+01,-0.86198850E-02, 0.94552236E-02, 0.55961576E-02,
              0.43859920E-02, 0.41296384E-02, 0.46669282E-02,-0.20577968E-02,
             -0.24387249E-03, },
            {-0.65158509E-01, 0.74515760E-01, 0.72828352E-01, 0.12489909E-01,
              0.58226571E-02,-0.11427455E-01,-0.15949465E-01,-0.65154112E-02,
              0.17882379E-02, },
            { 0.12957149E+01, 0.13960780E+00,-0.44050530E-01,-0.51063329E-01,
             -0.63741840E-01,-0.59596483E-01,-0.53228300E-01, 0.46188686E-01,
              0.11657926E-01, },
            {-0.15501007E+00,-0.51885819E+00,-0.24664447E+00,-0.34158930E-01,
             -0.38116850E-01, 0.12921122E-01, 0.49244039E-01, 0.17124636E-01,
             -0.53521986E-02, },
            {-0.20537572E+01,-0.26769561E+00, 0.49208454E-02, 0.11124684E+00,
              0.16032118E+00, 0.16892996E+00, 0.16521209E+00,-0.12701942E+00,
             -0.60623229E-01, },
            { 0.17818429E+00, 0.42037153E+00, 0.23076217E+00, 0.26306644E-01,
              0.27854430E-01,-0.14802126E-02,-0.40314917E-01,-0.73871380E-02,
              0.24690884E-02, },
            { 0.82360959E+00, 0.18576214E+00, 0.74848384E-01,-0.66371389E-01,
             -0.10588888E+00,-0.11814051E+00,-0.11794991E+00, 0.87646529E-01,
              0.56451142E-01, },
            { 0.19192193E-01,-0.34562644E+00,-0.26493758E+00, 0.78390101E-02,
             -0.45823748E-02, 0.45295575E-03, 0.74164444E-02,-0.11921156E-02,
              0.17643599E-02, },
            { 0.59155165E-04, 0.28991437E+00,-0.32774398E+00,-0.21361131E-01,
              0.55374289E-02, 0.53592436E-02,-0.18224832E-01,-0.22406343E-02,
              0.48251497E-02, },
            { 0.43235659E-02, 0.13838615E+00, 0.46312861E-01, 0.21590933E-01,
              0.16815718E-01,-0.19970588E-01, 0.64824068E-04,-0.19745342E-03,
              0.47364235E-01, },
            { 0.35681468E-01, 0.49083747E-01, 0.10480716E+00, 0.58407634E-02,
              0.63248561E-02, 0.33869767E-02,-0.22746570E-01, 0.86412365E-02,
              0.40659290E-01, },
            {-0.36743322E+00, 0.11957974E+01, 0.91784447E+00, 0.13865001E+00,
             -0.27913701E-01,-0.31462729E-01,-0.16293770E-01, 0.48463598E-01,
             -0.73108380E-03, },
            {-0.61130720E+00,-0.93153477E+00, 0.12323002E+01, 0.17962562E+00,
              0.92122052E-02,-0.88324815E-01, 0.12294372E+00,-0.67933217E-01,
             -0.18161764E-01, },
            { 0.24850307E+00,-0.11966048E+01,-0.56054062E+00,-0.38083974E-01,
             -0.12794180E+00, 0.15602164E+00,-0.32316092E+00, 0.64251274E-01,
             -0.52134240E+00, },
            { 0.38644767E+00, 0.52516818E+00,-0.71278828E+00,-0.22031011E+00,
              0.48846070E-01,-0.63622043E-01, 0.20911933E+00, 0.87847181E-01,
             -0.20944715E+00, },
            { 0.10246786E+01,-0.12937899E+01,-0.40522784E+00,-0.58681512E+00,
              0.13400623E+00, 0.13890813E+00, 0.50930053E-01,-0.12004627E-01,
             -0.18637219E-01, },
            { 0.22995977E+01,-0.29405490E+00,-0.19752228E+01,-0.36606318E+00,
             -0.36992472E-01, 0.18337315E+00,-0.16064256E+00, 0.12423857E+00,
              0.50655352E-02, },
            {-0.79235864E+00, 0.22702663E+01, 0.19082406E+01, 0.38968597E-01,
              0.33631064E-01,-0.34791511E+00, 0.11422637E+01,-0.18621103E+00,
              0.13048165E+01, },
            {-0.16866664E+01,-0.17086961E+01, 0.15004869E+01, 0.64295954E+00,
              0.89812361E-01, 0.19795394E+00,-0.47849467E+00,-0.32040602E+00,
              0.35970518E+00, },
            {-0.68489271E+00, 0.76091516E+00, 0.12355461E+00, 0.44687083E+00,
             -0.14212961E+00,-0.13827266E+00,-0.91967387E-02,-0.65421730E-01,
              0.32495581E-01, },
            {-0.19574792E+01, 0.92517078E+00, 0.13603250E+01, 0.23316291E+00,
             -0.27497089E-01,-0.11032374E+00, 0.18738519E-01,-0.33233728E-01,
             -0.10697253E-01, },
            { 0.59273267E+00,-0.12416369E+01,-0.19759161E+01, 0.18526608E-01,
              0.10760016E+00, 0.19402064E+00,-0.95974118E+00, 0.11718221E+00,
             -0.88802713E+00, },
            { 0.15903203E+01, 0.16918579E+01,-0.10219685E+01,-0.45557439E+00,
             -0.17555867E+00,-0.13787279E+00, 0.30494553E+00, 0.23491570E+00,
             -0.17701098E+00, },
            {-0.17107019E-01,-0.11062065E-01, 0.93461676E-02, 0.72881304E-01,
             -0.11495988E+00,-0.24580231E-02,-0.34615425E-02,-0.27623498E-02,
             -0.68666781E-02, },
            {-0.95973676E-02,-0.17075194E-01, 0.91627231E-02, 0.10242627E+00,
              0.92916168E-01, 0.24848136E-02,-0.11218095E-01,-0.55938442E-02,
              0.10607603E-03, },
            { 0.10907742E+00, 0.13264808E+00, 0.73107131E-01, 0.13814703E+00,
              0.24720404E-01, 0.25445605E-01,-0.19353466E-01,-0.68268122E-03,
              0.79990206E-02, },
            { 0.44637113E-02, 0.49218766E-01,-0.18194420E-01, 0.33405833E-01,
              0.20319274E+00, 0.36236916E-01, 0.17274117E-01,-0.63722855E-02,
              0.33721287E-01, },
            { 0.25666658E-01, 0.92364043E-01,-0.91084383E-01,-0.36218074E+00,
              0.19792330E+00,-0.52057378E-01, 0.40467951E-01,-0.31956337E-01,
              0.12260315E+00, },
            {-0.15529650E+00, 0.11941776E+00,-0.13577776E+00,-0.17587660E+00,
             -0.45683962E+00,-0.65770708E-01, 0.69410942E-01,-0.18389063E-01,
              0.85385451E-02, },
            {-0.37895921E+00,-0.73540902E+00,-0.56422465E-01,-0.47111732E+00,
             -0.20011030E+00,-0.17941417E-01, 0.27807336E-01, 0.76329634E-01,
             -0.42905848E-01, },
            { 0.33812630E+00,-0.11257858E+00, 0.52691784E-01, 0.99168252E-02,
             -0.50797993E+00,-0.69310442E-01,-0.12250749E+00, 0.10490370E+00,
             -0.24429388E+00, },
            {-0.96905559E-01,-0.29989800E+00, 0.95130719E-01, 0.83204031E-01,
             -0.46254903E+00, 0.10771059E+00,-0.16102755E-01, 0.71079575E-01,
             -0.19646353E+00, },
            { 0.21520536E+00,-0.41078597E+00, 0.15975517E+00, 0.48532072E+00,
              0.16240959E+00, 0.74154310E-01,-0.57246845E-01, 0.45167182E-01,
             -0.30693915E-01, },
            { 0.27020487E+00, 0.10784189E+01,-0.53013716E-01, 0.46552581E+00,
              0.92620172E-01, 0.51729448E-01,-0.36019329E-01,-0.14987502E+00,
              0.62644877E-01, },
            {-0.55153680E+00, 0.28454646E+00, 0.94641276E-01, 0.93170024E-01,
              0.38448083E+00, 0.47209587E-01, 0.20970257E+00,-0.13942525E+00,
              0.29898921E+00, },
            {-0.25534306E-01, 0.54413616E-02, 0.14065277E-01,-0.76298043E-02,
             -0.10420343E-01, 0.29778404E-01,-0.41098725E-01,-0.40198048E-02,
             -0.51609301E-02, },
            {-0.10902795E-01,-0.23414372E-01,-0.62101181E-02, 0.97503457E-02,
             -0.91798007E-02, 0.33085518E-01, 0.36749907E-01,-0.47427085E-02,
             -0.77581666E-02, },
            { 0.33572678E-01, 0.60052566E-01, 0.14006686E-01,-0.36195546E-01,
              0.13402758E-01, 0.33575904E-01, 0.35400663E-02,-0.38376546E-02,
              0.42043719E-02, },
            { 0.25882047E-01, 0.10295448E-01,-0.20480629E-02,-0.22618541E-01,
             -0.37360494E-02, 0.12422180E-02, 0.29957978E-01, 0.19331203E-02,
              0.71227252E-02, },
            { 0.44606984E-01,-0.10304903E+00,-0.50200410E-01,-0.11642319E-02,
              0.27468766E-02,-0.13490731E+00,-0.85447587E-01,-0.19251921E-02,
              0.38483702E-01, },
            { 0.11273085E-01, 0.55817258E-01,-0.40558748E-01,-0.12057238E-02,
              0.39492734E-02, 0.10076770E+00,-0.14639939E+00,-0.27176535E-02,
              0.11188758E-01, },
            {-0.19127030E-01,-0.16049713E-01, 0.92721246E-02, 0.23639505E-02,
             -0.56997489E-03, 0.52273925E-02, 0.54570334E-03, 0.32571722E-01,
              0.92297168E-02, },
            {-0.76535931E-02, 0.41365675E-02,-0.10613388E-01, 0.34963775E-02,
             -0.13046856E-02, 0.11616315E-02, 0.74619632E-02,-0.10147394E-01,
              0.32333374E-01, },
            { 0.38845401E-01, 0.55554281E-02,-0.41929283E-02,-0.15542838E-01,
             -0.18608436E-01,-0.68834010E-02, 0.12991249E-01, 0.47244698E-01,
              0.51083591E-01, },
            { 0.33533398E-01,-0.61769698E-01, 0.22310423E-01, 0.10246638E-01,
             -0.62199375E-02,-0.11598877E-01,-0.16985392E-01,-0.50428756E-01,
              0.41603886E-01, },
            { 0.17688370E-02,-0.36227442E-02, 0.84820492E-02, 0.28933492E-02,
             -0.58887932E-02, 0.68608033E-02, 0.30614494E-03,-0.15316576E-02,
             -0.24033547E-02, },
            { 0.92781493E-02, 0.14051887E-01,-0.10239172E-01, 0.53211111E-02,
             -0.21397618E-02, 0.44980948E-02, 0.54727206E-02, 0.70043919E-02,
              0.11711870E-01, },
            { 0.68015773E-02, 0.21883957E-02, 0.35151727E-02, 0.67476844E-02,
             -0.67525292E-02, 0.54354891E-02, 0.19866449E-02, 0.76344414E-02,
              0.59919464E-02, },
            {-0.26257581E-02,-0.48037629E-04,-0.10798140E-01, 0.56925311E-03,
              0.80262767E-02,-0.51455793E-03,-0.93577243E-03,-0.71163801E-03,
              0.78903772E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir15F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.56421361E+01,-0.81572428E-01, 0.20138618E+00, 0.42779669E-01,
              0.44377025E-01, 0.38025160E-02,-0.87150786E-03, 0.21708569E-01,
             -0.23025220E-01,-0.33594195E-02, 0.21446483E-03,-0.18263943E-02,
             -0.12104558E-01, },
            {-0.10026588E+01,-0.77602589E+00, 0.80115181E+00,-0.27182043E+00,
              0.54614949E+00,-0.18997940E+00, 0.82413152E-01,-0.38998052E-01,
             -0.97885408E-01, 0.13837354E+00, 0.10652158E+00,-0.15692274E+00,
              0.99289812E-01, },
            { 0.80236692E+01, 0.19602764E+01,-0.35586731E+01, 0.28472129E+00,
             -0.56597894E+00,-0.44183967E+00, 0.13710301E+00, 0.35544112E-01,
              0.41852084E+00, 0.63865769E+00,-0.19955039E+00, 0.11715313E+00,
              0.86535461E-01, },
            { 0.14375471E+02, 0.64524765E+01,-0.62800088E+01, 0.55624423E+01,
             -0.81383467E+01, 0.28389122E+01,-0.18190529E+01, 0.17703345E+00,
              0.13019991E+01,-0.19010282E+01,-0.25744607E+01, 0.22809548E+01,
             -0.81978828E+00, },
            {-0.53995434E+02,-0.20322048E+02, 0.18375986E+02,-0.74895936E+00,
             -0.51071935E+01, 0.81174737E+00,-0.39087531E+00,-0.14415859E+01,
             -0.10194235E+01,-0.38520768E+01, 0.11969794E+01,-0.14105285E+01,
             -0.62901032E+00, },
            {-0.18940151E+02,-0.23667875E+02, 0.88806610E+01,-0.27459188E+02,
              0.41921669E+02,-0.13132250E+02, 0.69411273E+01,-0.15238301E+01,
             -0.77653646E+01, 0.84578733E+01, 0.15979069E+02,-0.12099964E+02,
              0.12503768E+01, },
            { 0.11738895E+03, 0.68091232E+02,-0.39820190E+02,-0.16654693E+01,
              0.25928150E+02, 0.13787634E+01,-0.43914926E+00, 0.40016642E+01,
              0.47450104E+00, 0.83207235E+01,-0.34342322E+01, 0.43038507E+01,
              0.21257403E+01, },
            {-0.37460907E+02, 0.41022415E+02, 0.10812599E+02, 0.51976315E+02,
             -0.94334923E+02, 0.25150223E+02,-0.89467859E+01, 0.58774261E+01,
              0.20332344E+02,-0.17085917E+02,-0.38310886E+02, 0.27109829E+02,
              0.87873173E+00, },
            {-0.12372593E+03,-0.87597099E+02, 0.38698399E+02, 0.42182999E+01,
             -0.37059380E+02,-0.34062238E+01, 0.15188286E+01,-0.37199290E+01,
              0.36815086E+00,-0.74235358E+01, 0.42371931E+01,-0.49250031E+01,
             -0.28798954E+01, },
            { 0.87714088E+02,-0.30394653E+02,-0.28824894E+02,-0.42016949E+02,
              0.95767807E+02,-0.21253895E+02, 0.34847987E+01,-0.83288565E+01,
             -0.22973545E+02, 0.15845953E+02, 0.39268745E+02,-0.26755207E+02,
             -0.26882515E+01, },
            { 0.49962570E+02, 0.37968533E+02,-0.13776732E+02,-0.21174350E+01,
              0.16793774E+02, 0.16195391E+01,-0.80588609E+00, 0.10947104E+01,
             -0.22045150E+00, 0.22938368E+01,-0.17940499E+01, 0.19093058E+01,
              0.13130230E+01, },
            {-0.43963936E+02, 0.73540716E+01, 0.14685716E+02, 0.12209099E+02,
             -0.35778717E+02, 0.65739970E+01, 0.25819159E+00, 0.38785126E+01,
              0.92260513E+01,-0.54415417E+01,-0.14486173E+02, 0.96487923E+01,
              0.12716563E+01, },
            {-0.21124800E+00, 0.17043127E+01, 0.20464387E+01, 0.56599015E-02,
             -0.10268541E+00, 0.23923576E-01, 0.46511523E-01, 0.47292676E-01,
             -0.43041897E-02,-0.13252716E-01, 0.11830020E-02, 0.10728045E-01,
             -0.11813324E-01, },
            { 0.11713128E+00,-0.18258561E+01, 0.17008826E+01,-0.15215686E-01,
              0.11784354E+00,-0.43866588E-02,-0.22057734E-01,-0.63331686E-02,
              0.49470387E-01,-0.17963713E-01,-0.59963893E-02,-0.16910573E-01,
              0.36876634E-01, },
            {-0.67201293E+00, 0.13338754E+00,-0.21074166E+01, 0.51338410E+00,
             -0.38658473E+00, 0.81445105E-01,-0.95247170E-02,-0.41881129E+00,
              0.26037970E+00, 0.10410904E+00, 0.31723064E+00, 0.19097517E+00,
              0.45013998E-01, },
            {-0.20128696E+00, 0.15844090E+01,-0.83317590E+00, 0.54410454E-01,
              0.24460571E+00,-0.36854930E-01,-0.48869914E+00, 0.82281917E-01,
              0.58834499E+00, 0.92693098E-01, 0.62349554E-01, 0.23411550E+00,
              0.29134646E+00, },
            { 0.55470839E+01, 0.96163158E+01, 0.50997686E+01, 0.68665922E+00,
              0.65423720E-01,-0.24311846E+00,-0.12406635E+01,-0.11371589E+01,
              0.11067604E+01,-0.51531062E-01,-0.74737412E+00, 0.54248173E-01,
              0.63131139E-01, },
            {-0.68635064E+00,-0.12532818E+02, 0.31381702E+01, 0.58695197E+00,
             -0.89948498E-01,-0.52485317E+00, 0.11644162E+01,-0.24193650E+00,
              0.99850368E+00, 0.73659611E+00, 0.30205196E+00, 0.79808694E+00,
             -0.56268388E+00, },
            { 0.84152431E+01,-0.18351845E+01, 0.22851471E+02,-0.62919526E+01,
              0.38386998E+00,-0.11887341E+01,-0.32209761E+01, 0.62786980E+01,
             -0.32539845E+01, 0.33874832E-01,-0.68638234E+01,-0.39917052E+01,
              0.50753254E+00, },
            { 0.54675989E+01,-0.14342614E+02, 0.20977676E+02,-0.40119495E-01,
             -0.43394451E+01, 0.13764828E+01, 0.10244990E+02, 0.31333181E+00,
             -0.92915649E+01,-0.30012243E+01,-0.67161626E+00,-0.32468977E+01,
             -0.40313520E+01, },
            {-0.35370174E+02,-0.69950500E+02, 0.26502401E+02, 0.31762896E+01,
             -0.11033854E+01,-0.33225198E+01, 0.66381388E+01, 0.81779928E+01,
             -0.61022053E+01,-0.22130617E+00, 0.75198293E+01, 0.25410479E+00,
              0.76021153E+00, },
            {-0.13441768E+01, 0.27343557E+02, 0.47622552E+01,-0.13991564E+02,
              0.15639170E+01, 0.82513838E+01,-0.82063255E+01,-0.34090364E+01,
             -0.11663890E+02,-0.46683278E+01,-0.29037738E+01,-0.57367353E+01,
              0.13532746E+01, },
            {-0.48274441E+02, 0.83131844E+02,-0.96900970E+02, 0.27448704E+02,
              0.21074257E+02, 0.43562355E+01, 0.23491089E+02,-0.30603300E+02,
              0.98027277E+01,-0.17388166E+01, 0.41459393E+02, 0.21956772E+02,
             -0.74577127E+01, },
            {-0.57677406E+02, 0.61015911E+02,-0.67599236E+02,-0.76322885E+01,
              0.15360656E+02,-0.13954189E+02,-0.63971962E+02,-0.57594633E+00,
              0.46648548E+02, 0.18534496E+02,-0.11813200E+01, 0.14227717E+02,
              0.16818104E+02, },
            { 0.80797073E+02, 0.17325121E+03,-0.21808232E+03,-0.25066273E+02,
              0.77090278E+01, 0.19229862E+02,-0.16147860E+02,-0.21793207E+02,
              0.11739250E+02, 0.27427473E+01,-0.23950100E+02,-0.14955986E+01,
             -0.43279033E+01, },
            { 0.23412638E+02, 0.63171879E+02,-0.99681198E+02, 0.62766190E+02,
             -0.79878149E+01,-0.30435446E+02, 0.21497587E+02, 0.20054741E+02,
              0.38359371E+02, 0.92688770E+01, 0.10108423E+02, 0.14417062E+02,
             -0.20493574E+01, },
            { 0.14545422E+03,-0.31993140E+03, 0.20626822E+03,-0.59260872E+02,
             -0.98615173E+02,-0.91029177E+01,-0.61238270E+02, 0.69641464E+02,
             -0.53348732E+01, 0.55038424E+01,-0.10444603E+03,-0.51997677E+02,
              0.24604845E+02, },
            { 0.17732971E+03,-0.14906339E+03, 0.97182152E+02, 0.28640442E+02,
             -0.17308891E+02, 0.45831375E+02, 0.16195528E+03,-0.62566624E+01,
             -0.10428041E+03,-0.45196327E+02, 0.11912194E+02,-0.28687082E+02,
             -0.28267488E+02, },
            {-0.71918938E+02,-0.20264278E+03, 0.37785242E+03, 0.43567059E+02,
             -0.15192982E+02,-0.31795984E+02, 0.19159107E+02, 0.24578568E+02,
             -0.10395302E+02,-0.54584193E+01, 0.29443462E+02, 0.18643064E+01,
              0.73603716E+01, },
            {-0.46939167E+02,-0.19401114E+03, 0.19549149E+03,-0.98742599E+02,
              0.65470691E+01, 0.41411484E+02,-0.23003340E+02,-0.32262764E+02,
             -0.48849609E+02,-0.63948112E+01,-0.14122622E+02,-0.15711078E+02,
              0.33360174E+01, },
            {-0.20310941E+03, 0.43382684E+03,-0.22873076E+03, 0.61049896E+02,
              0.15113927E+03, 0.12178139E+02, 0.69404495E+02,-0.76307945E+02,
             -0.99914246E+01,-0.87610970E+01, 0.11714094E+03, 0.56842529E+02,
             -0.31153524E+02, },
            {-0.20313609E+03, 0.19473550E+03,-0.81566399E+02,-0.31190516E+02,
              0.46536431E+01,-0.60174667E+02,-0.17905360E+03, 0.15700250E+02,
              0.10812027E+03, 0.48076645E+02,-0.19411180E+02, 0.28377907E+02,
              0.18879490E+02, },
            { 0.18400129E+02, 0.92124519E+02,-0.19595229E+03,-0.22818579E+02,
              0.91224928E+01, 0.16459476E+02,-0.86445189E+01,-0.10113220E+02,
              0.36470227E+01, 0.31096530E+01,-0.12310074E+02,-0.65811318E+00,
             -0.39445660E+01, },
            { 0.27785633E+02, 0.12070577E+03,-0.10845207E+03, 0.50968784E+02,
              0.20172384E+01,-0.18991886E+02, 0.84511404E+01, 0.16190018E+02,
              0.21269348E+02, 0.98832011E+00, 0.67662487E+01, 0.63501000E+01,
             -0.22222323E+01, },
            { 0.10348282E+03,-0.20003775E+03, 0.10056244E+03,-0.24191299E+02,
             -0.75551025E+02,-0.66165800E+01,-0.28941280E+02, 0.32210011E+02,
              0.89458332E+01, 0.52425108E+01,-0.48151104E+02,-0.23473763E+02,
              0.13704450E+02, },
            { 0.76933624E+02,-0.97199905E+02, 0.34218758E+02, 0.94315853E+01,
             -0.39759991E+00, 0.27437757E+02, 0.72066147E+02,-0.95865564E+01,
             -0.42264874E+02,-0.18651270E+02, 0.93191538E+01,-0.11183516E+02,
             -0.35364709E+01, },
            { 0.97871020E-01, 0.20281778E+00, 0.90489715E-01,-0.95290929E+00,
             -0.27933320E+00, 0.52461393E-01, 0.72199428E-02, 0.15705125E-01,
              0.41039664E-01,-0.36827065E-01,-0.15596761E-01,-0.21006119E-01,
             -0.74302865E-03, },
            { 0.18413863E+00,-0.22145297E-01,-0.21845268E-01, 0.28004029E+00,
             -0.71528941E+00,-0.12112075E-01,-0.43858774E-02,-0.32223654E-02,
             -0.29513298E-02,-0.20680975E-01,-0.50281603E-02,-0.97669261E-02,
              0.16754756E-01, },
            {-0.50228141E-01, 0.86401850E+00, 0.37329832E+00,-0.43455347E-01,
             -0.37337765E+00, 0.31667180E-01,-0.33294806E+00, 0.19361353E+00,
             -0.14481428E+00, 0.17382690E+00,-0.27535016E-01,-0.52372944E-01,
              0.51665086E-01, },
            { 0.77041340E+00,-0.10802412E+00, 0.78209734E+00, 0.82496119E+00,
             -0.11585006E+00,-0.95668048E-01, 0.54860033E-01, 0.16625006E+00,
             -0.91622919E-01,-0.65615319E-01, 0.60415332E-03, 0.15374319E+00,
              0.46706021E-01, },
            { 0.54401726E+00,-0.93265307E+00,-0.56378675E+00, 0.10645507E+02,
              0.57229429E+00,-0.12630768E+01,-0.64789832E+00,-0.60353708E+00,
             -0.18424869E+00, 0.13993631E+00, 0.48576195E-01, 0.22784832E+00,
             -0.42130151E+00, },
            { 0.27307931E+00, 0.10153091E+01, 0.11288965E+01,-0.25885277E+01,
              0.76768188E+01, 0.73958021E+00, 0.39947239E+00,-0.11303554E+01,
              0.74285984E+00, 0.42960650E+00,-0.46359366E-02, 0.57497424E+00,
             -0.24941255E+00, },
            {-0.27576047E+00,-0.49189267E+01,-0.30248418E+01, 0.82247725E+01,
             -0.40766196E+01, 0.37377694E+00, 0.15920104E+01,-0.10867124E+01,
              0.17805321E+01,-0.15497746E+01, 0.73062021E+00, 0.45829788E-01,
              0.16331393E+00, },
            {-0.90494785E+01,-0.15745066E+00,-0.52452645E+01,-0.91057497E+00,
              0.86765213E+01, 0.23414004E+01,-0.92353123E+00,-0.28763181E+00,
              0.22783259E-01,-0.82801855E+00,-0.28931600E+00,-0.10174736E+01,
              0.16624613E+00, },
            {-0.10378919E+02, 0.58001833E+01,-0.22736102E+00,-0.23410555E+02,
              0.69524856E+01, 0.71683693E+01, 0.60270047E+01, 0.22774773E+01,
              0.14830394E+01, 0.16389118E+01,-0.30775643E-02,-0.78709882E+00,
              0.30579998E+01, },
            {-0.15465743E-02,-0.94203407E+00,-0.70205545E+01, 0.62845588E+01,
             -0.72446790E+01,-0.44833145E+01,-0.13172768E+01, 0.54477401E+01,
             -0.26917975E+01,-0.22063324E+01, 0.88429081E+00,-0.25067618E+01,
              0.70779270E+00, },
            {-0.41653025E+00, 0.77432489E+01, 0.44009709E+01,-0.26288893E+02,
              0.11670718E+02,-0.90083349E+00,-0.26713226E+01, 0.27358816E+01,
             -0.44873104E+01, 0.38639853E+01,-0.20221062E+01, 0.50795317E+00,
             -0.10303019E+01, },
            { 0.25454908E+02,-0.35198110E+00, 0.91766758E+01, 0.36100191E+00,
             -0.26752708E+02,-0.79657893E+01, 0.40400887E+01,-0.13829775E+00,
              0.47660896E+00, 0.37345989E+01, 0.49525353E+00, 0.16495343E+01,
             -0.32004577E+00, },
            { 0.26923609E+02,-0.12705561E+02, 0.77154827E+01, 0.47243571E+01,
             -0.21181990E+02,-0.15098260E+02,-0.13009450E+02,-0.24404039E+01,
             -0.28010442E+01,-0.60208449E+01,-0.58497292E+00, 0.13716583E+01,
             -0.60833530E+01, },
            {-0.75704546E+01,-0.52916546E+01, 0.16575653E+02,-0.68116546E+01,
             -0.28722118E+02, 0.83101597E+01, 0.55782187E+00,-0.92281694E+01,
              0.28159809E+01, 0.46327081E+01,-0.30307541E+01, 0.37459197E+01,
             -0.88170326E+00, },
            { 0.22301376E+00,-0.31962299E+01,-0.10003719E+01, 0.17513151E+02,
             -0.10250562E+02, 0.30598539E+00, 0.16885284E+01,-0.19954596E+01,
              0.31069324E+01,-0.28071928E+01, 0.13887666E+01,-0.67095476E+00,
              0.10415512E+01, },
            {-0.20710796E+02, 0.89437598E+00,-0.42360606E+01, 0.20117793E+01,
              0.17224691E+02, 0.66144185E+01,-0.45690055E+01,-0.18276887E+00,
             -0.40452820E+00,-0.32121379E+01,-0.92587329E-01,-0.48688596E+00,
             -0.14324804E+00, },
            {-0.18712111E+02, 0.78072848E+01,-0.87452765E+01, 0.12498962E+02,
              0.17300501E+02, 0.10375720E+02, 0.82283955E+01, 0.58782387E+00,
              0.11830972E+01, 0.48081841E+01, 0.77937973E+00,-0.73815542E+00,
              0.35488698E+01, },
            { 0.98545418E+01, 0.57349968E+01,-0.14113446E+02, 0.14990284E+01,
              0.34328514E+02,-0.51231594E+01, 0.11489596E+01, 0.57513089E+01,
             -0.11141500E+01,-0.30930831E+01, 0.25668890E+01,-0.20892141E+01,
              0.63647670E+00, },
            {-0.29918633E-01, 0.19186273E+00,-0.70395455E-01,-0.77079989E-01,
              0.31567018E-01,-0.38534543E+00,-0.50261647E+00,-0.21860762E-01,
              0.62904462E-01,-0.54080831E-02,-0.92705982E-02,-0.40175516E-01,
             -0.11157575E-01, },
            { 0.72898664E-01,-0.50168093E-01,-0.10570346E+00,-0.67150467E-02,
             -0.14133672E+00, 0.50753140E+00,-0.42581895E+00,-0.46848822E-01,
              0.22408159E-02, 0.83338562E-02, 0.41946103E-02,-0.58955308E-02,
             -0.26281827E-03, },
            {-0.26040110E+00, 0.76159227E+00, 0.23537507E-01,-0.80746651E-01,
             -0.17011431E+00, 0.11547460E+00,-0.15361232E+00, 0.17724212E+00,
              0.76225460E-01,-0.13381564E-01,-0.60893386E-02,-0.11518117E+00,
             -0.31129291E-01, },
            {-0.51667415E-01,-0.24441694E+00,-0.31773397E+00, 0.81835248E-01,
             -0.46549350E+00, 0.17142838E+00, 0.30523732E+00,-0.57707924E-01,
             -0.20797607E-01,-0.15126070E-01,-0.33176012E-01,-0.42346753E-01,
             -0.11050799E-01, },
            { 0.59186459E-01,-0.11252819E+01, 0.44319230E+00,-0.11809949E+00,
             -0.14963280E+00, 0.52666683E-01,-0.52684325E+00,-0.40326202E+00,
              0.12269677E+00, 0.28846711E+00, 0.12978290E+00, 0.32258502E+00,
             -0.14862968E+00, },
            {-0.80570722E+00, 0.17221979E+01, 0.46162105E+00,-0.34327608E+00,
              0.13253006E+01, 0.44412488E+00, 0.28846559E+00,-0.34307161E+00,
             -0.36704317E+00, 0.56617562E-01, 0.18177399E+00, 0.25099775E+00,
             -0.99185131E-01, },
            {-0.19901596E+00,-0.17126943E+01, 0.37629980E+00, 0.45409298E+00,
              0.27510104E+00, 0.15234837E-01, 0.65682179E+00,-0.45590812E+00,
             -0.21512812E+00, 0.90877265E-01,-0.33986196E-01, 0.28594887E+00,
              0.92194878E-01, },
            {-0.24209818E+00, 0.65725392E+00, 0.75583166E+00,-0.39450571E+00,
              0.12069463E+01,-0.75257272E+00,-0.52532136E+00, 0.99424221E-01,
              0.93963951E-01, 0.45424551E-02, 0.14965154E+00, 0.19529074E+00,
              0.89346744E-01, },
            { 0.76194292E+00, 0.21998913E+01,-0.61996305E+00, 0.38401404E+00,
              0.65990424E+00, 0.10685669E+01, 0.16874920E+01, 0.10111738E+01,
             -0.47734886E+00,-0.55955160E+00,-0.36574565E-01,-0.54719472E+00,
              0.21022716E+00, },
            { 0.96725523E+00,-0.21709857E+01,-0.58654177E+00, 0.52371088E-01,
             -0.19939440E+01,-0.13083379E+01, 0.97265768E+00, 0.10472490E+01,
              0.79788148E+00,-0.23003300E+00,-0.31732571E+00,-0.59758753E+00,
              0.17752629E-01, },
            {-0.48324671E-01,-0.39871577E-01, 0.35558338E-02,-0.26979163E-01,
              0.43359697E-01, 0.12917671E-01, 0.17010419E-01, 0.27531299E+00,
             -0.38564362E-01,-0.34878931E-02,-0.61485823E-02,-0.97274072E-02,
             -0.18774368E-01, },
            { 0.92524938E-01,-0.57997428E-01, 0.36104083E-01, 0.66196665E-01,
             -0.40818259E-01, 0.68578757E-02, 0.56706844E-02, 0.14657995E-01,
              0.25169230E+00, 0.17018361E-01,-0.13085437E-02, 0.40550678E-03,
             -0.75565781E-02, },
            {-0.27891763E-01,-0.61812479E-01,-0.66995524E-01, 0.60951114E-01,
             -0.16235442E+00,-0.18117538E+00,-0.63721985E-01, 0.19130240E+00,
              0.34930453E+00, 0.10013775E-03,-0.18040525E-01, 0.93262065E-02,
              0.29071178E-01, },
            { 0.37352061E+00, 0.23084080E+00,-0.11973947E+00, 0.39940480E-01,
              0.16606483E+00,-0.25743233E-01,-0.50882917E-01,-0.38867268E+00,
              0.90072393E-01,-0.18342920E-01,-0.16830478E-01, 0.65812245E-02,
             -0.11528600E-01, },
            {-0.17556234E+00,-0.39191529E-01, 0.14456748E-01,-0.31045396E-01,
             -0.29462801E-01, 0.17225457E-02, 0.81582461E-02, 0.32372467E-01,
             -0.37683394E-01, 0.90681992E-01, 0.16984081E+00,-0.56379475E-02,
             -0.16191302E-01, },
            { 0.32786489E-02, 0.13823365E-01,-0.58228850E-01, 0.46045277E-01,
             -0.32285951E-01,-0.29173903E-01,-0.13254940E-02, 0.33482354E-01,
              0.14193523E-01,-0.18655913E+00, 0.75768754E-01, 0.28235951E-01,
             -0.12667628E-01, },
            {-0.63085563E-01,-0.16578859E-01,-0.65637636E-02, 0.90851039E-02,
             -0.29999416E-01, 0.40609598E-01,-0.12365533E-01, 0.32348711E-01,
              0.48897574E-02,-0.33822318E-02,-0.95927659E-02,-0.10472462E+00,
             -0.49497215E-02, },
            { 0.13566672E-01, 0.51126843E-02, 0.69563393E-02, 0.21742018E-01,
              0.64151838E-01,-0.15110991E-01,-0.11155133E-01,-0.23500491E-02,
             -0.49487525E-02,-0.34136429E-01,-0.13634215E-01,-0.51828269E-02,
             -0.11964961E+00, },
            { 0.16280914E-01,-0.15323651E-01, 0.11045886E-01, 0.15388856E-01,
              0.71778158E-02, 0.13341840E-01, 0.86160125E-02, 0.35415865E-01,
              0.20430239E-01,-0.12541118E-01, 0.64188964E-02, 0.69954526E-02,
              0.14519762E-01, },
            { 0.26640084E-01,-0.81357203E-01,-0.82582489E-01, 0.33618689E-02,
             -0.22538289E-01,-0.19978173E-01, 0.16238056E-01,-0.12985623E-01,
              0.16977180E-01,-0.14756378E-01,-0.48074991E-03, 0.21765309E-02,
              0.14442624E-01, },
            { 0.29858338E-01, 0.81244618E-01,-0.43053649E-01, 0.27717860E-01,
              0.12420647E-01,-0.21559792E-01, 0.41899722E-01,-0.27755309E-01,
              0.32173231E-01,-0.28853366E-01, 0.96155703E-02, 0.35324396E-03,
             -0.10962152E-01, },
            {-0.35660431E-01,-0.20827377E-01,-0.27171995E-01,-0.39691202E-01,
              0.78898314E-02,-0.13305612E-01,-0.33514686E-02,-0.12471182E-01,
              0.22468641E-01, 0.17353624E-01,-0.10118406E-01,-0.17869116E-02,
              0.15280945E-01, },
         },
         {
            { 0.90913858E+01, 0.84599406E-01, 0.16230763E+00, 0.69487393E-01,
             -0.29570591E-02,-0.92661269E-02,-0.40300194E-01, 0.30152716E-01,
             -0.22117848E-01, 0.44199214E-02,-0.37470509E-02, 0.57405089E-02,
             -0.12053542E-01, },
            {-0.76227540E+00,-0.81212771E+00, 0.16045764E-01,-0.18294992E-01,
              0.17572674E+00, 0.21314566E+00, 0.27046995E-01, 0.82527503E-01,
             -0.28845841E+00, 0.24496240E-02, 0.97445555E-01,-0.39677333E-01,
              0.67152999E-01, },
            { 0.24650427E+02,-0.44397229E+00,-0.17870303E+01, 0.55547500E+00,
              0.10984249E+01, 0.28475463E+00, 0.69275779E+00,-0.18453959E+00,
              0.40410057E+00, 0.59705932E-01,-0.23535267E+00, 0.43878641E-01,
              0.11331538E+00, },
            {-0.10896392E+02, 0.80607300E+01, 0.99370747E+01, 0.28089604E+00,
             -0.36042504E+01,-0.39281123E+01,-0.43405023E+00,-0.12527297E+00,
              0.35088420E+01,-0.13881062E+01,-0.10145779E+01, 0.22173715E+00,
             -0.23965409E+00, },
            {-0.13814932E+03,-0.35444760E+01, 0.49864116E+01,-0.73278651E+01,
             -0.15269502E+02,-0.34637457E+00,-0.28207886E+01, 0.86998552E+00,
             -0.19370019E+01, 0.66754687E+00, 0.17929994E+01,-0.26302782E+00,
             -0.35202691E+00, },
            { 0.14604482E+03,-0.40838085E+02,-0.57767654E+02, 0.57909651E+01,
              0.29187016E+02, 0.18583534E+02, 0.96246183E+00,-0.31996384E+01,
             -0.18213390E+02, 0.86978769E+01, 0.45727201E+01, 0.29856852E+00,
             -0.23486688E+01, },
            { 0.25891187E+03, 0.25923477E+02,-0.39683554E+01, 0.22933840E+02,
              0.49968544E+02,-0.23614740E+01, 0.36843927E+01,-0.18787698E+01,
              0.36446035E+01,-0.35390708E+01,-0.51340876E+01, 0.95419985E+00,
              0.99823630E+00, },
            {-0.41596628E+03, 0.86813042E+02, 0.11431066E+03,-0.30645103E+02,
             -0.82427086E+02,-0.37951321E+02, 0.10900590E+01, 0.11784755E+02,
              0.41145611E+02,-0.19994535E+02,-0.94752426E+01,-0.28417885E+01,
              0.96225624E+01, },
            {-0.22837682E+03,-0.43467033E+02, 0.16895311E+01,-0.27585829E+02,
             -0.61419991E+02, 0.55255609E+01,-0.12157919E+01, 0.21039855E+01,
             -0.30910072E+01, 0.53918390E+01, 0.60075970E+01,-0.13708730E+01,
             -0.16586262E+01, },
            { 0.45341833E+03,-0.79230301E+02,-0.95844460E+02, 0.45088043E+02,
              0.94350952E+02, 0.35473499E+02,-0.38808095E+01,-0.14583707E+02,
             -0.40570156E+02, 0.19583448E+02, 0.89746466E+01, 0.42096753E+01,
             -0.11760089E+02, },
            { 0.78894501E+02, 0.21365410E+02,-0.97209376E+00, 0.11413410E+02,
              0.25590933E+02,-0.31731806E+01,-0.29654628E+00,-0.94074392E+00,
              0.10037756E+01,-0.26313784E+01,-0.24205432E+01, 0.61006391E+00,
              0.94114125E+00, },
            {-0.17161839E+03, 0.26147404E+02, 0.29388216E+02,-0.20523869E+02,
             -0.37670624E+02,-0.12390260E+02, 0.22395270E+01, 0.60695238E+01,
              0.14398425E+02,-0.68785167E+01,-0.31523423E+01,-0.18456868E+01,
              0.46553698E+01, },
            {-0.17224808E+00, 0.15295334E+01, 0.16319659E+01, 0.16117379E-01,
             -0.11191823E+00,-0.55230491E-01,-0.11635203E-01, 0.23330303E-01,
              0.14298861E-01, 0.73818795E-04,-0.13618111E-01, 0.40551308E-02,
             -0.13926427E-01, },
            { 0.25908184E+00,-0.15589666E+01, 0.15277148E+01,-0.52449096E-01,
              0.15074499E+00, 0.48045823E-02, 0.26224246E-01,-0.31880584E-01,
              0.30712897E-01,-0.15925413E-01, 0.30503899E-01,-0.15572473E-01,
              0.10377554E-01, },
            {-0.44175947E+00,-0.14467432E+00,-0.23857009E+01,-0.93523920E-01,
              0.39645356E+00, 0.34704551E+00,-0.23056675E+00,-0.74715160E-01,
             -0.10230460E-01, 0.67601532E-01, 0.24271284E+00,-0.83723627E-02,
              0.27769657E-01, },
            { 0.43120182E+00, 0.15971814E+01,-0.11812301E+01,-0.19957671E-01,
              0.79163861E+00, 0.32819316E+00, 0.47173291E-01, 0.10105979E+00,
              0.29491344E+00, 0.12614381E+00,-0.77709101E-01, 0.28120133E+00,
             -0.83021149E-01, },
            { 0.28239892E+01, 0.20100613E+02,-0.53325434E+01, 0.45510817E+00,
              0.10075530E+01, 0.19475687E+01, 0.87764084E+00,-0.28519857E+00,
             -0.21241654E+00,-0.31646341E+00,-0.12565173E+00,-0.32961044E-01,
              0.23946792E+00, },
            {-0.30458715E+01, 0.34423187E+01, 0.15719871E+02, 0.27821069E+01,
             -0.24117477E+01,-0.53116071E+00,-0.93346071E+00, 0.38857710E+00,
             -0.41621810E+00, 0.30858749E+00,-0.85805213E+00, 0.56975406E+00,
             -0.45797646E+00, },
            { 0.64062252E+01,-0.17833441E+02, 0.38745689E+02, 0.15681111E+01,
             -0.13188357E+02,-0.35303750E+01, 0.42505393E+01, 0.23458998E+01,
              0.36971977E+00,-0.15250417E+01,-0.25250385E+01, 0.11713949E+01,
             -0.25048587E+00, },
            {-0.47601018E+01,-0.14131312E+02, 0.16657829E+02, 0.26456211E+01,
             -0.20797195E+02,-0.33104246E+01, 0.13996363E+01,-0.86525661E+00,
             -0.32111042E+01,-0.41198349E+01,-0.53699982E+00,-0.44991617E+01,
              0.22725008E+01, },
            {-0.15645193E+02,-0.83720718E+02, 0.81215302E+02,-0.40556893E+01,
              0.73948927E+01,-0.12809274E+02,-0.93950243E+01, 0.31676583E+01,
              0.58742332E+01, 0.21229396E+01, 0.13009325E+01, 0.20239291E+01,
             -0.53447652E+00, },
            { 0.26276909E+02,-0.75438370E+02,-0.32025635E+02,-0.30627806E+02,
              0.26027321E+02, 0.33012908E+01, 0.51888561E+01,-0.51801095E+01,
              0.33771536E+01,-0.15328786E+01, 0.53555956E+01,-0.47251158E+01,
              0.42387228E+01, },
            {-0.45288258E+02, 0.15141853E+03,-0.23644781E+03,-0.37938159E+01,
              0.85512016E+02, 0.71108932E+01,-0.15790138E+02,-0.14872607E+02,
             -0.79568539E+01, 0.77140079E+01, 0.65682583E+01,-0.10868766E+02,
             -0.11020887E+01, },
            { 0.15142425E+02, 0.92271660E+02,-0.60111969E+02,-0.21405069E+02,
              0.12360047E+03, 0.14346115E+02,-0.23227024E+02, 0.11077701E+02,
              0.13304357E+02, 0.28648264E+02, 0.70955434E+01, 0.23707230E+02,
             -0.13942078E+02, },
            { 0.22773668E+02, 0.94970367E+02,-0.27014450E+03, 0.13208767E+02,
             -0.51395786E+02, 0.31416155E+02, 0.32453068E+02,-0.10277439E+02,
             -0.23690439E+02,-0.66481109E+01,-0.47724204E+01,-0.81575747E+01,
             -0.47356147E-01, },
            {-0.76574455E+02, 0.26398227E+03,-0.96341965E+02, 0.11079182E+03,
             -0.88914604E+02,-0.30642164E+01,-0.13225583E+02, 0.17785391E+02,
             -0.11198152E+02, 0.20659568E+01,-0.13020063E+02, 0.14299355E+02,
             -0.16164017E+02, },
            { 0.15472884E+03,-0.41988278E+03, 0.56520874E+03, 0.14640667E+01,
             -0.21045242E+03, 0.49664984E+01, 0.16372763E+02, 0.38117653E+02,
              0.34166718E+02,-0.15126781E+02,-0.92465889E+00, 0.30510057E+02,
              0.76909704E+01, },
            {-0.28304167E+02,-0.24170132E+03, 0.12176568E+03, 0.48061871E+02,
             -0.29643774E+03,-0.33637665E+02, 0.83702583E+02,-0.42672482E+02,
             -0.25475426E+02,-0.74082893E+02,-0.21089073E+02,-0.52561008E+02,
              0.36209114E+02, },
            { 0.58211665E+01,-0.10134482E+02, 0.36446863E+03,-0.14913530E+02,
              0.88607018E+02,-0.31856615E+02,-0.45381077E+02, 0.13251899E+02,
              0.32890629E+02, 0.99184685E+01, 0.62132945E+01, 0.10968967E+02,
              0.11417336E+01, },
            { 0.98582275E+02,-0.37036868E+03, 0.28047604E+03,-0.16251318E+03,
              0.11509062E+03,-0.60012803E+01, 0.18846554E+02,-0.23316351E+02,
              0.17238712E+02,-0.60839558E+00, 0.14205902E+02,-0.18185123E+02,
              0.25047588E+02, },
            {-0.21569096E+03, 0.48272430E+03,-0.61942755E+03,-0.80701262E-01,
              0.22225163E+03,-0.23381893E+02, 0.46022134E+01,-0.43023670E+02,
             -0.49467533E+02, 0.12000541E+02,-0.11647184E+02,-0.33798725E+02,
             -0.12033131E+02, },
            { 0.35190132E+02, 0.30408795E+03,-0.15077676E+03,-0.33036247E+02,
              0.31228766E+03, 0.41757622E+02,-0.11409177E+03, 0.61402256E+02,
              0.21682520E+02, 0.81727142E+02, 0.24596991E+02, 0.52422810E+02,
             -0.42727203E+02, },
            {-0.19147324E+02,-0.20564623E+02,-0.17347337E+03, 0.61360617E+01,
             -0.46458740E+02, 0.11141960E+02, 0.22063715E+02,-0.60768661E+01,
             -0.15222783E+02,-0.52150383E+01,-0.25772462E+01,-0.48167362E+01,
             -0.76443511E+00, },
            {-0.44532223E+02, 0.18376117E+03,-0.17429190E+03, 0.82279617E+02,
             -0.48177139E+02, 0.71741357E+01,-0.10823287E+02, 0.10417562E+02,
             -0.95174761E+01,-0.11475333E+00,-0.57856946E+01, 0.82347708E+01,
             -0.13050280E+02, },
            { 0.10409344E+03,-0.20178984E+03, 0.25817352E+03,-0.14944650E+00,
             -0.85363350E+02, 0.14998612E+02,-0.99766254E+01, 0.17693336E+02,
              0.23184240E+02,-0.29896560E+01, 0.85577002E+01, 0.12933767E+02,
              0.56063805E+01, },
            {-0.18578611E+02,-0.14634050E+03, 0.76995132E+02, 0.21686547E+01,
             -0.12222349E+03,-0.20490656E+02, 0.53517189E+02,-0.29491062E+02,
             -0.64766569E+01,-0.32628654E+02,-0.10126882E+02,-0.19581978E+02,
              0.18641266E+02, },
            { 0.73967703E-01,-0.14112779E+00, 0.17739686E+00,-0.10380149E+01,
              0.46758762E+00,-0.52029893E-01, 0.38553629E-01,-0.19935209E-01,
              0.24950037E-01,-0.20095922E-01,-0.28559000E-02,-0.27403604E-01,
             -0.97574195E-03, },
            { 0.11055296E+00,-0.16600928E+00,-0.32541618E-01,-0.43835488E+00,
             -0.84406424E+00,-0.22534750E-01, 0.29131178E-01,-0.11667867E-01,
             -0.34039460E-01,-0.32340210E-01, 0.11184864E-01,-0.23590121E-01,
              0.87832250E-02, },
            { 0.91149993E-01, 0.97471708E+00, 0.51438361E+00, 0.13833370E+01,
             -0.50698072E+00, 0.27014571E+00, 0.11282755E+00, 0.54294676E-01,
             -0.75866021E-02, 0.12068818E+00, 0.16455030E-01, 0.90567581E-01,
              0.75802878E-01, },
            { 0.34389898E+00, 0.55841506E+00, 0.90160257E+00, 0.12392099E+01,
              0.17093930E+01, 0.13808005E+00, 0.10166630E+00, 0.20844191E+00,
             -0.15551122E+00, 0.10595880E-01,-0.34317158E-01, 0.91806263E-01,
              0.14963972E-01, },
            { 0.16836854E+01, 0.24811437E+01,-0.16475384E+01, 0.17249215E+01,
              0.30530896E+01, 0.75067306E+00,-0.15300773E+01,-0.39954576E+00,
              0.12609008E+00,-0.23109366E+00,-0.14283510E+00, 0.11207330E+00,
             -0.23422834E+00, },
            { 0.22920127E+01,-0.15677989E+00, 0.10574989E+01,-0.47031479E+01,
              0.83145690E+00, 0.60532081E+00,-0.75335938E+00, 0.72133221E-01,
              0.13619457E+01, 0.90937072E+00,-0.11166501E+00, 0.18261883E+00,
             -0.90696365E-01, },
            {-0.31183465E+01,-0.60279012E+01,-0.40491314E+01, 0.21722064E+01,
             -0.14826589E+01,-0.20925791E+01,-0.12620353E+01,-0.50687373E+00,
             -0.72735786E+00,-0.18294009E+01, 0.10776019E+01,-0.65747297E+00,
             -0.52606915E-02, },
            {-0.83851500E+01,-0.95547777E+00,-0.60184135E+01,-0.85619497E+01,
              0.61534148E+00, 0.10074309E+01,-0.85972464E+00,-0.75723243E+00,
              0.14236768E+01,-0.71778768E+00,-0.21167088E-01,-0.60213548E+00,
              0.36463749E+00, },
            {-0.12601962E+02,-0.60230761E+01,-0.29406281E+01, 0.11637723E+02,
             -0.90889711E+01,-0.18631748E+01, 0.10832180E+02, 0.28277893E+01,
              0.15355181E+01, 0.26887271E+01,-0.92505731E-01, 0.11178880E+00,
              0.12021805E+01, },
            {-0.11314941E+02, 0.27979467E+01,-0.14773956E+02, 0.20741819E+02,
              0.14875138E+02,-0.32533488E+01, 0.46425400E+01,-0.30427511E+01,
             -0.68344517E+01,-0.52124023E+01,-0.58438945E+00,-0.31132221E+00,
             -0.11308281E+01, },
            { 0.78239317E+01, 0.12827962E+02, 0.60509434E+01,-0.11737222E+02,
              0.10860994E+02, 0.64112983E+01, 0.23935566E+01, 0.18167053E+01,
              0.33579202E+01, 0.52217593E+01,-0.41448812E+01, 0.17888294E+01,
             -0.79165143E+00, },
            { 0.24079092E+02,-0.46660581E+01, 0.13022430E+02, 0.13900978E+02,
             -0.91128855E+01,-0.40291896E+01, 0.34799480E+01, 0.18368626E+01,
             -0.34240825E+01, 0.28355694E+01, 0.16054691E+00, 0.70973271E+00,
             -0.11734304E+01, },
            { 0.24227192E+02, 0.57047434E+01, 0.18991457E+02,-0.39872898E+02,
              0.15416595E+01,-0.74747860E+00,-0.19494642E+02,-0.59575405E+01,
             -0.70911956E+01,-0.68836184E+01, 0.16940249E+01,-0.13717374E+01,
             -0.21887500E+01, },
            { 0.13924574E+02,-0.16587017E+01, 0.38369900E+02,-0.27500231E+02,
             -0.43234756E+02, 0.34732475E+01,-0.10127350E+02, 0.84686279E+01,
              0.12072734E+02, 0.89818316E+01, 0.23602653E+01,-0.55452925E+00,
              0.37565517E+01, },
            {-0.70313997E+01,-0.75539041E+01,-0.29214725E+01, 0.47263317E+01,
             -0.15518660E+02,-0.53729682E+01,-0.81937367E+00,-0.13753450E+01,
             -0.31669335E+01,-0.39806004E+01, 0.32713692E+01,-0.14239472E+01,
              0.86429405E+00, },
            {-0.19807907E+02, 0.82111835E+01,-0.75345130E+01,-0.12458882E+01,
              0.27180896E+01, 0.28920496E+01,-0.37135229E+01,-0.16157656E+01,
              0.26609850E+01,-0.22828846E+01,-0.28764549E+00, 0.47844201E-01,
              0.77955002E+00, },
            {-0.14075713E+02,-0.23986549E+01,-0.14268916E+02, 0.32392254E+02,
              0.10021720E+02, 0.22754657E+01, 0.98033123E+01, 0.34065833E+01,
              0.62487297E+01, 0.50818644E+01,-0.16391834E+01, 0.16020304E+01,
              0.12680324E+01, },
            {-0.18211569E+01,-0.56535420E+01,-0.28800457E+02, 0.73658066E+01,
              0.32735832E+02,-0.24511378E+00, 0.70626469E+01,-0.58881278E+01,
             -0.75028081E+01,-0.46169567E+01,-0.17588331E+01, 0.63301569E+00,
             -0.28427830E+01, },
            { 0.76349199E-01,-0.75198315E-01,-0.50871223E-02,-0.72312318E-01,
              0.48052013E-01,-0.58207494E+00,-0.75327557E+00,-0.42022016E-01,
              0.18120332E-01,-0.16880617E-01, 0.73444042E-02,-0.28194983E-01,
             -0.15638337E-01, },
            { 0.39461084E-01,-0.13488993E+00,-0.10055972E+00, 0.46493847E-01,
             -0.93732879E-01, 0.74668914E+00,-0.58071917E+00,-0.26067192E-01,
              0.20636749E-01, 0.86305896E-02, 0.54170616E-03,-0.23148715E-01,
             -0.52713491E-02, },
            {-0.97694404E-01, 0.93344510E+00, 0.29393701E-01, 0.73514744E-02,
             -0.34737319E+00, 0.26163518E+00, 0.36970001E+00, 0.14785510E+00,
             -0.55417154E-01, 0.31292816E-02, 0.78705736E-01,-0.81819005E-01,
              0.33747096E-01, },
            {-0.56032654E-01,-0.46219949E-01,-0.64890347E-01, 0.13689564E+00,
             -0.56143868E+00,-0.29345471E+00, 0.32432055E+00, 0.15689895E-01,
              0.65602973E-01, 0.60567804E-01,-0.10864880E-01,-0.10147065E+00,
             -0.35914693E-01, },
            { 0.67012018E+00, 0.42257118E+00, 0.19467359E-01, 0.89575797E-02,
             -0.10516388E+01, 0.78924227E+00, 0.37479475E+00,-0.40912420E+00,
              0.42230487E+00, 0.17606600E+00, 0.89392833E-01, 0.30513483E+00,
             -0.95638916E-01, },
            {-0.71605802E+00, 0.26266012E+01, 0.52779651E+00,-0.18734942E+00,
              0.10323124E+01,-0.91403717E+00, 0.53446960E+00,-0.13627081E+00,
             -0.66635752E+00,-0.12717301E+00, 0.21606894E+00, 0.22295696E+00,
              0.47271583E-01, },
            {-0.82142699E+00,-0.15194725E+01, 0.36633497E+00, 0.13400576E+00,
              0.29195154E+00, 0.53095353E+00, 0.40886421E-01,-0.17716572E+00,
              0.17803183E+00, 0.71642324E-01,-0.24857517E+00, 0.35629210E+00,
             -0.15162340E-01, },
            {-0.43191445E+00, 0.30332151E+00, 0.67259073E-01,-0.42061138E+00,
              0.12614172E+01,-0.33652726E+00, 0.57137632E+00,-0.20697652E+00,
             -0.75853616E-01,-0.11044272E+00, 0.11320280E+00, 0.41475806E+00,
              0.16160649E+00, },
            {-0.10161493E+00, 0.58983707E+00,-0.84979522E+00, 0.30489936E+00,
              0.25749736E+01,-0.91752881E+00,-0.53623151E-02, 0.80623126E+00,
             -0.10490124E+01,-0.38548842E+00, 0.44795712E-02,-0.58598870E+00,
              0.16041683E+00, },
            { 0.97775286E+00,-0.35922375E+01,-0.79198337E+00,-0.14946298E+00,
             -0.18443063E+01, 0.13141613E+01,-0.25741628E+00, 0.35331422E+00,
              0.10484736E+01, 0.61409295E-01,-0.38274255E+00,-0.52102482E+00,
             -0.16893646E+00, },
            { 0.26682647E-01,-0.44999905E-01,-0.30429330E-01,-0.24290942E-01,
              0.55203110E-01,-0.96770590E-02,-0.37339274E-01, 0.31498066E+00,
             -0.17359960E+00,-0.28275497E-01, 0.76047601E-02,-0.17485568E-01,
             -0.21701658E-01, },
            { 0.19067446E-01, 0.82480900E-01,-0.24417976E-01, 0.60341705E-01,
             -0.76737002E-01, 0.30098597E-01,-0.14800011E-01, 0.16694994E+00,
              0.29754844E+00, 0.60499981E-02,-0.65341271E-02,-0.12154019E-02,
             -0.32120442E-02, },
            {-0.31834999E+00,-0.10176746E+00,-0.12919512E+00, 0.57349347E-01,
             -0.38777705E-01,-0.16754001E+00, 0.71787603E-01, 0.55727322E-01,
              0.48854449E+00, 0.43341171E-01,-0.26241627E-01, 0.33198107E-01,
              0.25002802E-01, },
            { 0.26534075E+00, 0.13981649E+00, 0.81587816E-03,-0.29231967E-01,
              0.14419927E+00,-0.16403934E+00,-0.73877536E-01,-0.48972750E+00,
             -0.57703011E-01,-0.12619291E-01,-0.23607044E-01,-0.51552523E-01,
             -0.31952363E-01, },
            {-0.93206882E-01,-0.45213788E-02, 0.22315690E-02, 0.23447252E-02,
             -0.55114511E-01, 0.20760903E-01, 0.28086805E-02, 0.39981101E-01,
             -0.25381871E-01, 0.16350323E+00, 0.20666330E+00, 0.96418187E-02,
             -0.18114807E-01, },
            { 0.34151793E-01, 0.49445402E-01,-0.71330421E-01,-0.45091002E-02,
             -0.17531471E-01,-0.32351278E-01, 0.23442954E-02, 0.62810970E-02,
              0.84683262E-01,-0.18843912E+00, 0.15088977E+00, 0.14799476E-01,
             -0.90895388E-02, },
            { 0.39233349E-01, 0.67935758E-02, 0.17726658E-01, 0.13225300E-01,
              0.48658162E-01, 0.31417664E-01,-0.39813090E-01, 0.35820976E-01,
              0.27609885E-01, 0.68485285E-02,-0.42618301E-01,-0.11170206E+00,
              0.59159048E-01, },
            { 0.40888876E-01, 0.11376599E+00, 0.16162302E-01,-0.29816031E-01,
              0.62418427E-01,-0.13326484E-01,-0.32052249E-01,-0.99713812E-02,
              0.28484600E-01,-0.10098713E-02,-0.12852327E-01,-0.85706830E-01,
             -0.11604415E+00, },
            {-0.68437657E-02, 0.27410646E-02,-0.60129546E-01, 0.18074039E-01,
              0.18273348E-01, 0.19582406E-01,-0.92187580E-02, 0.13423235E-01,
              0.26095778E-01, 0.18187763E-01,-0.19140057E-01,-0.84481649E-02,
              0.25512718E-02, },
            { 0.86681031E-01,-0.42338647E-01,-0.20124212E-01,-0.68186219E-02,
             -0.59652332E-01, 0.32581393E-01,-0.32793738E-01,-0.16700966E-01,
              0.26844621E-01,-0.54568769E-02, 0.35884832E-02,-0.91454759E-02,
              0.69946642E-02, },
            { 0.57878252E-01, 0.23915995E-01, 0.12776043E-02, 0.33935215E-01,
             -0.24363872E-01,-0.43215800E-01,-0.14799070E-01,-0.12832981E-01,
              0.26710723E-01, 0.12159403E-02,-0.57877437E-02,-0.20406270E-01,
             -0.79121506E-02, },
            {-0.46333019E-01, 0.68355016E-01,-0.24444332E-01,-0.56031704E-01,
              0.86708978E-01, 0.14629347E-01, 0.63353748E-03, 0.32699492E-03,
              0.16688421E-01, 0.15034921E-01,-0.95454603E-03,-0.14304600E-01,
              0.29492131E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir15Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30829437E+01,-0.95623417E-03, 0.14071414E-01, 0.67336978E-02,
              0.16362238E-02, 0.31438053E-02, 0.11660823E-01,-0.21267915E-02,
             -0.41597779E-02, },
            {-0.42748913E+00, 0.24467647E-01, 0.60202345E-01, 0.19219477E-01,
             -0.36522668E-01,-0.30637039E-02,-0.21544660E-01,-0.25534466E-01,
              0.19556645E-01, },
            { 0.83016407E+00, 0.10246895E+00,-0.80737822E-01,-0.11050681E+00,
             -0.88695347E-01,-0.43044806E-01,-0.54705851E-01, 0.12666488E+00,
              0.29847229E-01, },
            { 0.79621750E+00,-0.35017520E+00,-0.20669055E+00,-0.10942011E+00,
              0.13936011E+00, 0.12803758E-02, 0.82603864E-01, 0.60857706E-01,
             -0.45860834E-01, },
            {-0.10769858E+01,-0.21080144E+00, 0.11818612E+00, 0.17931351E+00,
              0.22275142E+00, 0.15942883E+00, 0.81478491E-01,-0.35534877E+00,
             -0.45454629E-01, },
            {-0.48894572E+00, 0.31695545E+00, 0.18177135E+00, 0.96843764E-01,
             -0.10721572E+00, 0.11368042E-02,-0.67227729E-01,-0.28057726E-01,
              0.24773201E-01, },
            { 0.21707885E+00, 0.13230781E+00, 0.58709583E-02,-0.66172384E-01,
             -0.14059685E+00,-0.13175756E+00,-0.37203427E-01, 0.24195625E+00,
              0.17272590E-01, },
            { 0.22961402E-01,-0.11863714E+00,-0.35007322E+00, 0.12821091E-01,
             -0.12189340E-01,-0.68265543E-03, 0.62941569E-02,-0.44815377E-02,
             -0.26578419E-02, },
            { 0.14728447E-01, 0.37473145E+00,-0.96780010E-01,-0.27384372E-01,
              0.10581867E-01, 0.22666370E-02,-0.17687570E-01, 0.68567963E-02,
              0.72228303E-02, },
            {-0.12491125E+00,-0.71489297E-01, 0.97123340E-01, 0.34888726E-01,
              0.35563158E-02,-0.51503751E-01,-0.51774520E-01,-0.26012940E-01,
              0.61413504E-01, },
            { 0.35958029E-01, 0.11307500E-01,-0.75540908E-01, 0.68435147E-02,
              0.20799374E-01,-0.41120008E-01, 0.33827305E-01, 0.36062390E-01,
              0.14471351E-01, },
            {-0.33956641E+00, 0.54326749E+00, 0.19545006E+01, 0.74205472E-03,
             -0.80778316E-01,-0.77715360E-01,-0.50010346E-01, 0.22858028E+00,
              0.87244682E-01, },
            {-0.63539225E+00,-0.19151980E+01, 0.43575630E+00, 0.33090886E+00,
             -0.24209768E-02,-0.12964347E+00, 0.15143993E+00,-0.32082044E-01,
              0.21434763E-01, },
            { 0.14364071E+01, 0.52485019E-01,-0.16226660E+01,-0.25793177E+00,
             -0.36766967E+00, 0.67354387E+00, 0.12053780E+00, 0.38931045E+00,
             -0.57873863E+00, },
            { 0.38684511E+00, 0.14660304E+01, 0.38292372E+00, 0.73202103E-02,
              0.29484910E+00, 0.21471724E+00,-0.17507131E+00,-0.18583956E+00,
             -0.20088874E+00, },
            { 0.86518139E+00,-0.70394367E+00,-0.22149346E+01,-0.19477078E+00,
              0.22440962E+00, 0.24766719E+00, 0.11958317E+00,-0.63776320E+00,
             -0.41532379E+00, },
            { 0.21289845E+01, 0.17928966E+01,-0.76989651E+00,-0.93120408E+00,
             -0.15367731E+00, 0.42717677E+00,-0.41677803E+00, 0.60080402E-01,
             -0.10584719E+00, },
            {-0.35660176E+01,-0.61952978E+00, 0.36163361E+01, 0.49679780E+00,
              0.11595428E+01,-0.17288561E+01, 0.13930221E+00,-0.92723536E+00,
              0.12680798E+01, },
            {-0.96336013E+00,-0.35305631E+01,-0.10435582E+01,-0.91957055E-01,
             -0.86425048E+00,-0.35838720E+00, 0.15090416E+00, 0.41967389E+00,
              0.41672745E+00, },
            {-0.42783877E+00, 0.46139610E+00, 0.61414963E+00, 0.16084965E+00,
             -0.10869228E+00,-0.21873859E+00,-0.10980464E+00, 0.42953312E+00,
              0.41603583E+00, },
            {-0.16682128E+01,-0.58972288E-01, 0.53959042E+00, 0.71032184E+00,
              0.12974617E+00,-0.31926504E+00, 0.34129691E+00,-0.30647589E-01,
              0.58991492E-01, },
            { 0.24329290E+01, 0.10756953E+01,-0.27146754E+01,-0.27721620E+00,
             -0.90134066E+00, 0.11445333E+01,-0.27922100E+00, 0.58598697E+00,
             -0.78309441E+00, },
            { 0.52151656E+00, 0.27146018E+01, 0.98834449E+00, 0.13908830E+00,
              0.59049463E+00, 0.22968982E+00,-0.30222910E-01,-0.30343634E+00,
             -0.20505387E+00, },
            {-0.91355713E-02, 0.29706708E-02,-0.31272422E-02, 0.22377572E-02,
             -0.16692504E+00,-0.71510244E-02,-0.62363604E-02, 0.17446436E-02,
             -0.74244705E-02, },
            {-0.95079606E-03,-0.17749453E-01, 0.95796930E-02, 0.16611043E+00,
              0.34745239E-01,-0.31304036E-02,-0.20624200E-01,-0.34306971E-02,
              0.38484319E-02, },
            { 0.50973926E-01, 0.18939672E+00, 0.45086402E-01, 0.27533084E+00,
             -0.25414023E+00, 0.19070908E-01,-0.36821201E-01, 0.30477703E-01,
              0.12377640E-01, },
            {-0.62895715E-01, 0.65114923E-01, 0.16306715E-01, 0.36566287E+00,
              0.41203955E+00, 0.24824467E-01,-0.14574272E-01,-0.27033810E-02,
             -0.31155946E-01, },
            { 0.35454988E-01, 0.60413000E-02,-0.18199088E-01, 0.76078407E-01,
             -0.16603535E+00, 0.15787179E-01, 0.39357701E-02,-0.27306402E-01,
              0.80539152E-01, },
            {-0.28398255E+00, 0.76125801E-01,-0.20132874E+00, 0.19440579E+00,
             -0.70742458E-01, 0.12295902E-01, 0.14769582E+00, 0.96357893E-02,
              0.31613465E-01, },
            {-0.23468396E+00,-0.99008822E+00, 0.71097992E-01,-0.12017087E+01,
              0.50148809E+00,-0.10378635E+00, 0.77810965E-01,-0.91363624E-01,
             -0.11122116E+00, },
            { 0.67196661E+00,-0.91745377E-01, 0.13793392E+00,-0.85245258E+00,
             -0.14946311E+01,-0.28373862E-01,-0.16134294E-01, 0.76116845E-01,
              0.61346978E-01, },
            {-0.15103582E+00,-0.18279678E+00,-0.22154266E-01,-0.34824383E+00,
              0.38926736E+00, 0.52207313E-01, 0.68610243E-01, 0.53184792E-01,
             -0.12627009E+00, },
            { 0.39844444E+00,-0.27108863E+00, 0.32306048E+00,-0.42485851E+00,
             -0.19756165E+00,-0.24037840E-01,-0.15082456E+00, 0.23419553E-03,
             -0.81081510E-01, },
            { 0.24415769E+00, 0.13665473E+01,-0.11141641E+00, 0.11919484E+01,
             -0.32598576E+00, 0.14885177E+00,-0.22232583E-01, 0.77452868E-01,
              0.11268586E+00, },
            {-0.99769139E+00, 0.24313284E-01,-0.25469135E-01, 0.58631575E+00,
              0.13545561E+01, 0.77170582E-03, 0.12235342E+00,-0.77348314E-01,
              0.12083619E-01, },
            {-0.14412698E-01, 0.11934513E-01, 0.10500170E-01,-0.19674214E-01,
             -0.95874164E-02, 0.18518902E-01, 0.13004042E-01, 0.67524160E-02,
              0.29727495E-02, },
            {-0.11646161E-01,-0.37059538E-01,-0.13889452E-01, 0.17757954E-01,
             -0.19775612E-01,-0.20395935E-01, 0.16103785E-01,-0.75430768E-02,
              0.19750383E-02, },
            { 0.60488075E-01, 0.41547071E-01, 0.15083943E-01,-0.27579306E-01,
              0.15362628E-01, 0.45602046E-01, 0.11825938E-01, 0.46184431E-02,
              0.33622717E-02, },
            { 0.20156203E-01, 0.11467540E-01,-0.14934420E-01,-0.22406610E-01,
              0.51867850E-02,-0.21814423E-01, 0.54861706E-01,-0.98177865E-02,
              0.35655340E-02, },
            {-0.18438589E-01,-0.11210715E+00,-0.77830158E-01, 0.20839000E-01,
              0.28409932E-02,-0.75307317E-01,-0.12357456E+00,-0.36384016E-01,
              0.30860552E-02, },
            { 0.94958721E-02, 0.14035341E+00,-0.26280927E-01,-0.39201953E-01,
              0.32175556E-01, 0.13319464E+00,-0.63215457E-01, 0.32919317E-02,
              0.36654861E-02, },
            {-0.24621338E-01,-0.67079356E-02, 0.15665390E-01,-0.16562494E-02,
             -0.48217624E-02, 0.42397496E-02,-0.47429493E-02, 0.29692261E-01,
              0.93461300E-03, },
            {-0.17562089E-01, 0.82685566E-02,-0.17116127E-02, 0.37179261E-02,
             -0.40929308E-02, 0.27845460E-02, 0.11334968E-01,-0.89295959E-03,
              0.25569776E-01, },
            { 0.54277506E-01,-0.23874678E-01, 0.16083041E-01,-0.17546413E-01,
             -0.15783805E-01, 0.44590095E-02, 0.11278612E-01, 0.31083820E-01,
              0.20818269E-01, },
            { 0.31938594E-01,-0.67811191E-01, 0.65194801E-02, 0.25008731E-01,
              0.63123470E-02, 0.44389213E-02,-0.66350435E-03,-0.64168908E-01,
              0.25633674E-01, },
            { 0.53379317E-02, 0.13877489E-03, 0.42313864E-02,-0.18789083E-02,
             -0.55933762E-02, 0.72038206E-02,-0.30773245E-02,-0.21693164E-02,
             -0.23185892E-02, },
            { 0.16698711E-01, 0.15778277E-01,-0.79104668E-02, 0.66635874E-02,
             -0.56811650E-02, 0.11245554E-02, 0.19182810E-02, 0.65909019E-02,
             -0.23176591E-02, },
            { 0.10475382E-01, 0.17943383E-02, 0.12603857E-01, 0.36077972E-02,
             -0.10011117E-01, 0.22700520E-02, 0.29119304E-02, 0.28559086E-02,
             -0.42581139E-02, },
            { 0.80381511E-02, 0.52013630E-02,-0.15140937E-01, 0.62007951E-02,
              0.13680374E-01,-0.36166117E-02,-0.32075443E-02,-0.44594007E-02,
              0.68366108E-03, },
         },
         {
            { 0.27214456E+01,-0.59877997E-02, 0.48487321E-02, 0.11528316E-01,
              0.83281239E-03, 0.11694297E-02, 0.40655504E-02,-0.12881265E-02,
             -0.53722416E-02, },
            {-0.32532126E+00, 0.63253954E-01, 0.72628997E-01, 0.17123209E-01,
             -0.41282631E-01,-0.95136929E-03,-0.11847535E-01,-0.15334175E-01,
              0.25392929E-02, },
            { 0.11310599E+01, 0.56544151E-01,-0.42942472E-01,-0.46731703E-01,
             -0.11823614E-01,-0.18031187E-01,-0.14277424E-01, 0.86189993E-01,
              0.44520613E-01, },
            { 0.39499840E+00,-0.47727019E+00,-0.22805473E+00,-0.68672247E-01,
              0.14230086E+00,-0.33062331E-01, 0.33439875E-01, 0.30319517E-01,
             -0.60412716E-02, },
            {-0.15841742E+01, 0.48886333E-01, 0.48712164E-01, 0.24080880E-01,
             -0.28131995E-01, 0.11339512E+00, 0.49910150E-01,-0.21996433E+00,
             -0.95440678E-01, },
            {-0.17072472E+00, 0.39422408E+00, 0.19510727E+00, 0.54062922E-01,
             -0.10195932E+00, 0.36064230E-01,-0.24015130E-01,-0.89633605E-02,
              0.65473379E-02, },
            { 0.46993488E+00,-0.63174419E-01, 0.38383942E-01, 0.14852938E-01,
              0.32509878E-01,-0.10623866E+00,-0.43911122E-01, 0.13939916E+00,
              0.55493016E-01, },
            { 0.23913419E-01,-0.28132305E+00,-0.27718166E+00, 0.13927994E-01,
             -0.82893772E-02,-0.20239395E-02, 0.55825445E-02,-0.44232481E-02,
             -0.53990549E-04, },
            {-0.56995903E-02, 0.29952064E+00,-0.25091380E+00,-0.14151807E-01,
              0.12410180E-01, 0.29630894E-02,-0.19999536E-01, 0.42727250E-02,
              0.10325929E-01, },
            { 0.28787147E-01, 0.21169320E-01,-0.18652616E-01,-0.22639667E-02,
              0.15749900E-01,-0.35053190E-01,-0.11501440E-01,-0.28702248E-01,
              0.50729234E-01, },
            { 0.13469434E-01, 0.11982938E+00,-0.71723476E-01, 0.38758967E-01,
              0.65331496E-02,-0.35491824E-01, 0.78479424E-02, 0.58513922E-02,
              0.13235763E-01, },
            {-0.45237130E+00, 0.98882270E+00, 0.10768780E+01,-0.19403292E-01,
              0.32672111E-01, 0.33263270E-01,-0.44367231E-01, 0.15512651E+00,
             -0.33647958E-02, },
            {-0.50860500E+00,-0.11076277E+01, 0.85086280E+00, 0.18959832E+00,
             -0.88933870E-01,-0.87038308E-01, 0.14959402E+00,-0.85117795E-01,
             -0.10071342E+00, },
            { 0.16840751E+00,-0.55341917E+00,-0.18681356E+00, 0.46795160E-01,
             -0.36465406E+00, 0.34773061E+00,-0.56421123E-01, 0.41546983E+00,
             -0.37943959E+00, },
            { 0.48950046E+00, 0.17589639E+00, 0.51375562E+00,-0.19277123E+00,
              0.27112055E+00, 0.27518129E+00, 0.19615915E-03, 0.11075651E+00,
             -0.41206829E-01, },
            { 0.12937419E+01,-0.11020116E+01,-0.64798677E+00,-0.75120479E-01,
             -0.96557952E-01,-0.93400031E-01, 0.93651280E-01,-0.38147634E+00,
              0.10117626E-01, },
            { 0.18461685E+01, 0.29588601E+00,-0.11936399E+01,-0.57821631E+00,
              0.12842540E+00, 0.21288218E+00,-0.31878108E+00, 0.19306193E+00,
              0.21284461E+00, },
            {-0.75016105E+00, 0.11026198E+01, 0.58122629E+00,-0.15455300E+00,
              0.91149658E+00,-0.80885851E+00, 0.42632446E+00,-0.10352906E+01,
              0.75210643E+00, },
            {-0.14370759E+01,-0.69056630E+00,-0.11955296E+01, 0.39424512E+00,
             -0.76573443E+00,-0.49984264E+00,-0.18155196E+00,-0.34026536E+00,
             -0.81930935E-01, },
            {-0.85781968E+00, 0.72457278E+00,-0.22283772E-02, 0.71129121E-01,
              0.41598022E-01, 0.46153598E-01,-0.39733808E-01, 0.24633887E+00,
             -0.84825670E-02, },
            {-0.14810368E+01, 0.54091024E+00, 0.79484820E+00, 0.49584129E+00,
             -0.65858573E-01,-0.14391363E+00, 0.22043347E+00,-0.11268485E+00,
             -0.12970915E+00, },
            { 0.58107245E+00,-0.53847790E+00,-0.96545213E+00, 0.15799177E+00,
             -0.56166148E+00, 0.47301409E+00,-0.46004382E+00, 0.66261995E+00,
             -0.41957995E+00, },
            { 0.10437909E+01, 0.10547332E+01, 0.84057873E+00,-0.25376251E+00,
              0.53126270E+00, 0.30007434E+00, 0.16683571E+00, 0.24391644E+00,
              0.14497797E+00, },
            {-0.10355590E-01,-0.59964433E-02, 0.24918828E-02, 0.40248983E-01,
             -0.11974522E+00,-0.68878066E-02, 0.52973459E-03, 0.16610487E-02,
             -0.47631026E-02, },
            {-0.80973590E-02,-0.23301819E-01, 0.16890654E-01, 0.10705962E+00,
              0.74931182E-01, 0.11066468E-02,-0.15185948E-01,-0.26813566E-02,
              0.41845441E-02, },
            { 0.87334573E-01, 0.17434020E+00, 0.29009607E-01, 0.16647445E+00,
             -0.85534535E-01,-0.42410148E-02,-0.62163905E-02, 0.25638817E-01,
             -0.72888704E-02, },
            { 0.21463536E-01, 0.82178161E-01,-0.59707514E-02, 0.14109133E+00,
              0.29117456E+00,-0.18742292E-02,-0.28855884E-04, 0.78286685E-03,
             -0.20326763E-01, },
            {-0.23505557E-01, 0.50627660E-01,-0.42973336E-01,-0.24614643E+00,
              0.68570442E-01, 0.36111508E-01,-0.65725469E-02,-0.73447876E-01,
              0.73012590E-01, },
            {-0.22136711E+00, 0.12092493E+00,-0.19790530E+00,-0.13838495E-01,
             -0.37348422E+00,-0.22489665E-01, 0.12434304E+00,-0.84679516E-04,
              0.23986567E-02, },
            {-0.23487779E+00,-0.88596624E+00, 0.60993783E-01,-0.40152115E+00,
              0.38432073E-01, 0.25088888E-01,-0.47009036E-01,-0.10053905E+00,
              0.56201056E-01, },
            { 0.26335394E+00,-0.29947615E+00, 0.83331093E-01,-0.21753137E+00,
             -0.68338555E+00, 0.12236043E+00,-0.42072836E-01, 0.58362342E-01,
              0.13270973E-01, },
            {-0.91692032E-02,-0.24499676E+00, 0.31347323E-01, 0.55245612E-01,
             -0.95426172E-01,-0.17419253E-01, 0.71632415E-01, 0.13672970E+00,
             -0.10337002E+00, },
            { 0.32925227E+00,-0.42062321E+00, 0.24600771E+00, 0.48706904E-01,
              0.17111088E+00,-0.23537112E-01,-0.12686042E+00, 0.21002434E-01,
             -0.16690811E-01, },
            { 0.69676243E-01, 0.11886312E+01,-0.81857055E-01, 0.28926817E+00,
             -0.38532227E-01, 0.34839232E-01, 0.56403339E-01, 0.85259937E-01,
             -0.75022250E-01, },
            {-0.51328564E+00, 0.48400918E+00, 0.22977814E-01, 0.13689673E+00,
              0.42785177E+00,-0.13421567E+00, 0.10484700E+00,-0.73551662E-01,
              0.30889113E-01, },
            {-0.18197976E-01, 0.72729443E-02, 0.12053806E-01,-0.14377148E-01,
             -0.17075215E-01, 0.18109780E-01,-0.17175833E-01,-0.27712116E-02,
             -0.34635665E-03, },
            {-0.13854426E-01,-0.35453096E-01,-0.12559115E-01, 0.11651063E-01,
             -0.17490545E-01, 0.12271131E-01, 0.17724304E-01,-0.37607695E-02,
             -0.41570398E-02, },
            { 0.43509074E-01, 0.63937917E-01, 0.10682798E-01,-0.17563932E-01,
              0.50647352E-02, 0.68679936E-01, 0.10368119E-02, 0.18375491E-03,
              0.41448488E-02, },
            { 0.25249584E-01, 0.15560197E-01, 0.24144184E-02,-0.15688732E-01,
              0.11401277E-02,-0.10834954E-01, 0.62144563E-01,-0.63146567E-02,
              0.39895601E-02, },
            { 0.55234623E-02,-0.11332307E+00,-0.56373134E-01, 0.11459470E-01,
              0.22336069E-01,-0.10773219E+00,-0.61456602E-01,-0.12642963E-01,
              0.19825991E-01, },
            { 0.55266679E-02, 0.10650305E+00,-0.20077813E-01,-0.12482653E-01,
              0.13799644E-01, 0.59804551E-01,-0.83235875E-01,-0.92901830E-02,
              0.25422184E-03, },
            {-0.16490787E-01,-0.12288227E-01, 0.15624606E-01,-0.15590733E-02,
             -0.44173814E-03, 0.32792636E-02,-0.54032323E-02, 0.23649098E-01,
              0.65248716E-02, },
            {-0.10436538E-01, 0.51402622E-02,-0.10056754E-01, 0.46160161E-02,
             -0.17697068E-02, 0.29858900E-02, 0.63185222E-02,-0.59197075E-02,
              0.20331379E-01, },
            { 0.39676785E-01,-0.77944589E-02, 0.29491605E-02,-0.97978339E-02,
              0.13357931E-02, 0.31436787E-02, 0.15301985E-01, 0.28650684E-01,
              0.43868300E-01, },
            { 0.33646043E-01,-0.66603132E-01, 0.23036173E-01, 0.62092445E-02,
             -0.26638312E-02,-0.65160957E-02,-0.45676758E-02,-0.67288466E-01,
              0.19755878E-01, },
            { 0.65249880E-02,-0.74825557E-02, 0.43182308E-02, 0.69805081E-02,
             -0.67913304E-02, 0.64756777E-02,-0.16087731E-02, 0.93622401E-03,
             -0.20369282E-02, },
            { 0.11612056E-01, 0.10057593E-01,-0.81270486E-02, 0.65685674E-02,
              0.31718882E-02, 0.40957397E-02, 0.22977332E-02, 0.49984353E-02,
              0.35048136E-02, },
            { 0.72671412E-02, 0.54456848E-02, 0.84348572E-02, 0.48800143E-02,
             -0.11012992E-01,-0.88449667E-03, 0.15507995E-02, 0.17194895E-02,
             -0.80987962E-03, },
            {-0.22724713E-04, 0.86002506E-03,-0.96425554E-02, 0.45933654E-02,
              0.12882300E-01, 0.65020460E-03,-0.29192658E-02,-0.44633718E-02,
              0.21139076E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir16F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.53135514E+01,-0.61482415E-01, 0.75109832E-01, 0.34710813E-01,
              0.22684038E-01,-0.57048872E-02,-0.84173474E-02, 0.44471744E-01,
             -0.16074698E-01, 0.85667998E-03, 0.22791712E-03, 0.92006745E-02,
             -0.97777117E-02, },
            {-0.19316612E+01,-0.42007375E+00, 0.51300526E+00,-0.14258102E+00,
              0.31701854E+00,-0.88729514E-02, 0.24743211E+00,-0.80617800E-01,
             -0.12699057E+00,-0.17045531E-01, 0.84008286E-02,-0.75328685E-01,
              0.18828200E-01, },
            { 0.46763873E+01, 0.11320518E+01,-0.18580437E+01,-0.27329683E+00,
             -0.90213805E+00,-0.61084539E+00, 0.23694906E+00,-0.46116051E+00,
              0.34168661E+00, 0.33637661E+00,-0.19169091E+00,-0.11774270E+00,
              0.24918503E+00, },
            { 0.28097292E+02, 0.26754222E+01,-0.14834358E+01, 0.31701112E+01,
             -0.53412337E+01, 0.76283532E+00,-0.51419859E+01, 0.73323458E+00,
              0.33136436E+00, 0.64914894E+00, 0.56351509E-01, 0.17945926E+01,
              0.51091266E+00, },
            {-0.29699144E+02,-0.10468776E+02, 0.11762651E+02, 0.23983593E+01,
             -0.62706685E+00, 0.43435154E+01,-0.74329567E+00, 0.25396516E+01,
             -0.11957142E+01,-0.22866728E+01, 0.13279625E+01, 0.59059352E+00,
             -0.12482961E+01, },
            {-0.74450401E+02,-0.65370393E+01,-0.13106404E+02,-0.16474710E+02,
              0.30279955E+02,-0.35481794E+01, 0.26479380E+02,-0.35002706E+01,
              0.19646826E+01,-0.25204167E+01,-0.10204962E+01,-0.11526494E+02,
             -0.46005092E+01, },
            { 0.57885025E+02, 0.38123917E+02,-0.29894230E+02,-0.82557545E+01,
              0.11295097E+02,-0.11428127E+02,-0.59597737E+00,-0.69582043E+01,
              0.21485796E+01, 0.51051588E+01,-0.42251997E+01,-0.11586436E+01,
              0.25822914E+01, },
            { 0.64745682E+02, 0.56215553E+01, 0.52016819E+02, 0.30449734E+02,
             -0.71321335E+02, 0.47431431E+01,-0.54951233E+02, 0.77081962E+01,
             -0.85594635E+01, 0.24655552E+01, 0.34039283E+01, 0.28274258E+02,
              0.11945585E+02, },
            {-0.61265751E+02,-0.52918648E+02, 0.32995888E+02, 0.10468823E+02,
             -0.18554152E+02, 0.12917597E+02, 0.30349872E+01, 0.83194733E+01,
             -0.23343971E+01,-0.44490862E+01, 0.54735060E+01, 0.93166763E+00,
             -0.25764549E+01, },
            { 0.18634470E+01, 0.28486409E+01,-0.62127373E+02,-0.22524527E+02,
              0.73664352E+02,-0.11603888E+01, 0.50629311E+02,-0.74867887E+01,
              0.10972483E+02, 0.44247410E+00,-0.41348748E+01,-0.29279104E+02,
             -0.12263268E+02, },
            { 0.26340647E+02, 0.24130861E+02,-0.13063382E+02,-0.43806276E+01,
              0.87424278E+01,-0.52607641E+01,-0.19757253E+01,-0.35192804E+01,
              0.10667630E+01, 0.12665039E+01,-0.23884971E+01,-0.25039810E+00,
              0.10172306E+01, },
            {-0.17573338E+02,-0.41650991E+01, 0.24150318E+02, 0.54928117E+01,
             -0.27598324E+02,-0.83179241E+00,-0.17326010E+02, 0.26191671E+01,
             -0.46037669E+01,-0.10153755E+01, 0.16939994E+01, 0.10823385E+02,
              0.43800306E+01, },
            {-0.18334110E+00, 0.16058656E+01, 0.18451380E+01, 0.23984671E-01,
             -0.77645235E-01,-0.71523450E-02,-0.39501633E-02, 0.48090607E-01,
              0.10382491E-02,-0.10124075E-02,-0.98467534E-02, 0.16033929E-01,
             -0.86858859E-02, },
            { 0.11624959E+00,-0.18097712E+01, 0.16314628E+01,-0.21510409E-01,
              0.10980082E+00, 0.23303006E-01,-0.40793312E-02,-0.20446820E-01,
              0.77379011E-01,-0.18653555E-01, 0.93311332E-02,-0.12022865E-02,
              0.20225255E-01, },
            {-0.59684879E+00,-0.33499169E+00,-0.13415632E+01, 0.50301838E+00,
             -0.44692445E+00, 0.90694763E-01,-0.62957235E-01,-0.15851028E+00,
              0.21268469E+00, 0.11720025E+00, 0.27619761E+00, 0.42360511E-01,
              0.29771827E-01, },
            {-0.34701595E+00, 0.15442055E+01,-0.20425926E+00, 0.16489644E+00,
              0.11285269E+00, 0.20746952E+00,-0.82578465E-01, 0.18210283E+00,
              0.43843952E+00, 0.80315150E-01, 0.14707121E+00, 0.18282767E+00,
              0.11922825E+00, },
            { 0.64125867E+01, 0.73759170E+01, 0.92153292E+01,-0.22129582E+00,
             -0.95621055E+00, 0.99270296E+00,-0.32091096E+00,-0.50337327E+00,
              0.87898248E+00,-0.46062678E+00,-0.19489811E+00,-0.40433988E+00,
              0.12082740E+00, },
            {-0.33119065E+00,-0.12071861E+02, 0.27483230E+01,-0.50224137E+00,
             -0.92419714E+00,-0.15161744E+01, 0.67563355E+00, 0.55171251E+00,
             -0.55627638E+00, 0.47423026E+00, 0.14856794E+00, 0.11773849E+00,
             -0.19019896E+00, },
            { 0.57699900E+01, 0.14562860E+01, 0.80994415E+01,-0.70281959E+01,
              0.21438148E+01, 0.44327801E+00,-0.11645912E+01, 0.24570403E+01,
             -0.35969975E+01,-0.75412291E+00,-0.54496884E+01,-0.50559856E-01,
              0.71804142E+00, },
            { 0.59823146E+01,-0.11598491E+02, 0.66550407E+01, 0.14041714E+00,
             -0.62558069E+01,-0.17235842E+01, 0.17961844E+01,-0.20923460E+01,
             -0.71783361E+01,-0.25725663E+01,-0.15988770E+01,-0.18210272E+01,
             -0.11821673E+01, },
            {-0.42838211E+02,-0.50422752E+02,-0.27591564E+02, 0.72362642E+01,
              0.67548871E+01,-0.13753628E+02, 0.42630401E+01, 0.29996941E+01,
             -0.36755970E+01, 0.23023808E+01, 0.17583880E+01, 0.28728533E+01,
              0.73049802E+00, },
            {-0.27694557E+01, 0.49150024E+02,-0.67315660E+01, 0.31819255E+01,
              0.76478081E+01, 0.15908955E+02,-0.70412917E+01,-0.81684484E+01,
              0.24692976E+01,-0.28794234E+01,-0.24862816E+01,-0.21675971E+00,
             -0.54332763E+00, },
            {-0.26377701E+02, 0.61225288E+02,-0.36126202E+02, 0.34660233E+02,
              0.78493519E+01,-0.54490952E+01, 0.13300788E+02,-0.13281824E+02,
              0.18924454E+02, 0.30061636E+01, 0.32234249E+02,-0.37474296E+01,
             -0.85291586E+01, },
            {-0.58997833E+02, 0.78057648E+02, 0.12184751E+02,-0.11472317E+02,
              0.42532295E+02, 0.64326682E+01,-0.15016547E+02, 0.18534817E+02,
              0.30961195E+02, 0.14846240E+02, 0.53940797E+01, 0.65931735E+01,
              0.31726680E+01, },
            { 0.10302815E+03, 0.10765425E+03,-0.18549156E+02,-0.29647400E+02,
             -0.18905441E+02, 0.51032181E+02,-0.18045750E+02,-0.76261740E+01,
              0.33367565E+01,-0.33117504E+01,-0.46762824E+01,-0.70676723E+01,
             -0.52129989E+01, },
            { 0.19405296E+02,-0.37196632E+02,-0.30845879E+02,-0.12045132E+01,
             -0.24903931E+02,-0.52441917E+02, 0.21243822E+02, 0.32386875E+02,
             -0.61232271E+01, 0.69945936E+01, 0.10995952E+02,-0.16689281E+01,
              0.32438049E+01, },
            { 0.83281013E+02,-0.24887656E+03, 0.94408211E+02,-0.81634117E+02,
             -0.56857452E+02, 0.11548672E+02,-0.39307831E+02, 0.35773415E+02,
             -0.44039265E+02,-0.80293608E+01,-0.81032394E+02, 0.15755465E+02,
              0.29912725E+02, },
            { 0.17956213E+03,-0.25433278E+03,-0.10402868E+03, 0.38665817E+02,
             -0.10846452E+03,-0.16294807E+02, 0.43562103E+02,-0.65742599E+02,
             -0.52942467E+02,-0.32878132E+02,-0.90460014E+01,-0.10402137E+02,
             -0.99966192E+00, },
            {-0.99776047E+02,-0.10776513E+03, 0.93763397E+02, 0.41308418E+02,
              0.23109884E+02,-0.71873962E+02, 0.29615770E+02, 0.85741634E+01,
              0.15823607E+01, 0.56734997E+00, 0.39712129E+01, 0.65275764E+01,
              0.91736221E+01, },
            {-0.30059841E+02,-0.42904842E+02, 0.75276306E+02,-0.11297899E+02,
              0.29414515E+02, 0.68314957E+02,-0.24166710E+02,-0.47956135E+02,
              0.79749370E+01,-0.81939878E+01,-0.17279491E+02, 0.42007875E+01,
             -0.36308939E+01, },
            {-0.12822026E+03, 0.33393097E+03,-0.12717429E+03, 0.89460068E+02,
              0.95602333E+02,-0.56442857E+01, 0.46425873E+02,-0.45104004E+02,
              0.47921806E+02, 0.10032402E+02, 0.92091537E+02,-0.21654915E+02,
             -0.41252964E+02, },
            {-0.20665901E+03, 0.35750424E+03, 0.15833980E+03,-0.42081184E+02,
              0.11878425E+03, 0.21206402E+02,-0.51555927E+02, 0.93430550E+02,
              0.37203754E+02, 0.32163887E+02, 0.86474857E+01, 0.69033074E+01,
             -0.54782982E+01, },
            { 0.31105921E+02, 0.42761517E+02,-0.58713291E+02,-0.18812452E+02,
             -0.99255018E+01, 0.34368755E+02,-0.16233343E+02,-0.36081984E+01,
             -0.22873654E+01, 0.10863211E+01,-0.70161188E+00,-0.17797345E+01,
             -0.49097610E+01, },
            { 0.14821376E+02, 0.46732437E+02,-0.43093964E+02, 0.10727091E+02,
             -0.10283649E+02,-0.30857159E+02, 0.94555550E+01, 0.23968437E+02,
             -0.39821031E+01, 0.38799286E+01, 0.87662678E+01,-0.24545429E+01,
              0.99674916E+00, },
            { 0.70579102E+02,-0.14992651E+03, 0.64224586E+02,-0.36984921E+02,
             -0.49639442E+02,-0.14587626E+01,-0.19475060E+02, 0.20724131E+02,
             -0.19756748E+02,-0.44517136E+01,-0.38782745E+02, 0.96913366E+01,
              0.19572958E+02, },
            { 0.79299797E+02,-0.17635577E+03,-0.73442924E+02, 0.14171571E+02,
             -0.48244438E+02,-0.10028917E+02, 0.21243670E+02,-0.45515968E+02,
             -0.82364569E+01,-0.11869068E+02,-0.36658957E+01,-0.14194759E+01,
              0.46354752E+01, },
            { 0.10682828E+00, 0.15335342E+00, 0.24009713E-02,-0.94681269E+00,
             -0.20006712E+00, 0.35301302E-01,-0.26023917E-01,-0.17329574E-01,
              0.51625643E-01,-0.21257993E-01,-0.38532102E-02,-0.74463747E-02,
              0.90640932E-02, },
            { 0.19327568E+00,-0.18736569E-01, 0.22726156E-01, 0.25009137E+00,
             -0.72113854E+00, 0.99601615E-02,-0.55031674E-02,-0.26947472E-01,
              0.28790315E-02,-0.16322194E-01,-0.36808793E-03,-0.56895283E-02,
              0.75899302E-02, },
            {-0.55829501E+00, 0.74548769E+00, 0.33699301E+00,-0.38933989E+00,
             -0.11029156E+01, 0.14290258E+00,-0.13941409E-01, 0.13658305E+00,
             -0.56958534E-02, 0.14612201E+00,-0.36390368E-01, 0.34695819E-04,
              0.22696797E-01, },
            {-0.19936664E-01, 0.11800451E+00, 0.74418652E+00, 0.16391089E+01,
             -0.22145706E+00, 0.27896428E-01, 0.86451977E-01, 0.23700632E+00,
             -0.18414943E-01, 0.22026940E-03,-0.35097837E-01, 0.70859253E-01,
              0.16451158E-01, },
            { 0.78674906E+00, 0.40604296E+00, 0.13178561E+01, 0.94937210E+01,
              0.35943595E+00,-0.45602840E+00, 0.23751763E+00, 0.16641727E+00,
             -0.49453366E+00,-0.13433233E+00,-0.40668243E+00,-0.21240635E-01,
             -0.36346906E+00, },
            { 0.72496071E-01, 0.11956791E+01, 0.17677049E+00,-0.21435833E+01,
              0.64546747E+01, 0.49960443E+00, 0.50610906E+00,-0.10921649E+00,
              0.65923679E+00, 0.13551632E+00,-0.21280682E+00, 0.16277494E+00,
              0.11476960E+00, },
            { 0.45960088E+01,-0.63736382E+01,-0.17190388E+01, 0.82465286E+01,
              0.17915984E+01,-0.91923392E+00,-0.91493428E+00,-0.46042120E+00,
             -0.24589279E+00,-0.14275963E+01, 0.90770066E+00, 0.12878948E+00,
              0.27167577E+00, },
            {-0.16134089E+01,-0.31690547E+01,-0.59042554E+01,-0.66804452E+01,
              0.78570251E+01, 0.24046221E+01,-0.80951697E+00,-0.12263418E+01,
             -0.75266880E+00,-0.16484741E+01, 0.21213526E+00,-0.13156597E+00,
              0.31026939E+00, },
            {-0.10191396E+02,-0.35821579E+01,-0.67020483E+01,-0.21100809E+02,
              0.50712733E+01, 0.15026681E+01, 0.15030552E+01,-0.14051753E+01,
              0.43638811E+01, 0.14188287E+01, 0.20353065E+01, 0.83522245E-01,
              0.18233762E+01, },
            {-0.14807859E+00,-0.29033175E+01, 0.49544346E+00, 0.72752118E+01,
             -0.59718432E+01,-0.30980995E+01,-0.21497185E+01,-0.39590228E+00,
             -0.21902370E+01,-0.58068059E-01, 0.13292863E+01,-0.71125287E+00,
             -0.20930350E+01, },
            {-0.11946964E+02, 0.14842890E+02,-0.15085144E+00,-0.26015095E+02,
             -0.11381587E+00, 0.23868535E+01, 0.37237940E+01, 0.10595403E+01,
              0.15183357E+01, 0.36690791E+01,-0.32947772E+01,-0.13435316E+00,
             -0.86445820E+00, },
            { 0.70006118E+01, 0.88694229E+01, 0.11823129E+02, 0.99858646E+01,
             -0.25837082E+02,-0.93706493E+01, 0.37652469E+01, 0.22050340E+01,
              0.24930382E+01, 0.60392127E+01,-0.10250652E+01,-0.25152349E+00,
             -0.87281799E+00, },
            { 0.25511642E+02, 0.61066298E+01, 0.16452080E+02, 0.43983407E+01,
             -0.15019115E+02,-0.25166245E+01,-0.56915488E+01, 0.42986555E+01,
             -0.11266659E+02,-0.37101965E+01,-0.29099541E+01, 0.31099221E+00,
             -0.24174685E+01, },
            {-0.37529130E+01,-0.29049420E+01,-0.49277145E+00,-0.12283454E+02,
             -0.22927643E+02, 0.58630066E+01, 0.24241989E+01, 0.20458815E+01,
              0.14258064E+01,-0.27732790E+00,-0.25777242E+01, 0.11977600E+01,
              0.57707081E+01, },
            { 0.85478735E+01,-0.10505727E+02, 0.35088875E+01, 0.18057755E+02,
             -0.31088126E+01,-0.20258300E+01,-0.33768830E+01,-0.75583994E+00,
             -0.16262639E+01,-0.27519293E+01, 0.28147068E+01,-0.59420045E-01,
              0.66975570E+00, },
            {-0.69251733E+01,-0.75538425E+01,-0.68501964E+01,-0.25737543E+01,
              0.18046787E+02, 0.80261040E+01,-0.44093361E+01,-0.15761824E+01,
             -0.17725418E+01,-0.48975239E+01, 0.11016632E+01, 0.48760176E+00,
              0.51885700E+00, },
            {-0.17836386E+02,-0.32042096E+01,-0.13130729E+02, 0.11606612E+02,
              0.12947696E+02, 0.19340641E+01, 0.47053881E+01,-0.35811558E+01,
              0.81174049E+01, 0.28834205E+01, 0.10804650E+01,-0.32320696E+00,
              0.74705285E+00, },
            { 0.50985780E+01, 0.60607548E+01,-0.27176561E+01, 0.50116320E+01,
              0.27655184E+02,-0.36654398E+01, 0.24630740E+00,-0.14518211E+01,
              0.14266887E+00, 0.26619384E+00, 0.16181467E+01,-0.81003147E+00,
             -0.41553168E+01, },
            { 0.44094801E-01, 0.12518990E+00,-0.13474181E+00,-0.11440214E+00,
              0.75728670E-01,-0.35541511E+00,-0.50099200E+00,-0.55849489E-01,
              0.62280174E-01,-0.11978948E-01,-0.18889341E-03,-0.37581697E-01,
             -0.79547213E-02, },
            { 0.80496319E-01,-0.75641625E-01,-0.15093005E+00,-0.28906606E-01,
             -0.13633297E+00, 0.48335886E+00,-0.39124358E+00,-0.51529985E-01,
             -0.16674163E-01, 0.13538418E-01, 0.16163324E-02,-0.82436623E-02,
             -0.84085455E-04, },
            {-0.27754477E+00, 0.71311891E+00,-0.89700386E-01,-0.21103679E+00,
             -0.17687225E+00, 0.15123169E+00, 0.22682615E-01, 0.21121208E+00,
             -0.16462816E-01,-0.60570084E-01, 0.39293535E-01,-0.10021122E+00,
              0.95546981E-02, },
            {-0.29436976E+00,-0.28597829E+00,-0.23569670E+00,-0.61088581E-01,
             -0.30247924E+00, 0.42297073E-01, 0.29653084E+00,-0.27596239E-01,
             -0.62871657E-01,-0.64206328E-02,-0.10770185E+00,-0.35808660E-01,
             -0.16216652E-01, },
            {-0.14272901E+00,-0.95003837E+00, 0.14099309E+01, 0.34587759E+00,
             -0.58685106E+00,-0.53814942E+00,-0.16351078E+00, 0.94250903E-01,
              0.10865152E+00, 0.25800949E+00, 0.30995958E-01, 0.31330404E+00,
             -0.10657041E+00, },
            {-0.97229975E+00, 0.19243021E+01, 0.96367663E+00,-0.21867950E+00,
              0.12699347E+01, 0.10597437E+00,-0.20831758E+00,-0.16814458E+00,
              0.64819470E-01,-0.26897462E-01, 0.20455129E+00, 0.10330691E+00,
             -0.44147037E-01, },
            {-0.15029612E+00,-0.14936835E+01, 0.13109908E+00, 0.61500740E+00,
              0.24869451E+00,-0.14464337E+00, 0.21722548E+00,-0.44552392E+00,
              0.24300979E-02, 0.12858246E+00,-0.11394108E+00, 0.26291013E+00,
             -0.14204117E-01, },
            { 0.23137939E+00, 0.89680660E+00, 0.70877725E+00, 0.72342165E-01,
              0.90643167E+00,-0.44160914E+00,-0.38550001E+00, 0.47713518E-01,
              0.66411771E-01,-0.97876545E-02, 0.23822626E+00, 0.20062061E+00,
              0.58113962E-01, },
            { 0.85560966E+00, 0.20959485E+01,-0.21306541E+01,-0.25514224E+00,
              0.12787257E+01, 0.20966003E+01, 0.11189801E+01, 0.11804737E+00,
             -0.47789565E+00,-0.44577163E+00, 0.65643959E-01,-0.56391513E+00,
              0.21226558E+00, },
            { 0.11625001E+01,-0.24074843E+01,-0.14329216E+01,-0.33291936E-01,
             -0.19074674E+01,-0.82054245E+00, 0.16421086E+01, 0.59356934E+00,
             -0.19039523E-01,-0.14523591E+00,-0.37697402E+00,-0.36098781E+00,
              0.90530626E-01, },
            { 0.41392002E-01,-0.41367438E-01,-0.56517832E-01,-0.29723020E-01,
              0.45230564E-01, 0.29130369E-01,-0.52787438E-02, 0.22493723E+00,
             -0.37028406E-01,-0.76311780E-02, 0.66862693E-02,-0.14504310E-01,
             -0.11098138E-01, },
            { 0.60032666E-01,-0.52753162E-01, 0.69005713E-02, 0.17640466E-01,
             -0.42814646E-01, 0.10649805E-02,-0.16114800E-02, 0.41139279E-01,
              0.21040384E+00, 0.91859568E-02,-0.34672103E-02, 0.49885898E-02,
             -0.36266081E-02, },
            {-0.22909485E-01,-0.57594564E-01,-0.49758352E-01, 0.40956624E-01,
             -0.18536091E+00,-0.12244110E+00, 0.80677725E-01, 0.27833149E+00,
              0.26716733E+00,-0.18367633E-01,-0.48983745E-01,-0.45404307E-03,
              0.61136829E-02, },
            { 0.23329218E+00, 0.21748865E+00,-0.60024496E-01,-0.60173009E-01,
              0.15808193E+00, 0.48826919E-02,-0.13135211E-01,-0.34629282E+00,
              0.13930210E+00,-0.11054378E-01,-0.39946426E-01,-0.16428996E-01,
             -0.25691027E-01, },
            {-0.12797327E+00,-0.13176106E-01, 0.16579621E-01,-0.26598467E-01,
             -0.44626016E-01, 0.11000606E-01,-0.68844962E-02, 0.30088242E-01,
             -0.21648988E-01, 0.85243776E-01, 0.13738230E+00, 0.71503147E-02,
             -0.82375538E-02, },
            { 0.41204984E-02,-0.28827868E-02,-0.10035298E-01, 0.48501052E-01,
             -0.72628535E-01,-0.31241590E-01,-0.27600962E-02, 0.50012227E-01,
              0.15929714E-01,-0.15858254E+00, 0.73988296E-01, 0.60939570E-02,
             -0.19759694E-01, },
            {-0.29281452E-01,-0.43308523E-01,-0.50946777E-02, 0.35335834E-03,
             -0.43678591E-02, 0.27276650E-02,-0.24702514E-01, 0.39685573E-01,
              0.26893996E-01, 0.11443176E-02, 0.35606234E-03,-0.97904116E-01,
              0.17392356E-01, },
            { 0.29230772E-01, 0.30779630E-01, 0.66181175E-01, 0.33190589E-01,
              0.54724183E-01,-0.13985308E-01,-0.75963368E-02,-0.36636444E-02,
             -0.11163543E-01,-0.28938515E-01,-0.19950645E-01,-0.18290415E-01,
             -0.11451925E+00, },
            { 0.25555792E-02, 0.33711719E-02,-0.17088523E-01, 0.12914094E-01,
              0.22545582E-01,-0.11638348E-01, 0.84269810E-02, 0.31605013E-01,
              0.43525486E-02,-0.14523819E-01,-0.11092685E-01, 0.72272657E-03,
              0.65107662E-02, },
            { 0.73336594E-01,-0.65333776E-01,-0.23503136E-01, 0.45837099E-02,
              0.15234247E-01,-0.57932669E-02, 0.15654413E-02,-0.23898358E-01,
              0.22590397E-01,-0.16863436E-01,-0.25549070E-02,-0.54761791E-02,
              0.18356221E-01, },
            { 0.17226540E-01, 0.12007785E+00,-0.12394503E+00, 0.27426206E-01,
              0.75115939E-02,-0.23556633E-01, 0.14968693E-01,-0.76391217E-02,
              0.15911356E-01,-0.28241746E-01,-0.31705524E-02,-0.28219888E-01,
             -0.10984738E-01, },
            {-0.51771577E-01,-0.57098405E-02, 0.31669147E-01,-0.37942089E-01,
              0.34906853E-01, 0.10111721E-01,-0.26458624E-03,-0.22664923E-01,
              0.10375086E-01, 0.15457364E-01,-0.95300321E-02, 0.46850368E-02,
              0.10677025E-01, },
         },
         {
            { 0.80510998E+01,-0.25949422E-01, 0.21837340E+00, 0.76524161E-01,
             -0.26052505E-01,-0.19464865E-01,-0.11594040E-01, 0.40525004E-01,
             -0.83783437E-02, 0.77255415E-02,-0.43546511E-02, 0.13457161E-01,
             -0.19088319E-01, },
            {-0.62016821E+00,-0.67613447E+00, 0.24612109E+00,-0.29176807E+00,
              0.24085705E+00,-0.57791073E-01, 0.25819287E+00, 0.50516576E-01,
             -0.14634442E+00, 0.24345282E-02, 0.78144493E-02,-0.72793506E-01,
              0.15928514E-01, },
            { 0.17819746E+02, 0.14743440E+01,-0.27259543E+01,-0.13425981E+00,
              0.37347874E+00, 0.50698123E-02, 0.27457744E+00,-0.32091236E+00,
              0.19228476E+00, 0.13757847E+00,-0.13359857E+00,-0.24793717E+00,
              0.29714280E+00, },
            {-0.11858767E+02, 0.69419298E+01, 0.46298509E+01, 0.42721686E+01,
             -0.35434074E+01, 0.75084180E+00,-0.42302465E+01,-0.70975661E+00,
              0.99484891E+00,-0.40930372E+00,-0.47602188E-01, 0.14345567E+01,
              0.58592618E+00, },
            {-0.97920853E+02,-0.13692502E+02, 0.91235781E+01,-0.14460938E+01,
             -0.85480909E+01, 0.18136225E+01,-0.40499854E+00, 0.23285608E+01,
              0.21228252E+00,-0.17068177E+01, 0.10739388E+01, 0.17055205E+01,
             -0.15911946E+01, },
            { 0.13621458E+03,-0.31045456E+02,-0.30203131E+02,-0.16606495E+02,
              0.24689091E+02,-0.46569004E+01, 0.19129166E+02, 0.27777817E+01,
             -0.42672701E+01, 0.31055405E+01, 0.20506692E+00,-0.88754292E+01,
             -0.49245520E+01, },
            { 0.18260400E+03, 0.48245602E+02,-0.12680835E+02, 0.67058239E+01,
              0.30385052E+02,-0.77466555E+01,-0.24252276E+01,-0.74331207E+01,
             -0.30216341E+01, 0.50820293E+01,-0.34738324E+01,-0.39474041E+01,
              0.36654823E+01, },
            {-0.37284259E+03, 0.57534088E+02, 0.59615005E+02, 0.21586212E+02,
             -0.65765862E+02, 0.11232314E+02,-0.35051090E+02,-0.48223314E+01,
              0.10247663E+02,-0.89555550E+01, 0.19287635E-01, 0.21166868E+02,
              0.12303819E+02, },
            {-0.16647343E+03,-0.65894676E+02, 0.92080431E+01,-0.95255365E+01,
             -0.38242653E+02, 0.10814571E+02, 0.54566917E+01, 0.95939255E+01,
              0.48733854E+01,-0.54909315E+01, 0.43026366E+01, 0.36398020E+01,
             -0.39649410E+01, },
            { 0.40097147E+03,-0.42800575E+02,-0.49323971E+02,-0.69891477E+01,
              0.72595741E+02,-0.11016485E+02, 0.28427053E+02, 0.38504643E+01,
             -0.11370510E+02, 0.10496172E+02,-0.38129082E+00,-0.21485941E+02,
             -0.12095145E+02, },
            { 0.60411118E+02, 0.29807955E+02,-0.30553248E+01, 0.43338375E+01,
              0.16042967E+02,-0.49058828E+01,-0.29349284E+01,-0.42607260E+01,
             -0.22481444E+01, 0.19431134E+01,-0.17624894E+01,-0.11618649E+01,
              0.16431694E+01, },
            {-0.15156676E+03, 0.10127841E+02, 0.15031334E+02,-0.20478544E+01,
             -0.28253370E+02, 0.37018678E+01,-0.85423803E+01,-0.11203144E+01,
              0.45561490E+01,-0.42304101E+01, 0.19091371E+00, 0.78487253E+01,
              0.40897169E+01, },
            {-0.36684188E+00, 0.12926129E+01, 0.19531890E+01, 0.42412702E-01,
             -0.14427768E+00,-0.23068365E-01, 0.24992963E-02, 0.48195336E-01,
             -0.24228871E-02, 0.57797058E-03,-0.24764627E-01, 0.62236316E-02,
             -0.12336963E-01, },
            { 0.35460928E+00,-0.17541237E+01, 0.13221329E+01,-0.24655612E-04,
              0.17179137E+00, 0.25261961E-01, 0.78048166E-02,-0.28375529E-01,
              0.45270234E-01,-0.58390521E-02, 0.73639741E-02,-0.46915747E-02,
              0.16659260E-01, },
            {-0.69456911E+00, 0.28679106E+00,-0.34354627E+01, 0.52327073E+00,
              0.16810082E+00,-0.47624176E-02,-0.91403551E-01,-0.16691753E+00,
              0.12325200E-01, 0.15972370E+00, 0.14241000E+00, 0.19505944E-01,
              0.13720220E+00, },
            {-0.31720053E-02, 0.21107395E+01,-0.78734863E+00,-0.80523528E-02,
              0.38587731E+00, 0.15747765E+00, 0.84543660E-01, 0.20151427E+00,
              0.25630075E+00, 0.16800867E+00, 0.43668777E-01, 0.15243682E+00,
              0.24071131E-01, },
            { 0.73923025E+01, 0.16647623E+02,-0.37959800E+01, 0.78674835E+00,
              0.54041618E+00, 0.10678436E+01,-0.19122691E+00,-0.94048089E+00,
              0.97269678E+00,-0.43552950E+00, 0.79522692E-01,-0.37138738E-01,
             -0.29479025E-01, },
            {-0.34290457E+01,-0.85238671E+00, 0.14262810E+02, 0.15189956E+01,
             -0.36586006E+01,-0.14248925E+01,-0.20277178E+00, 0.64990950E+00,
             -0.36768889E+00, 0.84792279E-01,-0.21742781E+00,-0.16557669E-01,
              0.12499936E+00, },
            { 0.73744383E+01,-0.12027929E+02, 0.55477055E+02,-0.12374017E+02,
             -0.89327345E+01, 0.21117227E+01,-0.63658524E+00, 0.36965864E+01,
             -0.23416463E+00,-0.27068796E+01,-0.28190041E+01, 0.65999812E+00,
             -0.12989928E+01, },
            { 0.79730363E+01,-0.21980247E+02, 0.14381426E+02, 0.35661685E+01,
             -0.12320098E+02,-0.91450326E-01, 0.10625300E+01,-0.12245990E+01,
             -0.24598341E+01,-0.45958142E+01,-0.12662302E+01,-0.16750321E+01,
              0.63961756E+00, },
            {-0.44315506E+02,-0.68935883E+02, 0.51209675E+02,-0.96216927E+01,
              0.28568304E+01,-0.88125429E+01, 0.41833620E+01, 0.88361034E+01,
             -0.57595100E+01, 0.20263133E+01, 0.30380830E+00, 0.13347017E+01,
              0.20728569E+01, },
            { 0.27656660E+02,-0.24090044E+02,-0.49807522E+02,-0.16521885E+02,
              0.37152962E+02, 0.12428595E+02, 0.23935599E+01,-0.96629171E+01,
              0.31866329E+01,-0.14910942E+00,-0.12049040E+01, 0.13381022E+01,
             -0.26929400E+01, },
            {-0.27525063E+02, 0.94771980E+02,-0.32496066E+03, 0.86580139E+02,
              0.72477280E+02,-0.22168686E+02, 0.94462023E+01,-0.26052528E+02,
              0.84192997E+00, 0.15367923E+02, 0.14202004E+02,-0.84568520E+01,
              0.30221646E+01, },
            {-0.95830826E+02, 0.13345288E+03,-0.56603416E+02,-0.34000439E+02,
              0.71806602E+02,-0.46853423E+01,-0.18473898E+02, 0.10775526E+02,
              0.38866134E+01, 0.27761209E+02, 0.10534242E+02, 0.68109131E+01,
             -0.58825417E+01, },
            { 0.10576842E+03, 0.75757675E+02,-0.18107726E+03, 0.35365593E+02,
             -0.19049076E+02, 0.26869581E+02,-0.20695049E+02,-0.29340977E+02,
              0.12610971E+02,-0.15228617E+01,-0.24438107E+01,-0.53549356E+01,
             -0.92385302E+01, },
            {-0.89127052E+02, 0.11649306E+03, 0.99241447E+01, 0.59149529E+02,
             -0.12863148E+03,-0.36652939E+02,-0.66132178E+01, 0.37069744E+02,
             -0.13132428E+02,-0.28511498E+01, 0.12527537E+02,-0.64466619E+01,
              0.80187778E+01, },
            { 0.66444099E+02,-0.26490027E+03, 0.78100250E+03,-0.24294917E+03,
             -0.21248206E+03, 0.73090462E+02,-0.29586170E+02, 0.75238579E+02,
              0.11143709E+01,-0.39969158E+02,-0.29317850E+02, 0.27842539E+02,
              0.15493094E+01, },
            { 0.31403052E+03,-0.33532126E+03, 0.10940713E+03, 0.94474915E+02,
             -0.15778313E+03, 0.12858322E+02, 0.64601768E+02,-0.41679054E+02,
              0.12144041E+02,-0.63168148E+02,-0.34294418E+02,-0.11578428E+02,
              0.18609144E+02, },
            {-0.10748645E+03,-0.73261108E+01, 0.25742560E+03,-0.49046207E+02,
              0.29797882E+02,-0.34313000E+02, 0.34743034E+02, 0.39293594E+02,
             -0.13067580E+02,-0.31035950E+01, 0.34878824E+01, 0.69859462E+01,
              0.13996960E+02, },
            { 0.12487012E+03,-0.19734895E+03, 0.10182087E+03,-0.88575996E+02,
              0.17575227E+03, 0.44216854E+02, 0.75572181E+01,-0.53957497E+02,
              0.22110167E+02, 0.75596704E+01,-0.24996717E+02, 0.10142770E+02,
             -0.73031473E+01, },
            {-0.81616684E+02, 0.30472897E+03,-0.86462183E+03, 0.28986890E+03,
              0.26276410E+03,-0.95904716E+02, 0.36672890E+02,-0.93806793E+02,
             -0.43140082E+01, 0.47584782E+02, 0.27736128E+02,-0.35119770E+02,
             -0.99922857E+01, },
            {-0.39365311E+03, 0.40164426E+03,-0.12205155E+03,-0.98447792E+02,
              0.14291895E+03,-0.10083131E+02,-0.84789528E+02, 0.63032337E+02,
             -0.34860493E+02, 0.60776482E+02, 0.46811893E+02, 0.81084528E+01,
             -0.25287830E+02, },
            { 0.36671532E+02,-0.15971799E+02,-0.12797734E+03, 0.23251619E+02,
             -0.13679494E+02, 0.15429979E+02,-0.18713863E+02,-0.18289385E+02,
              0.52698426E+01, 0.33367665E+01,-0.13960910E+01,-0.28204956E+01,
             -0.69832816E+01, },
            {-0.59758083E+02, 0.11212505E+03,-0.80015900E+02, 0.45999847E+02,
             -0.81153366E+02,-0.18856775E+02,-0.28614230E+01, 0.26849733E+02,
             -0.12275286E+02,-0.47784538E+01, 0.14424383E+02,-0.51218362E+01,
              0.15706040E+01, },
            { 0.38644615E+02,-0.12562950E+03, 0.36313803E+03,-0.12401902E+03,
             -0.11666864E+03, 0.43718800E+02,-0.16167812E+02, 0.41872963E+02,
              0.27231398E+01,-0.20851528E+02,-0.99890919E+01, 0.15157324E+02,
              0.68729215E+01, },
            { 0.16797034E+03,-0.18619617E+03, 0.57807312E+02, 0.33903023E+02,
             -0.45472706E+02, 0.14138937E+01, 0.37569668E+02,-0.32199303E+02,
              0.21854696E+02,-0.20931599E+02,-0.22431284E+02,-0.17632341E+01,
              0.12280620E+02, },
            { 0.67577185E-02,-0.38690273E-01, 0.12910472E+00,-0.96712691E+00,
              0.28352141E+00,-0.23064802E-02,-0.67802047E-03,-0.94842110E-02,
              0.16349863E-01,-0.21919638E-01, 0.69284305E-03,-0.18507710E-01,
              0.65062190E-02, },
            { 0.98213874E-01,-0.86133815E-01,-0.43852173E-01,-0.25530991E+00,
             -0.75969112E+00, 0.14430066E-02, 0.11108719E-01,-0.26231494E-01,
             -0.16866442E-01,-0.13983527E-01, 0.39212308E-02,-0.13768057E-01,
              0.86285966E-02, },
            { 0.52556640E+00, 0.14199088E+01, 0.23482423E+00, 0.12398671E+01,
             -0.11701978E+01, 0.22943036E+00, 0.19906037E-01, 0.14976382E+00,
             -0.83794696E-02, 0.12005270E+00,-0.11553056E+00, 0.17222632E-01,
              0.61729096E-01, },
            { 0.25419283E+00, 0.50038338E+00, 0.77746254E+00, 0.17081389E+01,
              0.14862339E+01,-0.36264654E-01, 0.18140624E+00, 0.25878024E+00,
              0.94825074E-01, 0.19079769E-01,-0.11765101E-01, 0.11579243E+00,
              0.53563844E-01, },
            { 0.29051917E+01, 0.17165275E+01,-0.23254345E+00, 0.30379848E+01,
              0.28690646E+01, 0.66818811E-01,-0.14130628E+00,-0.79569364E+00,
              0.35532689E+00,-0.21288022E+00,-0.17567311E+00, 0.18693624E-01,
             -0.56478500E+00, },
            { 0.17097101E+01,-0.15680478E+01, 0.19375648E+01,-0.37112069E+01,
              0.15600331E+01, 0.15556124E+00,-0.19353969E+00,-0.24991187E+00,
              0.85250163E+00, 0.35902387E+00,-0.43576801E+00, 0.26326782E+00,
              0.53666756E-01, },
            {-0.71911683E+01,-0.10550578E+02, 0.53046633E-01, 0.34881504E+01,
              0.18754406E+01,-0.14867971E+01,-0.13179111E+01,-0.11313314E+01,
             -0.38169405E+00,-0.14188350E+01, 0.18088046E+01, 0.11685710E+00,
              0.16416541E+00, },
            {-0.77558784E+01,-0.51033101E+01,-0.53945317E+01,-0.79216189E+01,
              0.20668960E+01, 0.22548456E+01,-0.17693250E+01,-0.83705181E+00,
             -0.15927854E+01,-0.11773171E+01,-0.69688298E-01,-0.17553112E+00,
              0.24276218E+00, },
            {-0.18529564E+02,-0.49476762E+01,-0.80366945E+01,-0.59576052E+00,
             -0.12046759E+02, 0.84566593E+00, 0.37427566E+01, 0.49305139E+01,
             -0.11143556E+01, 0.25452406E+01,-0.23028915E+00, 0.14220794E+00,
              0.31511154E+01, },
            {-0.19755463E+00, 0.14758796E+02,-0.15246205E+02, 0.17099882E+02,
              0.68206816E+01,-0.16745818E+01, 0.30730574E+01, 0.30909929E+00,
             -0.25502460E+01,-0.20344477E+01, 0.21614161E+01,-0.16361508E+01,
             -0.19083089E+01, },
            { 0.17850922E+02, 0.23550045E+02,-0.43193951E+01,-0.17244564E+02,
              0.45090499E+01, 0.36393521E+01, 0.38667064E+01, 0.35419881E+01,
              0.14072553E+01, 0.40633569E+01,-0.58945403E+01,-0.35010526E+00,
             -0.12329621E+01, },
            { 0.25905699E+02, 0.12087793E+02, 0.13750561E+02, 0.92878151E+01,
             -0.13878139E+02,-0.71122155E+01, 0.60348301E+01, 0.10128769E+01,
              0.49910145E+01, 0.45441375E+01,-0.48133820E+00,-0.69957542E+00,
             -0.12999972E+01, },
            { 0.35084892E+02, 0.58390584E+01, 0.27396364E+02,-0.14684110E+02,
              0.16126560E+02,-0.30175450E+01,-0.90232534E+01,-0.10018678E+02,
              0.10488937E+01,-0.65118580E+01, 0.26540687E+01,-0.18480808E+00,
             -0.52721381E+01, },
            {-0.17716702E+02,-0.33102341E+02, 0.34046539E+02,-0.25810240E+02,
             -0.28638424E+02, 0.30514753E+01,-0.93732405E+01, 0.12338361E+00,
              0.99789155E+00, 0.36990767E+01,-0.32632139E+01, 0.32731326E+01,
              0.55128732E+01, },
            {-0.13897304E+02,-0.15651490E+02, 0.54399266E+01, 0.10493177E+02,
             -0.10791659E+02,-0.24484687E+01,-0.28806729E+01,-0.28858256E+01,
             -0.10917011E+01,-0.32969794E+01, 0.47821879E+01, 0.21285339E+00,
              0.13126817E+01, },
            {-0.23277504E+02,-0.86345367E+01,-0.10629125E+02, 0.21820576E+01,
              0.74627986E+01, 0.50995250E+01,-0.59439478E+01,-0.66435319E+00,
             -0.40929780E+01,-0.37793703E+01, 0.67755610E+00, 0.11436670E+01,
              0.11544902E+01, },
            {-0.19324162E+02,-0.31898675E+01,-0.20808168E+02, 0.17447025E+02,
             -0.33315101E+01, 0.21073177E+01, 0.58234086E+01, 0.62877498E+01,
             -0.35331634E+00, 0.48601122E+01,-0.26413305E+01, 0.93884856E-01,
              0.25990388E+01, },
            { 0.21725630E+02, 0.20295200E+02,-0.24284405E+02, 0.83008165E+01,
              0.26109770E+02,-0.13780729E+01, 0.82305450E+01, 0.23272064E+00,
              0.11319129E+01,-0.20607281E+01, 0.17115957E+01,-0.22145329E+01,
             -0.40296369E+01, },
            { 0.11448875E+00, 0.53372294E-01,-0.10701045E-01,-0.74131079E-01,
              0.42126238E-01,-0.47183973E+00,-0.62683088E+00,-0.49725015E-01,
              0.23906445E-01,-0.18615820E-01, 0.69484608E-02,-0.43719456E-01,
             -0.12143527E-01, },
            { 0.18777304E-03,-0.18922636E+00,-0.75974211E-01,-0.15254694E-01,
             -0.11909236E+00, 0.59676683E+00,-0.50491786E+00,-0.35921060E-01,
              0.58115725E-02, 0.59416122E-02, 0.95220320E-02,-0.17321225E-01,
              0.72834064E-03, },
            {-0.31445304E+00, 0.10098319E+01,-0.87438963E-01,-0.22055499E+00,
             -0.36901802E+00, 0.44668826E+00, 0.42858791E+00, 0.21535960E+00,
             -0.43418493E-01,-0.49117277E-02, 0.10932283E+00,-0.98627552E-01,
              0.37286639E-01, },
            {-0.16686460E+00,-0.16635004E+00,-0.32959878E-01,-0.37202913E-01,
             -0.50687611E+00,-0.32216674E+00, 0.34823060E+00, 0.52472215E-01,
              0.84805384E-01, 0.22642313E-01,-0.84570244E-01,-0.48971225E-01,
              0.45566838E-01, },
            {-0.63517235E-01,-0.32054949E+00, 0.87246031E-01, 0.14539485E+00,
             -0.43560812E+00, 0.26938108E+00, 0.18754629E-01,-0.34433404E+00,
              0.29875723E+00, 0.17472742E+00, 0.54090742E-01, 0.38365376E+00,
             -0.13843028E+00, },
            {-0.48946375E+00, 0.28971550E+01,-0.20743754E-01,-0.83255611E-01,
              0.11938328E+01, 0.55411711E-01, 0.60760695E+00,-0.29698908E+00,
             -0.55650949E+00, 0.20695293E-01, 0.36676876E-01, 0.14274941E+00,
             -0.98814443E-01, },
            {-0.46403563E+00,-0.19422119E+01, 0.36596009E+00, 0.68322176E+00,
              0.45094913E+00, 0.16012102E+00,-0.40091038E+00,-0.46582344E+00,
              0.89832827E-01, 0.94689272E-01,-0.34453103E+00, 0.32739177E+00,
             -0.56781594E-01, },
            {-0.38054120E-01, 0.52844387E+00, 0.19522023E+00, 0.33836249E-01,
              0.12656829E+01,-0.14137593E-01, 0.38812596E+00,-0.22139823E+00,
             -0.17572556E+00,-0.40329311E-01, 0.20087887E+00, 0.30446243E+00,
             -0.45953188E-01, },
            { 0.10138659E+01, 0.14339803E+01,-0.45818272E+00,-0.24776068E+00,
              0.95378476E+00, 0.11445934E+00, 0.66230971E+00, 0.79100257E+00,
             -0.76573831E+00,-0.34316742E+00, 0.80294624E-01,-0.70623606E+00,
              0.29327330E+00, },
            { 0.63751262E+00,-0.37546885E+01,-0.70940018E-01,-0.13639070E+00,
             -0.21821494E+01,-0.36584163E+00,-0.49187639E+00, 0.82650185E+00,
              0.85966331E+00,-0.19263405E+00,-0.79641826E-01,-0.47789237E+00,
              0.15595616E+00, },
            { 0.34587827E-01,-0.33244256E-01,-0.54627232E-01,-0.24314072E-01,
              0.39300185E-01, 0.12838095E-01,-0.27576180E-01, 0.22983019E+00,
             -0.15771292E+00,-0.23307070E-01, 0.17552825E-01,-0.18896224E-01,
             -0.17950874E-01, },
            { 0.24897996E-01,-0.81121093E-02,-0.34756999E-01, 0.31306081E-01,
             -0.67682348E-01, 0.25938660E-01,-0.15375219E-01, 0.15197173E+00,
              0.24201533E+00,-0.26705752E-02, 0.45459899E-02, 0.56280023E-02,
             -0.77642538E-02, },
            {-0.17782798E+00,-0.95260762E-01, 0.24179123E-01, 0.14758726E-01,
             -0.92235729E-01,-0.15296949E+00, 0.94236672E-01, 0.82640037E-01,
              0.50521767E+00, 0.33969387E-01,-0.62877722E-01, 0.20034479E-01,
              0.95262146E-02, },
            { 0.32055417E+00, 0.20525834E+00,-0.76902891E-02,-0.64364143E-01,
              0.19882102E+00,-0.81640065E-01,-0.71830384E-01,-0.49157616E+00,
             -0.57283647E-01, 0.12709680E-01,-0.30256083E-01,-0.39767653E-01,
             -0.55746748E-02, },
            {-0.16962911E+00,-0.24519250E-01, 0.53539060E-01,-0.22426639E-01,
             -0.52650891E-01, 0.57139243E-02,-0.12633940E-01, 0.43505147E-01,
             -0.18997150E-01, 0.14223887E+00, 0.16691910E+00, 0.28885426E-02,
             -0.21186968E-01, },
            { 0.75518645E-01, 0.52836422E-01,-0.81654236E-01, 0.27446862E-01,
             -0.50890606E-01,-0.17863950E-01,-0.73899115E-02, 0.33153400E-01,
              0.59054721E-01,-0.16145012E+00, 0.14286152E+00, 0.14969439E-01,
             -0.21417528E-01, },
            {-0.22653613E-01,-0.57655737E-01,-0.41951939E-01, 0.21979176E-01,
              0.39826494E-01, 0.30946506E-02,-0.23654750E-01, 0.27493823E-01,
              0.20307899E-01, 0.68600680E-03,-0.30355731E-01,-0.89675814E-01,
              0.58423135E-01, },
            { 0.93922287E-01, 0.68419144E-01, 0.33733767E-01,-0.72773839E-02,
              0.48324339E-01,-0.24635356E-02,-0.89860288E-02,-0.79068150E-02,
              0.14755895E-01,-0.16021999E-01,-0.71704332E-02,-0.60977753E-01,
             -0.10361492E+00, },
            { 0.12631768E-01,-0.49197290E-01, 0.24660140E-01, 0.43001071E-01,
              0.69695385E-02,-0.31781915E-03, 0.53595994E-02, 0.34497995E-01,
              0.24565097E-01,-0.50819240E-03,-0.10135818E-01,-0.16772541E-02,
              0.32331720E-02, },
            { 0.38050801E-01,-0.56574538E-01,-0.31323060E-01,-0.14262537E-01,
             -0.38575940E-01, 0.97641703E-02,-0.77285576E-02,-0.27659681E-01,
              0.20871909E-01,-0.15837995E-01, 0.18971974E-01,-0.38362518E-02,
              0.23249378E-01, },
            { 0.32181796E-01, 0.80461048E-01,-0.23041502E-01, 0.28507439E-01,
             -0.57646193E-01,-0.25178518E-01, 0.10214640E-01,-0.14632375E-02,
              0.22060374E-01,-0.57708239E-02, 0.16520452E-01,-0.30608475E-01,
             -0.77883643E-02, },
            {-0.38133394E-01, 0.41286569E-01, 0.27755771E-01,-0.51213089E-01,
              0.75025320E-01, 0.91396663E-02,-0.18474469E-02,-0.56077321E-02,
              0.64123259E-02, 0.22182658E-01,-0.65464103E-02,-0.81443517E-02,
              0.73059131E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir16Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30632045E+01,-0.57797851E-02, 0.91647813E-02, 0.41370955E-02,
             -0.19437239E-02, 0.24532698E-03, 0.60106777E-02,-0.21756680E-02,
             -0.31438239E-02, },
            {-0.41220626E+00, 0.29579294E-02, 0.18636411E-01, 0.37512437E-01,
             -0.35412803E-01, 0.96220765E-02,-0.19383488E-01,-0.20679396E-02,
              0.28241726E-02, },
            { 0.10044123E+01, 0.77929936E-01,-0.67139186E-01,-0.42513080E-01,
             -0.39895583E-01,-0.43510310E-01, 0.13425289E-01, 0.35680491E-01,
              0.17047495E-01, },
            { 0.43130204E+00,-0.24383721E+00,-0.48306435E-01,-0.16724011E+00,
              0.17723440E+00,-0.52420892E-01, 0.50915804E-01,-0.49679428E-02,
              0.87580746E-02, },
            {-0.12934322E+01, 0.17085487E-01, 0.12944695E+00,-0.12789106E-01,
              0.11405879E+00, 0.17716755E+00,-0.93909204E-01,-0.11515923E+00,
             -0.44955883E-01, },
            {-0.30758275E-01, 0.21726264E+00, 0.53814311E-01, 0.14063519E+00,
             -0.15107477E+00, 0.41386593E-01,-0.35365798E-01, 0.10928157E-01,
             -0.13283265E-01, },
            { 0.26758054E+00,-0.95864914E-01,-0.27972437E-01, 0.68929851E-01,
             -0.87709561E-01,-0.14526750E+00, 0.82934819E-01, 0.83224915E-01,
              0.33448048E-01, },
            { 0.21264717E-01,-0.10639626E+00,-0.32184544E+00, 0.10739344E-01,
             -0.15501696E-01, 0.23128907E-02,-0.49819713E-02, 0.26865129E-02,
              0.11654526E-02, },
            { 0.12853952E-02, 0.34511042E+00,-0.64418115E-01,-0.11180845E-01,
              0.79391981E-02, 0.22869739E-02,-0.77759842E-02, 0.74287609E-03,
              0.91153430E-02, },
            {-0.58623929E-01,-0.22482648E+00,-0.12649196E+00, 0.18788284E-01,
             -0.15773132E-01,-0.67276560E-01,-0.85366756E-01, 0.54488275E-01,
              0.43759987E-01, },
            { 0.15944013E-01, 0.91821142E-01,-0.17644204E+00, 0.27534859E-01,
              0.95351227E-02,-0.37926875E-01,-0.12159072E-02, 0.34995351E-01,
              0.62579088E-01, },
            {-0.14691392E+00, 0.48885843E+00, 0.11924573E+01,-0.97168460E-01,
             -0.39609086E-01,-0.14198712E+00, 0.43608997E-01, 0.92370085E-01,
             -0.18335907E-01, },
            {-0.61437911E+00,-0.14346495E+01,-0.14602734E-01, 0.17535535E+00,
             -0.40383213E-02,-0.13587080E+00,-0.25987262E-01,-0.29818421E-01,
             -0.44571441E-01, },
            { 0.85389566E+00, 0.11196742E+01,-0.79580927E+00, 0.13261000E-01,
             -0.15927894E+00, 0.72401887E+00, 0.55747724E+00,-0.33719021E+00,
             -0.25143474E+00, },
            { 0.10345604E+01, 0.16206180E+01, 0.10684304E+01,-0.67587040E-01,
              0.19474277E+00, 0.26491645E+00, 0.16167685E-01,-0.36782002E+00,
             -0.53570050E+00, },
            { 0.32261285E+00,-0.64737231E+00,-0.90475298E-01, 0.17187940E+00,
              0.13722499E+00, 0.50946969E+00,-0.93138762E-01,-0.32658824E+00,
             -0.66461273E-01, },
            { 0.22725639E+01, 0.69083470E+00, 0.48083431E+00,-0.49758098E+00,
             -0.84927388E-01, 0.52185225E+00, 0.19340727E+00, 0.32786693E-01,
              0.78221820E-01, },
            {-0.20661349E+01,-0.26972232E+01, 0.25274463E+01,-0.41764119E+00,
              0.75514770E+00,-0.17189139E+01,-0.10526706E+01, 0.63819683E+00,
              0.37426770E+00, },
            {-0.32621279E+01,-0.44142714E+01,-0.25552752E+01,-0.18484338E+00,
             -0.58688587E+00,-0.73055243E+00,-0.17521368E+00, 0.10259843E+01,
              0.11375989E+01, },
            {-0.73171481E-01, 0.42583603E+00,-0.10637463E+01,-0.14348125E+00,
             -0.74651487E-01,-0.45749301E+00, 0.43486070E-01, 0.28321269E+00,
              0.10276017E+00, },
            {-0.18371630E+01, 0.76073569E+00,-0.44730103E+00, 0.35803327E+00,
              0.80352314E-01,-0.40243587E+00,-0.17068449E+00, 0.65292539E-02,
             -0.63528381E-01, },
            { 0.13972975E+01, 0.22084525E+01,-0.22679477E+01, 0.48937282E+00,
             -0.68422282E+00, 0.11014562E+01, 0.64520347E+00,-0.39747542E+00,
             -0.17840618E+00, },
            { 0.23154075E+01, 0.34323814E+01, 0.20233507E+01, 0.38661447E+00,
              0.40558794E+00, 0.58963746E+00, 0.15238740E+00,-0.77325076E+00,
             -0.66182941E+00, },
            { 0.40216600E-02, 0.81703132E-02,-0.19011486E-01,-0.47476284E-01,
             -0.14503382E+00,-0.59457687E-02,-0.60895034E-02, 0.35825928E-02,
             -0.50007328E-02, },
            {-0.10892768E-01,-0.62809291E-03,-0.17020314E-03, 0.16835703E+00,
             -0.23484120E-01, 0.19963286E-02,-0.22885388E-01, 0.47483742E-02,
              0.63064597E-02, },
            { 0.95548488E-01, 0.15581782E+00, 0.46200905E-01, 0.19114015E+00,
             -0.25267342E+00, 0.37239853E-02, 0.66554118E-02, 0.20674267E-02,
             -0.10625846E-01, },
            { 0.15886465E-01,-0.15695851E-01,-0.18671377E-01, 0.40752470E+00,
              0.28736994E+00, 0.15669392E-01,-0.44707544E-01, 0.49465522E-02,
              0.24645288E-01, },
            {-0.16001102E+00, 0.39749850E-01, 0.11758360E+00, 0.14307669E+00,
             -0.21471813E+00, 0.79647778E-02, 0.41321028E-01,-0.32722354E-02,
              0.57371218E-01, },
            {-0.38416374E+00,-0.51390864E-01,-0.17938653E+00, 0.15849835E+00,
              0.80724046E-01,-0.29265678E-02, 0.96612900E-01,-0.67667104E-01,
              0.22370834E-01, },
            {-0.27011070E+00,-0.86970943E+00,-0.26902573E-01,-0.77681959E+00,
              0.69076121E+00, 0.41663114E-01,-0.10570973E+00, 0.84764361E-01,
              0.69376291E-02, },
            { 0.42020214E+00, 0.23183244E+00, 0.38358140E+00,-0.11147493E+01,
             -0.10411042E+01, 0.53920228E-01, 0.16102199E+00,-0.14279890E-01,
             -0.27331126E+00, },
            { 0.14651980E+00,-0.24204558E+00,-0.19398539E+00,-0.30382502E+00,
              0.62044007E+00, 0.47480121E-01,-0.15849266E-01,-0.10659328E-01,
             -0.80701113E-01, },
            { 0.72861421E+00,-0.92455231E-01, 0.29580644E+00,-0.54386604E+00,
             -0.24819843E+00, 0.39313916E-01,-0.49355179E-01, 0.44585172E-01,
             -0.85791960E-01, },
            { 0.11903773E+00, 0.12535146E+01, 0.82356326E-01, 0.67133921E+00,
             -0.63358366E+00,-0.49258057E-01, 0.10747476E+00,-0.10950665E+00,
              0.64388467E-02, },
            {-0.83330280E+00,-0.29580790E+00,-0.38530514E+00, 0.86506933E+00,
              0.80623895E+00,-0.13946185E+00,-0.10462884E+00, 0.11260598E-01,
              0.38791931E+00, },
            {-0.10402253E-01, 0.13346081E-01, 0.63475878E-02,-0.27291005E-01,
             -0.13256154E-02, 0.17733403E-01, 0.11105449E-01,-0.21581961E-03,
             -0.40799612E-03, },
            {-0.13009773E-01,-0.28726665E-01,-0.28361091E-01, 0.41243983E-02,
             -0.13775193E-01,-0.92879981E-02, 0.12643198E-01,-0.23260165E-02,
              0.20666828E-02, },
            { 0.73535770E-01, 0.55688366E-01, 0.99641196E-02,-0.35065774E-01,
              0.18284770E-01, 0.44654209E-01, 0.16372960E-01,-0.12499312E-02,
              0.70338654E-04, },
            { 0.43763254E-01,-0.15740022E-01,-0.13505678E-01,-0.87119667E-02,
              0.10261146E-01,-0.21680510E-01, 0.51671389E-01,-0.18428918E-01,
              0.62357122E-02, },
            {-0.52719351E-01,-0.13156223E+00,-0.79338402E-01, 0.49488008E-01,
              0.24957510E-01,-0.92251182E-01,-0.94008923E-01, 0.44068401E-02,
              0.33692043E-01, },
            {-0.60246535E-01, 0.86852990E-01, 0.20347420E-01,-0.17174080E-02,
              0.17042400E-01, 0.74888632E-01,-0.72979771E-01, 0.60993590E-03,
             -0.11566933E-02, },
            {-0.14518532E-01,-0.12442473E-01, 0.85507222E-02, 0.22372473E-02,
             -0.15677895E-02, 0.13057685E-01,-0.67228526E-02, 0.22530334E-01,
              0.29998149E-02, },
            {-0.19739617E-01, 0.10189696E-01,-0.83538266E-02, 0.88368857E-03,
             -0.46331584E-02, 0.45361207E-03, 0.11164929E-01,-0.60726954E-02,
              0.22200333E-01, },
            { 0.76710410E-01, 0.17538734E-01, 0.20874044E-01,-0.30737719E-01,
             -0.15248102E-01, 0.13212540E-01, 0.14518229E-01, 0.43812137E-01,
              0.32291189E-01, },
            {-0.13844927E-01,-0.34613054E-01, 0.59420760E-02, 0.18336574E-01,
              0.12362840E-01, 0.78165001E-03, 0.94650965E-02,-0.54954968E-01,
              0.48140671E-01, },
            { 0.21998519E-02,-0.30175578E-02,-0.21670593E-02, 0.24198068E-02,
             -0.30340797E-02, 0.98152589E-02,-0.65381783E-02, 0.18372700E-02,
             -0.32104759E-02, },
            { 0.15909625E-01, 0.17480101E-01, 0.73182187E-03, 0.53907954E-02,
             -0.10893663E-01,-0.14815384E-02, 0.32584106E-02, 0.13903450E-02,
             -0.29880791E-02, },
            { 0.18037504E-01,-0.34718488E-02, 0.86077778E-02, 0.51989104E-02,
             -0.11354344E-01, 0.33079886E-02,-0.26848447E-02, 0.73557561E-02,
              0.26780984E-03, },
            { 0.12174977E-02, 0.38145680E-02,-0.32383250E-02, 0.11604379E-01,
              0.10995209E-01,-0.69180368E-02, 0.18747009E-02,-0.57764836E-02,
             -0.31420870E-02, },
         },
         {
            { 0.27280636E+01,-0.75169103E-02, 0.63383449E-02, 0.95558316E-02,
             -0.28542201E-02,-0.33672303E-02, 0.22478853E-02, 0.12390611E-02,
             -0.25694482E-02, },
            {-0.34184477E+00, 0.26675481E-01, 0.36192123E-01, 0.17344089E-01,
             -0.27918063E-01, 0.46071778E-02,-0.19322153E-01, 0.14321898E-02,
             -0.45877192E-02, },
            { 0.11131347E+01, 0.26039990E-01,-0.86778462E-01,-0.36485463E-01,
              0.39370175E-01, 0.96002901E-02, 0.23496447E-01, 0.10478078E-01,
              0.15071011E-01, },
            { 0.26609913E+00,-0.37570027E+00,-0.76007955E-01,-0.84936224E-01,
              0.13073616E+00,-0.46946578E-01, 0.45213178E-01,-0.10088617E-01,
              0.15226900E-01, },
            {-0.13583761E+01, 0.24481584E+00, 0.19391088E+00,-0.25062395E-01,
             -0.17285113E+00, 0.35910089E-01,-0.83230503E-01,-0.49273174E-01,
             -0.39427180E-01, },
            {-0.86215436E-02, 0.33480579E+00, 0.59670329E-01, 0.72379611E-01,
             -0.10527881E+00, 0.40745031E-01,-0.27480911E-01, 0.10371030E-01,
             -0.11401369E-01, },
            { 0.24424902E+00,-0.25925013E+00,-0.73127270E-01, 0.64353183E-01,
              0.13510229E+00,-0.47481738E-01, 0.65662742E-01, 0.38441211E-01,
              0.27695235E-01, },
            { 0.26992045E-01,-0.27138302E+00,-0.25009316E+00, 0.13358043E-01,
             -0.99321315E-02,-0.14139374E-02,-0.65408170E-03, 0.11991658E-02,
              0.26755203E-02, },
            {-0.18171994E-01, 0.28280777E+00,-0.23699513E+00,-0.49411627E-02,
              0.77506038E-02, 0.32914598E-02,-0.10480203E-01, 0.53377454E-02,
              0.11933497E-01, },
            { 0.55335838E-01,-0.93998998E-01,-0.40524747E-01,-0.86126318E-02,
              0.32151006E-01,-0.25418943E-01,-0.56042250E-01, 0.22723371E-01,
              0.48075102E-01, },
            { 0.21400840E-02, 0.54945960E-02,-0.11576591E+00, 0.33427548E-01,
              0.10076365E-01,-0.27823556E-01, 0.87937471E-02, 0.16767824E-01,
              0.51116429E-01, },
            {-0.27947503E+00, 0.93270457E+00, 0.78121406E+00,-0.63873820E-01,
              0.13776053E-01,-0.41885123E-01,-0.14760400E-01, 0.49246795E-01,
             -0.38276151E-01, },
            {-0.38543758E+00,-0.12443132E+01, 0.63409525E+00, 0.10198897E+00,
             -0.18275555E-01,-0.13154987E+00, 0.44033036E-01,-0.12012166E+00,
             -0.11587386E+00, },
            {-0.24930301E+00, 0.41436785E+00,-0.42938063E+00, 0.14058124E+00,
             -0.51638871E+00, 0.26499906E+00, 0.30564678E+00,-0.35653755E-01,
             -0.30761656E+00, },
            { 0.96339178E+00, 0.17102709E+01, 0.64794040E+00,-0.14055821E+00,
              0.10397163E+00, 0.27128929E+00,-0.49434416E-01,-0.15722241E+00,
             -0.39522058E+00, },
            { 0.71117145E+00,-0.79698235E+00, 0.35370767E+00, 0.88413291E-01,
              0.65246043E-02, 0.25285155E+00, 0.90608612E-01,-0.12623823E+00,
              0.72437324E-01, },
            { 0.16189957E+01, 0.92443705E+00,-0.42564002E+00,-0.32307896E+00,
             -0.47557414E-01, 0.47201774E+00,-0.36377423E-01, 0.27063537E+00,
              0.27386218E+00, },
            { 0.49492338E+00,-0.11964587E+01, 0.11970491E+01,-0.51224530E+00,
              0.14279464E+01,-0.65490454E+00,-0.49377796E+00,-0.79987206E-01,
              0.57029313E+00, },
            {-0.29584346E+01,-0.44350519E+01,-0.13688515E+01, 0.90590253E-01,
             -0.28306559E+00,-0.70212859E+00,-0.48325151E-01, 0.42926690E+00,
              0.80071324E+00, },
            {-0.36841384E+00, 0.36017466E+00,-0.10382642E+01,-0.75529769E-01,
             -0.65040000E-01,-0.29723659E+00,-0.82611166E-01, 0.96393012E-01,
             -0.45457043E-01, },
            {-0.13651522E+01, 0.12593678E+00, 0.79192951E-01, 0.25124004E+00,
              0.63250221E-01,-0.38624293E+00, 0.94562238E-02,-0.15385661E+00,
             -0.18089309E+00, },
            {-0.32429627E+00, 0.10307864E+01,-0.13849024E+01, 0.49472693E+00,
             -0.10159550E+01, 0.43086207E+00, 0.25678623E+00, 0.93890756E-01,
             -0.33852810E+00, },
            { 0.21680369E+01, 0.36000590E+01, 0.99755543E+00, 0.11148506E+00,
              0.17308816E+00, 0.52738160E+00, 0.95052890E-01,-0.32424158E+00,
             -0.48464921E+00, },
            {-0.40881429E-02, 0.31389890E-02, 0.18125862E-02, 0.19379327E-01,
             -0.11340187E+00,-0.64218352E-02,-0.21440715E-02, 0.86666498E-03,
             -0.48338207E-02, },
            {-0.12114726E-01,-0.90989377E-02, 0.15213010E-01, 0.11901259E+00,
              0.44593573E-01, 0.35579519E-02,-0.16553327E-01, 0.31294818E-02,
              0.35295503E-02, },
            { 0.96299417E-01, 0.16172042E+00, 0.36571477E-01, 0.13853014E+00,
             -0.11524634E+00, 0.13647977E-01, 0.85650487E-02, 0.21059543E-02,
             -0.88550262E-02, },
            { 0.78764156E-01, 0.23560308E-01,-0.20360565E-01, 0.20476563E+00,
              0.23432060E+00, 0.29904551E-02,-0.21854280E-01, 0.12916533E-02,
              0.17185966E-01, },
            {-0.13958269E+00, 0.61252635E-01,-0.22732537E-01,-0.22977982E+00,
             -0.42527966E-01, 0.48952647E-01, 0.13864973E-01,-0.26283702E-01,
              0.68163075E-01, },
            {-0.25024179E+00,-0.60532674E-01,-0.19260155E+00, 0.39119691E-01,
             -0.26626396E+00,-0.44200186E-01, 0.90043500E-01,-0.57880651E-01,
              0.20709954E-01, },
            {-0.24203920E+00,-0.77736235E+00,-0.10913799E-01,-0.15193807E+00,
              0.20603824E+00,-0.31180762E-01,-0.11680545E+00, 0.67744017E-01,
              0.44562139E-01, },
            { 0.29965173E-01, 0.71288981E-01, 0.24611737E+00,-0.42808515E+00,
             -0.44561577E+00, 0.12265429E+00, 0.95303804E-01, 0.38496863E-01,
             -0.18627241E+00, },
            { 0.16500773E+00,-0.28346860E+00, 0.19127062E-01, 0.85767560E-01,
              0.18231672E+00,-0.66293105E-01, 0.29998733E-01, 0.59395704E-01,
             -0.82480334E-01, },
            { 0.46248379E+00,-0.12290555E+00, 0.23955330E+00,-0.17114018E+00,
              0.85420050E-01, 0.51798742E-01,-0.63089624E-01, 0.69193467E-01,
             -0.49095791E-01, },
            { 0.49044739E-01, 0.10698808E+01, 0.55030648E-01,-0.20411465E-01,
             -0.20164534E+00, 0.81138603E-01, 0.11107465E+00,-0.91995150E-01,
             -0.63594230E-01, },
            {-0.31178156E+00, 0.25681297E-02,-0.21641961E+00, 0.26155630E+00,
              0.19081634E+00,-0.18793817E+00,-0.63471727E-01,-0.60597725E-01,
              0.24745049E+00, },
            {-0.20798016E-01, 0.15350298E-01, 0.10828707E-01,-0.18462822E-01,
             -0.11293224E-01, 0.18085541E-01,-0.18076560E-01,-0.26808286E-02,
             -0.14639653E-02, },
            {-0.94376355E-02,-0.31904358E-01,-0.19534955E-01, 0.60758656E-02,
             -0.13963651E-01, 0.17930444E-01, 0.15960965E-01,-0.58656308E-03,
             -0.17829806E-02, },
            { 0.50320923E-01, 0.85855640E-01,-0.37249629E-03,-0.27981598E-01,
              0.27195103E-02, 0.81393354E-01, 0.11739279E-01,-0.58815950E-02,
             -0.22160930E-02, },
            { 0.30873053E-01,-0.98756980E-02, 0.89320733E-03,-0.10296550E-01,
             -0.80322934E-03,-0.16821079E-01, 0.74325636E-01,-0.80107599E-02,
              0.32815721E-02, },
            {-0.20960687E-01,-0.15283193E+00,-0.57994306E-01, 0.26183322E-01,
              0.26019171E-01,-0.12097655E+00,-0.35915900E-01,-0.33830630E-02,
              0.31820308E-01, },
            {-0.43779407E-01, 0.92425197E-01,-0.36082782E-02, 0.10040569E-01,
              0.69529493E-02, 0.14791758E-01,-0.99151865E-01,-0.59039001E-02,
             -0.60759163E-02, },
            {-0.11967236E-01,-0.15639639E-01, 0.11332542E-01, 0.19568345E-03,
             -0.19746975E-02, 0.52477429E-02,-0.68195597E-02, 0.18582528E-01,
              0.10505051E-01, },
            {-0.10374394E-01, 0.77973618E-02,-0.14768258E-01, 0.33124851E-02,
             -0.50527342E-02, 0.13333512E-02, 0.80117164E-02,-0.10721104E-01,
              0.16201984E-01, },
            { 0.51714681E-01, 0.20969117E-01, 0.20908019E-01,-0.21666186E-01,
             -0.65546874E-02, 0.10858584E-01, 0.19092079E-01, 0.34162614E-01,
              0.63728623E-01, },
            {-0.17780149E-01,-0.39414037E-01, 0.19805619E-01, 0.27769320E-02,
              0.32051459E-02,-0.13837150E-02, 0.38510279E-02,-0.78942724E-01,
              0.31475380E-01, },
            { 0.72736945E-02,-0.99098906E-02, 0.94700296E-03, 0.49195858E-02,
             -0.48104613E-02, 0.46512815E-02,-0.50494503E-02, 0.46232669E-02,
             -0.99206646E-03, },
            { 0.11354908E-01, 0.14260411E-01,-0.41098539E-02, 0.63321404E-02,
             -0.46809567E-02, 0.28002951E-02, 0.52125915E-02, 0.95991232E-03,
              0.42521297E-02, },
            { 0.14690466E-01, 0.30270962E-02, 0.53093689E-02, 0.74441070E-02,
             -0.92850029E-02, 0.12141704E-02,-0.84077264E-03, 0.41119582E-02,
              0.30829709E-02, },
            {-0.49859304E-02,-0.33042641E-03,-0.43395390E-02, 0.45590200E-02,
              0.91798063E-02,-0.12648865E-02, 0.12946193E-02,-0.44219219E-02,
              0.19899376E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir17F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.47669554E+01,-0.72017536E-01, 0.13942049E+00, 0.19542389E-01,
             -0.16752634E-01,-0.16510993E-01,-0.43158536E-03, 0.28941497E-01,
             -0.32931969E-01, 0.10719537E-02, 0.56658965E-02, 0.12240363E-01,
             -0.93685742E-02, },
            {-0.99394917E+00,-0.35927990E+00, 0.43325570E+00,-0.22593535E+00,
              0.49836481E+00,-0.91735348E-01, 0.10013926E+00,-0.16090275E+00,
             -0.16553175E+00,-0.32432500E-01,-0.38297892E-01,-0.82239382E-01,
             -0.42720828E-02, },
            { 0.12144961E+02, 0.79801214E+00,-0.25601573E+01,-0.68445534E-01,
              0.18775876E+00, 0.16825244E+00, 0.18251759E+00,-0.42203125E+00,
              0.31813495E-01, 0.75639360E-01,-0.42337403E+00,-0.16204225E+00,
              0.93350336E-01, },
            { 0.12359418E+02, 0.26116188E+01,-0.71795487E+00, 0.47592411E+01,
             -0.74272795E+01, 0.17137471E+01,-0.29525971E+01, 0.15546530E+01,
              0.12446737E+01, 0.55519021E+00,-0.42093495E+00, 0.19277966E+01,
              0.44809118E+00, },
            {-0.81837517E+02,-0.65713782E+01, 0.11112719E+02, 0.46813384E+00,
             -0.61830668E+01,-0.12024459E+01,-0.17724164E+01, 0.16484222E+01,
              0.20026155E+01,-0.97772136E-01, 0.29807785E+01, 0.10582246E+01,
             -0.70731813E+00, },
            {-0.72209411E+01,-0.77260118E+01,-0.15136260E+02,-0.22104202E+02,
              0.39128777E+02,-0.83586140E+01, 0.14526752E+02,-0.68457294E+01,
             -0.29666450E+01,-0.23397076E+01, 0.42374215E+01,-0.12021458E+02,
             -0.40962782E+01, },
            { 0.19595665E+03, 0.24501345E+02,-0.20282061E+02,-0.20203459E+01,
              0.23160069E+02, 0.31945641E+01, 0.48780818E+01,-0.33812382E+01,
             -0.74402051E+01,-0.85799468E+00,-0.81885357E+01,-0.29702859E+01,
              0.22015822E+01, },
            {-0.64242828E+02, 0.68489776E+01, 0.53867455E+02, 0.37935581E+02,
             -0.88741364E+02, 0.16450567E+02,-0.26521566E+02, 0.13563215E+02,
              0.29131477E+01, 0.35748591E+01,-0.11355737E+02, 0.28946178E+02,
              0.11574268E+02, },
            {-0.21447914E+03,-0.34901211E+02, 0.17094313E+02, 0.27981231E+01,
             -0.29804564E+02,-0.32461698E+01,-0.52834835E+01, 0.37156496E+01,
              0.90832834E+01, 0.20808840E+01, 0.93321514E+01, 0.36256206E+01,
             -0.27270098E+01, },
            { 0.11643464E+03, 0.41456490E+01,-0.62626999E+02,-0.26506950E+02,
              0.89930573E+02,-0.14143736E+02, 0.20422012E+02,-0.11808813E+02,
             -0.10933266E+01,-0.21905138E+01, 0.12235158E+02,-0.29777472E+02,
             -0.13072017E+02, },
            { 0.86740616E+02, 0.16090628E+02,-0.54182315E+01,-0.11867726E+01,
              0.12633715E+02, 0.11003165E+01, 0.20218921E+01,-0.16267680E+01,
             -0.36661091E+01,-0.12221850E+01,-0.37074435E+01,-0.15805085E+01,
              0.11606567E+01, },
            {-0.55534855E+02,-0.55097966E+01, 0.24141411E+02, 0.61227908E+01,
             -0.33446136E+02, 0.44106121E+01,-0.55466518E+01, 0.36874857E+01,
              0.85909329E-01, 0.44831970E+00,-0.46812630E+01, 0.11031492E+02,
              0.51603775E+01, },
            {-0.17869551E+00, 0.13910967E+01, 0.19192034E+01, 0.11573006E-01,
             -0.15976201E+00,-0.94560003E-02, 0.15693951E-01, 0.52722126E-01,
             -0.85014366E-02,-0.97233402E-02, 0.14873112E-02,-0.89176523E-03,
              0.52959193E-02, },
            { 0.14833479E+00,-0.18277854E+01, 0.14350961E+01, 0.30985156E-01,
              0.15200675E+00, 0.21492884E-01,-0.96385013E-02,-0.21557527E-01,
              0.68189330E-01, 0.68523240E-03,-0.16917478E-02,-0.20286474E-01,
              0.11845784E-01, },
            {-0.40418360E+00, 0.81630504E+00,-0.15781780E+01, 0.64889038E+00,
             -0.37128392E+00, 0.77008136E-01, 0.69604263E-01,-0.34736013E+00,
              0.32418558E+00, 0.33138849E-01, 0.30220786E+00,-0.44437163E-01,
              0.15520661E+00, },
            {-0.41341978E+00, 0.15498554E+01,-0.13387534E+00, 0.92664844E+00,
             -0.41743796E-01, 0.40571007E+00,-0.87894253E-01,-0.86675435E-01,
              0.41113243E+00, 0.72546639E-01, 0.17126939E+00, 0.89121640E-01,
              0.15699174E+00, },
            { 0.41381030E+01, 0.69626184E+01, 0.53660226E+01, 0.43502375E+00,
              0.14352136E+01, 0.10049860E+01,-0.11870514E+01,-0.10593939E+01,
              0.46142328E+00, 0.64426951E-01, 0.14919201E+00, 0.15421334E+00,
             -0.24009842E+00, },
            {-0.21884575E+01,-0.73620200E+01, 0.25206361E+01, 0.66348322E-01,
             -0.16518012E+01,-0.94297117E+00, 0.68582207E+00, 0.67841366E-01,
             -0.10439430E+01, 0.77655651E-01, 0.36230335E+00, 0.63397574E+00,
             -0.54638183E+00, },
            { 0.53683534E+01,-0.93815575E+01, 0.17646755E+02,-0.46907468E+01,
              0.32573049E+01, 0.13132706E+00,-0.36337693E+01, 0.69651113E+01,
             -0.32140777E+01, 0.10166423E+01,-0.68326788E+01,-0.89097261E+00,
             -0.18256758E+01, },
            { 0.64158630E+01,-0.17649351E+02, 0.56288085E+01,-0.14704763E+02,
             -0.18327432E+01,-0.44618912E+01, 0.16981554E+01, 0.28276656E+01,
             -0.73030243E+01,-0.36753397E+01,-0.28609183E+01,-0.84641004E+00,
             -0.27507968E+01, },
            {-0.22282154E+02,-0.46678989E+02,-0.15916713E+02, 0.96986282E+00,
             -0.12215256E+02,-0.11022174E+02, 0.93760414E+01, 0.61606741E+01,
             -0.46595785E+00,-0.15873922E+01,-0.12295980E+01,-0.86355746E+00,
              0.31697853E+01, },
            { 0.11682695E+02, 0.20335405E+02,-0.30051084E+01,-0.24479063E+01,
              0.95660801E+01, 0.90430307E+01,-0.53547616E+01,-0.27718673E+01,
              0.75180225E+01, 0.14638613E+00,-0.37047281E+01,-0.48737941E+01,
              0.41346002E+01, },
            {-0.28579910E+02, 0.87348892E+02,-0.86842323E+02, 0.14033395E+02,
             -0.40439680E-01, 0.79252595E+00, 0.26493780E+02,-0.41706631E+02,
              0.64251800E+01,-0.86240749E+01, 0.41827084E+02, 0.86550608E+01,
              0.81083364E+01, },
            {-0.62369034E+02, 0.10966636E+03, 0.57230368E+01, 0.72633636E+02,
              0.11584758E+02, 0.18804466E+02,-0.13280591E+02,-0.17550068E+02,
              0.34799728E+02, 0.25178934E+02, 0.15181874E+02, 0.13891364E+01,
              0.13887790E+02, },
            { 0.38604603E+02, 0.10207992E+03,-0.29230875E+02,-0.10485903E+02,
              0.40149216E+02, 0.39026348E+02,-0.27193357E+02,-0.14548298E+02,
             -0.53476868E+01, 0.69869070E+01, 0.42715588E+01, 0.28882394E+01,
             -0.10630176E+02, },
            {-0.17743690E+02, 0.38365269E+02,-0.44383980E+02, 0.14417901E+02,
             -0.26099285E+02,-0.29533575E+02, 0.11660138E+02, 0.13445920E+02,
             -0.22865898E+02,-0.37443502E+01, 0.11432396E+02, 0.13121328E+02,
             -0.13328956E+02, },
            { 0.83295593E+02,-0.27551205E+03, 0.20297762E+03,-0.27005564E+02,
             -0.36729797E+02,-0.95780029E+01,-0.72624352E+02, 0.10681412E+03,
              0.42937274E+01, 0.23383394E+02,-0.10540522E+03,-0.25344908E+02,
             -0.16387964E+02, },
            { 0.19820599E+03,-0.31570514E+03,-0.72060852E+02,-0.15624391E+03,
             -0.26451447E+02,-0.37654526E+02, 0.41089275E+02, 0.39953007E+02,
             -0.67739883E+02,-0.63928482E+02,-0.35235271E+02, 0.25527909E+01,
             -0.27388004E+02, },
            {-0.16694485E+02,-0.10372476E+03, 0.99343521E+02, 0.17906614E+02,
             -0.54899918E+02,-0.54147346E+02, 0.33478874E+02, 0.15286818E+02,
              0.11120189E+02,-0.10826112E+02,-0.68150578E+01,-0.49306626E+01,
              0.13716017E+02, },
            { 0.77929759E+01,-0.13942567E+03, 0.99441032E+02,-0.28234491E+02,
              0.29203568E+02, 0.39281445E+02,-0.71675448E+01,-0.22050737E+02,
              0.30663433E+02, 0.86184502E+01,-0.13269634E+02,-0.14924605E+02,
              0.19169945E+02, },
            {-0.11517070E+03, 0.34904071E+03,-0.23037421E+03, 0.29416445E+02,
              0.71594467E+02, 0.19957907E+02, 0.86077232E+02,-0.12215864E+03,
             -0.19588717E+02,-0.26783237E+02, 0.11752483E+03, 0.30323790E+02,
              0.14876017E+02, },
            {-0.23966219E+03, 0.41143326E+03, 0.11223404E+03, 0.15691782E+03,
              0.27349728E+02, 0.35871052E+02,-0.54274956E+02,-0.36422298E+02,
              0.56854435E+02, 0.68983749E+02, 0.36825100E+02,-0.73194642E+01,
              0.21632607E+02, },
            {-0.64012694E+01, 0.41037807E+02,-0.62049179E+02,-0.87074289E+01,
              0.26557791E+02, 0.25656239E+02,-0.14669993E+02,-0.59758453E+01,
             -0.57773538E+01, 0.55395350E+01, 0.37282646E+01, 0.29195380E+01,
             -0.61014247E+01, },
            { 0.11938375E+01, 0.93779282E+02,-0.58239254E+02, 0.16913059E+02,
             -0.99648829E+01,-0.18439867E+02, 0.14224555E-02, 0.11951701E+02,
             -0.14707023E+02,-0.52538481E+01, 0.50785527E+01, 0.62320628E+01,
             -0.98415794E+01, },
            { 0.59114883E+02,-0.15509395E+03, 0.99913818E+02,-0.13110655E+02,
             -0.38926403E+02,-0.11956065E+02,-0.37092178E+02, 0.51071323E+02,
              0.12083856E+02, 0.11144695E+02,-0.47916153E+02,-0.12860417E+02,
             -0.49043174E+01, },
            { 0.97261864E+02,-0.19500647E+03,-0.51127705E+02,-0.60541309E+02,
             -0.11995492E+02,-0.12932009E+02, 0.25311089E+02, 0.10825487E+02,
             -0.16840401E+02,-0.26814421E+02,-0.14202532E+02, 0.40943551E+01,
             -0.52332826E+01, },
            { 0.59779093E-01, 0.60269125E-01, 0.43412182E-02,-0.86104780E+00,
             -0.28442442E+00, 0.38802601E-01,-0.24069672E-01, 0.61457353E-02,
              0.26447732E-01,-0.10484729E-01,-0.16964691E-01,-0.17687945E-01,
              0.11072901E-01, },
            { 0.88018358E-01,-0.24598600E-01,-0.90877013E-03, 0.27645341E+00,
             -0.60979271E+00,-0.11868040E-01,-0.52261841E-02,-0.38976759E-01,
              0.42610688E-03, 0.50539705E-02,-0.27756458E-02,-0.37858905E-02,
              0.15917744E-01, },
            { 0.24479559E-01, 0.64580876E+00, 0.88117726E-01,-0.19167215E+00,
             -0.63693744E+00, 0.20008095E+00, 0.13274206E+00, 0.14893281E+00,
              0.11748412E+00, 0.23986910E+00,-0.11419377E+00, 0.55741396E-01,
              0.13194546E+00, },
            { 0.31363299E+00, 0.28544968E+00, 0.47836196E+00, 0.13769221E+01,
             -0.44855464E+00, 0.14112352E+00, 0.25663987E+00, 0.10438135E+00,
             -0.10474421E+00, 0.16550185E-01,-0.23519803E-01, 0.12023649E+00,
              0.87688930E-01, },
            { 0.15380675E+01, 0.14324548E+01, 0.13127990E+01, 0.88563509E+01,
              0.12689075E+01,-0.85907197E+00,-0.37547376E-01,-0.15601891E+00,
              0.33903081E-01, 0.14706784E+00, 0.37223580E+00, 0.14558417E+00,
             -0.20162193E+00, },
            { 0.20672157E+01, 0.85868543E+00, 0.20100529E+01,-0.15593593E+01,
              0.54521675E+01, 0.74721974E+00, 0.24748008E+00, 0.22941186E+00,
              0.41224465E+00, 0.10947402E+00, 0.19377372E+00, 0.53652622E-01,
             -0.26892835E+00, },
            {-0.85527790E+00,-0.59873829E+01, 0.82442212E+00, 0.72160811E+01,
             -0.55486888E+00,-0.11447293E+01,-0.21035647E+01,-0.41455364E+00,
             -0.13146853E+01,-0.19809873E+01, 0.17242732E+01,-0.57977825E+00,
             -0.89161646E+00, },
            {-0.37543094E+01,-0.52248607E+01,-0.48182120E+01,-0.46627312E+01,
              0.96200695E+01, 0.14079694E+01,-0.24980745E+01,-0.34461054E+00,
              0.31529835E+00,-0.14311078E+01, 0.84543675E+00,-0.79500079E+00,
              0.87758198E-01, },
            {-0.14999284E+02,-0.91342783E+01,-0.57922735E+01,-0.19058977E+02,
              0.25543198E+00, 0.45159869E+01, 0.29069436E+01, 0.59061384E+00,
             -0.35207078E+00,-0.85799408E+00,-0.22677574E+01,-0.35971817E+00,
              0.99695718E+00, },
            {-0.13004522E+02,-0.14058595E+01,-0.92240829E+01, 0.49196196E+01,
             -0.47094283E+01,-0.47235041E+01,-0.11311457E+01,-0.21206169E+01,
             -0.10868499E+01,-0.12165101E+01,-0.93641424E+00,-0.54299787E-01,
              0.10610462E+01, },
            { 0.39302894E+00, 0.14020730E+02,-0.63046670E+01,-0.23180601E+02,
              0.46417074E+01, 0.19860531E+01, 0.61314015E+01, 0.10478191E+01,
              0.39981802E+01, 0.44316478E+01,-0.51703086E+01, 0.15147123E+01,
              0.18553241E+01, },
            { 0.10761917E+02, 0.15200618E+02, 0.10380513E+02, 0.55815134E+01,
             -0.28787245E+02,-0.61082196E+01, 0.76987019E+01, 0.63724285E+00,
             -0.10391539E+00, 0.52535143E+01,-0.30186067E+01, 0.21368234E+01,
             -0.69694448E+00, },
            { 0.33612606E+02, 0.17597853E+02, 0.12092536E+02, 0.56274945E+00,
             -0.44205179E+01,-0.89462976E+01,-0.83171101E+01,-0.39124992E+00,
              0.78949368E+00, 0.13598933E+01, 0.44765353E+01, 0.34951830E+00,
             -0.16460216E+01, },
            { 0.24521830E+02,-0.49467854E+01, 0.14729335E+02,-0.94777203E+01,
             -0.24623814E+02, 0.99676733E+01, 0.11067480E+01, 0.53398237E+01,
              0.20645846E+00, 0.34774735E+01, 0.12603258E+01,-0.11288598E+00,
             -0.20366173E+01, },
            { 0.48746821E+00,-0.93833942E+01, 0.79512820E+01, 0.16153667E+02,
             -0.65026054E+01,-0.10142500E+01,-0.47579007E+01,-0.93598449E+00,
             -0.33259747E+01,-0.30019803E+01, 0.39455864E+01,-0.10857137E+01,
             -0.10412169E+01, },
            {-0.85157776E+01,-0.12449425E+02,-0.60487118E+01, 0.28868362E+00,
              0.19360544E+02, 0.52751713E+01,-0.72028527E+01,-0.67789567E+00,
             -0.26337564E+00,-0.44146733E+01, 0.24460630E+01,-0.18024994E+01,
              0.45497325E+00, },
            {-0.22045063E+02,-0.11040361E+02,-0.89635525E+01, 0.14214537E+02,
              0.54145708E+01, 0.58426008E+01, 0.63665800E+01,-0.24814361E+00,
             -0.69569623E+00,-0.63997215E+00,-0.28150806E+01, 0.38981918E-01,
              0.74708533E+00, },
            {-0.14405212E+02, 0.67465463E+01,-0.91972771E+01, 0.44302187E+01,
              0.29674311E+02,-0.69829130E+01, 0.78397954E+00,-0.36686268E+01,
              0.63624251E+00,-0.25999174E+01,-0.36813948E+00, 0.81634127E-01,
              0.15097922E+01, },
            { 0.19783931E-01, 0.11059830E+00,-0.11023654E+00,-0.75270355E-01,
              0.54907449E-01,-0.38886008E+00,-0.41173506E+00,-0.22116266E-01,
              0.64385377E-01, 0.88824835E-02,-0.96755140E-02,-0.36646105E-01,
             -0.59598195E-02, },
            { 0.64539403E-01,-0.11785930E+00,-0.12389943E+00,-0.48941523E-01,
             -0.10718920E+00, 0.40435502E+00,-0.40478399E+00,-0.61611693E-01,
              0.38958149E-03, 0.99619920E-03,-0.42985836E-02, 0.18320024E-02,
              0.19367568E-01, },
            {-0.32544461E+00, 0.51384830E+00,-0.25873068E+00,-0.29849005E+00,
             -0.10000073E+00,-0.15430472E-01, 0.58037211E-03, 0.10718274E+00,
             -0.16415427E-01,-0.15686333E-01, 0.29702716E-01,-0.29885124E-01,
             -0.15864320E-01, },
            {-0.14868063E+00,-0.11868401E+00,-0.14044303E+00,-0.17538846E+00,
             -0.13476025E+00, 0.57173114E-01, 0.18623371E+00,-0.32387108E-01,
             -0.34499161E-01,-0.32388661E-01,-0.80245815E-01, 0.56267913E-01,
              0.15138485E-01, },
            {-0.47861069E-01,-0.77162355E+00, 0.16360250E+01, 0.50554228E+00,
             -0.63924365E-01,-0.10863320E+00,-0.60092205E+00, 0.49819637E-01,
             -0.82271285E-02, 0.98940387E-01,-0.35185188E-01, 0.12983482E+00,
             -0.19542508E+00, },
            {-0.76338345E+00, 0.19630162E+01,-0.10691432E+00,-0.60503505E-01,
              0.13289427E+01, 0.43475190E+00, 0.11329929E+00,-0.10821795E-01,
             -0.62644839E-01, 0.12987889E+00, 0.13167375E+00, 0.77864654E-01,
             -0.24797525E+00, },
            { 0.24173480E+00,-0.10378166E+01, 0.63027930E+00, 0.74880987E+00,
              0.33605361E+00, 0.35392138E+00, 0.33974937E+00,-0.25777468E+00,
              0.61715249E-03, 0.44684775E-01,-0.11695752E+00, 0.11965482E+00,
              0.14105978E-01, },
            {-0.24457288E+00, 0.41991296E+00, 0.31134152E+00, 0.22164924E+00,
              0.72564882E+00,-0.43179068E+00,-0.61885383E-01, 0.49443066E-01,
              0.54404356E-01, 0.31738911E-01, 0.19574244E+00, 0.27847867E-02,
             -0.16355213E-01, },
            { 0.63239706E+00, 0.16469691E+01,-0.27302732E+01,-0.85581708E+00,
              0.37954584E+00, 0.11399052E+01, 0.17348310E+01, 0.94211541E-01,
             -0.21935873E+00,-0.23572168E+00, 0.16577183E+00,-0.19861729E+00,
              0.35016689E+00, },
            { 0.68477434E+00,-0.23878093E+01, 0.44941813E+00,-0.84276356E-01,
             -0.20325942E+01,-0.13022912E+01, 0.10930405E+01, 0.38093248E+00,
              0.19024202E+00,-0.33430806E+00,-0.23511814E+00,-0.28587615E+00,
              0.32850736E+00, },
            { 0.67689978E-01,-0.20780521E-01,-0.29032283E-01,-0.13323180E-01,
              0.28254269E-01, 0.84414100E-02,-0.15326820E-01, 0.25174028E+00,
             -0.40093143E-01, 0.27145154E-02,-0.66505717E-02,-0.13159065E-01,
             -0.18460391E-01, },
            { 0.69634892E-01,-0.45518778E-01, 0.13114978E-03,-0.10398069E-01,
             -0.32790072E-01,-0.10675941E-01,-0.56565069E-02, 0.37150957E-01,
              0.23021942E+00, 0.76370039E-02,-0.43002558E-02, 0.90373280E-02,
             -0.44722832E-02, },
            {-0.81032980E-02,-0.43692399E-01,-0.10039860E-01, 0.17762598E-01,
             -0.13719068E+00,-0.41710932E-01, 0.31929690E-01, 0.24983566E+00,
              0.21474642E+00,-0.19966746E-02, 0.15482435E-02,-0.80155046E-03,
              0.30076524E-01, },
            { 0.23039341E+00, 0.19240947E+00,-0.78822196E-01,-0.48037753E-01,
              0.23808508E+00,-0.85829999E-02,-0.15761973E-01,-0.27143875E+00,
              0.14898765E+00, 0.10034035E-01,-0.17262492E-01, 0.28459651E-02,
             -0.29851828E-01, },
            {-0.92098333E-01,-0.30134434E-01,-0.82727504E-03,-0.17205035E-01,
             -0.25473502E-01, 0.82982341E-02,-0.11164435E-01, 0.28170865E-01,
             -0.21050289E-01, 0.71516864E-01, 0.11752138E+00,-0.58911447E-02,
             -0.18760277E-01, },
            { 0.83058653E-02, 0.17495196E-01,-0.28589493E-01, 0.34168422E-01,
             -0.72476611E-01,-0.29382722E-02, 0.76811700E-02, 0.45096446E-01,
              0.30133612E-02,-0.14231350E+00, 0.89629523E-01, 0.86848531E-02,
             -0.99676484E-02, },
            {-0.47373831E-01,-0.30634219E-01, 0.16800816E-02, 0.18063534E-01,
             -0.41148610E-01, 0.12769086E-01,-0.67251893E-02, 0.29845899E-01,
              0.57053417E-02,-0.11818508E-02,-0.55645364E-02,-0.10559253E+00,
              0.37416615E-01, },
            { 0.55516444E-01, 0.15834859E-01, 0.20975683E-01, 0.30448465E-01,
              0.37793189E-01, 0.77342098E-02,-0.49891463E-02,-0.35565118E-02,
             -0.14904852E-01,-0.27537171E-01,-0.54248427E-02,-0.32515921E-01,
             -0.11524391E+00, },
            {-0.27006501E-01,-0.25322475E-01, 0.14191590E-01, 0.12788468E-01,
              0.57791113E-02,-0.54317171E-03,-0.15069270E-02, 0.22591196E-01,
             -0.38332655E-02,-0.25445297E-01, 0.13055151E-01, 0.71395235E-02,
              0.97428244E-02, },
            { 0.17405979E-01,-0.11919928E-01,-0.45696460E-02,-0.54349615E-02,
             -0.26084450E-02,-0.11787944E-01,-0.67170402E-02,-0.12048977E-01,
              0.12912758E-01,-0.51113814E-02, 0.12004339E-02, 0.12741836E-01,
              0.13009419E-01, },
            { 0.38855115E-02, 0.78748085E-01,-0.10889452E+00,-0.12783662E-01,
             -0.66305078E-02,-0.20848872E-01, 0.30470356E-01,-0.30254409E-01,
             -0.69153430E-02,-0.26659120E-01, 0.13413356E-01, 0.29060601E-02,
             -0.56839683E-02, },
            {-0.37495904E-01, 0.66720471E-02,-0.68216706E-02, 0.11667379E-01,
              0.37586469E-01, 0.30343615E-01,-0.39814594E-02,-0.21826498E-01,
              0.24977464E-01, 0.14075888E-01,-0.11293185E-01,-0.50641843E-02,
              0.14215916E-01, },
         },
         {
            { 0.80071468E+01, 0.23540020E-01, 0.14707807E+00, 0.61012927E-01,
             -0.19648800E-01,-0.79102488E-02,-0.12570567E-01, 0.60084999E-01,
             -0.31929113E-01, 0.15448347E-01,-0.50893174E-02, 0.16869301E-01,
             -0.57898671E-02, },
            {-0.21232369E+01,-0.46706444E+00,-0.10000438E+00,-0.10564299E+00,
              0.28760383E+00, 0.85470997E-01, 0.23103155E+00, 0.48029888E-02,
             -0.14235696E+00, 0.14473069E+00, 0.36030032E-01,-0.35992347E-01,
             -0.17253455E-01, },
            { 0.16923178E+02, 0.12374104E+01,-0.25139482E+01, 0.29843819E+00,
              0.55985516E+00, 0.69062692E+00, 0.48571926E+00,-0.87643987E+00,
              0.75207931E+00, 0.19640107E-01,-0.28679374E+00,-0.58915937E+00,
             -0.28327841E+00, },
            { 0.14983639E+02, 0.24837055E+01, 0.10404690E+02, 0.21364896E+01,
             -0.29136975E+01,-0.39514872E+00,-0.35909803E+01, 0.20056395E+00,
              0.15654199E+01,-0.34472339E+01,-0.59020323E+00, 0.87846285E+00,
              0.14721909E+01, },
            {-0.94911194E+02,-0.13008690E+02, 0.14315582E+02,-0.42997451E+01,
             -0.10056245E+02,-0.40604191E+01,-0.27439721E+01, 0.63601122E+01,
             -0.44736714E+01,-0.78184110E+00, 0.28047204E+01, 0.52159157E+01,
              0.21123674E+01, },
            { 0.30943274E+01,-0.11885342E+02,-0.64855026E+02,-0.10654855E+02,
              0.19201283E+02,-0.13944058E+01, 0.16869150E+02,-0.27809067E+01,
             -0.75264554E+01, 0.20107191E+02, 0.25993383E+01,-0.68293734E+01,
             -0.10687795E+02, },
            { 0.16992487E+03, 0.46145271E+02,-0.36491428E+02, 0.13116936E+02,
              0.34286945E+02, 0.78730626E+01, 0.54953218E+01,-0.18412989E+02,
              0.11174599E+02, 0.26052146E+01,-0.87555580E+01,-0.14695855E+02,
             -0.49863572E+01, },
            {-0.86712395E+02, 0.28112967E+02, 0.14459599E+03, 0.19851633E+02,
             -0.51848770E+02, 0.83207064E+01,-0.31751226E+02, 0.81608305E+01,
              0.16190430E+02,-0.46680210E+02,-0.44772248E+01, 0.18257042E+02,
              0.26758467E+02, },
            {-0.14461774E+03,-0.62235676E+02, 0.42808392E+02,-0.15741500E+02,
             -0.42179947E+02,-0.63173599E+01,-0.47827044E+01, 0.21872601E+02,
             -0.12338368E+02,-0.28421400E+01, 0.10545931E+02, 0.16512472E+02,
              0.45623174E+01, },
            { 0.12349963E+03,-0.25484196E+02,-0.13910371E+03,-0.15541058E+02,
              0.57951984E+02,-0.11481497E+02, 0.25678280E+02,-0.90002031E+01,
             -0.15541973E+02, 0.46808899E+02, 0.34660118E+01,-0.19979811E+02,
             -0.27687761E+02, },
            { 0.49242916E+02, 0.27691017E+02,-0.18267742E+02, 0.66256733E+01,
              0.17394239E+02, 0.18268524E+01, 0.15869142E+01,-0.90256739E+01,
              0.49349022E+01, 0.95967901E+00,-0.43152213E+01,-0.64874339E+01,
             -0.13755007E+01, },
            {-0.52184998E+02, 0.73005271E+01, 0.49102627E+02, 0.43267603E+01,
             -0.22683687E+02, 0.48243718E+01,-0.74036064E+01, 0.34105940E+01,
              0.54719052E+01,-0.16946058E+02,-0.10452632E+01, 0.77386184E+01,
              0.10162491E+02, },
            {-0.28234264E+00, 0.15047417E+01, 0.16470916E+01,-0.48499373E-02,
             -0.99328689E-01,-0.19733939E-01,-0.57898001E-02, 0.54610927E-01,
              0.14524219E-02, 0.27774854E-02, 0.19980098E-03, 0.76767760E-02,
             -0.83661648E-02, },
            { 0.36049801E+00,-0.15487713E+01, 0.14352394E+01,-0.28597059E-01,
              0.10506235E+00,-0.50372839E-01,-0.33528510E-01,-0.47886573E-01,
              0.57634916E-01,-0.36588252E-01, 0.12803383E-01,-0.29328017E-01,
              0.12255316E-01, },
            {-0.45544675E+00, 0.12475321E+01,-0.27271132E+01, 0.53092772E+00,
             -0.37300754E-01,-0.15156125E+00,-0.12215277E+00,-0.25236827E+00,
              0.13521501E+00, 0.36329892E+00, 0.22915353E+00,-0.17122564E+00,
              0.20348294E+00, },
            { 0.99758154E+00, 0.18983068E+01,-0.12141097E+00, 0.98354695E-02,
              0.15336272E+00, 0.17870606E+00,-0.72830565E-01, 0.17440289E+00,
              0.29191229E+00,-0.11306345E-01, 0.49621547E-02,-0.26877809E-01,
             -0.59098396E-02, },
            { 0.66613059E+01, 0.10923632E+02,-0.65970397E+00, 0.15522877E+01,
             -0.40954612E-01, 0.10254952E+01,-0.66243565E+00,-0.82247669E+00,
              0.49483865E+00,-0.44243667E+00,-0.27341887E+00,-0.15122773E+00,
             -0.44921702E-02, },
            {-0.38886521E+01, 0.54575652E+00, 0.10046308E+02, 0.23151209E+01,
             -0.10080287E+01, 0.23143321E+00, 0.23112388E+01, 0.82673109E+00,
             -0.13583063E+01, 0.51237226E+00,-0.59203404E+00, 0.10641146E+01,
             -0.16020350E-01, },
            { 0.38286211E+01,-0.29793097E+02, 0.47293575E+02,-0.99862328E+01,
             -0.12771469E+01, 0.57090254E+01, 0.90686393E+00, 0.46967688E+01,
             -0.48249456E+00,-0.78903108E+01,-0.37669451E+01, 0.33611767E+01,
             -0.28549347E+01, },
            {-0.42156825E+01,-0.18497114E+02, 0.43299971E+01, 0.12616365E+01,
             -0.74922557E+01, 0.18494391E+01, 0.33060460E+01,-0.33028397E+00,
             -0.44510746E+01,-0.27160680E+00, 0.20754679E+00, 0.16191085E+00,
              0.37127814E+00, },
            {-0.39047291E+02,-0.20917377E+02, 0.41411633E+02,-0.88546057E+01,
              0.36347630E+01,-0.46501169E+01, 0.43165107E+01, 0.63377647E+01,
             -0.79014248E+00, 0.20996790E+01, 0.26977179E+01, 0.49547371E-01,
             -0.80331719E+00, },
            { 0.24417505E+02,-0.50597050E+02,-0.12887446E+02,-0.25855635E+02,
              0.15060377E+02, 0.16648369E+01,-0.19734503E+02,-0.85645714E+01,
              0.11742890E+02, 0.24708913E+00, 0.30122676E+01,-0.88011379E+01,
             -0.45888126E-01, },
            {-0.12037977E+02, 0.18195720E+03,-0.30125452E+03, 0.56718910E+02,
              0.25235325E+02,-0.37847782E+02,-0.31361133E+00,-0.27802044E+02,
             -0.43944149E+01, 0.44352974E+02, 0.19799467E+02,-0.20778854E+02,
              0.12326245E+02, },
            {-0.41416504E+02, 0.12508257E+03,-0.62992988E+01,-0.16054337E+02,
              0.42519859E+02,-0.22424593E+02,-0.25135471E+02, 0.28161285E+01,
              0.19682430E+02, 0.15035913E+01,-0.35815947E+01, 0.37708406E+01,
             -0.11817546E+01, },
            { 0.89285217E+02,-0.69640999E+02,-0.17817468E+03, 0.21492142E+02,
             -0.17860403E+02, 0.67075515E+01,-0.15617906E+02,-0.20148720E+02,
             -0.40488234E+01,-0.17107679E+01,-0.83144178E+01, 0.30443573E+01,
              0.37432699E+01, },
            {-0.67462074E+02, 0.21829535E+03,-0.10266725E+03, 0.90444672E+02,
             -0.59647964E+02,-0.10293538E+02, 0.61297581E+02, 0.29657206E+02,
             -0.38800804E+02,-0.93168879E+01,-0.34729412E+01, 0.26925245E+02,
             -0.22017689E+01, },
            { 0.41186665E+02,-0.43663583E+03, 0.76321771E+03,-0.13260320E+03,
             -0.96994392E+02, 0.98809723E+02,-0.51419187E+01, 0.72419868E+02,
              0.21797657E+02,-0.10391363E+03,-0.44690708E+02, 0.50902679E+02,
             -0.22424061E+02, },
            { 0.19840215E+03,-0.35004047E+03,-0.27337153E+01, 0.44635445E+02,
             -0.85559502E+02, 0.67447411E+02, 0.67286049E+02,-0.17907366E+02,
             -0.35397381E+02, 0.19058682E+01, 0.13237237E+02,-0.18305573E+02,
              0.25211602E+00, },
            {-0.86747185E+02, 0.17347545E+03, 0.27344232E+03,-0.22800623E+02,
              0.27914158E+02,-0.33933644E+01, 0.26151587E+02, 0.27388149E+02,
              0.97586889E+01,-0.34674821E+01, 0.91410255E+01,-0.73688040E+01,
             -0.56035948E+01, },
            { 0.89096458E+02,-0.34042059E+03, 0.24030518E+03,-0.12773231E+03,
              0.86151863E+02, 0.17995501E+02,-0.78346107E+02,-0.41468628E+02,
              0.53289478E+02, 0.18957485E+02,-0.17327203E+01,-0.35620258E+02,
              0.71820140E+01, },
            {-0.68477409E+02, 0.45522491E+03,-0.86998053E+03, 0.13320493E+03,
              0.13828078E+03,-0.11296691E+03, 0.83289499E+01,-0.84006424E+02,
             -0.30078983E+02, 0.10894444E+03, 0.46073669E+02,-0.53107388E+02,
              0.18520054E+02, },
            {-0.27914142E+03, 0.44985599E+03,-0.11727840E+01,-0.40542358E+02,
              0.65773628E+02,-0.78929565E+02,-0.74833633E+02, 0.33005203E+02,
              0.26883795E+02,-0.96850271E+01,-0.17542385E+02, 0.28257006E+02,
              0.11094317E+01, },
            { 0.28169247E+02,-0.94577866E+02,-0.13985548E+03, 0.92719965E+01,
             -0.13244056E+02, 0.46153149E+00,-0.14713109E+02,-0.13108560E+02,
             -0.55679369E+01, 0.39470751E+01,-0.31173911E+01, 0.46963744E+01,
              0.28085086E+01, },
            {-0.41424831E+02, 0.17922871E+03,-0.13824788E+03, 0.62719311E+02,
             -0.40361134E+02,-0.99571943E+01, 0.35196617E+02, 0.20173746E+02,
             -0.25507154E+02,-0.10779361E+02, 0.30331249E+01, 0.17037985E+02,
             -0.52984161E+01, },
            { 0.38881653E+02,-0.17535535E+03, 0.36977036E+03,-0.48800972E+02,
             -0.67534874E+02, 0.46994465E+02,-0.36084449E+01, 0.35268539E+02,
              0.12913198E+02,-0.42146160E+02,-0.17815418E+02, 0.19775288E+02,
             -0.58432074E+01, },
            { 0.12580310E+03,-0.21518388E+03, 0.61443510E+01, 0.98084249E+01,
             -0.15733668E+02, 0.32032814E+02, 0.29372606E+02,-0.18224518E+02,
             -0.68510780E+01, 0.68954396E+01, 0.76545310E+01,-0.14303968E+02,
             -0.45466852E+00, },
            { 0.14912950E+00,-0.37197623E-01, 0.86089209E-01,-0.99689066E+00,
              0.34358004E+00, 0.41757068E-02, 0.11042265E-01,-0.31665795E-01,
              0.28434340E-01,-0.40736802E-01, 0.25358051E-01,-0.47866986E-02,
             -0.79530235E-02, },
            { 0.48351955E-01,-0.13903272E+00, 0.35033248E-01,-0.24141113E+00,
             -0.73474455E+00,-0.21165458E-02, 0.14233136E-01,-0.48750002E-01,
             -0.17404562E-01,-0.13931587E-01, 0.45621013E-02,-0.16081575E-01,
              0.18449978E-02, },
            { 0.15439525E+00, 0.10494125E+01, 0.53804463E+00, 0.16673172E+01,
             -0.14299715E+01, 0.68726435E-01,-0.16137257E+00, 0.13222758E+00,
              0.19399203E-01,-0.11860881E-01,-0.47297385E-01, 0.52892003E-01,
              0.16677070E-01, },
            { 0.36067051E+00, 0.11057934E+01, 0.61169720E+00, 0.19141263E+01,
              0.20964134E+01, 0.23646539E+00, 0.27404869E+00, 0.19712859E+00,
             -0.11474232E+00, 0.12215359E+00,-0.76190814E-01, 0.15382162E+00,
              0.43123838E-01, },
            { 0.10588363E+01, 0.11870986E+01, 0.88391542E+00, 0.23195529E+01,
              0.29367788E+01, 0.42034310E+00,-0.13270929E+01,-0.13814926E+00,
              0.26789281E+00,-0.76729983E-01,-0.62709659E+00,-0.13922800E+00,
             -0.22054443E+00, },
            { 0.26687746E+01, 0.14925655E+01,-0.10293322E+01,-0.60973663E+01,
              0.13214234E-01,-0.80502622E-01,-0.20025112E+00, 0.51012868E+00,
              0.36622775E+00, 0.50678074E+00,-0.61252302E+00, 0.36497283E+00,
              0.66087967E+00, },
            {-0.18513010E+01,-0.84644527E+01,-0.22534673E+01, 0.16729214E+01,
              0.59963422E+01, 0.83362734E+00,-0.32418764E+00,-0.12634010E+01,
             -0.21761373E+00,-0.32083201E+00, 0.10582196E+01,-0.40086418E-01,
              0.34409949E+00, },
            {-0.56983318E+01,-0.85301552E+01,-0.62716293E+01,-0.10336222E+02,
             -0.12337216E+01, 0.14525561E+01,-0.27670050E+01,-0.61272615E+00,
              0.89260936E-02,-0.17194246E+01, 0.48456728E+00,-0.75325119E+00,
             -0.12380633E+00, },
            {-0.10879601E+02, 0.18890285E+00,-0.12624624E+02, 0.85096445E+01,
             -0.14647897E+02,-0.11804613E+01, 0.11204756E+02, 0.11301163E+01,
             -0.65179712E+00, 0.20850821E+01, 0.33798697E+01, 0.68264633E+00,
              0.13606979E+01, },
            {-0.41109333E+01,-0.76429524E+01, 0.31357608E+01, 0.33289852E+02,
              0.18736217E+02, 0.52604097E+00, 0.32423337E+01,-0.39312522E+01,
             -0.42085305E+00,-0.20370185E+01, 0.46622753E+01,-0.20514877E+01,
             -0.67121148E+01, },
            { 0.33330195E+01, 0.19760462E+02, 0.13707290E+01,-0.15091409E+02,
             -0.44948006E+01,-0.37715893E+01, 0.19507859E+01, 0.41595783E+01,
              0.84614331E+00, 0.17263756E+01,-0.36580293E+01,-0.27227202E+00,
             -0.15076882E+01, },
            { 0.17205370E+02, 0.15522379E+02, 0.18444885E+02, 0.14610297E+02,
             -0.95373640E+01,-0.77903051E+01, 0.75731368E+01, 0.10421705E+01,
              0.14117401E+01, 0.51428719E+01,-0.15598035E+01, 0.11522195E+01,
              0.33083048E+00, },
            { 0.23841166E+02,-0.88205357E+01, 0.33722401E+02,-0.35630077E+02,
              0.22245089E+02, 0.98581381E-01,-0.23630219E+02,-0.24590368E+01,
             -0.60612363E+00,-0.56737971E+01,-0.57873211E+01,-0.94312793E+00,
             -0.20955913E+01, },
            {-0.14294488E+02, 0.16591463E+02,-0.36211152E+01,-0.58140308E+02,
             -0.52787479E+02,-0.87697142E+00,-0.86255169E+01, 0.79262271E+01,
             -0.16103133E+01, 0.19649650E+01,-0.10198181E+02, 0.28549724E+01,
              0.16818821E+02, },
            {-0.28521106E+01,-0.13874240E+02, 0.60465282E+00, 0.88589296E+01,
             -0.56679926E+01, 0.32743344E+01,-0.16497432E+01,-0.34437387E+01,
             -0.67558551E+00,-0.17088107E+01, 0.29519806E+01, 0.42225948E+00,
              0.13573281E+01, },
            {-0.15705806E+02,-0.77159276E+01,-0.15102554E+02,-0.16817579E+01,
              0.56376777E+01, 0.67993746E+01,-0.60265536E+01,-0.10358450E+01,
             -0.14570483E+01,-0.36746924E+01, 0.12936862E+01,-0.37998509E+00,
             -0.32503825E+00, },
            {-0.14731427E+02, 0.91323881E+01,-0.23522758E+02, 0.31281187E+02,
             -0.66921196E+01, 0.88023257E+00, 0.14684581E+02, 0.14542351E+01,
              0.11787513E+01, 0.40096455E+01, 0.31103556E+01, 0.42140290E+00,
              0.75577301E+00, },
            { 0.22154350E+02,-0.13379253E+02, 0.43898410E+00, 0.28501005E+02,
              0.40360855E+02, 0.30411851E+00, 0.64322257E+01,-0.44322767E+01,
              0.17975950E+01,-0.39312956E+00, 0.68860784E+01,-0.11438121E+01,
             -0.11947880E+02, },
            { 0.27685111E-01,-0.40053260E-01,-0.17839098E-01,-0.57812698E-01,
              0.70409186E-01,-0.50432491E+00,-0.61618298E+00,-0.51988091E-01,
              0.36189247E-01,-0.57952993E-01, 0.35770409E-01,-0.30052310E-01,
             -0.15522750E-01, },
            {-0.99234711E-02,-0.25411305E+00,-0.98816864E-01, 0.27671959E-01,
             -0.90068385E-01, 0.64124870E+00,-0.49534273E+00,-0.23844037E-01,
              0.90174731E-02, 0.16274264E-01, 0.13823560E-02,-0.16471019E-01,
              0.77255028E-02, },
            {-0.29735389E+00, 0.89244604E+00, 0.47042780E-01,-0.19908190E+00,
             -0.20957834E+00, 0.48498312E+00, 0.34726390E+00, 0.14254756E+00,
             -0.51067263E-01,-0.74147590E-01, 0.43129999E-01,-0.94624739E-02,
              0.69399804E-01, },
            {-0.27613023E+00, 0.39687451E-01,-0.22952503E+00,-0.17188592E+00,
             -0.37711376E+00,-0.37344018E+00, 0.44496518E+00, 0.49825110E-01,
              0.50268643E-01, 0.10268872E-01,-0.15096737E+00,-0.18276261E-02,
              0.57382856E-01, },
            { 0.55527079E+00, 0.49626112E+00, 0.41093677E+00,-0.36513752E+00,
             -0.92810732E+00, 0.45406213E+00, 0.39471567E+00,-0.30232587E+00,
              0.54300815E+00, 0.59649146E+00,-0.42978968E-01, 0.16226175E+00,
             -0.37429817E-01, },
            {-0.20489936E+00, 0.37278998E+01,-0.70701146E+00,-0.19198212E+00,
              0.71857435E+00,-0.99595582E+00, 0.54645717E+00,-0.43600601E+00,
             -0.46274224E+00, 0.24707445E-02, 0.20912142E+00, 0.10534024E+00,
             -0.15043882E-01, },
            {-0.54460263E+00,-0.15500151E+01, 0.16229698E-01, 0.42883778E+00,
              0.12063434E+00,-0.49502440E-01,-0.30448535E+00,-0.20664476E+00,
              0.19829325E+00, 0.26275936E+00,-0.19895925E+00, 0.21524690E+00,
             -0.15297687E+00, },
            { 0.29179329E+00, 0.17002033E+00, 0.72335619E+00, 0.31211472E+00,
              0.96879709E+00, 0.17340130E+00, 0.76188266E-01,-0.18247452E+00,
             -0.35377108E-01,-0.23166701E-01, 0.31477252E+00, 0.19326968E+00,
             -0.67122035E-01, },
            { 0.96205816E-01, 0.37230721E+00,-0.99004132E+00, 0.10045286E+01,
              0.21211245E+01,-0.90331696E-01, 0.57627037E-01, 0.65079659E+00,
             -0.12693567E+01,-0.96175170E+00, 0.58023080E-01,-0.35038722E+00,
              0.14534500E+00, },
            {-0.79609752E-01,-0.49538903E+01, 0.14437783E+01, 0.20078922E-01,
             -0.11813406E+01, 0.14132198E+01,-0.34010509E+00, 0.10204923E+01,
              0.75110745E+00,-0.16915599E+00,-0.33175179E+00,-0.33036423E+00,
             -0.20429282E-01, },
            {-0.10897010E-01,-0.49416184E-01,-0.44178925E-01,-0.84936572E-02,
              0.51277783E-01, 0.19495703E-01,-0.32920949E-01, 0.21518527E+00,
             -0.17939478E+00,-0.32887351E-01, 0.33539165E-01,-0.19388679E-01,
             -0.20020414E-01, },
            { 0.20877963E-01, 0.27526090E-01,-0.37257954E-01, 0.32353066E-01,
             -0.61392702E-01, 0.98865917E-02,-0.21773189E-01, 0.17957857E+00,
              0.23242453E+00, 0.58700140E-02, 0.39826785E-02, 0.25550261E-01,
              0.60049252E-03, },
            {-0.15558755E+00,-0.54346319E-01,-0.56051023E-01, 0.19358875E-01,
             -0.15182693E+00,-0.10598335E+00, 0.37775725E-01, 0.16895783E+00,
              0.51459491E+00,-0.84568355E-02,-0.67650639E-01, 0.27996898E-02,
             -0.11083936E-01, },
            { 0.30534738E+00, 0.23073228E+00,-0.16615911E+00,-0.73981099E-01,
              0.17580722E+00,-0.86328842E-01,-0.72317007E-02,-0.50074303E+00,
             -0.23022939E-02, 0.22110448E-02,-0.61923102E-01,-0.39555937E-01,
              0.32912306E-01, },
            {-0.12910460E+00,-0.22992328E-01,-0.24645030E-01,-0.16976858E-01,
             -0.48633263E-01, 0.24791237E-01,-0.44906000E-02, 0.44549942E-01,
             -0.13423373E-01, 0.14574012E+00, 0.18071000E+00, 0.22294498E-02,
             -0.21888547E-01, },
            { 0.82140617E-01, 0.60347050E-01,-0.69422051E-01, 0.34339996E-02,
             -0.79234779E-01,-0.41237101E-01,-0.28514843E-01, 0.24108853E-01,
              0.40748831E-01,-0.15513334E+00, 0.15307407E+00, 0.99850390E-02,
             -0.27682120E-01, },
            { 0.54472767E-01,-0.43857154E-01,-0.53891961E-01,-0.13284815E-01,
              0.35775442E-01, 0.40358426E-02,-0.27592603E-01, 0.31445228E-01,
              0.15491653E-01, 0.28190129E-02,-0.24603689E-01,-0.92626899E-01,
              0.67674026E-01, },
            { 0.71794264E-01, 0.95135733E-01, 0.50922348E-02,-0.13623560E-01,
              0.23757331E-01,-0.47744387E-02, 0.46965736E-02,-0.67392141E-02,
             -0.48691975E-02,-0.11122855E-02,-0.13454689E-01,-0.77131763E-01,
             -0.98624110E-01, },
            { 0.48160724E-01, 0.16543763E-01,-0.49138770E-01, 0.17461605E-01,
              0.26750349E-01, 0.39088377E-02, 0.55514900E-02, 0.43032456E-01,
              0.60738609E-02,-0.25063993E-02,-0.15855387E-01,-0.94612045E-02,
             -0.41879309E-03, },
            { 0.32233037E-01,-0.19281844E-01, 0.20792220E-01, 0.94251204E-02,
             -0.47330942E-01, 0.23148457E-01, 0.49970322E-02,-0.24953449E-01,
              0.21466170E-01,-0.37387648E-03, 0.19118547E-01, 0.32155174E-02,
              0.16403269E-01, },
            { 0.35219163E-01, 0.13113920E+00,-0.96734829E-01, 0.13746214E-01,
             -0.28147859E-01, 0.16202064E-01, 0.51043965E-01,-0.35983652E-02,
             -0.52156085E-02,-0.96815266E-02,-0.55470653E-02,-0.20149490E-01,
             -0.10250242E-01, },
            {-0.27565159E-01, 0.56680135E-01,-0.28701194E-01,-0.58173505E-02,
              0.27876016E-01, 0.27989307E-01,-0.13116185E-01,-0.12744517E-01,
              0.25175970E-01, 0.43719960E-02, 0.16126093E-02,-0.25717752E-01,
              0.67674026E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir17Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30802848E+01,-0.46980400E-02, 0.10328545E-01,-0.13646303E-02,
             -0.13921593E-02, 0.46656518E-02, 0.75738039E-02,-0.13919175E-02,
              0.18014414E-02, },
            {-0.73226583E+00, 0.12279886E-01, 0.34578342E-01, 0.33941451E-01,
              0.73212688E-03,-0.21163456E-01, 0.32289596E-02,-0.18333461E-01,
              0.91965441E-02, },
            { 0.79729569E+00, 0.13542826E+00,-0.11436112E+00,-0.70079923E-01,
              0.79188850E-02,-0.77631488E-01,-0.55221070E-01, 0.45033410E-01,
             -0.30216960E-01, },
            { 0.18268728E+01,-0.31209853E+00,-0.70338599E-01,-0.13058031E+00,
              0.52127116E-02, 0.59936713E-01,-0.18583011E-01, 0.56642335E-01,
              0.17623593E-02, },
            {-0.78476065E+00,-0.25712866E+00, 0.25612435E+00, 0.15128808E+00,
             -0.38681533E-01, 0.24744213E+00, 0.14993742E+00,-0.11375405E+00,
              0.66822320E-01, },
            {-0.13493091E+01, 0.30580220E+00, 0.73491491E-01, 0.89370720E-01,
              0.54565580E-02,-0.46355233E-01, 0.13624964E-01,-0.36061008E-01,
             -0.18066941E-01, },
            {-0.12063475E+00, 0.14794455E+00,-0.10244069E+00,-0.80330491E-01,
              0.43026242E-01,-0.18751287E+00,-0.10860286E+00, 0.70955373E-01,
             -0.43552943E-01, },
            { 0.86388132E-02,-0.10469254E+00,-0.35336661E+00, 0.68503232E-02,
             -0.22832712E-01, 0.74300254E-02,-0.25587229E-02, 0.51693595E-02,
             -0.12979574E-03, },
            { 0.82351686E-02, 0.37010682E+00,-0.82846217E-01,-0.23642721E-01,
              0.21374708E-01, 0.24881929E-02,-0.14439701E-01, 0.34175650E-02,
              0.15466106E-01, },
            {-0.13728087E+00,-0.90724178E-01,-0.25918829E+00, 0.55246770E-01,
              0.28049927E-01,-0.48293497E-01,-0.40341385E-01, 0.25727993E-01,
              0.12118476E-01, },
            { 0.87273709E-01, 0.31198278E+00,-0.16230865E+00, 0.16893700E-01,
              0.17322816E-02,-0.83513986E-02, 0.22217995E-01, 0.15569704E-01,
              0.63724130E-01, },
            { 0.55732697E-01, 0.28494719E+00, 0.15282595E+01,-0.20563173E+00,
              0.30643355E-01,-0.10549944E+00, 0.10248962E+00, 0.31379037E-01,
             -0.74367374E-02, },
            {-0.61427230E+00,-0.15684907E+01, 0.10502628E+00, 0.25299487E+00,
             -0.15594783E+00,-0.50039280E-01, 0.76203002E-02,-0.49784355E-01,
             -0.12944193E+00, },
            { 0.15699542E+01, 0.26262048E+00, 0.62663382E+00, 0.22692803E-01,
             -0.51594186E+00, 0.51364642E+00, 0.29827582E-01,-0.80818407E-01,
             -0.30650470E+00, },
            { 0.24862222E+00,-0.59349144E+00, 0.11204494E+01,-0.97968966E-01,
              0.33110031E+00, 0.38167056E-01,-0.16308677E+00, 0.54526995E-02,
             -0.50980163E+00, },
            {-0.47659847E+00,-0.55125346E-02,-0.95217299E+00, 0.47925776E+00,
             -0.42554024E-02, 0.30441788E+00,-0.30924687E+00,-0.12380421E+00,
             -0.70827238E-01, },
            { 0.20973926E+01, 0.99478257E+00, 0.27451342E+00,-0.55617458E+00,
              0.29574648E+00, 0.15973371E+00, 0.80506027E-01, 0.57603039E-01,
              0.32256550E+00, },
            {-0.38655005E+01,-0.90660799E+00,-0.59841734E+00,-0.50917405E+00,
              0.14092973E+01,-0.11617564E+01, 0.28998095E+00, 0.36635254E-01,
              0.80191475E+00, },
            {-0.10721380E+01, 0.55519497E+00,-0.26079962E+01, 0.24563982E+00,
             -0.10116768E+01,-0.10343829E+00, 0.17780776E+00,-0.81209339E-01,
              0.10334632E+01, },
            { 0.69367838E+00,-0.10222288E+00,-0.55200565E+00,-0.34082192E+00,
              0.15705701E-01,-0.24950403E+00, 0.24499443E+00, 0.78426957E-01,
              0.83322436E-01, },
            {-0.16341933E+01, 0.60357797E+00,-0.35521424E+00, 0.34881729E+00,
             -0.18382023E+00,-0.85821614E-01,-0.79250164E-01, 0.24138987E-02,
             -0.23403056E+00, },
            { 0.27014959E+01, 0.11459872E+01,-0.33060220E+00, 0.46477214E+00,
             -0.97078705E+00, 0.67035884E+00,-0.35578537E+00, 0.13218268E-01,
             -0.56398720E+00, },
            { 0.62231594E+00, 0.28681967E+00, 0.19950296E+01,-0.16715036E+00,
              0.70967084E+00, 0.80056027E-01,-0.58295801E-01, 0.67210004E-01,
             -0.60891330E+00, },
            {-0.86633563E-02,-0.13671836E-02,-0.19294512E-01,-0.29737541E-01,
             -0.15660122E+00,-0.86753992E-02, 0.21384902E-03, 0.20422880E-02,
             -0.96033914E-02, },
            { 0.10736866E-02, 0.13963470E-02, 0.80128183E-03, 0.16088326E+00,
             -0.86053144E-02, 0.60172728E-03,-0.22198915E-01,-0.37762453E-03,
              0.10805884E-02, },
            { 0.39740406E-01, 0.17161980E+00, 0.12527364E+00, 0.13852577E+00,
             -0.44440904E+00,-0.17778879E-01,-0.20324275E-01, 0.42344291E-01,
             -0.20028139E-01, },
            {-0.42737368E-01, 0.44157285E-01, 0.24216682E-01, 0.59995002E+00,
              0.19375703E+00, 0.41468341E-01,-0.40872436E-01,-0.31848896E-01,
             -0.14190458E-01, },
            {-0.27813461E-01, 0.34570392E-01, 0.91654710E-01, 0.33465780E-01,
             -0.93962334E-01, 0.60697462E-01,-0.45351591E-01,-0.58084037E-02,
              0.61074000E-01, },
            {-0.37716964E+00,-0.15043245E+00,-0.14817756E+00, 0.47543745E-01,
             -0.86728200E-01, 0.59861321E-01, 0.14234369E+00, 0.12710453E-01,
              0.17827239E-01, },
            {-0.24680117E+00,-0.86096442E+00,-0.37145874E+00,-0.49700645E+00,
              0.13920882E+01, 0.93504988E-01, 0.44289514E-01,-0.18447290E+00,
              0.17054563E-01, },
            { 0.70487481E+00,-0.10964639E-01, 0.19685283E+00,-0.18399014E+01,
             -0.80342036E+00,-0.18387800E+00, 0.39359804E-01, 0.17004316E+00,
             -0.85180625E-01, },
            {-0.19522337E-01,-0.21377654E+00,-0.14989559E+00,-0.97864687E-01,
              0.37941441E+00,-0.63822173E-01, 0.14001627E+00, 0.39011661E-02,
             -0.85912690E-01, },
            { 0.60058129E+00, 0.10312814E+00, 0.23901564E+00,-0.29843274E+00,
              0.10723197E-01,-0.10886107E+00,-0.15636367E+00,-0.39202116E-01,
             -0.35129011E-01, },
            { 0.25636184E+00, 0.11735374E+01, 0.45907420E+00, 0.33341327E+00,
             -0.12187766E+01,-0.59096042E-01,-0.59066504E-01, 0.20547853E+00,
              0.14204452E-01, },
            {-0.11265887E+01,-0.48129063E-01,-0.23462927E+00, 0.14852360E+01,
              0.70783460E+00, 0.19520159E+00, 0.13287680E+00,-0.18437938E+00,
              0.19562475E+00, },
            {-0.32194812E-01, 0.65475008E-02, 0.37131528E-02,-0.14261339E-01,
             -0.12782280E-02, 0.23382235E-01, 0.15923720E-01, 0.40608902E-04,
              0.66276081E-03, },
            { 0.31999865E-03,-0.42041160E-01,-0.23349874E-01, 0.11027430E-01,
             -0.22927441E-01,-0.20414442E-01, 0.19795489E-01,-0.69397581E-02,
              0.38036958E-02, },
            { 0.65187119E-01, 0.50789274E-01, 0.12329996E-01,-0.14014617E-01,
              0.19207697E-01, 0.54233067E-01, 0.19744506E-01, 0.45834291E-02,
              0.72671259E-02, },
            { 0.29568112E-01,-0.95958523E-02,-0.18973609E-01,-0.41401144E-02,
              0.17457606E-01,-0.15092639E-01, 0.75680852E-01,-0.13022618E-01,
              0.67012999E-02, },
            { 0.77274968E-02,-0.82368053E-01,-0.58961112E-01, 0.20770151E-01,
             -0.13150994E-02,-0.13049465E+00,-0.10597041E+00, 0.16197849E-01,
              0.15678909E-01, },
            {-0.72845094E-01, 0.12490302E+00,-0.20829698E-02,-0.25820648E-01,
              0.39102312E-01, 0.10707958E+00,-0.13309467E+00, 0.15458102E-01,
              0.37044154E-02, },
            {-0.31673741E-01,-0.62940731E-02, 0.84093548E-02,-0.41974136E-02,
             -0.33746944E-02, 0.80712140E-02,-0.77740918E-02, 0.31657413E-01,
              0.36994985E-02, },
            {-0.28113980E-01, 0.10798969E-01,-0.56396737E-02,-0.16900625E-02,
             -0.34986283E-02, 0.40706303E-02, 0.11442027E-01,-0.97587286E-02,
              0.27690606E-01, },
            { 0.79382271E-01,-0.30479811E-01, 0.32784846E-01,-0.26296770E-01,
             -0.10788017E-01, 0.10917043E-01, 0.14642074E-01, 0.22470199E-01,
              0.14176787E-01, },
            { 0.31412635E-01,-0.57085998E-01, 0.26893871E-01, 0.37701238E-01,
             -0.21372926E-02,-0.14738973E-01, 0.80593152E-03,-0.38642716E-01,
              0.31024845E-01, },
            { 0.50734030E-03,-0.64163841E-02,-0.12042396E-01, 0.22190325E-02,
             -0.28449393E-03, 0.62294272E-02,-0.91660712E-02, 0.21545021E-02,
             -0.10553715E-02, },
            { 0.14355371E-01, 0.12396547E-01,-0.10789397E-01,-0.44431705E-02,
             -0.94738062E-02, 0.47621662E-02, 0.40972382E-02, 0.32042819E-02,
             -0.99020661E-03, },
            { 0.49181809E-02, 0.23562347E-02, 0.40638875E-02, 0.27082455E-02,
             -0.65866150E-02, 0.62013259E-02, 0.16731292E-02, 0.18732373E-02,
             -0.48593818E-02, },
            { 0.13873120E-01,-0.13107671E-02,-0.19071303E-01, 0.38780418E-03,
              0.61205011E-02,-0.50388812E-02, 0.68964978E-03,-0.54935296E-02,
              0.18295188E-02, },
         },
         {
            { 0.27496719E+01,-0.11850115E-01, 0.73418082E-02, 0.64126854E-02,
             -0.29800744E-02, 0.10921643E-02, 0.41902238E-02, 0.14759843E-02,
             -0.22041446E-02, },
            {-0.51172680E+00, 0.35100680E-01, 0.40524267E-01, 0.30474843E-01,
             -0.65487823E-02,-0.66860467E-02,-0.82540857E-02,-0.16374506E-01,
              0.66717844E-02, },
            { 0.10756888E+01, 0.95197663E-01,-0.99020548E-01,-0.55270936E-01,
              0.56841303E-01,-0.30476358E-01,-0.44919342E-01, 0.41535314E-01,
              0.14882590E-01, },
            { 0.96904379E+00,-0.45030373E+00,-0.81589915E-01,-0.10445018E+00,
              0.13437195E-01,-0.15732037E-01, 0.23781391E-01, 0.42177223E-01,
             -0.97438293E-02, },
            {-0.12668188E+01,-0.14799927E-01, 0.23314634E+00, 0.87992266E-01,
             -0.20218949E+00, 0.13027802E+00, 0.13615835E+00,-0.11343505E+00,
             -0.31153904E-01, },
            {-0.63311547E+00, 0.41507858E+00, 0.64436302E-01, 0.67502894E-01,
              0.85417582E-02, 0.24351122E-01,-0.25230847E-01,-0.23825040E-01,
              0.86125854E-03, },
            { 0.16201258E+00,-0.35472833E-01,-0.86220764E-01,-0.37761636E-01,
              0.15207741E+00,-0.11216554E+00,-0.97735316E-01, 0.73955752E-01,
              0.15572720E-01, },
            { 0.19004470E-01,-0.27081847E+00,-0.25790614E+00, 0.95325941E-02,
             -0.15033994E-01, 0.27574443E-02,-0.25284474E-02, 0.38412474E-02,
              0.23758332E-02, },
            {-0.15574141E-01, 0.28764451E+00,-0.24723634E+00,-0.21190660E-01,
              0.19187475E-01,-0.14577296E-02,-0.13602387E-01, 0.47102780E-02,
              0.15887067E-01, },
            { 0.15916567E-01,-0.13094133E-01,-0.18341503E+00,-0.20860847E-01,
              0.39675791E-01,-0.68267519E-02,-0.66249087E-01,-0.62741726E-02,
              0.41506059E-01, },
            { 0.33226557E-01, 0.21242505E+00,-0.11735039E+00, 0.20106252E-01,
              0.14337774E-01, 0.52691698E-02, 0.18621637E-01,-0.20861311E-02,
              0.50332192E-01, },
            {-0.32793215E+00, 0.81487375E+00, 0.84182167E+00,-0.12698558E+00,
              0.33453379E-01,-0.25510749E-01, 0.89533143E-01, 0.24488725E-01,
             -0.31261619E-01, },
            {-0.37190884E+00,-0.10941412E+01, 0.68853664E+00, 0.24537718E+00,
             -0.12240302E+00,-0.18654227E-01, 0.40675946E-01,-0.11932018E+00,
             -0.14532743E+00, },
            { 0.39973900E-01,-0.20130175E+00, 0.98582131E+00, 0.46249530E+00,
             -0.45445198E+00, 0.95881127E-01, 0.30371648E+00, 0.18240003E+00,
             -0.36862892E+00, },
            { 0.49964002E+00,-0.55297488E+00, 0.81649625E+00,-0.15544052E+00,
              0.17243408E+00,-0.80211408E-01,-0.12309635E+00, 0.14403760E+00,
             -0.36307234E+00, },
            { 0.92932731E+00,-0.43920514E+00, 0.27505100E+00, 0.24573041E+00,
              0.31845255E-02, 0.79882346E-01,-0.29909259E+00,-0.48250318E-01,
              0.84246330E-01, },
            { 0.15329826E+01, 0.19085544E+00,-0.40697181E+00,-0.55546308E+00,
              0.18732370E+00, 0.17146086E-01, 0.14358910E-01, 0.27570710E+00,
              0.33728591E+00, },
            {-0.16367953E+00, 0.20679381E+00,-0.20812731E+01,-0.13375096E+01,
              0.10512470E+01,-0.18085140E+00,-0.35574475E+00,-0.50738007E+00,
              0.78306991E+00, },
            {-0.15351163E+01, 0.11400683E+01,-0.19037578E+01, 0.40057018E+00,
             -0.52411419E+00, 0.25429851E+00, 0.26375519E-01,-0.37648410E+00,
              0.65987456E+00, },
            {-0.54566395E+00, 0.74088342E-01,-0.10294523E+01,-0.16286100E+00,
             -0.72648101E-01,-0.86193547E-01, 0.27236110E+00, 0.20283155E-01,
             -0.70224538E-01, },
            {-0.12850333E+01, 0.83220738E+00,-0.85517123E-01, 0.36889663E+00,
             -0.78059815E-01, 0.37251141E-01,-0.59753969E-01,-0.16131189E+00,
             -0.22452673E+00, },
            { 0.14172664E+00, 0.15790805E+00, 0.79502696E+00, 0.10112376E+01,
             -0.61172265E+00, 0.34335397E-01, 0.59816524E-01, 0.33198726E+00,
             -0.48746240E+00, },
            { 0.10491930E+01,-0.24701853E+00, 0.15002282E+01,-0.29250708E+00,
              0.31989691E+00,-0.20179859E+00, 0.11427309E+00, 0.24893668E+00,
             -0.36306697E+00, },
            {-0.13550294E-01,-0.45300750E-02, 0.10528672E-02, 0.16152516E-01,
             -0.12370551E+00,-0.59309341E-02, 0.40165852E-02, 0.45750680E-03,
             -0.59628934E-02, },
            {-0.12362761E-01,-0.79272036E-02, 0.50656572E-02, 0.11978005E+00,
              0.42122323E-01, 0.62554074E-03,-0.15799779E-01,-0.46376414E-02,
              0.30326883E-02, },
            { 0.40974230E-01, 0.18386219E+00, 0.66352732E-01, 0.13339929E+00,
             -0.20539436E+00, 0.68436977E-02, 0.56280051E-02, 0.26018705E-01,
             -0.77441498E-02, },
            { 0.30746153E-01, 0.64624652E-01, 0.41689617E-02, 0.32169458E+00,
              0.19580451E+00, 0.16591450E-01,-0.15418100E-01,-0.22828069E-01,
              0.79204692E-02, },
            {-0.79213560E-01, 0.82305610E-01,-0.45376107E-01,-0.24927042E+00,
              0.59723146E-01, 0.52313942E-01,-0.31539883E-01,-0.39117865E-01,
              0.51201828E-01, },
            {-0.16604353E+00,-0.34118190E-01,-0.95448852E-01,-0.24441021E-01,
             -0.38288221E+00, 0.89257881E-02, 0.12591472E+00, 0.38770281E-01,
              0.11474446E-01, },
            {-0.95860370E-01,-0.84688097E+00,-0.15334813E+00,-0.17581967E+00,
              0.49173304E+00,-0.18058453E-01,-0.11134413E+00,-0.12453195E+00,
              0.18030409E-01, },
            { 0.21763852E+00,-0.97603641E-01, 0.88954560E-01,-0.84687620E+00,
             -0.32350314E+00, 0.37710898E-01, 0.73295427E-02, 0.13443394E+00,
             -0.16885924E+00, },
            { 0.10223650E+00,-0.29878554E+00, 0.48473828E-01, 0.14142914E+00,
             -0.96604209E-02,-0.75192235E-01, 0.95178723E-01, 0.81541315E-01,
             -0.61776754E-01, },
            { 0.29203841E+00,-0.17904034E+00, 0.66509947E-01,-0.33817422E-01,
              0.29854304E+00,-0.61735559E-01,-0.15191223E+00,-0.53076677E-01,
             -0.26007971E-01, },
            {-0.22676747E-01, 0.11153493E+01, 0.22664391E+00,-0.13381992E-01,
             -0.44242853E+00, 0.69775105E-01, 0.11397256E+00, 0.13336641E+00,
             -0.25085449E-01, },
            {-0.44642001E+00, 0.18625054E+00,-0.79894532E-02, 0.63482380E+00,
              0.70025556E-01,-0.66104084E-01, 0.85890360E-01,-0.14158376E+00,
              0.24665558E+00, },
            {-0.29041942E-01, 0.10848622E-01, 0.89042038E-02,-0.12965817E-01,
             -0.79534082E-02, 0.20859802E-01,-0.14016238E-01,-0.29190006E-02,
             -0.91342576E-03, },
            {-0.55187498E-02,-0.40372211E-01,-0.21171888E-01, 0.10458578E-01,
             -0.23211660E-01, 0.95567387E-02, 0.19028161E-01,-0.29276188E-02,
             -0.41621849E-02, },
            { 0.39665487E-01, 0.72864085E-01, 0.10825509E-01,-0.18992426E-01,
              0.10280470E-01, 0.83170965E-01, 0.42992984E-02, 0.93182252E-03,
              0.22329555E-02, },
            { 0.22737475E-01,-0.66708000E-02,-0.27572964E-02,-0.16470218E-01,
              0.29193019E-02,-0.60508568E-02, 0.88432148E-01,-0.63352720E-02,
              0.46102949E-02, },
            { 0.47638151E-02,-0.11769389E+00,-0.53772289E-01, 0.17596751E-01,
              0.23333792E-03,-0.14802746E+00,-0.58708407E-01, 0.86851157E-02,
              0.21828359E-01, },
            {-0.43172285E-01, 0.11592652E+00,-0.16070418E-02,-0.11035030E-01,
              0.17440675E-01, 0.57708938E-01,-0.13491406E+00,-0.18783395E-03,
              0.19392852E-01, },
            {-0.20160457E-01,-0.94742822E-02, 0.71759149E-02, 0.22431763E-03,
             -0.19019587E-02, 0.46471963E-02,-0.72945268E-02, 0.25696186E-01,
              0.10441859E-01, },
            {-0.15380017E-01, 0.67414884E-02,-0.12804105E-01, 0.24585058E-02,
             -0.51557599E-02, 0.37254030E-02, 0.76002465E-02,-0.13046168E-01,
              0.20730097E-01, },
            { 0.57154182E-01,-0.87949578E-02, 0.30844212E-01,-0.17324708E-01,
             -0.67216763E-03, 0.66427323E-02, 0.19945210E-01, 0.24971852E-01,
              0.30873109E-01, },
            { 0.13570253E-01,-0.68256535E-01, 0.30832380E-01, 0.22241361E-01,
             -0.16376127E-02,-0.12042126E-01,-0.16067507E-02,-0.53055108E-01,
              0.24462879E-01, },
            { 0.78554545E-02,-0.98696323E-02,-0.61809998E-02, 0.83083156E-02,
             -0.34189986E-02, 0.37059986E-02,-0.70023979E-02, 0.45059728E-02,
             -0.20708663E-02, },
            { 0.94396900E-02, 0.11253028E-01,-0.92065604E-02, 0.81361667E-03,
             -0.75608422E-03, 0.33125300E-02, 0.41626222E-02, 0.37212600E-02,
              0.41885455E-02, },
            { 0.60033705E-02, 0.81377747E-02, 0.13521597E-02, 0.25659315E-02,
             -0.80228671E-02, 0.31734100E-02, 0.90754894E-03, 0.38459085E-03,
             -0.12874377E-02, },
            {-0.29407945E-02,-0.58028940E-02,-0.11697264E-01, 0.18782510E-02,
              0.10815329E-01,-0.28420230E-02, 0.76707790E-03,-0.40857978E-02,
              0.25540697E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir18F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.51813760E+01,-0.10739883E+00, 0.18743367E+00, 0.31032026E-01,
             -0.31483453E-01, 0.46369401E-02,-0.17831052E-01, 0.25944985E-01,
             -0.39442893E-01, 0.11673883E-01, 0.13706181E-01, 0.52697590E-03,
             -0.25141253E-02, },
            {-0.16610043E+01,-0.63904011E+00, 0.52825075E+00,-0.35923615E+00,
              0.32526955E+00, 0.38260385E-01, 0.74715331E-01,-0.12071173E+00,
             -0.12027262E+00,-0.22109205E-01,-0.18548496E-02, 0.25158874E-02,
              0.43518778E-01, },
            { 0.11134710E+02, 0.21947019E+01,-0.38336279E+01, 0.41743028E+00,
              0.46625581E+00,-0.36733770E+00, 0.96669585E+00,-0.55345803E+00,
              0.46823847E+00,-0.48935539E+00,-0.64198560E+00,-0.11552801E+00,
              0.34517925E-01, },
            { 0.20840195E+02, 0.53288760E+01,-0.30419121E+01, 0.70807428E+01,
             -0.30199430E+01, 0.25945339E-01,-0.33274748E+01, 0.13975244E+01,
              0.53362620E+00, 0.27826801E+00, 0.33125132E-01, 0.62786597E+00,
             -0.52923745E+00, },
            {-0.79337822E+02,-0.14883757E+02, 0.22834850E+02,-0.51182194E+01,
             -0.11177418E+02, 0.16302793E+01,-0.79739599E+01, 0.32296658E+01,
             -0.12089566E+01, 0.36283174E+01, 0.44965773E+01, 0.86827540E+00,
             -0.31795710E+00, },
            {-0.60320335E+02,-0.19182150E+02,-0.33956966E+01,-0.32669117E+02,
              0.17268829E+02,-0.60134012E+00, 0.18945456E+02,-0.86268539E+01,
             -0.11386640E+01,-0.19328128E+01, 0.17548665E+01,-0.65107803E+01,
              0.11776744E+01, },
            { 0.18447604E+03, 0.42060055E+02,-0.56183029E+02, 0.14291060E+02,
              0.41250343E+02,-0.34360697E+01, 0.21798124E+02,-0.76566525E+01,
              0.13683516E+01,-0.94252605E+01,-0.11468615E+02,-0.22709808E+01,
              0.12029746E+01, },
            { 0.74724304E+02, 0.34598545E+02, 0.32904587E+02, 0.59504917E+02,
             -0.45993324E+02,-0.96503580E+00,-0.41161522E+02, 0.22533396E+02,
              0.26252196E+01, 0.52893238E+01,-0.82092028E+01, 0.18131802E+02,
             -0.36847976E+00, },
            {-0.19413287E+03,-0.49194946E+02, 0.61412071E+02,-0.15853179E+02,
             -0.53150242E+02, 0.39649019E+01,-0.23896528E+02, 0.79412022E+01,
             -0.10084863E+01, 0.10269410E+02, 0.12102466E+02, 0.25096750E+01,
             -0.16858004E+01, },
            {-0.42284904E+02,-0.29988331E+02,-0.47852016E+02,-0.47470257E+02,
              0.53297440E+02, 0.41365795E+01, 0.38657391E+02,-0.24937689E+02,
             -0.36407509E+01,-0.61866503E+01, 0.11591171E+02,-0.19596478E+02,
             -0.89709473E+00, },
            { 0.75968063E+02, 0.19798077E+02,-0.24477795E+02, 0.62807326E+01,
              0.22585154E+02,-0.18408319E+01, 0.91339474E+01,-0.30013289E+01,
              0.41616115E+00,-0.40143032E+01,-0.44904690E+01,-0.10010843E+01,
              0.77694696E+00, },
            { 0.95251894E+01, 0.10003698E+02, 0.20788961E+02, 0.13931005E+02,
             -0.21926090E+02,-0.26575904E+01,-0.13174495E+02, 0.97861652E+01,
              0.17670221E+01, 0.25980010E+01,-0.51927876E+01, 0.73529782E+01,
              0.56848937E+00, },
            {-0.20119770E+00, 0.15843167E+01, 0.18475239E+01,-0.47732759E-03,
             -0.13149090E+00,-0.13344645E-01, 0.26832491E-01, 0.30770525E-01,
             -0.20653838E-01, 0.10311009E-02, 0.70003867E-02,-0.62839808E-02,
             -0.99203400E-02, },
            { 0.70720062E-01,-0.16220536E+01, 0.15745577E+01, 0.24463184E-01,
              0.20518380E+00, 0.19001585E-01, 0.56619728E-02,-0.39436924E-02,
              0.51251244E-01,-0.10332140E-01, 0.11076461E-01,-0.14877492E-01,
              0.19403499E-01, },
            {-0.90413511E+00, 0.15951061E+01,-0.32156929E+00, 0.67641193E+00,
             -0.58352768E+00, 0.14265889E+00, 0.43839268E-01,-0.37467858E+00,
              0.20368481E+00, 0.66211246E-01, 0.16507994E+00, 0.45068026E-01,
              0.21851064E+00, },
            {-0.25522104E+00, 0.10744387E+01, 0.77017993E+00, 0.52008659E+00,
              0.31209913E+00, 0.15328670E+00,-0.16450721E+00, 0.30181509E-01,
              0.98174199E-01,-0.20944310E-01,-0.68538651E-01, 0.44294223E-01,
              0.10606261E+00, },
            { 0.43100529E+01, 0.10209758E+02, 0.10042840E+02, 0.17070240E+01,
              0.11407772E+00, 0.15409266E+01,-0.69876671E+00,-0.72542679E+00,
              0.74991882E+00,-0.28893912E+00,-0.34130789E-01, 0.30699372E+00,
              0.49604353E+00, },
            {-0.37056020E+00,-0.13278563E+02, 0.55308051E+01, 0.54367453E+00,
             -0.10339448E+01,-0.52278483E+00, 0.11216925E+01,-0.80142033E+00,
             -0.27339029E+00, 0.27401743E-01, 0.37330829E-02, 0.88551909E+00,
             -0.19130912E+00, },
            { 0.15317790E+02,-0.13764858E+02, 0.45487285E+01,-0.40844550E+01,
              0.10961870E+02,-0.27308998E+01,-0.39546218E+01, 0.59212508E+01,
             -0.20589113E+01,-0.41854739E+00,-0.23430898E+01,-0.11522582E+01,
             -0.30893290E+01, },
            { 0.34036417E+01,-0.14668043E+02,-0.73346114E+01,-0.88265228E+01,
             -0.11719311E+02,-0.30766165E+01, 0.42914267E+01,-0.19813076E+01,
             -0.91186047E+00,-0.14789377E+01, 0.34925082E+00, 0.82151622E+00,
             -0.14842312E+01, },
            {-0.21855589E+02,-0.65027206E+02,-0.30625284E+02,-0.61926742E+01,
             -0.35101626E+01,-0.17476919E+02, 0.71970272E+00, 0.40459504E+01,
             -0.35212376E+01, 0.19424437E+01,-0.18655061E+01,-0.29581990E+01,
             -0.35298369E+01, },
            { 0.45670981E+01, 0.37643353E+02,-0.25258648E+02,-0.83348694E+01,
              0.25785327E+01, 0.53328462E+01,-0.97375669E+01, 0.63686438E+01,
              0.81883103E+00,-0.19984325E+01,-0.17359768E+01,-0.85691080E+01,
              0.44630432E+00, },
            {-0.83068283E+02, 0.81330490E+02,-0.13745291E+02,-0.39484618E+01,
             -0.48257332E+02, 0.27937885E+02, 0.28801771E+02,-0.31976980E+02,
              0.32318890E+01, 0.10558324E+01, 0.14404617E+02, 0.91635990E+01,
              0.14174400E+02, },
            {-0.37831264E+02, 0.88615479E+02, 0.85113358E+02, 0.45270039E+02,
              0.76213264E+02, 0.17117207E+02,-0.29024532E+02, 0.17087751E+02,
              0.68407333E+00, 0.12642257E+02, 0.57222117E-01,-0.97055807E+01,
              0.69868464E+01, },
            { 0.33955742E+02, 0.13318353E+03,-0.21473759E+02, 0.25355750E+00,
              0.13797762E+02, 0.59645039E+02, 0.50354247E+01,-0.86467590E+01,
              0.35548909E+01,-0.50609388E+01, 0.10475528E+02, 0.11849051E+02,
              0.96025867E+01, },
            {-0.88309050E+01, 0.25575996E+02, 0.12401312E+02, 0.32052364E+02,
             -0.87352157E+00,-0.18661947E+02, 0.23053522E+02,-0.18508951E+02,
             -0.16507331E+01, 0.99664297E+01, 0.73132496E+01, 0.25814552E+02,
             -0.15154300E+01, },
            { 0.21601781E+03,-0.21782782E+03, 0.13970534E+02, 0.51641003E+02,
              0.82136398E+02,-0.96018127E+02,-0.80168480E+02, 0.78118790E+02,
              0.89466515E+01, 0.96090563E-01,-0.40393982E+02,-0.29548920E+02,
             -0.28057667E+02, },
            { 0.11636549E+03,-0.24491010E+03,-0.28724524E+03,-0.10040260E+03,
             -0.19873129E+03,-0.43567253E+02, 0.75220818E+02,-0.51479694E+02,
              0.52787185E+01,-0.37214550E+02,-0.12265165E+01, 0.29498684E+02,
             -0.13845214E+02, },
            {-0.66059847E+01,-0.12299647E+03, 0.11331442E+03, 0.16227594E+02,
             -0.18005981E+02,-0.79721268E+02,-0.89493876E+01, 0.85854511E+01,
              0.19708027E+01, 0.53805790E+01,-0.17981739E+02,-0.18486019E+02,
             -0.10671741E+02, },
            { 0.65107279E+01,-0.14133873E+03, 0.35762993E+02,-0.47331753E+02,
             -0.99259748E+01, 0.24949978E+02,-0.18619167E+02, 0.21813089E+02,
              0.29833362E+01,-0.15825300E+02,-0.94992371E+01,-0.31681845E+02,
              0.30539851E+01, },
            {-0.26602240E+03, 0.25943344E+03,-0.85672159E+01,-0.89114342E+02,
             -0.59151909E+02, 0.12887794E+03, 0.96713982E+02,-0.87241043E+02,
             -0.24014769E+02,-0.39270320E+01, 0.50494331E+02, 0.40384113E+02,
              0.24323547E+02, },
            {-0.12623465E+03, 0.30866837E+03, 0.37077533E+03, 0.10442806E+03,
              0.23109668E+03, 0.52738197E+02,-0.84341431E+02, 0.64608040E+02,
             -0.10220551E+02, 0.46504440E+02, 0.22121127E+00,-0.34443577E+02,
              0.11775063E+02, },
            {-0.12208918E+02, 0.45470352E+02,-0.71684982E+02,-0.12210121E+02,
              0.75946212E+01, 0.36560390E+02, 0.36604092E+01,-0.33766203E+01,
             -0.28442261E+01,-0.19091837E+01, 0.96889944E+01, 0.94606447E+01,
              0.40775514E+01, },
            {-0.14950647E+01, 0.93858498E+02,-0.31066874E+02, 0.23789320E+02,
              0.11685203E+02,-0.11222384E+02, 0.36598971E+01,-0.87570076E+01,
             -0.21024263E+01, 0.80605850E+01, 0.36843083E+01, 0.13835386E+02,
             -0.17752295E+01, },
            { 0.12323454E+03,-0.11393223E+03, 0.28638356E+01, 0.45115490E+02,
              0.14615677E+02,-0.59010235E+02,-0.42132607E+02, 0.35980793E+02,
              0.14087955E+02, 0.33986540E+01,-0.22851339E+02,-0.19381165E+02,
             -0.74500413E+01, },
            { 0.42877964E+02,-0.14127107E+03,-0.16386626E+03,-0.42104141E+02,
             -0.10067329E+03,-0.24085188E+02, 0.34423313E+02,-0.28762308E+02,
              0.50822926E+01,-0.20915609E+02, 0.88597655E+00, 0.13771997E+02,
             -0.35257232E+01, },
            { 0.42188969E-01, 0.94234280E-01,-0.10187069E-02,-0.95930177E+00,
             -0.19991234E+00, 0.58659386E-01, 0.31168277E-02,-0.84970305E-02,
              0.53929232E-01,-0.65914979E-02, 0.57587521E-02,-0.14161857E-01,
              0.11686263E-03, },
            { 0.43799311E-01, 0.11469412E-01, 0.45734111E-01, 0.27424732E+00,
             -0.63086385E+00, 0.69261733E-02, 0.10186278E-02,-0.43821827E-01,
              0.95467009E-02,-0.14789114E-01, 0.76593342E-02,-0.46657752E-02,
              0.73617133E-02, },
            { 0.49559942E+00, 0.82897675E+00, 0.47035179E+00, 0.17087352E+00,
             -0.12786779E+01, 0.15497789E-01,-0.29272199E+00, 0.50515074E-01,
              0.95190831E-01, 0.12434431E+00, 0.34540424E-02,-0.43415297E-01,
              0.11710780E+00, },
            { 0.91401197E-01, 0.34288970E+00, 0.77033257E+00, 0.19686368E+01,
              0.12333189E-02, 0.10140464E+00, 0.14774442E+00,-0.73128659E-02,
             -0.67382760E-01,-0.95939159E-01, 0.81297606E-01, 0.10078107E+00,
              0.75467929E-01, },
            { 0.15210741E+01, 0.97265679E+00,-0.11593559E+00, 0.81999540E+01,
              0.12909287E+00,-0.11842973E+01,-0.96201396E+00, 0.31217158E+00,
             -0.70322382E+00,-0.43442760E-01,-0.39182207E+00,-0.10794251E+00,
             -0.17477946E-01, },
            { 0.28209305E+01, 0.63783765E+00, 0.71414834E+00,-0.10531855E+01,
              0.46761045E+01, 0.80202073E+00, 0.12529795E-01, 0.56188387E+00,
             -0.26510638E+00, 0.43705702E+00, 0.11708755E+00,-0.95809540E-02,
              0.27964649E+00, },
            {-0.60612836E+01,-0.50309405E+01,-0.43330379E+01, 0.93652105E+01,
              0.22353725E+01, 0.19453353E+00, 0.13129736E+01,-0.80170625E+00,
             -0.83791220E+00,-0.26327208E+00, 0.91290653E+00, 0.65617907E+00,
             -0.94384933E+00, },
            {-0.78995579E+00,-0.36866460E+01,-0.43037763E+01,-0.60018635E+01,
              0.95902605E+01, 0.52022409E+00,-0.14935763E+01,-0.27530113E+00,
              0.18274926E+00, 0.21199745E+00,-0.55878675E+00,-0.25891751E+00,
              0.83397602E-03, },
            {-0.13319073E+02,-0.63696756E+01, 0.26987333E+01,-0.92368822E+01,
              0.32915807E+01, 0.60375071E+01, 0.76720295E+01,-0.18980306E+01,
              0.46611791E+01, 0.43055254E+00, 0.20732810E+01, 0.14235936E+01,
              0.31788766E+00, },
            {-0.14097284E+02,-0.46897783E+01,-0.82707202E+00,-0.41121367E+00,
              0.23824086E+01,-0.26390638E+01, 0.35508257E+00,-0.23547945E+01,
              0.10697405E+01,-0.21197121E+01,-0.27898791E+01, 0.12480121E+01,
             -0.18391019E+01, },
            { 0.15204087E+02, 0.78556147E+01, 0.88966312E+01,-0.30797503E+02,
              0.23227117E+01,-0.58457482E+00,-0.71814245E+00, 0.30694103E+01,
              0.22050869E+01,-0.34793630E+00,-0.29825599E+01,-0.19286658E+01,
              0.21556101E+01, },
            { 0.39697590E+01, 0.58501611E+01, 0.62696495E+01, 0.49510407E+01,
             -0.31089190E+02,-0.24617631E+01, 0.39447894E+01, 0.16149390E+01,
             -0.56727923E-01, 0.75868088E+00, 0.79545927E+00, 0.32469195E+00,
             -0.90026625E-01, },
            { 0.27027122E+02, 0.11467045E+02,-0.84330359E+01,-0.24010477E+02,
             -0.80087805E+01,-0.12543470E+02,-0.15589190E+02, 0.42210331E+01,
             -0.10726813E+02,-0.14858339E+01,-0.43507633E+01,-0.34340184E+01,
             -0.92487770E+00, },
            { 0.23143436E+02, 0.78808756E+01,-0.27895360E+01, 0.60927572E+01,
             -0.40888386E+02, 0.94554806E+00,-0.24960201E+01, 0.34536648E+01,
             -0.18722475E+01, 0.37619510E+01, 0.80734539E+01,-0.44939146E+01,
              0.28656080E+01, },
            {-0.11953168E+02,-0.27238479E+01,-0.52568307E+01, 0.21050129E+02,
             -0.68679848E+01, 0.32113910E+00,-0.58306420E+00,-0.27074990E+01,
             -0.17911482E+01, 0.77064550E+00, 0.22215376E+01, 0.15297557E+01,
             -0.14433203E+01, },
            {-0.47204428E+01,-0.10654535E+01,-0.18083954E+01, 0.15844553E+01,
              0.20997583E+02, 0.19634615E+01,-0.33727379E+01,-0.18170273E+01,
             -0.17987987E+00,-0.10957071E+01,-0.20409736E+00,-0.20305949E+00,
             -0.28999722E+00, },
            {-0.15884968E+02,-0.64883618E+01, 0.73605752E+01, 0.31496780E+02,
              0.69600091E+01, 0.86443911E+01, 0.96710367E+01,-0.29579740E+01,
              0.75072083E+01, 0.12868271E+01, 0.29590070E+01, 0.23414221E+01,
              0.84001136E+00, },
            {-0.11925214E+02,-0.50585051E+01, 0.18607297E+01,-0.73451662E+01,
              0.40590572E+02, 0.12435257E+01, 0.30310061E+01,-0.15552270E+01,
              0.12240763E+01,-0.21083274E+01,-0.60944834E+01, 0.37612362E+01,
             -0.10206302E+01, },
            { 0.64615183E-01, 0.78795552E-01,-0.15892947E+00,-0.10464617E+00,
              0.10353974E+00,-0.41167954E+00,-0.43749309E+00, 0.20638651E-02,
              0.74383467E-01,-0.43644945E-03,-0.90350350E-03,-0.19853504E-01,
             -0.59061525E-02, },
            { 0.81961580E-01,-0.21425916E+00,-0.38655456E-01,-0.88595822E-02,
             -0.14763263E+00, 0.46912336E+00,-0.45955449E+00,-0.42392034E-01,
             -0.80482624E-02, 0.18471889E-01,-0.14141251E-01, 0.55350456E-02,
              0.16753854E-01, },
            {-0.24682516E+00, 0.65146202E+00,-0.45703122E+00,-0.23654184E+00,
             -0.30039293E+00,-0.82151897E-01,-0.20593224E+00,-0.13945434E-01,
              0.16855291E+00, 0.32676257E-01, 0.80637790E-01,-0.79479448E-01,
             -0.61515283E-01, },
            {-0.29577321E+00,-0.25834915E+00, 0.53513624E-01, 0.29484292E-01,
             -0.17856269E+00, 0.19976071E+00, 0.16096137E+00,-0.13107130E+00,
             -0.10473933E+00,-0.19589696E-01, 0.30329490E-01, 0.64674303E-01,
             -0.38300402E-01, },
            {-0.60273278E+00,-0.20669493E+00, 0.82746637E+00, 0.46272588E+00,
             -0.10339417E+00, 0.15778165E+00,-0.30000055E+00,-0.28547913E+00,
              0.21799399E-01,-0.91843419E-01,-0.55076419E-02, 0.13837788E-01,
             -0.14949028E+00, },
            {-0.13389989E+01, 0.26402633E+01,-0.13628585E+01,-0.36753692E-01,
              0.15853026E+01, 0.13760415E+00, 0.79008830E+00,-0.23089467E+00,
              0.14146335E+00,-0.11153022E+00, 0.23030405E+00,-0.12568083E+00,
             -0.15787458E+00, },
            {-0.16602457E+00,-0.13449451E+01, 0.96737868E+00, 0.40215719E+00,
              0.72849149E+00, 0.41462141E+00, 0.71418375E+00, 0.10896401E+00,
             -0.35342368E+00,-0.11979083E+00,-0.10750519E+00, 0.23451699E+00,
              0.13568750E+00, },
            { 0.14382593E-01, 0.82095027E+00,-0.72794154E-01,-0.23630470E+00,
              0.75462168E+00,-0.65143400E+00, 0.49917012E-01, 0.24157877E+00,
              0.31928110E+00,-0.23850443E-01,-0.63626952E-01,-0.70730507E-01,
              0.35635930E-01, },
            { 0.16258487E+01, 0.63477248E+00,-0.11753286E+01,-0.52329248E+00,
              0.51585090E+00, 0.42656636E+00, 0.12154765E+01, 0.54212022E+00,
             -0.33579046E+00, 0.13298018E+00,-0.58626540E-01,-0.93792476E-01,
              0.21695583E+00, },
            { 0.19641677E+01,-0.34773593E+01, 0.24826541E+01,-0.47208840E+00,
             -0.22139666E+01,-0.79506546E+00,-0.24327517E+00, 0.67086709E+00,
             -0.30561268E-01, 0.17824247E+00,-0.29080093E+00, 0.13225445E+00,
              0.22372004E+00, },
            { 0.78067645E-01,-0.53872090E-01,-0.60876328E-01,-0.20090356E-01,
              0.19590052E-01, 0.37616719E-01,-0.21996699E-01, 0.30491668E+00,
             -0.21350857E-01, 0.73222145E-02, 0.10664701E-01,-0.17186703E-01,
             -0.79840180E-02, },
            { 0.54252598E-01,-0.42442694E-01, 0.41468631E-01, 0.11270581E-01,
             -0.79623416E-01,-0.27057895E-01, 0.20517232E-01, 0.19652473E-01,
              0.27900285E+00, 0.76455669E-02,-0.26138762E-02, 0.26082578E-02,
             -0.48174490E-02, },
            { 0.37738148E-01, 0.49505927E-01, 0.45610685E-01,-0.40265694E-01,
             -0.23026168E+00,-0.11767189E+00,-0.47708195E-01, 0.25821838E+00,
              0.16750067E+00,-0.34544688E-01,-0.48518635E-01, 0.20903081E-01,
              0.94813956E-02, },
            { 0.18504436E+00, 0.13707677E+00,-0.76823071E-01, 0.73215045E-01,
              0.25763610E+00,-0.57229973E-01,-0.43995194E-01,-0.27386624E+00,
              0.97568236E-01, 0.11287920E-01,-0.45958824E-01,-0.17281851E-01,
             -0.18543916E-01, },
            {-0.10852473E+00,-0.31554475E-01,-0.29181827E-01,-0.35038237E-01,
             -0.30168429E-01, 0.12917520E-01, 0.94219139E-02, 0.38543744E-02,
             -0.17429788E-01, 0.97568251E-01, 0.17038140E+00,-0.22542167E-02,
             -0.68581281E-02, },
            {-0.16812027E-01, 0.67123897E-01,-0.14856836E-01, 0.52298285E-01,
             -0.10310156E+00,-0.40235251E-01,-0.89705503E-02, 0.38926877E-01,
             -0.10838482E-02,-0.16985069E+00, 0.76311469E-01, 0.69150580E-02,
             -0.63182632E-02, },
            {-0.48873704E-01,-0.22341227E-01, 0.31494450E-01,-0.74725705E-02,
             -0.32325108E-01, 0.47166054E-02, 0.15750030E-01, 0.19931326E-01,
              0.61885430E-02,-0.89711556E-02,-0.22252218E-02,-0.13308422E+00,
              0.20251697E-01, },
            {-0.45145783E-02, 0.35775520E-01,-0.10440513E-01, 0.63793033E-01,
              0.43009203E-01,-0.34634177E-01,-0.14203522E-01, 0.64703203E-02,
             -0.13139685E-01,-0.32075319E-01,-0.19045237E-01,-0.12680903E-01,
             -0.13717172E+00, },
            { 0.29166550E-02, 0.11035411E-01, 0.24293691E-01, 0.25893958E-01,
             -0.13120287E-02,-0.17062493E-01,-0.69651459E-02, 0.15115200E-01,
             -0.12452925E-01,-0.28048914E-01,-0.81404569E-02, 0.14331407E-01,
              0.12644852E-01, },
            { 0.24846908E-01,-0.53721450E-01, 0.16348697E-01, 0.18280599E-01,
             -0.24520375E-01,-0.36987716E-02, 0.18260539E-01,-0.72437376E-02,
              0.14151407E-01,-0.22939686E-01, 0.24036895E-02, 0.15635617E-01,
              0.12054204E-01, },
            { 0.11425200E-01, 0.64702697E-01,-0.10536327E+00, 0.15012830E-01,
             -0.28842267E-01, 0.11042488E-01, 0.13311166E-01,-0.21481495E-02,
              0.73922775E-02,-0.16001927E-01, 0.10468615E-01,-0.66283136E-02,
             -0.15343135E-02, },
            {-0.30380554E-01,-0.51950170E-02,-0.93821893E-02, 0.29672759E-01,
              0.30377174E-01, 0.16042337E-01,-0.11458547E-01,-0.95287329E-02,
              0.13136099E-01, 0.89334687E-02,-0.17647015E-01,-0.20479793E-02,
              0.17140299E-01, },
         },
         {
            { 0.84415445E+01, 0.26355572E-01, 0.16610390E+00, 0.69635049E-01,
              0.77834222E-02,-0.25760174E-01,-0.21441121E-01, 0.21861907E-01,
             -0.43986645E-01, 0.11674735E-01, 0.56952261E-02, 0.28399102E-01,
             -0.13653146E-01, },
            {-0.68217140E+00,-0.62160009E+00, 0.28212479E+00,-0.24901924E+00,
              0.21900199E+00, 0.24887722E-01, 0.72718799E-01,-0.13336083E+00,
             -0.78881793E-01, 0.14562245E+00, 0.74521393E-01,-0.27514610E-01,
              0.10257617E+00, },
            { 0.17437782E+02, 0.79321623E+00,-0.35100787E+01, 0.84811628E-01,
              0.70552386E-01, 0.47351521E+00, 0.83660269E+00,-0.47608498E+00,
              0.52839464E+00,-0.69492771E-02,-0.20629799E+00,-0.32985020E+00,
              0.31779465E+00, },
            {-0.81596603E+01, 0.80805292E+01, 0.59097290E+01, 0.42979136E+01,
             -0.42025328E+01,-0.14056302E+01,-0.12244359E+01, 0.21699390E+01,
              0.52206904E+00,-0.24186170E+01,-0.21609974E+01, 0.93807650E+00,
             -0.84871131E+00, },
            {-0.89978989E+02,-0.97939730E+01, 0.20942669E+02,-0.55313463E+01,
             -0.76858745E+01,-0.21651825E+00,-0.45114717E+01, 0.38345082E+01,
             -0.16243830E+01,-0.89623278E+00, 0.11692472E+01, 0.93201983E+00,
             -0.17749968E+01, },
            { 0.10546338E+03,-0.44166641E+02,-0.43877007E+02,-0.16669228E+02,
              0.34683037E+02, 0.49835873E+01, 0.66487484E+01,-0.12427212E+02,
             -0.19051713E+01, 0.11908665E+02, 0.12564396E+02,-0.36486394E+01,
              0.21449244E+01, },
            { 0.13710806E+03, 0.36479836E+02,-0.52778606E+02, 0.21492033E+02,
              0.30154282E+02,-0.53919759E+01, 0.90821161E+01,-0.11428968E+02,
              0.14775454E+01, 0.36835520E+01,-0.33135943E+01,-0.38186175E+00,
              0.41576314E+01, },
            {-0.28806189E+03, 0.10074989E+03, 0.97997749E+02, 0.22048018E+02,
             -0.10003945E+03,-0.56709027E+01,-0.15209171E+02, 0.29163034E+02,
              0.34512579E+01,-0.24924685E+02,-0.28658514E+02, 0.41394582E+01,
             -0.29731119E+01, },
            {-0.92897758E+02,-0.49600414E+02, 0.61100552E+02,-0.29214361E+02,
             -0.39369938E+02, 0.10823676E+02,-0.76445956E+01, 0.13982789E+02,
              0.32520753E+00,-0.46346078E+01, 0.43214622E+01,-0.86068255E+00,
             -0.42901020E+01, },
            { 0.30256940E+03,-0.99049271E+02,-0.91175804E+02,-0.90334110E+01,
              0.11542465E+03, 0.17341838E+01, 0.15239948E+02,-0.29818983E+02,
             -0.28255911E+01, 0.23323519E+02, 0.28412251E+02,-0.63854128E+00,
              0.26830745E+01, },
            { 0.24605694E+02, 0.22045202E+02,-0.26002682E+02, 0.13213135E+02,
              0.16756659E+02,-0.57091608E+01, 0.22745020E+01,-0.59658070E+01,
             -0.66697073E+00, 0.18124434E+01,-0.19950210E+01, 0.59119934E+00,
              0.16144733E+01, },
            {-0.11073072E+03, 0.35112747E+02, 0.30812071E+02,-0.38981718E+00,
             -0.46086620E+02, 0.31463280E+00,-0.55018616E+01, 0.11076973E+02,
              0.82669556E+00,-0.80218763E+01,-0.10252825E+02,-0.76681507E+00,
             -0.11369579E+01, },
            {-0.24553841E+00, 0.17193551E+01, 0.16381207E+01, 0.33417907E-01,
             -0.87884702E-01,-0.44835400E-01, 0.15925823E-02, 0.49698550E-01,
             -0.38588453E-01,-0.72526657E-02, 0.15999507E-01, 0.17281709E-01,
             -0.23321385E-01, },
            { 0.24167123E+00,-0.16023287E+01, 0.16606361E+01,-0.45682013E-01,
              0.11704248E+00,-0.18465983E-01, 0.11106687E-01,-0.60868151E-02,
              0.61217647E-01,-0.20567151E-01, 0.24912491E-01,-0.28536351E-01,
              0.29306814E-01, },
            {-0.10224926E+01, 0.14925382E+01,-0.19260072E+01, 0.61678416E+00,
              0.13985469E-01, 0.10479474E+00,-0.49587530E-02,-0.18433394E+00,
              0.11050394E+00, 0.21527071E+00, 0.22213718E+00, 0.34993712E-01,
              0.57391070E-01, },
            { 0.10419737E+01, 0.66445446E+00, 0.14948963E+01, 0.38029797E-01,
              0.27230132E+00, 0.42719600E+00,-0.11771204E+00, 0.30687597E+00,
              0.19784185E+00, 0.22538267E+00, 0.75966686E-01, 0.23985893E-01,
              0.45384716E-01, },
            { 0.53208251E+01, 0.75190620E+01,-0.22172494E+01,-0.12778753E+00,
              0.56682128E-01, 0.18827899E+01, 0.72104907E+00,-0.80092132E+00,
              0.77964395E+00,-0.24134891E+00,-0.46352610E+00, 0.28311205E+00,
              0.10742375E+01, },
            {-0.41045341E+01, 0.57498431E+01, 0.83945713E+01, 0.23423393E+01,
             -0.14150887E+01,-0.54918933E+00,-0.10221728E+01, 0.26822236E+00,
             -0.12929893E+01, 0.97145742E+00,-0.42177832E+00, 0.68537563E+00,
             -0.58753085E+00, },
            { 0.14998870E+02,-0.16657772E+02, 0.37549644E+02,-0.95424957E+01,
             -0.54443731E+01,-0.14911759E+01, 0.17992316E+01, 0.51960616E+01,
              0.45644295E+00,-0.39563315E+01,-0.49043798E+01, 0.12105607E+00,
              0.70748419E+00, },
            {-0.60123081E+01, 0.27362645E+01,-0.13724414E+02, 0.12894962E+01,
             -0.14900351E+02,-0.52518587E+01, 0.34944665E+01,-0.55088906E+01,
             -0.13576542E+01,-0.56548223E+01,-0.14211199E+01,-0.34091601E+00,
              0.21933839E+00, },
            {-0.29305424E+02, 0.13374522E+02, 0.44242222E+02, 0.50953383E+01,
              0.65631719E+01,-0.14111629E+02,-0.65622892E+01, 0.39515307E+01,
             -0.34545369E+01,-0.12505515E+00, 0.29348438E+01,-0.32276697E+01,
             -0.90113583E+01, },
            { 0.43037151E+02,-0.89085785E+02, 0.11644012E+02,-0.24075853E+02,
              0.19164951E+02, 0.77643185E+01, 0.86020861E+01,-0.46943688E+01,
              0.93667440E+01,-0.74655390E+01, 0.14679497E+01,-0.47342935E+01,
              0.18102796E+01, },
            {-0.85623055E+02, 0.92817024E+02,-0.23375591E+03, 0.44671661E+02,
              0.45343346E+02, 0.83043785E+01,-0.13587065E+02,-0.29296474E+02,
             -0.10942899E+02, 0.20286388E+02, 0.29787083E+02,-0.44528899E+00,
             -0.79419918E+01, },
            {-0.14370024E+02,-0.23429798E+02, 0.72968773E+02,-0.15687455E+02,
              0.10109998E+03, 0.26097256E+02,-0.26932791E+02, 0.35635754E+02,
             -0.34211284E+00, 0.35934021E+02, 0.61775665E+01, 0.22777042E+01,
             -0.23924460E+01, },
            { 0.57428696E+02,-0.17855730E+03,-0.15708559E+03,-0.22408588E+02,
             -0.40592297E+02, 0.40333138E+02, 0.14209915E+02,-0.89817600E+01,
              0.57742748E+01, 0.57205853E+01,-0.63581610E+01, 0.10133678E+02,
              0.26636444E+02, },
            {-0.14794983E+03, 0.30004440E+03,-0.19929813E+03, 0.84621696E+02,
             -0.69814247E+02,-0.29103554E+02,-0.25958654E+02, 0.20315407E+02,
             -0.28830307E+02, 0.23460487E+02,-0.31325624E+01, 0.13810139E+02,
             -0.21702385E+01, },
            { 0.23797343E+03,-0.23116394E+03, 0.57393121E+03,-0.83598228E+02,
             -0.12368264E+03,-0.19707600E+02, 0.34384033E+02, 0.70111160E+02,
              0.39034550E+02,-0.41825325E+02,-0.75330421E+02,-0.22089810E+01,
              0.25101606E+02, },
            { 0.10284250E+03, 0.69117165E+02,-0.15603969E+03, 0.39777988E+02,
             -0.25606186E+03,-0.60250473E+02, 0.77766205E+02,-0.99222710E+02,
              0.16326077E+02,-0.91902000E+02,-0.83012609E+01,-0.85460234E+01,
              0.51328115E+01, },
            {-0.39939327E+02, 0.31202109E+03, 0.21929730E+03, 0.33807743E+02,
              0.73635201E+02,-0.49816437E+02,-0.86225977E+01, 0.10290629E+02,
             -0.33790231E+01,-0.12311239E+02, 0.48300257E+01,-0.12229668E+02,
             -0.32109089E+02, },
            { 0.20875215E+03,-0.39679617E+03, 0.38521735E+03,-0.12168412E+03,
              0.91634483E+02, 0.41888870E+02, 0.34228813E+02,-0.33702576E+02,
              0.39293419E+02,-0.32169025E+02, 0.41854286E+01,-0.18185402E+02,
              0.11737410E+01, },
            {-0.29656155E+03, 0.26091183E+03,-0.63637982E+03, 0.62630611E+02,
              0.13731348E+03, 0.20191547E+02,-0.35748486E+02,-0.76228554E+02,
             -0.50450886E+02, 0.36267136E+02, 0.83756058E+02, 0.57476335E+01,
             -0.32509529E+02, },
            {-0.15768517E+03,-0.66360435E+02, 0.12792127E+03,-0.30300962E+02,
              0.27846167E+03, 0.64009613E+02,-0.94658752E+02, 0.12312606E+03,
             -0.32563572E+02, 0.10454442E+03, 0.17428432E+01, 0.15118477E+02,
             -0.33578801E+01, },
            { 0.47491126E+01,-0.15489160E+03,-0.10394126E+03,-0.16034281E+02,
             -0.40621128E+02, 0.22269804E+02,-0.31044576E+00,-0.47275276E+01,
              0.55015851E-01, 0.71959572E+01,-0.86488312E+00, 0.50432911E+01,
              0.13676483E+02, },
            {-0.10171490E+03, 0.18161353E+03,-0.21165048E+03, 0.60878799E+02,
             -0.37977203E+02,-0.20529871E+02,-0.16520407E+02, 0.18732552E+02,
             -0.19229500E+02, 0.15826124E+02,-0.22000675E+01, 0.87317457E+01,
             -0.31610513E+00, },
            { 0.13370317E+03,-0.11240935E+03, 0.26245120E+03,-0.15337643E+02,
             -0.54398438E+02,-0.74560561E+01, 0.13048158E+02, 0.30778091E+02,
              0.21956047E+02,-0.10538049E+02,-0.33810032E+02,-0.32820210E+01,
              0.14922835E+02, },
            { 0.75884071E+02, 0.17452677E+02,-0.31479095E+02, 0.36800358E+01,
             -0.11180386E+03,-0.25325590E+02, 0.41215767E+02,-0.55677891E+02,
              0.18475767E+02,-0.44003674E+02, 0.18647236E+01,-0.90560875E+01,
              0.22126122E+00, },
            { 0.59121020E-01,-0.38882446E-01, 0.10154726E+00,-0.99066663E+00,
              0.50862730E+00, 0.37865343E-02,-0.13369663E-01,-0.14525892E-01,
              0.73863156E-02,-0.30057946E-01, 0.92732534E-02,-0.24269311E-01,
              0.67565674E-02, },
            {-0.24163937E-01,-0.20808218E+00, 0.18645240E-02,-0.51261657E+00,
             -0.78584683E+00,-0.29707689E-01,-0.29643602E-02,-0.43101724E-01,
              0.92033194E-02,-0.20032544E-01, 0.12454584E-01,-0.10859430E-01,
              0.14212405E-01, },
            {-0.20849434E+00, 0.50730848E+00, 0.54870909E+00, 0.17681440E+01,
             -0.10749120E+01, 0.42078006E+00,-0.26223657E+00, 0.33963424E+00,
              0.82626760E-01, 0.92897750E-01,-0.60189381E-01,-0.76034263E-01,
              0.18571973E-01, },
            { 0.42013156E+00, 0.22726022E+00, 0.90818733E+00, 0.16409394E+01,
              0.20787294E+01, 0.16166556E+00, 0.25318131E+00, 0.91371499E-03,
             -0.12026191E+00, 0.11957009E-01,-0.17968928E-01, 0.20824531E+00,
              0.10581681E+00, },
            { 0.18979921E+01, 0.71559265E-01, 0.89219481E+00,-0.64803618E+00,
              0.19108589E+01, 0.61305553E-01,-0.76099968E+00, 0.66523030E-02,
              0.38209528E+00,-0.37404042E+00,-0.24693862E-01,-0.15020821E-01,
             -0.21455880E+00, },
            { 0.42901478E+01, 0.27137227E+01, 0.18731354E+01,-0.25761666E+01,
             -0.37876692E-01, 0.42818061E+00,-0.39081356E+00, 0.64281893E+00,
              0.43348350E-01, 0.28045836E+00,-0.51489584E-01, 0.33649018E+00,
              0.13021860E+00, },
            { 0.61204880E+00,-0.18564703E+01,-0.39929540E+01, 0.96122243E-01,
              0.16075824E+01,-0.39626312E+01, 0.11777747E+01,-0.33050480E+01,
             -0.11639585E+01,-0.89641178E+00, 0.77148652E+00, 0.10184772E+01,
              0.13489571E+00, },
            {-0.77427740E+01,-0.11954879E+01,-0.10486695E+02,-0.58320417E+01,
             -0.18617257E+01, 0.74399304E+00,-0.28453026E+01,-0.58667099E+00,
              0.11166003E+01,-0.77045232E+00, 0.64419031E+00,-0.96420527E+00,
             -0.97287238E+00, },
            {-0.14225009E+02, 0.82643967E+01,-0.12470506E+02, 0.23383005E+02,
             -0.65144601E+01, 0.16238860E+01, 0.87209654E+01, 0.49000898E+00,
             -0.16810759E+01, 0.40386558E+01,-0.54124254E+00, 0.10109010E+00,
              0.19018241E+01, },
            {-0.21727472E+02,-0.96902895E+01,-0.16478348E+02, 0.13098322E+02,
              0.13914478E+02,-0.20104990E+01, 0.37096817E+01,-0.27436793E+01,
             -0.43885440E-01,-0.13556918E+01, 0.84448802E+00,-0.14343554E+01,
             -0.20107965E+01, },
            {-0.89594114E+00, 0.10083544E+01, 0.83134689E+01,-0.92639027E+01,
              0.61116376E+01, 0.10380750E+02,-0.26651862E+01, 0.87228279E+01,
              0.38335390E+01, 0.22555156E+01,-0.13236306E+01,-0.31068852E+01,
             -0.87246418E+00, },
            { 0.23570587E+02,-0.42556520E+01, 0.28354004E+02, 0.33738234E+01,
             -0.40777335E+01,-0.38387177E+01, 0.82230272E+01, 0.30383515E+01,
             -0.22789021E+01, 0.26750424E+01,-0.16830761E+01, 0.11110405E+01,
              0.25620415E+01, },
            { 0.29727594E+02,-0.24605175E+02, 0.28340710E+02,-0.59987556E+02,
             -0.22684792E+00,-0.73963881E+01,-0.18819962E+02,-0.99585795E+00,
              0.24640315E+01,-0.93489466E+01, 0.17905924E+01, 0.29973716E+00,
             -0.48281670E+01, },
            { 0.32501541E+02, 0.18787926E+02, 0.36723728E+02,-0.15535311E+02,
             -0.37230057E+02, 0.20521283E+01,-0.89159498E+01, 0.26074269E+01,
             -0.13309879E+01, 0.20207708E+01,-0.26565912E+01, 0.20663900E+01,
              0.51164727E+01, },
            {-0.40171739E+00, 0.11958817E+01,-0.70398526E+01, 0.46672320E+01,
             -0.12406041E+02,-0.77181621E+01, 0.25220163E+01,-0.60711088E+01,
             -0.32043297E+01,-0.13135966E+01, 0.27531669E+00, 0.25722663E+01,
              0.94680876E+00, },
            {-0.20347054E+02, 0.94924231E+01,-0.21017023E+02, 0.51578898E+01,
              0.91074489E-01, 0.31714413E+01,-0.65355906E+01,-0.32284212E+01,
              0.13073928E+01,-0.20606008E+01, 0.10267882E+01,-0.13762021E+00,
             -0.18398629E+01, },
            {-0.18292107E+02, 0.18030376E+02,-0.15053249E+02, 0.42048473E+02,
              0.93764210E+01, 0.66455450E+01, 0.11065763E+02, 0.21237047E+00,
             -0.12657822E+01, 0.58830400E+01,-0.11597257E+01,-0.38042429E+00,
              0.34563775E+01, },
            {-0.12867409E+02,-0.16812653E+02,-0.25532946E+02, 0.17166659E+01,
              0.26918324E+02,-0.10111470E+00, 0.59210205E+01, 0.11573966E+00,
              0.15903757E+01,-0.98362416E+00, 0.20214050E+01,-0.12598089E+01,
             -0.36495934E+01, },
            { 0.28310280E-01, 0.48973437E-01,-0.48619840E-01,-0.72511427E-01,
              0.58058472E-02,-0.57069111E+00,-0.71659225E+00,-0.18692106E-02,
              0.42772554E-01,-0.31100757E-01, 0.13589856E-01,-0.41641410E-01,
             -0.74718855E-02, },
            {-0.28761316E-01,-0.29289359E+00,-0.91049284E-01, 0.28099608E-01,
             -0.29132949E-01, 0.69334489E+00,-0.55960542E+00,-0.21222344E-01,
              0.24635848E-01, 0.21211600E-01,-0.15531605E-02, 0.26932620E-02,
              0.16090758E-01, },
            { 0.20954400E+00, 0.73779237E+00,-0.42621502E+00,-0.23507123E+00,
             -0.37294054E+00, 0.17519316E+00, 0.57431810E-01,-0.47046978E-01,
              0.15076196E-01,-0.54287319E-02, 0.58389850E-01,-0.86825825E-01,
             -0.17713912E-01, },
            {-0.23256958E+00,-0.20628646E+00, 0.74116617E-01, 0.43207306E-01,
             -0.11664169E+00,-0.47020305E-01, 0.75928085E-01,-0.98211251E-01,
              0.93043060E-03, 0.55244029E-01,-0.46837240E-01,-0.95012728E-02,
             -0.13064935E-01, },
            { 0.87182951E+00, 0.98459274E-02, 0.13477666E+01, 0.84805533E-01,
             -0.90552852E-01, 0.11711707E+01, 0.11748557E+01,-0.12035688E+00,
              0.12323441E+00, 0.49199241E+00,-0.20814496E+00, 0.27044243E+00,
             -0.61024815E-01, },
            {-0.74558862E-01, 0.37783556E+01,-0.26144087E+00, 0.38623143E-01,
              0.53248072E+00,-0.10437831E+01, 0.90481389E+00,-0.44594201E+00,
             -0.40208280E+00,-0.60124949E-01,-0.48902173E-01,-0.11441462E+00,
             -0.19880749E+00, },
            {-0.13543634E+01,-0.10415508E+01, 0.12016990E+01, 0.31708425E+00,
              0.61210239E+00, 0.58425140E+00, 0.41064870E+00, 0.33918127E+00,
              0.11345342E-01, 0.93263805E-01,-0.17820968E+00, 0.18072975E+00,
              0.96731916E-01, },
            { 0.32778770E+00, 0.61795223E+00, 0.38936533E-01,-0.50377148E+00,
              0.31532276E+00,-0.38423333E+00, 0.74498868E+00, 0.17342281E+00,
              0.20069090E+00,-0.13456018E+00, 0.16104504E+00, 0.72342478E-01,
              0.79420209E-01, },
            {-0.31249234E+00, 0.92828923E+00,-0.26748803E+01, 0.70246682E-01,
              0.58934122E+00,-0.16105400E+01,-0.14977306E+01,-0.18931707E-01,
             -0.31899455E+00,-0.82864511E+00, 0.36326492E+00,-0.42674014E+00,
             -0.46573725E-01, },
            { 0.81409059E-01,-0.52734656E+01, 0.76651889E+00,-0.32308713E+00,
             -0.10347863E+01, 0.13800038E+01,-0.11256927E+01, 0.89915168E+00,
              0.71173912E+00, 0.48457179E-01, 0.25199077E+00, 0.14797410E+00,
              0.20402835E+00, },
            { 0.24246454E-01, 0.33298835E-01,-0.51717691E-01,-0.14058649E-01,
              0.17725911E-01,-0.72037068E-03,-0.37894547E-02, 0.30802149E+00,
             -0.16611566E+00,-0.17199373E-01, 0.13114575E-01,-0.76011149E-02,
             -0.20986781E-01, },
            { 0.43441545E-01, 0.33683844E-01,-0.34691116E-02, 0.59443168E-01,
             -0.51950090E-01, 0.14247886E-01,-0.17108589E-01, 0.14792427E+00,
              0.28507856E+00, 0.35790657E-02, 0.32111532E-02, 0.13911074E-02,
             -0.73967529E-02, },
            {-0.19056965E+00,-0.73565212E-02,-0.24821372E+00, 0.15964326E-01,
             -0.97779892E-01,-0.89113951E-01, 0.58998290E-01, 0.14977667E+00,
              0.37016949E+00,-0.46572641E-01,-0.31782843E-01,-0.32574493E-02,
              0.12899466E-01, },
            { 0.37356037E+00, 0.15084852E+00,-0.12477553E+00, 0.29166037E-01,
              0.21163251E+00,-0.11922070E+00,-0.11764953E+00,-0.38922527E+00,
              0.26210438E-01, 0.71545686E-02,-0.61420333E-01,-0.20083405E-01,
             -0.11609336E-01, },
            {-0.15726325E+00, 0.20085622E-01,-0.33992976E-01,-0.16686320E-01,
             -0.15578196E-01,-0.89996764E-02, 0.52281570E-01, 0.24140203E-01,
             -0.13938872E-01, 0.15512536E+00, 0.19358249E+00, 0.49229502E-02,
             -0.39371615E-02, },
            { 0.14834774E+00, 0.25739465E-01,-0.83402745E-01, 0.18065708E-01,
             -0.55616297E-01,-0.52165301E-02,-0.17087549E-01, 0.21788107E-01,
              0.38194988E-01,-0.19343592E+00, 0.16540502E+00,-0.46783774E-02,
             -0.27857188E-01, },
            { 0.57891991E-01,-0.67955382E-01, 0.42646252E-01, 0.22683810E-01,
              0.79109436E-02, 0.28160231E-01,-0.27141111E-01, 0.41366834E-01,
             -0.74150283E-02, 0.70931404E-02,-0.23410007E-01,-0.12303041E+00,
              0.72064400E-01, },
            { 0.10394365E+00, 0.57514329E-01,-0.21865284E-01, 0.21552900E-02,
              0.56937981E-01,-0.20307319E-02, 0.91909728E-03,-0.14761715E-01,
              0.11957444E-01,-0.14504174E-01,-0.67575327E-02,-0.97156227E-01,
             -0.12803575E+00, },
            {-0.39362207E-01,-0.32046013E-01,-0.55135102E-02, 0.48773892E-01,
              0.35422139E-01,-0.15434346E-01,-0.21533076E-01, 0.26582329E-01,
             -0.12274754E-01,-0.12937960E-01,-0.11175702E-01, 0.53416751E-02,
              0.67404481E-02, },
            { 0.73514357E-01,-0.38842116E-01, 0.64909115E-01, 0.56321872E-02,
             -0.52885335E-01, 0.41649371E-01,-0.11770996E-02,-0.26402390E-01,
              0.25709886E-01,-0.13926198E-02, 0.27511716E-01, 0.84407153E-02,
              0.18437810E-01, },
            { 0.43839946E-01, 0.51004354E-01,-0.11362996E+00,-0.25646824E-01,
             -0.38826078E-01,-0.55158515E-02, 0.62045278E-02,-0.30429460E-01,
              0.18208286E-01,-0.11582433E-01, 0.17949345E-01,-0.15534251E-01,
              0.51459920E-03, },
            {-0.43065980E-01, 0.55039483E-02, 0.27189227E-02, 0.13957473E-01,
              0.87091923E-01, 0.10603105E-01, 0.77987841E-03,-0.24959501E-02,
              0.29643327E-01, 0.11370151E-01,-0.18617267E-01,-0.15555475E-01,
              0.72064400E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir18Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30766914E+01,-0.92550134E-02, 0.13718505E-01, 0.25895583E-02,
              0.33574869E-03, 0.78745289E-02, 0.86438525E-02,-0.60896222E-02,
             -0.74246735E-03, },
            {-0.50321376E+00, 0.22363435E-01, 0.60551178E-01, 0.54579623E-01,
             -0.33105440E-01,-0.27306173E-01,-0.16869433E-01,-0.11755095E-01,
              0.34014808E-02, },
            { 0.10928845E+01, 0.24034415E+00,-0.11382686E+00,-0.20328368E+00,
             -0.24367245E-01,-0.92177808E-01,-0.59212130E-01, 0.92386231E-01,
              0.51138084E-02, },
            { 0.11338211E+01,-0.36324242E+00,-0.20184082E+00,-0.19428015E+00,
              0.12432440E+00, 0.92264690E-01, 0.44125546E-01, 0.36674652E-01,
              0.35087094E-02, },
            {-0.19719753E+01,-0.64643306E+00, 0.26046559E+00, 0.49098560E+00,
              0.92803150E-01, 0.28286198E+00, 0.15693869E+00,-0.19568925E+00,
             -0.20824329E-02, },
            {-0.74346578E+00, 0.34690487E+00, 0.19643439E+00, 0.14222169E+00,
             -0.96713386E-01,-0.67438833E-01,-0.23977172E-01,-0.20441962E-01,
             -0.10335058E-01, },
            { 0.92457843E+00, 0.46374896E+00,-0.11210353E+00,-0.29201853E+00,
             -0.79533793E-01,-0.21276265E+00,-0.10755368E+00, 0.10385718E+00,
             -0.11526103E-02, },
            { 0.57956658E-03,-0.12857877E+00,-0.37720919E+00, 0.71668392E-02,
             -0.15985873E-01, 0.35551095E-02,-0.34057416E-03, 0.14429940E-04,
              0.77074156E-02, },
            { 0.16609097E-01, 0.39636013E+00,-0.11226575E+00,-0.17938472E-01,
              0.14924302E-01,-0.20589654E-02,-0.79512065E-02, 0.89895235E-04,
              0.10084794E-01, },
            {-0.84668107E-01, 0.76170884E-01, 0.11493012E+00, 0.58139663E-01,
             -0.90520024E-01,-0.49981788E-01,-0.52669294E-01, 0.67823958E-02,
              0.27890796E-01, },
            { 0.11318666E+00,-0.73709123E-01, 0.20983368E-01,-0.12963885E-01,
              0.24992188E-01,-0.32588847E-02, 0.19097012E-02, 0.37765738E-01,
              0.79129241E-01, },
            { 0.79023778E-01, 0.42699623E+00, 0.21301763E+01,-0.24369007E+00,
             -0.14716794E+00,-0.96097685E-01,-0.25467388E-01, 0.65074660E-01,
             -0.70087478E-01, },
            {-0.67493325E+00,-0.20482519E+01, 0.44933867E+00, 0.34022969E+00,
             -0.98558955E-01, 0.39792750E-01, 0.53783178E-01,-0.29130100E-03,
             -0.84323861E-01, },
            { 0.10565306E+01,-0.90792572E+00,-0.15945410E+01, 0.52398443E-01,
              0.60421151E+00, 0.41048077E+00,-0.23732608E-01, 0.20138571E+00,
             -0.27786821E+00, },
            {-0.57795197E-01, 0.20736563E+01,-0.39351213E+00, 0.12668329E+00,
              0.84951319E-01, 0.28178258E-01,-0.23711748E-01,-0.16345520E+00,
             -0.57944131E+00, },
            {-0.20411162E+00,-0.13376068E+00,-0.29449091E+01, 0.41258937E+00,
              0.56290030E+00, 0.31165561E+00, 0.20544067E+00,-0.21796127E+00,
              0.47752853E-01, },
            { 0.20179317E+01, 0.24129705E+01,-0.70617986E+00,-0.83160406E+00,
              0.22868297E+00,-0.11227272E+00,-0.42454861E-01,-0.91994748E-01,
              0.24011090E+00, },
            {-0.28024802E+01, 0.13774416E+01, 0.36972830E+01,-0.59003949E+00,
             -0.13606600E+01,-0.84793615E+00, 0.57913697E+00,-0.74747223E+00,
              0.55327827E+00, },
            {-0.22295168E+00,-0.52054830E+01, 0.60161120E+00,-0.30097663E+00,
             -0.41190431E+00,-0.10745584E+00,-0.66798493E-01, 0.33825681E+00,
              0.10680364E+01, },
            { 0.27225360E+00,-0.13481601E+00, 0.12781663E+01,-0.17782266E+00,
             -0.43654701E+00,-0.28031033E+00,-0.21654996E+00, 0.18418586E+00,
              0.50952464E-01, },
            {-0.13763983E+01,-0.62647796E+00, 0.42017722E+00, 0.53561598E+00,
             -0.22507687E+00, 0.83096340E-01,-0.32817584E-01, 0.12693739E+00,
             -0.21098842E+00, },
            { 0.21004002E+01,-0.19308670E+00,-0.28420036E+01, 0.45582688E+00,
              0.97629905E+00, 0.47537673E+00,-0.57095116E+00, 0.58855778E+00,
             -0.29641420E+00, },
            { 0.67884624E-01, 0.40048332E+01,-0.36370404E-01, 0.15666211E+00,
              0.35521010E+00, 0.11299833E+00, 0.86527780E-01,-0.22095665E+00,
             -0.55126411E+00, },
            {-0.88387989E-02,-0.10312352E-01,-0.61113387E-02,-0.20343257E-01,
             -0.15003505E+00,-0.10647348E-01,-0.78879838E-03, 0.66963251E-03,
             -0.76038600E-02, },
            { 0.41395663E-02,-0.34112083E-02, 0.10996495E-01, 0.17316753E+00,
             -0.10291023E-01,-0.73754769E-02,-0.14292508E-01,-0.21934167E-02,
              0.18509438E-02, },
            { 0.63759461E-01, 0.23891982E+00, 0.63553944E-01, 0.18635046E+00,
             -0.57529569E+00, 0.23344524E-01,-0.40877607E-01, 0.17844956E-01,
             -0.39308965E-01, },
            {-0.51171139E-01, 0.10197276E+00, 0.71252682E-02, 0.72727478E+00,
              0.15855432E+00, 0.49558837E-01,-0.45434017E-01,-0.18780010E-01,
              0.50435401E-02, },
            { 0.15530051E-01, 0.14170410E+00,-0.61096840E-01,-0.37090287E-01,
             -0.10134680E+00, 0.43753270E-01, 0.16147131E-01, 0.11646984E-01,
              0.15542092E-01, },
            {-0.36590850E+00,-0.32496876E-02,-0.23358011E+00,-0.45408837E-01,
             -0.12131418E+00, 0.92704944E-01, 0.10798852E+00,-0.10122197E-01,
              0.33092615E-03, },
            {-0.40528265E+00,-0.12369304E+01,-0.24165794E-01,-0.74351096E+00,
              0.17940987E+01,-0.18066379E+00, 0.10244836E+00,-0.25582047E-01,
              0.19211514E+00, },
            { 0.76301038E+00,-0.20792608E+00, 0.27201289E+00,-0.23394880E+01,
             -0.62358546E+00,-0.24235266E+00, 0.63761316E-01, 0.15015985E+00,
             -0.11706612E+00, },
            {-0.10061766E+00,-0.37011167E+00, 0.14104469E+00,-0.26392074E-01,
              0.25978979E+00, 0.53946935E-02, 0.15027906E-01,-0.45017198E-01,
              0.68766944E-03, },
            { 0.53705126E+00,-0.14921010E+00, 0.35961419E+00,-0.32242209E-01,
              0.42051267E-01,-0.12954605E+00,-0.89045294E-01, 0.19702064E-04,
             -0.23211284E-01, },
            { 0.49777216E+00, 0.15834332E+01, 0.39794613E-01, 0.63238031E+00,
             -0.15466079E+01, 0.28480631E+00,-0.73767789E-01, 0.22335300E-01,
             -0.21944922E+00, },
            {-0.12544603E+01, 0.71595207E-01,-0.24489100E+00, 0.20173306E+01,
              0.51294672E+00, 0.28983241E+00, 0.99735051E-01,-0.18368557E+00,
              0.18341650E+00, },
            {-0.42410571E-01,-0.52171294E-02, 0.19085895E-02,-0.95514078E-02,
              0.75655179E-02, 0.16171318E-01, 0.16939323E-01, 0.22008182E-02,
             -0.19602047E-02, },
            {-0.60261209E-02,-0.52024610E-01,-0.89434842E-02, 0.10327846E-01,
             -0.25033876E-01,-0.14511695E-01, 0.23334123E-01,-0.61932360E-02,
              0.28024365E-02, },
            { 0.82942434E-01, 0.45942049E-01, 0.93686730E-02,-0.41635599E-01,
              0.30467382E-01, 0.41355919E-01, 0.26757890E-01, 0.10544643E-01,
              0.57960446E-02, },
            { 0.58556028E-01,-0.45138933E-02,-0.20666830E-01,-0.21553095E-01,
              0.11791925E-01,-0.26258841E-01, 0.53928331E-01,-0.11998913E-01,
              0.21382065E-01, },
            { 0.77182382E-01,-0.48827209E-01,-0.91188706E-01, 0.14237206E-01,
             -0.10213549E-01,-0.94213583E-01,-0.15220276E+00,-0.12513259E-02,
              0.21464357E-01, },
            {-0.15465503E-01, 0.13954592E+00,-0.39753947E-01,-0.45017153E-01,
              0.57116088E-01, 0.15541755E+00,-0.12666646E+00, 0.11608529E-01,
              0.80760103E-02, },
            {-0.41384086E-01,-0.10227784E-01, 0.80588767E-02,-0.46442337E-02,
             -0.21833975E-02, 0.53496994E-02,-0.68548233E-02, 0.41308586E-01,
              0.59219892E-02, },
            {-0.27170403E-01, 0.62266630E-02,-0.46112244E-02,-0.12002573E-01,
              0.16872867E-02, 0.85898452E-02, 0.16054403E-01,-0.50134775E-02,
              0.41393556E-01, },
            { 0.69590531E-01,-0.37438667E-02, 0.20416295E-01,-0.48384883E-01,
             -0.18482743E-01,-0.26152568E-03, 0.12390666E-01, 0.16554868E-01,
              0.15767444E-01, },
            { 0.27102111E-01,-0.50822388E-01, 0.19041890E-01, 0.45237873E-01,
             -0.61440300E-02,-0.86206961E-02,-0.41923290E-02,-0.37406992E-01,
              0.99841179E-02, },
            {-0.49510780E-02,-0.68882159E-02,-0.39861626E-02,-0.19369667E-02,
             -0.11413929E-02, 0.49859667E-02,-0.92119956E-02, 0.47657973E-03,
              0.34508729E-03, },
            { 0.18523857E-01, 0.17887400E-01,-0.11042763E-01,-0.10345280E-01,
             -0.11657286E-01, 0.51686917E-02, 0.27109310E-02, 0.35306651E-02,
              0.11049252E-02, },
            { 0.36995022E-02, 0.33790923E-02, 0.36772604E-02,-0.24134079E-02,
             -0.38045826E-02, 0.39723795E-02, 0.10341560E-02, 0.54815668E-02,
             -0.16316446E-02, },
            { 0.22392916E-01, 0.57393694E-02,-0.18493544E-01, 0.23872721E-03,
              0.48304014E-02,-0.32244358E-03, 0.17225273E-02,-0.32498674E-02,
             -0.64918457E-03, },
         },
         {
            { 0.27327437E+01,-0.15785564E-01, 0.97018043E-02, 0.48172278E-02,
             -0.28546432E-02, 0.45274911E-02, 0.34621279E-02,-0.15102919E-02,
             -0.27329123E-02, },
            {-0.25903657E+00, 0.55780217E-01, 0.65728329E-01, 0.32537010E-01,
             -0.25246153E-01,-0.58772466E-02,-0.21703182E-01,-0.43969019E-02,
             -0.69272384E-03, },
            { 0.11714929E+01, 0.16338734E+00,-0.89373522E-01,-0.10331998E+00,
              0.13328508E-01,-0.56765247E-01,-0.25298420E-01, 0.44980552E-01,
              0.17588584E-01, },
            { 0.33108100E+00,-0.49494028E+00,-0.18589833E+00,-0.10610254E+00,
              0.78505076E-01,-0.88712927E-02, 0.33990052E-01, 0.10129429E-01,
              0.43824976E-02, },
            {-0.17302923E+01,-0.27473363E+00, 0.17417511E+00, 0.22498931E+00,
             -0.11839691E-01, 0.17832267E+00, 0.89640006E-01,-0.92783161E-01,
             -0.12903627E-01, },
            {-0.15330756E+00, 0.43055135E+00, 0.16011506E+00, 0.73783807E-01,
             -0.51124744E-01, 0.16581139E-01,-0.11291335E-01,-0.22684636E-02,
             -0.43424224E-02, },
            { 0.63714796E+00, 0.18350784E+00,-0.42438373E-01,-0.12726226E+00,
             -0.10468658E-01,-0.13165224E+00,-0.66993155E-01, 0.44999924E-01,
             -0.64978427E-02, },
            { 0.79358937E-02,-0.29707021E+00,-0.28992331E+00, 0.80166208E-02,
             -0.15802497E-01,-0.19063946E-02, 0.49845497E-02, 0.27219285E-03,
              0.67873574E-02, },
            { 0.62516500E-03, 0.32199055E+00,-0.27686164E+00,-0.15336918E-01,
              0.17286168E-01, 0.12508899E-02,-0.14910120E-01, 0.12231509E-02,
              0.14304347E-01, },
            { 0.16834082E-01, 0.10278944E+00,-0.63203834E-01, 0.10091765E-01,
             -0.13688602E-01,-0.95551871E-02,-0.36618944E-01,-0.14316996E-01,
              0.35773881E-01, },
            { 0.76245844E-01, 0.77808850E-01,-0.12024795E-01, 0.12149659E-01,
              0.17534763E-01, 0.12295829E-02,-0.60971705E-02, 0.14225282E-01,
              0.82393371E-01, },
            {-0.15534867E+00, 0.80440164E+00, 0.12503365E+01,-0.12217534E+00,
             -0.85220747E-01,-0.29276054E-01,-0.48868369E-01, 0.28717835E-01,
             -0.10331116E+00, },
            {-0.52347887E+00,-0.13314192E+01, 0.89782566E+00, 0.28488895E+00,
             -0.12012097E+00,-0.21137984E-01, 0.12435994E+00,-0.14031794E-01,
             -0.16417257E+00, },
            { 0.76694377E-01,-0.93182540E+00, 0.25457191E+00, 0.19702044E+00,
              0.48383981E-01, 0.41341368E-01,-0.14367604E-01, 0.36359438E+00,
             -0.25973442E+00, },
            { 0.16930665E+00, 0.43561876E+00, 0.67502417E-01,-0.75919375E-01,
             -0.20566436E-01, 0.45070466E-01, 0.74356943E-01,-0.47366299E-01,
             -0.56628042E+00, },
            { 0.58145106E+00,-0.34549749E+00,-0.13091027E+01, 0.17559958E+00,
              0.43319133E+00, 0.16244146E+00, 0.20602815E+00,-0.41776557E-01,
              0.30710989E+00, },
            { 0.17932121E+01, 0.97565883E+00,-0.11525728E+01,-0.74104398E+00,
              0.24997143E+00, 0.52191265E-01,-0.24265365E+00,-0.61648171E-01,
              0.39528716E+00, },
            {-0.39080113E+00, 0.15962688E+01,-0.17618662E+00,-0.74545747E+00,
             -0.20102003E+00,-0.15913907E-01, 0.36910924E+00,-0.10008817E+01,
              0.41477209E+00, },
            {-0.10000013E+01,-0.15626558E+01,-0.33813745E+00, 0.20472082E+00,
              0.47138073E-01,-0.14620386E+00,-0.33509147E+00, 0.74930072E-01,
              0.10780282E+01, },
            {-0.40786570E+00, 0.21695097E-02, 0.61571985E+00,-0.87222412E-01,
             -0.40252182E+00,-0.18569061E+00,-0.16891396E+00, 0.19968783E-01,
             -0.22394384E+00, },
            {-0.13754953E+01,-0.63793459E-02, 0.60612941E+00, 0.53394854E+00,
             -0.20959119E+00,-0.59920955E-01, 0.10459418E+00, 0.97062908E-01,
             -0.27976444E+00, },
            { 0.39142886E+00,-0.65925485E+00,-0.55752856E+00, 0.61146140E+00,
              0.21277002E+00,-0.66732608E-01,-0.38353363E+00, 0.67082864E+00,
             -0.16077600E+00, },
            { 0.87837434E+00, 0.18233119E+01, 0.43625033E+00,-0.18887997E+00,
             -0.19754047E-01, 0.14874180E+00, 0.31835011E+00,-0.25374444E-01,
             -0.59807515E+00, },
            {-0.35858257E-02,-0.18426422E-01, 0.11940397E-01, 0.35249278E-01,
             -0.11406703E+00,-0.12094727E-01, 0.12369556E-02, 0.28486021E-02,
             -0.41241115E-02, },
            {-0.44571543E-02,-0.12385954E-01, 0.10621492E-01, 0.11439821E+00,
              0.52063718E-01,-0.31545039E-02,-0.91192275E-02,-0.23166044E-02,
              0.18062951E-02, },
            { 0.77467859E-01, 0.19298270E+00, 0.57610362E-02, 0.17876288E+00,
             -0.28498143E+00, 0.33583455E-01,-0.21250669E-02, 0.40939944E-02,
             -0.17955484E-01, },
            { 0.18605869E-01, 0.88920861E-01,-0.16398907E-01, 0.37881327E+00,
              0.19283801E+00, 0.35964742E-01,-0.57503371E-03,-0.16725548E-01,
              0.14335946E-01, },
            {-0.39326932E-01, 0.15444148E+00,-0.13056353E+00,-0.27956468E+00,
              0.53259443E-01, 0.70716918E-01, 0.21401586E-01,-0.33642214E-01,
              0.21223655E-01, },
            {-0.20482560E+00, 0.36653712E-01,-0.11544807E+00,-0.96398272E-01,
             -0.40236861E+00, 0.14114498E-01, 0.74856617E-01, 0.13251967E-01,
              0.58944183E-02, },
            {-0.22919865E+00,-0.93992049E+00, 0.22374582E+00,-0.53515422E+00,
              0.77677137E+00,-0.16614506E+00,-0.86691096E-01, 0.29328503E-01,
              0.12252356E+00, },
            { 0.30133858E+00,-0.19458076E+00, 0.14997202E+00,-0.10709431E+01,
             -0.48711553E+00,-0.11093743E+00,-0.55874560E-01, 0.11211873E+00,
             -0.14200687E+00, },
            { 0.17139059E-01,-0.40209791E+00, 0.19722582E+00, 0.75891189E-01,
             -0.96643463E-01,-0.82315318E-01,-0.36470420E-02, 0.47883753E-01,
             -0.17165303E-01, },
            { 0.32197592E+00,-0.29618844E+00, 0.99401399E-01, 0.21706817E+00,
              0.22001296E+00,-0.53230442E-01,-0.68153217E-01,-0.20923691E-01,
             -0.37170000E-01, },
            { 0.10419927E+00, 0.12077615E+01,-0.25412044E+00, 0.44525284E+00,
             -0.71311611E+00, 0.24982084E+00, 0.11983401E+00,-0.38160987E-01,
             -0.16348417E+00, },
            {-0.60997337E+00, 0.27142987E+00,-0.35729647E-01, 0.90510237E+00,
              0.34794793E+00, 0.13886355E+00, 0.16090356E+00,-0.10832496E+00,
              0.19660074E+00, },
            {-0.28633850E-01,-0.51487325E-03, 0.12486504E-01,-0.59260265E-02,
             -0.30928615E-02, 0.16670747E-01,-0.18133426E-01, 0.13137640E-02,
             -0.29191643E-02, },
            {-0.73401621E-02,-0.43932945E-01,-0.60264287E-02, 0.11825816E-01,
             -0.20122048E-01, 0.13404754E-01, 0.25719734E-01,-0.27513448E-02,
             -0.39640418E-02, },
            { 0.54108258E-01, 0.62069230E-01, 0.32028572E-02,-0.30705713E-01,
              0.21409197E-01, 0.68470776E-01, 0.18883282E-01, 0.12196345E-02,
             -0.59491460E-03, },
            { 0.44463780E-01, 0.15322219E-02,-0.11973175E-01,-0.23743760E-01,
              0.23724630E-02,-0.18886378E-01, 0.65881431E-01,-0.61001866E-02,
              0.12625720E-01, },
            { 0.42619128E-01,-0.87218329E-01,-0.71985729E-01, 0.61123907E-02,
             -0.80230413E-02,-0.94753467E-01,-0.10718403E+00, 0.44616999E-03,
              0.18477114E-01, },
            {-0.49196170E-02, 0.10727092E+00,-0.36542352E-01,-0.19530224E-01,
              0.29819097E-01, 0.11121344E+00,-0.11566354E+00,-0.12570542E-02,
              0.71676592E-02, },
            {-0.28708642E-01,-0.12917655E-01, 0.10160701E-01,-0.15056655E-02,
             -0.51682289E-02, 0.13105695E-02,-0.69624763E-02, 0.31344198E-01,
              0.13837171E-01, },
            {-0.15254789E-01, 0.17759033E-02,-0.10812398E-01,-0.43772645E-02,
             -0.18710297E-02, 0.43232990E-02, 0.13123615E-01,-0.15094435E-01,
              0.28170321E-01, },
            { 0.47917835E-01, 0.23561502E-02, 0.78526502E-02,-0.31751979E-01,
             -0.63430383E-02,-0.38912459E-02, 0.14013788E-01, 0.21223193E-01,
              0.36825228E-01, },
            { 0.24586657E-01,-0.54429136E-01, 0.31562790E-01, 0.22680206E-01,
              0.70566137E-03,-0.58405814E-02,-0.98891770E-02,-0.44559006E-01,
              0.10059295E-01, },
            { 0.96068269E-03,-0.11706586E-01, 0.22053521E-02, 0.32256660E-02,
             -0.55763084E-02, 0.31189667E-02,-0.83871549E-02, 0.37916502E-03,
              0.65405335E-03, },
            { 0.91339229E-02, 0.15419393E-01,-0.98529020E-02,-0.21402466E-02,
             -0.34411531E-02, 0.44119060E-02, 0.64586932E-02, 0.16170664E-02,
              0.65145995E-02, },
            { 0.20680716E-02, 0.92757810E-02, 0.41878442E-02, 0.99740468E-03,
             -0.10964575E-01, 0.11091142E-03, 0.14305855E-02, 0.14857621E-02,
              0.76322776E-03, },
            { 0.71551851E-02,-0.16748263E-02,-0.11113628E-01, 0.20171816E-02,
              0.70238686E-02, 0.14452589E-02, 0.47303335E-03,-0.38445606E-02,
              0.29371867E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir19F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.60445695E+01,-0.13494897E+00, 0.96523471E-01, 0.30800914E-01,
             -0.25691688E-02,-0.24381364E-01,-0.20869752E-01, 0.36958508E-01,
             -0.24559852E-01, 0.99607911E-02, 0.18155642E-02, 0.16870741E-04,
             -0.14172103E-01, },
            {-0.72683138E+00,-0.18670639E+00, 0.51046336E+00,-0.36924446E+00,
             -0.19951835E+00,-0.26578641E+00, 0.36712724E+00,-0.13204169E+00,
             -0.27474661E-02, 0.22398602E-01,-0.15810436E+00, 0.48819307E-01,
              0.84509522E-01, },
            { 0.13992862E+02, 0.23779745E+01,-0.22226188E+01, 0.20291131E+01,
              0.90447301E+00, 0.73813629E+00, 0.10136298E+01,-0.49498451E+00,
              0.72660756E+00,-0.77001888E+00,-0.32229844E+00, 0.19383496E+00,
              0.34973943E+00, },
            { 0.33502454E+00,-0.41584191E+01,-0.30603714E+01, 0.73569479E+01,
              0.64584503E+01, 0.61389766E+01,-0.67864485E+01, 0.15375274E+01,
              0.11104467E+01, 0.45180345E+00, 0.31834459E+01,-0.39800099E+00,
             -0.16252377E+01, },
            {-0.98461967E+02,-0.11804735E+02, 0.13932345E+02,-0.21968882E+02,
             -0.18346836E+02,-0.64648104E+01,-0.44017797E+01, 0.21018789E+01,
             -0.50670667E+01, 0.55518346E+01, 0.14706390E+01,-0.12416881E+01,
             -0.32495730E+01, },
            { 0.24509441E+02, 0.18994970E+02,-0.48482704E+01,-0.31010715E+02,
             -0.24011599E+02,-0.33442169E+02, 0.34668297E+02,-0.85968723E+01,
             -0.92274466E+01,-0.22783225E+01,-0.16358894E+02, 0.99501222E+00,
              0.78107009E+01, },
            { 0.21546281E+03, 0.30437609E+02,-0.30536016E+02, 0.66628876E+02,
              0.65871796E+02, 0.18404390E+02, 0.62645345E+01,-0.42978358E+01,
              0.12689314E+02,-0.13794364E+02,-0.23129451E+01, 0.28023379E+01,
              0.10037551E+02, },
            {-0.63695538E+02,-0.22283279E+02, 0.36996841E+02, 0.50817825E+02,
              0.28390305E+02, 0.72428009E+02,-0.71910110E+02, 0.21617928E+02,
              0.24796774E+02, 0.37806880E+01, 0.33501648E+02,-0.13203067E+01,
             -0.16414623E+02, },
            {-0.21614815E+03,-0.35832664E+02, 0.27470688E+02,-0.79447624E+02,
             -0.83696152E+02,-0.21164886E+02,-0.25409181E+01, 0.43153014E+01,
             -0.12989493E+02, 0.14197570E+02, 0.14302781E+01,-0.27501328E+01,
             -0.12051160E+02, },
            { 0.50732521E+02, 0.19075854E+01,-0.50831139E+02,-0.35323662E+02,
             -0.81260357E+01,-0.68726357E+02, 0.65579124E+02,-0.23836714E+02,
             -0.26761435E+02,-0.26312616E+01,-0.30050358E+02, 0.11140531E+01,
              0.15877548E+02, },
            { 0.83067856E+02, 0.14992576E+02,-0.86386423E+01, 0.32848270E+02,
              0.35260712E+02, 0.85520496E+01,-0.31619912E+00,-0.16816278E+01,
              0.46327853E+01,-0.52116356E+01,-0.27778086E+00, 0.10079776E+01,
              0.49246526E+01, },
            {-0.10440203E+02, 0.58246608E+01, 0.21354177E+02, 0.85120945E+01,
             -0.24493570E+01, 0.23852114E+02,-0.21972420E+02, 0.94112186E+01,
              0.10090175E+02, 0.65944129E+00, 0.98899641E+01,-0.44724047E+00,
             -0.57489376E+01, },
            {-0.63381091E-01, 0.20475085E+01, 0.15625848E+01,-0.13729405E-01,
             -0.65938756E-01,-0.35532121E-01,-0.29464888E-02, 0.79189196E-01,
              0.18604737E-02, 0.96583366E-02, 0.30324928E-01, 0.14240874E-02,
             -0.19405013E-01, },
            {-0.23046065E-01,-0.14310176E+01, 0.19285048E+01,-0.16601916E-01,
              0.18041897E+00,-0.12310948E-01,-0.66375285E-01, 0.31838231E-01,
             -0.65732598E-02,-0.13759874E-01, 0.18407125E-01,-0.24185097E-01,
              0.10961994E-01, },
            {-0.74787430E-01, 0.16353496E+01,-0.22924951E+00, 0.47504836E+00,
             -0.96047330E+00,-0.33705956E+00, 0.11720680E+00,-0.35202015E-01,
              0.46376839E+00,-0.10265652E-01,-0.14193834E+00, 0.55663763E-02,
              0.19628496E+00, },
            {-0.12596104E+00, 0.14582266E+01,-0.98657086E-01, 0.94749078E-01,
              0.78419155E+00,-0.29752413E-01,-0.19540650E+00, 0.36470249E+00,
              0.27288207E+00, 0.13509390E+00,-0.56428380E-01,-0.11677991E+00,
             -0.53491343E-01, },
            { 0.19545885E+01, 0.11435436E+02, 0.84868221E+01, 0.32103235E+00,
              0.12959203E+01, 0.88761157E+00, 0.49820134E+00,-0.11861235E+01,
             -0.71958087E-01,-0.25178078E+00,-0.12668631E+01,-0.35598647E-01,
              0.32306072E+00, },
            {-0.32294741E+01,-0.95716171E+01, 0.12515850E+02, 0.23365383E+01,
              0.91575392E-01,-0.88255316E+00, 0.17265298E+01,-0.47084525E+00,
              0.12570558E+01, 0.15405376E+00, 0.85922658E+00, 0.83317697E+00,
             -0.39139026E+00, },
            { 0.11067981E+01,-0.20162714E+02, 0.14454757E+02,-0.86500759E+01,
              0.16342585E+02, 0.64312339E+01,-0.42658453E+01,-0.13463537E+00,
             -0.77128196E+01, 0.24014208E+01, 0.39491782E+01, 0.46762183E+00,
             -0.34915750E+01, },
            { 0.46240864E+01,-0.27222712E+02, 0.80155439E+01,-0.32276535E+01,
             -0.17326429E+02, 0.45502667E+01, 0.41100821E+01,-0.76522098E+01,
             -0.50376501E+01,-0.25007749E+01, 0.33137116E+00, 0.32376506E+01,
              0.32104585E+00, },
            { 0.24788020E+01,-0.65718430E+02, 0.30561613E+02, 0.72700047E+01,
             -0.19604420E+02,-0.72012863E+01,-0.44401312E+01, 0.93486271E+01,
              0.33760922E+01, 0.81467517E-01, 0.93344440E+01, 0.74155289E+00,
             -0.13300089E+01, },
            { 0.34563679E+02,-0.28905020E+02,-0.76245705E+02,-0.21742231E+02,
             -0.73547716E+01, 0.10644203E+02,-0.10774131E+02, 0.45778413E+01,
             -0.11885620E+02,-0.60007966E+00,-0.11258645E+02,-0.67847219E+01,
              0.38254924E+01, },
            {-0.19042896E+02, 0.11207795E+03,-0.28545294E+02, 0.46980576E+02,
             -0.61684097E+02,-0.24524796E+02, 0.27351971E+02, 0.67426047E+01,
              0.44498322E+02,-0.20013243E+02,-0.25409010E+02,-0.27430830E+01,
              0.19376114E+02, },
            {-0.52723091E+02, 0.99575500E+02,-0.13353540E+02, 0.95136633E+01,
              0.99035027E+02,-0.30965115E+02,-0.18851999E+02, 0.40043255E+02,
              0.25284658E+02, 0.14075030E+02, 0.11309584E+01,-0.20960064E+02,
             -0.15822057E+01, },
            {-0.48169975E+02, 0.94947174E+02,-0.23541681E+03,-0.43397125E+02,
              0.64458717E+02, 0.22686687E+02, 0.10289151E+02,-0.30886675E+02,
             -0.14552165E+02, 0.30687971E+01,-0.24410389E+02,-0.24599547E+01,
              0.13256648E+01, },
            {-0.10717301E+03, 0.22662590E+03, 0.13886285E+03, 0.76127411E+02,
              0.35750031E+02,-0.36212978E+02, 0.21812931E+02,-0.18747749E+02,
              0.38481537E+02, 0.29832237E+01, 0.37119461E+02, 0.18805193E+02,
             -0.13755708E+02, },
            { 0.97456581E+02,-0.27367038E+03,-0.62144730E+02,-0.11783699E+03,
              0.86211739E+02, 0.28845253E+02,-0.65351234E+02,-0.25876451E+02,
             -0.10882457E+03, 0.56715382E+02, 0.64410553E+02, 0.43075614E+01,
             -0.45408981E+02, },
            { 0.15016524E+03,-0.12772850E+03,-0.25741722E+02,-0.16592799E+01,
             -0.24547108E+03, 0.70262955E+02, 0.35663345E+02,-0.83696991E+02,
             -0.53332424E+02,-0.35324783E+02,-0.94713192E+01, 0.54111397E+02,
              0.49722629E+01, },
            { 0.10506150E+03,-0.35246891E+02, 0.37882608E+03, 0.76265465E+02,
             -0.77831039E+02,-0.30161860E+02,-0.64913325E+01, 0.43002579E+02,
              0.21760742E+02,-0.60918565E+01, 0.27057970E+02, 0.25830400E+01,
              0.15233262E+01, },
            { 0.13615527E+03,-0.35940277E+03,-0.10190280E+03,-0.10915962E+03,
             -0.63814045E+02, 0.46641560E+02,-0.12899118E+02, 0.29911598E+02,
             -0.51161182E+02,-0.58345952E+01,-0.46785748E+02,-0.20363272E+02,
              0.19093109E+02, },
            {-0.17207220E+03, 0.30198877E+03, 0.17715979E+03, 0.13847585E+03,
             -0.45248798E+02,-0.24912093E+01, 0.66965683E+02, 0.35798885E+02,
              0.11744838E+03,-0.66378052E+02,-0.71170624E+02,-0.11229161E+01,
              0.46958759E+02, },
            {-0.15008180E+03, 0.48165199E+02, 0.66421715E+02,-0.11201522E+02,
              0.28107013E+03,-0.61901672E+02,-0.33674416E+02, 0.75491982E+02,
              0.52029320E+02, 0.41378738E+02, 0.16554775E+02,-0.61346951E+02,
             -0.63890009E+01, },
            {-0.64716988E+02,-0.60834203E+01,-0.18212560E+03,-0.42207047E+02,
              0.31869562E+02, 0.14135931E+02,-0.47822112E+00,-0.21003960E+02,
             -0.10799576E+02, 0.31178300E+01,-0.10885387E+02,-0.69965732E+00,
             -0.20979679E+01, },
            {-0.62125351E+02, 0.17012572E+03, 0.26723660E+02, 0.53963669E+02,
              0.37830799E+02,-0.20305946E+02,-0.10621082E+01,-0.15930128E+02,
              0.23917713E+02, 0.34242952E+01, 0.20282827E+02, 0.73727045E+01,
             -0.88406191E+01, },
            { 0.97107460E+02,-0.12479255E+03,-0.10509427E+03,-0.61543755E+02,
              0.48847432E+01,-0.85133734E+01,-0.24883598E+02,-0.16887283E+02,
             -0.46089825E+02, 0.27499294E+02, 0.28551277E+02,-0.10849844E+01,
             -0.17723860E+02, },
            { 0.46780483E+02, 0.82450275E+01,-0.36148045E+02, 0.53822312E+01,
             -0.12206293E+03, 0.17168350E+02, 0.13732118E+02,-0.24196617E+02,
             -0.19565023E+02,-0.18214315E+02,-0.86782436E+01, 0.25485151E+02,
              0.25449452E+01, },
            { 0.95903993E-01, 0.12766166E+00,-0.18699701E+00,-0.10404546E+01,
             -0.33024654E-01, 0.64295433E-01,-0.23013584E-01,-0.10285793E-01,
              0.37326079E-01, 0.31895945E-02, 0.30733405E-01,-0.10633793E-01,
              0.97853728E-02, },
            { 0.52426510E-01,-0.93061021E-02,-0.71888246E-01, 0.19813684E+00,
             -0.72505492E+00,-0.19345094E-02,-0.19068209E-01,-0.36769819E-01,
              0.53982814E-02,-0.14521895E-01, 0.23320390E-01,-0.12640823E-02,
              0.60315416E-02, },
            { 0.46948263E+00, 0.12817408E+01, 0.52438563E+00, 0.58714247E+00,
             -0.23273043E+00, 0.46726218E+00,-0.28142142E+00, 0.50233081E-01,
             -0.11847415E+00,-0.78053989E-01,-0.36326803E-01,-0.14682151E+00,
              0.96851006E-01, },
            { 0.68315560E+00, 0.54635304E+00, 0.54994017E+00, 0.64071083E+00,
              0.38787290E+00, 0.12569439E+00, 0.33838964E+00,-0.25519842E+00,
             -0.18448704E+00,-0.18360077E+00,-0.28392771E+00, 0.12267494E+00,
              0.58491524E-01, },
            { 0.66295362E+00,-0.11465982E+01, 0.44839840E+01, 0.63690023E+01,
             -0.21015067E+01,-0.15754128E+01, 0.21561065E+00, 0.82991850E+00,
             -0.24949779E+00, 0.12380320E+00,-0.32161883E+00,-0.23786856E+00,
             -0.31155005E+00, },
            { 0.14305441E+01, 0.57962000E+00, 0.32569954E+01,-0.28792398E+01,
              0.34138973E+01,-0.71248835E+00,-0.62034968E-01, 0.14145756E+01,
              0.23729173E-01,-0.19871333E+00,-0.66435295E+00, 0.17309137E-01,
              0.12985046E-02, },
            {-0.47963367E+01,-0.89218206E+01,-0.36218772E+01, 0.90268250E+01,
             -0.87401462E+00,-0.29467752E+01, 0.10264589E+01,-0.11269215E+01,
              0.13553836E+01, 0.10891480E+01, 0.17733830E+01, 0.15696553E+01,
             -0.47471687E+00, },
            {-0.50660028E+01,-0.33092887E+01,-0.32716267E+01,-0.14229895E+01,
              0.11620011E+02, 0.96331590E+00,-0.26582198E+01, 0.10442986E+01,
              0.50223678E+00, 0.14703425E+01, 0.40092273E+01,-0.53163874E+00,
             -0.32783002E+00, },
            {-0.91051664E+01, 0.17604088E+02,-0.22866295E+02,-0.28144238E+01,
              0.20614243E+02, 0.94886522E+01,-0.77871466E+00,-0.35348945E+01,
              0.17768615E+01,-0.16262704E+01,-0.12231375E+00, 0.21009269E+01,
              0.29577000E+01, },
            {-0.74458656E+01, 0.18080215E+01,-0.13265327E+02, 0.81773024E+01,
              0.70696664E+01, 0.32722304E+01, 0.41111064E+01,-0.59407196E+01,
             -0.56068152E+00, 0.22052250E+01, 0.37547357E+01,-0.12558265E+01,
              0.11403291E+01, },
            { 0.10656018E+02, 0.17880117E+02, 0.63293304E+01,-0.30708920E+02,
              0.19644450E+01, 0.56080647E+01,-0.20938640E+01, 0.40487800E+01,
             -0.39456055E+01,-0.30156834E+01,-0.52191610E+01,-0.42830544E+01,
              0.71889979E+00, },
            { 0.10659558E+02, 0.14922249E+01, 0.54132767E+01, 0.28934119E+01,
             -0.37208477E+02,-0.38359447E+01, 0.57908921E+01,-0.21831961E+00,
              0.91132575E+00,-0.33404295E+01,-0.10999828E+02, 0.95123515E-01,
              0.62245464E+00, },
            { 0.21184179E+02,-0.49009296E+02, 0.41135643E+02,-0.23144958E+02,
             -0.44450764E+02,-0.21969942E+02, 0.31099708E+01, 0.57433114E+01,
             -0.42183104E+01, 0.41270490E+01, 0.23304346E+01,-0.47546310E+01,
             -0.74363651E+01, },
            { 0.14705822E+02,-0.11686625E+02, 0.15904526E+02,-0.11272635E+02,
             -0.37194714E+02,-0.40984364E+01,-0.12301613E+02, 0.77951894E+01,
              0.56480533E+00,-0.55646262E+01,-0.60771265E+01, 0.43992543E+01,
             -0.39650044E+01, },
            {-0.80576096E+01,-0.10058566E+02,-0.42752757E+01, 0.22247026E+02,
             -0.14402002E+01,-0.33199692E+01, 0.17421428E+01,-0.34508293E+01,
              0.32978044E+01, 0.24247870E+01, 0.38595335E+01, 0.32630551E+01,
             -0.44972682E+00, },
            {-0.69929409E+01, 0.46824846E+01,-0.20134742E+01,-0.25459471E+01,
              0.26119644E+02, 0.29910052E+01,-0.36391468E+01,-0.93128592E+00,
             -0.17195288E+01, 0.21541712E+01, 0.83281469E+01, 0.72017050E+00,
             -0.27138808E+00, },
            {-0.15138537E+02, 0.36185287E+02,-0.22484388E+02, 0.22932758E+02,
              0.28119076E+02, 0.15488616E+02,-0.30330801E+01,-0.34491429E+01,
              0.28721263E+01,-0.30168860E+01,-0.23003569E+01, 0.31897192E+01,
              0.54570012E+01, },
            {-0.10177624E+02, 0.91815348E+01,-0.54682875E+01, 0.65655260E+01,
              0.30150761E+02, 0.12102582E+01, 0.89690657E+01,-0.31353066E+01,
              0.28226671E+00, 0.41323018E+01, 0.27218044E+01,-0.37113197E+01,
              0.31892061E+01, },
            {-0.56241829E-01, 0.73317058E-01,-0.23292311E+00,-0.61917238E-01,
              0.11740326E+00,-0.28218538E+00,-0.69195360E+00, 0.36205132E-01,
              0.38509484E-01, 0.25892833E-01, 0.39562065E-01,-0.88990778E-02,
             -0.71549118E-02, },
            { 0.10465758E+00,-0.25458544E+00,-0.65254413E-01, 0.92262980E-02,
             -0.97085372E-01, 0.68069780E+00,-0.31799850E+00,-0.16797258E-01,
              0.35726070E-01, 0.86577646E-02,-0.35667919E-01, 0.55345451E-02,
             -0.37628175E-02, },
            {-0.15063886E+00, 0.63173950E+00, 0.58640532E-01,-0.11310439E+00,
             -0.56205934E+00,-0.12007036E+00,-0.60520858E+00,-0.13778128E+00,
              0.53372163E-01, 0.55302665E-01, 0.43478370E-01, 0.42523559E-01,
             -0.29654521E-01, },
            {-0.46249099E-01,-0.18232642E+00, 0.18445110E+00, 0.11651336E+00,
             -0.13888021E+00, 0.77584624E+00,-0.20802420E+00,-0.27784950E+00,
             -0.12532072E+00,-0.95613144E-01,-0.53957444E-01, 0.21425946E-01,
             -0.73493482E-02, },
            { 0.32895875E+00, 0.57959902E+00, 0.17125003E+01, 0.64916812E-01,
             -0.66671216E+00, 0.42497212E+00, 0.10747848E+01,-0.29168382E+00,
             -0.33504639E-02,-0.39606754E-01,-0.48538277E+00,-0.46966869E-01,
              0.79471610E-01, },
            {-0.63520753E+00, 0.35119760E+01,-0.12184794E+01, 0.14708529E+00,
              0.32728174E+00,-0.13452771E+01, 0.91351074E+00,-0.10723860E+00,
             -0.46364650E+00,-0.17759089E+00, 0.33546487E-03,-0.68300068E-01,
              0.33596814E-01, },
            {-0.32629791E+00,-0.10383365E+01, 0.18383627E+00,-0.12193326E+00,
              0.95989013E+00, 0.17567472E+00, 0.10174921E+01, 0.54608107E+00,
             -0.69135964E-01,-0.16146509E+00,-0.69958568E-01,-0.12977321E+00,
              0.74624196E-01, },
            {-0.34265670E+00, 0.70992500E+00,-0.20130862E+00,-0.87979764E+00,
              0.27189708E+00,-0.16422273E+01, 0.42830017E+00, 0.45904860E+00,
              0.39268953E+00, 0.97692028E-01, 0.53528871E-01,-0.25272183E-02,
             -0.82533369E-02, },
            { 0.30163246E+00,-0.75499952E+00,-0.23318834E+01, 0.28877845E+00,
              0.12862626E+01,-0.79786736E+00,-0.10814675E+01, 0.10795548E+00,
             -0.14806841E+00, 0.30484816E-01, 0.60364807E+00, 0.24701132E+00,
             -0.16554859E+00, },
            { 0.75564712E+00,-0.49911842E+01, 0.25221035E+01,-0.50133252E+00,
             -0.11462419E+00, 0.19772404E+01,-0.17406584E+01, 0.17858028E+00,
              0.75754690E+00, 0.32222256E+00, 0.12655260E+00, 0.14917502E+00,
             -0.11977191E+00, },
            { 0.76102793E-01,-0.43147720E-01,-0.81818521E-01, 0.19644612E-01,
              0.33825956E-01, 0.25999684E-01, 0.32359265E-01, 0.29988968E+00,
             -0.63901916E-01, 0.23167776E-02, 0.22926815E-01,-0.77330978E-02,
              0.10746138E-02, },
            { 0.12264740E+00, 0.20447599E-01, 0.59562452E-01, 0.19856852E-01,
             -0.11300006E+00,-0.29754141E-01,-0.44352575E-02, 0.56839380E-01,
              0.25084686E+00,-0.56135736E-03,-0.54703793E-02, 0.48170928E-02,
             -0.36291028E-02, },
            { 0.24246810E-01, 0.65477557E-01, 0.12209361E+00,-0.99905565E-01,
             -0.29400787E+00,-0.25406672E-01, 0.31711336E-01, 0.17040537E+00,
              0.24152596E+00,-0.77344932E-01, 0.33627842E-01,-0.97614238E-02,
              0.20799657E-01, },
            { 0.12910360E+00, 0.27060756E-01,-0.15518239E+00, 0.15885320E+00,
              0.30369675E+00,-0.49701214E-01,-0.10029802E+00,-0.28022015E+00,
              0.36929357E-02, 0.24210740E-01,-0.16487345E-01, 0.94972895E-02,
             -0.56310687E-02, },
            {-0.11048770E+00,-0.25132010E-01,-0.74848481E-01,-0.36955964E-01,
             -0.84306849E-02, 0.59861795E-03, 0.43695938E-01, 0.46816003E-02,
             -0.12226735E-01, 0.69531336E-01, 0.23915982E+00,-0.52309283E-02,
             -0.14996283E-01, },
            { 0.29184736E-01, 0.55157773E-01, 0.20157386E-01, 0.30791920E-01,
             -0.58250558E-01,-0.56456961E-02,-0.21537563E-01, 0.28616695E-01,
             -0.26698606E-02,-0.25978240E+00, 0.62796988E-01, 0.22548081E-02,
             -0.23580464E-02, },
            { 0.47859617E-01,-0.17690947E-01,-0.53106416E-02,-0.21001626E-01,
             -0.39911261E-02, 0.21716015E-01,-0.17059458E-01, 0.28004622E-01,
              0.10446360E-01,-0.44645024E-02,-0.42491471E-02,-0.13451762E+00,
             -0.59785281E-03, },
            {-0.11083220E-01, 0.54354630E-01,-0.40297960E-02, 0.53323619E-01,
              0.48851684E-01, 0.93341740E-02,-0.30359656E-05,-0.13635743E-01,
             -0.13263462E-01,-0.91869375E-02,-0.23183506E-01, 0.10026116E-02,
             -0.14598797E+00, },
            {-0.39637778E-01, 0.34675904E-01,-0.13779508E-01, 0.31649983E-02,
              0.24430851E-01,-0.10841477E-01,-0.87321568E-02, 0.24531895E-01,
             -0.39002334E-02,-0.10708721E-01,-0.32987449E-01, 0.29985204E-02,
              0.48053171E-02, },
            { 0.86180255E-01, 0.39627044E-02, 0.22435104E-01, 0.21348750E-01,
             -0.34228362E-01,-0.53658383E-03, 0.33991758E-01,-0.37611157E-01,
              0.23497487E-02,-0.50931578E-02, 0.65411784E-03, 0.11545370E-01,
             -0.11957943E-02, },
            {-0.27704688E-01, 0.11326194E-01,-0.12809221E+00, 0.25609735E-01,
             -0.43673404E-01,-0.19919191E-01, 0.57441801E-01, 0.13651283E-02,
              0.99236434E-02, 0.11110033E-02,-0.88214837E-02, 0.10099307E-01,
              0.23763576E-03, },
            {-0.32913532E-01, 0.35655253E-01,-0.93667448E-01, 0.22362642E-01,
              0.35769939E-01, 0.23009181E-01, 0.10751249E-01, 0.21540695E-02,
             -0.16343425E-02,-0.27889894E-01,-0.19826155E-01,-0.16842147E-02,
              0.94351284E-02, },
         },
         {
            { 0.89623709E+01, 0.19730670E-01, 0.21596248E+00, 0.50009090E-01,
             -0.21762836E-02,-0.36576837E-01,-0.37319519E-01, 0.31727146E-01,
             -0.24356844E-01, 0.29576533E-01, 0.85489517E-02,-0.13777472E-01,
             -0.97738234E-02, },
            { 0.33469999E+00,-0.24697515E+00, 0.48163033E+00,-0.35931331E+00,
             -0.15751718E-01,-0.13071296E+00, 0.29412560E-01,-0.13171205E+00,
             -0.74888915E-01,-0.31300247E-01,-0.69453977E-01, 0.31701785E-01,
              0.48052177E-01, },
            { 0.24180008E+02,-0.59165543E+00,-0.37481437E+01, 0.27094533E+01,
              0.17237262E+01, 0.33680838E+00, 0.12361391E+01,-0.40080991E+00,
              0.24777052E+00,-0.11386623E+01,-0.26397088E+00, 0.51696777E+00,
              0.26516542E+00, },
            {-0.45573082E+01, 0.22108464E+01,-0.33646340E+01, 0.54219594E+01,
              0.28218365E+01, 0.23740580E+01,-0.16911587E+01, 0.25113804E+01,
              0.21021688E+01, 0.10237397E+00, 0.14552329E+01,-0.97442865E-01,
             -0.11147535E+01, },
            {-0.11681603E+03, 0.18470577E+01, 0.18251852E+02,-0.25811161E+02,
             -0.20601797E+02,-0.18826652E+01,-0.56787043E+01, 0.29563882E+01,
             -0.94193526E-01, 0.73134317E+01, 0.93215400E+00,-0.28211844E+01,
             -0.22883477E+01, },
            { 0.19154861E+02,-0.18602942E+02, 0.20144350E+02,-0.16414492E+02,
             -0.60010877E+01,-0.12694320E+02, 0.97623625E+01,-0.14265865E+02,
             -0.14238819E+02, 0.14062835E+00,-0.69437256E+01,-0.15784086E+00,
              0.63612428E+01, },
            { 0.18642741E+03, 0.41710634E+01,-0.36739273E+02, 0.75611092E+02,
              0.65239151E+02, 0.62713485E+01, 0.90178194E+01,-0.86719627E+01,
             -0.24287522E+01,-0.17567657E+02,-0.38892239E+00, 0.58335896E+01,
              0.67666426E+01, },
            {-0.25843210E+02, 0.54234547E+02,-0.64408768E+02, 0.10873847E+02,
             -0.72628431E+01, 0.26647579E+02,-0.19763281E+02, 0.33570370E+02,
              0.35707272E+02,-0.12532645E-01, 0.12849774E+02, 0.64608556E+00,
             -0.15312591E+02, },
            {-0.14083356E+03,-0.12821358E+02, 0.34492329E+02,-0.88502136E+02,
             -0.78154541E+02,-0.92759399E+01,-0.48082018E+01, 0.10566218E+02,
              0.48087645E+01, 0.18066017E+02,-0.13238797E+01,-0.51708665E+01,
             -0.80225296E+01, },
            { 0.59901586E+01,-0.64209244E+02, 0.83126961E+02, 0.10725402E+02,
              0.23379948E+02,-0.23843412E+02, 0.16165714E+02,-0.34889957E+02,
             -0.37377113E+02,-0.10515053E+01,-0.10207627E+02,-0.58196390E+00,
              0.16365570E+02, },
            { 0.43592426E+02, 0.73433652E+01,-0.12409451E+02, 0.35991268E+02,
              0.31727673E+02, 0.46962137E+01, 0.23304205E+00,-0.45286808E+01,
             -0.25499396E+01,-0.67354417E+01, 0.10447700E+01, 0.16423652E+01,
              0.32990837E+01, },
            { 0.50915036E+01, 0.26708822E+02,-0.35995022E+02,-0.10378606E+02,
             -0.12888997E+02, 0.75558434E+01,-0.44697514E+01, 0.13253450E+02,
              0.13915670E+02, 0.88236904E+00, 0.29039440E+01, 0.16834038E+00,
             -0.63849540E+01, },
            {-0.26160195E+00, 0.16475309E+01, 0.18607447E+01, 0.27880883E-01,
             -0.80495954E-01,-0.39654482E-01, 0.19887313E-01, 0.60452126E-01,
             -0.19153394E-01, 0.13035006E-01, 0.31249687E-01,-0.75434586E-02,
             -0.15125268E-01, },
            { 0.27377906E+00,-0.17763206E+01, 0.15071578E+01,-0.19758046E-01,
              0.10182646E+00,-0.26982676E-01,-0.23550984E-01, 0.36379886E-02,
             -0.36529673E-02, 0.11719952E-01, 0.14171905E-01,-0.14677171E-01,
             -0.14988882E-02, },
            {-0.11673796E+01, 0.21813788E+01,-0.90088534E+00, 0.22404641E+00,
             -0.33333287E+00,-0.52520949E-01, 0.29254219E+00, 0.10636131E+00,
              0.68046279E-01, 0.12188297E-02,-0.82268119E-01,-0.86118124E-01,
              0.65291584E-01, },
            { 0.17419060E+01, 0.69445306E+00, 0.54693860E+00,-0.17051892E-01,
              0.58553481E+00,-0.22795539E-01, 0.11075487E+00, 0.25393158E+00,
              0.15103640E+00, 0.17391805E+00,-0.11656546E+00,-0.77986225E-01,
             -0.58590692E-01, },
            { 0.73945637E+01, 0.12272065E+02,-0.10930670E+02, 0.13587241E+01,
              0.17232165E+01, 0.35481679E+00, 0.61459994E+00,-0.78423440E+00,
              0.48749742E+00,-0.26622641E+00,-0.94569325E+00, 0.14302284E+00,
              0.17925684E+00, },
            {-0.94064894E+01, 0.11073936E+02, 0.13409169E+02, 0.13028538E+01,
              0.65245140E+00, 0.19972022E+00, 0.22320145E+00, 0.11397737E+01,
              0.71284533E+00,-0.11684898E+01, 0.29646003E+00, 0.81502265E+00,
             -0.25439862E-01, },
            { 0.23263647E+02,-0.16287771E+02, 0.19715300E+02,-0.70920930E+01,
              0.32130179E+01, 0.31002333E+01,-0.58355947E+01,-0.65897214E+00,
              0.11365482E+00,-0.26643327E+00, 0.25300441E+01, 0.20917065E+01,
             -0.21054513E+01, },
            {-0.25114740E+02,-0.66058850E+01, 0.86541386E+01,-0.41537158E-01,
             -0.17567032E+02, 0.47047324E+01,-0.51503760E+00,-0.63852286E+01,
             -0.29741337E+01,-0.44764342E+01, 0.62373561E+00, 0.24865687E+01,
              0.47261235E+00, },
            {-0.46814045E+02,-0.12041885E+02, 0.99752289E+02,-0.64495907E+01,
             -0.92173414E+01,-0.46798882E+01,-0.54954991E+01, 0.61131954E+01,
             -0.16522959E+01,-0.54570884E+00, 0.57664261E+01,-0.13666019E+01,
             -0.71777904E+00, },
            { 0.84022469E+02,-0.11194759E+03,-0.39507432E+01,-0.16203300E+02,
             -0.66421180E+01,-0.14821443E+00,-0.78339767E+00,-0.83775568E+01,
             -0.87370195E+01, 0.92152634E+01,-0.33313084E+01,-0.74441700E+01,
              0.18862633E+00, },
            {-0.14273080E+03, 0.62041721E+02,-0.79482536E+02, 0.51631123E+02,
             -0.57821947E+00,-0.19178213E+02, 0.29178068E+02, 0.33035190E+01,
             -0.28263030E+01, 0.34217317E-01,-0.18571112E+02,-0.10850002E+02,
              0.14315627E+02, },
            { 0.12893007E+03,-0.33273048E+01,-0.50037964E+02,-0.18768166E+01,
              0.10671618E+03,-0.27278667E+02, 0.50431371E+01, 0.36296223E+02,
              0.19505959E+02, 0.25242184E+02, 0.24296870E+01,-0.16299253E+02,
             -0.30654261E+01, },
            { 0.11099849E+03,-0.12857597E+03,-0.27917538E+03, 0.11415404E+02,
              0.13686012E+02, 0.19129147E+02, 0.14277333E+02,-0.19621246E+02,
              0.15207518E+01, 0.56465182E+01,-0.13264232E+02, 0.55628519E+01,
              0.15896335E+01, },
            {-0.26209372E+03, 0.31271347E+03,-0.18712296E+03, 0.61504059E+02,
              0.26230099E+02,-0.22624297E+01, 0.91929603E+00, 0.19573435E+02,
              0.32240540E+02,-0.23829140E+02, 0.96026239E+01, 0.23164410E+02,
             -0.11098813E+01, },
            { 0.37926721E+03,-0.13912029E+03, 0.11385747E+03,-0.15078870E+03,
             -0.28678265E+02, 0.41981728E+02,-0.61708336E+02,-0.99502068E+01,
              0.10072963E+02, 0.43251381E+01, 0.50344307E+02, 0.20761368E+02,
             -0.38384724E+02, },
            {-0.31183640E+03, 0.80717461E+02, 0.11177303E+03, 0.91798124E+01,
             -0.26519113E+03, 0.57131924E+02,-0.17250916E+02,-0.80624062E+02,
             -0.53986412E+02,-0.57842266E+02,-0.15444742E+02, 0.40284004E+02,
              0.98377438E+01, },
            {-0.10636666E+03, 0.27624472E+03, 0.33518640E+03,-0.90313139E+01,
             -0.29404740E+01,-0.29340532E+02,-0.13130365E+02, 0.27019056E+02,
              0.83808237E+00,-0.92540722E+01, 0.13138417E+02,-0.91429548E+01,
             -0.16192185E+01, },
            { 0.33913440E+03,-0.35920078E+03, 0.38479999E+03,-0.92527618E+02,
             -0.46624046E+02, 0.41366453E+01, 0.18275695E+01,-0.17915739E+02,
             -0.46472927E+02, 0.25177954E+02,-0.10816146E+02,-0.29169777E+02,
              0.18046658E+01, },
            {-0.44847644E+03, 0.16557732E+03,-0.59701851E+02, 0.19198531E+03,
              0.51862465E+02,-0.38352905E+02, 0.59586216E+02, 0.13888144E+02,
             -0.13456491E+02,-0.92949009E+01,-0.56630112E+02,-0.15786907E+02,
              0.44390171E+02, },
            { 0.35611423E+03,-0.13994145E+03,-0.12520427E+03,-0.78124218E+01,
              0.29974640E+03,-0.48495983E+02, 0.19427450E+02, 0.77247574E+02,
              0.66220634E+02, 0.59678249E+02, 0.23805574E+02,-0.41955189E+02,
             -0.13117631E+02, },
            { 0.34567719E+02,-0.14869554E+03,-0.14342838E+03, 0.30032880E+01,
             -0.36467550E+01, 0.15035783E+02, 0.32215376E+01,-0.13186405E+02,
             -0.12452888E+01, 0.43014369E+01,-0.46899452E+01, 0.50963411E+01,
              0.54561722E+00, },
            {-0.15541122E+03, 0.14388922E+03,-0.21015965E+03, 0.47509956E+02,
              0.29349554E+02,-0.19194942E+01,-0.30574071E+01, 0.53069601E+01,
              0.22960079E+02,-0.95035095E+01, 0.42254791E+01, 0.12840061E+02,
             -0.78551632E+00, },
            { 0.19336099E+03,-0.79162109E+02, 0.28272572E+01,-0.89598595E+02,
             -0.26287643E+02, 0.12450287E+02,-0.21709846E+02,-0.69439921E+01,
              0.60890908E+01, 0.54649796E+01, 0.22320847E+02, 0.36850126E+01,
             -0.18504169E+02, },
            {-0.15068378E+03, 0.72892563E+02, 0.55283623E+02,-0.76779521E+00,
             -0.12916338E+03, 0.13477338E+02,-0.61723790E+01,-0.26636496E+02,
             -0.29567837E+02,-0.23007311E+02,-0.11395707E+02, 0.15494074E+02,
              0.59043760E+01, },
            { 0.97768597E-01,-0.21706764E+00,-0.14598057E+00,-0.98067135E+00,
              0.56929809E+00, 0.43482479E-01, 0.11748247E-01,-0.63081672E-02,
              0.20523420E-01,-0.19572712E-01, 0.17729735E-03, 0.11943498E-02,
              0.66722082E-02, },
            {-0.22512754E-01,-0.17544612E+00,-0.19724552E+00,-0.56630099E+00,
             -0.77338606E+00,-0.35187386E-01, 0.10012046E-01,-0.27610054E-01,
              0.89784892E-03,-0.95248339E-03, 0.29487306E-01, 0.15489024E-01,
              0.32603152E-02, },
            { 0.90035915E+00, 0.50192165E+00, 0.39231557E+00, 0.72755909E+00,
              0.34172869E+00, 0.64472872E+00,-0.24112161E-01,-0.37124623E-01,
             -0.18246342E+00,-0.95317356E-01,-0.20045063E-01,-0.81074119E-01,
              0.11073198E-01, },
            { 0.55172008E+00, 0.20146661E+00, 0.44691011E+00, 0.65697357E-01,
              0.77415019E+00, 0.60446817E-01, 0.51282352E+00,-0.15366629E+00,
             -0.22197446E+00,-0.59145335E-01,-0.21178664E+00, 0.11207618E+00,
             -0.13108963E-01, },
            { 0.10676382E+01, 0.29645822E+01, 0.37952478E+01,-0.78678852E+00,
              0.32439947E+01,-0.16431471E+01,-0.11421096E+01, 0.17130587E+00,
              0.37664506E+00, 0.55448914E+00, 0.10976688E+00,-0.48771808E+00,
             -0.32246506E+00, },
            { 0.27298226E+01, 0.24008121E+01, 0.60759602E+01,-0.47121968E+01,
             -0.16266237E+01,-0.13724278E+00,-0.14916685E+01, 0.16014088E+01,
              0.21585748E+00,-0.35771871E+00,-0.89440823E+00,-0.45400912E+00,
             -0.29001749E+00, },
            {-0.74832025E+01,-0.29968793E+01,-0.42970152E+01, 0.20118179E+01,
             -0.21555855E+01,-0.44935822E+01,-0.38260967E+00,-0.88122147E+00,
              0.21959436E+01, 0.85298419E+00, 0.65500510E+00, 0.64462608E+00,
              0.34080345E-01, },
            {-0.58604355E+01,-0.17552921E+01,-0.46983767E+01,-0.19348935E+01,
              0.24940045E+01, 0.21718979E+01,-0.40221734E+01, 0.73567528E+00,
              0.19161574E+01,-0.13397838E+00, 0.25529764E+01,-0.50712591E+00,
              0.23999058E+00, },
            {-0.10549742E+02,-0.37017422E+01,-0.24552065E+02, 0.14676774E+02,
             -0.11914489E+02, 0.10997478E+02, 0.74938984E+01,-0.25663313E+00,
             -0.35813992E+01,-0.28400996E+01,-0.93953723E+00, 0.33387854E+01,
              0.20984364E+01, },
            {-0.14897774E+02,-0.62908635E+01,-0.32800335E+02, 0.18528563E+02,
              0.12417139E+02,-0.69420964E+00, 0.11475649E+02,-0.82309322E+01,
             -0.28264165E+01, 0.27254322E+01, 0.45255299E+01, 0.22271841E+01,
              0.21908436E+01, },
            { 0.17852583E+02, 0.30649593E+01, 0.11477947E+02,-0.67365303E+01,
              0.28456860E+01, 0.10238112E+02, 0.33093038E+00, 0.40141964E+01,
             -0.65592384E+01,-0.24417315E+01,-0.14772207E+01,-0.15737491E+01,
             -0.74067026E-01, },
            { 0.15307250E+02,-0.26185241E+01, 0.12405679E+02, 0.62960510E+01,
             -0.67733917E+01,-0.70792394E+01, 0.91405792E+01, 0.88415743E-03,
             -0.37179437E+01, 0.90935570E+00,-0.69149547E+01, 0.18560132E+00,
             -0.67503846E+00, },
            { 0.21599215E+02,-0.46994729E+01, 0.46027321E+02,-0.36294868E+02,
              0.96739960E+01,-0.24945257E+02,-0.13759644E+02,-0.24616983E-01,
              0.89603739E+01, 0.54308796E+01, 0.24014454E+01,-0.70196056E+01,
             -0.44327598E+01, },
            { 0.24843397E+02, 0.92978916E+01, 0.56579742E+02,-0.19334846E+02,
             -0.24256990E+02, 0.30407271E+01,-0.25068012E+02, 0.13572435E+02,
              0.68599362E+01,-0.67986002E+01,-0.75372796E+01,-0.38506787E+01,
             -0.46823874E+01, },
            {-0.14843636E+02, 0.19837296E+01,-0.11109138E+02, 0.31957550E+01,
             -0.17539186E+01,-0.67461467E+01, 0.35601419E+00,-0.37653399E+01,
              0.55061636E+01, 0.21600037E+01, 0.83115494E+00, 0.11146947E+01,
             -0.17095937E+00, },
            {-0.12286270E+02, 0.98170996E+01,-0.81901531E+01,-0.38168242E+01,
              0.12576302E+01, 0.55010004E+01,-0.60531683E+01,-0.10407491E+01,
              0.18965904E+01,-0.94105482E+00, 0.53762965E+01, 0.51917458E+00,
              0.65497190E+00, },
            {-0.13637304E+02, 0.65925388E+01,-0.23569542E+02, 0.25342230E+02,
              0.71028113E+00, 0.16614403E+02, 0.75715604E+01, 0.19589573E-01,
             -0.64481812E+01,-0.36762245E+01,-0.19363295E+01, 0.46145444E+01,
              0.30328979E+01, },
            {-0.13089158E+02,-0.10095703E+02,-0.30684874E+02, 0.39198732E+01,
              0.14732669E+02,-0.24675801E+01, 0.16160910E+02,-0.72000704E+01,
             -0.45575032E+01, 0.50762730E+01, 0.37089520E+01, 0.21473510E+01,
              0.29847896E+01, },
            {-0.14737675E-01,-0.25679311E-01,-0.16136080E+00,-0.33218786E-01,
              0.27244406E-01,-0.47106108E+00,-0.91915029E+00, 0.38202498E-01,
              0.29088611E-01, 0.31560920E-02, 0.18914854E-01, 0.52554188E-02,
             -0.11544408E-02, },
            { 0.72901808E-01,-0.22339118E+00, 0.53658471E-01, 0.85221520E-02,
             -0.27125636E-01, 0.88938648E+00,-0.52879435E+00, 0.87448508E-02,
              0.30740051E-01, 0.10852590E-01,-0.27724359E-01, 0.31586990E-02,
             -0.96590295E-02, },
            { 0.20115213E+00, 0.69436580E+00,-0.17900576E+00, 0.66310994E-01,
             -0.75164104E+00, 0.13775501E+00,-0.21529463E+00,-0.14966124E+00,
              0.16769144E+00, 0.77871263E-01, 0.51278979E-01, 0.15873125E-01,
              0.21192819E-01, },
            { 0.77322774E-01,-0.42061176E-01, 0.45077723E+00, 0.66685408E-01,
             -0.15208274E+00, 0.44830185E+00,-0.12592126E+00,-0.13366275E+00,
             -0.90220653E-01,-0.61910573E-01,-0.25508961E-01, 0.12957442E-01,
             -0.12587028E-01, },
            { 0.74074906E+00, 0.65904367E+00, 0.15433567E+01,-0.63742287E-02,
             -0.98790757E-01, 0.18118404E+01, 0.15583321E+01,-0.43290365E+00,
             -0.20866205E+00, 0.95957085E-01,-0.65422811E-01,-0.14005326E+00,
             -0.20921381E-01, },
            {-0.11043867E+01, 0.32856829E+01,-0.19886183E+01, 0.40116802E+00,
              0.26653954E+00,-0.17676287E+01, 0.24231007E+01,-0.21656232E+00,
             -0.57338196E+00,-0.14849383E+00, 0.14925669E+00,-0.52936438E-01,
              0.73655564E-02, },
            {-0.71928334E+00,-0.94114923E+00, 0.78959525E+00,-0.13111615E+00,
              0.16068588E+01,-0.33002041E-01, 0.42055130E+00, 0.56223714E+00,
             -0.28989446E+00,-0.17253461E+00, 0.78965602E-02,-0.94987266E-01,
             -0.70541412E-01, },
            {-0.24910378E+00, 0.25402689E+00,-0.95344973E+00,-0.80586588E+00,
              0.43270764E+00,-0.10406414E+01, 0.31384522E+00, 0.15633456E+00,
              0.34068555E+00, 0.12060667E+00, 0.44971276E-01, 0.29588670E-01,
              0.46364356E-01, },
            {-0.51294982E+00,-0.29964095E+00,-0.28561344E+01,-0.11806239E+00,
              0.18849920E+00,-0.34402866E+01,-0.21050479E+01, 0.34388900E+00,
              0.30473164E+00,-0.14386865E+00,-0.43631759E-01, 0.33641663E+00,
             -0.22661610E-01, },
            { 0.15949913E+01,-0.41355290E+01, 0.32266247E+01,-0.85630828E+00,
             -0.50795478E+00, 0.28716083E+01,-0.42127266E+01, 0.28837740E+00,
              0.93107861E+00, 0.27240244E+00,-0.13775127E+00, 0.13223094E+00,
             -0.13103773E+00, },
            {-0.49044993E-01,-0.24295989E-02,-0.71430229E-01, 0.35413116E-01,
              0.47498848E-01,-0.13454735E-01, 0.98276092E-02, 0.32753569E+00,
             -0.18088369E+00, 0.19360161E-02, 0.13374137E-01,-0.94158677E-02,
             -0.30937691E-02, },
            { 0.55371165E-01, 0.24911972E-01, 0.41393071E-01, 0.39408021E-01,
             -0.64998358E-01,-0.18124787E-01,-0.27285462E-01, 0.16500179E+00,
              0.29878315E+00, 0.13220957E-02,-0.14755523E-02, 0.31277381E-02,
              0.15555121E-03, },
            {-0.56562969E-02, 0.10716920E-01, 0.58698531E-01,-0.22612128E-02,
             -0.11211977E+00,-0.11326517E+00, 0.88223264E-01, 0.13487896E+00,
              0.22745983E+00,-0.64147316E-01, 0.13011377E-01,-0.15176603E-01,
              0.37889306E-01, },
            { 0.18017766E+00, 0.76493047E-01,-0.57414532E-01,-0.15384177E-01,
              0.19273417E+00,-0.67927875E-01,-0.18422922E+00,-0.22047012E+00,
              0.14405789E-01, 0.12134630E-01,-0.35937384E-01,-0.15340685E-02,
              0.19010877E-01, },
            {-0.14448559E+00,-0.33150800E-02,-0.10132004E-01, 0.44721063E-01,
             -0.38510407E-02, 0.65288940E-02, 0.30064268E-01,-0.13620453E-01,
             -0.25559587E-01, 0.13974449E+00, 0.27066708E+00,-0.86301874E-03,
             -0.15927410E-01, },
            { 0.85362434E-01, 0.48635934E-01,-0.30831050E-01, 0.35901424E-02,
             -0.16469320E-01,-0.93949521E-02, 0.62366962E-02, 0.42513665E-02,
              0.16225785E-01,-0.26053265E+00, 0.12521908E+00,-0.24030069E-02,
             -0.12820498E-02, },
            { 0.42795010E-01,-0.88504680E-01,-0.15416031E-01, 0.35100665E-01,
              0.10118750E-01, 0.25384672E-01,-0.40251631E-02, 0.18797457E-02,
             -0.41919933E-02,-0.62470161E-02,-0.65072584E-02,-0.15718541E+00,
              0.42815138E-01, },
            { 0.51770523E-01, 0.23641726E-01,-0.22212747E-01, 0.52224710E-02,
              0.48885681E-01, 0.18276574E-01,-0.21028422E-01,-0.14288486E-01,
             -0.28683112E-02, 0.25293042E-03,-0.83165923E-02,-0.44567961E-01,
             -0.16046225E+00, },
            {-0.36254805E-01,-0.75680152E-01, 0.11980511E-01, 0.15882116E-03,
              0.22734884E-01,-0.10775125E-01,-0.22750691E-01, 0.10671292E-01,
             -0.29391658E-02,-0.48969565E-02,-0.22469327E-01, 0.37602372E-02,
              0.80211431E-03, },
            { 0.21320259E-01,-0.17746110E-01, 0.20988828E-02,-0.17451579E-01,
             -0.29435143E-01, 0.10905027E-01,-0.15734235E-01,-0.13093308E-01,
              0.74879108E-02,-0.76296972E-02, 0.12714958E-01, 0.12751132E-01,
              0.24891517E-02, },
            {-0.13963948E-01, 0.43464817E-01,-0.57666734E-01, 0.98980134E-02,
             -0.56504559E-01,-0.33702858E-01, 0.28680092E-01, 0.24618777E-01,
              0.21952307E-01, 0.17210137E-01,-0.88103274E-02, 0.80375150E-02,
              0.10196300E-01, },
            {-0.33362068E-01,-0.18013632E-01,-0.50624602E-01, 0.88627376E-02,
              0.10242135E+00, 0.11714541E-01, 0.93254829E-02, 0.93967225E-02,
             -0.15307956E-01,-0.24555840E-01,-0.17302781E-01, 0.19131057E-02,
              0.42815138E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir19Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30401049E+01,-0.11382078E-01, 0.13699237E-01, 0.13619756E-01,
             -0.15344954E-03,-0.39020018E-03, 0.33493678E-02,-0.11851718E-02,
              0.34948639E-03, },
            {-0.82672872E-01,-0.32174747E-01, 0.38680803E-01, 0.55962011E-01,
              0.32362810E-03,-0.41486818E-01,-0.11401795E-01,-0.68960893E-02,
              0.10805070E-01, },
            { 0.14065900E+01, 0.15397176E+00,-0.18210937E+00,-0.31921729E+00,
             -0.54489747E-02, 0.57063509E-01, 0.12826703E-01,-0.53385347E-02,
              0.62804609E-02, },
            { 0.22763485E+00,-0.14647670E+00,-0.14446510E+00,-0.12818100E+00,
              0.18414596E-01, 0.14890826E+00, 0.45125451E-01, 0.18452298E-01,
             -0.38831923E-01, },
            {-0.28037069E+01,-0.28717342E+00, 0.53387308E+00, 0.77303892E+00,
              0.49879473E-01,-0.18796812E+00,-0.46676476E-01,-0.11675812E-02,
             -0.18786209E-01, },
            {-0.20530869E+00, 0.14871514E+00, 0.13572893E+00, 0.65998554E-01,
             -0.31058760E-01,-0.12001267E+00,-0.36234986E-01,-0.10483883E-01,
              0.33103835E-01, },
            { 0.15812294E+01, 0.16315398E+00,-0.34262037E+00,-0.47133777E+00,
             -0.53432129E-01, 0.13809934E+00, 0.39644614E-01, 0.98345075E-02,
              0.13404462E-01, },
            { 0.11054382E-01,-0.15015076E+00,-0.34573376E+00, 0.15864993E-01,
              0.12899942E-02, 0.38709242E-02,-0.75407006E-04, 0.21587815E-02,
              0.61787381E-02, },
            { 0.11099253E-01, 0.36393246E+00,-0.14289594E+00,-0.51004002E-02,
              0.82212426E-02, 0.36659569E-02,-0.50246581E-02, 0.64045354E-03,
              0.93177753E-02, },
            {-0.18676747E+00, 0.13912369E+00, 0.92345759E-01, 0.45837265E-01,
             -0.24934011E-01,-0.26602110E-01,-0.30030532E-01, 0.49917046E-02,
              0.12582732E-01, },
            { 0.13173369E+00,-0.10819068E+00, 0.15030783E+00,-0.23040416E-01,
              0.13999088E-01,-0.53261607E-02,-0.43439619E-01,-0.90245353E-02,
              0.10945592E+00, },
            { 0.33914685E-01, 0.47822487E+00, 0.16972907E+01,-0.97674638E-01,
             -0.25269112E+00,-0.44677760E-01, 0.67510251E-02, 0.25692284E-01,
             -0.82349569E-01, },
            {-0.35780886E+00,-0.18699245E+01, 0.49952665E+00, 0.71142484E-02,
              0.22229070E-01,-0.12980822E+00, 0.70356480E-02,-0.14538261E-01,
             -0.14582597E+00, },
            { 0.21972475E+01,-0.59621954E+00,-0.96930653E+00,-0.94870478E-01,
              0.64373899E+00, 0.14172481E+00,-0.12882969E+00, 0.14451261E+00,
             -0.11963338E+00, },
            {-0.23363806E+00, 0.16209319E+01,-0.70229024E+00, 0.12993802E+00,
             -0.48182800E-01, 0.18711828E+00, 0.38275993E+00, 0.53427994E-01,
             -0.70009434E+00, },
            {-0.24579667E+00, 0.24078412E-01,-0.16323544E+01,-0.19505771E+00,
              0.68362409E+00, 0.15810391E+00, 0.11939487E+00,-0.77476732E-01,
              0.17349911E+00, },
            { 0.10717825E+01, 0.19524435E+01,-0.70215613E+00, 0.40747777E+00,
             -0.14923325E+00, 0.51048869E+00, 0.74851766E-01,-0.32693017E-01,
              0.38650307E+00, },
            {-0.58413815E+01, 0.16597109E+00, 0.27413237E+01, 0.22096719E+00,
             -0.18455598E+01,-0.18398350E+00, 0.69714391E+00,-0.38460970E+00,
              0.25037786E+00, },
            { 0.16911915E+00,-0.42069697E+01, 0.91982204E+00,-0.53290057E+00,
              0.25880048E+00,-0.68226004E+00,-0.97130626E+00, 0.19385042E-01,
              0.14533879E+01, },
            { 0.40299878E+00,-0.48112997E+00, 0.52594155E+00, 0.29526755E+00,
             -0.43825641E+00,-0.12281952E+00,-0.15333907E+00, 0.56682952E-01,
             -0.86398900E-01, },
            {-0.75084627E+00,-0.45546556E+00, 0.44570705E+00,-0.53101832E+00,
              0.12392121E-01,-0.47902694E+00,-0.12242783E+00, 0.79856642E-01,
             -0.28643292E+00, },
            { 0.42676086E+01, 0.75077915E+00,-0.22261817E+01,-0.27955213E+00,
              0.13397520E+01, 0.14111669E+00,-0.55246305E+00, 0.23521648E+00,
             -0.18291542E+00, },
            {-0.12076446E+00, 0.32277856E+01,-0.13160841E+00, 0.44515896E+00,
             -0.28191578E+00, 0.62046933E+00, 0.68323207E+00,-0.10502839E+00,
             -0.91661125E+00, },
            { 0.27267998E-02,-0.37316999E-02, 0.35979025E-03, 0.47388568E-03,
             -0.13095340E+00, 0.21076865E-03, 0.52354215E-02, 0.26401207E-02,
             -0.60893549E-02, },
            {-0.61952616E-02, 0.12154616E-02, 0.40964223E-02, 0.15712863E+00,
              0.24770111E-01, 0.38494696E-02,-0.92166457E-02, 0.62270334E-03,
             -0.54044821E-02, },
            { 0.42403363E-01, 0.16611996E+00, 0.80097161E-01, 0.56755996E+00,
             -0.33011404E+00, 0.29526917E-01,-0.65970644E-01,-0.28990030E-01,
              0.46952786E-02, },
            {-0.80264293E-01, 0.20948894E-01,-0.29410383E-01, 0.50466293E+00,
              0.66722238E+00, 0.76955371E-01,-0.33780285E-02,-0.26567908E-01,
              0.40409271E-01, },
            {-0.15232775E+00, 0.13101333E+00,-0.91356911E-01, 0.89222513E-01,
             -0.36949329E-02,-0.17322950E-01, 0.55812236E-01,-0.38744360E-01,
              0.16779214E-01, },
            {-0.24620678E+00,-0.79602763E-01,-0.20739435E+00,-0.15338077E+00,
             -0.47849964E-01,-0.14388233E-01, 0.94271958E-01,-0.29404018E-01,
              0.18598402E-01, },
            {-0.18558075E+00,-0.83184230E+00, 0.13150723E-01,-0.22259934E+01,
              0.66283786E+00,-0.31915799E-01, 0.28466687E+00, 0.21174742E+00,
             -0.12655893E-01, },
            { 0.94725484E+00, 0.19020595E+00, 0.40080431E+00,-0.13141518E+01,
             -0.27080829E+01,-0.34854394E+00,-0.15705083E+00, 0.14084280E+00,
             -0.21892302E+00, },
            { 0.17485859E+00,-0.35491335E+00, 0.11176544E+00,-0.19686691E+00,
             -0.70896432E-01, 0.47556851E-01,-0.74512035E-01, 0.51269028E-01,
             -0.11006738E-01, },
            { 0.36457384E+00,-0.93383693E-04, 0.29480577E+00, 0.35257223E+00,
             -0.10449865E+00, 0.26363550E-01,-0.74307755E-01, 0.54344729E-01,
             -0.63038334E-01, },
            { 0.20447034E+00, 0.11452413E+01,-0.16736257E+00, 0.20206418E+01,
             -0.30654338E+00, 0.28216397E-01,-0.26526570E+00,-0.28426144E+00,
              0.90763699E-02, },
            {-0.14910082E+01,-0.37116313E+00,-0.24615522E+00, 0.88091087E+00,
              0.25278471E+01, 0.36680901E+00, 0.36790523E+00,-0.11417954E+00,
              0.26816338E+00, },
            {-0.24069045E-01,-0.37140232E-02, 0.59421258E-02,-0.14975986E-01,
              0.38167016E-03, 0.24848035E-01,-0.14859644E-01, 0.81578468E-03,
             -0.90723382E-02, },
            {-0.75797653E-02,-0.47509167E-01,-0.53818156E-02, 0.15393181E-01,
             -0.22171626E-01, 0.20891447E-01, 0.37496552E-01,-0.90530521E-03,
             -0.27469555E-02, },
            { 0.87826133E-01, 0.61553407E-01, 0.20494960E-01,-0.68606257E-01,
              0.31154953E-01,-0.22373168E-01, 0.19868614E-01,-0.36711791E-02,
              0.80109155E-02, },
            { 0.58815431E-01,-0.10893079E-02,-0.23224903E-01,-0.16784027E-01,
             -0.12939301E-02,-0.17937096E-01,-0.30362142E-01,-0.11328247E-01,
              0.18137729E-01, },
            {-0.83675431E-02,-0.66944212E-01,-0.99108607E-01, 0.53046636E-01,
              0.11491635E-02,-0.97788624E-01,-0.19152731E+00,-0.73964251E-02,
              0.37721645E-01, },
            {-0.16437825E-01, 0.10426622E+00,-0.13097627E-01,-0.31341720E-01,
              0.33724148E-01, 0.20263115E+00,-0.17441693E+00,-0.13966219E-01,
              0.81566256E-03, },
            {-0.30277975E-01,-0.14341920E-01, 0.34723729E-02, 0.66115148E-02,
             -0.40517421E-02, 0.80762738E-02, 0.77816513E-02, 0.50833821E-01,
              0.10282873E-01, },
            {-0.26500834E-01, 0.64238147E-02,-0.28022844E-02,-0.14088667E-01,
             -0.54610550E-03, 0.28645764E-02, 0.12587353E-01,-0.20940429E-01,
              0.47031794E-01, },
            { 0.57847179E-01, 0.32054439E-01, 0.12395522E-01,-0.61136406E-01,
             -0.11976909E-01, 0.27411242E-03, 0.54735653E-02, 0.39643612E-01,
              0.14503116E-01, },
            {-0.14846103E-01,-0.43443698E-01,-0.62784306E-02, 0.66622496E-01,
              0.10906368E-01,-0.16276255E-01,-0.74495091E-02,-0.27881341E-01,
              0.35522427E-01, },
            {-0.26395649E-02,-0.95528336E-02,-0.20187974E-02, 0.48234980E-02,
             -0.36433239E-02, 0.73964102E-02, 0.43250853E-02,-0.13721441E-04,
             -0.82670292E-03, },
            { 0.13747018E-01, 0.26495285E-01,-0.12389235E-01, 0.13934656E-03,
             -0.12922279E-01, 0.92590265E-02, 0.30354068E-02, 0.11110849E-02,
              0.40769838E-02, },
            { 0.33883248E-02, 0.76762335E-02, 0.87840520E-02, 0.22252684E-02,
             -0.17388515E-02, 0.65976731E-02,-0.13436445E-02, 0.53181122E-02,
              0.17218638E-02, },
            { 0.12250467E-01, 0.88767400E-02,-0.20778641E-01, 0.91415541E-02,
             -0.31969624E-02, 0.82466059E-03, 0.72336243E-02,-0.23431671E-02,
              0.52077025E-02, },
         },
         {
            { 0.27045255E+01,-0.13861433E-01, 0.10853585E-01, 0.13674901E-01,
              0.45255958E-02, 0.96232325E-04, 0.28539707E-02,-0.50630595E-04,
             -0.10792224E-02, },
            { 0.33882726E-01, 0.16726006E-01, 0.41411079E-01, 0.21762999E-01,
              0.13571894E-01,-0.18095523E-01,-0.13972076E-01,-0.26165762E-02,
             -0.44640922E-03, },
            { 0.13936986E+01, 0.10243688E+00,-0.12836154E+00,-0.18260194E+00,
             -0.46525557E-01, 0.44762935E-01, 0.57317317E-02, 0.11237833E-01,
              0.40574851E-02, },
            {-0.14981104E+00,-0.25437304E+00,-0.14702123E+00,-0.51582754E-01,
             -0.47143351E-01, 0.47940180E-01, 0.27028264E-01, 0.77501182E-02,
             -0.80215288E-02, },
            {-0.24028523E+01,-0.13707371E+00, 0.35199860E+00, 0.39832273E+00,
              0.85571006E-01,-0.12622128E+00,-0.67097712E-02,-0.48906956E-01,
              0.55381274E-02, },
            { 0.77955551E-01, 0.19398201E+00, 0.13605739E+00, 0.29710919E-01,
              0.28151933E-01,-0.32603387E-01,-0.11237349E-01,-0.34127152E-02,
              0.13379143E-01, },
            { 0.11976627E+01, 0.86586386E-01,-0.22247823E+00,-0.23125169E+00,
             -0.42487212E-01, 0.83616689E-01, 0.73067646E-03, 0.40582355E-01,
             -0.96359570E-02, },
            { 0.17483035E-01,-0.30285546E+00,-0.29345882E+00, 0.99322964E-02,
             -0.17893454E-02, 0.10022582E-02, 0.11504443E-02, 0.16644030E-02,
              0.72499416E-02, },
            { 0.13540727E-02, 0.32227382E+00,-0.29328412E+00,-0.80812844E-02,
              0.39960979E-02, 0.65438594E-02,-0.11646176E-01, 0.39998051E-02,
              0.15032205E-01, },
            {-0.36074266E-01, 0.11459168E+00, 0.90131857E-01, 0.35216373E-02,
             -0.96506746E-02,-0.11520932E-01,-0.12881164E-01, 0.99653052E-03,
              0.34462839E-01, },
            { 0.89647539E-01, 0.63354815E-02, 0.12553091E+00, 0.24806114E-01,
              0.23215953E-01, 0.13266702E-01, 0.39352668E-02, 0.90735722E-02,
              0.86630374E-01, },
            {-0.22010055E+00, 0.76199836E+00, 0.10411052E+01, 0.79058155E-01,
             -0.99691860E-01,-0.41243132E-01, 0.57794455E-01,-0.16476742E-02,
             -0.94741918E-01, },
            {-0.32839933E+00,-0.12193962E+01, 0.87339061E+00,-0.75224414E-02,
              0.83490223E-01,-0.14356729E+00, 0.13259062E+00,-0.44737361E-01,
             -0.20198126E+00, },
            { 0.54963994E+00,-0.52403355E+00,-0.74558896E+00,-0.96110024E-01,
              0.27973866E+00, 0.15445522E-01,-0.24832048E+00, 0.65243810E-01,
             -0.22887573E+00, },
            {-0.19111110E-01, 0.47740996E+00,-0.63515687E+00,-0.18528543E+00,
             -0.21004455E+00, 0.61389271E-01, 0.34061927E-01,-0.37664600E-01,
             -0.56488287E+00, },
            { 0.61485970E+00,-0.31153601E+00,-0.85383409E+00,-0.55827850E+00,
              0.24451426E+00, 0.20632704E+00,-0.12735528E+00, 0.49136013E-01,
              0.29886141E+00, },
            { 0.12342763E+01, 0.80860949E+00,-0.12855700E+01, 0.31017226E+00,
             -0.24868594E+00, 0.44440651E+00,-0.30813733E+00, 0.23854533E-01,
              0.45323089E+00, },
            {-0.15821571E+01, 0.43670562E+00, 0.22294593E+01, 0.54407924E+00,
             -0.86818773E+00,-0.27222033E-02, 0.87385136E+00,-0.10560450E+00,
              0.39706621E+00, },
            {-0.65258825E+00,-0.15170186E+01, 0.12435969E+01, 0.19180834E+00,
              0.70927751E+00,-0.27004617E+00,-0.20135774E+00, 0.98844670E-01,
              0.11511528E+01, },
            {-0.35777101E+00,-0.10984187E+00, 0.54976183E+00, 0.51074892E+00,
             -0.15392701E+00,-0.20111597E+00, 0.95634453E-01,-0.59104066E-01,
             -0.22803462E+00, },
            {-0.10545499E+01,-0.27652279E-01, 0.84639686E+00,-0.34254837E+00,
              0.83599433E-01,-0.39442429E+00, 0.17786869E+00, 0.49440950E-01,
             -0.29044276E+00, },
            { 0.11647863E+01, 0.19355892E+00,-0.19237444E+01,-0.51354492E+00,
              0.63591230E+00, 0.36586221E-01,-0.67268568E+00, 0.16708907E-01,
             -0.21348131E+00, },
            { 0.78157800E+00, 0.13963382E+01,-0.70771319E+00,-0.20638825E-01,
             -0.56146246E+00, 0.25821999E+00, 0.18338437E+00,-0.10435095E+00,
             -0.70114332E+00, },
            { 0.33377009E-02,-0.85557243E-02, 0.53263279E-02, 0.48212931E-01,
             -0.96097551E-01, 0.16747259E-02, 0.45310468E-02, 0.21136762E-02,
             -0.20766114E-02, },
            {-0.66840369E-02,-0.10856335E-01,-0.23787534E-02, 0.92386574E-01,
              0.68768784E-01, 0.10402149E-02,-0.86118113E-02, 0.32925498E-03,
              0.57025033E-03, },
            { 0.51790807E-01, 0.13119574E+00, 0.47563083E-01, 0.28507447E+00,
             -0.11513667E+00, 0.31583652E-01,-0.17520612E-01,-0.31817120E-01,
              0.82542188E-02, },
            { 0.19457083E-01, 0.16882727E-01,-0.44377230E-01, 0.21419054E+00,
              0.39532006E+00, 0.44337548E-01, 0.33838246E-01,-0.26187995E-01,
              0.26135845E-01, },
            {-0.13655464E+00, 0.12530467E+00,-0.43442167E-01,-0.23065427E+00,
              0.20435360E+00,-0.63712358E-01, 0.23470277E-01,-0.50777260E-01,
              0.77459794E-02, },
            {-0.15135796E+00, 0.11079714E-01,-0.65361388E-01,-0.20480011E+00,
             -0.31951770E+00,-0.43420333E-01, 0.58943622E-01,-0.19627888E-01,
             -0.11768199E-01, },
            {-0.15009120E+00,-0.68192893E+00, 0.62561356E-01,-0.10389147E+01,
              0.10595896E+00,-0.14849147E-01, 0.53832039E-01, 0.23531134E+00,
             -0.12635956E-01, },
            { 0.28327590E+00, 0.38813151E-01, 0.31918576E+00,-0.46337992E+00,
             -0.14411012E+01,-0.15281472E+00,-0.21893854E+00, 0.14097042E+00,
             -0.14544623E+00, },
            { 0.15800914E+00,-0.34267643E+00, 0.18755988E-02,-0.56562793E-01,
             -0.48462987E+00, 0.95949911E-01,-0.30947629E-01, 0.87747708E-01,
             -0.19303991E-01, },
            { 0.24605586E+00,-0.13650826E+00, 0.21064965E-01, 0.58473659E+00,
             -0.75870594E-02, 0.39007965E-01,-0.65296404E-01, 0.41345578E-01,
             -0.19478334E-01, },
            { 0.78472018E-01, 0.91951132E+00,-0.16342767E+00, 0.91248029E+00,
              0.87954272E-02,-0.59618726E-02,-0.40209666E-01,-0.31057978E+00,
              0.18426050E-01, },
            {-0.60459119E+00,-0.10414231E-01,-0.18814395E+00, 0.27712169E+00,
              0.13081124E+01, 0.17085785E+00, 0.34071115E+00,-0.14231364E+00,
              0.17387938E+00, },
            {-0.20386552E-01, 0.21182620E-02, 0.10177098E-01,-0.11378626E-01,
             -0.98456312E-02, 0.34398962E-01,-0.42581428E-01, 0.34251876E-03,
             -0.68609137E-02, },
            {-0.90750623E-02,-0.32029830E-01,-0.47561587E-02, 0.14946333E-01,
             -0.16567832E-01, 0.42375118E-01, 0.41583791E-01, 0.13377547E-03,
             -0.61774603E-02, },
            { 0.45967977E-01, 0.61088778E-01, 0.14942785E-01,-0.37811134E-01,
              0.15868828E-01, 0.47008302E-02, 0.46885707E-02, 0.11044603E-02,
              0.83790062E-03, },
            { 0.33965636E-01, 0.12285841E-01,-0.15363421E-01,-0.20721937E-01,
             -0.11257234E-01,-0.34942648E-02,-0.12950346E-01,-0.59640319E-02,
              0.12831651E-01, },
            { 0.60491147E-03,-0.74729361E-01,-0.69819920E-01, 0.26307790E-01,
              0.17833915E-01,-0.10292423E+00,-0.14374946E+00,-0.44204816E-02,
              0.32788616E-01, },
            {-0.15078791E-01, 0.62297020E-01,-0.27057173E-01,-0.22026932E-01,
              0.26395356E-01, 0.16220757E+00,-0.12398390E+00,-0.14915660E-01,
              0.10779776E-01, },
            {-0.22808822E-01,-0.15335506E-01, 0.21026686E-02, 0.44900649E-02,
             -0.13017196E-02, 0.27200764E-02, 0.10563135E-02, 0.33134330E-01,
              0.16995138E-01, },
            {-0.15292642E-01, 0.24188119E-02,-0.86551392E-02,-0.19452726E-02,
             -0.34849253E-02, 0.35679105E-02, 0.10176241E-01,-0.22009404E-01,
              0.29706346E-01, },
            { 0.43338250E-01, 0.19471711E-01, 0.19346902E-03,-0.33331003E-01,
             -0.12334583E-01,-0.12015219E-01, 0.14972220E-02, 0.39651532E-01,
              0.19308904E-01, },
            { 0.14977311E-02,-0.34002740E-01, 0.13617956E-02, 0.26084950E-01,
              0.37475578E-02,-0.10403191E-01,-0.17990541E-01,-0.22997433E-01,
              0.27994959E-01, },
            { 0.22576404E-03,-0.16536338E-01,-0.75725868E-03, 0.69075981E-02,
             -0.14428189E-02, 0.61770566E-02, 0.60216250E-03,-0.43525919E-03,
             -0.13888464E-02, },
            { 0.90428498E-02, 0.19430207E-01,-0.10300137E-01, 0.35099406E-02,
             -0.42426982E-02, 0.60566198E-02, 0.52108942E-02,-0.11467599E-03,
              0.65590143E-02, },
            { 0.49651740E-02, 0.11645567E-01, 0.83609400E-02, 0.52899164E-02,
             -0.47522220E-02, 0.55126837E-02, 0.35592569E-02, 0.34613642E-02,
              0.36597401E-02, },
            { 0.45108609E-02, 0.48094862E-02,-0.14215787E-01, 0.53761145E-02,
              0.45948825E-02, 0.83546591E-03, 0.53303521E-02,-0.26749300E-02,
              0.29056252E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir20F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.67417641E+01,-0.11025394E+00, 0.42666834E-01, 0.69521941E-01,
              0.38874124E-02, 0.29258806E-01,-0.31417802E-01, 0.35931028E-01,
              0.11918333E-01, 0.29975636E-01, 0.11452777E-01,-0.29228499E-01,
             -0.76055541E-02, },
            { 0.17436563E+01, 0.40657678E+00, 0.59440559E+00,-0.81161804E-01,
             -0.12061444E+00,-0.42257652E-01,-0.16804051E-01,-0.13078281E+00,
              0.99271558E-01,-0.51369164E-01,-0.85818343E-01, 0.20449582E-01,
             -0.34776747E-01, },
            { 0.15906835E+02, 0.21122408E+01,-0.30194516E+01,-0.18164214E+00,
              0.21544621E+01,-0.77579105E+00, 0.10776812E+01,-0.65275151E+00,
             -0.18329655E+00,-0.10741085E+01,-0.21992724E+00, 0.54272276E+00,
              0.26474300E+00, },
            {-0.21567724E+02,-0.10071519E+02,-0.46050444E+01, 0.22639155E+01,
              0.77192564E+01, 0.70966089E+00,-0.21318839E+01, 0.47265267E+00,
             -0.24995817E-01, 0.65379196E+00, 0.20589564E+01,-0.17014968E+00,
              0.56020799E-02, },
            {-0.10173342E+03,-0.68225641E+01, 0.18339714E+02,-0.96500959E+01,
             -0.22774458E+02, 0.70135794E+01,-0.57714853E+01, 0.20234113E+01,
              0.98741543E+00, 0.70813155E+01, 0.75227982E+00,-0.31741402E+01,
             -0.28904328E+01, },
            { 0.65307861E+02, 0.35732182E+02, 0.99326601E+01,-0.70676713E+01,
             -0.35474052E+02,-0.48456573E+01, 0.10269098E+02, 0.14416782E+01,
             -0.37578681E+01,-0.17822330E+01,-0.10741420E+02, 0.18747391E+01,
              0.18222950E+01, },
            { 0.20845084E+03, 0.83894491E+01,-0.37941162E+02, 0.42702366E+02,
              0.68335434E+02,-0.22162090E+02, 0.11464567E+02,-0.11181591E+01,
             -0.30024517E+01,-0.17626598E+02,-0.13377714E+00, 0.74559193E+01,
              0.94865475E+01, },
            {-0.69698105E+02,-0.40740063E+02,-0.12088245E+02, 0.55066094E+01,
              0.60161949E+02, 0.12805878E+02,-0.14906711E+02,-0.69451771E+01,
              0.13351114E+02, 0.15886544E+01, 0.22924721E+02,-0.65486469E+01,
             -0.64081602E+01, },
            {-0.19967574E+03,-0.47658625E+01, 0.32122219E+02,-0.59845806E+02,
             -0.79271217E+02, 0.26973694E+02,-0.94852943E+01,-0.13761377E+01,
              0.42310858E+01, 0.18761662E+02,-0.14811943E+01,-0.75402808E+01,
             -0.11900676E+02, },
            { 0.14291371E+02, 0.10281889E+02, 0.12892766E+02, 0.20141015E+01,
             -0.43246784E+02,-0.13862740E+02, 0.66849313E+01, 0.83780088E+01,
             -0.16522486E+02,-0.29590270E+00,-0.21744144E+02, 0.82283382E+01,
              0.76309428E+01, },
            { 0.74218613E+02, 0.12529610E+01,-0.98135452E+01, 0.27000690E+02,
              0.31554647E+02,-0.11102576E+02, 0.27184381E+01, 0.10648201E+01,
             -0.20641975E+01,-0.71944165E+01, 0.10889814E+01, 0.27322817E+01,
              0.50644026E+01, },
            { 0.97484446E+01, 0.43847666E+01,-0.68491726E+01,-0.25755978E+01,
              0.10981911E+02, 0.51441727E+01, 0.12638378E+00,-0.31936567E+01,
              0.68834958E+01,-0.72948217E-01, 0.75957398E+01,-0.34138052E+01,
             -0.30203903E+01, },
            {-0.17938741E-01, 0.19350545E+01, 0.15287569E+01, 0.26359454E-01,
             -0.69553903E-02, 0.48199920E-02, 0.15078617E-01, 0.53562142E-01,
              0.21571161E-01, 0.32001279E-01, 0.37023298E-01,-0.33349071E-01,
              0.10461510E-01, },
            {-0.71816571E-01,-0.14951198E+01, 0.19070232E+01,-0.90120353E-01,
              0.83764896E-01,-0.22245713E-01,-0.61692059E-01, 0.29028347E-01,
             -0.15067049E-03, 0.12577567E-01,-0.22862754E-02, 0.29112538E-01,
             -0.23672460E-01, },
            {-0.51051003E+00, 0.87592429E+00, 0.11411304E+01, 0.22836588E+00,
              0.10007082E+00,-0.85580871E-02, 0.14482203E+00, 0.23493987E+00,
              0.12527980E+00,-0.16819620E+00,-0.48568074E-01, 0.24008988E+00,
              0.96231982E-01, },
            { 0.10973550E+00, 0.93418971E-01,-0.70641696E-01,-0.53768361E+00,
              0.48843423E+00, 0.34973875E+00,-0.26007736E+00, 0.36855582E-01,
              0.41532871E+00, 0.89809448E-01,-0.23388444E-01, 0.20598353E+00,
              0.13492872E+00, },
            { 0.54216437E+01, 0.19527584E+02, 0.53556581E+01,-0.20208032E+01,
              0.27149031E+01, 0.11244015E+01,-0.42923173E+00,-0.90001625E+00,
             -0.22988829E+00,-0.79169679E+00, 0.95905103E-01, 0.54770768E+00,
              0.40796738E-01, },
            {-0.49918337E+01,-0.68213148E+01, 0.20141775E+02, 0.33453064E+01,
              0.20416534E+01,-0.18846661E+00, 0.78771967E+00, 0.12231276E+01,
              0.63756788E+00,-0.58407247E+00,-0.35017237E+00,-0.12065351E+01,
             -0.24117194E+00, },
            { 0.15664848E+02,-0.58232422E+01, 0.16471388E+01,-0.85785646E+01,
              0.65734754E+01, 0.10979421E+01,-0.61561337E+01,-0.41306434E+01,
             -0.25767187E-01, 0.22381325E+01, 0.16197391E+01,-0.47094798E+01,
             -0.19486105E+01, },
            { 0.54404092E+01,-0.23112692E+02, 0.87230244E+01, 0.41379757E+01,
             -0.82567225E+01,-0.59944463E+00, 0.49454799E+01, 0.98528072E-01,
             -0.69704366E+01,-0.26175516E+01, 0.21983194E+01,-0.85137612E+00,
             -0.48446989E+01, },
            {-0.34786564E+02,-0.13385123E+03, 0.53099270E+02, 0.98495770E+01,
             -0.21858513E+02,-0.12796616E+02, 0.21972885E+01, 0.71941557E+01,
              0.17378716E+01, 0.38053644E+01,-0.41692481E+01,-0.43841825E+01,
             -0.76313013E+00, },
            { 0.52958324E+02,-0.62617264E+02,-0.13794221E+03,-0.20468203E+02,
             -0.26598701E+02, 0.76006281E+00,-0.45324337E+00,-0.73798909E+01,
             -0.68499522E+01, 0.43220587E+01, 0.26352139E+01, 0.11029831E+02,
              0.39386227E+01, },
            {-0.10278553E+03,-0.11709525E+02, 0.43092846E+02, 0.72482162E+02,
             -0.24375433E+02,-0.46230431E+01, 0.37790916E+02, 0.18252401E+02,
              0.32258278E+00,-0.98822670E+01,-0.65691562E+01, 0.26157690E+02,
              0.14764631E+02, },
            {-0.63290554E+02, 0.11773388E+03,-0.50047344E+02,-0.31375063E+02,
              0.31207630E+02,-0.13806309E+02,-0.21457752E+02,-0.76790338E+01,
              0.33817787E+02, 0.16895014E+02,-0.18257301E+02,-0.33989453E+01,
              0.31438547E+02, },
            { 0.73195847E+02, 0.30527811E+03,-0.27625671E+03,-0.13173969E+02,
              0.52327705E+02, 0.43725922E+02,-0.53056192E+01,-0.26080379E+02,
             -0.80343571E+01,-0.59193101E+01, 0.15645749E+02, 0.14612919E+02,
              0.19771434E+01, },
            {-0.16780865E+03, 0.33273395E+03, 0.31980917E+03, 0.61927944E+02,
              0.10291243E+03,-0.41476986E+00,-0.10810330E+02, 0.16019861E+02,
              0.22894283E+02,-0.96919661E+01,-0.76568937E+01,-0.35144093E+02,
             -0.13620372E+02, },
            { 0.27711322E+03, 0.76293732E+02,-0.24416869E+03,-0.23412332E+03,
              0.59856181E+01, 0.27036533E+01,-0.85468987E+02,-0.32569973E+02,
             -0.93770189E+01, 0.22770227E+02, 0.79303355E+01,-0.62524040E+02,
             -0.43537697E+02, },
            { 0.18402339E+03,-0.25016977E+03, 0.72703697E+02, 0.10605528E+03,
             -0.51503254E+02, 0.54460190E+02, 0.38909683E+02, 0.36796562E+02,
             -0.68024529E+02,-0.41103397E+02, 0.53216114E+02, 0.20626389E+02,
             -0.76586349E+02, },
            {-0.53355736E+02,-0.31135284E+03, 0.40809729E+03, 0.22601173E+01,
             -0.48498844E+02,-0.60832169E+02, 0.60337324E+01, 0.38279335E+02,
              0.14871668E+02, 0.28498704E+01,-0.20230135E+02,-0.19879827E+02,
             -0.18497735E+01, },
            { 0.21465652E+03,-0.48994763E+03,-0.32978488E+03,-0.86516556E+02,
             -0.15399602E+03,-0.53305179E+00, 0.26561083E+02,-0.14605586E+02,
             -0.29598482E+02, 0.74684277E+01, 0.98743944E+01, 0.45331917E+02,
              0.17090784E+02, },
            {-0.33723950E+03,-0.97724586E+02, 0.38985550E+03, 0.31390698E+03,
              0.46843403E+02, 0.78864837E+01, 0.82173851E+02, 0.24357899E+02,
              0.20966228E+02,-0.26853407E+02,-0.14723213E+01, 0.67029839E+02,
              0.53401054E+02, },
            {-0.18740387E+03, 0.25239400E+03,-0.10743400E+02,-0.13630225E+03,
              0.44591602E+02,-0.69500046E+02,-0.34749329E+02,-0.57859837E+02,
              0.60593739E+02, 0.43181362E+02,-0.64212608E+02,-0.31543680E+02,
              0.79737732E+02, },
            { 0.79219427E+01, 0.12230792E+03,-0.19162122E+03, 0.30479712E+01,
              0.14701435E+02, 0.29905254E+02,-0.25530334E+01,-0.18944752E+02,
             -0.87824144E+01,-0.98065168E-01, 0.87094259E+01, 0.93237696E+01,
              0.71487147E+00, },
            {-0.98071724E+02, 0.22642126E+03, 0.13037717E+03, 0.42995682E+02,
              0.77473969E+02, 0.44055769E+00,-0.17098782E+02, 0.44686174E+01,
              0.12881230E+02,-0.12471923E+01,-0.46582799E+01,-0.20446924E+02,
             -0.70700450E+01, },
            { 0.15115994E+03, 0.37530956E+02,-0.19583463E+03,-0.14785715E+03,
             -0.36896564E+02,-0.74833941E+01,-0.28641958E+02,-0.58377399E+01,
             -0.12369586E+02, 0.12254857E+02,-0.14293088E+01,-0.26398525E+02,
             -0.23067722E+02, },
            { 0.59227791E+02,-0.99076767E+02,-0.24037251E+02, 0.57410080E+02,
             -0.17986750E+02, 0.29093678E+02, 0.13013299E+02, 0.29289719E+02,
             -0.19803263E+02,-0.16676153E+02, 0.27448059E+02, 0.15189789E+02,
             -0.29994450E+02, },
            { 0.11391176E+00, 0.11553436E+00,-0.13704097E+00,-0.10008467E+01,
             -0.24247713E-01, 0.66881222E-02,-0.69679026E-02,-0.18138342E-01,
              0.27570529E-01, 0.35871942E-01, 0.73015732E-02,-0.69541764E-02,
              0.26003541E-01, },
            {-0.63337181E-02, 0.59983626E-01,-0.62968001E-01,-0.78753326E-02,
             -0.75065094E+00,-0.37398331E-01,-0.44260029E-01,-0.36373198E-01,
              0.74006170E-02,-0.10022034E-02, 0.77789617E-02, 0.12514464E-01,
             -0.10064091E-01, },
            { 0.86539829E+00, 0.32539025E+00,-0.91669343E-01,-0.73207444E+00,
              0.26933301E-01, 0.13879360E+00,-0.26311162E+00,-0.37850922E+00,
             -0.21304099E+00,-0.14258491E+00,-0.13731001E+00,-0.30468129E-01,
              0.92138372E-01, },
            { 0.98262489E+00,-0.15070742E+00, 0.71532860E-01,-0.49188623E+00,
             -0.57117575E+00, 0.43504018E+00, 0.19652760E+00,-0.13652391E-02,
             -0.67386083E-01,-0.13397166E+00,-0.20012897E+00, 0.14705022E+00,
              0.10473770E-01, },
            { 0.65290123E+00,-0.26946282E+01, 0.31813004E+01, 0.30024579E+01,
             -0.75441279E-01,-0.15592440E+01,-0.21163237E+00,-0.12697160E+00,
             -0.19921806E+00,-0.66176975E+00, 0.10346518E+00,-0.18251108E+00,
              0.17797954E-01, },
            { 0.18768592E+00,-0.25953684E+01, 0.43567185E+01,-0.59679967E+00,
              0.19172314E+01,-0.20841491E+00, 0.28366706E+00, 0.23806677E+01,
             -0.16260111E+00, 0.13395765E+00,-0.83577555E+00,-0.23089887E+00,
             -0.83205447E-03, },
            {-0.53307676E+01, 0.61688054E+00, 0.14718913E+01, 0.10001513E+02,
              0.29644825E+01,-0.11001223E+01, 0.14558415E+01, 0.26624279E+01,
              0.21787982E+01, 0.13549827E+01, 0.16713557E+01,-0.15284111E+00,
             -0.52187556E+00, },
            {-0.72269955E+01, 0.91789746E+00, 0.22233882E+01,-0.45572978E+00,
              0.10729968E+02,-0.29951901E+01,-0.21556865E+00, 0.35207564E+00,
              0.13136928E+01, 0.57696944E+00, 0.23771760E+01,-0.17125853E+01,
             -0.86974633E+00, },
            {-0.36376338E+01, 0.24080009E+02,-0.22185778E+02, 0.18825312E+02,
              0.14852665E+02, 0.13149285E+02, 0.78031552E+00, 0.31474850E+01,
              0.90485042E+00, 0.16362022E+01,-0.48030415E+00, 0.18465122E+01,
             -0.83016783E+00, },
            {-0.18757300E+01, 0.21337709E+02,-0.17421923E+02,-0.10823549E+02,
              0.15082483E+02,-0.15371397E+01,-0.41927969E+00,-0.11322726E+02,
              0.11449955E+01,-0.39108786E+00, 0.54798636E+01, 0.16054937E+01,
              0.82940835E+00, },
            { 0.90842657E+01,-0.63919301E+01,-0.66001620E+01,-0.23100986E+02,
             -0.12172362E+02, 0.22952433E+01,-0.42773733E+01,-0.51716614E+01,
             -0.57398787E+01,-0.32952104E+01,-0.44157414E+01, 0.97596109E+00,
              0.49134457E+00, },
            { 0.14462518E+02,-0.52479367E+01,-0.77367229E+01, 0.63962259E+01,
             -0.26448858E+02, 0.69845829E+01,-0.13445467E+01,-0.86556637E+00,
             -0.37851818E+01,-0.59372693E+00,-0.62924967E+01, 0.47564263E+01,
              0.29345119E+01, },
            { 0.36482801E+01,-0.54122810E+02, 0.51179451E+02,-0.67288895E+02,
             -0.36518002E+02,-0.31562344E+02,-0.62648946E+00,-0.97915506E+01,
             -0.14420407E+01,-0.11084070E+01, 0.88024229E+00,-0.53258376E+01,
              0.19962699E+01, },
            { 0.75823536E+01,-0.47701122E+02, 0.20454948E+02, 0.31747019E+02,
             -0.50241886E+02, 0.81888828E+01, 0.43072930E+00, 0.17209768E+02,
             -0.32296555E+01,-0.40862790E+00,-0.11610971E+02,-0.27689962E+01,
             -0.20991716E+01, },
            {-0.50960884E+01, 0.83177357E+01, 0.55298500E+01, 0.15825745E+02,
              0.13216482E+02,-0.14331723E+01, 0.36789041E+01, 0.29861968E+01,
              0.41942501E+01, 0.21892414E+01, 0.33499987E+01,-0.10724663E+01,
              0.32862183E-01, },
            {-0.85228519E+01, 0.73134866E+01, 0.61695185E+01,-0.86385641E+01,
              0.17545185E+02,-0.52412457E+01, 0.18066072E+01, 0.71043044E+00,
              0.29276321E+01,-0.61344326E-01, 0.45447369E+01,-0.35246754E+01,
             -0.23307092E+01, },
            {-0.14690654E+01, 0.35724678E+02,-0.35846252E+02, 0.50520844E+02,
              0.23054344E+02, 0.21617199E+02, 0.42514130E+00, 0.79941406E+01,
              0.58302629E+00, 0.34610984E+00,-0.60215044E+00, 0.39643900E+01,
             -0.13188493E+01, },
            {-0.92348213E+01, 0.30514215E+02,-0.52573867E+01,-0.22224283E+02,
              0.36993191E+02,-0.71565070E+01,-0.77821231E+00,-0.85328360E+01,
              0.27982295E+01, 0.89062178E+00, 0.76935654E+01, 0.13601083E+01,
              0.11444426E+01, },
            {-0.74351348E-01, 0.11168007E+00,-0.12678429E+00,-0.30625785E-01,
              0.81547037E-01,-0.26790616E+00,-0.77349359E+00, 0.18570652E-01,
              0.38729448E-01, 0.77069057E-02, 0.12018093E-01,-0.19658063E-01,
             -0.86077964E-02, },
            { 0.12409092E+00,-0.10666509E+00, 0.45101717E-01, 0.15352032E-01,
              0.32349147E-01, 0.78267974E+00,-0.29498765E+00,-0.52898012E-01,
              0.61348882E-01, 0.11466916E-01, 0.13180029E-02, 0.22439856E-01,
              0.54106647E-02, },
            { 0.23398851E+00, 0.64493269E+00,-0.25810570E+00,-0.12363273E+00,
             -0.66201246E+00,-0.27819201E+00,-0.30108163E+00,-0.24624993E+00,
              0.83835699E-01,-0.74149808E-02, 0.13186117E+00,-0.16377572E-01,
             -0.61401571E-02, },
            {-0.10936357E+00,-0.29475041E-01, 0.53700852E+00, 0.50398592E-01,
             -0.11941080E+00, 0.84602892E+00,-0.45599923E+00,-0.29877833E+00,
             -0.10263894E+00,-0.88044167E-01,-0.85591517E-01, 0.64158496E-02,
              0.54729246E-02, },
            { 0.34892559E+00, 0.13616563E-01, 0.68630129E+00, 0.92597771E+00,
             -0.78667665E+00, 0.75528562E+00, 0.13148088E+01, 0.89991748E-01,
             -0.31526265E+00, 0.76205105E-01,-0.15212275E+00, 0.11211587E+00,
              0.32542087E-01, },
            {-0.11553695E+01, 0.24110079E+01,-0.11516421E+01, 0.59397310E+00,
             -0.87734061E+00,-0.15284475E+01, 0.15883173E+01, 0.54522204E+00,
             -0.40533349E+00, 0.17719109E-01,-0.13833664E+00,-0.10189562E+00,
             -0.79786859E-01, },
            {-0.51425898E+00,-0.15127133E+01, 0.96630716E+00, 0.11464612E+00,
              0.11833749E+01, 0.31068289E+00,-0.74194200E-01, 0.72967577E+00,
             -0.67832701E-01,-0.77846984E-03,-0.20988442E+00,-0.51187862E-01,
              0.38249537E-01, },
            {-0.78501157E-01, 0.32682669E+00,-0.10028919E+01,-0.68409115E+00,
              0.19638067E+00,-0.12582335E+01, 0.71608400E+00, 0.40842831E+00,
              0.26023892E+00, 0.20802367E+00, 0.18907779E+00,-0.95071457E-01,
             -0.14524133E-01, },
            { 0.70622310E-01,-0.23873582E-01,-0.46930566E+00,-0.11361483E+01,
              0.12044519E+01,-0.13642393E+01,-0.16782405E+01,-0.27571243E+00,
              0.62265939E+00,-0.27575913E+00, 0.26088655E+00,-0.17131169E+00,
             -0.96791862E-02, },
            { 0.16397073E+01,-0.35624073E+01, 0.21964977E+01,-0.69169527E+00,
              0.17876031E+01, 0.23638010E+01,-0.29829438E+01,-0.65094346E+00,
              0.50062829E+00,-0.78845382E-01, 0.20135802E+00, 0.35707735E-01,
              0.43332722E-01, },
            {-0.73506460E-02,-0.20373339E-01,-0.59456583E-02, 0.47092013E-01,
              0.59794255E-01,-0.22777202E-01, 0.34936242E-01, 0.28207561E+00,
             -0.12549119E+00, 0.68888762E-02, 0.73659769E-02,-0.16957685E-01,
             -0.69591464E-02, },
            { 0.60888998E-01, 0.48170742E-01, 0.90240315E-02,-0.23745389E-02,
             -0.88016756E-01, 0.73502366E-02,-0.11265669E-02, 0.12217470E+00,
              0.24448656E+00,-0.23252072E-02,-0.70067816E-02,-0.43141777E-02,
             -0.13638792E-01, },
            {-0.21918522E-01, 0.39478119E-01, 0.21353072E-01,-0.17231616E+00,
             -0.24383087E+00,-0.18098043E+00, 0.78112222E-01,-0.96717022E-01,
             -0.21291029E+00,-0.78414723E-01, 0.27862417E-01, 0.96148206E-02,
              0.23770265E-01, },
            {-0.13406623E-01,-0.97504027E-01,-0.22518177E-01, 0.13743083E+00,
              0.13689157E+00,-0.19897047E-01,-0.10379756E+00, 0.18668231E+00,
             -0.28371491E-01,-0.66896580E-01,-0.45131464E-01, 0.52789491E-01,
             -0.40041968E-01, },
            {-0.66618256E-01, 0.42511921E-01,-0.38794965E-01,-0.10955274E-01,
              0.30871356E-01, 0.52086008E-02, 0.24081491E-01,-0.67935861E-02,
             -0.13217943E-01, 0.82531139E-01, 0.28200027E+00,-0.76102950E-02,
             -0.90683848E-02, },
            { 0.23629055E-02,-0.10296953E-02,-0.14055242E-01,-0.31999350E-01,
             -0.57006430E-01, 0.18182561E-01,-0.55214246E-02, 0.49740784E-02,
             -0.54625701E-01,-0.27094826E+00, 0.51207963E-01, 0.45274422E-02,
              0.16952408E-01, },
            { 0.41054886E-01,-0.54486718E-01,-0.10627154E-01,-0.28977985E-01,
              0.76224119E-02, 0.58471583E-01,-0.90225777E-02, 0.22696028E-01,
             -0.16175291E-01, 0.21533300E-02, 0.36474864E-02,-0.14331755E+00,
             -0.11911416E-01, },
            {-0.35727635E-01,-0.25858898E-01,-0.81544844E-02, 0.45217113E-02,
              0.82371887E-02, 0.27359465E-01,-0.12122327E-01,-0.90430006E-02,
             -0.25242612E-01,-0.10398197E-01,-0.11534831E-01, 0.29744126E-01,
             -0.14588501E+00, },
            {-0.27973065E-01,-0.17960723E-01,-0.72880231E-01,-0.13211422E-01,
              0.74935732E-02, 0.40451571E-01,-0.13288229E-02, 0.23200864E-01,
             -0.21652490E-01, 0.10630924E-01,-0.19871013E-01,-0.89811655E-02,
              0.44909436E-02, },
            { 0.15431918E-01, 0.17651659E-01, 0.37659165E-01, 0.40000812E-02,
              0.67411782E-02, 0.39261781E-01, 0.64045368E-02, 0.15160269E-01,
              0.13118289E-01,-0.23408243E-01,-0.30519590E-02,-0.95153069E-02,
             -0.35423130E-01, },
            { 0.18906865E-01, 0.64482591E-02,-0.96611172E-01, 0.15293910E-01,
             -0.54000065E-01, 0.30757926E-01, 0.59357233E-01, 0.36398362E-01,
              0.25581049E-02, 0.11665761E-01, 0.56256289E-02, 0.57567651E-02,
              0.31496417E-02, },
            {-0.46144247E-01, 0.41687142E-01,-0.34603193E-01, 0.53481121E-01,
              0.11668256E-01,-0.49698236E-02, 0.57190461E-02, 0.17202956E-02,
             -0.39180489E-02,-0.26724448E-01,-0.32743976E-01, 0.17875217E-01,
             -0.13148283E-01, },
         },
         {
            { 0.97347717E+01, 0.80199363E-02, 0.88190660E-01, 0.34591865E-01,
             -0.20683805E-01, 0.29348714E-01,-0.21960216E-01,-0.94146077E-02,
              0.15242774E-02, 0.18784342E-01,-0.42741881E-02,-0.11803485E-01,
              0.89114830E-02, },
            { 0.17238567E+01,-0.14606644E+00,-0.23781560E+00,-0.18891615E+00,
             -0.83614878E-01, 0.48989985E-01,-0.23082246E+00,-0.45915204E+00,
              0.52839011E-01, 0.65929607E-01,-0.27021214E+00, 0.59767954E-01,
              0.45384470E-01, },
            { 0.26139549E+02,-0.26177337E+01,-0.18818336E+01, 0.28918128E+01,
              0.26626935E+01,-0.13387190E+01, 0.35804051E+00, 0.61114317E+00,
             -0.39724249E+00,-0.34554461E+00, 0.11107710E-01, 0.46736574E+00,
             -0.14963689E+00, },
            {-0.14329063E+02, 0.17789959E+01, 0.11053965E+02, 0.40939870E+01,
              0.74688954E+01, 0.81202614E+00, 0.18834094E+01, 0.78406534E+01,
              0.12138880E+01,-0.18455257E+01, 0.44079385E+01,-0.87374485E+00,
             -0.19130919E+01, },
            {-0.12161191E+03, 0.20769333E+02, 0.10549552E+02,-0.25936485E+02,
             -0.22852907E+02, 0.10790629E+02, 0.59407198E+00,-0.15447999E+01,
              0.52851825E+01, 0.17490312E+00, 0.17242504E+01,-0.27120554E+01,
              0.20165932E+00, },
            { 0.56903176E+01,-0.30547829E+02,-0.53477966E+02,-0.13067127E+02,
             -0.37363407E+02,-0.77788720E+01,-0.83473549E+01,-0.36846283E+02,
             -0.10496927E+02, 0.98110600E+01,-0.20411940E+02, 0.33325286E+01,
              0.11927084E+02, },
            { 0.18244954E+03,-0.48165424E+02,-0.24666487E+02, 0.77004440E+02,
              0.64335304E+02,-0.29243431E+02,-0.68443341E+01,-0.68185717E+00,
             -0.17612892E+02, 0.27443240E+01,-0.67463579E+01, 0.52837248E+01,
              0.84665340E+00, },
            { 0.83355324E+02, 0.10590702E+03, 0.85216660E+02, 0.10837219E+02,
              0.71135712E+02, 0.20155458E+02, 0.19259611E+02, 0.72910683E+02,
              0.27517429E+02,-0.19637804E+02, 0.41009953E+02,-0.52474489E+01,
             -0.27441065E+02, },
            {-0.12251460E+03, 0.41699898E+02, 0.27045620E+02,-0.92491447E+02,
             -0.72843239E+02, 0.32114170E+02, 0.11262262E+02, 0.40949612E+01,
              0.22106949E+02,-0.48072939E+01, 0.86024685E+01,-0.41298609E+01,
             -0.18827738E+01, },
            {-0.14486723E+03,-0.13257008E+03,-0.48221794E+02, 0.35999949E+01,
             -0.58804531E+02,-0.19716801E+02,-0.20199421E+02,-0.64824928E+02,
             -0.29064964E+02, 0.16903173E+02,-0.37667240E+02, 0.34804225E+01,
              0.26898163E+02, },
            { 0.30944574E+02,-0.11548166E+02,-0.11504376E+02, 0.38749798E+02,
              0.28696703E+02,-0.12365782E+02,-0.53707695E+01,-0.24697113E+01,
             -0.94244776E+01, 0.22112002E+01,-0.35927393E+01, 0.10858067E+01,
              0.97322190E+00, },
            { 0.68757813E+02, 0.55788937E+02, 0.54013424E+01,-0.52038832E+01,
              0.17593607E+02, 0.63505483E+01, 0.76194353E+01, 0.21402803E+02,
              0.10745119E+02,-0.52722220E+01, 0.12947476E+02,-0.74531823E+00,
             -0.95251179E+01, },
            {-0.11819898E+00, 0.16981829E+01, 0.17257081E+01,-0.46752773E-01,
              0.31282842E-01,-0.48106876E-02, 0.42170644E-01, 0.12124905E-01,
             -0.19874832E-01, 0.12629480E-01, 0.39389998E-01,-0.12933319E-01,
             -0.14372598E-01, },
            { 0.10223607E+00,-0.16555769E+01, 0.15552585E+01,-0.12264324E+00,
              0.65852158E-01,-0.47787797E-03,-0.17770162E-01, 0.24633724E-01,
             -0.14257873E-01,-0.19858574E-01, 0.25402587E-01,-0.80514261E-02,
              0.68274671E-02, },
            {-0.99085343E+00, 0.17199130E+01,-0.47460574E+00,-0.74100316E+00,
              0.10574152E+01, 0.11296365E+00, 0.82022703E+00,-0.18898286E+00,
              0.36352620E+00, 0.80850042E-01,-0.19919536E+00,-0.40538307E-01,
              0.16510320E+00, },
            { 0.11990190E+01, 0.51340390E-01, 0.75055975E+00,-0.12824067E+01,
              0.86559242E+00,-0.28388435E-02, 0.14062119E-01, 0.24629071E+00,
              0.20781083E+00,-0.11486780E+00, 0.91260616E-02, 0.73089264E-01,
             -0.38038802E-02, },
            { 0.51901360E+01, 0.11364080E+02,-0.14781400E+02, 0.34115415E+01,
              0.17582846E+01, 0.49295741E+00, 0.74259001E+00, 0.60993052E+00,
              0.43994176E+00, 0.49398413E+00,-0.82156599E-01, 0.19255295E+00,
              0.72789848E+00, },
            {-0.10661777E+02, 0.13685346E+02, 0.17330500E+02, 0.17598805E+01,
              0.15718225E+01,-0.35067117E+00,-0.24391625E+01, 0.87025601E+00,
              0.10643625E+01, 0.49751109E+00,-0.65840626E+00, 0.33091986E+00,
             -0.13313615E+01, },
            { 0.21538338E+02,-0.17283089E+02, 0.23691774E+02, 0.43560233E+01,
             -0.13055443E+02, 0.54472351E+00,-0.12533862E+02, 0.69577451E+01,
             -0.59899955E+01,-0.24915967E+01, 0.48114142E+01,-0.15907498E+00,
             -0.37714081E+01, },
            {-0.12313134E+02,-0.22617418E+02, 0.24490969E+01, 0.22901756E+02,
             -0.17509409E+02, 0.51498137E+01, 0.57923192E+00,-0.68807111E+01,
             -0.37194932E+01, 0.33164258E+01,-0.77732426E+00,-0.33252787E-01,
             -0.22301903E+01, },
            {-0.31336721E+02, 0.28626883E+01, 0.13524348E+03,-0.29654032E+02,
             -0.11670376E+02,-0.11883139E+02,-0.67173586E+01,-0.26616092E+01,
             -0.46543760E+01,-0.62572255E+01,-0.36758001E+01,-0.12831590E+00,
             -0.83014069E+01, },
            { 0.10399525E+03,-0.14304045E+03,-0.14625183E+02,-0.12641206E+02,
             -0.18179705E+02, 0.63950963E+01, 0.24394505E+02,-0.26621683E+01,
             -0.96777048E+01,-0.36132805E+01, 0.60474663E+01,-0.52178507E+01,
              0.14800138E+02, },
            {-0.13617873E+03, 0.98202393E+02,-0.52166687E+02, 0.82724130E+00,
              0.79597404E+02,-0.97800264E+01, 0.51829216E+02,-0.55843555E+02,
              0.30942001E+02, 0.15340666E+02,-0.32310665E+02, 0.47805400E+01,
              0.16678785E+02, },
            { 0.71018166E+02, 0.11442221E+03, 0.75697632E+01,-0.13351144E+03,
              0.80042564E+02,-0.19554031E+02,-0.23364079E+01, 0.45533413E+02,
              0.25092596E+02,-0.27485874E+02, 0.11551980E+02,-0.85284805E+01,
              0.22155190E+02, },
            { 0.72216705E+02,-0.18108649E+03,-0.35639267E+03, 0.91767868E+02,
              0.27649412E+02, 0.51445087E+02, 0.16393131E+02, 0.84795976E+00,
              0.11156778E+02, 0.21300432E+02, 0.15415002E+02,-0.20493467E+01,
              0.29226500E+02, },
            {-0.33076926E+03, 0.39775549E+03,-0.19134296E+03, 0.43584145E+02,
              0.63088699E+02,-0.25370869E+02,-0.73108444E+02, 0.75435942E+00,
              0.30592009E+02, 0.11364077E+02,-0.18345686E+02, 0.19485804E+02,
             -0.48085079E+02, },
            { 0.35883054E+03,-0.27099554E+03,-0.71656799E+02,-0.42929935E+02,
             -0.21730473E+03, 0.31643620E+02,-0.90837830E+02, 0.16265657E+03,
             -0.70991798E+02,-0.36282787E+02, 0.85148849E+02,-0.18301117E+02,
             -0.27730669E+02, },
            {-0.21271365E+03,-0.24034344E+03,-0.90665169E+02, 0.33108362E+03,
             -0.14937883E+03, 0.15747417E+02, 0.58647780E+01,-0.11471978E+03,
             -0.71360039E+02, 0.83697266E+02,-0.38742779E+02, 0.33933643E+02,
             -0.68442345E+02, },
            {-0.65536263E+02, 0.34354947E+03, 0.40484613E+03,-0.11488379E+03,
             -0.30950413E+02,-0.81028809E+02,-0.16655785E+02, 0.55178194E+01,
             -0.82527332E+01,-0.27838104E+02,-0.20135136E+02, 0.50235019E+01,
             -0.40293743E+02, },
            { 0.42467825E+03,-0.45499557E+03, 0.41682208E+03,-0.65872879E+02,
             -0.86527466E+02, 0.35884357E+02, 0.88252113E+02, 0.32965231E+01,
             -0.39156567E+02,-0.14947558E+02, 0.21975187E+02,-0.26912621E+02,
              0.60366070E+02, },
            {-0.41629010E+03, 0.33380893E+03, 0.27333212E+03, 0.82168999E+02,
              0.25991666E+03,-0.40061562E+02, 0.73255257E+02,-0.19695262E+03,
              0.74450417E+02, 0.36899387E+02,-0.95648888E+02, 0.25082708E+02,
              0.18237814E+02, },
            { 0.29320822E+03, 0.22077357E+03, 0.16044160E+03,-0.35729318E+03,
              0.12526577E+03, 0.13969834E+02,-0.11527355E+02, 0.12376456E+03,
              0.87678589E+02,-0.10517470E+03, 0.47443066E+02,-0.45187057E+02,
              0.84595184E+02, },
            { 0.20050932E+02,-0.17802786E+03,-0.16851566E+03, 0.49642143E+02,
              0.13681282E+02, 0.42664768E+02, 0.62386675E+01,-0.44329839E+01,
              0.11935354E+01, 0.12374495E+02, 0.84155293E+01,-0.32429960E+01,
              0.19101906E+02, },
            {-0.19180679E+03, 0.18370607E+03,-0.23252710E+03, 0.34046066E+02,
              0.40737797E+02,-0.16559662E+02,-0.37816505E+02,-0.23657224E+01,
              0.17279062E+02, 0.67948833E+01,-0.89706383E+01, 0.12578118E+02,
             -0.25909058E+02, },
            { 0.17511346E+03,-0.14709790E+03,-0.17955647E+03,-0.46925884E+02,
             -0.11192869E+03, 0.18169199E+02,-0.22695395E+02, 0.84683502E+02,
             -0.29096594E+02,-0.13504688E+02, 0.38522060E+02,-0.11487776E+02,
             -0.34478912E+01, },
            {-0.14182312E+03,-0.71934738E+02,-0.81841087E+02, 0.13787189E+03,
             -0.40718723E+02,-0.16116909E+02, 0.78342290E+01,-0.48472740E+02,
             -0.38284924E+02, 0.46472549E+02,-0.19452414E+02, 0.19880957E+02,
             -0.36534100E+02, },
            { 0.72543673E-01,-0.91503099E-01,-0.11593870E+00,-0.89834797E+00,
              0.61833626E+00, 0.47313958E-01, 0.14711141E-02, 0.10285539E-01,
             -0.22091963E-02,-0.20506838E-01, 0.34120739E-01,-0.22531549E-01,
             -0.78072742E-03, },
            {-0.96214540E-01,-0.54838866E-01,-0.14777187E+00,-0.64081067E+00,
             -0.69528621E+00,-0.30307973E-01,-0.33019181E-01, 0.77059716E-02,
             -0.97604236E-02, 0.16391926E-02, 0.17542893E-02, 0.45385696E-02,
             -0.20407379E-01, },
            { 0.10522875E+01, 0.18861838E+00,-0.25081074E-01,-0.14823078E+01,
              0.23610927E+00, 0.65894437E+00,-0.24199405E+00,-0.87921269E-01,
             -0.54795289E+00,-0.14422113E+00,-0.10366818E+00,-0.18718950E+00,
              0.21545326E-01, },
            { 0.42069238E+00,-0.22901510E+00,-0.91535009E-01, 0.25210852E-01,
             -0.13556504E+01, 0.37180072E+00, 0.70798683E+00, 0.22921542E-01,
              0.14435732E+00, 0.44424638E-01,-0.13737643E+00, 0.19060686E-01,
              0.10951717E+00, },
            { 0.14380064E+01,-0.18441191E+00, 0.46994653E+01,-0.21012435E+01,
              0.58279996E+01,-0.13754472E+01,-0.14721313E+01, 0.12456217E+00,
             -0.15493175E+00, 0.83261234E+00,-0.55525309E+00,-0.44475816E-01,
              0.58509374E+00, },
            { 0.19653126E+01,-0.75903958E+00, 0.46786284E+01,-0.82793903E+01,
             -0.22516363E+01, 0.15342182E+01,-0.47626281E+00, 0.35320488E+00,
              0.90423584E+00,-0.91187485E-01,-0.61380106E+00,-0.55390352E+00,
              0.45646802E+00, },
            {-0.69860544E+01,-0.81822312E+00,-0.18467587E+01, 0.76645608E+01,
              0.19130971E+01,-0.54477286E+01, 0.15927750E+01, 0.45357141E+00,
              0.58195391E+01, 0.14146249E+01, 0.78945208E+00, 0.14831167E+01,
             -0.24750523E+00, },
            {-0.21675032E+00, 0.16968607E+01,-0.38511195E+01,-0.57537832E+01,
              0.97383308E+01,-0.13728514E+01,-0.47436857E+01,-0.27548201E+01,
             -0.15055325E+01, 0.76347262E-01, 0.40537009E+00,-0.25368646E+00,
             -0.12546912E+01, },
            {-0.13612036E+02, 0.10055150E+02,-0.42518852E+02, 0.18920910E+02,
             -0.25449715E+02, 0.61947532E+01, 0.10574360E+02, 0.17470124E+01,
             -0.16524613E+00,-0.48321452E+01, 0.26428554E+01, 0.16906739E+01,
             -0.48421707E+01, },
            {-0.15996868E+02, 0.18508348E+02,-0.23283045E+02, 0.35497883E+02,
              0.12486041E+02,-0.13592185E+02, 0.23057771E+01,-0.33479509E+01,
             -0.91297178E+01, 0.20603864E+01, 0.30273182E+01, 0.47495813E+01,
             -0.24556153E+01, },
            { 0.14484992E+02,-0.12824783E+01, 0.20726955E+01,-0.13500336E+02,
             -0.13579147E+02, 0.12507603E+02,-0.44871597E+01,-0.13822854E+01,
             -0.16225626E+02,-0.32869406E+01,-0.13542657E+01,-0.36956699E+01,
              0.49570027E+00, },
            {-0.28008471E+01,-0.72233214E+01, 0.16578836E+02, 0.20905169E+02,
             -0.18679207E+02, 0.21797931E+01, 0.85508757E+01, 0.10466122E+02,
              0.37626474E+01,-0.83420086E+00,-0.52542210E+00, 0.81247646E+00,
              0.30173633E+01, },
            { 0.29996330E+02,-0.24423285E+02, 0.98381447E+02,-0.40458511E+02,
              0.31081457E+02,-0.10232323E+02,-0.22379601E+02,-0.84146442E+01,
              0.16865023E+01, 0.94843407E+01,-0.36858835E+01,-0.51620426E+01,
              0.11019040E+02, },
            { 0.38974052E+02,-0.47553543E+02, 0.32823261E+02,-0.45407307E+02,
             -0.18267160E+02, 0.32208038E+02,-0.38544316E+01, 0.74274793E+01,
              0.22328951E+02,-0.70309253E+01,-0.59846663E+01,-0.10427266E+02,
              0.56140108E+01, },
            {-0.11142906E+02, 0.50988245E+01,-0.87233227E+00, 0.10051833E+02,
              0.17175941E+02,-0.81438684E+01, 0.35580683E+01, 0.14108999E+01,
              0.12470690E+02, 0.20998702E+01, 0.89491493E+00, 0.26432147E+01,
             -0.34282866E+00, },
            { 0.32492125E+01, 0.95376787E+01,-0.15285566E+02,-0.19982004E+02,
              0.12006571E+02,-0.14975014E+01,-0.40201006E+01,-0.91834297E+01,
             -0.27835941E+01, 0.51126963E+00, 0.23976563E+00,-0.81174558E+00,
             -0.18622959E+01, },
            {-0.20657648E+02, 0.16039457E+02,-0.63869217E+02, 0.26869730E+02,
             -0.87734652E+01, 0.46892033E+01, 0.14171228E+02, 0.80630875E+01,
             -0.16122298E+01,-0.58862243E+01, 0.13981988E+01, 0.40675755E+01,
             -0.73429766E+01, },
            {-0.29788891E+02, 0.30969652E+02,-0.10377958E+02, 0.15621193E+02,
              0.83370752E+01,-0.22060728E+02, 0.21002872E+01,-0.47829390E+01,
             -0.15391683E+02, 0.58026686E+01, 0.41257248E+01, 0.70328178E+01,
             -0.43682208E+01, },
            {-0.13153026E+00, 0.52396119E-01,-0.78708120E-01,-0.31984996E-01,
              0.66879772E-01,-0.44668841E+00,-0.11997780E+01, 0.44919256E-01,
             -0.21612016E-01,-0.85290587E-02, 0.40041875E-01,-0.18274922E-01,
             -0.12108567E-01, },
            { 0.68851382E-01,-0.16586338E+00, 0.10567039E+00,-0.35489373E-01,
              0.10825994E+00, 0.11110218E+01,-0.48024797E+00,-0.32602344E-01,
              0.30053170E-01, 0.14435068E-01,-0.43142173E-01, 0.28743999E-01,
             -0.16354781E-03, },
            { 0.43814236E+00, 0.41187561E+00,-0.30113024E+00, 0.17926472E+00,
             -0.76448750E+00,-0.22111153E-01,-0.31990391E+00,-0.39657485E-01,
              0.26836851E+00, 0.31483594E-01, 0.17101410E+00,-0.48243791E-01,
              0.58657944E-01, },
            { 0.29777789E+00, 0.70694327E-01, 0.46847123E+00, 0.16531505E+00,
             -0.68414465E-01, 0.86738020E+00,-0.25612119E+00,-0.27293500E+00,
              0.91270700E-01, 0.98998006E-02,-0.54274175E-01, 0.55843040E-01,
              0.87239005E-01, },
            { 0.14875149E+01, 0.18058693E+00, 0.96503448E+00, 0.38504772E-01,
             -0.81080878E+00, 0.23628514E+01, 0.31373041E+01, 0.35168904E+00,
              0.17957918E+00, 0.99717319E-01, 0.12352602E+00,-0.28581694E-01,
              0.32931817E+00, },
            {-0.18351794E+01, 0.33663924E+01,-0.17299489E+01, 0.30873060E+00,
             -0.10293868E+01,-0.27952218E+01, 0.29568193E+01, 0.23304486E+00,
              0.48831758E+00,-0.16388200E-01, 0.26351678E+00,-0.23727889E+00,
              0.21022537E+00, },
            {-0.37114632E+00,-0.61679882E+00, 0.70735806E+00,-0.31027344E+00,
              0.14951218E+01,-0.35590678E+00,-0.21566993E-02, 0.32540190E+00,
             -0.41896856E+00,-0.11011457E+00,-0.22550862E+00,-0.13376597E+00,
             -0.11999568E+00, },
            {-0.62257117E+00,-0.84297568E-01,-0.13334845E+01,-0.10768685E+01,
              0.16588950E+00,-0.11461290E+01,-0.81029370E-01, 0.43433130E+00,
             -0.16242850E+00,-0.12601896E-01, 0.15622550E+00,-0.24858958E+00,
             -0.19455425E+00, },
            {-0.25591044E+01, 0.25164348E+00,-0.15365639E+01,-0.79542905E-01,
              0.11995534E+01,-0.45441856E+01,-0.45316739E+01,-0.10119162E+01,
             -0.80300719E-01,-0.52209429E-01,-0.35192704E+00, 0.85065439E-01,
             -0.56523353E+00, },
            { 0.30118430E+01,-0.47145052E+01, 0.29010694E+01,-0.36481506E+00,
              0.11654499E+01, 0.43329029E+01,-0.52999201E+01,-0.24453495E+00,
             -0.10941246E+01, 0.39213534E-01,-0.34023368E+00, 0.27661625E+00,
             -0.51622367E+00, },
            {-0.63605845E-01, 0.50896596E-01, 0.11700895E-01, 0.43969844E-01,
              0.13695911E+00,-0.28410681E-01,-0.93208402E-02, 0.34158775E+00,
             -0.23341532E+00, 0.18170308E-01, 0.19188978E-02,-0.67197052E-02,
              0.79181036E-02, },
            { 0.70990942E-01, 0.41841712E-01, 0.20218294E-02, 0.81035674E-01,
             -0.52733000E-01,-0.91757327E-02,-0.15658233E-01, 0.20307919E+00,
              0.37615097E+00,-0.61855074E-02, 0.74328687E-02, 0.91638947E-02,
             -0.92550581E-02, },
            { 0.22771759E+00,-0.20004077E-01,-0.19363023E+00,-0.27965687E-02,
              0.73775589E-02,-0.16548780E+00, 0.24623014E-01, 0.59029967E-01,
             -0.32041299E+00,-0.62310666E-01,-0.22864670E-01, 0.68012923E-02,
              0.41548245E-01, },
            { 0.16248761E+00,-0.53943157E-01,-0.10124869E+00, 0.88020787E-01,
              0.84402587E-03,-0.98183513E-01,-0.16950405E+00, 0.16585672E+00,
              0.22753301E+00,-0.47868274E-01,-0.62641405E-01, 0.39978638E-01,
             -0.17434882E-01, },
            {-0.12619768E+00, 0.46246566E-01,-0.21488875E-01, 0.72626024E-01,
              0.24535481E-01, 0.37770525E-01, 0.66443910E-02,-0.34785960E-01,
              0.48899220E-03, 0.11222392E+00, 0.31389341E+00,-0.94639612E-02,
              0.29937252E-02, },
            { 0.18342680E+00,-0.42065293E-01, 0.56304694E-02,-0.50527258E-02,
             -0.67958511E-01,-0.14655957E-01, 0.18668728E-01, 0.40048421E-02,
             -0.52483652E-01,-0.29388085E+00, 0.83209299E-01, 0.99197831E-02,
              0.39488561E-02, },
            { 0.73701084E-01,-0.11414203E+00, 0.50474752E-01, 0.52794013E-02,
              0.51152506E-02, 0.31500902E-01, 0.14944745E-01, 0.27513613E-02,
             -0.20441337E-01,-0.17187309E-01, 0.23239015E-01,-0.15333304E+00,
              0.10820904E-01, },
            { 0.10246071E+00,-0.10813680E-01,-0.28045259E-01, 0.85988389E-02,
             -0.66896558E-01, 0.20317493E-01,-0.25311206E-01, 0.21137524E-01,
             -0.59515103E-02,-0.73466594E-02,-0.46075849E-03,-0.14882740E-01,
             -0.15001261E+00, },
            {-0.62333126E-01,-0.11848945E+00,-0.74348420E-01,-0.13404631E-02,
             -0.22817828E-01,-0.10830126E-01,-0.48877955E-02, 0.12237122E-01,
             -0.30541401E-01, 0.57756705E-02,-0.10295756E-01, 0.12080544E-01,
             -0.14406471E-01, },
            { 0.41248910E-01, 0.44591889E-01, 0.36592875E-01, 0.55447145E-03,
              0.19915281E-01,-0.63508269E-02,-0.47519132E-01, 0.28801892E-01,
              0.77612209E-02,-0.59790271E-02, 0.73863971E-02, 0.16617060E-01,
             -0.22494165E-01, },
            {-0.71962871E-01,-0.41219030E-01,-0.70982397E-01, 0.12210120E-01,
             -0.87522060E-01,-0.47411237E-01, 0.57425763E-01, 0.45690551E-01,
              0.26248202E-01, 0.28013919E-01,-0.16828964E-02, 0.15796809E-01,
              0.20284114E-01, },
            {-0.47389165E-01, 0.37087288E-01,-0.58108825E-01,-0.54313848E-02,
              0.10483098E+00, 0.13532057E-01,-0.20340554E-01,-0.50312970E-02,
             -0.42387201E-02,-0.89632040E-02,-0.32773022E-01, 0.21721894E-01,
              0.10820904E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir20Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30495803E+01, 0.35076546E-02, 0.76724417E-02, 0.87806750E-02,
              0.47458550E-02, 0.25473754E-02, 0.29852835E-02,-0.12277642E-03,
             -0.32025625E-03, },
            { 0.10599100E+00,-0.21536589E-01, 0.46008382E-01, 0.78201421E-01,
              0.22680618E-01,-0.46905663E-01,-0.96271224E-02, 0.12562236E-01,
              0.24986399E-01, },
            { 0.10458715E+01,-0.25089502E-01,-0.17962436E+00,-0.17404799E+00,
             -0.22410804E-01, 0.27813714E-01,-0.18103598E-01,-0.58276366E-01,
              0.26434981E-02, },
            { 0.17178220E+00,-0.15449823E+00,-0.10333893E+00,-0.18914133E+00,
             -0.13402216E+00, 0.16679610E+00, 0.46815984E-01,-0.91324374E-02,
             -0.10088239E+00, },
            {-0.18385901E+01, 0.77625453E-01, 0.55216676E+00, 0.35753915E+00,
             -0.87231752E-02,-0.51372502E-01, 0.34072302E-01, 0.17208675E+00,
             -0.13502965E-01, },
            {-0.32891473E+00, 0.15413617E+00, 0.46012118E-01, 0.10862464E+00,
              0.12537768E+00,-0.12952726E+00,-0.47462337E-01,-0.35870597E-02,
              0.86061329E-01, },
            { 0.89103073E+00,-0.37754770E-01,-0.38592705E+00,-0.17952418E+00,
              0.40533390E-01, 0.47559445E-02,-0.18648680E-01,-0.11294576E+00,
              0.12865890E-01, },
            {-0.69007248E-03,-0.18691148E+00,-0.33661303E+00, 0.48614144E-02,
              0.11631002E-01,-0.95324073E-03, 0.85280649E-02, 0.36868376E-02,
              0.60911463E-02, },
            { 0.32471757E-01, 0.35279295E+00,-0.18637425E+00,-0.75623016E-02,
             -0.97096683E-02, 0.31888080E-02,-0.51301736E-02,-0.62356037E-02,
              0.11819322E-01, },
            {-0.84096253E-01, 0.25142324E+00,-0.35519764E-01, 0.26370287E-01,
              0.97039536E-01,-0.31170715E-01,-0.37270427E-01, 0.46906836E-01,
              0.65398109E-02, },
            { 0.45785110E-01, 0.13975869E+00, 0.26513165E+00,-0.29776646E-01,
             -0.25009252E-01, 0.36971882E-01,-0.30679654E-01, 0.35375282E-01,
              0.11662492E+00, },
            { 0.15254179E+00, 0.67530102E+00, 0.15002052E+01, 0.58707792E-01,
             -0.19350794E+00, 0.69749169E-01,-0.31615756E-01,-0.64090937E-01,
             -0.20062795E+00, },
            {-0.83293790E+00,-0.15851485E+01, 0.79647267E+00, 0.14819029E+00,
              0.13229819E+00,-0.22473099E-01, 0.55494748E-01, 0.13741881E+00,
             -0.30662367E+00, },
            { 0.15882361E+01,-0.10474651E+01, 0.56707895E+00, 0.21353865E+00,
             -0.70312959E+00, 0.67766070E-01, 0.13480249E+00,-0.26861522E+00,
              0.10558596E+00, },
            { 0.42528209E+00,-0.81927925E+00,-0.13487948E+01,-0.28105810E+00,
              0.56431311E+00,-0.22810178E+00, 0.15924788E+00,-0.40170720E+00,
             -0.70476109E+00, },
            {-0.50016040E+00,-0.57117230E+00,-0.95467204E+00,-0.63309115E+00,
              0.40046284E+00,-0.30428410E+00, 0.42816602E-01, 0.86010098E-01,
              0.67586058E+00, },
            { 0.25389247E+01, 0.97007108E+00,-0.13116745E+01,-0.12303472E+00,
             -0.53631330E+00, 0.64055860E-01,-0.72015218E-01,-0.42832959E+00,
              0.86087793E+00, },
            {-0.44755645E+01, 0.10564411E+01,-0.66151208E+00,-0.45737475E+00,
              0.15295455E+01, 0.55482633E-01,-0.10898267E+00, 0.44265333E+00,
             -0.39993185E+00, },
            {-0.14845783E+01, 0.12734603E+01, 0.23328686E+01, 0.94133908E+00,
             -0.13304437E+01, 0.53523225E+00,-0.28171411E+00, 0.10899057E+01,
              0.11757891E+01, },
            { 0.43657407E+00, 0.14657229E+00, 0.85669927E-01, 0.69275594E+00,
             -0.22207183E+00, 0.28957173E+00, 0.16122398E-02, 0.24003390E-01,
             -0.55424219E+00, },
            {-0.20080192E+01, 0.30063149E+00, 0.88270497E+00,-0.69684193E-01,
              0.51002681E+00,-0.63769467E-01, 0.87251076E-02, 0.35308439E+00,
             -0.61427212E+00, },
            { 0.32225552E+01,-0.25058147E-02, 0.50169885E+00, 0.13464716E+00,
             -0.10105824E+01,-0.71476690E-01, 0.32011800E-01,-0.21982677E+00,
              0.30498490E+00, },
            { 0.11992681E+01,-0.96706653E+00,-0.11388683E+01,-0.73485720E+00,
              0.80431402E+00,-0.39570120E+00, 0.16278829E+00,-0.83236420E+00,
             -0.58864450E+00, },
            {-0.14675558E-01,-0.23623409E-01,-0.17950738E-01, 0.66034064E-01,
             -0.11922826E+00,-0.54179137E-02, 0.14375948E-01,-0.60841944E-02,
             -0.26708576E-02, },
            { 0.43763379E-02,-0.10887975E-01,-0.71096844E-02, 0.11585610E+00,
              0.90701804E-01, 0.55126945E-03, 0.56899362E-02, 0.31451138E-02,
             -0.52986522E-02, },
            { 0.10125032E+00, 0.17959675E+00, 0.10625629E+00,-0.50184567E-01,
              0.10304492E+00, 0.24058525E-01,-0.32390207E-01, 0.75553027E-02,
             -0.20764861E-01, },
            {-0.11460812E+00, 0.79707839E-01,-0.29399197E-02,-0.67522340E-01,
              0.13258697E+00, 0.49736477E-01,-0.78869835E-02,-0.39157858E-02,
              0.24828205E-01, },
            { 0.10480853E+00, 0.19343019E+00, 0.17289962E+00,-0.31678677E+00,
             -0.17360756E+00, 0.37843499E-01, 0.62656075E-01,-0.18113919E-02,
             -0.78926899E-01, },
            {-0.26774183E+00,-0.22751540E-01,-0.13300699E+00, 0.24421448E+00,
             -0.40151697E+00,-0.44680607E-01, 0.90509988E-01,-0.45797311E-01,
             -0.19253580E-01, },
            {-0.53374827E+00,-0.90245152E+00,-0.14048404E+00, 0.31762448E+00,
             -0.75603151E+00,-0.17422198E+00, 0.16830654E+00,-0.67688487E-01,
              0.83993100E-01, },
            { 0.92552233E+00,-0.17285006E-01, 0.26109925E+00, 0.74910748E+00,
             -0.50195232E-01,-0.18097821E+00,-0.59655149E-01, 0.36261849E-01,
             -0.16167022E+00, },
            {-0.18626483E+00,-0.31044501E+00,-0.34045997E+00, 0.30551198E+00,
              0.83555371E-01,-0.10926255E+00,-0.14400959E+00, 0.99276425E-02,
              0.17333107E+00, },
            { 0.29973921E+00,-0.38495310E-01, 0.25320417E+00,-0.16226833E+00,
              0.34776387E+00, 0.68651758E-01,-0.18148483E+00, 0.94451770E-01,
              0.44157453E-01, },
            { 0.68308866E+00, 0.12142456E+01, 0.23941616E-01,-0.49178913E+00,
              0.10317271E+01, 0.30297923E+00,-0.16258200E+00, 0.87277547E-01,
             -0.87765060E-01, },
            {-0.12280098E+01,-0.17898367E+00,-0.24613486E+00,-0.11506540E+01,
             -0.37660384E+00, 0.14181741E+00, 0.15465291E+00,-0.45992363E-01,
              0.17110234E+00, },
            {-0.21097982E-01, 0.60008105E-03,-0.18510208E-02,-0.62552849E-02,
             -0.93422104E-02, 0.60288724E-01,-0.40334366E-01,-0.43196464E-02,
             -0.61084577E-02, },
            {-0.83072307E-02,-0.34225453E-01,-0.11824392E-01,-0.13751618E-02,
             -0.17261762E-01, 0.43971878E-01, 0.76659046E-01, 0.33106832E-02,
              0.45388611E-02, },
            { 0.75911164E-01, 0.46962660E-01, 0.21635788E-01,-0.77638470E-01,
              0.22221388E-01,-0.79871550E-01,-0.12672771E-01, 0.25498134E-02,
              0.12646123E-02, },
            { 0.51471148E-01, 0.33606584E-02,-0.13351038E-01,-0.24537463E-01,
             -0.35670411E-01, 0.14631917E-01,-0.90372808E-01, 0.99615008E-03,
              0.22813002E-01, },
            { 0.21750353E-01,-0.83270617E-01,-0.18755151E-02, 0.53586468E-01,
              0.93389477E-03,-0.19740278E+00,-0.11719642E+00, 0.33343490E-01,
              0.22319425E-01, },
            {-0.29066566E-02, 0.41417848E-01,-0.10801823E-01, 0.28476706E-01,
              0.10270041E-01, 0.15855379E+00,-0.24726029E+00, 0.14735273E-01,
             -0.27862186E-01, },
            {-0.23555486E-01,-0.40901415E-02, 0.60059442E-02, 0.97556552E-02,
              0.91812024E-02,-0.14488046E-02, 0.16592178E-02, 0.36357131E-01,
              0.13751269E-01, },
            {-0.29764045E-01, 0.92727989E-02,-0.86509585E-02,-0.18609468E-01,
             -0.27944657E-03, 0.31447997E-02, 0.75850124E-02,-0.18039657E-01,
              0.38740944E-01, },
            { 0.79681963E-01, 0.48902291E-02, 0.57677603E-02,-0.33417843E-01,
             -0.68668877E-02, 0.22215862E-02,-0.40455712E-02, 0.36775917E-02,
             -0.22278046E-01, },
            { 0.10155325E-01,-0.54381907E-01, 0.22469945E-01, 0.53363744E-01,
             -0.10555824E-01,-0.46219719E-02,-0.16068283E-02, 0.35136204E-01,
              0.11071405E-01, },
            { 0.18289034E-02,-0.75299381E-02,-0.14482981E-03, 0.35247686E-02,
              0.42061866E-02, 0.51727840E-02, 0.17811510E-02,-0.10641967E-01,
              0.48779794E-02, },
            { 0.15797319E-01, 0.18416291E-01, 0.16457362E-05,-0.16946379E-02,
             -0.99487100E-02, 0.13055210E-01,-0.14801215E-02,-0.56088273E-02,
             -0.40782345E-02, },
            {-0.35834808E-02, 0.71111410E-02, 0.12105495E-01,-0.45448786E-03,
              0.62631634E-02, 0.36953585E-02,-0.28422426E-02,-0.19898398E-02,
              0.48590722E-02, },
            { 0.11628112E-01, 0.28534012E-03,-0.12218224E-01, 0.24151376E-02,
             -0.16151644E-03, 0.75605242E-02, 0.45826556E-02,-0.46821288E-02,
              0.12317485E-02, },
         },
         {
            { 0.26966269E+01,-0.14578769E-05, 0.50493279E-02, 0.65831305E-02,
              0.43787737E-02, 0.39015978E-02, 0.89700258E-03, 0.85402245E-03,
             -0.68476156E-03, },
            { 0.11744526E+00, 0.12448850E-01, 0.51003762E-01, 0.40138353E-01,
              0.26125014E-01,-0.16533557E-01,-0.29090958E-01,-0.17052057E-02,
              0.80826618E-02, },
            { 0.11973118E+01,-0.65078013E-01,-0.99028312E-01,-0.81597298E-01,
             -0.32280181E-01,-0.22603732E-01,-0.10447549E-01,-0.41825704E-01,
              0.23424169E-02, },
            { 0.84804960E-01,-0.19440410E+00,-0.12508395E+00,-0.10045742E+00,
             -0.13333432E+00, 0.45456056E-01, 0.10219767E+00, 0.23704873E-01,
             -0.32805737E-01, },
            {-0.18993030E+01, 0.22647896E+00, 0.26471585E+00, 0.15323690E+00,
              0.31060221E-01, 0.91907240E-01, 0.40878788E-01, 0.95126577E-01,
              0.45649102E-02, },
            {-0.19986725E+00, 0.13906373E+00, 0.85632414E-01, 0.60318705E-01,
              0.11465530E+00,-0.33805624E-01,-0.83187349E-01,-0.24020078E-01,
              0.27057443E-01, },
            { 0.82170755E+00,-0.12471489E+00,-0.17876776E+00,-0.71481489E-01,
              0.27813890E-02,-0.87954432E-01,-0.34027167E-01,-0.52087925E-01,
             -0.90684919E-02, },
            { 0.43652612E-02,-0.33954412E+00,-0.28059766E+00,-0.48793945E-03,
              0.57276776E-02,-0.32016952E-02, 0.50360449E-02, 0.29207736E-02,
              0.43764645E-02, },
            { 0.85188961E-02, 0.30349514E+00,-0.33877447E+00,-0.28169504E-02,
             -0.11583043E-01, 0.14881253E-02,-0.76294229E-02,-0.18964122E-02,
              0.11345063E-01, },
            {-0.12247941E-01, 0.17714764E+00,-0.34096591E-01,-0.22695521E-01,
              0.67024931E-01,-0.39165456E-01,-0.30592242E-01, 0.23541650E-01,
              0.10291560E-01, },
            { 0.50195772E-01, 0.15112165E+00, 0.20722289E+00, 0.34769014E-01,
             -0.29959407E-01, 0.41639198E-01,-0.29532194E-01, 0.25578812E-01,
              0.78463838E-01, },
            {-0.60606830E-01, 0.10363140E+01, 0.86959022E+00, 0.15487760E+00,
             -0.99790499E-01, 0.41185062E-01, 0.15034962E-01,-0.47623415E-01,
             -0.14509769E+00, },
            {-0.57434636E+00,-0.97933829E+00, 0.12581741E+01, 0.25207497E-01,
              0.18004075E+00,-0.21284509E-01, 0.77311493E-01, 0.64989790E-01,
             -0.20855135E+00, },
            { 0.47468442E+00,-0.10885421E+01, 0.38821119E+00, 0.23283729E+00,
             -0.50951570E+00, 0.23506916E+00, 0.35243019E-01,-0.17403507E+00,
             -0.23718674E-01, },
            { 0.26270482E+00,-0.70497757E+00,-0.13304210E+01,-0.53473514E+00,
              0.28150737E+00,-0.26304531E+00, 0.18870257E+00,-0.23507071E+00,
             -0.43555272E+00, },
            { 0.11272904E+00,-0.99514472E+00,-0.34449834E+00,-0.71797216E+00,
              0.18339764E+00,-0.18274480E+00,-0.63537367E-01, 0.10146949E+00,
              0.48926970E+00, },
            { 0.20371649E+01, 0.22109611E+00,-0.20750749E+01, 0.11111619E+00,
             -0.56142098E+00, 0.42209644E-01,-0.19956113E+00,-0.19299477E+00,
              0.54810613E+00, },
            {-0.14952495E+01, 0.17684418E+01,-0.49218816E+00,-0.28709847E+00,
              0.11091636E+01,-0.42815658E+00, 0.12084533E+00, 0.37503195E+00,
             -0.81880488E-01, },
            {-0.14637735E+01, 0.10478177E+01, 0.26719489E+01, 0.13296674E+01,
             -0.46114388E+00, 0.65784061E+00,-0.44182307E+00, 0.56460422E+00,
              0.67381424E+00, },
            {-0.28567558E-01, 0.46754566E+00, 0.17383933E+00, 0.67441458E+00,
             -0.10554819E+00, 0.19837098E+00, 0.47730759E-01,-0.50233286E-01,
             -0.40128100E+00, },
            {-0.17196779E+01, 0.35429364E+00, 0.14362482E+01,-0.16386373E+00,
              0.45979795E+00,-0.21841004E-01, 0.13400900E+00, 0.14516471E+00,
             -0.39125723E+00, },
            { 0.10732856E+01,-0.75785559E+00, 0.52866328E+00, 0.71639992E-01,
             -0.73546070E+00, 0.24522778E+00,-0.13535358E+00,-0.22979885E+00,
              0.11654176E+00, },
            { 0.15114303E+01,-0.77876258E+00,-0.15880039E+01,-0.89949375E+00,
              0.22964896E+00,-0.49463138E+00, 0.30698773E+00,-0.40895736E+00,
             -0.29977861E+00, },
            {-0.79477411E-02,-0.18472305E-01,-0.12967197E-01, 0.94843686E-01,
             -0.88151120E-01,-0.51070806E-02, 0.13504693E-01,-0.41395505E-02,
              0.24751091E-03, },
            {-0.44946186E-02,-0.16740741E-01,-0.41124807E-02, 0.76202437E-01,
              0.10344878E+00,-0.22537799E-02, 0.43183658E-02, 0.64705717E-04,
             -0.59640518E-03, },
            { 0.95618159E-01, 0.92954643E-01, 0.61767250E-01,-0.79024971E-01,
              0.11346289E+00, 0.15616628E-01,-0.56508440E-02, 0.42031724E-02,
              0.24400151E-02, },
            { 0.96911388E-02, 0.29432042E-01,-0.46471480E-01,-0.12849204E+00,
              0.22337712E-01, 0.15172200E-01, 0.14756573E-01,-0.10275276E-01,
              0.32632593E-01, },
            { 0.90240715E-02, 0.98447271E-01, 0.13910589E+00,-0.49684674E+00,
              0.73422439E-01, 0.23828603E-02, 0.33573525E-02,-0.72285314E-02,
             -0.70686325E-01, },
            {-0.16957556E+00, 0.50366666E-01,-0.11485599E+00,-0.34002479E-01,
             -0.50488788E+00,-0.42284265E-01, 0.40896658E-01,-0.22395281E-02,
             -0.41483179E-01, },
            {-0.39100450E+00,-0.52364653E+00, 0.81161167E-02, 0.23826459E+00,
             -0.69935983E+00,-0.49828898E-01, 0.58817074E-01,-0.30699246E-01,
              0.18457994E-02, },
            { 0.18260169E+00, 0.12305915E-01, 0.35234398E+00, 0.78882325E+00,
              0.17531563E+00, 0.12401869E-01,-0.94772555E-01, 0.70517570E-01,
             -0.18616913E+00, },
            {-0.20249626E-01,-0.18173020E+00,-0.28802195E+00, 0.27763471E+00,
             -0.33017445E+00,-0.40717527E-01,-0.36766961E-01, 0.36674999E-01,
              0.13604657E+00, },
            { 0.24803011E+00,-0.15678068E+00, 0.20198597E+00, 0.30329382E+00,
              0.25746182E+00, 0.61767373E-01,-0.93926691E-01, 0.18352006E-01,
              0.62546603E-01, },
            { 0.35427305E+00, 0.75491345E+00,-0.11663892E+00,-0.31980485E+00,
              0.90525955E+00, 0.89059651E-01,-0.78885183E-01, 0.31197926E-01,
             -0.51024295E-02, },
            {-0.33548266E+00, 0.36326089E-03,-0.39684558E+00,-0.10250442E+01,
             -0.45261109E+00,-0.34287624E-01, 0.12944759E+00,-0.95721230E-01,
              0.19820654E+00, },
            {-0.17688449E-01, 0.15564424E-02, 0.39224289E-02,-0.38294392E-02,
             -0.75140465E-02, 0.59528705E-01,-0.48654277E-01,-0.22800365E-02,
             -0.31075822E-02, },
            {-0.96650571E-02,-0.21461820E-01, 0.77744690E-03, 0.16463975E-02,
             -0.11097819E-01, 0.51459324E-01, 0.68256907E-01, 0.39644483E-02,
             -0.15337553E-02, },
            { 0.34351684E-01, 0.43128196E-01, 0.19956021E-01,-0.36301594E-01,
              0.99364547E-02,-0.61444476E-01,-0.24400521E-01, 0.44092600E-03,
              0.13028901E-02, },
            { 0.22622351E-01, 0.11014455E-01,-0.81500681E-02,-0.17104017E-01,
             -0.29482944E-01, 0.26210617E-01,-0.75117700E-01,-0.28545759E-03,
              0.11840697E-01, },
            { 0.17646819E-01,-0.77693120E-01, 0.38775124E-02, 0.28927619E-01,
              0.27532193E-02,-0.16366033E+00,-0.96197531E-01, 0.21273768E-01,
              0.86615868E-02, },
            { 0.72159092E-02, 0.12788316E-01,-0.31918742E-01, 0.10852541E-01,
              0.17516689E-01, 0.12468155E+00,-0.17470151E+00,-0.41061863E-02,
             -0.12752778E-02, },
            {-0.18442726E-01,-0.91943759E-02, 0.67788684E-02, 0.14435091E-03,
              0.78480067E-02, 0.98523422E-04, 0.22894247E-02, 0.25241941E-01,
              0.15407712E-01, },
            {-0.16817285E-01, 0.31028625E-02,-0.10441542E-01,-0.79444088E-02,
             -0.17821643E-02, 0.52367494E-03, 0.82076322E-02,-0.18905919E-01,
              0.29213514E-01, },
            { 0.44744335E-01,-0.13247139E-02,-0.37604002E-02,-0.16222805E-01,
             -0.89065684E-02,-0.51037166E-02,-0.28314355E-02, 0.18366748E-01,
             -0.27277077E-01, },
            { 0.19684851E-01,-0.35009183E-01, 0.23218991E-01, 0.28978951E-01,
             -0.10054998E-01,-0.95502073E-02,-0.97779622E-02, 0.31590566E-01,
              0.18811375E-01, },
            { 0.27152249E-02,-0.10953413E-01, 0.43081818E-02, 0.26744686E-02,
              0.12518663E-02, 0.79071680E-02,-0.18357651E-02,-0.90767127E-02,
              0.21241107E-02, },
            { 0.64085410E-02, 0.12395338E-01,-0.43195123E-02, 0.27304080E-02,
             -0.55480246E-02, 0.65472890E-02, 0.11660623E-02,-0.26218852E-02,
             -0.10962778E-02, },
            { 0.18541915E-03, 0.11536265E-01, 0.10362318E-01, 0.24194026E-02,
             -0.11380968E-02, 0.27860261E-02, 0.36632185E-03,-0.13429242E-02,
              0.37920703E-02, },
            { 0.48007774E-02, 0.10027210E-02,-0.10543189E-01, 0.72369934E-02,
             -0.16516771E-02, 0.35164719E-02, 0.81353862E-03,-0.30095526E-02,
              0.13707984E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir21F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.63694148E+01,-0.34792226E-01,-0.74101150E-01, 0.35132650E-01,
              0.24849806E-01, 0.11001032E-01,-0.19688884E-01, 0.26043868E-01,
              0.10386006E-01,-0.32306511E-02, 0.13962947E-01,-0.10103378E-01,
              0.18723549E-01, },
            { 0.21955316E+01,-0.39681266E-02, 0.18197685E+00, 0.11352257E+00,
             -0.31861094E+00,-0.15868060E-02, 0.39689042E-01, 0.17209752E-01,
              0.19868961E+00,-0.80792844E-01,-0.11767738E+00, 0.12445180E+00,
             -0.55129856E-01, },
            { 0.97781305E+01, 0.63362435E-01,-0.53614479E+00,-0.49567866E+00,
              0.11741468E+01,-0.63698125E+00, 0.52839708E+00,-0.11437672E+00,
             -0.15261206E+00,-0.13389719E+00,-0.36991826E+00, 0.11427352E+00,
             -0.65866762E+00, },
            {-0.32809158E+02,-0.52905566E+00, 0.56250453E+00, 0.25619733E+00,
              0.76901011E+01,-0.76617765E+00, 0.13428245E+00,-0.12091900E+01,
             -0.15285416E+01, 0.67826205E+00, 0.23068154E+01,-0.12790626E+01,
             -0.21327354E+00, },
            {-0.58018158E+02, 0.77569704E+01, 0.58701005E+01,-0.37597203E+01,
             -0.11486585E+02, 0.50069208E+01,-0.19475280E+01,-0.26410751E+01,
              0.11136000E+01, 0.12512259E+00, 0.17224880E+01,-0.64933658E+00,
              0.42313733E+01, },
            { 0.81844147E+02,-0.18177500E+02,-0.10697339E+02, 0.40396552E+01,
             -0.31275728E+02, 0.73205338E+01,-0.39177847E+01, 0.58170772E+01,
              0.57865572E+01,-0.22364533E+01,-0.11624065E+02, 0.43845687E+01,
              0.30469108E+01, },
            { 0.10854137E+03,-0.30877756E+02,-0.10837447E+02, 0.22226160E+02,
              0.34705902E+02,-0.13085101E+02, 0.15620646E+01, 0.11406905E+02,
             -0.31094887E+01, 0.85883933E+00,-0.26042700E+01, 0.11922537E+01,
             -0.10150283E+02, },
            {-0.44580307E+02, 0.79383812E+02, 0.24413515E+02,-0.21338619E+02,
              0.50065228E+02,-0.21513105E+02, 0.15347665E+02,-0.10429506E+02,
             -0.11754140E+02, 0.34370358E+01, 0.24936632E+02,-0.73416333E+01,
             -0.82282839E+01, },
            {-0.10028139E+03, 0.37588978E+02, 0.43946161E+01,-0.33876656E+02,
             -0.40892410E+02, 0.13890780E+02, 0.78339022E+00,-0.14895038E+02,
              0.35314486E+01,-0.15978297E+01, 0.11808842E+01,-0.71185213E+00,
              0.10538705E+02, },
            {-0.50124062E+02,-0.10354129E+03,-0.18342447E+02, 0.30102610E+02,
             -0.34501987E+02, 0.25331213E+02,-0.20694256E+02, 0.83078852E+01,
              0.11584894E+02,-0.26216393E+01,-0.24104069E+02, 0.61813402E+01,
              0.85840073E+01, },
            { 0.37392719E+02,-0.14476900E+02, 0.99496180E+00, 0.15965063E+02,
              0.16501513E+02,-0.52093725E+01,-0.93629724E+00, 0.62317777E+01,
             -0.14138385E+01, 0.73162246E+00, 0.71325324E-01, 0.48246250E-01,
             -0.39943352E+01, },
            { 0.43037422E+02, 0.42916580E+02, 0.38246312E+01,-0.13166988E+02,
              0.83462524E+01,-0.10465692E+02, 0.90876627E+01,-0.24948664E+01,
             -0.43121109E+01, 0.83627832E+00, 0.86244040E+01,-0.20590615E+01,
             -0.31272595E+01, },
            {-0.40967345E-01, 0.18442563E+01, 0.17633563E+01,-0.41515935E-01,
              0.14598442E-01, 0.52640415E-02, 0.33678550E-01, 0.18317280E-01,
             -0.30224023E-01,-0.20597542E-01, 0.24734342E-01,-0.10589097E-01,
              0.24328345E-01, },
            {-0.22360490E+00,-0.19206479E+01, 0.18999895E+01,-0.99210821E-01,
              0.13044933E-01, 0.23915380E-01, 0.10539087E-01, 0.34584550E-02,
              0.12267416E-01, 0.57058195E-02, 0.42165192E-02, 0.10772912E-01,
             -0.26614875E-01, },
            {-0.84906530E+00, 0.15865498E+01, 0.22927699E+01,-0.33673322E+00,
              0.19166987E+00, 0.64818418E+00, 0.42091948E+00,-0.16618843E+00,
              0.78511827E-01,-0.64768136E-01,-0.39221168E-01,-0.89224152E-01,
              0.13744977E+00, },
            {-0.14338664E+01,-0.13094720E+01, 0.17051713E+01,-0.10853502E+01,
              0.19608435E+00, 0.16197301E+00, 0.21957265E+00,-0.31434265E+00,
              0.37495288E+00, 0.34630176E-01,-0.90445802E-01, 0.12040331E+00,
              0.17347548E-01, },
            { 0.52947850E+01, 0.10844605E+02, 0.37868409E+01,-0.10532480E+01,
              0.36165705E+01,-0.88409197E+00, 0.49687558E+00,-0.61486661E+00,
              0.60421681E+00, 0.21092279E+00,-0.32828119E+00,-0.47752506E+00,
             -0.94027740E+00, },
            {-0.25222619E+01,-0.46317253E+01, 0.80145826E+01, 0.24095893E+01,
              0.39263377E+01, 0.25864679E+00,-0.13400965E+01, 0.52035743E+00,
              0.38246587E+00,-0.92407811E+00,-0.74844398E-01,-0.70236456E+00,
              0.10894667E+00, },
            { 0.22070520E+02,-0.17723101E+02,-0.98362646E+01, 0.23332732E+01,
              0.64828506E+01,-0.10151888E+02,-0.41051803E+01, 0.79824430E+00,
             -0.15293568E+01,-0.26827996E+01, 0.52764028E+00, 0.14580549E+01,
             -0.35977669E+01, },
            { 0.25647627E+02,-0.80371361E+01,-0.18844767E+02, 0.13793572E+02,
              0.82348639E+00, 0.21338460E+01,-0.32136986E+01, 0.57288928E+01,
             -0.49189863E+01, 0.13956453E+01, 0.25248592E+01,-0.16098522E+00,
             -0.32238259E+01, },
            {-0.37831741E+02,-0.82785957E+02, 0.40492912E+02, 0.49759912E+01,
             -0.32693497E+02, 0.38788466E+01,-0.55101314E+01, 0.52948971E+01,
             -0.33069551E+01,-0.27049429E+01, 0.15614638E+01, 0.44462080E+01,
              0.58480530E+01, },
            { 0.43074551E+02,-0.35585445E+02,-0.63008469E+02,-0.10148061E+02,
             -0.38213272E+02,-0.12484533E+01, 0.11518609E+02,-0.16986067E+01,
             -0.44723053E+01, 0.66183820E+01, 0.27471104E+00, 0.51703401E+01,
              0.16862423E+01, },
            {-0.12163273E+03, 0.35768089E+02, 0.79624352E+02,-0.42255273E+00,
             -0.29189657E+02, 0.46464653E+02, 0.11797403E+02,-0.35026555E+01,
              0.71749539E+01, 0.28277515E+02, 0.89573674E-01,-0.11701196E+02,
              0.20683783E+02, },
            {-0.14859586E+03, 0.73930824E+02, 0.46461655E+02,-0.73587746E+02,
             -0.38935528E+02,-0.21367317E+02, 0.22717581E+02,-0.38625607E+02,
              0.22611212E+02,-0.13838289E+02,-0.19563032E+02,-0.54408684E+01,
              0.22669931E+02, },
            { 0.96786781E+02, 0.19943217E+03,-0.22654292E+03,-0.25076122E+01,
              0.99311028E+02,-0.32880912E+01, 0.18322615E+02,-0.17872757E+02,
              0.32761443E+01, 0.10708526E+02,-0.29951110E+01,-0.12494042E+02,
             -0.13923942E+02, },
            {-0.15908409E+03, 0.22274886E+03, 0.14706615E+03, 0.20086601E+02,
              0.12031584E+03,-0.89654094E+00,-0.32396168E+02,-0.99818575E+00,
              0.18598759E+02,-0.17126841E+02, 0.17582542E+00,-0.13738128E+02,
             -0.75380640E+01, },
            { 0.28013394E+03,-0.26947153E+02,-0.30688678E+03,-0.24610758E+02,
              0.25527044E+02,-0.90596825E+02,-0.53650866E+01, 0.11049999E+02,
             -0.16220327E+02,-0.88406433E+02,-0.52042227E+01, 0.37062027E+02,
             -0.48077995E+02, },
            { 0.34042722E+03,-0.20572089E+03,-0.69570465E+02, 0.17793367E+03,
              0.15674860E+03, 0.55941879E+02,-0.64308510E+02, 0.11241123E+03,
             -0.44273415E+02, 0.41367416E+02, 0.56130985E+02, 0.22120152E+02,
             -0.57441715E+02, },
            {-0.10435168E+03,-0.22239352E+03, 0.35573297E+03,-0.10317237E+02,
             -0.12436950E+03,-0.36222422E+01,-0.26919477E+02, 0.25896906E+02,
              0.43949404E+01,-0.15245827E+02, 0.24285519E+01, 0.14503664E+02,
              0.14726946E+02, },
            { 0.22333339E+03,-0.35747256E+03,-0.15495203E+03,-0.19590113E+02,
             -0.15038443E+03, 0.54151440E+01, 0.37789562E+02, 0.71866360E+01,
             -0.28933886E+02, 0.18707529E+02,-0.73916203E+00, 0.14262703E+02,
              0.99755716E+01, },
            {-0.29553113E+03, 0.18611118E+02, 0.45680972E+03, 0.50157089E+02,
              0.23859198E+02, 0.80800880E+02,-0.16151365E+02,-0.16563154E+02,
              0.17871490E+02, 0.11023319E+03, 0.84442797E+01,-0.48368690E+02,
              0.49598927E+02, },
            {-0.32021780E+03, 0.24028613E+03, 0.77732681E+02,-0.19011111E+03,
             -0.21534552E+03,-0.59049309E+02, 0.76389923E+02,-0.14133389E+03,
              0.37059875E+02,-0.50378239E+02,-0.66598373E+02,-0.30074066E+02,
              0.62589401E+02, },
            { 0.41164993E+02, 0.96164986E+02,-0.17758118E+03, 0.96721039E+01,
              0.54648762E+02, 0.41163754E+01, 0.14427828E+02,-0.13174432E+02,
             -0.55630002E+01, 0.72903724E+01,-0.63024837E+00,-0.61418190E+01,
             -0.56469097E+01, },
            {-0.10944715E+03, 0.17815541E+03, 0.63606415E+02, 0.73619609E+01,
              0.64106522E+02,-0.35921402E+01,-0.15911514E+02,-0.52352424E+01,
              0.14696769E+02,-0.72762308E+01, 0.22202390E+00,-0.47575579E+01,
             -0.41712513E+01, },
            { 0.11644561E+03,-0.11094306E+02,-0.22673264E+03,-0.28864283E+02,
             -0.28778976E+02,-0.27439207E+02, 0.14110433E+02, 0.86871843E+01,
             -0.75935063E+01,-0.47815193E+02,-0.36480932E+01, 0.22022610E+02,
             -0.18813353E+02, },
            { 0.10188473E+03,-0.10228133E+03,-0.39122402E+02, 0.73224983E+02,
              0.96420273E+02, 0.22319057E+02,-0.32373447E+02, 0.63122818E+02,
             -0.10419510E+02, 0.21842417E+02, 0.27861082E+02, 0.13659900E+02,
             -0.24893913E+02, },
            { 0.17436193E-01,-0.93482248E-01,-0.10793444E+00,-0.82785124E+00,
             -0.84534422E-01,-0.26766161E-03, 0.16654745E-01,-0.30994598E-01,
             -0.11604786E-01, 0.13113019E-01,-0.68038772E-03, 0.19080637E-02,
              0.26341386E-01, },
            { 0.13251559E-01,-0.39075937E-01, 0.66035584E-01, 0.18244596E-01,
             -0.78173661E+00,-0.11768348E-01,-0.53608539E-02,-0.28954308E-01,
              0.22645099E-01, 0.34665675E-02, 0.17233107E-01, 0.14634843E-01,
             -0.29612660E-01, },
            { 0.50487894E+00, 0.22164828E+00,-0.24378799E+00,-0.20211710E+00,
              0.94408840E-01, 0.48972866E+00,-0.23864597E-01,-0.27299488E+00,
             -0.34467158E+00,-0.14131469E+00,-0.20775779E+00,-0.61760709E-01,
             -0.61581101E-01, },
            { 0.99109298E+00, 0.17821749E+00, 0.82714492E+00,-0.94977474E+00,
             -0.61216378E+00, 0.36868054E+00, 0.22957662E+00,-0.12434895E+00,
              0.49470678E-01,-0.90065990E-02,-0.12500758E+00, 0.15639452E-01,
             -0.57548754E-01, },
            { 0.26369982E+01, 0.56819504E+00, 0.27612624E+01, 0.47631841E+01,
              0.26233945E+01,-0.35643741E+00,-0.49008372E+00, 0.86467236E+00,
              0.32177621E+00,-0.11061669E+00, 0.54952288E+00, 0.32997720E-01,
             -0.70062494E+00, },
            { 0.11445436E+01,-0.10629308E+01, 0.18839436E+01,-0.18157120E+01,
              0.57845397E+01,-0.11576208E+00, 0.76202619E+00, 0.12073851E+01,
              0.22669688E-01,-0.58742684E+00,-0.53316754E+00,-0.86090761E+00,
              0.12829594E+00, },
            {-0.12164210E+00,-0.11267498E+00, 0.39087934E+01, 0.14429278E+01,
              0.49848485E+01,-0.38739052E+01, 0.32264167E+00, 0.28122408E+01,
              0.33922617E+01, 0.10338417E+01, 0.14853586E+01, 0.56888652E+00,
              0.14110199E+00, },
            {-0.83919153E+01,-0.36753819E+01,-0.42955184E+01, 0.33921528E+01,
              0.80299025E+01,-0.31461635E+01,-0.15146472E+01, 0.62426537E+00,
              0.49127710E+00, 0.91559422E+00, 0.12294981E+01,-0.11871793E+01,
             -0.17221960E+00, },
            {-0.10749142E+02, 0.90448103E+01,-0.22625639E+02,-0.52351284E+01,
              0.13716090E+02, 0.13832469E+00, 0.71097839E+00,-0.13742863E+01,
             -0.29770541E+01,-0.49114960E+00,-0.18319699E+01, 0.83288515E+00,
              0.42117639E+01, },
            {-0.74518118E+01, 0.20720312E+02,-0.12630376E+02,-0.16478212E+02,
             -0.81918631E+01,-0.70487130E+00,-0.88454456E+01,-0.63079481E+01,
             -0.16885533E+01, 0.26288922E+01, 0.37352061E+01, 0.57972150E+01,
              0.16058006E+01, },
            {-0.51203427E+01,-0.25459335E+01,-0.16263302E+02,-0.22843137E+01,
             -0.19856340E+02, 0.82789288E+01,-0.49226767E+00,-0.64728532E+01,
             -0.86539001E+01,-0.22938893E+01,-0.35961869E+01,-0.14950491E+01,
              0.24064369E+00, },
            { 0.16074701E+02, 0.12596722E+02, 0.77019687E+01,-0.36065149E+01,
             -0.19908415E+02, 0.71198821E+01, 0.27998273E+01,-0.71557932E-01,
             -0.21365404E+01,-0.29742064E+01,-0.35151474E+01, 0.40621052E+01,
              0.22940133E+01, },
            { 0.13587055E+02,-0.25019388E+02, 0.54193542E+02,-0.97644215E+01,
             -0.52721302E+02, 0.34096539E+01, 0.70795673E+00,-0.32761893E+01,
              0.75477500E+01, 0.20922942E+01, 0.12183950E+01,-0.31462874E+01,
             -0.88257294E+01, },
            { 0.19144426E+02,-0.55912872E+02, 0.20548437E+02, 0.58517231E+02,
             -0.10258888E+02, 0.43046956E+01, 0.22909641E+02, 0.11171076E+02,
              0.56796517E+01,-0.34555612E+01,-0.77860579E+01,-0.11080912E+02,
             -0.49921532E+01, },
            { 0.50711031E+01, 0.36411526E+01, 0.15481095E+02, 0.27508340E+01,
              0.21559464E+02,-0.52467356E+01,-0.18340144E+00, 0.42208877E+01,
              0.59293571E+01, 0.12054518E+01, 0.28709695E+01, 0.10667130E+01,
             -0.38529634E+00, },
            {-0.73677697E+01,-0.12092901E+02,-0.48593469E+01,-0.27885952E+01,
              0.14997449E+02,-0.45441275E+01,-0.15541584E+01,-0.77496177E+00,
              0.20404882E+01, 0.22645693E+01, 0.28346663E+01,-0.32707062E+01,
             -0.27267530E+01, },
            {-0.64316149E+01, 0.16303364E+02,-0.38149586E+02, 0.12910197E+02,
              0.42223900E+02,-0.46120901E+01,-0.76996171E+00, 0.53413100E+01,
             -0.56315417E+01,-0.12916847E+01, 0.31741199E+00, 0.23768795E+01,
              0.58335042E+01, },
            {-0.17753616E+02, 0.41826595E+02,-0.77663765E+01,-0.46996101E+02,
              0.16903778E+02,-0.39931262E+01,-0.17180531E+02,-0.67168827E+01,
             -0.45003757E+01, 0.13160985E+01, 0.48622327E+01, 0.64542894E+01,
              0.36038089E+01, },
            {-0.22608569E+00, 0.22383168E-01,-0.21983469E+00, 0.38383208E-01,
              0.10683269E+00,-0.27958038E+00,-0.72073120E+00,-0.13659807E-01,
             -0.94429441E-02,-0.40688934E-02, 0.14465627E-01,-0.12307435E-01,
              0.84743900E-02, },
            { 0.77261649E-01, 0.52328430E-01,-0.33979189E-01, 0.24548575E-01,
              0.15614870E+00, 0.68249077E+00,-0.32020846E+00,-0.23883097E-01,
              0.48189063E-01,-0.71453056E-02, 0.65714070E-02,-0.42875452E-03,
             -0.49787788E-02, },
            { 0.16318841E+00, 0.40782076E+00,-0.13131221E+00,-0.78088120E-01,
             -0.43878078E+00,-0.45414293E+00,-0.18014473E+00, 0.51335122E-01,
              0.92360303E-01, 0.70260219E-01, 0.40587511E-01, 0.57136446E-01,
             -0.38172573E-01, },
            {-0.32295987E+00,-0.36206935E-01, 0.54615408E+00, 0.15255278E+00,
              0.10668567E-02, 0.59954655E+00,-0.58115119E+00,-0.14064194E+00,
             -0.59324689E-02, 0.17862687E-02,-0.16501710E+00,-0.68641931E-01,
              0.14574032E-01, },
            { 0.22033205E+01,-0.42505041E+00, 0.14705453E+01, 0.99113494E-01,
             -0.91881227E+00, 0.16044205E+01, 0.17543232E+01, 0.45916408E+00,
              0.27655859E-02,-0.15211327E+00, 0.62531821E-01, 0.22174665E+00,
             -0.89840271E-01, },
            {-0.98237640E+00, 0.61098063E+00,-0.93203463E-01, 0.24999487E+00,
             -0.16881289E+01,-0.12344168E+01, 0.17829297E+01, 0.27928630E+00,
             -0.16599249E-01, 0.71338467E-01,-0.15047702E+00, 0.21899357E+00,
              0.51836595E-01, },
            { 0.34149848E-01,-0.11778055E+01, 0.41532448E+00, 0.15049255E+00,
              0.86920917E+00, 0.51964009E+00,-0.20884304E+00, 0.89135788E-01,
             -0.16401221E+00,-0.20817281E+00,-0.11379710E+00,-0.21582046E+00,
              0.54201890E-01, },
            { 0.65804678E+00, 0.50968662E-01,-0.83104873E+00,-0.68715566E+00,
              0.75480483E-01,-0.84142131E+00, 0.86892831E+00, 0.25009820E+00,
             -0.44691864E-01, 0.41970626E-01, 0.42782474E+00, 0.50102556E-02,
             -0.18369292E+00, },
            {-0.30072827E+01, 0.10538676E+01,-0.18198307E+01, 0.25968856E+00,
              0.99997920E+00,-0.23609509E+01,-0.17743442E+01,-0.73138219E+00,
              0.11850427E+00, 0.35192615E+00,-0.23847862E+00,-0.57276577E+00,
              0.14691012E+00, },
            { 0.11595041E+01,-0.71592605E+00, 0.47009271E+00, 0.22766799E+00,
              0.25950408E+01, 0.12390486E+01,-0.26281445E+01,-0.37698331E+00,
             -0.21818750E+00,-0.15531531E+00, 0.19898316E+00,-0.42630193E+00,
             -0.24251965E+00, },
            { 0.75564645E-01,-0.15819146E-02,-0.17070128E-01, 0.62731832E-01,
              0.80122389E-01,-0.19053893E-01,-0.75502822E-03, 0.20281678E+00,
             -0.16062942E+00,-0.61040525E-02, 0.32858420E-01,-0.53751818E-02,
              0.11075356E-01, },
            { 0.84059753E-01,-0.26932992E-02, 0.44004381E-01,-0.30800799E-01,
              0.69791381E-02, 0.19021643E-01,-0.30156031E-01, 0.18149000E+00,
              0.21475334E+00,-0.98471157E-02,-0.55861087E-02, 0.48251478E-02,
             -0.33545550E-02, },
            { 0.15172923E+00,-0.10189057E+00, 0.25645059E-01,-0.52431297E-01,
             -0.31783830E-01,-0.11050274E+00, 0.11658210E-01,-0.22356764E-02,
             -0.46832764E+00,-0.37870400E-01,-0.47675385E-02,-0.38657390E-01,
              0.74187763E-01, },
            { 0.38575072E-01,-0.21887577E+00, 0.43412637E-01, 0.79081804E-01,
              0.39201990E-01, 0.11154531E-01,-0.64383946E-01, 0.39611334E+00,
             -0.98804720E-02, 0.39850544E-01,-0.27133226E-01, 0.39764382E-01,
             -0.42043138E-01, },
            {-0.27693447E-02, 0.97743109E-01,-0.95391117E-01, 0.11043071E-01,
              0.31084713E-01,-0.40372331E-02, 0.13213970E-01,-0.24805332E-01,
              0.15137902E-01, 0.56641247E-01, 0.22465751E+00,-0.23782956E-02,
              0.96832955E-03, },
            { 0.56767878E-02, 0.27197037E-01, 0.41327935E-01,-0.92214338E-01,
             -0.47889195E-01,-0.41702380E-02,-0.81137735E-02, 0.41315701E-01,
             -0.75267494E-01,-0.20449100E+00, 0.59991557E-01,-0.68433122E-02,
              0.14528462E-01, },
            { 0.14191374E-01,-0.28540870E-01, 0.25324695E-01,-0.79455242E-01,
             -0.23113195E-01, 0.44245183E-01, 0.34820773E-01,-0.90411389E-02,
             -0.25287878E-01, 0.31344425E-02, 0.10355319E-01,-0.10612031E+00,
              0.21018282E-01, },
            {-0.77975765E-02,-0.76840571E-02, 0.15142567E-01,-0.22948675E-01,
             -0.47588389E-01, 0.11881715E-01,-0.63865617E-01, 0.30425686E-01,
             -0.82677780E-02, 0.21936271E-01, 0.12532286E-01,-0.44801785E-02,
             -0.13164043E+00, },
            {-0.46902079E-01,-0.13513894E-01, 0.16624622E-03,-0.26316926E-01,
             -0.12186967E-02, 0.28734948E-01,-0.22959083E-02, 0.10504897E-01,
             -0.78018638E-02, 0.84232548E-02,-0.90076961E-02,-0.66834749E-02,
              0.13984973E-01, },
            {-0.61138693E-01, 0.11805169E-01, 0.78700714E-01, 0.18015917E-01,
              0.28624056E-01,-0.76267403E-02,-0.46754625E-01,-0.22486155E-02,
              0.15880592E-01, 0.64636157E-02, 0.14192480E-01,-0.67947758E-03,
             -0.35506718E-01, },
            {-0.25783082E-01, 0.39371252E-01,-0.79181977E-02,-0.47959839E-02,
             -0.34519922E-01,-0.25952054E-01,-0.92375092E-02, 0.23945650E-01,
             -0.36872244E-02, 0.38071219E-01, 0.14242783E-01, 0.10948616E-01,
              0.21540886E-01, },
            {-0.42456158E-01, 0.11031415E-01, 0.34051061E-01, 0.57469562E-01,
              0.45886550E-01, 0.13773274E-01,-0.64351404E-03,-0.16607651E-01,
              0.18327732E-01,-0.26219755E-02,-0.77091181E-02, 0.13056586E-01,
             -0.54514217E-02, },
         },
         {
            { 0.94509640E+01, 0.10107969E+00, 0.95546111E-01, 0.34959666E-01,
             -0.25399696E-01, 0.18447235E-01,-0.18753277E-01,-0.21336963E-01,
              0.20273566E-01,-0.74352040E-02, 0.86263977E-02,-0.19587195E-01,
             -0.25863366E-02, },
            { 0.17576997E+01,-0.20958441E+00, 0.20107350E+00, 0.26883551E+00,
             -0.69596463E+00, 0.10864495E+00,-0.57094645E-01,-0.11880032E+00,
              0.32511491E+00,-0.73277988E-01,-0.19179016E+00, 0.55194344E-01,
             -0.48167195E-01, },
            { 0.29966228E+02,-0.28514745E+01,-0.27479420E+01, 0.16417269E+00,
              0.23762665E+01,-0.24790712E+00, 0.38372591E-01, 0.68017262E+00,
             -0.64208418E+00, 0.19077405E+00,-0.19376384E+00, 0.37586835E+00,
              0.30274448E+00, },
            {-0.79763260E+01, 0.40185118E+01, 0.27582934E+01,-0.44167314E+01,
              0.16556219E+02,-0.10127513E+01,-0.14381729E-01, 0.10526127E+01,
             -0.36751299E+01, 0.16371242E+01, 0.41598487E+01,-0.55727178E+00,
              0.25336441E+00, },
            {-0.15115533E+03, 0.20107765E+02, 0.16070599E+02,-0.69206400E+01,
             -0.15509940E+02,-0.65745495E-01, 0.61048579E+00,-0.44110274E+01,
              0.59234338E+01,-0.19652240E+01, 0.83969170E+00,-0.19654928E+01,
             -0.22424846E+01, },
            {-0.51554680E+02,-0.49286888E+02,-0.15149562E+02, 0.23192854E+02,
             -0.80976891E+02, 0.17976050E+01, 0.31771953E+01,-0.22073927E+01,
              0.17565590E+02,-0.79421062E+01,-0.22982227E+02, 0.20799625E+01,
              0.11556675E+01, },
            { 0.26708896E+03,-0.49775471E+02,-0.31839897E+02, 0.28108734E+02,
              0.37422558E+02, 0.36079819E+01,-0.32567403E+01, 0.10001911E+02,
             -0.17851255E+02, 0.60564766E+01,-0.76352984E+00, 0.37349446E+01,
              0.60045552E+01, },
            { 0.22564854E+03, 0.16248386E+03, 0.12195727E+02,-0.51379269E+02,
              0.16426907E+03, 0.24047847E+01,-0.12291703E+02,-0.43176427E+00,
             -0.39148830E+02, 0.16007685E+02, 0.51836742E+02,-0.38026702E+01,
             -0.54714437E+01, },
            {-0.22505072E+03, 0.46291023E+02, 0.24638092E+02,-0.39034763E+02,
             -0.37432152E+02,-0.72609534E+01, 0.49040074E+01,-0.89484711E+01,
              0.20959160E+02,-0.69502692E+01,-0.58399487E+00,-0.28382370E+01,
             -0.64945631E+01, },
            {-0.28448126E+03,-0.19788374E+03, 0.13808954E+02, 0.50430801E+02,
             -0.14956322E+03,-0.69601116E+01, 0.16451906E+02, 0.45007586E+01,
              0.39577290E+02,-0.14559422E+02,-0.51503658E+02, 0.33227439E+01,
              0.67428279E+01, },
            { 0.75200890E+02,-0.13670172E+02,-0.64882317E+01, 0.17861347E+02,
              0.13128088E+02, 0.39957590E+01,-0.23009706E+01, 0.26648011E+01,
             -0.84337873E+01, 0.26480327E+01, 0.69651747E+00, 0.70226914E+00,
              0.24378769E+01, },
            { 0.11641260E+03, 0.81135353E+02,-0.14079868E+02,-0.17968302E+02,
              0.50422318E+02, 0.35681407E+01,-0.73284211E+01,-0.27828982E+01,
             -0.14672050E+02, 0.49403863E+01, 0.18719210E+02,-0.10883954E+01,
             -0.26159148E+01, },
            {-0.12469147E-01, 0.13936483E+01, 0.20418575E+01,-0.22939935E-01,
              0.33805311E-01, 0.10366075E-01, 0.52215006E-01,-0.36718503E-01,
             -0.45215521E-01, 0.20861069E-02, 0.31817243E-01,-0.47385283E-02,
             -0.63512935E-02, },
            {-0.79864658E-01,-0.21967120E+01, 0.13169786E+01,-0.49137101E-01,
              0.79261839E-01, 0.56432478E-01,-0.11213575E-01, 0.39652806E-01,
             -0.29792653E-01, 0.64580194E-02,-0.23676369E-02, 0.21759911E-01,
             -0.17042236E-01, },
            {-0.10706545E+01, 0.23217793E+01, 0.22575028E+01,-0.52305853E+00,
              0.33758911E+00, 0.40237907E+00, 0.70796597E+00,-0.25653306E+00,
              0.18688981E+00,-0.25361739E-01, 0.87322183E-01, 0.24367885E+00,
              0.40593699E-01, },
            {-0.15108147E+00,-0.16911287E+01, 0.46142992E+00,-0.10303921E+01,
              0.93612778E+00,-0.41574039E-02, 0.11723827E+00,-0.15480372E+00,
              0.25974426E-01,-0.10854245E+00,-0.14955416E+00, 0.20788647E+00,
              0.12922181E+00, },
            { 0.45731859E+01, 0.24417906E+02,-0.21377110E+02,-0.15830410E+00,
              0.28918290E+01,-0.90342098E+00,-0.10628958E+01, 0.70779812E+00,
              0.28248230E+00, 0.51816136E+00,-0.36014694E+00,-0.21408524E+00,
              0.44734198E+00, },
            {-0.10066270E+02, 0.24201607E+02, 0.22228245E+02, 0.18787283E+01,
              0.35598913E+00,-0.81920505E+00,-0.45490566E+00, 0.21012618E+00,
              0.17021557E+01,-0.23625430E+00,-0.15785757E+00,-0.44948339E+00,
             -0.50422323E+00, },
            { 0.25363052E+02,-0.22388351E+02,-0.88549417E+00,-0.22121559E+00,
              0.23813438E+01,-0.34516563E+01,-0.12210044E+02, 0.55636535E+01,
             -0.32453895E+01, 0.19233570E+00,-0.14384522E+01,-0.48079529E+01,
             -0.13344591E+01, },
            { 0.43639555E+01,-0.82982798E+01, 0.11524791E+02, 0.18317818E+02,
             -0.13572060E+02, 0.43871188E+01,-0.21051328E+01, 0.19665152E+01,
             -0.18783902E+01, 0.23498127E+01, 0.49273577E+01,-0.22936010E+01,
             -0.35927687E+01, },
            {-0.33341885E+02,-0.11396944E+03, 0.19362669E+03, 0.21216767E+01,
             -0.21863722E+02, 0.87932034E+01, 0.66870275E+01,-0.20778823E+01,
              0.39056785E+01,-0.55137281E+01, 0.73257333E+00, 0.10095549E+01,
             -0.58631840E+01, },
            { 0.10840975E+03,-0.21727190E+03,-0.68914238E+02,-0.21362246E+02,
             -0.86977644E+01, 0.82116756E+01, 0.49080110E+01, 0.17874563E+00,
             -0.12050731E+02, 0.51833916E+00, 0.31218398E+01, 0.36935799E+01,
              0.80403614E+01, },
            {-0.15367139E+03, 0.11878262E+03, 0.51804714E+02, 0.41489388E+02,
             -0.92071486E+01, 0.51096478E+01, 0.64470383E+02,-0.44764999E+02,
              0.16702759E+02, 0.28305548E+00, 0.86546431E+01, 0.29645947E+02,
              0.80854740E+01, },
            {-0.86085358E+01, 0.41116173E+02,-0.47850399E+02,-0.11364082E+03,
              0.49419270E+02,-0.27409754E+02, 0.12218876E+02,-0.15123549E+02,
              0.17315872E+02,-0.15716614E+02,-0.34133495E+02, 0.81285315E+01,
              0.23876568E+02, },
            { 0.82036057E+02, 0.18187785E+03,-0.55807098E+03,-0.10879265E+02,
              0.61412319E+02,-0.27611143E+02,-0.19147324E+02,-0.71760750E+00,
             -0.27551128E+02, 0.17640097E+02,-0.56628734E-01,-0.14020643E+01,
              0.21747429E+02, },
            {-0.34376413E+03, 0.61911011E+03, 0.88020509E+00, 0.80663330E+02,
              0.33479347E+02,-0.29412020E+02,-0.12201406E+02,-0.44523067E+01,
              0.36756096E+02, 0.16952411E+01,-0.14172467E+02,-0.11445619E+02,
             -0.29532711E+02, },
            { 0.39695712E+03,-0.31497382E+03,-0.28771411E+03,-0.17258989E+03,
             -0.86242523E+01, 0.11642606E+02,-0.14586752E+03, 0.13591988E+03,
             -0.39550259E+02,-0.15524073E+01,-0.22016012E+02,-0.76522240E+02,
             -0.19635260E+02, },
            {-0.40375984E+02,-0.67061432E+02, 0.28259424E+02, 0.29492319E+03,
             -0.67607613E+02, 0.62555237E+02,-0.28627241E+02, 0.50263222E+02,
             -0.52197632E+02, 0.39175499E+02, 0.90898705E+02,-0.12516282E+02,
             -0.61318790E+02, },
            {-0.77070221E+02,-0.11974761E+03, 0.69099872E+03, 0.18479025E+02,
             -0.77280106E+02, 0.33186859E+02, 0.23168581E+02, 0.75805955E+01,
              0.50765301E+02,-0.22174515E+02,-0.21018499E+00, 0.45265010E+00,
             -0.30652090E+02, },
            { 0.42491214E+03,-0.73491571E+03, 0.15072490E+03,-0.11781622E+03,
             -0.46392895E+02, 0.41384529E+02, 0.10866575E+02, 0.84837999E+01,
             -0.48217793E+02,-0.58226819E+01, 0.23120079E+02, 0.13756743E+02,
              0.40079506E+02, },
            {-0.45912537E+03, 0.38094446E+03, 0.50466553E+03, 0.24710139E+03,
              0.43194183E+02,-0.31068718E+02, 0.14789594E+03,-0.17133972E+03,
              0.43186558E+02, 0.93547171E+00, 0.24350426E+02, 0.87021637E+02,
              0.20689638E+02, },
            { 0.12518957E+03, 0.21702751E+02, 0.59070374E+02,-0.33266949E+03,
              0.31520334E+02,-0.63018490E+02, 0.29843792E+02,-0.66842224E+02,
              0.62429546E+02,-0.40911572E+02,-0.10445174E+03, 0.77937250E+01,
              0.66863708E+02, },
            { 0.23716257E+02, 0.26841919E+02,-0.31155273E+03,-0.99036121E+01,
              0.36229214E+02,-0.13366701E+02,-0.94885092E+01,-0.60028815E+01,
             -0.28343122E+02, 0.97320232E+01,-0.21416096E+00, 0.32326192E+00,
              0.14711193E+02, },
            {-0.18264751E+03, 0.31139908E+03,-0.10778708E+03, 0.57750397E+02,
              0.20604429E+02,-0.19741055E+02,-0.31399972E+01,-0.45069766E+01,
              0.22200363E+02, 0.40623951E+01,-0.12418812E+02,-0.53976035E+01,
             -0.18255384E+02, },
            { 0.19313271E+03,-0.16603587E+03,-0.27830505E+03,-0.11844001E+03,
             -0.29174128E+02, 0.17652523E+02,-0.55044979E+02, 0.75791016E+02,
             -0.17480881E+02, 0.37095240E+00,-0.96177435E+01,-0.35863873E+02,
             -0.79011192E+01, },
            {-0.83767029E+02, 0.14769975E+02,-0.53073029E+02, 0.13477794E+03,
             -0.10499868E+01, 0.24099756E+02,-0.11736704E+02, 0.30020025E+02,
             -0.25523891E+02, 0.15339497E+02, 0.43580658E+02,-0.11596680E+01,
             -0.26010191E+02, },
            {-0.24157926E-01,-0.34682450E+00,-0.18939769E+00,-0.64226544E+00,
              0.55502748E+00,-0.18019242E-02, 0.22090345E-02,-0.17449288E-02,
             -0.28906872E-01, 0.43518983E-01,-0.22777419E-02, 0.19499291E-01,
              0.26698962E-01, },
            {-0.66926338E-01,-0.13376318E+00,-0.66897385E-01,-0.60208154E+00,
             -0.70030004E+00,-0.11302370E-01, 0.22896841E-01, 0.18626200E-01,
              0.16040893E-01, 0.18582191E-01,-0.50697462E-02, 0.14439798E-01,
             -0.25851393E-01, },
            { 0.14338864E+00,-0.20765391E+00, 0.20112073E-01,-0.19753712E+01,
              0.36831617E+00, 0.33009917E+00,-0.17354338E+00, 0.28956002E-02,
             -0.41839543E+00,-0.34900645E-02,-0.25787419E+00, 0.39183512E-01,
              0.11534757E+00, },
            { 0.57123256E+00, 0.13649093E+00, 0.26264960E-01,-0.47044334E+00,
             -0.22277884E+01, 0.34276772E+00, 0.26841480E+00,-0.31208158E+00,
              0.21924429E+00,-0.73497251E-01,-0.54401502E-01, 0.14796603E-01,
             -0.10233092E-01, },
            { 0.26954265E+01, 0.43012781E+01, 0.75683284E+01,-0.86731186E+01,
              0.55410342E+01,-0.34549937E+00,-0.13151923E+01, 0.79660070E+00,
              0.91225111E+00,-0.41482779E+00, 0.31972158E+00,-0.65595812E+00,
             -0.98114915E-01, },
            { 0.17693305E+01, 0.10319415E+01, 0.57046089E+01,-0.44164987E+01,
             -0.76197066E+01, 0.69258058E+00,-0.58791065E+00,-0.44445005E+00,
              0.10734792E+01,-0.10839006E+01, 0.18576832E+00,-0.56307358E+00,
             -0.81509627E-01, },
            { 0.19725923E+00, 0.28090067E+01, 0.29866964E-01, 0.65746832E+01,
             -0.11214925E+01,-0.20306814E+01, 0.75946403E+00, 0.37662435E+00,
              0.41457515E+01, 0.45575309E+00, 0.18497028E+01,-0.14746182E+00,
             -0.16188046E+01, },
            {-0.33737960E+01,-0.32643378E+00, 0.13863986E+01, 0.35408862E+01,
              0.94330416E+01,-0.89115298E+00,-0.36629066E+00, 0.16838450E+01,
             -0.21706350E+01, 0.13435290E+01, 0.55284905E+00,-0.85979068E+00,
             -0.11613747E+01, },
            {-0.12316348E+02,-0.62260914E+01,-0.65585999E+02, 0.51939068E+02,
             -0.18281101E+02, 0.18273875E+00, 0.63747220E+01,-0.31539526E+01,
             -0.72625136E+01,-0.33774602E+00,-0.64533216E+00, 0.45377502E+01,
             -0.94161624E+00, },
            {-0.12553596E+02, 0.13752454E+02,-0.35581814E+02, 0.14409949E+02,
              0.47376873E+02,-0.34375093E+01, 0.25728254E+01, 0.43016977E+01,
             -0.10728865E+02, 0.57186842E+01,-0.11502291E+01, 0.40079656E+01,
              0.17250158E+01, },
            {-0.35611405E+01,-0.97855129E+01,-0.59383321E+01,-0.11810493E+02,
             -0.76464243E+01, 0.37008057E+01,-0.16280932E+01,-0.65146512E+00,
             -0.10992208E+02,-0.11756465E+01,-0.40906620E+01,-0.35428894E+00,
              0.46568985E+01, },
            { 0.40749516E+01,-0.54280490E+00,-0.29149284E+01,-0.16286656E+01,
             -0.16713770E+02, 0.47613055E+00,-0.20696537E+01,-0.19630735E+01,
              0.53624787E+01,-0.44261832E+01,-0.13592387E+01, 0.33461597E+01,
              0.42439528E+01, },
            { 0.19575254E+02,-0.26728251E+01, 0.15306610E+03,-0.97623291E+02,
              0.11181700E+02, 0.33377397E+01,-0.11486933E+02, 0.41499352E+01,
              0.16440626E+02, 0.37850554E+01,-0.18400379E+00,-0.10662279E+02,
              0.31876299E+01, },
            { 0.31439281E+02,-0.47513309E+02, 0.64092300E+02,-0.10152738E+02,
             -0.89982887E+02, 0.79072332E+01,-0.60005236E+01,-0.98099136E+01,
              0.26642679E+02,-0.94802303E+01, 0.32628336E+01,-0.80624046E+01,
             -0.41372290E+01, },
            { 0.24437640E+01, 0.10388881E+02, 0.68885064E+01, 0.12287474E+02,
              0.17508669E+02,-0.15349623E+01, 0.13399680E+01, 0.89213789E-01,
              0.77907591E+01, 0.79130781E+00, 0.28016553E+01, 0.53495371E+00,
             -0.36855667E+01, },
            { 0.68383670E+00, 0.14101259E+00, 0.12219146E+01,-0.87792912E+01,
              0.14048577E+02,-0.40416885E-01, 0.34328499E+01, 0.43522057E+00,
             -0.37442031E+01, 0.36730547E+01, 0.10144058E+01,-0.31850066E+01,
             -0.35276756E+01, },
            {-0.11864492E+02, 0.65092659E+01,-0.10362085E+03, 0.60457390E+02,
              0.88302031E+01,-0.48034692E+01, 0.68933558E+01,-0.17215779E+01,
             -0.11243499E+02,-0.31618342E+01, 0.66602868E+00, 0.76481528E+01,
             -0.22676835E+01, },
            {-0.26041988E+02, 0.39193073E+02,-0.33144814E+02,-0.53187180E+01,
              0.56564884E+02,-0.62643838E+01, 0.45099568E+01, 0.61439466E+01,
             -0.18840761E+02, 0.46164675E+01,-0.28137391E+01, 0.49374666E+01,
              0.28067627E+01, },
            {-0.25802639E+00,-0.10362916E+00,-0.14398941E+00, 0.58753002E-01,
              0.12338717E+00,-0.46615648E+00,-0.90974379E+00, 0.48583642E-01,
             -0.38379822E-01, 0.27855985E-01, 0.10477463E-02,-0.31467213E-02,
              0.46931063E-02, },
            { 0.13942069E+00, 0.12052406E-01, 0.13556781E+00,-0.54092322E-01,
              0.15451410E+00, 0.88882434E+00,-0.48357457E+00,-0.32771014E-01,
              0.11079021E-01,-0.23551969E-01,-0.56003863E-02,-0.15953735E-01,
              0.25094349E-02, },
            { 0.20348473E+00, 0.49829680E+00,-0.55318739E-01,-0.92830181E-01,
             -0.18149967E+00,-0.54599792E+00,-0.81165332E+00, 0.41849393E-01,
              0.22468764E+00, 0.12768607E-01,-0.45675814E-01, 0.34527790E-01,
              0.90832688E-01, },
            { 0.19808038E+00, 0.24423756E-01, 0.55058855E+00, 0.96684530E-01,
              0.14878929E+00, 0.10547436E+01,-0.62313193E+00,-0.34745380E+00,
              0.50657775E-01,-0.10458727E+00,-0.26829256E-01,-0.22105072E-01,
              0.96443847E-01, },
            { 0.31103616E+01, 0.71752518E+00, 0.43198234E+00,-0.83313775E+00,
             -0.53591412E+00, 0.21094110E+01, 0.14961641E+01, 0.52341461E+00,
              0.56767017E+00,-0.88358931E-01,-0.25685966E-01,-0.62317005E-02,
              0.82329907E-01, },
            {-0.23380406E+01, 0.12645091E+01,-0.13486705E+01, 0.41103089E+00,
             -0.19716953E+01,-0.13170983E+01, 0.23266773E+01, 0.89183159E-01,
              0.99011421E+00, 0.14776289E+00,-0.96008055E-01, 0.22073540E+00,
              0.14917378E-01, },
            { 0.83357565E-01,-0.13930918E+01, 0.16632780E+00, 0.31476656E+00,
              0.29862192E+00,-0.63334607E-01, 0.78159118E+00,-0.13982970E-01,
             -0.37854099E+00,-0.40592089E-01, 0.14077954E+00,-0.17048199E+00,
             -0.20473443E+00, },
            {-0.39951530E+00, 0.38091745E-01,-0.17263715E+01,-0.53047764E+00,
             -0.42756036E-01,-0.14003088E+01, 0.27357009E+00, 0.70787126E+00,
             -0.17673521E+00, 0.21879543E+00, 0.80344655E-01,-0.11453548E+00,
             -0.28751436E+00, },
            {-0.55554628E+01, 0.63528180E-01, 0.22511274E+00, 0.16389220E+01,
              0.68950057E-01,-0.38856034E+01,-0.18219573E+01,-0.12218167E+01,
             -0.72119141E+00, 0.12372471E-01, 0.60922831E-01,-0.13905819E-01,
             -0.64103089E-01, },
            { 0.34650526E+01,-0.16690048E+01, 0.22416811E+01,-0.33316332E+00,
              0.29141891E+01, 0.15945005E+01,-0.39390171E+01, 0.14417328E+00,
             -0.18604945E+01,-0.11784468E+00,-0.16993752E-01,-0.41759604E+00,
             -0.18586282E+00, },
            { 0.81547573E-02, 0.49226344E-01, 0.81611373E-01, 0.27063418E-01,
              0.56695234E-01, 0.29357164E-02, 0.14539908E-01, 0.30828521E+00,
             -0.25350037E+00, 0.32837324E-01, 0.15210060E-01,-0.29200683E-02,
              0.44837701E-02, },
            { 0.63956775E-01, 0.18061155E-01, 0.10958803E-01,-0.12470478E-01,
              0.35997184E-02,-0.25639519E-01,-0.23812983E-01, 0.26933238E+00,
              0.31840917E+00,-0.11766966E-02,-0.27984953E-02,-0.12501192E-02,
             -0.80634849E-02, },
            { 0.14839776E+00,-0.10045361E+00, 0.11778341E-01, 0.67010596E-01,
              0.10346636E+00,-0.71791507E-01, 0.17526034E-01, 0.13578543E+00,
             -0.68693244E+00,-0.82893714E-01, 0.53929277E-02, 0.23169667E-02,
              0.82021438E-01, },
            { 0.72574534E-01,-0.60684308E-01,-0.86023420E-01, 0.90452895E-01,
              0.34990810E-01,-0.15001757E+00,-0.22994665E-01, 0.55149466E+00,
              0.16523851E+00,-0.21080254E-01,-0.10785421E-01, 0.24345174E-01,
             -0.54076161E-01, },
            {-0.14327258E-01, 0.63852847E-01,-0.25183151E-01, 0.29911581E-01,
              0.37023772E-01, 0.14488783E-01,-0.15269484E-01,-0.60717769E-01,
              0.13028531E-01, 0.14004098E+00, 0.26168507E+00,-0.16771069E-01,
              0.91431160E-02, },
            { 0.12888256E+00, 0.18391632E-02,-0.29771950E-01,-0.61108533E-01,
              0.42996099E-02,-0.14745311E-01,-0.18162554E-01, 0.26890341E-01,
             -0.88421255E-01,-0.26770899E+00, 0.11803225E+00,-0.93428353E-02,
              0.73491596E-02, },
            { 0.42300068E-01,-0.45630768E-01, 0.12768690E-01,-0.31926125E-01,
              0.13576030E-01, 0.21456007E-01, 0.21313623E-01,-0.25462057E-01,
             -0.38112130E-01,-0.38594943E-02, 0.38873557E-01,-0.10141932E+00,
              0.36759220E-01, },
            { 0.96034102E-01, 0.24284817E-01,-0.10095998E+00,-0.96809138E-02,
             -0.16556034E-01, 0.69916956E-02,-0.73311739E-01, 0.47385052E-01,
             -0.22124328E-01,-0.81969174E-02, 0.11834414E-01,-0.44802796E-01,
             -0.12607549E+00, },
            {-0.29452965E-01,-0.18742386E-01,-0.76822206E-01, 0.27353512E-01,
              0.59363917E-02, 0.14614475E-01,-0.26352966E-01,-0.18402827E-02,
             -0.14121162E-01,-0.11845143E-01,-0.38414609E-02,-0.61304052E-02,
             -0.21523267E-02, },
            { 0.61670173E-01, 0.75453207E-01, 0.65484559E-02,-0.33840731E-01,
              0.21261007E-01,-0.11356534E-01,-0.17962535E-02, 0.23999993E-01,
              0.22581514E-01,-0.91374582E-02,-0.62641094E-03,-0.11860891E-01,
             -0.49161039E-01, },
            {-0.25092568E-01, 0.65070137E-01,-0.17101344E-01,-0.30995658E-01,
             -0.48828226E-01,-0.25985174E-01, 0.12641994E-01, 0.94131026E-02,
              0.50771046E-01, 0.10564291E-01,-0.72512310E-02,-0.25239275E-02,
              0.16289363E-01, },
            {-0.34294993E-01,-0.21341737E-01,-0.14802243E-01, 0.52205998E-01,
              0.39956398E-01, 0.15908470E-01, 0.47674975E-02, 0.53186305E-02,
              0.79527143E-02,-0.18043905E-01, 0.74231257E-02, 0.87468559E-02,
              0.36759220E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir21Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30515985E+01, 0.18279366E-01, 0.17321778E-01, 0.13490525E-02,
              0.54005985E-02,-0.12811268E-02,-0.19620711E-02, 0.16095781E-02,
              0.14036379E-02, },
            { 0.36024821E+00,-0.26675237E-01, 0.60383171E-01, 0.85761592E-01,
              0.32015789E-01,-0.64746398E-02,-0.23822611E-01, 0.12654256E-02,
              0.19042149E-01, },
            { 0.84638625E+00,-0.16989470E+00,-0.25357148E+00,-0.44812657E-01,
             -0.90480018E-02, 0.13199283E-01,-0.49734190E-02,-0.90144098E-01,
             -0.46962071E-01, },
            {-0.31312117E+00,-0.13555598E+00,-0.11967962E+00,-0.20256715E+00,
             -0.13136697E+00, 0.48089355E-01, 0.87422676E-01, 0.93227997E-02,
             -0.46409301E-01, },
            {-0.11959354E+01, 0.43471059E+00, 0.63000107E+00, 0.10890940E+00,
             -0.57202525E-01,-0.42029072E-01, 0.43181811E-01, 0.22912052E+00,
              0.13201462E+00, },
            {-0.73961414E-01, 0.12270621E+00, 0.51998906E-01, 0.10816619E+00,
              0.11332868E+00,-0.50417311E-01,-0.68393320E-01,-0.12005514E-01,
              0.23018699E-01, },
            { 0.42550960E+00,-0.27673936E+00,-0.39599004E+00,-0.66320188E-01,
              0.74589588E-01, 0.29583050E-01,-0.38249172E-01,-0.14549865E+00,
             -0.96091665E-01, },
            {-0.81040589E-02,-0.14526182E+00,-0.32619151E+00, 0.83900504E-02,
              0.61066765E-02,-0.79448353E-02, 0.88535789E-02, 0.61606448E-02,
              0.44978694E-02, },
            { 0.34204829E-01, 0.34175068E+00,-0.15657048E+00,-0.35750011E-02,
              0.16515717E-02, 0.34812370E-02,-0.10691710E-01,-0.10663674E-01,
              0.23136511E-02, },
            {-0.77340782E-01, 0.29641938E+00, 0.46389941E-02, 0.60751557E-01,
              0.13668638E+00, 0.73995506E-02, 0.36284521E-01, 0.43617327E-01,
             -0.78822457E-03, },
            {-0.23608712E-01, 0.67810323E-02, 0.38915452E+00,-0.11137573E+00,
             -0.67211337E-01, 0.10261582E+00,-0.20004436E-01,-0.96985581E-03,
              0.74450895E-01, },
            { 0.24482937E+00, 0.68339103E+00, 0.12487189E+01,-0.17763805E+00,
             -0.11472326E+00, 0.13915654E+00,-0.96532442E-01,-0.11791184E+00,
             -0.16605750E+00, },
            {-0.76662284E+00,-0.14496020E+01, 0.84100413E+00, 0.16240078E+00,
              0.76123066E-01,-0.33510583E-04, 0.83549894E-01, 0.23060144E+00,
             -0.16083211E+00, },
            { 0.14090075E+01,-0.13704493E+01, 0.82855475E+00, 0.23537932E+00,
             -0.70777285E+00,-0.28288743E+00,-0.19857667E+00,-0.45078415E+00,
             -0.24828870E-01, },
            { 0.88282007E+00,-0.41681415E+00,-0.20059464E+01, 0.15218285E+00,
              0.75277686E+00,-0.67411071E+00, 0.11159578E+00,-0.46602711E-01,
             -0.56245166E+00, },
            {-0.76950717E+00,-0.10647525E+01,-0.34677619E+00, 0.63541669E+00,
              0.15894818E+00,-0.45485401E+00, 0.20325267E+00, 0.27888203E+00,
              0.53756195E+00, },
            { 0.23074839E+01, 0.57445264E+00,-0.17256494E+01,-0.20322417E+00,
             -0.25591567E+00,-0.21160703E-01,-0.19882099E+00,-0.64506978E+00,
              0.51714456E+00, },
            {-0.38324997E+01, 0.20495992E+01,-0.12175716E+01,-0.10217855E+01,
              0.12773292E+01, 0.99618143E+00, 0.27631152E+00, 0.99627531E+00,
              0.13893232E-01, },
            {-0.26958106E+01, 0.34171972E+00, 0.34515679E+01, 0.34862801E+00,
             -0.16408192E+01, 0.13690850E+01,-0.63893899E-01, 0.18333472E+00,
              0.11947597E+01, },
            { 0.52510667E+00, 0.79661417E+00,-0.27414501E+00,-0.59034508E+00,
             -0.29286742E-01, 0.33197421E+00,-0.11383864E+00,-0.16956963E+00,
             -0.43929490E+00, },
            {-0.18387926E+01, 0.63808429E+00, 0.13265423E+01,-0.18234240E-01,
              0.18122619E+00, 0.92859911E-02, 0.90937927E-01, 0.46281740E+00,
             -0.37837014E+00, },
            { 0.27536213E+01,-0.96891516E+00, 0.93679023E+00, 0.81044078E+00,
             -0.73057920E+00,-0.86545914E+00,-0.13849419E+00,-0.62656611E+00,
              0.88995732E-02, },
            { 0.20475008E+01,-0.35984865E+00,-0.18440652E+01,-0.44248223E+00,
              0.98484981E+00,-0.84341449E+00,-0.35579182E-01,-0.14597042E+00,
             -0.76683801E+00, },
            {-0.10489090E-01,-0.39877072E-01,-0.26295239E-01, 0.73930889E-01,
             -0.10012757E+00,-0.76692938E-02, 0.43066703E-02,-0.10926161E-01,
             -0.48446790E-02, },
            { 0.18568033E-01,-0.17637935E-01,-0.15594788E-01, 0.10350185E+00,
              0.84533416E-01,-0.65629520E-02,-0.97999501E-03, 0.40845182E-02,
             -0.18347696E-02, },
            { 0.30939199E-01, 0.14365974E+00, 0.14420240E+00,-0.77935576E-01,
              0.16382048E+00,-0.65682996E-02,-0.39891735E-01, 0.59632142E-02,
             -0.52234367E-01, },
            {-0.16920580E+00, 0.51358785E-01, 0.32620181E-02,-0.74220739E-01,
              0.15272126E-01, 0.58219019E-01, 0.18921616E-01, 0.17532919E-01,
              0.32072235E-01, },
            { 0.11713439E-01, 0.24454077E+00, 0.22597702E+00,-0.16024730E+00,
             -0.13617696E+00, 0.23809751E-01, 0.65214179E-01, 0.66272601E-01,
             -0.77504419E-01, },
            {-0.36200440E+00, 0.25575509E-01,-0.46344455E-01, 0.24129643E+00,
             -0.14844865E+00,-0.63828021E-01, 0.11078907E+00,-0.16530167E-01,
             -0.12369106E-01, },
            {-0.18463311E+00,-0.74005991E+00,-0.46722332E+00, 0.24123137E+00,
             -0.84606230E+00, 0.38615707E-01, 0.21995303E+00,-0.47678851E-01,
              0.23793846E+00, },
            { 0.10495080E+01, 0.73820189E-01, 0.17725090E-01, 0.56766880E+00,
              0.60886084E-02,-0.12156435E+00,-0.15492921E+00,-0.12488942E+00,
             -0.24417043E+00, },
            { 0.22861373E-01,-0.35652456E+00,-0.40377608E+00, 0.73468529E-01,
              0.20436759E+00,-0.63278258E-01,-0.14106607E+00,-0.82917646E-01,
              0.15102735E+00, },
            { 0.48683444E+00,-0.13043688E+00, 0.14397170E+00,-0.34078661E+00,
              0.98404596E-02, 0.12281654E+00,-0.21639533E+00, 0.21982001E-01,
              0.36892843E-01, },
            { 0.22279333E+00, 0.97342873E+00, 0.53715861E+00,-0.29357713E+00,
              0.11556523E+01, 0.34201711E-01,-0.19304839E+00, 0.81136659E-01,
             -0.24921562E+00, },
            {-0.13376596E+01,-0.14052005E+00, 0.35047200E-01,-0.92397130E+00,
             -0.17662610E+00,-0.17981412E-01, 0.24125864E+00, 0.16264918E+00,
              0.29551035E+00, },
            {-0.21483745E-01,-0.36245286E-02,-0.66329092E-02, 0.25388638E-02,
              0.10015321E-01, 0.40678948E-01,-0.17979136E-01,-0.35519518E-02,
             -0.52801091E-02, },
            { 0.23592529E-02,-0.10888859E-01,-0.72884997E-02,-0.15401221E-01,
             -0.82034264E-02, 0.22992514E-01, 0.55773426E-01, 0.39712242E-02,
              0.38829453E-02, },
            { 0.69172211E-01, 0.48699997E-01, 0.27636139E-01,-0.61428368E-01,
             -0.52963924E-02,-0.12855162E+00,-0.45400564E-01, 0.16215649E-02,
             -0.12359414E-01, },
            { 0.48297949E-01,-0.49748858E-02,-0.98275091E-03,-0.19215355E-01,
             -0.17720040E-01, 0.45080960E-01,-0.12306380E+00, 0.50776438E-02,
              0.61110142E-02, },
            { 0.41643456E-01,-0.11696455E+00, 0.58375955E-01, 0.44971615E-01,
             -0.28971106E-01,-0.11677378E+00,-0.12909587E+00, 0.26876843E-01,
             -0.87210005E-02, },
            {-0.27737666E-01,-0.31307798E-01,-0.67466116E-02, 0.54379940E-01,
             -0.14625589E-02, 0.13169987E+00,-0.15068077E+00, 0.17126866E-01,
             -0.85981060E-02, },
            {-0.20431230E-01,-0.40790266E-02, 0.63434639E-02, 0.17711798E-01,
              0.10949640E-01,-0.42260378E-02, 0.21843901E-02, 0.74000163E-02,
              0.10929640E-01, },
            {-0.27122835E-01, 0.13527165E-01,-0.41228514E-02,-0.12625435E-01,
             -0.37137333E-02,-0.90636835E-02, 0.47529996E-02,-0.73364242E-02,
              0.13638129E-01, },
            { 0.87972768E-01, 0.10461537E-01,-0.20852067E-01,-0.11302185E-01,
             -0.18262224E-01,-0.84132282E-02,-0.51975157E-02, 0.13170406E-02,
             -0.52717347E-01, },
            { 0.91586038E-02,-0.67112334E-01, 0.20436235E-01, 0.37548594E-01,
             -0.96599301E-02, 0.67724632E-02,-0.25683215E-02, 0.61321169E-01,
              0.49731391E-02, },
            { 0.30659314E-03,-0.44321292E-02, 0.46804580E-02,-0.64085319E-03,
              0.67579322E-02, 0.63220300E-02,-0.32938600E-02,-0.14760738E-01,
              0.72151292E-02, },
            { 0.10289571E-01, 0.13664791E-01,-0.10606593E-01,-0.77491556E-02,
             -0.10887925E-01, 0.14865423E-01, 0.28957613E-02,-0.37721093E-02,
             -0.77854278E-02, },
            {-0.13292247E-01, 0.18373661E-02, 0.13984727E-01, 0.48237410E-02,
             -0.23889195E-03, 0.75594457E-05,-0.27488936E-02,-0.73596179E-02,
              0.46203346E-02, },
            { 0.75970124E-02,-0.96835792E-02,-0.18875452E-01,-0.29838530E-02,
             -0.54170112E-02, 0.12925678E-01, 0.10607135E-02,-0.30239653E-02,
             -0.59676557E-02, },
         },
         {
            { 0.26920488E+01, 0.11389978E-01, 0.98317824E-02, 0.31620429E-02,
              0.71236417E-02, 0.11406111E-02,-0.56722080E-02,-0.48866012E-04,
              0.11990918E-03, },
            { 0.32997811E+00, 0.15148186E-01, 0.51400617E-01, 0.34719396E-01,
              0.39471131E-01, 0.79672048E-02,-0.34432359E-01,-0.70640777E-03,
              0.12702998E-01, },
            { 0.10057878E+01,-0.11234321E+00,-0.14979783E+00,-0.79824487E-02,
             -0.41622274E-01, 0.64391503E-02, 0.55315364E-01,-0.57274159E-01,
             -0.10790603E-01, },
            {-0.18790220E+00,-0.25727227E+00,-0.12212502E+00,-0.53840321E-01,
             -0.13131250E+00,-0.32435574E-01, 0.12784509E+00, 0.15400990E-01,
             -0.22077980E-01, },
            {-0.11739050E+01, 0.24142879E+00, 0.30385274E+00, 0.25107432E-02,
              0.12746825E-01,-0.26694337E-01,-0.99654526E-01, 0.15097810E+00,
              0.38390014E-01, },
            {-0.10511706E+00, 0.20301551E+00, 0.83070330E-01, 0.77093919E-02,
              0.10205883E+00, 0.23708317E-01,-0.99959821E-01,-0.15592851E-01,
              0.25066223E-02, },
            { 0.24683730E+00,-0.11758252E+00,-0.16027445E+00, 0.14066710E-02,
              0.35752703E-01, 0.23185330E-01, 0.48782002E-01,-0.95009811E-01,
             -0.37437104E-01, },
            { 0.88183937E-03,-0.30879867E+00,-0.27182159E+00, 0.20393897E-02,
              0.69514830E-02,-0.93122534E-02, 0.51981998E-02, 0.39402079E-02,
              0.44796327E-02, },
            { 0.87271491E-02, 0.28625375E+00,-0.31935695E+00, 0.52287388E-02,
             -0.67754048E-02, 0.86630229E-02,-0.77996161E-02,-0.60802847E-02,
              0.78002373E-02, },
            {-0.49862538E-01, 0.26634568E+00, 0.43451972E-01,-0.30126229E-01,
              0.80953494E-01,-0.31576782E-01, 0.30366642E-01, 0.42546973E-01,
              0.15590327E-01, },
            {-0.92389202E-02, 0.49252599E-01, 0.31745583E+00,-0.11614788E-01,
             -0.22583406E-01, 0.92642047E-01,-0.35509120E-02, 0.13614683E-01,
              0.56305062E-01, },
            { 0.99343164E-02, 0.11374707E+01, 0.70455754E+00, 0.17588671E-01,
             -0.75575434E-01, 0.13827311E+00,-0.49321633E-01,-0.75004660E-01,
             -0.13349947E+00, },
            {-0.42829427E+00,-0.84019780E+00, 0.13067011E+01, 0.34150391E-03,
              0.15719564E+00,-0.40955950E-01, 0.75470291E-01, 0.13132016E+00,
             -0.17925552E+00, },
            { 0.84138435E+00,-0.15912498E+01, 0.10511795E+00, 0.54206944E+00,
             -0.38688609E+00, 0.17358598E+00,-0.36708114E+00,-0.45672667E+00,
             -0.17131349E+00, },
            { 0.51538002E+00,-0.22939309E+00,-0.19646254E+01,-0.31394237E+00,
              0.19157244E+00,-0.59996855E+00,-0.45292720E-01,-0.13538666E+00,
             -0.45756468E+00, },
            {-0.15232117E+00,-0.15745953E+01, 0.19509153E+00, 0.83998501E-01,
              0.52878518E-01,-0.39462078E+00, 0.11502753E+00, 0.21379741E+00,
              0.39088184E+00, },
            { 0.13416333E+01,-0.28084823E+00,-0.21697867E+01, 0.53006239E-01,
             -0.35013482E+00, 0.56974604E-02,-0.25478348E+00,-0.39096361E+00,
              0.53477621E+00, },
            {-0.23757496E+01, 0.26613855E+01, 0.50327115E-01,-0.13042608E+01,
              0.67580837E+00,-0.25880238E+00, 0.89902413E+00, 0.10861776E+01,
              0.29366979E+00, },
            {-0.18501281E+01, 0.81386849E-01, 0.37214670E+01, 0.11096135E+01,
             -0.27120560E+00, 0.12532763E+01, 0.18778534E+00, 0.33471352E+00,
              0.96769452E+00, },
            { 0.14908600E+00, 0.10968636E+01,-0.32124093E+00,-0.12884712E+00,
              0.20785201E-01, 0.26191479E+00,-0.83107598E-01,-0.16051687E+00,
             -0.30176720E+00, },
            {-0.10500849E+01, 0.92894047E+00, 0.14747008E+01,-0.54342680E-01,
              0.21156456E+00, 0.44004083E-01, 0.19760658E+00, 0.30182552E+00,
             -0.39921561E+00, },
            { 0.17400059E+01,-0.12493542E+01, 0.42433235E+00, 0.91900015E+00,
             -0.38044804E+00, 0.54337408E-01,-0.64055729E+00,-0.72182900E+00,
             -0.13823685E+00, },
            { 0.16223973E+01,-0.39147210E+00,-0.21358540E+01,-0.86738002E+00,
              0.11883757E+00,-0.81355983E+00,-0.14370240E+00,-0.23106740E+00,
             -0.60141885E+00, },
            {-0.41002692E-02,-0.36998648E-01,-0.17324733E-01, 0.11522978E+00,
             -0.77249430E-01,-0.40560430E-02, 0.70232498E-02,-0.66030943E-02,
             -0.17369712E-02, },
            { 0.50536189E-02,-0.27059956E-01,-0.90134060E-02, 0.70231602E-01,
              0.11070350E+00,-0.50629838E-02, 0.32779330E-02,-0.49949071E-03,
              0.14835848E-02, },
            { 0.78637168E-01, 0.81079379E-01, 0.11170681E+00,-0.13736880E+00,
              0.11822002E+00,-0.20365615E-02, 0.56556158E-03, 0.83278008E-02,
             -0.18613596E-01, },
            {-0.28584436E-01,-0.58529172E-02,-0.46521258E-01,-0.11188855E+00,
             -0.86472198E-01, 0.11980531E-01, 0.47342207E-01, 0.28845016E-02,
              0.28383438E-01, },
            { 0.11538816E-01, 0.23314317E+00, 0.19854650E+00,-0.54327124E+00,
              0.38741905E-01, 0.24814070E-02,-0.65545565E-02, 0.54088697E-01,
             -0.54012053E-01, },
            {-0.19324923E+00, 0.67288898E-01,-0.59418734E-01,-0.33771223E-02,
             -0.44705462E+00,-0.62183533E-01, 0.37801322E-01, 0.20709286E-01,
             -0.37422135E-01, },
            {-0.37016487E+00,-0.48482874E+00,-0.33297902E+00, 0.26727724E+00,
             -0.59645987E+00, 0.24300680E-01,-0.60210197E-03,-0.69528729E-01,
              0.92760593E-01, },
            { 0.19207671E+00, 0.27005580E+00, 0.30531994E+00, 0.62059212E+00,
              0.20099737E+00, 0.89472935E-01,-0.26362386E+00,-0.48994768E-01,
             -0.22086474E+00, },
            { 0.96757943E-02,-0.35891131E+00,-0.41006809E+00, 0.39764428E+00,
             -0.12385339E+00,-0.45176540E-01,-0.23520408E-01,-0.71112022E-01,
              0.90213269E-01, },
            { 0.28137875E+00,-0.18450230E+00, 0.15791111E+00, 0.12017469E+00,
              0.28672829E+00, 0.10803044E+00,-0.11098100E+00,-0.24466727E-01,
              0.60332600E-01, },
            { 0.31297663E+00, 0.70864713E+00, 0.34029236E+00,-0.29005626E+00,
              0.79561400E+00, 0.39008945E-01, 0.32641575E-01, 0.97112492E-01,
             -0.11558538E+00, },
            {-0.27881745E+00,-0.29964480E+00,-0.33964044E+00,-0.84683907E+00,
             -0.30862889E+00,-0.19399051E+00, 0.30272609E+00, 0.92557348E-01,
              0.24453788E+00, },
            {-0.15125610E-01,-0.41656359E-02,-0.18647785E-03, 0.33162646E-02,
              0.50839242E-02, 0.39834607E-01,-0.29238209E-01,-0.27017531E-03,
             -0.12010454E-02, },
            {-0.41408944E-02,-0.13943053E-02, 0.43188082E-02,-0.89726988E-02,
             -0.38272680E-02, 0.31500991E-01, 0.41226041E-01, 0.43756245E-02,
             -0.18315912E-03, },
            { 0.28158762E-01, 0.54722499E-01, 0.17690169E-01,-0.28583935E-01,
             -0.11942321E-01,-0.10031212E+00,-0.44474870E-01, 0.96164370E-03,
             -0.99050319E-02, },
            { 0.20313956E-01, 0.83643049E-02, 0.24804203E-02,-0.18918399E-02,
             -0.15124847E-01, 0.43693494E-01,-0.10090125E+00, 0.51212185E-02,
              0.10049676E-02, },
            { 0.44255398E-01,-0.71331427E-01, 0.36873396E-01, 0.20218041E-01,
             -0.17782997E-01,-0.11550901E+00,-0.89868911E-01, 0.13469202E-01,
             -0.11585846E-01, },
            {-0.35603975E-02,-0.43277528E-01,-0.30855784E-01, 0.36533542E-01,
              0.49907360E-02, 0.10209156E+00,-0.10187620E+00, 0.46443972E-02,
              0.65317191E-02, },
            {-0.13696849E-01,-0.11126058E-01, 0.41133645E-02, 0.43337820E-02,
              0.92311008E-02,-0.39588860E-02, 0.16066917E-02, 0.65573328E-02,
              0.13953523E-01, },
            {-0.12618544E-01, 0.56935060E-02,-0.90064071E-02,-0.48857494E-02,
             -0.64519118E-02,-0.68315426E-02, 0.63782968E-02,-0.12648813E-01,
              0.11987502E-01, },
            { 0.40514901E-01, 0.76482454E-02,-0.16679959E-01,-0.22868705E-02,
             -0.15372105E-01,-0.84379911E-02, 0.35069739E-02, 0.13873799E-01,
             -0.56473475E-01, },
            { 0.20609386E-01,-0.42281467E-01, 0.22169946E-01, 0.14798193E-01,
             -0.78803375E-02,-0.38305379E-02, 0.14281864E-02, 0.57916559E-01,
              0.14591149E-01, },
            { 0.25254390E-02,-0.38154577E-02, 0.52015893E-02,-0.14573345E-02,
              0.38699114E-02, 0.55529494E-02,-0.52860053E-02,-0.12854257E-01,
              0.48506586E-02, },
            { 0.82307914E-02, 0.14202736E-01,-0.94194580E-02, 0.12523378E-02,
             -0.89414315E-02, 0.87305941E-02, 0.66426750E-02,-0.21206427E-02,
             -0.49932897E-02, },
            {-0.50855689E-02, 0.87064495E-02, 0.80340123E-02, 0.87542385E-02,
             -0.17597808E-02, 0.23883770E-03, 0.21871652E-03,-0.43078689E-02,
              0.17831495E-02, },
            { 0.10356435E-01,-0.30754986E-02,-0.14237600E-01, 0.38996674E-02,
             -0.63463412E-02, 0.10078620E-01, 0.10248976E-02, 0.13287502E-04,
             -0.20263768E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }


   double CCIR ::
   ccir22F2(int cond, int deg, int ord)
   {
      static const double F2[F2SolarActCond][F2MaxDegree][F2MaxOrder] {
         {
            { 0.58184547E+01, 0.57026275E-01,-0.92279583E-01, 0.19769210E-02,
             -0.63549969E-02, 0.41342468E-03,-0.16677534E-01, 0.29840269E-02,
              0.19395292E-01,-0.31526904E-02, 0.83086165E-02,-0.21525295E-01,
             -0.12466762E-02, },
            { 0.10098135E+01,-0.13401215E+00,-0.33394384E+00, 0.53832471E+00,
             -0.33256567E+00,-0.14199045E+00, 0.17917547E+00, 0.53553913E-01,
              0.36476406E+00,-0.17950524E+00, 0.83029037E-02, 0.52691013E-01,
             -0.10528484E+00, },
            { 0.10512217E+02,-0.11169873E+01, 0.69445834E-01,-0.11768398E+01,
              0.15896691E+01,-0.46408096E+00, 0.20050450E+00,-0.11498708E+00,
             -0.57279074E+00,-0.55499069E-01,-0.22052056E+00, 0.46984813E+00,
             -0.40550134E-02, },
            {-0.88619490E+01, 0.20526533E+01, 0.73109612E+01,-0.79561610E+01,
              0.10884787E+02, 0.16878909E+01,-0.36661744E+01,-0.14041631E+01,
             -0.40023403E+01, 0.28041875E+01, 0.11619287E+01,-0.47040588E+00,
              0.75334054E+00, },
            {-0.62287876E+02, 0.10522943E+02, 0.70986325E+00, 0.65067163E+01,
             -0.11479047E+02, 0.16810986E+01, 0.92432690E+00,-0.25512171E+00,
              0.36646883E+01,-0.23676340E+00, 0.22571785E-01,-0.15189878E+01,
              0.11747854E+01, },
            {-0.51741142E+02,-0.26478231E+02,-0.41005913E+02, 0.41161312E+02,
             -0.59304089E+02,-0.69479570E+01, 0.18469704E+02, 0.71893792E+01,
              0.17059683E+02,-0.13025754E+02,-0.89780149E+01, 0.14681500E+01,
              0.57374662E+00, },
            { 0.13726009E+03,-0.30932428E+02, 0.32160757E+01,-0.12439873E+02,
              0.30461220E+02,-0.47828910E+00,-0.68936372E+01, 0.25059500E+01,
             -0.91692266E+01, 0.18327304E+01, 0.26999156E+01, 0.10919596E+01,
             -0.47839489E+01, },
            { 0.24011169E+03, 0.88412567E+02, 0.86547073E+02,-0.93710808E+02,
              0.12796277E+03, 0.12891210E+02,-0.38673927E+02,-0.13718105E+02,
             -0.34361622E+02, 0.25967365E+02, 0.23587971E+02,-0.25956814E+01,
             -0.78749714E+01, },
            {-0.15224947E+03, 0.34043217E+02,-0.10899738E+02, 0.94556570E+01,
             -0.33112156E+02,-0.28386922E+01, 0.11093059E+02,-0.39166231E+01,
              0.96845121E+01,-0.29332969E+01,-0.52627816E+01, 0.10346909E+01,
              0.64608569E+01, },
            {-0.31515054E+03,-0.10496167E+03,-0.78569489E+02, 0.95572739E+02,
             -0.12133438E+03,-0.10754658E+02, 0.36431686E+02, 0.10801248E+02,
              0.32190319E+02,-0.23463760E+02,-0.25738005E+02, 0.27011609E+01,
              0.12155417E+02, },
            { 0.65124084E+02,-0.12487495E+02, 0.69259715E+01,-0.23780632E+01,
              0.12499528E+02, 0.21069338E+01,-0.53715777E+01, 0.17919514E+01,
             -0.36225002E+01, 0.14143248E+01, 0.27828605E+01,-0.10895599E+01,
             -0.28614464E+01, },
            { 0.13440363E+03, 0.41123314E+02, 0.26135244E+02,-0.35684410E+02,
              0.42126480E+02, 0.32188382E+01,-0.12763022E+02,-0.28629279E+01,
             -0.11248032E+02, 0.79024072E+01, 0.99840336E+01,-0.11786457E+01,
             -0.55230813E+01, },
            { 0.46197325E-01, 0.18396910E+01, 0.18488530E+01,-0.72137311E-01,
             -0.22965059E-01, 0.49167216E-01,-0.88025229E-02,-0.14172700E-02,
              0.23798379E-02, 0.10697534E-01,-0.86828216E-03,-0.11820203E-01,
             -0.45507951E-02, },
            {-0.34407914E+00,-0.19840239E+01, 0.18451532E+01,-0.71499117E-01,
             -0.29341163E-01,-0.38065258E-02, 0.12800088E-02, 0.35247207E-01,
             -0.29336326E-01,-0.28935510E-02, 0.34494635E-01, 0.21810072E-01,
             -0.35707127E-01, },
            {-0.14954211E+01, 0.16685070E+01, 0.26021440E+01,-0.25061376E-01,
             -0.11775342E+00, 0.18140525E+00, 0.39146900E+00,-0.52104045E-01,
              0.17152113E+00,-0.24431722E+00, 0.83082154E-01, 0.10321481E+00,
             -0.50523303E-01, },
            {-0.29371035E+01,-0.11318464E+01, 0.17996328E+01,-0.12266654E+01,
              0.44546545E+00,-0.17469050E-01, 0.10839576E+00, 0.13065223E+00,
             -0.14671960E+00,-0.17891817E-01, 0.82936779E-01,-0.53874258E-01,
              0.32858126E-01, },
            { 0.35966065E+01, 0.80477695E+01, 0.35369372E+01,-0.24907688E-01,
              0.49452472E+01,-0.24985828E+01, 0.10750397E+01, 0.63011825E-01,
             -0.19682705E+00, 0.48551191E-01,-0.24114858E+00, 0.44627540E-01,
              0.18766768E+00, },
            {-0.88128072E+00,-0.42972250E+01, 0.32859540E+01, 0.79657030E+00,
              0.30905221E+01, 0.28533414E-01,-0.50964904E+00,-0.72284150E+00,
              0.72344583E+00,-0.41555977E+00,-0.27640811E+00,-0.11061147E+01,
             -0.24223343E+00, },
            { 0.26371164E+02,-0.15519335E+02,-0.16091352E+02,-0.61467066E+01,
              0.11827186E+02,-0.18683293E+01,-0.10109256E+02,-0.41066766E+00,
             -0.26999035E+01, 0.11567605E+01,-0.66600835E+00,-0.19529310E+01,
              0.17002133E+00, },
            { 0.52747944E+02,-0.59905572E+01,-0.19065039E+02, 0.19325472E+02,
             -0.37808802E+01, 0.25225880E+01,-0.50770384E+00,-0.24967108E+01,
              0.18975933E+01, 0.15414751E+01, 0.46599278E+00, 0.26762016E+01,
             -0.23129802E+01, },
            {-0.29581627E+02,-0.67340530E+02, 0.28578785E+02,-0.18400332E+01,
             -0.45590832E+02, 0.19642672E+02,-0.11388482E+02,-0.14384048E+01,
              0.23746016E+01,-0.25869808E+01, 0.12232533E+01,-0.22738206E+00,
             -0.13948896E+01, },
            { 0.25422518E+02,-0.24761139E+02,-0.18647608E+02,-0.35908458E+01,
             -0.27576836E+02, 0.63180453E+00, 0.12264181E+01, 0.73593297E+01,
             -0.56782942E+01, 0.42023597E+01, 0.94908339E+00, 0.99610100E+01,
              0.74102955E+01, },
            {-0.13301096E+03, 0.29004534E+02, 0.96729050E+02, 0.52969921E+02,
             -0.57015472E+02, 0.43563184E+00, 0.61022591E+02,-0.60143486E-01,
              0.14871103E+02, 0.36482079E+01, 0.28649278E+01, 0.86638842E+01,
             -0.38194213E+01, },
            {-0.28812790E+03, 0.46469360E+02, 0.46227028E+02,-0.10875101E+03,
             -0.70202547E+00,-0.15389408E+02, 0.14878047E+01, 0.11239145E+02,
             -0.60284514E+01,-0.76406460E+01,-0.82917595E+01,-0.18456875E+02,
              0.17502781E+02, },
            { 0.90780685E+02, 0.17434750E+03,-0.17242982E+03, 0.74431643E+01,
              0.14255406E+03,-0.58141239E+02, 0.39145481E+02, 0.34620700E+01,
             -0.12430976E+02, 0.10840864E+02,-0.43649893E-01, 0.59502614E+00,
              0.22078130E+01, },
            {-0.10493087E+03, 0.18064041E+03, 0.12570059E+02, 0.93478651E+01,
              0.80625313E+02,-0.71573329E+00, 0.36804316E+01,-0.22530621E+02,
              0.18506762E+02,-0.12973787E+02,-0.15200816E+01,-0.31336437E+02,
             -0.28821777E+02, },
            { 0.29102225E+03,-0.17519779E+02,-0.29855054E+03,-0.15717352E+03,
              0.88891266E+02, 0.25252760E+02,-0.15087782E+03, 0.76023388E+01,
             -0.37483158E+02,-0.23462990E+02,-0.69535470E+01,-0.14723260E+02,
              0.18481215E+02, },
            { 0.65221857E+03,-0.11774995E+03,-0.61445251E+02, 0.26809067E+03,
              0.39994270E+02, 0.31199490E+02,-0.64933023E+01,-0.18023024E+02,
              0.64583664E+01, 0.12623926E+02, 0.27021164E+02, 0.45272427E+02,
             -0.46571922E+02, },
            {-0.11661733E+03,-0.21271249E+03, 0.27106604E+03,-0.73082743E+01,
             -0.18257878E+03, 0.73000687E+02,-0.54757530E+02,-0.16007293E+01,
              0.22605629E+02,-0.14365945E+02,-0.50904408E+01,-0.98563850E+00,
              0.61151135E+00, },
            { 0.15628786E+03,-0.29974243E+03, 0.10640742E+02,-0.10775609E+02,
             -0.95066612E+02,-0.36515400E+01,-0.11473593E+02, 0.27438049E+02,
             -0.24867657E+02, 0.15603787E+02, 0.12937165E+01, 0.39795895E+02,
              0.39510033E+02, },
            {-0.29016129E+03, 0.28823552E+01, 0.39856253E+03, 0.19402602E+03,
             -0.45091358E+02,-0.53616592E+02, 0.16660764E+03,-0.15165434E+02,
              0.42398968E+02, 0.36244057E+02, 0.75232472E+01, 0.95507259E+01,
             -0.28676897E+02, },
            {-0.64159631E+03, 0.12781446E+03, 0.57046253E+02,-0.29240640E+03,
             -0.71489220E+02,-0.24908289E+02, 0.13778471E+02, 0.10944452E+02,
             -0.66494685E+00,-0.74688768E+01,-0.32826694E+02,-0.46453716E+02,
              0.50829025E+02, },
            { 0.52968678E+02, 0.99466309E+02,-0.13427264E+03, 0.10737036E+01,
              0.81903267E+02,-0.32730858E+02, 0.26825422E+02,-0.69327599E+00,
             -0.12845870E+02, 0.59912968E+01, 0.45008116E+01, 0.76359075E+00,
             -0.18522649E+01, },
            {-0.80060051E+02, 0.15137988E+03,-0.81274099E+01, 0.41646919E+01,
              0.38812267E+02, 0.42529483E+01, 0.73921485E+01,-0.11849993E+02,
              0.11516630E+02,-0.64914279E+01,-0.51537508E+00,-0.17529139E+02,
             -0.18062862E+02, },
            { 0.10677107E+03, 0.52536047E+00,-0.18758057E+03,-0.85391655E+02,
              0.35786375E+00, 0.30250124E+02,-0.67851524E+02, 0.84298763E+01,
             -0.17312017E+02,-0.17777077E+02,-0.27602580E+01,-0.14503183E+01,
              0.14062240E+02, },
            { 0.22694662E+03,-0.50785782E+02,-0.25865496E+02, 0.11572225E+03,
              0.35236393E+02, 0.62496061E+01,-0.90702019E+01,-0.16877422E+01,
             -0.16750351E+01, 0.10357891E+01, 0.13722246E+02, 0.17021441E+02,
             -0.19445265E+02, },
            {-0.39433226E-01,-0.21142434E+00,-0.90724351E-02,-0.83350801E+00,
             -0.50699193E-01, 0.27291071E-01,-0.22223895E-01,-0.42128101E-01,
              0.17714033E-01, 0.50699499E-01,-0.18712536E-01,-0.25406671E-02,
              0.20049289E-01, },
            {-0.49931601E-01,-0.20490995E-01, 0.89790568E-01, 0.51811721E-01,
             -0.87662041E+00,-0.31257950E-01, 0.40552104E-02,-0.14721376E-01,
             -0.69894153E-02,-0.72437213E-02, 0.23188520E-01, 0.26874647E-01,
             -0.39877024E-01, },
            {-0.24794403E+00,-0.22231767E+00, 0.30068848E+00, 0.26426491E+00,
              0.85828662E+00, 0.70318066E-01,-0.19365221E+00,-0.67670614E-01,
             -0.44199413E+00,-0.11491100E+00,-0.18138404E+00, 0.90832096E-02,
              0.69318712E-02, },
            { 0.56248826E+00, 0.47823495E+00, 0.41573855E+00,-0.15119504E+01,
              0.67963108E-01,-0.13004091E+00, 0.64840458E-01,-0.91815062E-01,
             -0.15601902E+00,-0.78684457E-01,-0.90477422E-01, 0.76826736E-01,
             -0.65387972E-01, },
            { 0.36128531E+01, 0.25138476E+01, 0.35854378E+00, 0.26517680E+01,
              0.23037329E+01,-0.11147922E+01,-0.63770419E+00, 0.86852974E+00,
             -0.41232470E+00,-0.61307818E+00, 0.48212731E+00,-0.85765906E-01,
             -0.57749575E+00, },
            { 0.26269414E+01,-0.84287798E+00, 0.11433274E+01,-0.36928189E+01,
              0.54051328E+01, 0.74675167E+00,-0.73716527E+00,-0.66338859E-01,
              0.11560968E+01,-0.29391339E+00, 0.12733983E+00,-0.10972692E+01,
              0.59222376E+00, },
            { 0.34999065E+01, 0.44444332E+01,-0.94481927E+00,-0.35326798E+01,
             -0.27436817E+01, 0.56982708E+00, 0.76956064E-01, 0.31902781E+00,
              0.52106791E+01, 0.92866051E+00, 0.87636429E+00,-0.33926573E+00,
             -0.11445957E+01, },
            {-0.32386546E+01,-0.44973111E+01,-0.15676624E+01, 0.10158686E+02,
              0.16184865E+00, 0.16182853E+01,-0.21049386E+00,-0.28675815E+00,
              0.18338075E+01, 0.79065192E+00, 0.12128632E+01,-0.13835089E+01,
             -0.13644750E-01, },
            {-0.20781828E+02, 0.17002146E+01,-0.11406076E+02, 0.90912104E+01,
              0.13688050E+02, 0.40151911E+01, 0.44199939E+01,-0.25502865E+01,
              0.19699675E+01, 0.12704998E+01,-0.20803688E+01, 0.20040412E+01,
              0.38943079E+01, },
            {-0.19840216E+02, 0.19784559E+02,-0.86485300E+01,-0.49333038E+01,
             -0.39710073E+01,-0.40674095E+01, 0.28137231E+01, 0.38876519E+01,
             -0.90402184E+01, 0.19475243E+01,-0.10716798E+01, 0.62051377E+01,
             -0.18414352E+01, },
            {-0.83312197E+01,-0.13289581E+02,-0.21736002E+01, 0.84501009E+01,
              0.49111623E+00,-0.37395921E+01, 0.19115800E+01,-0.31825442E-01,
             -0.13818624E+02,-0.18343800E+01,-0.11181920E+01, 0.80240744E+00,
              0.43642979E+01, },
            { 0.33570621E+01, 0.12437586E+02, 0.45811825E+01,-0.16603788E+02,
              0.22749799E+00,-0.40944538E+01,-0.53237623E+00, 0.23649323E+01,
             -0.41487660E+01,-0.17448486E+01,-0.36165757E+01, 0.43271375E+01,
              0.10559268E+01, },
            { 0.39056721E+02,-0.16090921E+02, 0.33891582E+02,-0.35210052E+02,
             -0.50438084E+02,-0.49294519E+01,-0.90522985E+01, 0.19701451E+01,
             -0.41518707E+01,-0.19429389E+00, 0.29413283E+01,-0.64302478E+01,
             -0.75884647E+01, },
            { 0.49585617E+02,-0.54486118E+02, 0.14613009E+02, 0.31883018E+02,
             -0.19175350E+02, 0.78750348E+01,-0.41738634E+01,-0.11004766E+02,
              0.19877577E+02,-0.26470737E+01, 0.21634009E+01,-0.10771997E+02,
              0.26160250E+01, },
            { 0.42416420E+01, 0.10615758E+02, 0.38284779E+01,-0.26582875E+01,
              0.61848521E+01, 0.39846966E+01,-0.23324945E+01,-0.27493328E+00,
              0.98930330E+01, 0.91430855E+00, 0.42881787E+00,-0.45336166E+00,
             -0.38754070E+01, },
            { 0.24860692E+01,-0.11103153E+02,-0.55986381E+01, 0.30626101E+01,
              0.12912285E+01, 0.31475737E+01, 0.94185495E+00,-0.24471545E+01,
              0.26365435E+01, 0.11146898E+01, 0.30409727E+01,-0.34792864E+01,
             -0.11173525E+01, },
            {-0.24812998E+02, 0.12293331E+02,-0.24448639E+02, 0.25972631E+02,
              0.41211086E+02, 0.13715389E+01, 0.59663386E+01,-0.71985565E-01,
              0.29825644E+01,-0.41931191E+00,-0.13509992E+01, 0.51593585E+01,
              0.43710918E+01, },
            {-0.40055676E+02, 0.40998569E+02,-0.62064986E+01,-0.28469561E+02,
              0.22959063E+02,-0.49932957E+01, 0.19243966E+01, 0.76674228E+01,
             -0.12809521E+02, 0.61497426E+00,-0.14364125E+01, 0.57304015E+01,
             -0.13400353E+01, },
            {-0.21295029E+00,-0.11400111E+00,-0.14825793E+00, 0.73622137E-01,
              0.64809769E-01,-0.29826665E+00,-0.60644519E+00,-0.13423300E-01,
              0.10754849E-01, 0.98052928E-02,-0.27152034E-02,-0.18311009E-01,
              0.64233541E-02, },
            {-0.17358117E-01, 0.90460517E-01,-0.55983938E-01,-0.37680583E-02,
              0.10464079E+00, 0.60645211E+00,-0.32411674E+00,-0.19527392E-01,
              0.57433564E-01,-0.14313875E-01, 0.14264599E-01,-0.12788296E-01,
              0.12462893E-01, },
            { 0.72629370E-01, 0.26425889E+00,-0.72310239E-01, 0.14027588E+00,
             -0.22800760E+00,-0.82698578E+00,-0.59944493E+00, 0.41647565E-01,
              0.10132378E+00, 0.10920783E+00,-0.92539199E-01, 0.54342553E-01,
              0.23273796E-01, },
            {-0.79490080E-01,-0.41849073E-01, 0.11703410E+00, 0.22500437E+00,
              0.79866014E-01, 0.79547960E+00,-0.68940175E+00,-0.91365695E-01,
              0.26474414E-01, 0.18633839E-01,-0.69830693E-01, 0.42015348E-01,
              0.70129656E-02, },
            { 0.20651813E+01, 0.10757015E+01, 0.79591483E+00,-0.10709347E+01,
             -0.39959517E+00, 0.12449454E+01, 0.65018201E+00, 0.57454801E+00,
             -0.10787249E+00,-0.11827233E+00, 0.89393541E-01, 0.38926962E+00,
              0.20278770E-01, },
            {-0.84885871E+00, 0.50201140E-01, 0.41724554E+00, 0.10930064E+00,
             -0.19860851E+01,-0.40959147E+00, 0.14579872E+01, 0.10777475E+00,
              0.18702915E+00, 0.58261946E-01,-0.15377997E+00, 0.27000561E+00,
             -0.32803483E-01, },
            { 0.41750300E+00,-0.89985055E+00, 0.11675123E+00, 0.33188909E-01,
              0.56400406E+00, 0.12467124E+01, 0.11199267E+01,-0.13293042E-01,
             -0.56087691E-01,-0.28695655E+00, 0.14087248E+00,-0.16224995E+00,
             -0.25811439E-01, },
            { 0.51898807E+00, 0.17738800E+00,-0.20662893E+00,-0.74402356E+00,
             -0.51481444E-02,-0.16159163E+01, 0.10972348E+01, 0.44918306E-01,
             -0.20699680E+00,-0.11631241E-01, 0.20945826E+00,-0.16005199E+00,
             -0.11305244E+00, },
            {-0.41412516E+01,-0.14561827E+01,-0.11091290E+01, 0.17685617E+01,
              0.66921726E-01,-0.16768463E+01, 0.23649099E+00,-0.10593188E+01,
              0.12860221E+00, 0.10053814E+00,-0.29987147E+00,-0.75887227E+00,
              0.55327021E-01, },
            { 0.18062499E+01, 0.15113790E+00,-0.36903124E-01, 0.28656241E+00,
              0.33028457E+01,-0.43000481E+00,-0.17351261E+01,-0.96321538E-01,
             -0.62929451E+00,-0.54888729E-01, 0.19858769E+00,-0.39517498E+00,
             -0.23517720E-01, },
            { 0.74075326E-01, 0.55286419E-01, 0.37950128E-01, 0.51930584E-01,
              0.40860217E-01,-0.80779605E-02,-0.85725700E-02, 0.18911599E+00,
             -0.20170498E+00, 0.79165380E-02, 0.14078500E-01,-0.48967579E-03,
              0.23352634E-01, },
            { 0.89454815E-01, 0.45219720E-02, 0.19925103E-01,-0.36571357E-01,
              0.47071739E-02,-0.22226086E-01, 0.12251128E-01, 0.23481008E+00,
              0.19869673E+00,-0.52124937E-02,-0.27196895E-01, 0.11811548E-01,
              0.11335118E-01, },
            { 0.20631571E+00,-0.40896397E-01,-0.66134050E-01,-0.51769245E-01,
              0.11315651E+00,-0.29828817E-01,-0.70822775E-01,-0.71775690E-01,
             -0.42620280E+00,-0.58472637E-01,-0.11232210E-01,-0.11249758E-01,
              0.20316778E-01, },
            { 0.17420180E-01,-0.14574118E+00,-0.10202885E+00, 0.72233818E-01,
              0.11322709E-01,-0.20416332E-01,-0.24119224E-02, 0.33355662E+00,
             -0.65747298E-01,-0.34695771E-02,-0.40769741E-01, 0.66911586E-01,
             -0.37780598E-01, },
            { 0.43868124E-01, 0.14381094E+00,-0.61921760E-01,-0.12998655E-01,
              0.31560197E-01, 0.31114614E-02, 0.38460427E-03,-0.28693313E-01,
              0.34101967E-01, 0.90243518E-01, 0.18531388E+00,-0.13973088E-01,
              0.10807764E-01, },
            { 0.87856077E-01, 0.35964809E-01, 0.23326820E-01,-0.84922142E-01,
             -0.19689238E-01, 0.94046257E-02, 0.14910235E-02, 0.26861498E-01,
             -0.73656917E-01,-0.17092833E+00, 0.78958347E-01, 0.87024560E-02,
              0.12014927E-01, },
            { 0.82137808E-02,-0.21466317E-01, 0.53214639E-01,-0.74789733E-01,
              0.19152014E-01, 0.21119369E-01, 0.25541728E-01, 0.88134193E-03,
             -0.14540546E-01, 0.78659160E-02, 0.15383034E-01,-0.64573243E-01,
              0.21328622E-01, },
            { 0.28671172E-01, 0.20090424E-01,-0.94851188E-01,-0.24305644E-01,
             -0.72831862E-01,-0.18191807E-01,-0.27933948E-01, 0.21245504E-01,
             -0.38015399E-01, 0.13354953E-01, 0.61733276E-02, 0.53486875E-02,
             -0.67182191E-01, },
            { 0.49329074E-02, 0.14485019E-01,-0.57947930E-01,-0.16713040E-01,
              0.52156751E-02, 0.26821880E-01,-0.57450021E-02,-0.63059921E-02,
             -0.14616561E-01,-0.94541983E-03,-0.15653398E-01,-0.63756146E-02,
              0.19463095E-02, },
            { 0.91823854E-01, 0.34876380E-01,-0.81078401E-02, 0.40082075E-01,
              0.11814687E-01,-0.21329770E-01, 0.22458946E-01, 0.65466752E-02,
              0.32364109E-02,-0.13744369E-01, 0.66302335E-02, 0.16155699E-01,
             -0.14738605E-01, },
            { 0.45109160E-01, 0.52861501E-01,-0.88001676E-01, 0.24178792E-01,
             -0.26224427E-01,-0.27569726E-01, 0.28009566E-02,-0.12213375E-01,
             -0.14918844E-02, 0.38230043E-01,-0.10143310E-01, 0.97960318E-02,
              0.72819605E-02, },
            {-0.12117780E+00, 0.23717241E-01,-0.16449519E-01, 0.21570399E-01,
             -0.11265770E-02, 0.23510948E-01, 0.17620750E-01,-0.14561802E-01,
              0.44145947E-02,-0.19181985E-01, 0.15150579E-01, 0.23060735E-01,
             -0.34930815E-02, },
         },
         {
            { 0.85668230E+01, 0.58971159E-01,-0.37359368E-01, 0.98172938E-02,
              0.60634501E-02,-0.19980328E-01,-0.32437317E-01, 0.73017133E-02,
              0.21622829E-01,-0.31204394E-02, 0.86839320E-02,-0.15189710E-01,
              0.15739934E-02, },
            { 0.15422637E+01,-0.45887977E+00,-0.63496351E+00, 0.17954226E+00,
             -0.36354408E+00,-0.81500508E-01, 0.53280359E-02, 0.13243176E+00,
              0.38387778E+00,-0.11102082E+00,-0.78610063E-01, 0.94472550E-01,
             -0.58259502E-01, },
            { 0.24310478E+02,-0.11059129E+01,-0.73916423E+00,-0.92551410E+00,
              0.17832867E+01, 0.35105866E+00, 0.37704727E+00,-0.31123796E+00,
             -0.53970373E+00, 0.86808112E-02,-0.72665714E-01, 0.46467739E+00,
             -0.98924175E-01, },
            { 0.20755560E+02, 0.67352271E+01, 0.14065329E+02,-0.25990703E+01,
              0.10409377E+02, 0.16016326E+01,-0.47564018E+00,-0.34755881E+01,
             -0.49684534E+01, 0.15748028E+01, 0.26321311E+01,-0.12655041E+01,
              0.15727635E+00, },
            {-0.12931139E+03, 0.12252737E+02, 0.60598216E+01, 0.50476084E+01,
             -0.11773267E+02,-0.44770432E+01, 0.58385706E+00, 0.23724539E+01,
              0.37232220E+01,-0.10170099E+01,-0.81260860E-01,-0.29869394E+01,
              0.22409637E+01, },
            {-0.23669112E+03,-0.51905815E+02,-0.72058403E+02, 0.12511720E+02,
             -0.56770985E+02,-0.10931894E+02, 0.33293872E+01, 0.20718521E+02,
              0.24318001E+02,-0.57881908E+01,-0.15459870E+02, 0.46247511E+01,
              0.34875414E+01, },
            { 0.24901990E+03,-0.39882587E+02,-0.83957424E+01,-0.10404231E+02,
              0.27809801E+02, 0.16726303E+02,-0.68507929E+01,-0.69119296E+01,
             -0.97388639E+01, 0.47765851E+01, 0.10716473E+01, 0.72030778E+01,
             -0.82405844E+01, },
            { 0.65161255E+03, 0.14923309E+03, 0.14101414E+03,-0.29154617E+02,
              0.12519636E+03, 0.29351564E+02,-0.84194279E+01,-0.48260624E+02,
             -0.53140247E+02, 0.90673313E+01, 0.35299232E+02,-0.72856526E+01,
             -0.14410473E+02, },
            {-0.23240266E+03, 0.45167839E+02,-0.96814597E+00, 0.99426565E+01,
             -0.26336687E+02,-0.23381857E+02, 0.11163598E+02, 0.86249828E+01,
              0.10363331E+02,-0.69799447E+01,-0.19014372E+01,-0.71312876E+01,
              0.10706991E+02, },
            {-0.70311969E+03,-0.17120168E+03,-0.11967136E+03, 0.32664330E+02,
             -0.12050853E+03,-0.33125164E+02, 0.86441698E+01, 0.48691650E+02,
              0.51921913E+02,-0.65182605E+01,-0.34950245E+02, 0.55114870E+01,
              0.18803883E+02, },
            { 0.85038689E+02,-0.16290762E+02, 0.40023508E+01,-0.37510366E+01,
              0.83846464E+01, 0.10919796E+02,-0.52057114E+01,-0.38280964E+01,
             -0.38365119E+01, 0.32329717E+01, 0.99229783E+00, 0.24394760E+01,
             -0.46375208E+01, },
            { 0.26612119E+03, 0.67721451E+02, 0.37346931E+02,-0.13712147E+02,
              0.41968533E+02, 0.13200320E+02,-0.30124371E+01,-0.17847174E+02,
             -0.18521812E+02, 0.17858783E+01, 0.12569484E+02,-0.17031465E+01,
             -0.80002728E+01, },
            {-0.13380323E+00, 0.13366905E+01, 0.19891349E+01,-0.10753992E-01,
              0.60303468E-01, 0.12918174E-01, 0.21262087E-01,-0.86736670E-02,
             -0.13964107E-01, 0.11481032E-01, 0.10142981E-01, 0.72076609E-02,
             -0.85020000E-02, },
            {-0.14243269E-01,-0.21352170E+01, 0.13255059E+01,-0.50178140E-01,
             -0.54384332E-01, 0.46576783E-01, 0.34636784E-01, 0.48959581E-02,
             -0.16821539E-01, 0.46431450E-02, 0.12051069E-01, 0.19954719E-01,
             -0.32330859E-01, },
            {-0.12649831E+01, 0.14815663E+01, 0.32681105E+01,-0.28774256E+00,
              0.42832842E+00, 0.53366369E+00, 0.69917524E+00,-0.12640499E+00,
              0.27914149E+00,-0.86231783E-01, 0.41828956E-01, 0.22610457E+00,
             -0.11857753E+00, },
            {-0.59289962E+00,-0.26590743E+01, 0.71710646E+00,-0.80467057E+00,
              0.91069084E+00, 0.17542484E+00, 0.10409195E+00,-0.57189096E-01,
              0.43423317E-01,-0.84960863E-01,-0.95633030E-01, 0.33903435E-01,
             -0.36507253E-01, },
            { 0.83079939E+01, 0.22010695E+02,-0.11524427E+02,-0.11183747E+00,
              0.25016727E+01,-0.13266135E+01, 0.22817478E+00, 0.46875328E+00,
              0.34386232E+00, 0.40068373E+00,-0.31280607E+00,-0.36988357E+00,
              0.16799913E+00, },
            {-0.83424835E+01, 0.11889293E+02, 0.16917753E+02, 0.39813349E+00,
              0.21293635E+01,-0.16580629E+00,-0.91171318E+00,-0.19491872E-01,
              0.62579691E+00,-0.31269869E+00,-0.14522402E+00,-0.69244891E+00,
             -0.35517254E+00, },
            { 0.23320499E+02, 0.19454935E+01,-0.32308598E+02,-0.22719015E+00,
             -0.26350050E+01,-0.75710707E+01,-0.13849431E+02, 0.13357112E+01,
             -0.52987828E+01,-0.12208441E+01, 0.56600362E+00,-0.38908503E+01,
              0.15157390E+01, },
            { 0.16239506E+02, 0.26882658E+02, 0.11697211E+02, 0.13142451E+02,
             -0.13128712E+02, 0.12815480E+01,-0.26485260E+01, 0.10252703E+01,
             -0.71769851E+00, 0.26684380E+01, 0.33325875E+01, 0.19368449E+00,
             -0.84864247E+00, },
            {-0.65441612E+02,-0.12118099E+03, 0.10928064E+03,-0.38240664E+01,
             -0.23437611E+02, 0.10801532E+02,-0.57041898E+01,-0.27291293E+01,
             -0.15678360E+01,-0.73939724E+01, 0.37768626E+00, 0.33962661E+00,
             -0.93000340E+00, },
            { 0.93348366E+02,-0.10296486E+03,-0.70953468E+02,-0.35590768E+00,
             -0.17237814E+02, 0.94308928E-01, 0.38772573E+01, 0.26037455E+01,
             -0.46013012E+01, 0.14253412E+01, 0.20683482E+01, 0.66785064E+01,
              0.57986798E+01, },
            {-0.11509787E+03,-0.40946972E+02, 0.20788942E+03, 0.28071583E+02,
              0.24814009E+02, 0.32777500E+02, 0.80104485E+02,-0.11489333E+02,
              0.29405239E+02, 0.16883720E+02,-0.67208619E+01, 0.20167179E+02,
             -0.86552134E+01, },
            {-0.99561371E+02,-0.15522029E+03,-0.83673103E+02,-0.80494255E+02,
              0.58310280E+02,-0.12448786E+02, 0.17183468E+02,-0.89103413E+01,
              0.53705993E+01,-0.13733030E+02,-0.22880342E+02,-0.23245771E+01,
              0.86334391E+01, },
            { 0.17955624E+03, 0.24499542E+03,-0.33905188E+03, 0.16499292E+02,
              0.74469521E+02,-0.33439671E+02, 0.23443146E+02, 0.36955311E+01,
             -0.19126467E+00, 0.27561241E+02, 0.56984258E+01, 0.44640994E+01,
              0.42115948E+00, },
            {-0.30066251E+03, 0.29662677E+03, 0.81036797E+02,-0.27805751E-01,
              0.43939407E+02, 0.27983515E+01,-0.60298972E+01,-0.10772779E+02,
              0.15233603E+02,-0.58941245E+00,-0.87549553E+01,-0.21512075E+02,
             -0.19019295E+02, },
            { 0.25668509E+03, 0.95829285E+02,-0.60892999E+03,-0.11896392E+03,
             -0.95018196E+02,-0.59214664E+02,-0.19752927E+03, 0.38122326E+02,
             -0.69384659E+02,-0.56728783E+02, 0.22776094E+02,-0.43739380E+02,
              0.22105671E+02, },
            { 0.22645404E+03, 0.37045316E+03, 0.17037305E+03, 0.21269356E+03,
             -0.11409075E+03, 0.32559406E+02,-0.47628788E+02, 0.30940620E+02,
             -0.16595905E+02, 0.25164560E+02, 0.59392143E+02, 0.48828721E+01,
             -0.25786690E+02, },
            {-0.19930544E+03,-0.22929100E+03, 0.46098517E+03,-0.23051146E+02,
             -0.99358955E+02, 0.43438587E+02,-0.36220837E+02, 0.10952334E+01,
              0.52028384E+01,-0.36286861E+02,-0.15389333E+02,-0.10731896E+02,
              0.32523794E+01, },
            { 0.36962787E+03,-0.36779376E+03,-0.39542661E+01,-0.51138026E+00,
             -0.43776100E+02,-0.79711547E+01, 0.35742667E+01, 0.16140493E+02,
             -0.21209721E+02,-0.38114667E+01, 0.13219823E+02, 0.27240120E+02,
              0.23259232E+02, },
            {-0.26610745E+03,-0.69780006E+02, 0.81476440E+03, 0.17288275E+03,
              0.13693073E+03, 0.46457741E+02, 0.21848393E+03,-0.50798943E+02,
              0.72315247E+02, 0.74823509E+02,-0.31317862E+02, 0.41232723E+02,
             -0.24342672E+02, },
            {-0.20460796E+03,-0.41652863E+03,-0.12226570E+03,-0.24583488E+03,
              0.10287257E+03,-0.36260223E+02, 0.59697350E+02,-0.42112240E+02,
              0.20332176E+02,-0.18127939E+02,-0.65813683E+02,-0.29394605E+01,
              0.30914236E+02, },
            { 0.77954071E+02, 0.84637459E+02,-0.22712354E+03, 0.10199258E+02,
              0.47287090E+02,-0.19806791E+02, 0.19231525E+02,-0.30108094E+01,
             -0.38124545E+01, 0.15798557E+02, 0.10139494E+02, 0.67266974E+01,
             -0.32354667E+01, },
            {-0.15642979E+03, 0.16548746E+03,-0.24017464E+02, 0.17524619E+00,
              0.14203983E+02, 0.56363974E+01,-0.46004733E+00,-0.83661394E+01,
              0.10115620E+02, 0.35576582E+01,-0.64790502E+01,-0.11826179E+02,
             -0.96805029E+01, },
            { 0.10178647E+03, 0.11225664E+02,-0.39543845E+03,-0.84101143E+02,
             -0.65480949E+02,-0.12942043E+02,-0.88712433E+02, 0.23320219E+02,
             -0.27256409E+02,-0.34397594E+02, 0.15014293E+02,-0.13734142E+02,
              0.94665298E+01, },
            { 0.60337921E+02, 0.17995731E+03, 0.21066154E+02, 0.10197303E+03,
             -0.35869057E+02, 0.15114552E+02,-0.27445200E+02, 0.19399147E+02,
             -0.83749247E+01, 0.40309701E+01, 0.26498037E+02,-0.13782138E-01,
             -0.13060991E+02, },
            {-0.10923450E-01,-0.34306788E+00,-0.50401580E-01,-0.53487885E+00,
              0.29230735E+00, 0.21476941E-01, 0.15144962E-01,-0.43948866E-01,
             -0.17394183E-01, 0.35130534E-01,-0.16292399E-01, 0.19373255E-01,
             -0.59662841E-03, },
            {-0.71579292E-01,-0.52842975E-01,-0.63118017E-02,-0.39636719E+00,
             -0.73986328E+00, 0.37425833E-02, 0.59765212E-01,-0.49656001E-02,
             -0.27551942E-02,-0.15596611E-01, 0.61703688E-02, 0.14522822E-01,
             -0.32979961E-01, },
            {-0.84795365E-02, 0.16810313E-02,-0.11047077E+00,-0.18073320E+01,
              0.75538784E+00, 0.17500804E+00,-0.17733023E+00,-0.16572930E+00,
             -0.33800665E+00,-0.15694118E+00,-0.16335554E+00, 0.71416780E-01,
             -0.57195537E-01, },
            { 0.38296199E+00, 0.77749258E+00,-0.32756321E-01,-0.16789720E+01,
             -0.18577414E+01, 0.17541359E+00,-0.14601539E-01, 0.33322370E-02,
              0.29737126E-01,-0.87182283E-01,-0.31497914E-01, 0.12684085E-01,
             -0.11295188E-01, },
            { 0.23203259E+01, 0.53199558E+01, 0.16164668E+01,-0.67294307E+01,
              0.70609212E+01,-0.22695801E+00,-0.15492891E+01, 0.11764526E+01,
              0.21189202E+00,-0.77632385E+00, 0.94330899E-01,-0.52504981E+00,
              0.80921426E-01, },
            { 0.32638378E+01,-0.18616734E+00, 0.25722713E+01,-0.63221340E+01,
             -0.21414373E+01, 0.87206262E+00,-0.12179788E+01, 0.13763864E+00,
              0.67562330E+00,-0.19984256E+00, 0.40309739E+00,-0.78906339E+00,
              0.50545210E+00, },
            { 0.94253534E+00, 0.12683750E+01, 0.62755954E+00, 0.30156040E+01,
              0.10006471E+01, 0.20007719E+00, 0.20150834E+00, 0.14695404E+01,
              0.41846933E+01, 0.12572116E+01, 0.65958399E+00,-0.71903515E+00,
             -0.69292140E+00, },
            {-0.96949649E+00,-0.45164185E+01, 0.29109006E+01, 0.85274162E+01,
              0.41857915E+01,-0.27851799E+00, 0.91058570E+00,-0.63607663E+00,
             -0.55645533E-01, 0.10949482E+01, 0.38305825E+00,-0.72221702E+00,
             -0.53886354E+00, },
            {-0.11483661E+02,-0.18448803E+02,-0.25420229E+02, 0.41966877E+02,
             -0.18985954E+02,-0.14435635E+01, 0.80026455E+01,-0.42318783E+01,
             -0.10838451E+01, 0.28619409E+01, 0.61609662E+00, 0.37960072E+01,
             -0.52905315E+00, },
            {-0.17035336E+02, 0.18621317E+02,-0.21966009E+02, 0.18735846E+02,
              0.16747011E+02,-0.54307652E+01, 0.46619482E+01, 0.13719676E+01,
             -0.60610504E+01, 0.19186600E+01,-0.33431869E+01, 0.48622260E+01,
             -0.19352443E+01, },
            {-0.21383023E+01,-0.43643160E+01,-0.42401314E+01,-0.24073770E+01,
             -0.16128111E+02,-0.27644095E+01, 0.22688612E+00,-0.28903916E+01,
             -0.11218066E+02,-0.23697209E+01,-0.47793251E+00, 0.14431590E+01,
              0.33283553E+01, },
            {-0.29100385E+01, 0.97868328E+01,-0.62205763E+01,-0.72675920E+01,
             -0.31473615E+01, 0.19133319E+00,-0.38154783E+01, 0.26445303E+01,
             -0.82178062E+00,-0.30948792E+01,-0.84835994E+00, 0.25783031E+01,
              0.23090553E+01, },
            { 0.19434116E+02, 0.30117950E+02, 0.67169952E+02,-0.83444626E+02,
              0.97705097E+01, 0.59132814E+01,-0.15886705E+02, 0.57373362E+01,
              0.21024210E+01,-0.35821984E+01,-0.25029049E+01,-0.88817778E+01,
              0.14393607E+01, },
            { 0.31481159E+02,-0.55233753E+02, 0.44615513E+02,-0.14347218E+02,
             -0.38026085E+02, 0.13057509E+02,-0.83902435E+01,-0.42682991E+01,
              0.14140669E+02,-0.35403111E+01, 0.79202223E+01,-0.92711287E+01,
              0.31921399E+01, },
            {-0.40178025E+00, 0.42423429E+01, 0.44495778E+01, 0.65041461E+01,
              0.24042871E+02, 0.30938346E+01,-0.44691342E+00, 0.15059996E+01,
              0.77578301E+01, 0.13139611E+01,-0.14930587E+00,-0.67745137E+00,
             -0.30641334E+01, },
            { 0.71513062E+01,-0.80155268E+01, 0.27227871E+01,-0.77887855E+01,
              0.47823057E+01,-0.47966767E-01, 0.38374758E+01,-0.23107648E+01,
              0.14483806E+01, 0.24450035E+01, 0.82637250E+00,-0.23080869E+01,
             -0.19962353E+01, },
            {-0.12261738E+02,-0.18203463E+02,-0.47682377E+02, 0.55126305E+02,
              0.94886637E+01,-0.50771351E+01, 0.10601325E+02,-0.30876780E+01,
             -0.16409191E+01, 0.16010164E+01, 0.18693397E+01, 0.63582640E+01,
             -0.12191840E+01, },
            {-0.22224463E+02, 0.43476688E+02,-0.25098253E+02,-0.41798453E+01,
              0.30020361E+02,-0.10327440E+02, 0.55453873E+01, 0.28609169E+01,
             -0.91859627E+01, 0.17712349E+01,-0.55576038E+01, 0.54289842E+01,
             -0.16860573E+01, },
            {-0.18644652E+00,-0.13026768E+00,-0.51389337E-01, 0.13972405E+00,
              0.26259443E-01,-0.48280844E+00,-0.59422386E+00,-0.21665383E-01,
             -0.14493400E-01, 0.15758840E-01,-0.17352704E-02,-0.41003264E-02,
              0.18915856E-02, },
            { 0.57341747E-01, 0.42945143E-01, 0.96227340E-01,-0.34298085E-01,
              0.13546824E+00, 0.63187897E+00,-0.47779465E+00, 0.20380679E-02,
              0.22282990E-01,-0.25250018E-01, 0.17265640E-01,-0.25108820E-01,
              0.94693340E-02, },
            {-0.25878796E+00, 0.28488120E+00,-0.16580707E+00, 0.12818602E+00,
              0.35997484E-01,-0.64699262E+00,-0.10717692E+01, 0.10131285E+00,
              0.16753370E+00, 0.32581434E-01,-0.11407858E+00, 0.59960652E-01,
             -0.77041015E-02, },
            {-0.41792631E-01, 0.13099888E+00, 0.21938293E+00,-0.11784480E+00,
              0.17003253E+00, 0.11506057E+01,-0.71615899E+00,-0.11056098E+00,
              0.15374933E+00, 0.68769425E-01, 0.45401078E-01, 0.56390647E-01,
              0.55900361E-01, },
            { 0.28251824E+01, 0.44229245E+00,-0.81067193E+00,-0.12398359E+01,
             -0.28553767E-01, 0.22159548E+01,-0.54317749E+00, 0.75851673E+00,
              0.27122670E+00,-0.17409934E+00, 0.50550703E-01, 0.21512508E+00,
              0.78091890E-01, },
            {-0.14035779E+01, 0.84380841E+00,-0.64104086E+00, 0.46829499E-01,
             -0.14327688E+01, 0.25394753E+00, 0.19480610E+01,-0.91148943E-01,
              0.78488475E+00, 0.22460079E+00,-0.24968466E+00, 0.46386969E+00,
             -0.64009469E-03, },
            { 0.98801070E+00,-0.11307433E+01, 0.52810138E+00, 0.87130740E-01,
              0.50161190E-01,-0.13517044E+00, 0.17106220E+01,-0.20004056E+00,
             -0.23602618E+00,-0.27457481E-01, 0.24470349E+00,-0.14111978E+00,
              0.38745444E-01, },
            { 0.47567675E+00,-0.22344865E+00,-0.75762641E+00,-0.57572670E-01,
              0.84235653E-01,-0.19831989E+01, 0.25760263E+00, 0.19424839E+00,
             -0.48381361E+00,-0.13142878E+00,-0.88607289E-01,-0.23240808E+00,
             -0.17395243E+00, },
            {-0.52569547E+01, 0.56680167E+00, 0.16630803E+01, 0.17917036E+01,
             -0.68722159E+00,-0.39917276E+01, 0.18925166E+01,-0.13140194E+01,
             -0.42374143E+00, 0.26679030E+00,-0.19116408E+00,-0.40941295E+00,
             -0.66891648E-01, },
            { 0.19922812E+01,-0.13392367E+01, 0.13449210E+01, 0.34894937E+00,
              0.20699844E+01,-0.15274907E+01,-0.31885908E+01, 0.30454439E+00,
             -0.15603627E+01,-0.34676239E+00, 0.25211805E+00,-0.71817315E+00,
             -0.18247532E-01, },
            { 0.26140720E-01, 0.61536264E-01, 0.58766458E-01, 0.60080968E-01,
              0.25750723E-01,-0.20359196E-01, 0.31155759E-01, 0.21450184E+00,
             -0.28345245E+00, 0.30602589E-01, 0.73734480E-02,-0.73152762E-04,
              0.24243210E-01, },
            { 0.19760441E-01, 0.36050626E-02, 0.31682935E-01,-0.42440861E-01,
              0.21866074E-01,-0.36474712E-01,-0.28035508E-02, 0.29943243E+00,
              0.22910598E+00, 0.11597273E-01,-0.11098400E-01, 0.64757066E-02,
              0.79353563E-02, },
            { 0.14553243E+00,-0.74297085E-01,-0.25932562E-01,-0.46754810E-02,
              0.16648775E+00,-0.20828387E-01,-0.15960755E-02,-0.84260628E-02,
             -0.61570007E+00,-0.76635070E-01, 0.38172156E-01,-0.21436466E-02,
              0.33362638E-01, },
            { 0.16782244E+00,-0.23373500E+00,-0.76538667E-01,-0.46066750E-01,
              0.44487558E-01,-0.98707631E-01,-0.62737954E-02, 0.47552589E+00,
              0.33932384E-01,-0.10383098E-01, 0.54402561E-02, 0.42593040E-01,
             -0.40717162E-01, },
            {-0.34468144E-01, 0.99032059E-01,-0.15680047E-01, 0.16594915E-01,
              0.60266782E-01,-0.96315397E-02,-0.12239870E-01,-0.44720203E-01,
              0.24779214E-01, 0.17988123E+00, 0.16323383E+00,-0.13639979E-01,
              0.24354503E-01, },
            { 0.13295417E+00, 0.36607604E-01, 0.11295780E-01,-0.85317560E-01,
              0.23435000E-01,-0.11025268E-01,-0.20789802E-01, 0.22085607E-01,
             -0.84365681E-01,-0.16373871E+00, 0.14517102E+00,-0.26270184E-02,
              0.10327833E-01, },
            { 0.39756168E-01,-0.21894038E-01, 0.32299142E-01,-0.25709143E-01,
              0.46975818E-01, 0.91380738E-02, 0.45137787E-02,-0.18453874E-01,
             -0.16742310E-01, 0.62881177E-02, 0.31858400E-01,-0.52852482E-01,
              0.64126670E-01, },
            { 0.70708215E-01, 0.26414242E-01,-0.95734000E-01,-0.51715512E-01,
             -0.61222430E-01, 0.32585312E-02,-0.72614186E-01, 0.29171487E-01,
             -0.30535296E-01,-0.62957483E-05, 0.48962277E-02,-0.69674492E-01,
             -0.63961051E-01, },
            { 0.24603065E-02, 0.32394640E-01,-0.59477240E-01,-0.60001020E-02,
             -0.37934547E-03, 0.28015230E-01,-0.36198471E-01, 0.15542215E-01,
             -0.82169725E-02,-0.18009149E-01,-0.15005712E-01,-0.68974853E-02,
             -0.44691041E-02, },
            { 0.16535794E-01, 0.56647848E-01,-0.21545598E-01,-0.41591849E-01,
             -0.52183713E-02,-0.25275357E-01, 0.10705411E-01, 0.35601035E-01,
              0.87656369E-02,-0.97603891E-02, 0.26778039E-02, 0.52772003E-03,
             -0.49273390E-02, },
            {-0.10546016E-02, 0.60169496E-01,-0.38740799E-01,-0.21469520E-01,
              0.29543184E-02,-0.40258944E-01,-0.13036972E-01, 0.18623425E-01,
              0.90010501E-02, 0.14114746E-01,-0.19131180E-01,-0.65653445E-03,
              0.10507869E-01, },
            { 0.22839565E-01,-0.56603742E-02, 0.63457802E-01, 0.13852383E-01,
              0.19432070E-01, 0.25225332E-01, 0.21424852E-01,-0.25062202E-02,
              0.82456255E-02,-0.16133005E-01, 0.12762873E-01, 0.22965312E-01,
              0.64126670E-01, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < F2MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < F2MaxOrder);
      return F2[cond][deg][ord];
   }


   double CCIR ::
   ccir22Fm3(int cond, int deg, int ord)
   {
      static const double Fm3[F2SolarActCond][FM3MaxDegree][FM3MaxOrder] {
         {
            { 0.30660009E+01, 0.84245279E-02, 0.10416747E-01,-0.79152789E-02,
              0.97902566E-02, 0.27440805E-02,-0.18010708E-02,-0.44805654E-02,
              0.44099679E-02, },
            { 0.47937575E+00,-0.73863333E-02, 0.62492259E-01, 0.71070671E-01,
              0.24661029E-01,-0.76728272E-02,-0.34576587E-01,-0.55801393E-02,
              0.24911568E-01, },
            { 0.72640646E+00,-0.13610825E+00,-0.92151836E-01, 0.57366926E-01,
             -0.79799727E-01, 0.15536819E-01, 0.29552156E-01,-0.23152215E-01,
             -0.71591372E-02, },
            {-0.52047455E+00,-0.18002425E+00,-0.14835778E+00,-0.15261011E+00,
             -0.84134936E-01, 0.74266516E-01, 0.12540767E+00, 0.69134543E-02,
             -0.10695435E+00, },
            {-0.78587282E+00, 0.42692187E+00, 0.14050616E+00,-0.10526546E+00,
              0.16430844E+00,-0.38810812E-01,-0.95510431E-01, 0.89601807E-01,
              0.10011222E-01, },
            {-0.18745262E-01, 0.15723652E+00, 0.10078381E+00, 0.68045869E-01,
              0.75673684E-01,-0.72365344E-01,-0.91716684E-01,-0.11222346E-02,
              0.89619219E-01, },
            { 0.83576031E-01,-0.30042040E+00,-0.12336441E-01, 0.59537537E-01,
             -0.84578089E-01, 0.17366186E-01, 0.79638004E-01,-0.66486605E-01,
             -0.10213455E-01, },
            {-0.24552050E-02,-0.10713033E+00,-0.34063554E+00, 0.15499518E-03,
              0.16007205E-01,-0.43798913E-02, 0.56936443E-02, 0.65028886E-02,
              0.10104384E-01, },
            { 0.11042719E-01, 0.37260795E+00,-0.10771591E+00, 0.56015593E-02,
             -0.42281896E-02,-0.25236963E-02,-0.11927215E-02,-0.10831781E-01,
              0.66218302E-02, },
            {-0.12299374E+00, 0.37923911E+00, 0.25219777E+00, 0.50730992E-01,
              0.70875049E-01,-0.12099678E-01, 0.36568884E-01, 0.57058934E-01,
             -0.18947596E-01, },
            {-0.35674047E-01,-0.31694239E+00, 0.45843518E+00,-0.12794103E+00,
             -0.69748282E-01, 0.42822126E-01, 0.30879781E-01,-0.51876172E-01,
              0.43667890E-01, },
            { 0.11940227E+00, 0.61209160E+00, 0.10119880E+01,-0.12578768E+00,
             -0.23280199E+00, 0.84285967E-01,-0.31325568E-01,-0.14911862E+00,
             -0.10552527E+00, },
            {-0.42771018E+00,-0.15437865E+01, 0.72490811E+00,-0.41361243E-01,
              0.11341509E+00, 0.58602658E-02,-0.10867080E-01, 0.23384985E+00,
             -0.18132639E+00, },
            { 0.14408733E+01,-0.15181103E+01,-0.70733821E+00, 0.22425801E+00,
             -0.20626904E+00,-0.23165996E+00,-0.24311604E+00,-0.75157189E+00,
              0.32163310E+00, },
            { 0.11034918E+01, 0.19018683E+01,-0.25283425E+01, 0.54272127E+00,
              0.81747144E+00,-0.33536407E+00,-0.28131485E+00, 0.44513670E+00,
             -0.38100833E+00, },
            {-0.49559730E+00,-0.76724958E+00, 0.64328122E+00, 0.37971288E+00,
              0.38863850E+00,-0.20415752E+00,-0.27450986E-01, 0.35379422E+00,
              0.24708341E+00, },
            { 0.15503995E+01, 0.70807004E+00,-0.15437074E+01, 0.22590764E+00,
             -0.22412081E+00, 0.70388354E-01, 0.28750276E-01,-0.59116602E+00,
              0.65819973E+00, },
            {-0.31314054E+01, 0.15656358E+01, 0.18508843E+01,-0.10490452E+01,
              0.13597289E+00, 0.96911991E+00, 0.34504092E+00, 0.18334630E+01,
             -0.94316882E+00, },
            {-0.32625735E+01,-0.48891945E+01, 0.43436418E+01,-0.89692688E+00,
             -0.22180891E+01, 0.81275958E+00, 0.57701594E+00,-0.92923886E+00,
              0.10023088E+01, },
            { 0.46660012E+00, 0.33790904E+00,-0.12450566E+01,-0.27982497E+00,
             -0.11594565E+00, 0.78048654E-01, 0.53353399E-01,-0.21656552E+00,
             -0.16291541E+00, },
            {-0.14640079E+01, 0.67885804E+00, 0.12005756E+01,-0.24355866E+00,
              0.63787431E-01,-0.11045193E+00,-0.53045399E-01, 0.37856793E+00,
             -0.54621232E+00, },
            { 0.18558358E+01,-0.15444054E+00,-0.95597863E+00, 0.83078742E+00,
              0.86885981E-01,-0.86581147E+00,-0.98496750E-01,-0.12377061E+01,
              0.71731114E+00, },
            { 0.23264565E+01, 0.32293904E+01,-0.22380531E+01, 0.50838184E+00,
              0.16552072E+01,-0.55498052E+00,-0.29978287E+00, 0.58354074E+00,
             -0.75011027E+00, },
            {-0.18672661E-02,-0.25456157E-01, 0.60223136E-02, 0.49720477E-01,
             -0.96478693E-01,-0.38978413E-02,-0.22143553E-03,-0.35477506E-02,
             -0.55410183E-03, },
            {-0.34474754E-02,-0.11412907E-01,-0.14703928E-01, 0.10629820E+00,
              0.38479723E-01,-0.11775076E-01, 0.24791460E-02,-0.13214642E-04,
              0.54156147E-02, },
            { 0.17747490E-01, 0.85842289E-01, 0.12783460E-01,-0.80395052E-02,
              0.82914688E-01, 0.52115154E-01,-0.59585259E-02,-0.34316640E-01,
             -0.31203479E-01, },
            {-0.15008713E+00, 0.13022913E-01,-0.13114069E+00,-0.98083436E-01,
              0.43264881E-01, 0.23754107E-01, 0.40249381E-01,-0.25870558E-01,
              0.10604724E-01, },
            {-0.11669767E+00, 0.18879850E+00, 0.85860342E-01,-0.80694489E-01,
             -0.14608112E+00,-0.38452525E-01, 0.11943401E+00, 0.36737777E-01,
             -0.83100669E-01, },
            {-0.98170832E-01,-0.25824536E-01,-0.13470170E-01, 0.16278113E+00,
              0.14445356E+00,-0.37913606E-01, 0.26959060E-01, 0.54383244E-01,
             -0.33156373E-01, },
            {-0.69396205E-01,-0.29353243E+00, 0.12254417E+00,-0.11734211E-01,
             -0.65991062E+00,-0.22865216E+00,-0.37296597E-01, 0.19098966E+00,
              0.12578192E+00, },
            { 0.10914133E+01, 0.47314721E+00, 0.71602601E+00, 0.67153996E+00,
             -0.39010674E+00, 0.86237691E-01,-0.38608781E+00, 0.10881895E+00,
             -0.13271818E-01, },
            { 0.24741711E+00,-0.29912946E+00,-0.27786815E+00, 0.58144867E-01,
              0.35581678E+00, 0.53345066E-01,-0.22732742E+00,-0.30918676E-01,
              0.11908955E+00, },
            { 0.87417528E-01,-0.19209103E-02, 0.12066233E+00,-0.37097234E+00,
             -0.29199532E+00, 0.99688828E-01,-0.63901879E-01,-0.95861524E-01,
              0.24224207E-01, },
            {-0.14814958E-01, 0.35126498E+00,-0.10779124E+00,-0.37479263E-01,
              0.10758766E+01, 0.33757868E+00, 0.12894204E+00,-0.21282887E+00,
             -0.11465584E+00, },
            {-0.14006575E+01,-0.77323854E+00,-0.76969266E+00,-0.10458283E+01,
              0.45393547E+00,-0.20995061E+00, 0.51989406E+00,-0.76540701E-01,
             -0.34076910E-01, },
            {-0.14458183E-01, 0.55377367E-02, 0.10118624E-01, 0.43726047E-02,
              0.30203222E-02, 0.26917916E-01,-0.27824063E-01, 0.38613998E-02,
             -0.23288135E-02, },
            { 0.97673871E-02,-0.12221950E-01,-0.20525286E-02,-0.16824726E-01,
              0.66913562E-04, 0.29891016E-01, 0.39633669E-01, 0.86441580E-02,
              0.57085617E-02, },
            { 0.60076401E-01, 0.44391785E-01, 0.75440123E-02,-0.47574192E-01,
             -0.25747169E-01,-0.13596806E+00,-0.25348425E-01,-0.39103729E-03,
             -0.79124048E-02, },
            { 0.47816981E-01,-0.12213377E-01,-0.60907356E-02,-0.16249452E-01,
             -0.81891660E-02, 0.20783218E-01,-0.12515883E+00,-0.63580880E-02,
             -0.93365340E-02, },
            { 0.35757348E-01,-0.11628059E+00, 0.23941850E-01, 0.39817769E-01,
              0.22874350E-01,-0.14632468E+00,-0.48552625E-01, 0.21140071E-01,
             -0.29413914E-01, },
            {-0.10383629E+00,-0.25948141E-01,-0.37469424E-01, 0.13439357E-01,
             -0.65478263E-02, 0.93140125E-01,-0.17407711E+00,-0.22653142E-01,
              0.20707196E-01, },
            {-0.12050121E-01,-0.76118793E-03, 0.14033116E-02, 0.10667692E-01,
              0.31927912E-02,-0.39311568E-02, 0.34947079E-02, 0.25789489E-02,
              0.31379892E-02, },
            {-0.14101000E-01, 0.78286082E-02,-0.55754250E-02,-0.79506766E-02,
              0.20916839E-02, 0.22414438E-02, 0.73174662E-02,-0.58244873E-03,
              0.37509066E-02, },
            { 0.89371942E-01, 0.39174829E-01,-0.15044765E-01,-0.11416791E-02,
             -0.28407650E-01,-0.12282236E-01, 0.21022290E-02, 0.14036150E-01,
             -0.10413602E+00, },
            {-0.34115105E-02,-0.33254713E-01, 0.22143330E-01, 0.35554647E-01,
             -0.46583945E-02, 0.82491385E-02, 0.43415073E-02, 0.11600922E+00,
              0.77101225E-02, },
            {-0.15329777E-02,-0.16291803E-02,-0.26436499E-02,-0.49016397E-02,
              0.12201109E-02, 0.16907734E-02, 0.32629367E-03,-0.15686175E-01,
              0.86215120E-02, },
            { 0.87878658E-02, 0.13618468E-01,-0.13227126E-01,-0.82996050E-02,
             -0.15946573E-01, 0.71573956E-02, 0.56405202E-02,-0.54664928E-02,
             -0.11713528E-01, },
            {-0.12917928E-02, 0.49148039E-02, 0.13711536E-01, 0.31793679E-02,
              0.66184965E-02, 0.20440640E-02,-0.92955929E-03,-0.69237831E-02,
             -0.55623794E-03, },
            { 0.28278935E-03,-0.72326208E-02,-0.10317585E-01, 0.10494902E-03,
             -0.11032540E-01, 0.24561610E-02, 0.84329043E-02,-0.13802494E-02,
             -0.10716310E-01, },
         },
         {
            { 0.26969762E+01, 0.32681650E-02, 0.43882988E-02,-0.27695221E-02,
              0.10836555E-01,-0.36222459E-03,-0.40186797E-02,-0.25219587E-02,
             -0.58713165E-04, },
            { 0.43824396E+00, 0.19661849E-02, 0.37536167E-01, 0.40424865E-01,
              0.34500234E-01,-0.12376064E-01,-0.37711073E-01,-0.25790085E-02,
              0.21678755E-01, },
            { 0.90846723E+00,-0.32327816E-01,-0.90745442E-01, 0.30759597E-01,
             -0.99251971E-01, 0.25366189E-01, 0.51679712E-01,-0.10518884E-01,
              0.47520380E-01, },
            {-0.36741725E+00,-0.22690707E+00,-0.80392964E-01,-0.10627037E+00,
             -0.96276939E-01, 0.56504987E-01, 0.13875304E+00,-0.96518273E-03,
             -0.10057620E+00, },
            {-0.98529118E+00, 0.91059320E-01, 0.14998367E+00,-0.79339772E-01,
              0.17853460E+00,-0.37562929E-01,-0.12783723E+00, 0.34081671E-01,
             -0.13274518E+00, },
            {-0.14129069E-01, 0.19939514E+00, 0.66745520E-01, 0.63963860E-01,
              0.75332023E-01,-0.48900589E-01,-0.10383722E+00, 0.38989398E-02,
              0.86292423E-01, },
            { 0.10910988E+00,-0.51262561E-01,-0.39895140E-01, 0.59805356E-01,
             -0.85471421E-01, 0.12946113E-01, 0.86364143E-01,-0.23241360E-01,
              0.85489251E-01, },
            { 0.68126281E-03,-0.26906112E+00,-0.25801969E+00,-0.36406654E-03,
              0.15821831E-01,-0.58499649E-02, 0.49455757E-02, 0.19876708E-02,
              0.70102816E-02, },
            { 0.10385459E-01, 0.28189835E+00,-0.27379474E+00, 0.99617587E-02,
             -0.94898148E-02, 0.23341458E-02,-0.17798019E-02,-0.35258995E-02,
              0.95521472E-02, },
            {-0.59638537E-01, 0.22104724E+00, 0.11115756E+00, 0.36462764E-02,
              0.68865590E-01,-0.22449013E-01, 0.57117860E-02, 0.45325380E-01,
              0.41912273E-02, },
            {-0.27711060E-01,-0.10434671E+00, 0.26874310E+00,-0.27662922E-01,
             -0.48578799E-01, 0.34111135E-01, 0.36623143E-01,-0.33080466E-02,
              0.34407139E-01, },
            {-0.86527253E-02, 0.10166690E+01, 0.33211759E+00,-0.69901399E-01,
             -0.18758866E+00, 0.96067429E-01,-0.20856796E-01,-0.89351051E-01,
             -0.11115274E+00, },
            {-0.27616104E+00,-0.77658612E+00, 0.11817397E+01,-0.40753499E-01,
              0.15753447E+00,-0.38351923E-01, 0.15354547E-01, 0.11337685E+00,
             -0.17565700E+00, },
            { 0.70202410E+00,-0.12175164E+01,-0.54803377E+00, 0.31004715E+00,
             -0.24121673E+00, 0.54169376E-02,-0.16837779E+00,-0.57409120E+00,
             -0.40699039E-01, },
            { 0.73859042E+00, 0.11418145E+01,-0.18006949E+01,-0.92229664E-01,
              0.44709337E+00,-0.23921192E+00,-0.38646665E+00, 0.41334989E-03,
             -0.26683557E+00, },
            {-0.14896899E+00,-0.11365073E+01, 0.14847133E+01, 0.33789396E+00,
              0.27518693E+00,-0.24134621E+00,-0.20836785E-01, 0.24304211E+00,
              0.32803184E+00, },
            { 0.98338109E+00,-0.47739923E+00,-0.19460961E+01, 0.17087240E-01,
             -0.25491282E+00, 0.74909970E-01,-0.80111466E-01,-0.28729579E+00,
              0.55402631E+00, },
            {-0.14751050E+01, 0.17943174E+01, 0.17780763E+01,-0.91258276E+00,
              0.23318025E+00, 0.18268088E+00, 0.40549916E+00, 0.14381841E+01,
              0.42043656E-01, },
            {-0.24171760E+01,-0.34031210E+01, 0.33957868E+01, 0.46847907E+00,
             -0.11220565E+01, 0.62169796E+00, 0.88092864E+00, 0.88447511E-01,
              0.57009745E+00, },
            { 0.19369581E+00, 0.60008752E+00,-0.14411573E+01,-0.28396797E+00,
             -0.81249483E-01, 0.11603205E+00, 0.37423864E-01,-0.16771762E+00,
             -0.25206971E+00, },
            {-0.87629116E+00, 0.11575415E+01, 0.13158292E+01, 0.61591063E-01,
              0.67997426E-01,-0.37556443E-01, 0.51347751E-01, 0.18010110E+00,
             -0.45261535E+00, },
            { 0.80545521E+00,-0.58170259E+00,-0.83205831E+00, 0.67988551E+00,
             -0.48437756E-02,-0.24384318E+00,-0.24711609E+00,-0.99533564E+00,
             -0.74051530E-03, },
            { 0.19566715E+01, 0.21233406E+01,-0.18595179E+01,-0.41028827E+00,
              0.87477672E+00,-0.47322938E+00,-0.53535795E+00,-0.10144442E+00,
             -0.34023517E+00, },
            { 0.19387048E-02,-0.24992781E-01, 0.41351970E-02, 0.82355030E-01,
             -0.10060746E+00, 0.10565569E-02, 0.61386190E-02,-0.56847963E-02,
              0.28516669E-03, },
            { 0.18765301E-02,-0.18433956E-01,-0.57170093E-02, 0.96418172E-01,
              0.74403822E-01,-0.49711904E-02, 0.73961457E-02,-0.26749184E-02,
              0.35287787E-02, },
            { 0.44248577E-01, 0.44711590E-01, 0.35021894E-01,-0.10435295E+00,
              0.12232722E+00, 0.40250495E-01, 0.19099291E-01, 0.10738099E-02,
             -0.69058025E-02, },
            {-0.51917080E-01,-0.27200999E-01,-0.10501927E+00,-0.16478807E+00,
             -0.83345883E-01,-0.19203559E-02, 0.45351192E-01,-0.21076348E-01,
              0.86140940E-02, },
            {-0.10319305E+00, 0.13298593E+00, 0.73425964E-01,-0.46246278E+00,
              0.13967551E+00,-0.45703530E-01, 0.76875174E-02, 0.67371532E-01,
             -0.46121821E-01, },
            {-0.71389146E-01,-0.35766073E-01,-0.66930741E-01,-0.13037997E+00,
             -0.29186937E+00,-0.43604024E-01,-0.38124863E-01, 0.46981771E-01,
             -0.33583380E-01, },
            {-0.22479945E+00,-0.17325100E+00,-0.42389341E-01, 0.63497208E-01,
             -0.65707499E+00,-0.18000965E+00,-0.13928051E+00, 0.92657882E-03,
              0.32833587E-01, },
            { 0.35454518E+00, 0.56296510E+00, 0.54718381E+00, 0.87280089E+00,
             -0.95135480E-01, 0.18839495E+00,-0.28471857E+00, 0.99687912E-01,
             -0.74194811E-01, },
            { 0.22021563E+00,-0.19814751E+00,-0.26210853E+00, 0.38916111E+00,
             -0.11701600E+00, 0.43736532E-01,-0.51089481E-01,-0.89396305E-01,
              0.56751575E-01, },
            { 0.84288068E-01, 0.46980049E-01, 0.16849743E+00, 0.14809281E+00,
              0.15139215E+00, 0.80832951E-01, 0.28207000E-01,-0.66889457E-01,
              0.30166352E-01, },
            { 0.12673169E+00, 0.26919132E+00, 0.78022078E-01, 0.44204059E-03,
              0.88714296E+00, 0.26615989E+00, 0.19613440E+00,-0.19267652E-01,
             -0.56138389E-01, },
            {-0.44904345E+00,-0.76422292E+00,-0.58892214E+00,-0.11297759E+01,
              0.21760568E+00,-0.28634107E+00, 0.33985862E+00,-0.90833008E-01,
              0.72062925E-01, },
            {-0.51934798E-02, 0.74422540E-03, 0.79966486E-02, 0.72337938E-02,
             -0.33158786E-03, 0.32955918E-01,-0.36173228E-01, 0.37724695E-02,
             -0.41374913E-03, },
            {-0.21712191E-02,-0.42578488E-03, 0.46815402E-02,-0.11746973E-01,
              0.20427755E-02, 0.37616257E-01, 0.33100210E-01, 0.46383860E-02,
              0.91368466E-03, },
            { 0.12895065E-01, 0.50809748E-01, 0.29385984E-02,-0.18984521E-01,
             -0.27869942E-01,-0.12483256E+00,-0.34310818E-01,-0.27956211E-02,
             -0.34609209E-04, },
            { 0.11988606E-01,-0.21017001E-02, 0.25302023E-02,-0.96779130E-02,
             -0.14333853E-01, 0.32475434E-01,-0.11795449E+00,-0.32607217E-02,
             -0.88285496E-02, },
            { 0.10866594E-01,-0.90187959E-01, 0.33737600E-01, 0.10237814E-01,
              0.22867156E-01,-0.17877857E+00,-0.42381451E-01, 0.12767127E-01,
             -0.93460009E-02, },
            {-0.30316098E-01,-0.55458799E-01,-0.43160249E-01, 0.22116292E-01,
             -0.88928174E-02, 0.67580663E-01,-0.13444774E+00,-0.15160068E-01,
              0.16901061E-01, },
            {-0.74068061E-02,-0.93656145E-02, 0.16188228E-02, 0.28720102E-02,
              0.59500979E-02,-0.38546906E-02, 0.22760713E-02, 0.99050291E-02,
              0.50792447E-02, },
            {-0.11978722E-01, 0.60989028E-02,-0.10759148E-01,-0.51578460E-02,
             -0.61472147E-02,-0.13215247E-02, 0.77774036E-02,-0.41620326E-02,
              0.12450544E-01, },
            { 0.33062402E-01, 0.28125867E-01,-0.74927374E-02, 0.32324712E-02,
             -0.23429561E-01,-0.13820422E-02, 0.71330271E-02, 0.28685573E-01,
             -0.87745167E-01, },
            { 0.55558784E-02,-0.37983757E-01, 0.23901636E-01, 0.20743832E-01,
             -0.81198774E-02,-0.32056179E-02, 0.81567001E-02, 0.96058153E-01,
              0.33794027E-01, },
            { 0.55779296E-03,-0.44976049E-02,-0.40536787E-03,-0.57768053E-02,
              0.36962274E-02, 0.12729283E-02,-0.32060328E-02,-0.11466083E-01,
              0.39026923E-02, },
            { 0.30240275E-02, 0.14327827E-01,-0.91552837E-02,-0.15580239E-02,
             -0.14593671E-01, 0.58387034E-02, 0.71950480E-02,-0.30479324E-02,
             -0.67118248E-02, },
            {-0.39168904E-02, 0.10714657E-01, 0.97695896E-02, 0.44566900E-02,
              0.11315702E-02,-0.24044179E-03,-0.57686231E-03,-0.59982366E-02,
             -0.22883762E-02, },
            { 0.37045740E-02,-0.60077994E-02,-0.99031944E-02, 0.84312021E-03,
             -0.11128205E-01, 0.24483956E-02, 0.49151871E-02, 0.20255843E-03,
             -0.45773280E-02, },
         },
      };
      GNSSTK_ASSERT(cond >= 0 && cond < F2SolarActCond);
      GNSSTK_ASSERT(deg >= 0 && deg < FM3MaxDegree);
      GNSSTK_ASSERT(ord >= 0 && ord < FM3MaxOrder);
      return Fm3[cond][deg][ord];
   }
}
